# liza-crypto
This library provides cryptographic operations for Liza, intended for
use both by Liza itself and by downstream systems that consume data from
Liza.

The version number of this library mirrors that of the parent Liza
project; version changes do not necessarily mean that changes to liza-crypto
have been made.

## When Do I Use This Library?
If your system received data that starts with "Encrypted:" followed by what
appears to be random data, then you _must_ use this library to decrypt it.
_Do not_ attempt to decrypt it on your own without use of liza-crypto.

If you are working on Liza and need to perform some cryptographic operation,
this library _must_ be used.

If you need to encrypt data, and your data is not sourced from or used in
conjunction with data sourced from Liza, then you _must not_ use this
library; the design decisions made for this library are specific to the
backwards-compatibility requirements of Liza's bucket format and do not make
sense for all systems, nor will this library commit to the support of other
systems.

## Session Keys and Key Encryption Keys (KEKs)
While efforts have been made to present an API that is relatively easy to
use and encapsulates most of the cryptographic details, correct use of the
API does require understanding a couple concepts.  Within the context of
this library:

<!--
  Note: this deliberately omits details about the keys that are not
  necessary for use of the API.  It is important to keep this section
  simple: encourage proper use of the library, but do not distract from it.
-->

  1. A _session key_ is a randomly generated key used to encrypt bucket
     data for a single quote.  A new session key is generated for each
     quote, and session keys are not to be re-used across quotes.
  2. A _key encryption key_ (KEK) is a key that is used to encrypt session
     keys.  There should be one KEK for each downstream system.  KEKs are
     re-used across many quotes.

Session keys must _never_ be stored plaintext—they must always encrypted
with a KEK.  Exposing a session key will compromise encrypted data for the
entire quote.  Exposing a KEK compromises all quotes with session keys
encrypted to that KEK.

If a KEK does not yet exist for your system, _do not generate one
yourself_, as it must be generated and stored securely; consult your
organization's documentation for information on how to invoke a secure
process for key generation and storage.  The key must be 256 bits in length.

## Usage In Downstream Systems
A _downstream system_ is any system that receives data that was sourced by
Liza.  These data are usually referred to as "bucket data", though they may
have gone through one or more layers of mapping before they reach your
service.

When integrating this library into downstream system, you can expect that:

  1. Encrypted session keys are sent to the downstream system as an array
     of strings, alongside bucket data; and
  2. The KEK for the downstream system is loaded from a trusted key store.
     This KEK is the same for every quote.

_You should never attempt to decrypt session keys yourself;_ this risks
leaking the key to other parts of the system, or to other systems entirely.

If the KEK is encoded (e.g. base64), you must decode it into a binary
`Buffer` before use.

### Decryption
Data that has been encrypted using `liza-crypto` is able to be decrypted
with the KEK of any service that is an intended recipient of encrypted data.

```js
import {decryptorFromEncryptedSessionKeys} from 'liza-crypto';

// Attempt to create a decryptor from the first available session key.
// At least one session key must be able to be decrypted using `kek`.
const decryptor = decryptorFromEncryptedSessionKeys(
  kek,                     // Buffer
  encrypted_session_keys,  // string[]
);

// The decryptor can decrypt any `ciphertext` that has been encrypted using
// the session key.
const decrypted = decryptor.decrypt(ciphertext);

// Retrieving the plaintext requires invoking `getDecryptedText` to
// emphasize the security risks of using such data.
const plaintext = decrypted.getDecryptedText();
```

_Always_ use `decryptorFromEncryptedSessionKeys` to create a `Decryptor`
instance.  It is important that you treat the array of session keys as a black
box—never attempt to select one of the values in the array yourself, even if
it has only a single element.  This array is used to support key rotation
of KEKs, and failure to present the entire array to `liza-crypto` will cause
decryption failures during key rotation periods.


### Encryption
First, be sure that it is appropriate to encrypt data using this system:

  1. If an already-encrypted datum does not require any modification and
     needs to be propagated to another system, simply propagate the existing
     value.
  2. If you need to modify an encrypted value, you may decrypt it and
     encrypt the modified value using this library.
  3. If you need to need augment (or "enrich") quote data, and the
     additional data ought to be confidential, and those data ought to be
     considered part of the quote as if collected by Liza, then you may
     encrypt those data with this library.
  4. If your system does not receive data sourced from Liza, or if you are
     publishing data to a system that does not receive data sourced from
     Liza, you _must not_ utilize this library.  See [When Do I Use This
     Library](#when-do-i-use-this-library) above.

```js
import {encryptorFromEncryptedSessionKeys} from 'liza-crypto';

// Attempt to create an encryptor from the first available session key.
// At least one session key must be able to be decrypted using `kek`.
const encryptor = encryptorFromEncryptedSessionKeys(
  kek,                     // Buffer
  encrypted_session_keys,  // string[]
);

// The encryptor will encrypt the provided plaintext using the session key
// discovered above.
const ciphertext = encryptor.encrypt(plaintext);
```

_Always_ use `encryptorFromEncryptedSessionKeys` to create an `Encryptor`
instance.  It is important that you treat the array of session keys as a black
box—never attempt to select one of the values in the array yourself, even if
it has only a single element.  This array is used to support key rotation
of KEKs, and failure to present the entire array to `liza-crypto` will cause
session key decryption failures during key rotation periods.

## Local Development
`liza-crypto` is automatically set up as part of the parent project.
If you wish to work on `liza-crypto` in isolation, simply run `bootstrap`:

```sh
# Set up environment after first clone
./bootstrap
```

To build and run the test suite, run `make check`.

To build API documentation, run `make html`.

## Debugging
`liza-crypto` provides a dev-tools file that can be utilized to quickly
encrypt and decrypt data using the librarys provided features. To access this
run:

```sh
make tsc #Builds any changes that might be there
node ./bin/dev-tools.js
```

This will open up a JS REPL, typing `.help` will provide you with a list
of functions to choose from.

### Modifying `package.json`
`package.json` is generated from `package.json.in`; you should update the
latter instead, and then either re-bootstrap or run `./config.status` to
re-generate it (or re-run `bootstrap`).

## Purpose Of This Library
This library exists for two fundamental reasons:

  1. To straddle both the producer and consumer of encrypted data; and
  2. To provide small, easily-auditable module for sensitive cryptographic
     code.

Since this library is used both by Liza (the producer) and downstream
systems (the consumers), Liza is both free to vary its implementation and
fully encapsulate those implementation details; this fits with the library's
philosophy of providing no configurable options, but it does force coupling
of downstream systems with Liza.  However, where this coupling is not
appropriate (where a downstream system shouldn't be considered to be part of
a single coherent system), a downstream system can decrypt data and choose
to utilize its own format.
