import {describe, it} from 'mocha';
import {BinaryLike, createCipheriv, randomBytes} from 'crypto';
import {Decryptor, Encryptor} from '.';
import fc from 'fast-check';
import {expect, assert} from 'chai';
import {
  DecryptionError,
  InvalidPrefixError,
  KeyLengthError,
  MessageLengthError,
} from './types';
import {keyBits} from './config';

const mockEncryptor = (encoding: 'hex' | 'binary') => {
  return (sessionKey: BinaryLike) => (plaintext: string) => {
    const iv = randomBytes(16);
    const cipher = createCipheriv('aes-256-cbc', sessionKey, iv);
    return cipher.update(plaintext, 'utf8', encoding) + cipher.final(encoding);
  };
};

// Generates a test that looks for a specific instance of an error
// eslint-disable-next-line @typescript-eslint/no-explicit-any
const testErrorClass = (e: any, errorClass: any) =>
  assert(
    e instanceof errorClass,
    `Expected ${errorClass.name} to be thrown, got ${e.name} instead`
  );

describe('Decryptor', () => {
  const sessionKey = randomBytes(keyBits / 8);

  it('fails given a key of an incorrect length', () => {
    const tooShort = randomBytes(3);
    const tooLong = randomBytes(keyBits / 8 + 1);

    expect(() => new Decryptor(tooShort)).to.throw(
      KeyLengthError,
      tooShort.length * 8 + '-bit'
    );

    expect(() => new Decryptor(tooLong)).to.throw(
      KeyLengthError,
      tooLong.length * 8 + '-bit'
    );
  });

  it('should decrypt text given a key', () => {
    const textEncryptor = new Encryptor(sessionKey);

    try {
      const decryptor = new Decryptor(sessionKey);
      assert(true, 'Decryptor successfully initialized');

      fc.assert(
        fc.property(fc.string(), actual => {
          const ciphertext = textEncryptor.encrypt(actual);
          const plaintext = decryptor.decrypt(ciphertext);
          expect(plaintext).to.not.be.null;
          expect(plaintext?.getDecryptedText()).to.equal(actual);
        })
      );
    } catch (e) {
      assert(false, `Decryption error: ${e}`);
    }
  });

  it('should error if decryption fails', () => {
    const testEncoding = mockEncryptor('hex')(sessionKey);

    try {
      const decryptor = new Decryptor(sessionKey);
      assert(true, 'Decryptor successfully initialized');

      fc.assert(
        fc.property(fc.string(), actual => {
          const ciphertext = testEncoding(actual);

          try {
            decryptor.decrypt(`Encrypted:${ciphertext}`);
            assert(false, 'Expected ciphertext decryption to fail');
          } catch (e) {
            assert(true, 'Cipher text failed to decrypt');
            testErrorClass(e, DecryptionError);
          }
        })
      );
    } catch (e) {
      assert(false, `Decryption error: ${e}`);
    }
  });

  it('throws InvalidPrefix Error when text is incorrectly created', () => {
    const testEncoding = mockEncryptor('binary')(sessionKey);

    try {
      const decryptor = new Decryptor(sessionKey);
      assert(true, 'Decryptor successfully initialized');

      fc.assert(
        fc.property(fc.string(), actual => {
          const ciphertext = testEncoding(actual);

          try {
            decryptor.decrypt(ciphertext);
            assert(false, 'Expected ciphertext decryption to fail');
          } catch (e) {
            assert(true, 'Cipher text failed to decrypt');
            testErrorClass(e, InvalidPrefixError);
          }
        })
      );
    } catch (e) {
      assert(false, `Decryption error: ${e}`);
    }
  });

  it('throws MessageLengthError on incorrect message size', () => {
    try {
      const decryptor = new Decryptor(sessionKey);
      assert(true, 'Decryptor successfully initialized');

      fc.assert(
        fc.property(fc.string({maxLength: 16}), actual => {
          try {
            decryptor.decrypt(`Encrypted:${actual}`);
            assert(false, 'Expected ciphertext decryption to fail');
          } catch (e) {
            assert(true, 'Cipher text failed to decrypt');
            testErrorClass(e, MessageLengthError);
          }
        })
      );
    } catch (e) {
      assert(false, `Decryption error: ${e}`);
    }
  });
});
