/**
 * Opinionated decryption primitive
 *
 * This is the dual of {@link encrypt.ts}.
 *
 * @module
 */

import {BinaryLike, createDecipheriv} from 'crypto';
import {
  Key,
  ProtectedString,
  InvalidPrefixError,
  DecryptionError,
  MessageLengthError,
  assertKeyLength,
} from './types';
import {Buffer} from 'buffer';
import {
  algorithm,
  plaintextMagicBytes,
  ciphertextPrefix,
  blockSizeBytes,
} from './config';

/** Base64-encoded string. */
type Base64 = string;

/**
 * Symmetric decryption primitive accepting a ciphertext previously
 * encrypted with an {@link Encryptor}.
 *
 * This system may dynamically decrypt ciphertexts of previous formats, but
 * there is no guarantee that a ciphertext that is not in the latest format
 * will be able to be decrypted indefinitely.  It is important not to make
 * assumptions about the format or encoding of the ciphertext; let this
 * system handle that for you.
 */
export class Decryptor {
  constructor(private readonly key: Key) {
    assertKeyLength(key);
  }

  /**
   * Decrypt `ciphertext` previously encrypted using an {@link Encryptor}.
   *
   * Decryption will fail if the `ciphertext` is not in the format returned
   * by a previous {@link Encryptor.encrypt} operation.  Note that a
   * human-readable "Encrypted" prefix is part of the ciphertext and must be
   * provided.
   */
  public decrypt(ciphertext: string): ProtectedString {
    const {iv, message} = this.splitBuffer(
      this.afterPrefix(ciphertext, ciphertextPrefix)
    );
    try {
      const decipher = createDecipheriv(algorithm, <BinaryLike>this.key, iv);
      const decryptedText =
        decipher.update(message, undefined, 'utf8') + decipher.final('utf8');

      return new ProtectedString(
        this.afterPrefix(decryptedText, plaintextMagicBytes)
      );
    } catch (e) {
      // Invalid decryption
      throw new DecryptionError();
    }
  }

  // Helper function to allocate buffers for messages
  private splitBuffer(ciphertext: Base64): {iv: Buffer; message: Buffer} {
    const buff = Buffer.from(ciphertext, 'base64');
    if (buff.length < blockSizeBytes) {
      throw new MessageLengthError();
    }

    return {
      iv: buff.slice(0, blockSizeBytes),
      message: buff.slice(blockSizeBytes),
    };
  }

  // Helper function to split a string to after its decryption prefix
  private afterPrefix(s: string, prefix: string): string {
    if (!s.startsWith(prefix)) {
      throw new InvalidPrefixError();
    }
    return s.substring(prefix.length);
  }
}
