/**
 * Opinionated encryption primitive
 *
 * This is the dual of {@link decrypt}.
 *
 * @module
 */

import {createCipheriv, randomBytes} from 'crypto';
import {
  algorithm,
  ciphertextPrefix,
  plaintextMagicBytes,
  blockSizeBytes,
} from './config';
import {assertKeyLength, Key} from './types';

/**
 * Symmetric encryption primitive producing base64-encoded ciphertext.
 *
 * The only configurable value for this system is the key; all other
 * cipher parameters are hard-coded.
 *
 * While decryption supports older ciphertext formats during a migration
 * period, {@link Encryptor} has only one mode of operation representing the
 * latest version of Liza's encryption standard.  It is not possible to, nor
 * is it desirable to, encrypt using earlier formats.
 *
 * For more information on the chosen parameters, see {@link algorithm}.
 */
export class Encryptor {
  constructor(private readonly key: Key) {
    assertKeyLength(key);
  }

  /**
   * Encrypt the provided `plaintext`, producing a base64-encoded ciphertext.
   *
   * _Do not_ provide binary data to this function.  If you intend to
   * encrypt a key, first choose a string representation (e.g. base64).
   */
  public encrypt(plaintext: string): string {
    const iv = randomBytes(blockSizeBytes);
    const cipher = createCipheriv(algorithm, this.key, iv);

    const enc =
      cipher.update(plaintextMagicBytes + plaintext, 'utf8', 'binary') +
      cipher.final('binary');

    const buff = Buffer.alloc(enc.length + iv.length);
    iv.copy(buff);
    buff.write(enc, iv.length, 'binary');

    return `${ciphertextPrefix}${buff.toString('base64')}`;
  }
}
