import {expect} from 'chai';

import {randomBytes} from 'crypto';
import {
  encryptSessionKey,
  decryptorFromEncryptedSessionKeys,
  encryptorFromEncryptedSessionKeys,
  createSessionKey,
} from './session';
import {Encryptor} from './encrypt';
import {SessionKeyDecryptionError} from './types';
import {Decryptor} from './decrypt';

const kek = randomBytes(32);

describe('createEncryptedSessionKey', () => {
  // This is nothing more than a sanity check to make sure we haven't
  // accidentally hard-coded values; it does not serve as an alternative to
  // just looking at the definition of `createEncryptedSessionKey`—which was
  // intentionally kept very simple—to verify that it is defined as
  // intended.
  //
  // Usually, nondeterminism ought to be eliminated by abstracting over it,
  // but it was an explicit decision to couple the CSPRNG with the SUT so
  // that callers cannot mess up its configuration.  Further, there's not
  // much SUT left to test without it.
  it('should return a unique value each time', () => {
    // The chosen sample size is arbitrary; there is no use making it larger
    // since the purpose of this test isn't to test whether the SUT utilizes
    // a proper CSPRNG.  To determine that, read the implementation of
    // `createEncryptedSessionKey`.
    const sample_size = 10;

    const unique_samples = new Set();

    for (let i = 0; i < sample_size; i++) {
      unique_samples.add(encryptSessionKey(kek, createSessionKey()));
    }

    // Duplicates will result in a set length < n.
    expect(unique_samples.size).to.equal(sample_size);
  });

  // This is an attempt to sensure that our key generation is doing
  // something same.
  it('produces at least a 256-bit value', () => {
    const expected_bits = 256;
    const expected_bytes = expected_bits / 8;

    const result = encryptSessionKey(kek, createSessionKey());

    // Encrypting adds additional information (increases the length); this
    // is just a sanity check
    expect(result.length).to.be.greaterThan(expected_bytes);
  });
});

describe('decryptorFromSessionKeys', () => {
  const dummyEncSessionKey = encryptSessionKey(kek, createSessionKey());
  const dummyEncSessionKeys = [dummyEncSessionKey];

  it('should initialize with a valid session key', () => {
    // These types are intentional; they assert that the properties
    // that are tested with these objects hold.
    const enc: Encryptor = encryptorFromEncryptedSessionKeys(
      kek,
      dummyEncSessionKeys
    );
    const dec: Decryptor = decryptorFromEncryptedSessionKeys(
      kek,
      dummyEncSessionKeys
    );

    const plaintext = 'foo';
    const result = dec.decrypt(enc.encrypt(plaintext)).getDecryptedText();

    expect(result).to.equal(plaintext);
  });

  it('should error if no session keys are provided', () => {
    expect(() => decryptorFromEncryptedSessionKeys(kek, [])).to.throw(
      SessionKeyDecryptionError
    );
  });

  it('should error if no valid session keys are found', () => {
    expect(() => decryptorFromEncryptedSessionKeys(kek, ['badkey'])).to.throw(
      SessionKeyDecryptionError
    );
  });
});
