/**
 * Session key layer
 *
 * This module builds atop of the cryptographic primitives in
 * {@link encrypt} and {@link decrypt} to provide an abstraction over the
 * concept of session keys, and in doing so introduces the concept of a key
 * encryption key (KEK).  Specifically, this module provides:
 *
 *   1. A means for instantiating an {@link Encryptor} and {@link Decryptor}
 *      from a KEK and an array of encrypted session keys; and
 *   2. A means for generating and encrypting session keys.
 *
 * Encrypted session keys are accepted as an array to support key rotation
 * of KEKs; the first session key that is able to be decrypted is used to
 * instantiate an {@link Encryptor}/{@link Decryptor}, and others are ignored.
 *
 * Let _kᵢ_ be some KEK used to encrypt session key _s_, and let _s'ᵢ_ represent
 * the encrypted session key.  The following diagram illustrates the
 * contents of the encrypted session key array _S'_ of length _n_, where
 * _0≤i<n_ and _n≥1_.
 *
 * ```text
 *            ____________ s ________________
 *           |             |                 |
 *           v             v                 v
 *         (Enc) <- k₀   (Enc) <- k₁   …   (Enc) <- kₙ
 *           |             |                 |
 *           v             v                 v
 *  S' = [  s'₀           s'₁          …    s'ₙ         ]
 * ```
 *
 * Session keys are encrypted using the same cipher and parameters as all
 * other data handled by this library.  For more information, see the
 * {@link config} module.
 *
 * Session Key Concept
 * -------------------
 * A _session key_ is a randomly generated symmetric key that is used to
 * encrypt and decrypt field data.  Every intended recipient of encrypted
 * data holds a KEK, and the session key is encrypted to each KEK.  All
 * encrypted session keys are published alongside the ciphertext for field
 * data, and downstream recipients use their respective KEKs to decrypt the
 * data.
 *
 * To support rotation of KEKs, each recipient has an _array_ of encrypted
 * session keys, as shown above.  When a KEK needs to be rotated for a
 * downstream system, the session key is first decrypted using the original
 * KEK and then re-encrypted using the new KEK.  Both the old and the new
 * encrypted session key are stored in the array.
 *
 * TODO: Document the protocol and provide further rationale once key
 * rotation is introduced to this system.
 *
 * @module
 */

import {randomBytes} from 'crypto';
import {keyBytes} from './config';
import {Decryptor} from './decrypt';
import {Encryptor} from './encrypt';
import {Key, ProtectedString, SessionKeyDecryptionError} from './types';

type Kek = Buffer & Key;
type SessionKey = Buffer & Key;
export type EncryptedSessionKey = string;
export type EncodedSessionKey = string;

/** Base64-encode the provided binary session key. */
const encodeSessionKey = (skey: SessionKey) => skey.toString('base64');

/**
 * Given a decrypted, base64-encoded session key, return a binary
 * `SessionKey`.
 */
const decodeSessionKey = (s: EncodedSessionKey): SessionKey =>
  Buffer.from(s, 'base64');

const decryptSessionKeyWith =
  (dec: Decryptor) =>
  (skey: EncryptedSessionKey): SessionKey =>
    decodeSessionKey(dec.decrypt(skey).getDecryptedText());

/**
 * Generate a new base64-encoded random 256-bit session key.
 *
 * It is an explicit design decision to hardcode the CSPRNG—the security of
 * this system rests in the session key, so it is important that there is no
 * opportunity for misconfiguration.
 */
export const createSessionKey = (): ProtectedString =>
  new ProtectedString(encodeSessionKey(randomBytes(keyBytes)));

/** Encrypt a session key to a key encryption key (KEK). */
export const encryptSessionKey = (
  kek: Kek,
  key: ProtectedString
): EncryptedSessionKey => {
  const enc = new Encryptor(kek);
  return enc.encrypt(key.getDecryptedText());
};

/** Decrypt a session key with a key encrypting key (KEK). */
export const decryptSessionKey = (
  kek: Kek,
  key: EncodedSessionKey
): ProtectedString => {
  const dec = new Decryptor(kek);
  return dec.decrypt(key);
};

/**
 * Ignore argument and do nothing.
 *
 * This provides a self-documenting way to do nothing with a value, leaving
 * no question was to whether this was intended.  It is code, not
 * documentation, so it cannot get out-of-sync with the implementation.
 */
const ignoreError = (_e: unknown): void => {};

const fromEncryptedSessionKeys =
  <T>(ctor: (session_key: Key) => T) =>
  (kek: Key, encrypted_session_keys: EncryptedSessionKey[]): T => {
    const decryptSessionKey = decryptSessionKeyWith(new Decryptor(kek));

    for (const enc_skey of encrypted_session_keys) {
      try {
        // return on first match to avoid unnecessary processing
        return ctor(decryptSessionKey(enc_skey));
      } catch (e) {
        ignoreError(e);
      }
    }

    throw new SessionKeyDecryptionError();
  };

/**
 * Create a new `Decryptor` from an available session key.
 *
 * From an array of encrypted session keys, instantiate a `Decryptor` using
 * the first session key that is able to be decrypted using the
 * provided `kek`.  If no session key can be decrypted, throw a
 * `SessionKeyDecryptionError`.
 *
 * The array of session keys allows for KEK rotation; it is important that
 * `Decryptor` be instantiated using this function rather than constructing
 * it directly.
 */
export const decryptorFromEncryptedSessionKeys: (
  kek: Key,
  encrypted_session_keys: EncryptedSessionKey[]
) => Decryptor = fromEncryptedSessionKeys(
  session_key => new Decryptor(session_key)
);

/**
 * Create a new `Encryptor` from an available session key.
 *
 * From an array of encrypted session keys, instantiate a `Encryptor` using
 * the first session key that is able to be decrypted using the
 * provided `kek`.  If no session key can be decrypted, throw a
 * `SessionKeyDecryptionError`.
 *
 * The array of session keys allows for KEK rotation; it is important that
 * `Encryptor` be instantiated using this function rather than constructing
 * it directly.
 */
export const encryptorFromEncryptedSessionKeys: (
  kek: Key,
  encrypted_session_keys: EncryptedSessionKey[]
) => Encryptor = fromEncryptedSessionKeys(
  session_key => new Encryptor(session_key)
);
