import {Bucket, BucketData, BucketFieldValue} from './Bucket';
import {StagingBucket} from './StagingBucket';
import {DelayedRecursionError} from './DelayedRecursionError';

export class DelayedStagingBucket extends StagingBucket {
  /**
   * Queued data
   */
  private _queued: BucketData = {};

  /**
   * Delay timer id
   */
  private _timer: NodeJS.Timeout | 0 = 0;

  /**
   * Whether #processValues is still on the stack
   */
  private _processing = 0;

  constructor(bucket: Bucket) {
    super(bucket);
  }

  public setValues(data: BucketData): this {
    for (const name in data) {
      if (this._queued[name] === undefined) {
        this._queued[name] = [];
      }

      // merge individual indexes
      this.merge(data[name], this._queued[name]);
    }

    this._setTimer();
    return this;
  }

  private _setTimer() {
    // no need to re-set timers
    if (this._timer) {
      return;
    }

    // invoke when stack clears
    this._timer = setTimeout(() => {
      // just in case something prevented us from decrementing the
      // lock (e.g. an exception)
      this._processing = 0;
      this.processValues();
    }, 0);
  }

  /**
   * Retrieve the data that will result after a merge
   *
   * This should be used sparingly, since if this is called before data is
   * actually merged into the bucket, then it is possible that the values will
   * change after validations are run.
   */
  public getPendingDataByName(name: string, diff: BucketData) {
    diff = diff || this._queued;

    const pending = super.getDataByName(name);
    if (!(this._queued[name] || diff[name])) {
      return pending;
    }

    // merge the queued data
    this.merge(this._queued[name] || diff[name], pending, true);
    return pending;
  }

  public getDataByName(name: string) {
    // if enqueued data is requested, then we have no choice but to merge to
    // ensure that the data is up-to-date
    if (this._queued[name] !== undefined) {
      this.processValues();
    }

    return super.getDataByName(name);
  }

  public getData() {
    this.processValues();
    return super.getData();
  }

  public each(c: (data: BucketFieldValue, name: string) => void) {
    this.processValues();
    return super.each(c);
  }

  public getFilledDiff() {
    this.processValues();
    return super.getFilledDiff();
  }

  public hasIndex(name: string, i: number) {
    this.processValues();
    return super.hasIndex(name, i);
  }

  public processValues() {
    // if no timer is set, then we have no data
    if (!this._timer) {
      return this;
    }

    // since additional data may be queued as a consequence of the below
    // set, prepare for it by providing an empty queue
    this._withProcessLock(() => {
      const oldqueue = this._queued;
      this._queued = {};
      this._timer = 0;

      super.setValues(oldqueue);
    });

    return this;
  }

  /**
   * Increment lock for duration of function F
   *
   * The lock is incremented before calling F and decremented after F
   * returns.  The maximum lock count is hard-coded to 5, after which an
   * exception will be thrown.  The intent of this value is to provide a
   * little bit of flexibility to hook implementations that allows for
   * some gentle recursion, but preempts hooks when it looks like things
   * may be going awry.
   *
   * @param {Function} f function to call
   *
   * @throws {DelayedRecursionError} when lock count exceeds maximum
   *
   * @return {undefined}
   */
  private _withProcessLock(f: () => void) {
    // protect against runaway recursion from hooks
    if (this._processing === 5) {
      // throwing the exception will halt execution of the stack if
      // uncaught, meaning _processing will never be fully
      // decremented, so clear it out
      this._processing = 0;

      throw new DelayedRecursionError(
        'Recursion on DelayedStagingBucket#processValues ' +
          '(likely caused by a bucket hook)'
      );
    }

    // note that this isn't guarded with a try/catch, because we handle
    // resetting the lock in a couple other places---this is more
    // fool-proof, since catching an exception mail fail in certain
    // circumstances (e.g. when reaching a resource limit, like
    // stack/memory)
    this._processing++;
    f();
    this._processing--;
  }
}
