/**
 * Staging key/value store
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter as EventEmitterBase} from 'events';
// eslint-disable-next-line @typescript-eslint/no-var-requires
const EventEmitter: typeof EventEmitterBase = require('../events').EventEmitter;

import type {Bucket, BucketData, BucketFieldValue} from './Bucket';

/**
 * Stages and merges values into underlying key/value store
 */
export class StagingBucket extends EventEmitter implements Bucket {
  /**
   * Triggered when data in the bucket is updated, before it's committed
   * @type {string}
   */
  static readonly EVENT_UPDATE = 'update';
  static readonly EVENT_STAGING_PRE_UPDATE = 'preStagingUpdate';
  static readonly EVENT_STAGING_UPDATE = 'stagingUpdate';
  static readonly EVENT_PRE_COMMIT = 'preCommit';
  static readonly EVENT_COMMIT = 'commit';
  static readonly EVENT_STAGING_PRE_REVERT = 'preRevert';
  static readonly EVENT_STAGING_REVERT = 'revert';
  static readonly EVENT_STAGING_POST_REVERT = 'postRevert';

  /**
   * Contains staged (uncommitted) data
   * @type {Object.<string,Array>}
   */
  private _staged: BucketData = {};

  /**
   * Represents the current state of the bucket for fast retrieval
   * @type {Object.<string,Array>}
   */
  private _curdata: BucketData = {};

  /**
   * Whether data is staged but not committed
   *
   * Ah, brining back the "dirty" term from the good 'ol days of the "dirty
   * bucket"!
   *
   * @type {boolean}
   */
  private _dirty = false;

  /**
   * Prevent setCommittedValues from bypassing staging
   * @type {boolean}
   */
  private _noStagingBypass = false;

  /**
   * Initializes staging bucket with the provided data bucket
   *
   * @param {Bucket} bucket bucket in which to store data
   *
   * @return {undefined}
   */
  constructor(private _bucket: Bucket) {
    super();
    this._init();
  }

  private _init() {
    // forward events
    this._bucket.on(StagingBucket.EVENT_UPDATE, (data: unknown) => {
      this.emit(StagingBucket.EVENT_UPDATE, data);
    });
    this._initState();
  }

  private _initState() {
    // ensure that we don't modify the original data
    const retdata = Object.create(this._bucket.getData());

    this._curdata = retdata;
    this._dirty = false;
  }

  public merge(
    src: BucketFieldValue | null,
    dest: BucketFieldValue | null,
    nonull = false
  ): void {
    nonull = !!nonull;
    if (!src || !dest) return;

    const len = src.length;

    for (let i = 0; i < len; i++) {
      const data = src[i];

      // don't merge if it's undefined or if null and nulls were not
      // permitted
      if (data === undefined) {
        continue;
      } else if (nonull && data === null) {
        // nulls mark the end of the vector
        dest.length = i;
        break;
      }

      // merge with bucket data
      dest[i] = data;
    }
  }

  /**
   * Analgous to setValues(), but immediately commits the changes
   *
   * This still calls setValues() to ensure all events are properly kicked
   * off.
   */
  public setCommittedValues(data: BucketData, _?: boolean, __?: boolean): this {
    if (this._noStagingBypass) {
      return this.setValues(data);
    }

    this._bucket.setValues(data);

    // no use in triggering a pre-update, since these values are
    // already committed
    this.emit(StagingBucket.EVENT_STAGING_UPDATE, data);

    return this;
  }

  /**
   * Prevent #setCommittedValues from bypassing staging
   *
   * When set, #setCommittedValues will act as an alias of #setValues.
   *
   * @return {StagingBucket} self
   */
  public forbidBypass(): this {
    this._noStagingBypass = true;
    return this;
  }

  /**
   * Determine whether values have changed
   *
   * If all values are identical to the current bucket values returns
   * `false`.  Otherwise, `true`.
   *
   * Because JSON serializes all undefined values to `null`, only the
   * final null in a diff is considered terminating; the rest are
   * converted into `undefined`.
   *
   * This creates a fresh copy of the data to prevent unexpected
   * side-effects.
   *
   * @param {Object.<string,Array>} data        key/value data or diff
   *
   * @return {boolean} whether a change was recognized
   */
  private _parseChanges(data: Record<string, unknown>): {
    changed: boolean;
    result: BucketData;
  } {
    // generate a new object so that we don't mutate the original, which
    // may be used elsewhere
    const result: BucketData = {};

    let changed = false;

    for (const name in data) {
      const values = Array.prototype.slice.call(data[name], 0);
      const cur = this._curdata[name] || [];
      const len = this._length(values);
      const has_null = len !== values.length;

      result[name] = values;

      const merge_len_change = has_null && len < cur.length;

      // quick change check; but we still have to keep processing the
      // data to generate a proper diff and process non-terminating
      // nulls
      if (merge_len_change) {
        changed = true;
      }

      for (let index = 0; index < len; index++) {
        // see #_length; if ending in a null, this is offset by -1
        const last_index = index === len;

        // only the last null is terminating; the rest are
        // interpreted as `undefined' (because JSON serializes
        // undefined values to `null')
        if (values[index] === null && !last_index) {
          values[index] = undefined;
        }

        if (values[index] === undefined) {
          continue;
        }

        if (!this._deepEqual(values[index], cur[index])) {
          changed = true;
          continue;
        }

        // unchanged
        values[index] = undefined;
      }

      // if nothing is left, remove entirely
      if (!values.some(x => x !== undefined)) {
        delete result[name];
      }
    }

    return {changed, result};
  }

  /**
   * Get actual length of vector
   *
   * This considers when the last element of the vector is a null value,
   * which is a truncation indicator.
   *
   * @param {Array} values value vector
   *
   * @return {number} length of vector considering truncation
   */
  private _length(values: unknown[]): number {
    if (values[values.length - 1] === null) {
      return values.length - 1;
    }

    return values.length;
  }

  /**
   * Recursively check for equality of two vavlues
   *
   * This only recognizes nested arrays (vectors).  They are not
   * traditionally encountered in the bucket, but may exist.
   *
   * The final comparison is by string equality, since bucket values are
   * traditionally strings.
   *
   * @param {*} a first vector or scalar
   * @param {*} b second vector or scalar
   *
   * @return {boolean} whether `a` and `b` are equal
   */
  private _deepEqual(a: unknown | unknown[], b: unknown | unknown[]): boolean {
    if (Array.isArray(a)) {
      if (!Array.isArray(b) || a.length !== b.length) {
        return false;
      }

      return a
        .map((item, i) => this._deepEqual(item, b[i]))
        .every(res => res === true);
    }

    return '' + a === '' + b;
  }

  /**
   * Explicitly sets the contents of the bucket
   *
   * Because JSON serializes all undefined values to `null`, only the
   * final null in a diff is considered terminating; the rest are
   * converted into `undefined`.  Therefore, it is important that all
   * truncations include no elements in the vector after the truncating null.
   *
   * @param {Object.<string,Array>} given_data associative array of the data
   *
   * @return {Bucket} self
   */
  public setValues(given_data: BucketData): this {
    const {changed, result: data} = this._parseChanges(given_data);

    if (!changed) {
      return this;
    }

    this.emit(StagingBucket.EVENT_STAGING_PRE_UPDATE, data);

    for (const name in data) {
      const item = data[name];

      // initialize as array if necessary
      if (this._staged[name] === undefined) {
        this._staged[name] = [];
      }

      // since _curdata's prototype is a reference to the internal data of
      // the other bucket (gah!---for perf reasons), we need to take care
      // to ensure that we do not modify it...this accomplishes that
      if (Object.hasOwnProperty.call(this._curdata, name) === false) {
        if (this._curdata[name] !== undefined) {
          this._curdata[name] = Array.prototype.slice.call(
            this._curdata[name],
            0
          );
        } else {
          this._curdata[name] = [];
        }
      }

      // merge with previous values
      this.merge(item, this._staged[name]);

      // we do not want nulls in our current representation of the
      // data
      this.merge(item, this._curdata[name], true);
    }

    this._dirty = true;
    this.emit(StagingBucket.EVENT_STAGING_UPDATE, data);

    return this;
  }

  /**
   * Overwrites values in the original bucket
   *
   * @param {Object.<string,Array>} data associative array of the data
   *
   * @return {StagingBucket} self
   */
  public overwriteValues(data: BucketData): this {
    const new_data: BucketData = {};

    for (const name in data) {
      new_data[name] = Array.prototype.slice.call(data[name], 0);

      // a terminating null ensures all data is overwritten, rather than
      // just the beginning indexes
      new_data[name]?.push(null);
    }

    return this.setValues(new_data);
  }

  /**
   * Returns staged data
   *
   * @return {Object.<string,Array>}
   */
  public getDiff(): BucketData {
    return this._staged;
  }

  /**
   * Returns a field-oriented diff filled with all values rather than a
   * value-oriented diff
   *
   * Only the fields that have changed are returned. Each field contains its
   * actual value---not the diff representation of what portions of the field
   * have changed.
   *
   * return {Object} filled diff
   */
  public getFilledDiff(): BucketData {
    const ret: BucketData = {};

    // return each staged field
    for (const field in this._staged) {
      // retrieve the diff for this field
      ret[field] = Array.prototype.slice.call(this._staged[field], 0);
    }

    return ret;
  }

  /**
   * Reverts staged changes, preventing them from being committed
   *
   * This will also generate a diff and raise the same events that would be
   * raised by setting values in the conventional manner, allowing reverts to
   * transparently integrate with the remainder of the system.
   *
   * @return {StagingBucket} self
   */
  public revert(evented?: boolean): this {
    evented = evented === undefined ? true : !!evented;

    const data: BucketData = {};

    // generate data for this revert (so that hooks may properly handle it)
    for (const name in this._staged) {
      const curstaged = this._staged[name];
      const orig = this._bucket.getDataByName(name);

      data[name] = [];
      for (const i in curstaged) {
        const index = +i;

        // if the original value is undefined, then we want to remove
        // the value entirely, *not* set it to undefiend (which would
        // affect the length of the array)
        if (orig?.[index] === undefined) {
          delete data[name]?.[index];
          continue;
        }

        // data[name] is set above, therefore safe assertion
        data[name]![index] = orig[index];
      }
    }

    if (evented) {
      this.emit(StagingBucket.EVENT_STAGING_PRE_REVERT, data);
      this.emit(StagingBucket.EVENT_STAGING_PRE_UPDATE, data);
    }

    this._staged = {};
    this._initState();

    // everything after this point is evented
    if (!evented) {
      return this;
    }

    // trigger revert after update (since we did preRevert before update;
    // this also allows logic to disable further updates; DO NOT CHANGE
    // ORDER WITHOUT LOOKING AT WHAT USES THIS!)
    this.emit(StagingBucket.EVENT_STAGING_UPDATE, data);
    this.emit(StagingBucket.EVENT_STAGING_REVERT, data);

    // a distinct event lets hooks know that a revert has been completed
    // (which may be useful for allowing asychronous data to be
    // automatically committed following a revert, rather than once again
    // allowing the staging bucket to be considred dirty)
    this.emit(StagingBucket.EVENT_STAGING_POST_REVERT, data);

    return this;
  }

  /**
   * Commits staged changes, merging them with the bucket
   *
   * @return {StagingBucket} self
   */
  public commit(store?: {old: BucketData}): this {
    const old = this._staged;

    this.emit(StagingBucket.EVENT_PRE_COMMIT);

    this._bucket.setValues(this._staged);
    this._staged = {};

    this.emit(StagingBucket.EVENT_COMMIT);

    this._initState();

    if (typeof store === 'object') {
      store.old = old;
    }

    return this;
  }

  /**
   * Clears all data from the bucket
   *
   * @return {Bucket} self
   */
  public clear(): this {
    this._bucket.clear();
    return this;
  }

  /**
   * Calls a function for each each of the values in the bucket
   *
   * @param {function( Object, number )} callback function to call for each
   *                                              value in the bucket
   *
   * @return {Bucket} self
   */
  public each(callback: (data: BucketFieldValue, name: string) => void): this {
    for (const name in this._curdata) {
      callback(this._curdata[name], name);
    }

    return this;
  }

  /**
   * Returns the data for the requested field
   *
   * WARNING: This can be a potentially expensive operation if there is a
   * great deal of staged data. The staged data is merged with the bucket data
   * on each call. Do not make frequent calls to retrieve the same data. Cache
   * it instead.
   *
   * @param {string} name field name (with or without trailing brackets)
   *
   * @return {Array} data for the field, or empty array if none
   */
  public getDataByName(name: string): BucketFieldValue | null {
    if (this._curdata[name]) {
      // important: return a clone so that operations on this data doesn't
      // modify the bucket without us knowing!
      return Array.prototype.slice.call(this._curdata[name]);
    }

    return [];
  }

  /**
   * Returns original bucket data by name, even if there is data staged atop
   * of it
   *
   * There is no additional overhead of this operation versus getDataByName()
   *
   * @param {string} name field name (with or without trailing brackets)
   *
   * @return {Array} data for the field, or empty array if none
   */
  public getOriginalDataByName(name: string): BucketFieldValue | null {
    return this._bucket.getDataByName(name);
  }

  /**
   * Returns the data as a JSON string
   *
   * @return {string} data represented as JSON
   */
  public getDataJson(): string {
    return this._bucket.getDataJson();
  }

  /**
   * Return raw bucket data
   *
   * todo: remove; breaks encapsulation
   *
   * @return {Object} raw bucket data
   */
  public getData(): BucketData {
    return this._curdata;
  }

  /**
   * Calls a function for each each of the values in the bucket matching the
   * given predicate
   *
   * @param {function(string)}           pred predicate
   * @param {function( Object, number )} c    function to call for each
   *                                          value in the bucket
   *
   * @return {StagingBucket} self
   */
  public filter(
    pred: (val: string) => boolean,
    c: (data: BucketFieldValue, name: string) => void
  ): this {
    this.each((data, name) => {
      if (pred(name)) {
        c(data, name);
      }
    });
    return this;
  }

  public hasIndex(name: string, i: number): boolean {
    return this._curdata?.[name]?.[i] !== undefined;
  }

  public isDirty(): boolean {
    return this._dirty;
  }
}
