/**
 * Liza Defaulting Bucket Layer
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {ClassData} from '../../client/Cmatch';
import {BucketData, BucketDefaults, BucketFieldValue} from '../Bucket';
import {LayeredBucket} from './LayeredBucket';

export class DefaultingLayeredBucket implements LayeredBucket {
  /**
   * Whether or not this layer is active
   */
  private _isActive = false;

  /**
   * Create a new DefaultingLayeredBucket
   *
   * @param innerLayer - a layer to decorate
   * @param when       - a classification for when this layer is applicable
   * @param defaults   - default values to apply
   */
  constructor(
    private readonly _innerLayer: LayeredBucket,
    private readonly _when: string,
    private readonly _defaults: BucketDefaults
  ) {}

  /**
   * Update the class data
   *
   * @param classData - the class data to update
   */
  public updateClassData(classData: ClassData): this {
    // an invalid classification should be prevented by static analysis but it
    // isn't right now. Default to false if it is invalid. We should log if this
    // happens as well
    this._isActive = classData[this._when]?.is ?? false;

    this._innerLayer.updateClassData(classData);

    return this;
  }

  /**
   * Return raw bucket data
   *
   * TODO: remove; breaks encapsulation
   *
   * @return raw bucket data
   */
  public getData(): BucketData {
    const innerData = this._innerLayer.getData();
    if (!this._isActive) {
      return innerData;
    }

    Object.entries(this._defaults).forEach(([k, v]) => {
      const existingData: BucketFieldValue | undefined = innerData[k];
      innerData[k] = this.alignIndexes(existingData, v);
    });

    return innerData;
  }

  /**
   * Returns the data for the requested field
   *
   * @param name - name of the field (with or without trailing brackets)
   *
   * @return data for the field
   */
  public getDataByName(name: string): BucketFieldValue | null {
    const innerData = this._innerLayer.getDataByName(name);

    if (!this._isActive) {
      return innerData;
    }

    // If there is no default for this field, fall back
    // to an empty string to prevent alignIndexes returning
    // undefined for the default value of the bucket field
    const defaultToApply = this._defaults[name] || '';
    return this.alignIndexes(innerData, defaultToApply);
  }

  /**
   * Align indexes on default values to match the indexes of the source data
   *
   * There may be fields that have multiple indexes where we need to apply
   * defaults across those indexes. If there was no source data found then it
   * seems reasonable to use a single index because that it the default for the
   * bucket
   *
   * @param src        - source data
   * @param defaultVal - the default to apply
   *
   * @returns defaulted bucket data
   */
  private alignIndexes(
    src: BucketFieldValue | null,
    defaultVal: string
  ): BucketFieldValue {
    if (!src) {
      return [defaultVal];
    }

    return src.map(cur => (cur === '' ? defaultVal : cur));
  }
}
