/**
 * Client Interface
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This interface was created to satisfy MongoServerDao and may not be
 * sufficiently general for other database abstractions.
 */

import {Program} from '../program/Program';
import {CmatchData} from './Cmatch';
import {ClientQuote} from './quote/ClientQuote';
import {PositiveInteger} from '../numeric';
import {Ui} from '../ui/Ui';

/**
 * Client abstraction
 */
export interface ClientInterface {
  program: Program;

  /**
   * Saves a quote to the database
   */
  getQuote(): ClientQuote;

  /**
   * Handle error events
   *
   * Ideally, this should never happen. This method indicates an error that
   * could not be properly handled by another part of the system. Let the user
   * know that this should not be happening and trigger our own error event.
   *
   * @param e - error
   */
  handleError(error: Error): void;

  /**
   * Update cmatch data where necessary
   *
   * @param cmatch - the updated class match data
   */
  cmatchUpdate(cmatch: CmatchData): void;

  /**
   * Whether there is a current step loaded
   */
  hasCurrentStep: () => boolean;

  /**
   * Get the current step id
   *
   * @returns the current step id
   */
  getCurrentStepId: () => PositiveInteger;

  /**
   * Returns the UI object
   */
  getUi: () => Ui | undefined;

  /**
   * Handles client-side events
   */
  handleEvent(
    event_name: string,
    data: Record<string, unknown>,
    callback?: (data: any) => void,
    error_callback?: (err: Error) => void
  ): ClientInterface;
}
