/**
 * Data API mediator
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {ClientQuote} from '../quote/ClientQuote';
import {DataApiManager} from '../../dapi/DataApiManager';
import {DataValidator} from '../../validate/DataValidator';
import {MissingDataError} from '../../dapi/MissingDataError';
import {PositiveInteger} from '../../numeric';
import {Ui} from '../../ui/Ui';

export interface UiFieldData {
  value: string | number;
  label: string | number;
  label_id: string;
}

export type DapiResult = Record<string, Record<string, string> | boolean> & {
  ___empty: boolean;
};

export type DapiUpdate = (
  name: string,
  i: number | string,
  results: DapiResult,
  map: Record<string, string>
) => void;

/**
 * Mediate updates to system state based on DataAPI request status and
 * results
 *
 * The UI will be updated to reflect the options returned by DataAPI
 * requests.  When a field is cleared of all options, any errors on that
 * field will be cleared.
 */
export class DataApiMediator {
  /** Optional callback for update. Set during monitor */
  private _update_cb: DapiUpdate | null = null;
  /**
   * Initialize mediator
   *
   * The provided DataValidator DATA_VALIDATOR must be the same validator
   * used to produce errors on fields to ensure that its state can be
   * appropriately cleared.
   *
   * DAPI_MAP stores destination:source field mappings, where source is
   * the result of the DataAPI call and destination is the target field in
   * which to store those data.
   *
   * Since the active quote changes at runtime, this constructor accepts a
   * quote function QUOTEF to return the active quote.
   *
   * @param dapi_map       - field source and destination map
   * @param quotef         - nullary function returning quote
   * @param ui             - UI
   * @param data_validator - data validator
   */
  constructor(
    private _dapi_map: Record<string, Record<string, string>>,
    private _quotef: () => ClientQuote,
    private _ui?: Ui,
    private _data_validator?: DataValidator
  ) {
    if (typeof _dapi_map !== 'object') {
      throw TypeError('dapi_map must be a key/value object');
    }
  }

  /**
   * Hook given DataApiManager
   *
   * Handled events are updateFieldData, clearFieldData, and fieldLoaded.
   *
   * @param dapi_manager - manager to hook
   */
  public monitor(
    dapi_manager: DataApiManager,
    update_cb: DapiUpdate | null = null
  ) {
    this._update_cb = update_cb;
    dapi_manager.on(
      'updateFieldData',
      this._updateFieldData.bind(this, dapi_manager)
    );

    dapi_manager.on('clearFieldData', this._clearFieldOptions.bind(this));
    dapi_manager.on('fieldLoaded', this._clearFieldFailures.bind(this));

    return this;
  }

  /**
   * Set field options
   *
   * If the bucket value associated with NAME and INDEX are in the result
   * set RESULTS, then it will be selected.  Otherwise, the first result
   * in RESULTS will be selected, if any.  If there are no results in
   * RESULTS, the set value will be the empty string.
   *
   * @param dapi_manager - DataAPI manager
   * @param name         - field name
   * @param index        - field index
   * @param val_label    - value and label
   * @param results      - DataAPI result set
   */
  private _updateFieldData(
    dapi_manager: DataApiManager,
    name: string,
    index: string | -1,
    val_label: UiFieldData[],
    results: DapiResult
  ) {
    const group = this._ui?.getCurrentStep?.()?.getElementGroup(name);

    // invoke optional cb
    this._update_cb &&
      this._update_cb(name, index, results, this._dapi_map[name] ?? {});

    // allow the stack to clear before setting data to allow any
    // existing bucket processing to complete before hooks are kicked
    // off yet again (which, in practice, could otherwise result in
    // infinite recursion depending on what the hooks are doing), and to
    // allow the UI to update with any new elements we might be about to
    // populate
    setTimeout(() => {
      const quote = this._quotef();
      const existing = <string[]>quote.getDataByName(name) || [];

      let indexes: string[] = [];

      // index of -1 indicates that all indexes should be affected
      if (index === -1) {
        indexes = existing;
      } else {
        indexes[+index] = index;
      }

      // keep existing value if it exists in the result set, otherwise
      // use the first value of the set
      const field_update = indexes.map((_: unknown, i: number) =>
        results[existing[i]] ? existing[i] : this._getDefaultValue(val_label)
      );

      indexes.forEach((_: unknown, i: number) =>
        group?.setOptions(name, i, val_label, existing[i])
      );

      const label_id = val_label[0]?.label_id || '';

      const update = this._populateWithMap(
        dapi_manager,
        name,
        indexes,
        quote,
        label_id
      );

      update[name] = field_update;

      quote.setData(update);
    });
  }

  /**
   * Generate bucket update with field expansion data
   *
   * If multiple indexes are provided, updates will be merged.  If
   * expansion data are missing, then the field will be ignored.  If a
   * destination field is populated such that auto-expanding would
   * override that datum, then that field will be excluded from the
   * expansion.  Labels are exempt from this rule, since they are
   * considered to be married to the value; labels are not
   * user-modifiable.
   *
   * @param dapi_manager - manager responsible for fields
   * @param name         - field name
   * @param indexes      - field indexes
   * @param quote        - source quote
   * @param label_id     - name of label field
   */
  private _populateWithMap(
    dapi_manager: DataApiManager,
    name: string,
    indexes: string[],
    quote: ClientQuote,
    label_id: string
  ): Record<string, (string | number)[]> {
    const map = this._dapi_map[name];

    // calculate field expansions for each index, which contains an
    // object suitable as-is for use with Quote#setData
    const expansions = indexes.map((_, i) => {
      try {
        return dapi_manager.getDataExpansion(
          name,
          <PositiveInteger>i,
          quote,
          map,
          false,
          {}
        );
      } catch (e) {
        if (e instanceof MissingDataError) {
          // this value is ignored below
          return undefined;
        }

        throw e;
      }
    });

    // produce a final update that merges each of the expansions
    return expansions.reduce(
      (update: Record<string, string[]>, expansion, i) => {
        // it's important that we check here instead of using #filter on
        // the array so that we maintain index association
        if (expansion === undefined) {
          return update;
        }

        // merge each key individually
        Object.keys(expansion).forEach(key => {
          const existing = (quote.getDataByName(key) || [])[i];

          // if set and non-empty, then it's already populated and we
          // must leave the value alone (so as not to override
          // something the user directly entered)
          if (key !== label_id && existing !== undefined && existing !== '') {
            return;
          }

          update[key] = update[key] || [];
          update[key][i] = expansion[key][i];
        });

        return update;
      },
      {}
    );
  }

  /**
   * Clear field options
   *
   * @param name  - field name
   * @param index - field index
   */
  private _clearFieldOptions(name: string, index: number) {
    if (this._ui === undefined) {
      return;
    }

    const group = this._ui.getCurrentStep()?.getElementGroup(name);

    // ignore unknown fields
    if (!group) {
      return;
    }

    group.clearOptions(name, index);
  }

  /**
   * Clear field failures
   *
   * @param name  - field name
   * @param index - field index
   */
  private _clearFieldFailures(name: string, index: number) {
    if (this._data_validator === undefined) {
      return;
    }

    this._data_validator.clearFailures({[name]: [index]});
  }

  /**
   * Determine default value for result set
   *
   * @param val_label - value and label
   *
   * @return default value for result set
   */
  private _getDefaultValue(val_label: UiFieldData[]) {
    // default to the empty string if no results were returned
    if (val_label.length === 0) {
      return '';
    }

    return (val_label[0] || {}).value;
  }
}
