/**
 * Fetch rate event handler
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventHandler, EventHandlerCb} from './EventHandler';
import {Client} from '../Client';
import {ClientAction, ClientActionType} from '../action/ClientAction';
import {RateRequestResult} from '../../server/service/RatingService';
import {ClientDataProxy} from '../ClientDataProxy';
import {ClientQuote} from '../quote/ClientQuote';

/**
 * Handles fetchRate events
 */
export class FetchRateEventHandler implements EventHandler {
  /**
   * Initializes with client that will delegate the event
   *
   * @param _client     - client object
   * @param _data_proxy - used for fetchRate requests
   */
  constructor(
    private readonly _client: Client,
    private readonly _data_proxy: ClientDataProxy
  ) {}

  /**
   * Handles fetchRate events
   *
   * @param type - event id; ignored
   * @param c    - continuation to invoke on completion
   * @param data - event data
   */
  handle(
    _type: ClientActionType,
    c: EventHandlerCb,
    _data: ClientAction
  ): this {
    // if we're in the process of saving, then wait until all saves are
    // complete before continuing
    if (this._client.isSaving()) {
      // defer rating until after saving is complete
      this._client.once('postSaveAll', () => this._fetchRateResults(c));

      return this;
    }

    this._fetchRateResults(c);

    return this;
  }

  /**
   * Call a server endpoint to retrieve rating results
   *
   * @param c - a function to call when we are done fetching
   */
  private _fetchRateResults(c: EventHandlerCb): void {
    const quote = this._client.getQuote();

    // grab the rates from the server for the already posted quote data
    this._data_proxy.get(
      quote.getId() + '/fetchRate',
      (response: RateRequestResult, err: Error) => {
        if (err) {
          // error; do not provide rate data
          c(err, null);
          return;
        }

        this._saveRateResults(response, quote, c);
      }
    );
  }

  private _saveRateResults(
    response: RateRequestResult,
    quote: ClientQuote,
    c: any
  ) {
    const {
      initialRatedDate: initial_rated = <UnixTimestamp>0,
      lastRatedDate: last_rated = <UnixTimestamp>0,
      data = {},
    } = response.content || {};

    const step_id = quote.getCurrentStepId();

    // Set the data in the underlying bucket that the staging bucket uses
    quote.refreshData(data);

    // Set the data in staging as well so that its values are updated
    quote.setData(data);

    quote.setInitialRatedDate(initial_rated);
    quote.setLastPremiumDate(last_rated);

    // invalidate the step to force emptying of the bucket, ensuring
    // that data is updated on the screen when it's re-rated (will
    // be undefined if the step hasn't been loaded yet, in which
    // case it doesn't need to be invalidated)
    const stepui = this._client.getUi().getStep(step_id);

    if (stepui !== undefined) {
      stepui.invalidate();
      stepui.emptyBucket();
    }

    c(null, response.content.data);
  }
}
