/**
 *  Value set event handler
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventHandler, EventHandlerCb} from './EventHandler';
import {ClientActionType} from '../action/ClientAction';
import {ClientQuote} from '../quote/ClientQuote';
import {Client} from '../Client';
import {DelayedStagingBucket} from '../../bucket/DelayedStagingBucket';

export interface ClientActionSetData {
  elementName: string;
  indexes: number[];
  value: string[] | string;
}

/**
 * Set field values
 */
export class ValueSetEventHandler implements EventHandler {
  /**
   * Initializes with client that will delegate the event
   *
   * @param client - client object
   */
  constructor(protected readonly _client: Client) {}

  /**
   * Set value of specified fields
   *
   * If the destination index exceeds the number of indexes of the source
   * field, the last available source index will be used.  This matches
   * the behavior of assertions.
   *
   * If the value is a string, the value will be set on all the indexes
   * of the destination field.
   *
   * @param type - event id; ignored
   * @param c    - continuation to invoke on completion
   * @param data - event data
   */
  handle(
    _type: ClientActionType,
    c: EventHandlerCb,
    data: ClientActionSetData
  ): this {
    const quote: ClientQuote = this._client.getQuote();
    const maxi: number = data.value.length - 1;

    const set_data: string[] = data.indexes.reduce(
      (result: string[], desti: number) => {
        result[desti] =
          typeof data.value === 'string'
            ? data.value
            : data.value[Math.min(desti, maxi)];
        return result;
      },
      []
    );

    quote.setData({[data.elementName]: set_data});
    quote.visitData(bucket => {
      // Program is expecting a DelayedStagingBucket
      const staging_bucket = <DelayedStagingBucket>bucket;
      staging_bucket.processValues();
      c();
    });

    return this;
  }
}
