/**
 * Client-side quote
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {BaseQuote} from '../../quote/BaseQuote';
import {PositiveInteger} from '../../numeric';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {QuoteTransport} from '../transport/QuoteTransport';
import {StagingBucket} from '../../bucket/StagingBucket';

export type Data = Record<string, any>;
export type VisibilityScalar = PositiveInteger;
export type VisibilityVector = VisibilityScalar[];
export type VisibilityMatrix = VisibilityVector[];

/**
 * Controller for the program client
 */
export declare class ClientQuote extends BaseQuote {
  constructor(
    quote: BaseQuote,
    data: CommonObject,
    staging_callback: (bucket: QuoteDataBucket) => QuoteDataBucket
  );
  /**
   * Set the classifier to be used for data classification
   *
   * The classifier should return an object containing all classifications and
   * a single boolean value per classification.
   */
  setClassifier(known_fields: any, classifier: any): ClientQuote;

  /**
   * Returns all bucket data
   */
  getData(): Data;

  /**
   * Returns data from the quote
   */
  getDataByName(name: string): any[];

  /**
   * Set data on the quote
   */
  setDataByName(name: string, values: string[]): void;

  /**
   * Returns most recent classifications
   */
  getLastClassify(): Data;

  /**
   * Commits changes to quote and attempts to save
   *
   * @param transport - Transport for the data
   * @param callback  - Function to call with results
   */
  autosave(transport: QuoteTransport, callback?: any): this;

  /**
   * Invalidate any pending autosaves
   */
  invalidateAutosave(): void;

  /**
   * Returns whether the quote is locked from modifications
   *
   * @return true if locked, otherwise false
   */
  isLocked(): boolean;

  /**
   * Sets the top visited step id
   *
   * If the provided step id is less than the current step, then the current
   * step id is used instead.
   *
   * @param step_id - ID of the step to set
   */
  setTopVisitedStepId(step_id: PositiveInteger): void;

  /**
   * Returns the id of the highest step the quote has reached
   *
   * @return top visited step id
   */
  getTopVisitedStepId(): PositiveInteger;

  /**
   * Hooks classify event and immediately triggers the continuation with the
   * last classification data, if any
   *
   * Useful if the classification event has already been kicked off, but the
   * data is needed (prevents race conditions).
   *
   * We provide this method rather than a getter to enforce encapsulation (we
   * won't consider it violated here since event handlers normally have access
   * to that data anyway)---this signifies intent.
   *
   * @param {function(Object)} c continuation
   */
  onClassifyAndNow(c: any): this;

  /**
   * Sets an explicit lock, providing a reason for doing so
   *
   * @param reason - lock reason
   * @param step   - step that user may not navigate prior
   *
   * @return self
   */
  setExplicitLock(reason: string, step?: number): this;

  /**
   * Visits staging data
   */
  visitData(visitor: (bucket: StagingBucket) => void): void;

  /**
   * Returns the id of the current step
   *
   * @return id of current step
   */
  getCurrentStepId(): PositiveInteger;

  /**
   * Force a classification
   */
  forceClassify(): void;

  /**
   * Set quote data without considering it to be a modification
   *
   * Set underlying quote data DATA without triggering a client-side
   * modifications.  The intended us of this is to refresh quote
   * data from the server, which has already been saved.
   *
   * DATA will completely overwrite existing values; it will not
   * merge indexes like normal updates.
   *
   * @param data - data to overwrite
   */
  refreshData(data: Data): this;
}
