/**
 * Retrieves API data from bucket
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import type {DataApi, DataApiInput, DataApiResult} from './DataApi';
import type {Bucket} from '../bucket/Bucket';

/**
 * Retrieve data from the bucket
 */
export class BucketDataApi implements DataApi {
  /**
   * Initialize data API
   *
   * @param _bucket - Bucket to use as data source
   * @param _params - dapi parameters
   */
  constructor(
    private readonly _bucket: Bucket,
    private readonly _params: Record<string, string>
  ) {}

  /**
   * Request data from the bucket
   *
   * @param data     - request params
   * @param callback - server response callback
   */
  request(_data: DataApiInput, callback: NodeCallback<DataApiResult>): this {
    const rows: DataApiResult = [];

    for (const i in this._params) {
      const field = this._params[i],
        fdata = this._bucket.getDataByName(field);

      // There should be static anaylsis to ensure that fdata is a valid
      // bucket field, which means that #getDataByName should not return null
      // here. If it does, return an error.
      if (!fdata) {
        callback(
          new Error(
            `Bucket data api lookup for field "${field}" returned null`
          ),
          null
        );

        return this;
      }

      for (const index in fdata) {
        rows[index] = rows[index] || {};
        rows[index][field] = fdata[index];
      }
    }

    callback(null, rows);

    return this;
  }
}
