/**
 * Generic interface for data transmission
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {TokenId} from '../server/token/Token';

export enum DapiAction {
  TRIGGER = 'TRIGGER',
  TRIGGER_PENDING = 'TRIGGER_PENDING',
  PEND = 'PEND',
}

// const {Interface} = require('easejs');

/**
 * Result of DataAPI call
 *
 * This seemingly pointless type exists to emphasize that the result of all
 * DataAPI requests is and must be an array.  Overlooking this has been the
 * source of bugs in the past.
 */
export type DataApiResult = DataApiResultItem[];

/**
 * Individual item of DataAPI result
 *
 * Each result contains a set of key/value pairs.  Usually, the value is a
 * string or number, but more complex structures may be used server-side.
 */
export type DataApiResultItem = Record<string, any>;

/**
 * Inputs to the DataAPI
 *
 * Since data originate from the bucket, all values are expected to be
 * strings.
 */
export type DataApiInput = Record<string, string | string[]> | string;

/** Name of DataAPI */
export type DataApiName = NominalType<string, 'DataApiName'>;

export type DataApiType = 'rest' | 'local' | 'quote';

export type DapiParam = {
  name: string;
  default: {
    type: string;
    value: string;
  };
};

export type DataApiDefinition = {
  type: DataApiType;
  source: string;
  method: string;
  enctype: string;
  params: Record<string, DapiParam>;
  retvals: string[];
  static: Record<string, any>[];
  static_nonempty: boolean;
  static_multiple: boolean;
};

/**
 * Provides a generic interface for data transmission. The only assumption that
 * a user of this API shall make is that data may be sent and received in some
 * arbitrary, implementation-defined format, and that every request for data
 * shall yield some sort of response via a callback.
 */
export interface DataApi {
  request(
    data: DataApiInput | null,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    action?: DapiAction,
    token_id?: TokenId
  ): this;
}
