/**
 * Instantiate appropriate DataApi
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

'use strict';

const Class = require('easejs').Class;
const XhrHttpImpl = require('./http/XhrHttpImpl');

const {BucketDataApi} = require('./BucketDataApi');
const {HttpDataApi} = require('./http/HttpDataApi');
const {ParamCheckedDataApi} = require('./ParamCheckedDataApi');
const {ResponseCheckedDataApi} = require('./ResponseCheckedDataApi');
const {StaticAdditionDataApi} = require('./StaticAdditionDataApi');
const {WebhookBucketDataApi} = require('./WebhookBucketDataApi');

/**
 * Instantiates the appropriate DataApi object for the given service type
 */
module.exports = Class('DataApiFactory', {
  /**
   * Return a DataApi instance for the requested service type
   *
   * The source and method have type-specific meaning; that is, "source"
   * may be a URL and "method" may be get/post for a RESTful service.
   *
   * @param {string} type     service type (e.g. "rest")
   * @param {Object} desc     API description
   * @param {Bucket} bucket   active bucket
   * @param {string} api_name dapi name
   *
   * @return {DataApi} appropriate DataApi instance
   */
  'public fromType': function (type, desc, bucket, api_name) {
    return this.descLookup(api_name, desc).then(descl => {
      const static_data = descl['static'] || [];
      const nonempty = !!descl.static_nonempty;
      const multiple = !!descl.static_multiple;

      const api = this.createDataApi(type, descl, bucket);

      // We want to do param validation for the webhook DAPI
      // but not response validation
      if (type === 'webhook') {
        return new ParamCheckedDataApi(api, descl);
      }

      return new ParamCheckedDataApi(
        new ResponseCheckedDataApi(
          new StaticAdditionDataApi(api, nonempty, multiple, static_data),
          descl
        ),
        descl
      );
    });
  },

  /**
   * Look up dapi descriptor from configuration
   *
   * The default implementation just echoes back the given descriptor.
   *
   * @param {string} api_name dapi identifier
   * @param {Object} desc     given descriptor
   *
   * @return {Object} looked up descriptor
   */
  'virtual protected descLookup'(api_name, desc) {
    return Promise.resolve(desc);
  },

  /**
   * Create DataApi instance
   *
   * @param {string} type   API type
   * @param {Object} desc   API descriptor
   * @param {Bucket} bucket data bucket
   *
   * @return {DataApi}
   */
  'virtual protected createDataApi'(type, desc, bucket) {
    const source = desc.source || '';
    const method = desc.method || '';
    const enctype = desc.enctype || '';
    const downstream_system_id = desc.downstream_system_id || '';

    switch (type) {
      case 'rest':
        return this._createHttp(source, method, enctype);

      case 'local':
        // currently, only local bucket data sources are supported
        if (source !== 'bucket') {
          throw Error('Unknown local data API source: ' + source);
        }

        return new BucketDataApi(bucket, desc.retvals);

      case 'webhook':
        // currently, only bucket data sources are supported
        if (method !== 'bucket' && method !== 'oauth') {
          throw Error('Unsupported webhook method: ' + method);
        }

        if (method === 'oauth') {
          return this.createOAuthDapi(source, downstream_system_id);
        }

        return new WebhookBucketDataApi(bucket);
      default:
        throw Error('Unknown data API type: ' + type);
    }
  },

  /**
   * Create HttpDataApi instance
   *
   * The `Base` is intended to allow for the caller to mix traits in.
   *
   * @param {string} source  URL
   * @param {string} method  HTTP method
   * @param {string} enctype MIME media type (for POST)
   *
   * @return {HttpDataApi}
   */
  'private _createHttp'(source, method, enctype) {
    const impl = this.createHttpImpl();

    return new HttpDataApi(source, method.toUpperCase(), impl, enctype);
  },

  /**
   * Create HttpImpl
   *
   * This is simply intended to allow subtypes to override the type.
   *
   * @return {XhrHttpImpl}
   */
  'virtual protected createHttpImpl'() {
    // eslint-disable-next-line no-undef
    return XhrHttpImpl(XMLHttpRequest);
  },

  /**
   * Create HttpImpl
   *
   * This is simply intended to allow subtypes to override the type.
   *
   * @return {XhrHttpImpl}
   */
  'virtual protected createOAuthDapi'() {
    throw new Error('Client-side OAuth not implemented');
  },
});
