/**
 *  Restricts usage of a DataApi such that only declared params may be used in requests,
 *  and that required params must be used.
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter} from 'events';
import {
  DapiAction,
  DataApi,
  DataApiDefinition,
  DataApiInput,
  DataApiResult,
  DapiParam,
} from './DataApi';

// eslint-disable-next-line @typescript-eslint/no-var-requires
const {MissingParamError} = require('../error/MissingParamError');

/**
 * Restricts usage of a DataApi such that only declared params may be used in requests,
 * and that required params must be used.
 *
 * Also provides support for param defaults. Perhaps this logic doesn't belong
 * here, in which case it should be extracted into a separate decorator.
 */
export class ParamCheckedDataApi extends EventEmitter implements DataApi {
  /**
   * Available params and their defaults
   */
  private _params: Record<string, DapiParam>;

  /**
   * Initialize API restriction
   *
   * @param {DataApi} data_api data API to wrap
   * @param {Object}  desc     API description
   */
  constructor(private readonly _api: DataApi, desc: DataApiDefinition) {
    super();
    this._params = desc.params || {};
  }

  /**
   * Request data from the service
   *
   * @param {Object=}           data     request params
   * @param {function(Object)=} callback server response callback
   *
   * @return {DataApi} self
   */
  public request(
    data: DataApiInput,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    action?: DapiAction
  ): this {
    data = data || {};
    callback = callback || function () {};

    if (typeof data === 'string') {
      throw Error('ParamCheckedDataApi cannot be used with string requests');
    }

    // check the given params; if any are missing, then it should be
    // considered to be an error
    const reqdata = this._requestParamCheck(data);

    this._api.request(reqdata, callback, id, action);

    return this;
  }

  /**
   * Check request params to ensure that they are (a) known, (b) all required
   * params are provided and (c) contain valid values
   *
   * A separate object is returned to prevent side-effects.
   *
   * @param {Object} data request data
   *
   * @return {Object} request data to be sent
   */
  private _requestParamCheck(
    data: Record<string, string | string[]>
  ): Record<string, string | string[]> {
    const ret: Record<string, string | string[]> = {};

    for (const name in data) {
      // fail on unknown params
      if (!this._params[name]) {
        throw Error('Unknown param: ' + name);
      }
    }

    // yes, there are more efficient ways than looping through the above and
    // now through this, but the actual XHR itself will make it negligable,
    // so I'm going for clarity
    for (const name in this._params) {
      const def_data = this._params[name]['default'],
        def_val = (def_data && def_data.value) || '';

      // if the data is set, then we're good
      if (data[name]) {
        ret[name] = data[name];
        continue;
      }

      // the data is not set; if there's no default, then this is an error
      if (!def_val) {
        throw new MissingParamError('Missing param: ' + name);
      }

      // sorry---we only support string defaults for now
      if (def_data.type !== 'string') {
        throw Error('Only string param defaults are currently supported');
      }

      // use default
      ret[name] = def_val;
    }

    // return a separate object to prevent side-effects
    return ret;
  }
}
