/**
 *  Checks response for Data API
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter} from 'events';
import {
  DapiAction,
  DataApi,
  DataApiDefinition,
  DataApiInput,
  DataApiResult,
} from './DataApi';

/**
 * Ensures that the server responds with the
 * expected values of the Data API
 */
export class ResponseCheckedDataApi extends EventEmitter implements DataApi {
  /**
   * Expected return params
   */
  private _retvals: string[];

  /**
   * Initialize API restriction
   *
   * @param {DataApi} data_api data API to wrap
   * @param {Object}  desc     API description
   */
  constructor(private readonly _api: DataApi, desc: DataApiDefinition) {
    super();
    this._retvals = desc.retvals || [];
  }

  /**
   * Request data from the service and checks response
   *
   * @param {Object=}           data     request params
   * @param {function(Object)=} callback server response callback
   *
   * @return {DataApi} self
   */
  public request(
    data: DataApiInput,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    action?: DapiAction
  ): this {
    data = data || {};
    callback = callback || function () {};

    // make the request and check the response
    this._api.request(
      data,
      (err, response) =>
        callback(err, this._checkResponse(err, response, callback)),
      id,
      action
    );

    return this;
  }

  /**
   * Check the response data to ensure that all the expected params have been
   * returned for each item
   *
   * The response data should be an array of objects; if this is not the case
   * for the service in question, another decorator should be used to
   * transform the data *before* it gets to this decorator.
   *
   * The callback is not invoked by this method; instead, it will be passed to
   * any error events that may be emitted, allowing the handler to associate
   * it with the original request and invoke it manually if necessary.
   *
   * @param {Error|null}     err      error, if any
   * @param {Array.<Object>} response response data
   * @param {Function}       callback callback to be called with response
   *
   * @return {Object} original object if validations passed; otherwise {}
   */
  private _checkResponse(
    err: NullableError,
    response: DataApiResult | null,
    callback: NodeCallback<DataApiResult>
  ): DataApiResult {
    // the response should be an array; otherwise, we cannot process it to
    // see if the return data is valid (since it would not be in the
    // expected format---if the format needs conversion, a separate
    // decorator should handle that job *before* the data gets to this one)
    //
    // since ES5 isn't an option, we'll stick with this dirty hack
    if (err || !response || !response.slice) {
      this.emit(
        'error',
        err || TypeError('Response data is not an array'),
        callback,
        response
      );

      return [];
    }

    // for each item returned, check to ensure that it contains the expected
    // params
    let i = this._retvals.length;
    while (i--) {
      // check all items for the given param
      const param = this._retvals[i],
        response_err = this._checkAllFor(param, response);

      // were any errors found (these would be the failed indexes)
      if (response_err.length > 0) {
        this.emit(
          'error',
          TypeError(
            'Return data missing param ' +
              param +
              ': ' +
              response_err.join(', ')
          ),
          callback,
          response
        );

        // param was not found; data is invalid
        return [];
      }
    }

    // everything looks good
    return response;
  }

  /**
   * Checks each item for the given param
   *
   * If all goes well, the caller should expect that this method will return
   * an empty array.
   *
   * @param {string}         param    name of param
   * @param {Array.<Object>} response response data to scan
   *
   * @return {Array.<number>} indexes of failed response items, if any
   */
  private _checkAllFor(param: string, response: DataApiResult): number[] {
    let i = response.length;
    const err = [];

    while (i--) {
      if (response[i][param] === undefined) {
        // since we're looping in reverse, unshift instead of push
        err.unshift(i);
      }
    }

    return err;
  }
}
