/**
 * Adds static data to API response
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter} from 'events';
import {TokenId} from '../server/token/Token';
import {DapiAction, DataApi, DataApiInput, DataApiResult} from './DataApi';

/**
 * Prepends static data to a response
 */
export class StaticAdditionDataApi extends EventEmitter implements DataApi {
  /**
   * Prevents augmenting empty sets
   */
  private _nonempty = false;

  /**
   * Only append if more than 1 result
   */
  private _multiple = false;

  /**
   * Initialize static API response data
   *
   * @param data_api    - data API to wrap
   * @param nonempty    - to append if any results
   * @param multiple    - to append if more than 1 result
   * @param static_data - static data to prepend
   */
  constructor(
    private readonly _api: DataApi,
    nonempty = false,
    multiple = false,
    private readonly _static: DataApiResult[] = []
  ) {
    super();

    this._nonempty = !!nonempty;
    this._multiple = !!multiple;
  }

  /**
   * Request data from the service
   *
   * @param data     - request params
   * @param callback - server response callback
   * @param id       - the dapi id
   * @param action   - the action to perform for the dapi
   * @param token_id - a token associated with the dapi
   */
  request(
    data: DataApiInput | null,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    action?: DapiAction,
    token_id?: TokenId
  ): this {
    data = data || {};
    callback = callback || function () {};

    this._api.request(
      data,
      (err, response) => {
        // if the data are invalid, do nothing
        if (!Array.isArray(response)) {
          callback.call(this, err, response);
          return;
        }

        // return the response with our data
        callback.call(this, err, this._unshiftData(response));
      },
      id,
      action,
      token_id
    );

    return this;
  }

  /**
   * Unshifts the static data onto the given data set
   *
   * Be warned: for performance reasons, this alters the actual array that was
   * passed in (rather than returning a new one); if this is a problem, change
   * it (or add a flag to the ctor).
   *
   * @param data - data to augment
   *
   * @return augmented data
   */
  private _unshiftData(data: DataApiResult) {
    // if the nonempty flag is set, then we should not augment empty sets
    if (data.length === 0 && this._nonempty) {
      return data;
    }

    // if multiple flag is set but result contains < 2 results, do
    // not augment
    if (data.length < 2 && this._multiple) {
      return data;
    }

    // note that this modifies the actual reference!
    let i = this._static.length;
    while (i--) {
      data.unshift(this._static[i]);
    }

    return data;
  }
}
