/**
 * Facilitates webhooks by setting bucket values
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {TokenId} from '../server/token/Token';
import {Bucket, BucketData} from '../bucket/Bucket';
import {DapiAction, DataApi, DataApiInput, DataApiResult} from './DataApi';

/**
 * Set bucket values so that downstream systems are able to use these values for
 * triggering requests and responding to the quote server with webhooks
 */
export class WebhookBucketDataApi implements DataApi {
  /**
   * The key that will contain aggregated webhook info
   */
  public static WEBHOOK_AGG_KEY = 'webhook_trigger_aggregate';

  /**
   * Initialize Webhook Bucket Data API
   *
   * @param bucket - a bucket to save data to
   */
  constructor(private readonly _bucket: Bucket) {}

  /**
   * Delete any data that the client should have no knowledge of
   *
   * This performs a deletion of the data in-place
   *
   * @param data - the data to sanitize
   */
  public static sanitizeClientData(data: BucketData): void {
    const agg = data[WebhookBucketDataApi.WEBHOOK_AGG_KEY];

    if (!agg) {
      return;
    }

    for (const key of agg) {
      delete data[key + ''];
    }

    delete data[WebhookBucketDataApi.WEBHOOK_AGG_KEY];
  }

  /**
   * Request data from the service
   *
   * @param data     - request params
   * @param callback - server response callback
   * @param id       - unique dapi identifier
   * @param action   - the action for a dapi to take
   * @param token    - a token associated with a dapi call
   */
  public request(
    _data: DataApiInput,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    _action?: DapiAction,
    token?: TokenId
  ): this {
    if (!token) {
      throw new Error('Webhook bucket dapi triggered without a valid token');
    }

    // Set an aggregate so that downstream values do not have to loop through
    // the entire bucket
    const key = `webhook_trigger_${id}`;
    const agg =
      this._bucket.getDataByName(WebhookBucketDataApi.WEBHOOK_AGG_KEY) ?? [];

    if (!agg.includes(key)) {
      agg.push(key);
    }

    this._bucket.setValues({
      [WebhookBucketDataApi.WEBHOOK_AGG_KEY]: agg,
      [key]: [token],
    });

    // Indicate that there is no data to save but there was not an error either
    callback(null, null);

    return this;
  }
}
