/* TODO auto-generated eslint ignore, please fix! */
/* eslint no-unused-vars: "off" */
/**
 * Data transmission over HTTP(S)
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {DataApi, DataApiInput, DataApiResult} from '../DataApi';
import {HttpImpl} from './HttpImpl';

/**
 * HTTP request abstraction. Does minor validation, but delegates to a specific
 * HTTP implementation for the actual request.
 */
export class HttpDataApi implements DataApi {
  // RFC 2616 methods
  public static RFCMETHODS: string[] = [
    'DELETE',
    'GET',
    'HEAD',
    'OPTIONS',
    'POST',
    'PUT',
    'TRACE',
  ];

  /**
   * Initialize Data API with destination and HTTP implementation
   *
   * The supplied HTTP implementation will be used to perform the HTTP
   * requests, which permits the user to use whatever implementation works
   * well with their existing system.
   *
   * Default `enctype` is `application/x-www-form-urlencoded`.
   *
   * TODO: Accept URI encoder.
   *
   * @param url     - destination URL
   * @param method  - RFC-2616-compliant HTTP method
   * @param impl    - HTTP implementation
   * @param enctype - MIME media type
   *
   * @throws {TypeError} when non-HttpImpl is provided
   */
  constructor(
    private readonly _url: string,
    private readonly _method: string,
    private readonly _impl: HttpImpl,
    private _enctype: string = ''
  ) {
    if (this._enctype === '') {
      this._enctype = 'application/x-www-form-urlencoded';
    }

    this._validateMethod(_method);
  }

  /**
   * Perform an asynchronous request and invoke the callback with the reply
   *
   * DATA must be either a string or an object; the latter is treated as a
   * key-value parameter list, which will have each key and value
   * individually URI-encoded and converted into a string, delimited by
   * ampersands.  `null` may be used to indicate that no data should be
   * sent.
   *
   * In the event of an error, the first parameter is the error; otherwise, it
   * is null. The return data shall not be used in the event of an error.
   *
   * The return value shall be a raw string; conversion to other formats must
   * be handled by a wrapper.
   *
   * @param data - request params or post data
   *
   * @param callback - continuation upon reply
   *
   * @throws {TypeError} on validation failure
   */
  request(data: DataApiInput, callback: NodeCallback<DataApiResult>): this {
    // null is a good indicator of "I have no intent to send any data";
    // empty strings and objects are not, since those are valid data
    if (data === null) {
      data = {};
    }

    this._validateDataType(data);

    const url = this._url.replace(/{([^}]+)}/g, (match, field) => {
      const value = typeof data === 'object' ? data[field] : undefined;

      return !value || !value[0] || typeof value !== 'string' ? match : value;
    });

    this.requestData(
      url,
      this._method,
      data,
      (err: NullableError, resp: string | null) =>
        this._tryParse(err, resp, callback)
    );

    return this;
  }

  /**
   * Request data from underlying HttpImpl
   *
   * Subtypes may override this method to alter any aspect of the request
   * before sending.
   *
   * @param {string}                    url      destination URL
   * @param {string}                    method   RFC-2616-compliant HTTP method
   * @param {Object|string}             data     request params
   * @param {function(?Error, ?string)} callback server response callback
   *
   * @return {HttpDataApi} self
   */
  protected requestData(
    url: string,
    method: string,
    data: DataApiInput,
    callback: NodeCallback<string>
  ): HttpImpl {
    return this._impl.requestData(url, method, this.encodeData(data), callback);
  }

  /**
   * Ensures that the provided method conforms to RFC 2616
   *
   * @param {string} method HTTP method
   * @return {string} provided method
   *
   * @throws {Error} on non-conforming method
   */
  private _validateMethod(method: string) {
    if (!HttpDataApi.RFCMETHODS.includes(method)) {
      throw Error('Invalid RFC 2616 method: ' + method);
    }
  }

  /**
   * Validates that the provided data type is accepted by the Data API
   *
   * @param {*} data data to validate
   * @return {undefined}
   *
   * @throws {TypeError} on validation failure
   */
  private _validateDataType(data: DataApiInput) {
    const type = typeof data;

    if (!(type === 'string' || type === 'object')) {
      throw TypeError(
        'Data must be a string of raw data or object containing ' +
          'key-value params'
      );
    }
  }

  /**
   * Generate params for URI from key-value `data`
   *
   * Conversion depends on the MIME type (enctype) with which this instance
   * was initialized.  For example, `application/x-www-form-urlencoded`
   * will result in a urlencoded string, whereas `application/json` will
   * simply be serialized.
   *
   * If `data` is not an object, it will be returned as a string datum.
   *
   * @param data - key-value request params
   *
   * @return generated URI, or empty if no keys
   */
  protected encodeData(data: DataApiInput): string {
    if (typeof data !== 'object') {
      return '' + data;
    }

    if (this._method !== 'POST') {
      return this._urlEncode(data);
    }

    switch (this._enctype) {
      case 'application/x-www-form-urlencoded':
        return this._urlEncode(data);

      case 'application/json':
        return JSON.stringify(data);

      default:
        throw Error('Unknown enctype for POST: ' + this._enctype);
    }
  }

  /**
   * urlencode each key of provided object
   *
   * @param obj - key/value
   *
   * @return urlencoded string, joined with '&'
   */
  private _urlEncode(obj: Record<string, string | string[]>): string {
    // This cast should be removed. The underlying code is technically wrong
    // because this can be called with a string[] as the value but
    // encodeURIComponent only takes a string.
    const to_encode = <Record<string, string>>obj;

    return Object.keys(to_encode)
      .map(
        key =>
          encodeURIComponent(key) + '=' + encodeURIComponent(to_encode[key])
      )
      .join('&');
  }

  /**
   * Attempt to parse SRC as JSON and invoke callback according to the
   * rules of `#request`
   *
   * @param {?Error}             err      response error
   * @param {string}             src      JSON string
   * @param {function(?Error,*)} callback continuation
   *
   * @return {undefined}
   */
  private _tryParse(
    resp_err: NullableError,
    src: string | null,
    callback: NodeCallback<DataApiResult>
  ) {
    try {
      if (src === null) {
        throw new Error('Received null response');
      }

      const data = JSON.parse(src);

      callback(resp_err, data);
    } catch (e) {
      const parse_err =
        e instanceof Error ? e : new Error('Unknown JsonResponse Error format');

      // parsing failed; provide response text in addition to
      // original data so that the caller can handle how they
      // please
      callback(this._getReturnError(resp_err, parse_err), []);

      return;
    }
  }

  /**
   * Produce the parse error, or a combined error containing both the
   * original and parse errors
   *
   * @param {?Error} orig  response error
   * @param {Error}  parse parse error
   *
   * @return {Error} parse error or combined error
   */
  private _getReturnError(orig: NullableError, parse: Error) {
    if (!orig) {
      return parse;
    }

    const e = Error(
      'Multiple errors occurred; Original: "' +
        orig +
        '"; Parsed: "' +
        parse +
        '"'
    );

    return e;
  }
}
