/**
 *  Spoofed API assets
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {DocumentData} from '../server/db/MongoServerDao';
import {PositiveInteger, ts_ctor} from '../numeric';
import {UserSession} from '../server/request/UserSession';
import {UserRequest} from '../server/request/UserRequest';

/** Data required to spoof a session */
export interface ApiSessionTemplate {
  agentId: PositiveInteger;
  agentName: string;
  agentEntityId: PositiveInteger;
  agencyNumber: string;
  sessionId?: string;
}

/**
 * Extension of Session with some fluff for Dullahan
 */
export interface ApiSession extends UserSession {
  toDocumentData(): DocumentData;
  id(): string | null;
}

/**
 * Spoof a user request
 *
 * @param session - Session Interface
 *
 * @return spoofed user request
 **/
export const getApiRequest =
  (session: UserSession) =>
  (session_id: string): UserRequest => {
    const request = {
      getSession: () => session,
      getOrigin: () => 'dullahan',
      getSessionId: () => session_id,
      getSessionIdName: () => 'PHPSESSID',
      getUserAgent: () => '',
      getRemoteAddr: () => '',
      consumePostData: () => request,
      getGetData: () => ({}),
      end: () => request,
      setContentType: () => request,
      getUri: () => '',
      noCache: () => request,
      setResponseCode: (_code: number) => request,
      error: (_error: unknown, _data: unknown, _encoding?: string) => undefined,
      accepted: (_data?: unknown, _encoding?: string) => undefined,

      // EventEmitter stubs
      addListener: () => request,
      on: () => request,
      once: () => request,
      prependListener: () => request,
      prependOnceListener: () => request,
      removeListener: () => request,
      off: () => request,
      removeAllListeners: () => request,
      setMaxListeners: () => request,
      getMaxListeners: () => 0,
      listeners: () => [],
      rawListeners: () => [],
      emit: () => true,
      eventNames: () => [],
      listenerCount: () => 0,
      has_skey: () => false,
    };

    // This is a temporary hack. We should not be creating a whole mock object
    // to begin with
    return <UserRequest>(<unknown>request);
  };

/**
 * Determine if a session template contains proper fields
 *
 * @param template - hopefully an API session template
 *
 * @throws if any fields are not properly set
 */
export const validate_session_template = (template: CommonObject): void => {
  template = template ?? {};

  const missing = [
    'agentId',
    'agentName',
    'agentEntityId',
    'agencyNumber',
  ].filter(field => {
    return template[field] === undefined;
  });

  if (missing.length) {
    throw new Error(
      `Missing session data for the following fields: ${missing.join(', ')}`
    );
  }
};

/**
 * Get a spoofed delta to that triggers the delta processor
 */
export const getApiDelta = () => ({
  'rdelta.data': {
    data: {},
    concluding_save: false,
    timestamp: ts_ctor(),
    step_id: -1,
  },
});

/**
 * Delay for a set amount of time
 *
 * @param ms - milliseconds to delay
 */
export const delay = (ms: number) =>
  new Promise<void>(resolve => setTimeout(resolve, ms));
