/**
 * Dullahan Factory
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {BaseQuote} from '../quote/BaseQuote';
import {ClientInterface} from '../client/ClientInterface';
import {ClientQuote} from '../client/quote/ClientQuote';
import {CmatchVisibility} from '../client/CmatchVisibility';
import {Cmatch} from '../client/Cmatch';
import {DapiUpdate, DataApiMediator} from '../client/dapi/DataApiMediator';
import {DapiUpdateError} from './validate/DapiValidator';
import {DataApiManager} from '../../src/dapi/DataApiManager';
import {DataValidator} from '../validate/DataValidator';
import {DelayedStagingBucket} from '../bucket/DelayedStagingBucket';
import {DullahanClient} from './client/DullahanClient';
import {FieldClassMatcher} from '../field/FieldClassMatcher';
import {FieldResetter} from '../client/FieldResetter';
import {Program} from '../program/Program';
import {QuoteDataBucket} from '../bucket/QuoteDataBucket';
import {ServerSideQuote} from '../server/quote/ServerSideQuote';
import {StagingBucket} from '../bucket/StagingBucket';
import {StepDataSetter} from './step/StepDataSetter';
import {StepSimulator} from './step/StepSimulator';
import {Step} from '../step/Step';
import {UserRequest} from '../server/request/UserRequest';
import {Validator} from './validate/Validator';
import {multiSort} from '../sort/MultiSort';
import {defer} from '../util/defer';

/* eslint-disable @typescript-eslint/no-var-requires */
const clientQuote = require('../client/quote/ClientQuote');
const createDataApiManager = require('../dapi/DataApiManager');
const createDataApiFactory = require('../server/request/DullahanClientDataApiFactory');
const createClientDependencyFactory = require('../client/ClientDependencyFactory');
/* eslint-enable @typescript-eslint/no-var-requires */

/**
 * Dependencies needed for indication creation
 *
 * Note: Not fully complete. Many dependencies are handled separately from this
 * helper
 **/
export interface IndicationCreateDeps {
  createValidator: (
    program: Program,
    step_simulator: StepSimulator
  ) => Validator;

  createClientDapiManager: (request: UserRequest) => DataApiManager;

  createDapiMediator: (
    dapi_map: Record<string, Record<string, string>>,
    quotef: () => ClientQuote
  ) => DataApiMediator;

  createClientQuote: (
    quote: BaseQuote,
    staging_ctor?: (bucket: QuoteDataBucket) => StagingBucket
  ) => ClientQuote;

  createCmatch: (p: Program) => (q: ClientQuote) => Cmatch;

  createDataValidator: (p: Program) => DataValidator;

  createDapiOutputValidator: (
    bucket: DelayedStagingBucket
  ) => (program: Program) => (error_log: DapiUpdateError[]) => DapiUpdate;

  createStepSimulator: (
    quote: ServerSideQuote,
    program: Program,
    fields: string[]
  ) => StepSimulator;

  createStepDataSetter: (p: Program) => StepDataSetter;
}

export const createClientQuote = (
  quote: BaseQuote,
  staging_ctor?: (bucket: QuoteDataBucket) => StagingBucket
): ClientQuote =>
  clientQuote(
    quote,
    {
      data: quote.getBucket().getData(),
      meta: {
        created_by_username: [quote.getCreatedByUserName()],
        last_updated_by_username: [quote.getLastUpdatedByUserName()],
      },
      currentStepId: quote.getCurrentStepId(),
      topVisitedStepId: quote.getTopVisitedStepId(),
      agentId: quote.getAgentId(),
      agentName: quote.getAgentName(),
      agentEntityId: quote.getAgentEntityId(),
      startDate: quote.getStartDate(),
      initialRatedDate: quote.getInitialRatedDate(),
      lastPremDate: quote.getLastPremiumDate(),
      imported: quote.isImported(),
      bound: quote.isBound(),
      needsImport: quote.needsImport(),
      explicitLock: quote.getExplicitLockReason(),
      explicitLockStepId: quote.getExplicitLockStep(),
    },
    staging_ctor ??
      ((bucket: QuoteDataBucket): StagingBucket => new StagingBucket(bucket))
  );

export const createValidator = (
  program: Program,
  step_simulator: StepSimulator
) =>
  new Validator(
    program,
    (id: number, quote: ServerSideQuote) => new Step(id, quote, multiSort),
    step_simulator
  );

export const createCmatch = (program: Program) => (quote: ClientQuote) => {
  const client = createClient(program)(quote);

  return new Cmatch(
    createFieldClassMatcher(program.whens),
    program,
    client,
    createCmatchVisibility(client),
    createFieldResetter(client)
  );
};

export const createClientDapiManager =
  (base_url: string) => (request: UserRequest) =>
    createDataApiManager(createDataApiFactory(base_url, request));

export const createDapiMediator = (
  dapi_map: Record<string, Record<string, string>>,
  quotef: () => ClientQuote
) => new DataApiMediator(dapi_map, quotef);

const createClient = (program: Program) => (quote: ClientQuote) =>
  new DullahanClient(program, quote);

export const createDataValidator = (p: Program): DataValidator => {
  const client_dependency_factory = createClientDependencyFactory();

  return client_dependency_factory.createDataValidator(
    p.meta.qtypes,
    client_dependency_factory.createFieldValidator()
  );
};

const createFieldClassMatcher = (fields: Record<string, any>) =>
  new FieldClassMatcher(fields);

const createCmatchVisibility = (client: ClientInterface) =>
  new CmatchVisibility(client);

const createFieldResetter = (client: ClientInterface) =>
  new FieldResetter(client);

export const createStepDataSetter = () => new StepDataSetter(defer);

export const createStepSimulator = (
  quote: ServerSideQuote,
  program: Program,
  fields: string[]
) => new StepSimulator(quote, program, fields);
