/**
 *  Rate Poller
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {
  RateRequestResult,
  RatingService,
} from '../../server/service/RatingService';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {Either, isLeft} from 'fp-ts/lib/Either';
import {EventEmitter} from 'events';
import {RateResult} from '../../server/rater/Rater';

export const pollForRates =
  (
    emitter: EventEmitter,
    delay: (ms: number) => Promise<void>,
    rating_service: RatingService,
    getPollingSettings: (start_timestamp: number) => {
      timeout_ms: number;
      interval_ms: number;
    }
  ) =>
  async (
    start_timestamp: number,
    reloadQuote: () => Promise<Either<unknown, ServerSideQuote>>,
    rdata: RateResult
  ): Promise<RateResult> => {
    let timed_out = false;

    const {timeout_ms, interval_ms} = getPollingSettings(start_timestamp);

    emitter.emit('rate-polling-begin', {
      interval_ms,
      timeout_ms,
    });

    const checkForServerSideDapiResponses = async () => {
      if (timed_out) {
        emitter.emit('rating-timeout');
        return;
      }

      // Multiple saves were made in the background (e.g. DAPI calls) so we
      // should pull a fresh version of the quote from the database before
      // we send the response.
      const quote = await errorWrapper(reloadQuote);
      const context = {quote_id: quote.getId()};
      const fetch_rate_results = await rating_service.fetchRate(quote);

      rdata = fetch_rate_results.content.data;

      if (donePolling(fetch_rate_results)) {
        emitter.emit('rating-completed', context);
        return;
      }

      emitter.emit('pending-rate', {
        ...context,
        pending: fetch_rate_results.content.data.suppliers_pending_rate,
      });

      await delay(interval_ms);
      await checkForServerSideDapiResponses();
    };

    await Promise.race([
      checkForServerSideDapiResponses(),
      delay(timeout_ms).then(() => (timed_out = true)),
    ]);

    return rdata;
  };

/**
 * Determine whether to continue polling for rates
 *
 * This should really go away in favor of a more fully featured client event
 * handler for Dullahan that mimics the UI client event handler
 *
 * @param results - Fetch rate results
 *
 * @returns whether to continue polling
 */
const donePolling = (results: RateRequestResult): boolean =>
  results.actions[0]?.then?.action !== 'fetchRate';

/**
 * Wrap a TaskEither response
 *
 * The is so that fp-ts logic is abstracted away somewhat from the main logic
 *
 * @param fn - a TaskEither response
 *
 * @returns the result of a successful response or throws an error
 */
const errorWrapper = (fn: () => Promise<Either<unknown, ServerSideQuote>>) =>
  fn().then(quote => {
    if (isLeft(quote)) {
      throw new Error('Could not reload quote: ' + quote.left);
    }

    return quote.right;
  });

/**
 * Get polling settings
 *
 * We will timeout in headless after a set amount of timme, if we have reached
 * that time then return whatever we can and don't wait for any more rates
 **/
export const getPollingSettings =
  (ts_ctor: () => UnixTimestamp, api_timeout_ms: number, interval_ms: number) =>
  (start_timestamp: number): {timeout_ms: number; interval_ms: number} => {
    const elapsed_s = ts_ctor() - start_timestamp;
    const elapsed_ms = elapsed_s * 1000;
    const timeout_ms = api_timeout_ms - elapsed_ms;

    return {
      timeout_ms,
      interval_ms,
    };
  };
