/**
 *  Indication Controller
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {BucketData} from '../../bucket/Bucket';
import {AsyncQuoteFactory} from '../quote/QuoteFactory';
import {DapiUpdateError} from '../validate/DapiValidator';
import {DelayedStagingBucket} from '../../bucket/DelayedStagingBucket';
import {EventEmitter} from 'events';
import {FeatureFlag} from '../../system/flags/FeatureFlag';
import {HttpClient, InvalidUrlError} from '../../system/network/HttpClient';
import {IndicationCreateDeps} from '../DullahanFactory';
import {IndicationCreationManager} from '../indication/IndicationCreationManager';
import {ProgramFactory} from '../program/ProgramFactory';
import {RateResult} from '../../server/rater/Rater';
import {RatingService} from '../../server/service/RatingService';
import {Request, Response} from 'express';
import {Response as HttpResponse} from 'node-fetch';
import {InitExistingDocument} from '../../server/quote/loader';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {UserSession} from '../../server/request/UserSession';
import {ts_ctor} from '../../numeric';
import {
  generateValidationMessages,
  validateProgramChange,
} from '../../validate/ProgramChangeValidation';
import {getApiRequest, validate_session_template} from '../ApiTools';
import {
  asyncServerDao,
  MongoServerDao,
  ServerDaoSaveError,
} from '../../server/db/MongoServerDao';
import {Either} from 'fp-ts/lib/Either';
import {ServiceSecretMap} from '../../server/service/KekService';

export const indication = {
  /**
   * Create a new indication
   *
   * @param manager                - manager for indication "create" operations
   * @param emitter                - event emitter
   * @param http                   - HTTP client
   * @param flag                   - feature flag
   * @param createSession          - creates a session
   * @param creation_factory       - indication creation deps
   * @param program_factory        - used to generate programs on demand
   * @param quote_factory          - create quotes
   * @param dao                    - DAO for mongo
   * @param rating_service         - facilitates rating
   * @param request                - request object
   * @param response               - response object
   */
  create:
    (
      manager: IndicationCreationManager,
      emitter: EventEmitter,
      http: HttpClient,
      flag: FeatureFlag,
      createSession: (id: string) => Promise<UserSession>,
      creation_factory: IndicationCreateDeps,
      program_factory: ProgramFactory,
      quote_factory: AsyncQuoteFactory,
      dao: MongoServerDao,
      rating_service: RatingService,
      initExistingDocument: InitExistingDocument,
      pollForRates: (
        start_timestamp: number,
        reloadQuote: () => Promise<Either<unknown, ServerSideQuote>>,
        rdata: RateResult
      ) => Promise<RateResult>,
      gcp_file_path: string,
      keks: ServiceSecretMap
    ) =>
    async (request: Request, response: Response): Promise<void> => {
      let webhook = '';
      const {
        bucket: raw_bucket,
        session: request_session,
        metabucket: meta = {},
      } = request.body;

      const session_id = request_session.sessionId;
      const start_timestamp = ts_ctor();

      try {
        if (typeof request?.query?.callback === 'string') {
          webhook = decodeURIComponent(request.query.callback);
        }

        http.validateUrl(webhook);

        /**
         * Ideally the request session should contain a PHPSESSID. If it doesn't,
         * it means that DAPIs will not kick off properly. We want to break
         * gracefully in this scenario and still attempt to rate
         */
        validate_session_template(request_session);
      } catch (e) {
        let err_msg = 'Unknown error initializing indication';

        if (e instanceof Error) {
          err_msg = e.message;
        }

        const message =
          e instanceof InvalidUrlError
            ? 'Request must contain a valid webhook URL.'
            : err_msg;

        response.status(422).send({message});

        emitter.emit('error', e);
        return;
      }

      // Send acceptance back to originator
      response.status(202).send({message: 'Request received.'});

      emitter.emit('response-sent', {http_code: 202});

      const session = await createSession(session_id);

      // This isn't ideal but until we have a better solution for initializing
      // quotes with certain values on a program level we will have to do it.
      // Once we have a solution that removes the logic in QuoteInitializer in the
      // Connector repo, we can clean this up
      //
      // This should be resolved when DEV-10354 is completed
      if (session.isInternal()) {
        raw_bucket['publish_quote_ind'] = ['0'];
      }

      try {
        filterEmptyValues(raw_bucket);

        const client_dapi_manager = creation_factory
          .createClientDapiManager(getApiRequest(session)(session_id))
          // NB: Hooking this event prevents failures from crashing the daemon
          .on('error', (error, name) => {
            emitter.emit('dapi-failure', {
              error: error + '',
              name: name ?? 'unknown',
              type: 'client',
            });
          });

        const request_fields = Object.keys(raw_bucket ?? {});
        const {bucket, program} = program_factory.createProgram(
          client_dapi_manager,
          emitter,
          raw_bucket
        );

        const quote: ServerSideQuote = await quote_factory(dao)(
          manager.quoteMaker(bucket)
        )(program)(session)(flag)({
          agentId: session.agentId(),
          agentName: session.agentName(),
          agentEntityId: session.agentEntityId(),
          startDate: start_timestamp,
        })(keks);

        // if we're not internal, filter out the internal questions from the
        // quote bucket to ensure they don't cause required_field errors;
        if (!session.isInternal()) {
          const bucket_data = quote.getBucket().getData();
          for (const id in program.internal) {
            delete bucket_data[id];
          }
        }

        const quote_id = quote.getId();
        const staging_bucket = new DelayedStagingBucket(bucket);

        const client_quote = creation_factory.createClientQuote(
          quote,
          () => staging_bucket
        );
        const cmatch = creation_factory.createCmatch(program)(client_quote);
        cmatch.hookClassifier();
        quote.classify();
        client_quote.forceClassify();

        program.initQuote(bucket, true);

        const step_simulator = creation_factory.createStepSimulator(
          quote,
          program,
          request_fields
        );

        const validator = creation_factory.createValidator(
          program,
          step_simulator
        );

        // FIXME: this should be used when DEV-9728 is implemented
        const validation_messages = {};

        /*
          CAUTION: Now that the dullahan_client_dapi feature flag is removed from our code,
          we ideally shouldn't need to clear these client DAPI related fields from the bucket
          before running the client DAPIs. None of these fields need to be mapped in
          the bucket transformation because we can rely on the Client DAPI results to
          give them values. However, if we stop mapping these fields in the bucket transformation,
          indications with more than one location are likely to break.
          As part of invoking the quote_factory above, we rely on classifications to determine
          non-applicable fields, which we clear the values of. If fields which some of these classifications
          depend on, like state_verify, are cleared before this determination of non-applicable fields,
          then many values will be cleared for "non-applicable" fields, which later become
          applicable because of client dapi response data. This will result in required fields
          and assertion errors for locations other than the first location, even when this field
          data was provided by the user making the indication request, as well as a simple loss of
          data provided in an indication request. An example question which could lose its user
          provided data is latest_operation_hour. We must address this problem before
          unmapping in the bucket transformation and removing this ugly clearing below.
        */
        staging_bucket.overwriteValues({
          address_confirm: [''],
          crime_score: [''],
          crime_success: [''],
          location_success: [''],
          address_verify: [''],
          streetaddress_verify: [''],
          city_verify: [''],
          state_verify: [''],
          zipcode_verify: [''],
          business_state: [''],
          business_county: [''],
          google_county: [''],
          county_verify: [''],
          liza_start_date_location: [''],
          applicant_state: [''],
          applicant_county: [''],
          applicant_country: [''],
          headquarters_state: [''],
          prop_interest_state: [''],
        });
        await manager.initialQuoteSave(quote)(dao)(program)(bucket)(meta)();
        emitter.emit('quote-save-success', {quote_id});

        const data_validator = creation_factory.createDataValidator(program);
        const step_setter = creation_factory.createStepDataSetter(program);

        const validation_callback = (
          vdiff: CommonObject,
          failures: CommonObject
        ) => {
          validateProgramChange(program)(cmatch)(
            staging_bucket,
            vdiff,
            failures,
            step_setter.getCurrentStep.bind(step_setter),
            // FIXME: this should be implemented during DEV-9728
            () => {},
            generateValidationMessages,
            validation_messages
          );
        };

        client_quote.on('preDataUpdate', diff => {
          data_validator
            .validate(diff, undefined, validation_callback)
            .catch(e => {
              throw e;
            });
        });

        const dapi_mediator = creation_factory.createDapiMediator(
          program.dapimap ?? {},
          () => client_quote
        );

        const dapi_validation_failures: DapiUpdateError[] = [];

        dapi_mediator.monitor(
          client_dapi_manager,
          creation_factory.createDapiOutputValidator(staging_bucket)(program)(
            dapi_validation_failures
          )
        );

        const {
          failed_assertions = [],
          missing_required = [],
          triggered_events = [],
        } = await step_setter.setDataByStep(
          bucket,
          staging_bucket,
          quote,
          step_simulator,
          client_dapi_manager,
          validator,
          cmatch,
          () => <UnixTimestamp>0
        );

        /*
        Some agents assigned to a quote via agent_email will not
        yet be registered as agents in the Connector. If they do register,
        we have a webhook which finds all quotes assigned to them and
        publishes these quotes to their agency. As part of this publishing,
        we trigger a data delta so that downstream systems like risk
        search pick up the agency change. Data deltas are not processed
        by the delta processor if the current step is ahead of the top
        saved step. This is to prevent unvalidated data from polluting our
        downstream systems.
        For indications, we don't set the current step and it defaults to
        the first step. Setting the top saved step here as the first step ID
        ensures that the data delta triggered in the publish gets forwarded
        on by the delta processor. Also, the fact that it's only the first step
        mitigates the problems prevented by the delta processor's step check.
        If a user can't go back to a previous step, there shouldn't be a
        scenario where unvalidated data can get through on a step save.

        The check for agent_email serves as a feature flag and can be removed after testing
      */
        if (quote.getDataByName('agent_email')[0] !== '') {
          quote.setTopSavedStepId(program.getFirstStepId());
          await asyncServerDao(dao).saveQuoteState(quote);
        }

        if (dapi_validation_failures.length) {
          emitter.emit('quote-validation-warning', {
            quote_id,
            dapi_validation_failures,
          });
          throw new Error('Client dapi validation failure');
        }

        await manager.initialQuoteSave(quote)(dao)(program)(bucket)(meta)();
        emitter.emit('quote-save-success', {quote_id});

        if (failed_assertions.length > 0 || missing_required.length > 0) {
          emitter.emit('quote-validation-failure', {
            quote_id,
            missing_required,
            failed_assertions,
          });

          await sendResponse(emitter)(http)(webhook, {
            bucket: quote.getBucket().getData(),
            rdata: {},
            validation_errors: [...missing_required, ...failed_assertions],
            sso: session.isInternal(),
          });

          return;
        }

        await manager.invokeServerDapis(program)(quote)(session)(session_id)(
          bucket
        )(emitter)(dao)(gcp_file_path);

        let rdata = await rating_service.request(session, quote);

        if (triggered_events.length > 0) {
          emitter.emit('quote-validation-warning', {
            quote_id,
            triggered_events,
          });
        }

        await asyncServerDao(dao).saveQuote(quote, undefined, {});
        // This is purposely called without the meta argument so it updates the
        // DAPI values that are in memory.
        await asyncServerDao(dao).saveQuoteMetaDirect(quote, undefined, true);
        await dao.publishQuote(quote, true, true);

        rdata = await pollForRates(
          start_timestamp,
          initExistingDocument(dao)(program)(session)(flag)(quote)({
            startDate: quote.getStartDate(),
          })(keks),
          rdata
        );

        emitter.emit('callback-sent', {target: webhook});

        await sendResponse(emitter)(http)(webhook, {
          bucket: quote.getBucket().getData(),
          rdata,
          validation_errors: [...missing_required, ...failed_assertions],
          sso: session.isInternal(),
        });
      } catch (e) {
        if (e instanceof ServerDaoSaveError) {
          emitter.emit('quote-save-failure', {quote_id: e.id});
        } else {
          emitter.emit('error', e);
        }

        let error = 'Unknown indication error';

        if (e instanceof Error) {
          error = e.toString();
        }

        await sendResponse(emitter)(http)(webhook, {error});
      }
    },
};

/**
 * Remove keys that do not contain any data
 *
 * We want to remove any keys from the bucket that have empty data so that when
 * we initialize defaults it is not hindered by the presence of an empty array
 *
 * @param data - data to filter
 */
const filterEmptyValues = (data: CommonObject) => {
  for (const key in data) {
    const val = data[key];

    if (!Array.isArray(val) || val.length === 0) {
      delete data[key];
    }
  }
};

/**
 * Handle event emission for a webhook response
 *
 * @param emitter  - event emitter
 * @param response - webhook response
 */
const handleWebhookResponse =
  (emitter: EventEmitter) =>
  (response: HttpResponse, is_error_callback = false) => {
    if (!response.ok) {
      emitter.emit('callback-failure', {
        is_error_callback,
        error: response.status,
      });
    } else {
      emitter.emit('callback-success', {
        is_error_callback,
        response,
      });
    }
  };

// All of the request/response logic should really be moved into its own class
// this one is doing far too much
declare type WebhookErrorResponse = {
  error: string;
};

declare type WebhookFinishedResponse = {
  bucket: BucketData;
  rdata: RateResult;
  validation_errors: string[];
  sso: boolean;
};

declare type WebhookResponse = WebhookErrorResponse | WebhookFinishedResponse;

/**
 * Send a webhook callback when an error has occurred
 *
 * @param emitter - event emitter
 * @param http    - http client
 * @param url     - callback url
 * @param payload - http payload
 * @returns
 */
const sendResponse =
  (emitter: EventEmitter) =>
  (http: HttpClient) =>
  async (url: string, payload: WebhookResponse) => {
    try {
      handleWebhookResponse(emitter)(await http.post(url, payload), true);
    } catch (error) {
      emitter.emit('error', error);
    }
  };
