/**
 *  Service Controller
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter} from 'events';
import {Request, Response} from 'express';
import CopyRequestAssertionError from '../../error/CopyRequestAssertionError';
import InitRequestAssertionError from '../../error/InitRequestAssertionError';
import InitRateRequestAssertionError from '../../error/InitRateRequestAssertionError';
import {assertCopyRequest} from '../validate/copy_validator';
import {assertInitRequest} from '../validate/init_validator';
import {assertInitRateRequest} from '../validate/init_rate_validator';
import {createUserSession} from '../../server/ServerFactory';
import SessionError from '../../error/SessionError';
import {
  programCopyCleanerLoader,
  programLoader,
} from '../../program/ProgramLoader';
import {initNewDocument, initExistingDocument} from '../../server/quote/loader';
import {DocumentData, MongoServerDao} from '../../server/db/MongoServerDao';
import {FeatureFlag} from '../../system/flags/FeatureFlag';
import {isLeft} from 'fp-ts/Either';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {
  ClassificationResult,
  Program,
  QuestionTypes,
} from '../../program/Program';
import {PositiveInteger, ts_ctor} from '../../numeric';
import {createQuoteQuick} from '../../server/quote/QuoteFactory';
import {ServiceSecretMap} from '../../server/service/KekService';
import {UserSession} from '../../server/request/UserSession';
import {BucketDataValidator} from '../../../src/validate/BucketDataValidator';

type EventContext = Record<string, string | PositiveInteger>;

export class ServiceController {
  constructor(
    private readonly createDataValidator: (
      q_types: QuestionTypes
    ) => BucketDataValidator,
    private readonly emitter: EventEmitter,
    private readonly createSession: typeof createUserSession,
    private readonly progLoader: typeof programLoader,
    private readonly progCopyCleanerLoader: typeof programCopyCleanerLoader,
    private readonly createNewDocument: typeof initNewDocument,
    private readonly dao: MongoServerDao,
    private readonly flag: FeatureFlag,
    private readonly keks: ServiceSecretMap
  ) {}
  /**
   * Copy a submission
   */
  public copy =
    (createSourceQuote: typeof initExistingDocument) =>
    async (request: Request, response: Response): Promise<void> => {
      let event_context: EventContext = {};

      try {
        const session = await this.getSession(request.body.session_id);
        event_context = this.createContext(request, session);
        this.emitter.emit('copy-request-received', event_context);

        const program = this.progLoader(request.body.program)();
        assertCopyRequest(request.body);

        const quoteCreator = (id: number): ServerSideQuote =>
          createQuoteQuick(id, program);

        const quote = quoteCreator(request.body.quote_id);

        const source_quote_either = await createSourceQuote(this.dao)(program)(
          session
        )(this.flag)(quote)(<DocumentData>{})(this.keks)();
        if (isLeft(source_quote_either)) {
          throw new Error(`${source_quote_either.left}`);
        }
        const source_quote = source_quote_either.right;

        if (!source_quote.isAuthorized(session)) {
          throw new Error('Not authorized to use quote');
        }

        if (request.body.program !== source_quote.getProgramId()) {
          throw new CopyRequestAssertionError(
            'Cannot copy quote across programs'
          );
        }

        await this.mergeAndSave(
          request,
          response,
          event_context,
          source_quote,
          session,
          program,
          'copy'
        );

        return;
      } catch (e) {
        let err: Error;
        if (e instanceof CopyRequestAssertionError) {
          response.status(422);
          err = e;
        } else if (e instanceof SessionError) {
          response.status(401);
          err = e;
        } else if (e instanceof Error) {
          switch (e.message) {
            case 'Quote not found':
              response.status(404);
              break;
            case 'Not authorized to use quote':
              response.status(403);
              break;
            default:
              response.status(500);
          }
          err = e;
        } else {
          err = new Error(`${e}`);
          response.status(500);
        }

        response.json(err.message);
        this.emitter.emit('error', err);
        this.emitter.emit('copy-request-unsuccessful', event_context);

        return;
      }
    };

  /**
   * Init a submission with the posted data
   */
  public init = async (request: Request, response: Response): Promise<void> => {
    // If we have a quote ID, we can just copy the quote, otherwise we just
    // want to bootstrap a new one with the data posted.
    if (request.body.quote_id) {
      // The 307 response code will preserve the HTTP method from the
      // original request. https://http.cat/status/307
      response.redirect(307, '/copy');
      return;
    }
    let event_context: EventContext = {};

    try {
      const session = await this.getSession(request.body.session_id);
      event_context = this.createContext(request, session);
      this.emitter.emit('init-request-received', event_context);

      assertInitRequest(request.body);

      const source_quote = await this.initNewQuote(
        request.body.program,
        session
      );
      const program = source_quote.getProgram();

      await this.mergeAndSave(
        request,
        response,
        event_context,
        source_quote,
        session,
        program,
        'init'
      );

      return;
    } catch (e) {
      let err: Error;
      if (e instanceof InitRequestAssertionError) {
        response.status(422);
        err = e;
      } else if (e instanceof SessionError) {
        response.status(401);
        err = e;
      } else if (e instanceof Error) {
        response.status(500);
        err = e;
      } else {
        err = new Error(`${e}`);
        response.status(500);
      }

      response.json(err.message);

      this.emitter.emit('error', err);
      this.emitter.emit('init-request-unsuccessful', event_context);

      return;
    }
  };

  /**
   * Init a submission and rate it with the posted data
   * TODO: REMOVE STATIC DEFINITION ON RELEASE OF DEV-16997
   */
  public initRate =
    (createSourceQuote: typeof initExistingDocument) =>
    async (request: Request, response: Response): Promise<void> => {
      let event_context: EventContext = {};

      try {
        const session = await this.getSession(request.body.session_id);
        event_context = this.createContext(request, session);
        this.emitter.emit('initRate-request-received', event_context);

        const program = this.progLoader(request.body.program)();
        assertInitRateRequest(request.body);

        const quoteCreator = (id: number): ServerSideQuote =>
          createQuoteQuick(id, program);

        const quote = quoteCreator(request.body.quote_id);

        const source_quote_either = await createSourceQuote(this.dao)(program)(
          session
        )(this.flag)(quote)(<DocumentData>{})(this.keks)();
        if (isLeft(source_quote_either)) {
          throw new Error(`${source_quote_either.left}`);
        }
        const source_quote = source_quote_either.right;

        if (!source_quote.isAuthorized(session)) {
          throw new Error('Not authorized to use quote');
        }

        if (request.body.program !== source_quote.getProgramId()) {
          throw new InitRateRequestAssertionError(
            'Cannot initRate quote across programs'
          );
        }

        await this.mergeAndSave(
          request,
          response,
          event_context,
          source_quote,
          session,
          program,
          'initRate'
        );

        return;
      } catch (e) {
        let err: Error;
        if (e instanceof InitRateRequestAssertionError) {
          response.status(422);
          err = e;
        } else if (e instanceof SessionError) {
          response.status(401);
          err = e;
        } else if (e instanceof Error) {
          switch (e.message) {
            case 'Quote not found':
              response.status(404);
              break;
            case 'Not authorized to use quote':
              response.status(403);
              break;
            default:
              response.status(500);
          }
          err = e;
        } else {
          err = new Error(`${e}`);
          response.status(500);
        }

        response.json(err.message);
        this.emitter.emit('error', err);
        this.emitter.emit('initRate-request-unsuccessful', event_context);

        return;
      }
    };

  private mergeAndSave = async (
    request: Request,
    response: Response,
    event_context: EventContext,
    source_quote: ServerSideQuote,
    session: UserSession,
    program: Program,
    type: 'init' | 'copy' | 'initRate'
  ) => {
    const cleanBucket = this.progCopyCleanerLoader(request.body.program);
    const target_quote = await this.initNewQuote(request.body.program, session);

    event_context.target_quote_id = target_quote.getId();
    const bucket = target_quote.getBucket();
    const bucket_data = source_quote.getBucket().getData();

    // Feature flagging in this way works because the copy cleaner
    // won't affect select values if the select value metadata is empty
    const select_question_metadata = (await this.flag.isEnabled(
      'validate_select_values'
    ))
      ? {
          static_select_questions: program.meta.staticSelectQuestions,
          valid_select_values: program.meta.qdata,
          question_defaults: program.defaults,
        }
      : {
          static_select_questions: {},
          valid_select_values: {},
          question_defaults: {},
        };

    const clean_bucket_data = cleanBucket(
      bucket_data,
      session,
      source_quote.getAgencyNumber(),
      request.body.bucket,
      select_question_metadata
    );

    const data_validator = this.createDataValidator(program.meta.qtypes);
    const combined_bucket = data_validator.format({
      ...clean_bucket_data,
      ...request.body.bucket,
    });
    const valid_data = data_validator.validate(
      combined_bucket,
      this.validateCallback(combined_bucket)
    );

    bucket.clear();
    // Set new values for aligned bucket here
    bucket.setValues(valid_data);

    this.saveQuote(
      response,
      target_quote,
      session,
      request.body.program,
      request.body.origin,
      request.body.policy_id,
      event_context,
      type,
      type === 'copy' ? request.body.quote_id : undefined
    );
  };

  private createContext = (request: Request, session: UserSession) => ({
    source_quote_id: request.body.quote_id,
    policy_id: request.body.policy_id,
    program: request.body.program,
    user_id: (session.userName && session.userName()) ?? '',
    agent_code: (session.agencyNumber && session.agencyNumber()) ?? '',
  });

  private getSession = async (session_id: string): Promise<UserSession> => {
    const session = await this.createSession(session_id);
    if (!session.isLoggedIn()) {
      throw new SessionError('User is not logged in');
    }

    return session;
  };

  private initNewQuote = async (
    program_name: string,
    session: UserSession
  ): Promise<ServerSideQuote> => {
    const program = this.progLoader(program_name)();
    const quoteCreator = (id: number): ServerSideQuote =>
      createQuoteQuick(id, program);

    const target_quote_either = await this.createNewDocument(this.dao)(
      quoteCreator
    )(program)(session)(this.keks)(this.flag)(<DocumentData>{})();
    if (isLeft(target_quote_either)) {
      throw new Error('Failed to create inited quote');
    }

    return target_quote_either.right;
  };

  /**
   * Save a quote that was created
   *
   * Note: this also handles the HTTP response if the quote is created successfully.
   */
  private saveQuote = (
    response: Response,
    quote: ServerSideQuote,
    session: UserSession,
    program_name: string,
    origin: string,
    policy_id: string,
    event_context: EventContext,
    quote_created_type: 'copy' | 'init' | 'initRate',
    quote_id?: PositiveInteger
  ): void => {
    this.dao.saveQuote(
      quote,
      <ClassificationResult>{},
      () => {
        this.dao.saveQuoteMetaDirect(
          quote,
          {
            origin: [origin],
            retail_agency: [session.agencyNumber() ?? ''],
            liza_access_email: [session.userName() ?? ''],
            created_by_username: [session.userName() ?? ''],
            created_by_roles: session.userRoles(),
            last_updated_by_username: [session.userName() ?? ''],
            last_updated_by_roles: session.userRoles(),
            liza_access_groups: session.isInternal()
              ? ['']
              : [session.agencyNumber() ?? ''],
          },
          () => {
            response.status(200).json({id: quote.getId()});
            this.emitter.emit(
              `${quote_created_type}-request-successful`,
              event_context
            );
          },
          () => {
            throw new Error(
              `Failed to save ${quote_created_type} quote metadata`
            );
          }
        );
      },
      () => {
        throw new Error(`Failed to save ${quote_created_type} quote`);
      },
      {
        agentEntityId: session.agentEntityId(),
        programId: program_name,
        copyFromQuote: quote_id,
        copyFromPolicy: policy_id,
        currentStepId: 0,
        published: false,
        importedInd: 0,
        notifyInd: 0,
        startDate: ts_ctor(),
        syncDate: '',
        syncInd: 0,
        topVisitedStepId: 0,
        data: quote.getBucket().getData(),
        rdelta: {
          data: [
            {
              data: {},
              concluding_save: false,
              timestamp: ts_ctor(),
              step_id: -1,
            },
          ],
        },
      },
      {},
      !session.isInternal()
    );
  };

  private validateCallback =
    (bucket: Record<string, any>) =>
    (name: string, value: string, i: string, error: Error) => {
      this.emitter.emit('quote-validation-warning', {
        field: `${name}[${i}]`,
        value,
        error,
      });
      // Reset the value to an empty string so invalid data is not in the bucket.
      bucket[name][+i] = '';
    };
}
