/**
 * Generate pending DAPIs
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter} from 'events';
import {Collection} from 'mongodb';
import {getApiRequest} from '../ApiTools';
import {Bucket, BucketData, EmptyBucketData} from '../../bucket/Bucket';
import {DataProcessor} from '../../server/request/DataProcessor';
import {Program} from '../../program/Program';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {Store} from '../../store/Store';
import {bucketClearer} from '../../bucket/BucketClear';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {StagingBucket} from '../../bucket/StagingBucket';
import {createDapiManager} from '../../server/dapi/ServerDataApiManagerFactory';
import {filter as bucket_filter} from '../../bucket/bucket_filter';
import {ts_ctor} from '../../numeric';
import {DapiMetaSource, FieldData} from '../../server/meta/DapiMetaSource';
import {UserSession} from '../../server/request/UserSession';
import {MongoTokenDao} from '../../server/token/MongoTokenDao';
import {PriorityLog} from '../../server/log/PriorityLog';

/**
 * Used to generate DAPIs
 *
 * @param _base_url   - URL prefix for DAPI requests
 * @param _timeout    - max time to wait for DAPIs
 * @param _config     - config used by the DAPI manager
 * @param _collection - DB used by the DAPI manager
 * @param _emitter    - event emitter
 */
export class DapiFactory {
  constructor(
    private _base_url: string,
    private _timeout: number,
    private _config: Store,
    private _collection: Collection,
    private _emitter: EventEmitter
  ) {}

  /**
   * Get the DAPIs and meta fields to clear
   *
   * @param program - compiled rating program
   * @param logger  - for making log entries
   * @param quote   - related quote
   * @param session - user's sessions
   * @param bucket  - raw bucket data
   *
   * @return pending dapis and meta fields to clear
   */
  public getDapis(program: Program) {
    return (logger: PriorityLog) =>
      (quote: ServerSideQuote) =>
      (session: UserSession) =>
      (session_id: string) =>
      (bucket: BucketData) =>
      (
        gcp_file_path: string
      ): {dapis: Promise<FieldData>[]; meta_update: EmptyBucketData} => {
        const dapi_manager = createDapiManager(this._base_url)(this._config)(
          new MongoTokenDao(this._collection, 'dapitok', ts_ctor)
        )(gcp_file_path);

        const data_processor = new DataProcessor(
          logger,
          bucket_filter,
          dapi_manager,
          new DapiMetaSource(() => new QuoteDataBucket()),
          (bucket: Bucket) => new StagingBucket(bucket),
          bucketClearer('meta.')
        );

        const on_processing_error = (error: unknown, name?: string) => {
          this._emitter.emit('dapi-failure', {
            quote_id: quote.getId(),
            error,
            name: name ?? 'unknown',
            type: 'server',
          });
        };

        const {dapis, meta_update} = data_processor.processDiff(
          bucket,
          getApiRequest(session)(session_id),
          program,
          new QuoteDataBucket(),
          quote,
          undefined,
          on_processing_error
        );

        return {
          dapis: dapis.map(this._timeoutDapis(this._timeout)),
          meta_update,
        };
      };
  }

  /**
   * Wrap a DAPI in a timeout
   *
   * @param timeout - timeout in ms
   * @param dapi    - pending dapi
   *
   * @return timeout-wrapped DAPIs
   */
  private _timeoutDapis(timeout: number) {
    return (dapi: Promise<FieldData | void>) =>
      new Promise<FieldData>((resolve, reject) => {
        const timer = setTimeout(() => {
          clear(reject)(`DAPI took longer than ${timeout} ms to return.`);
        }, timeout);

        /** Automatically clear the timeout when fulfilling a promise */
        const clear =
          (callback: (payload: any) => void) => (payload: unknown) => {
            clearTimeout(timer);
            callback(payload);
          };

        dapi
          .then(result => {
            if (!result) {
              reject();
            }

            return result;
          })
          .then(clear(resolve))
          .catch(e => {
            clear(reject)(e.message ?? 'DAPI failed');
          });
      });
  }
}
