/**
 *  Indication Creation Manager
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Bucket, BucketData} from '../../bucket/Bucket';
import {DapiFactory} from '../dapi/DapiFactory';
import {MissingParamError} from '../../error/MissingParamError';
import {EventEmitter} from 'events';
import {Program} from '../../program/Program';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {UserSession} from '../../server/request/UserSession';
import {asyncServerDao, MongoServerDao} from '../../server/db/MongoServerDao';
import {FieldData} from '../../server/meta/DapiMetaSource';
import {SavedDelta} from '../../bucket/delta';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {PriorityLog} from '../../server/log/PriorityLog';
import {createQuoteQuick} from '../../server/quote/QuoteFactory';

/* eslint-disable @typescript-eslint/no-var-requires */
const {Class} = require('easejs');
/* eslint-enable @typescript-eslint/no-var-requires */

export class IndicationCreationManager {
  constructor(
    private readonly _logger: PriorityLog,
    private readonly _dapi_factory: DapiFactory,
    private readonly _constructSaveData: (
      quote: ServerSideQuote,
      delta: BucketData
    ) => {save_data: BucketData; rdelta: SavedDelta<BucketData>}
  ) {}
  /**
   * Used to generate quotes
   *
   * @param bucket  - bucket
   * @param program - program
   * @param id      - quote id
   *
   * @return quote
   */
  public quoteMaker(bucket: QuoteDataBucket) {
    return (program: Program) =>
      (id: number): ServerSideQuote =>
        createQuoteQuick(id, program, bucket);
  }

  /**
   * Save the quote and meta once all resources are established
   */
  public initialQuoteSave(quote: ServerSideQuote) {
    return (dao: MongoServerDao) =>
      (program: Program) =>
      (bucket: Bucket) =>
      (meta: Bucket) =>
      async (): Promise<ServerSideQuote> => {
        const data = {
          programId: program.getId(),
          data: {...bucket.getData()},
        };

        await asyncServerDao(dao).saveQuote(quote, data);
        return asyncServerDao(dao).saveQuoteMetaDirect(quote, meta);
      };
  }

  /**
   * Helper that does all DAPI operations
   */
  public invokeServerDapis(program: Program) {
    return (quote: ServerSideQuote) =>
      (session: UserSession) =>
      (session_id: string) =>
      (bucket: Bucket) =>
      (emitter: EventEmitter) =>
      (dao: MongoServerDao) =>
      async (gcp_file_path: string): Promise<void> => {
        const onDapiError = (x: unknown) => {
          // This is an error that is an expected part of the server side dapi
          // workflow. It is triggered when a required field is not present in
          // the bucket. That does not necessarily constitute an error because
          // we may simply not be using that dapi in this request.
          if (Class.isA(MissingParamError, x)) {
            return;
          }

          emitter.emit('dapi-failure', {
            quote_id: quote.getId(),
            error: x,
            type: 'server',
          });
        };

        const onDapiSuccess = async (x: FieldData) => {
          if (x.data) {
            quote.setMetadata(x.data);
          }

          const delta = x.delta;

          if (!delta) {
            return;
          }

          const {save_data, rdelta} = this._constructSaveData(quote, delta);

          asyncServerDao(dao).saveQuote(quote, save_data, {
            'rdelta.data': rdelta,
          });
        };

        const {dapis, meta_update} = this._dapi_factory.getDapis(program)(
          this._logger
        )(quote)(session)(session_id)({
          ...bucket.getData(),
        })(gcp_file_path);

        // Remove the "meta:" prefix from these fields before setting them on the
        // quote in memory. This is a little hacky and isn't necessary in the
        // quote server because this is used to save the fields to the database
        // and is then loaded automatically into the metabucket on the next
        // request
        const stripped_meta_update = Object.fromEntries(
          Object.entries(meta_update).map(([k, v]) => [
            k.replace('meta.', ''),
            v,
          ])
        );

        quote.setMetadata(stripped_meta_update);

        await Promise.allSettled([
          asyncServerDao(dao).mergeData(quote, meta_update),
          ...dapis.map(x =>
            x.then(onDapiSuccess).catch(err => onDapiError(err))
          ),
        ]);
      };
  }
}
