/**
 *  ProgramFactory
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {BucketDataValidator} from '../../validate/BucketDataValidator';
import {DataApiManager} from '../../dapi/DataApiManager';
import {Program, QuestionTypes} from '../../program/Program';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {EventEmitter} from 'events';
import {applyBucketDefaults} from '../../server/quote/loader';

type BucketData = Record<string, string[]>;

/**
 * Generates a new program on demand
 */
export class ProgramFactory {
  // Todo: DEV-16587 Allow for encryption here
  private fakeEncryptBucketValue = (x: string) => x;

  /**
   * Create a new ProgramFactory
   *
   * @param programMaker    - creates a program
   * @param bucketMaker     - creates a bucket
   * @param createFormatter - create a data formatter
   */
  constructor(
    private readonly programMaker: (dapi_manager: DataApiManager) => Program,
    private readonly bucketMaker: () => QuoteDataBucket,
    private readonly createFormatter: (
      types: QuestionTypes
    ) => BucketDataValidator
  ) {}

  /**
   * Create a new program
   *
   * @param bucket_data - base bucket data
   *
   * @return a program and bucket
   */
  public createProgram(
    dapi_manager: DataApiManager,
    emitter: EventEmitter,
    bucket_data: BucketData = {}
  ): {bucket: QuoteDataBucket; program: Program} {
    const program = this.programMaker(dapi_manager);
    const formatter = this.createFormatter(program.meta.qtypes);

    const validated = formatter.validate(
      bucket_data,
      (name: string, value: string, i: string, error: Error) => {
        emitter.emit('quote-validation-warning', {
          field: `${name}[${i}]`,
          value,
          error,
        });
        // Reset the value to an empty string so invalid data is not in the bucket.
        bucket_data[name][+i] = '';
      }
    );

    const bucket = this.bucketMaker().setValues(
      applyBucketDefaults(program)(validated)(this.fakeEncryptBucketValue)()
    );

    return {bucket, program};
  }
}
