import {Cmatch} from '../../client/Cmatch';
import {DataApiManager} from '../../dapi/DataApiManager';
import {DelayedStagingBucket} from '../../bucket/DelayedStagingBucket';
import {PositiveInteger} from '../../numeric';
import {Program} from '../../program/Program';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {StepSimulator, StepSummary} from './StepSimulator';
import {ValidationResults, Validator} from '../validate/Validator';

/** Format used to conduct an update per step */
interface StepUpdateData {
  id: PositiveInteger;
  update: CommonObject;
}

export type StepDataSetterFactory = (p: Program) => StepDataSetter;

/**
 * Sets bucket data per step
 *
 * This class uses a StepSimulator to know all applicable steps for a given data
 * source and then sets the data for each step on a DelayedStagingBucket.
 *
 * Additionally this class hooks and responds when updates are made to the
 * DelayedStagingBucket
 */
export class StepDataSetter {
  /**
   * Instantiate a new step data setter
   */
  constructor(
    private readonly _defer: (timeout_ms?: number) => Promise<void>
  ) {}

  /* Keep track of current step so external parties can check the progress */
  private _current_step: PositiveInteger = <PositiveInteger>0;

  public getCurrentStep() {
    return this._current_step;
  }

  /**
   * Populate the staging bucket and handle client DAPIs
   *
   * @param source         - source bucket where data come from
   * @param bucket         - staging bucket built up per step
   * @param quote          - server-side quote
   * @param step_simulator - helper to simulate each step
   * @param dapi_manager   - data API manager
   * @param validator      - validate field and step changes
   * @param cmatch         - class match
   * @param ts_ctor        - gets current timestamp
   */
  public async setDataByStep(
    source: QuoteDataBucket,
    bucket: DelayedStagingBucket,
    quote: ServerSideQuote,
    step_simulator: StepSimulator,
    dapi_manager: DataApiManager,
    validator: Validator,
    cmatch: Cmatch,
    ts_ctor: () => UnixTimestamp
  ) {
    let results: ValidationResults | null = null;

    const step_updates = this.getStepUpdates(
      await step_simulator.getSteps(cmatch),
      source,
      cmatch
    );

    for (const update of step_updates) {
      await this._performUpdate(bucket, update, cmatch, validator);
      await this._pendingDapis(dapi_manager, ts_ctor);

      results = validator.validateStep(source, quote, cmatch, update.id);

      bucket.commit();
    }

    return (
      results || {
        failed_assertions: [],
        missing_required: [],
        triggered_events: [],
      }
    );
  }

  /**
   * Generate the data for setting each step on demand
   *
   * @param steps  - a summary of each step
   * @param source - data source for setting step data
   * @param cmatch - class match
   *
   * @return step updates
   */
  private getStepUpdates(
    steps: StepSummary[],
    source: QuoteDataBucket,
    cmatch: Cmatch
  ): StepUpdateData[] {
    return steps.map(step => {
      const {id, fields} = step;

      const update = fields
        // we only want data for applicable fields for this step
        .filter(field => cmatch.getMatches()[field]?.any ?? false)
        // get values for the keys from the raw request bucket
        .reduce((carry: CommonObject, field: string) => {
          carry[field] = source.getDataByName(field) ?? [];

          return carry;
        }, {});

      return {id, update};
    });
  }

  /**
   * Perform an update for a step
   *
   * @param bucket      - bucket to build up
   * @param step_update - step data for the update
   * @param cmatch      - class match
   * @param validator   - validator
   */
  private async _performUpdate(
    bucket: DelayedStagingBucket,
    step_update: StepUpdateData,
    cmatch: Cmatch,
    validator: Validator
  ): Promise<unknown> {
    const {id, update} = step_update;
    this._current_step = id;

    for (const field in update) {
      validator.validateField(bucket, cmatch, id, field);
    }

    // Need to defer so DAPIs have a chance to kick off
    return this._defer();
  }

  /**
   * Facilitate deferral until pending DAPIs are done
   *
   * @param manager    - DAPI manager
   * @param ts_ctor    - gets current timestamp
   * @param timeout_ms - optional timeout
   */
  private async _pendingDapis(
    manager: DataApiManager,
    ts_ctor: () => UnixTimestamp,
    timeout_ms = 5000
  ) {
    const getPending = () =>
      Object.values(manager.getPendingApiCalls()).filter(
        val => val && val.uid > ts_ctor() - timeout_ms
      );

    while (getPending().length) {
      await this._defer(100);
    }
  }
}
