/**
 *  Step Simulator
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {Cmatch, CmatchData} from '../../client/Cmatch';
import {PositiveInteger} from '../../numeric';
import {Program} from '../../program/Program';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';

type StepClosure = (
  id: PositiveInteger,
  fields: string[],
  cmatch: CmatchData
) => void;

export interface StepSummary {
  id: PositiveInteger;
  fields: string[];
}

/** Expose step operations for situations where a UI is unavailable */
export class StepSimulator {
  /**
   * Create a new Step Simulator
   *
   * @param _quote    - related quote
   * @param _program  - rating program
   * @param _fields   - all field names
   */
  constructor(
    private _quote: ServerSideQuote,
    private _program: Program,
    private _fields: string[]
  ) {}

  /**
   * Allow operations to take place for each step
   *
   * @param closure - function to run for each step
   * @param cmatch  - class match
   */
  public async each(closure: StepClosure, cmatch: Cmatch) {
    const steps = await this.getSteps(cmatch);

    for (const step of steps) {
      closure(step.id, step.fields, cmatch.getMatches());
    }
  }

  /**
   * Get step data
   *
   * @param cmatch - class match
   *
   * @return step data
   */
  public async getSteps(cmatch: Cmatch): Promise<StepSummary[]> {
    const result = [];
    const {fields_by_step, top_step} = this._getFieldsByStep(
      this._quote,
      cmatch.getMatches(),
      this._fields
    );

    for (let id = this._program.getFirstStepId(); id <= top_step; id++) {
      result.push({id, fields: fields_by_step[id] ?? []});
    }

    return result;
  }

  /**
   * Get a grouping of applicable and requested fields grouped by step
   *
   * Sort fields by what was requested as this represents the input to our
   * system and thus the data that should be asserted against. We will not
   * consider fields if they were requested but not applicable as these will be
   * reset
   *
   * This function will also filter inapplicable steps
   *
   * We also return the top applicable step id so that we don't have to loop
   * again later to find it
   *
   * @param quote          - the quote to get top applicable step for
   * @param cdata          - class match data
   * @param request_fields - the fields that were initially sent in
   *
   * @returns a tuple containing applicable and requested fields grouped by step
   *          and the top step id
   */
  private _getFieldsByStep(
    quote: ServerSideQuote,
    cdata: CmatchData,
    request_fields: string[]
  ): {fields_by_step: Record<number, string[]>; top_step: PositiveInteger} {
    let top_step: PositiveInteger = this._program.getFirstStepId();

    const field_state = Object.keys(quote.getFieldState() ?? {});

    const fields_by_step = request_fields
      // We do not want to do anything with the "publish_quote_ind" since
      // Dullahan does not allow a supplier to be chosen. This will
      // eventually not be the case, but we need to "init" the quote earlier.
      .filter(field => field !== 'publish_quote_ind')
      .filter(field => {
        // We need this special consideration for live search fields because
        // their underlying questions will never be visible. Instead we need
        // to check for the visibility of their anchors
        return (
          field_state.includes(field) ||
          field_state.includes(field.replace('_input', '_anchor'))
        );
      })
      .map(field => ({field, step_id: this._program.qstep[field]}))
      .filter(({step_id}) => this._program.isStepVisible(cdata, step_id))
      .reduce((result: Record<number, string[]>, {field, step_id}) => {
        result[step_id] = result[step_id] ?? [];
        result[step_id].push(field);

        top_step = step_id > top_step ? step_id : top_step;

        return result;
      }, {});

    return {fields_by_step, top_step};
  }
}
