import {DapiResult} from '../../client/dapi/DataApiMediator';
import {DelayedStagingBucket} from '../../bucket/DelayedStagingBucket';
import {Program} from '../../program/Program';

export interface DapiUpdateError {
  field: string;
  current_value: unknown;
  new_value: unknown;
  message: string;
}

/**
 * Validate output from the DAPI against pre-existing bucket data
 *
 * This uses a mechanism to accumulate the errors because this function is
 * passed as a callback to an event system. Throwing errors are done by
 * inspecting the error log after the events using this function have been
 * handled.
 *
 * There are a few different possibilities when we get matches back in a DAPI:
 *
 * - If no results are returned then we do nothing.
 * - If one result is returned then we use it.
 * - If one of the results is an exact match then we will use it.
 * - If many results are exact matches then we will use the first one.
 * - If no results match what we expect then we will just use the first
 *   non-empty result. This last option is done because these Dullahan does not
 *   allow for user interaction like the regular UI does so the user cannot
 *   pick the correct result (e.g. multiple cities returned). Since these are
 *   indications and the answers do not need to be perfect, we can just guess at
 *   the correct answer and let the user confirm or fix it when they finally
 *   enter the UI.
 *
 * @param bucket     - bucket
 * @param program    - compiled program
 * @param errors     - list of errors that occur
 * @param name       - name of field
 * @param i          - bucket index to apply the results to
 * @param results    - formatted results
 * @param map        - mapping from DAPI output to much field
 */
export const validateDapiOutput =
  (bucket: DelayedStagingBucket) =>
  (program: Program) =>
  (errors: DapiUpdateError[]) =>
  (
    name: string,
    i: number | string,
    results: DapiResult,
    map: Record<string, string>
  ): void => {
    const keys = Object.keys(results);
    const result_count = keys.filter(key => key !== '___empty').length;

    // DAPI returned no results
    if (!result_count) {
      return;
    }

    const index = typeof i === 'string' ? parseInt(i) : i;
    const value = bucket.getDataByName(name) ?? [];

    const indexes = index === -1 ? Object.keys(value).map(Number) : [index];
    for (const index of indexes) {
      const new_value =
        /* It might be possible for there to be more than one CI match, but ATM
         it hasn't been noticed and is not considered likely. This comment is
         just a reminder that this situation could arise and will be ignored. */
        result_count === 1
          ? keys[0]
          : keys.find(is_ci_match(value?.[index])) ?? keys.find(x => x !== '');

      // Can't detect the value for the main param
      if (new_value === undefined) {
        const message = 'New value is undefined';
        const current_value = value[index];
        errors.push({field: name, current_value, new_value, message});
        continue;
      }

      // Prefill the update with the existing value
      const update: Record<string, unknown[]> = {[name]: value};

      // Update the main param
      update[name][index] = new_value;

      const result = results[new_value];

      // Set all mapped fields' values
      for (const field in map) {
        const value = bucket.getDataByName(field) ?? [];
        const result_key = map[field];
        const new_value = (<Record<string, string>>result)[result_key];

        update[field] = value;

        if (
          typeof value[index] === 'undefined' ||
          is_default(program)(field)(value[index]) ||
          is_ci_match(value[index])(String(new_value))
        ) {
          update[field][index] = new_value;
        } else {
          const message = 'Current value prevents new value from being set';
          const current_value = value[index];
          errors.push({field, current_value, new_value, message});
          continue;
        }
      }

      bucket.setValues(update);
      bucket.processValues();
      bucket.commit();
    }
  };

const is_default = (program: Program) => (key: string) => (value: unknown) =>
  program.defaults[key] === value;

const is_ci_match = (value: unknown) => (comparison: string) =>
  typeof value === 'string' && value.toLowerCase() === comparison.toLowerCase();
