/**
 *  Indication Validator
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {ClientQuote} from '../../client/quote/ClientQuote';
import {Cmatch} from '../../client/Cmatch';
import {DelayedStagingBucket} from '../../bucket/DelayedStagingBucket';
import {PositiveInteger} from '../../numeric';
import {Program, TriggerEvent} from '../../program/Program';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {ServerSideQuote} from '../../server/quote/ServerSideQuote';
import {StepSimulator} from '../step/StepSimulator';
import {Step} from '../../step/Step';
import {StagingBucket} from '../../bucket/StagingBucket';

export interface ValidationResults {
  failed_assertions: string[];
  missing_required: string[];
  triggered_events: string[];
}

export type ValidatorFactory = (
  program: Program,
  step_simulator: StepSimulator
) => Validator;

/**
 * Ensure the validity of the quote
 */
export class Validator {
  /** Storing this as a member assumes the failures can/should accumulate */
  private _results: ValidationResults = {
    failed_assertions: [],
    missing_required: [],
    triggered_events: [],
  };

  /**
   * Create a new validator instance
   *
   * @param _program        - program
   * @param _getStep        - ctor for getting the current step
   * @param _step_simulator - simulates UI steps
   */
  constructor(
    private readonly _program: Program,
    private readonly _getStep: (id: number, quote: ServerSideQuote) => Step,
    private readonly _step_simulator: StepSimulator
  ) {}

  /**
   * Validate quote data
   *
   * Check for missing required fields or failed assertions
   *
   * @param quote        - server side quote
   * @param client_quote - client side quote
   * @param cmatch       - class match
   *
   * @returns missing required, failed assertions, or attempted trigger calls
   */
  async validate(
    quote: ServerSideQuote,
    client_quote: ClientQuote,
    cmatch: Cmatch
  ): Promise<ValidationResults> {
    const validate_step =
      (bucket: QuoteDataBucket) => (id: PositiveInteger, fields: string[]) => {
        // Program is expecting a DelayedStagingBucket so for now we're just
        // going to pretend that is what it is getting because the server does
        // not want or need that functionality
        const staging_bucket = <DelayedStagingBucket>(<unknown>bucket);

        staging_bucket.getPendingDataByName = (name, _) =>
          bucket.getDataByName(name);

        fields.forEach((field: string) =>
          this.validateField(staging_bucket, cmatch, id, field)
        );

        this.validateStep(bucket, quote, cmatch, id);
      };

    client_quote.visitData((bucket: StagingBucket) => {
      this._step_simulator.each(
        validate_step(<QuoteDataBucket>(<unknown>bucket)),
        cmatch
      );
    });

    return this._results;
  }

  /**
   * Validate an individual field
   *
   * @param bucket  - delayed staging bucket
   * @param cmatch  - class matcher
   * @param step_id - current step's ID
   * @param field   - field name
   */
  public validateField(
    bucket: DelayedStagingBucket,
    cmatch: Cmatch,
    step_id: PositiveInteger,
    field: string
  ): ValidationResults {
    const change_failures = this._program.change(
      step_id,
      field,
      bucket,
      {},
      cmatch,
      this._triggerCb.bind(this)
    );

    this._recordFailures(change_failures || {}, step_id);

    return this._results;
  }

  /**
   * Validate an entire step
   *
   * @param bucket  - bucket
   * @param quote   - server side quote
   * @param cmatch  - class matcher
   * @param step_id - current step's ID
   */
  public validateStep(
    bucket: QuoteDataBucket,
    quote: ServerSideQuote,
    cmatch: Cmatch,
    step_id: PositiveInteger
  ): ValidationResults {
    const failures = this._program.submit(step_id, bucket, cmatch.getMatches());

    this._recordFailures(failures || {}, step_id);

    const forward_failures = this._program.forward(
      step_id,
      bucket,
      cmatch.getMatches(),
      this._triggerCb.bind(this)
    );

    this._recordFailures(forward_failures || {}, step_id);

    this._getStep(step_id, quote)
      .setRequiredFieldNames(this._program.requiredFields[step_id])
      .getRequired(cmatch.getMatches(), false)
      .forEach(missing => {
        this._results.missing_required.push(
          `Step ${step_id}: ${missing[0]} missing index ${missing[1]}`
        );
      });

    return this._results;
  }

  /**
   * Record failures to a shared record
   *
   * @param failures - failures to record
   * @param step_id  - step where failure occurred
   */
  private _recordFailures(failures: CommonObject, step_id: PositiveInteger) {
    for (const key in failures) {
      this._results.failed_assertions.push(
        `Step ${step_id}: ${key}: ${failures[key]}`
      );
    }
  }

  /**
   * Expose a trigger callback used by the program for change and forward events
   *
   * @param event - event name
   * @param field - field name
   * @param value - new values
   * @param index - field index
   */
  private _triggerCb(
    event: TriggerEvent,
    field: string,
    value: string,
    index: PositiveInteger[]
  ) {
    // This is unsupported at the moment it should be resolved in DEV-9728
    this._results.triggered_events.push(
      `Triggering event '${event}' on field
    '${field}[${index}]' with value '${value}'.`
    );
  }
}
