/**
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {isPositiveInteger, PositiveInteger} from '../../numeric';
import InitRateRequestAssertionError from '../../error/InitRateRequestAssertionError';
import * as Validator from '../../util/common_validators';

export type InitRateRequestBody = {
  quote_id: PositiveInteger;
  policy_id: string;
  session_id: string;
  program: Validator.AlphaNumeric;
  bucket: Validator.Bucket;
  origin: Validator.AlphaNumeric;
};

export function assertInitRateRequest(
  body: unknown
): asserts body is InitRateRequestBody {
  if (typeof body !== 'object' || body === null) {
    throw new InitRateRequestAssertionError('Init rate request must an object');
  }

  const obj: Record<string, unknown> = <Record<string, unknown>>body;

  if (
    typeof obj.quote_id !== 'number' ||
    !isPositiveInteger(obj.quote_id) ||
    obj.quote_id === 0
  ) {
    throw new InitRateRequestAssertionError(
      'Init rate quote ID must be a positive integer'
    );
  }

  if (!Validator.isNonEmptyString(obj.session_id)) {
    throw new InitRateRequestAssertionError(
      'Init rate session ID must be a non-empty string'
    );
  }

  if (
    !Validator.isAlphaNumeric(obj.program) ||
    !Validator.isNonEmptyString(obj.program)
  ) {
    throw new InitRateRequestAssertionError(
      'Init rate program must be a non-empty string'
    );
  }

  if (
    !Validator.isAlphaNumeric(obj.origin) ||
    !Validator.isNonEmptyString(obj.origin)
  ) {
    throw new InitRateRequestAssertionError(
      'Init rate origin must be a non-empty string'
    );
  }

  if (!Validator.isString(obj.policy_id)) {
    throw new InitRateRequestAssertionError(
      'Init rate policy ID must be a string'
    );
  }
}
