/**
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {CopyRequestBody} from './copy_validator';
import * as Validator from '../../util/common_validators';
import InitRequestAssertionError from '../../error/InitRequestAssertionError';

/**
 * The init request body is the same as the copy request body, but without a quote quote_id
 */
export type InitRequestBody = Omit<CopyRequestBody, 'quote_id'>;

export function assertInitRequest(
  body: unknown
): asserts body is InitRequestBody {
  if (typeof body !== 'object' || body === null) {
    throw new InitRequestAssertionError('Init request must an object');
  }

  const obj: Record<string, unknown> = <Record<string, unknown>>body;

  if (obj.quote_id) {
    throw new InitRequestAssertionError('Init must not have a quote_id');
  }

  if (!Validator.isNonEmptyString(obj.session_id)) {
    throw new InitRequestAssertionError(
      'Init session ID must be a non-empty string'
    );
  }

  if (
    !Validator.isAlphaNumeric(obj.program) ||
    !Validator.isNonEmptyString(obj.program)
  ) {
    throw new InitRequestAssertionError(
      'Init program must be a non-empty string'
    );
  }

  if (
    !Validator.isAlphaNumeric(obj.origin) ||
    !Validator.isNonEmptyString(obj.origin)
  ) {
    throw new InitRequestAssertionError(
      'Init origin must be a non-empty string'
    );
  }

  if (!Validator.isString(obj.policy_id)) {
    throw new InitRequestAssertionError('Init policy ID must be a string');
  }

  if (!Validator.isBucket(obj.bucket)) {
    throw new InitRequestAssertionError('Init bucket invalid');
  }
}
