/**
 * Reduce field predicate results into vectors and flags
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// TODO: It is not appropriate to import client-specific modules from a
// generic module; move ClassData.
import {ClassData} from '../client/Cmatch';
import {PositiveInteger} from '../numeric';

export type FieldData = {
  [index: string]: FieldMatch | ClassData;
};

export type FieldMatch = {
  all: boolean;
  any: boolean;
  indexes: number[];
};

/**
 * Generate match vector for fields given field predicates and
 * classification results
 *
 * TODO: Support for multiple predicates on fields is for
 * backwards-compatibility with older classification systems; newer systems
 * generate a single classification representing the visibility of the
 * field, allowing the classification reduction complexity and logic to stay
 * within TAME.  Much of the complexity in this class can therefore be
 * removed in the future.
 */
export class FieldClassMatcher {
  /**
   * Initialize matcher with a list of fields and their classifications
   */
  constructor(private _fields: Record<string, string[]> = {}) {}

  /**
   * Generate classification match array for each field
   *
   * Any index for any field will be considered to be a match if the index
   * is classified as each of the field's required classifications.
   */
  public match(classes: ClassData): FieldData {
    return Object.keys(this._fields).reduce(
      (cmatch: FieldData, id: string) => {
        (cmatch[id] = this._reduceFieldMatches(this._fields[id], classes)),
          cmatch;

        return cmatch;
      },
      {__classes: classes}
    );
  }

  /**
   * Reduce field class matches to a vector
   *
   * All field predicates in FIELDC will be reduced and combined into a
   * single vector representing the visibility of each index.
   */
  private _reduceFieldMatches(
    fieldc: string[],
    classes: ClassData
  ): FieldMatch {
    const vis: PositiveInteger[] = [];

    let all = true;
    let hasall = false;

    // determine if we have a match based on the given classifications
    for (const c in fieldc) {
      const data = classes[fieldc[c]] || {is: false, indexes: 0};

      // this will ensure that, if we've already determined some sort
      // of visibility, that encountering a scalar will still manage
      // to affect previous results even if it is the last
      // classification that we are checking
      const [thisall, indexes] = Array.isArray(data.indexes)
        ? [false, data.indexes]
        : [true, vis];

      let alltrue = !thisall || data.indexes === 1;

      // if no indexes apply for the given classification (it may be a
      // pure boolean), then this variable will be true if any only if
      // all of them are true. Note that we only want to take the
      // value of thisall if we're on our first index, as if hasall is
      // empty thereafter, then all of them certainly aren't true!
      hasall = hasall || (thisall && +c === 0);

      indexes.forEach((_, i) => {
        // default to visible; note that, if we've encountered any
        // "all index" situations (scalars), then we must only be
        // true if the scalar value was true
        vis[i] = <PositiveInteger>(
          +(
            (!hasall || all) &&
            (vis[i] === undefined ? <PositiveInteger>1 : vis[i]) &&
            this._reduceMatch(thisall ? data.indexes : indexes[i])
          )
        );
      });

      alltrue = alltrue && vis.every(x => x);
      all = all && alltrue;
    }

    // store the classification match data for assertions, etc
    return {
      all: all,
      any: all || vis.some(x => !!x),
      indexes: vis,
    };
  }

  /**
   * Reduces the given scalar or vector
   *
   * If a scalar is provided, then it is immediately returned. Otherwise, the
   * vector is reduced using a logical or and the integer representation of
   * the boolean value returned.
   *
   * This is useful for classification matrices since each match will be a
   * vector and we wish to consider any match within that vector to be a
   * positive match.
   */
  private _reduceMatch(
    result: PositiveInteger | PositiveInteger[] | PositiveInteger[][]
  ): boolean {
    if (!Array.isArray(result)) {
      return !!result;
    }

    return result.some(x => this._reduceMatch(x));
  }
}
