/**
 * Contains Program base class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Bucket, BucketData, BucketDefaults} from '../bucket/Bucket';
import {CmatchData, ClassData} from '../client/Cmatch';
import {DataApiDefinition, DataApiResult} from '../dapi/DataApi';
import {Data} from '../client/quote/ClientQuote';
import {PositiveInteger} from '../numeric';
import {StagingBucket} from '../bucket/StagingBucket';
import {DelayedStagingBucket} from '../bucket/DelayedStagingBucket';
import {QuoteDataBucket} from '../bucket/QuoteDataBucket';

export type ClassificationResult = {[index: string]: any};
export type ClassificationRetain = Record<string, any>;
export type QuestionTypes = Record<string, any>;
export type AnswerRefs = Record<string, string>;
export type SelectData = Record<string, Record<string | number, string>>;
export type DataApiDefinitions = Record<string, DataApiDefinition>;
export type ServerSideDapi = {
  name: string;
  when: string;
  value: string;
  mapsrc: Record<string, string>;
  mapdest: Record<string, string>;
};
export type MetaField = {desc: string; dapi: ServerSideDapi};
export type MetaFields = Record<string, MetaField>;

export type EncryptedKeys = Record<string, string[]>;

type Step = {
  title: string;
  id: string;
  continueLabel: string;
  backLabel: string;
  type: string;
  section: string;
  groups: string[];
};

type StepName = string;
type FieldName = string;

type TriggerEvent =
  | 'set'
  | 'visit'
  | 'forward'
  | 'change'
  | 'beforeLoad'
  | 'submit'
  | 'navFreeze'
  | 'rate'
  | 'dapi';

type TriggerCallBack = (
  event: TriggerEvent,
  field: FieldName,
  value: string,
  index: PositiveInteger[]
) => void;

type BucketLayerData = {
  /**
   * The id of the bucket layer
   */
  id: string;

  /**
   * A classification representing when the layer is applicable
   */
  when: string;

  /**
   * Data representing defaults for the layer
   */
  defaults: BucketDefaults;
};

export declare abstract class Program {
  apis: DataApiDefinitions;

  autosave: boolean;

  bucketLayers: BucketLayerData[];

  classifierModulePath: string;

  clearNaFields: boolean;

  copyQuoteUrl: string;

  cretain: ClassificationRetain;

  /**
   * Default field values
   */
  defaults: BucketDefaults;

  downstreamServices: string[];

  encryptedKeys: EncryptedKeys;

  flags: Record<string, string>;

  groupFields: Record<string, string[]>;

  groupWhens: Record<string, string>;

  id: string;

  internal: Record<string, boolean>;

  isInternal: boolean;

  kbclear: CommonObject;

  lockOnImport: boolean;

  newDocumentUrl: string;

  /**
   * A mapping of question ids to the server-side dapis that should be called
   * when those fields are changed
   */
  mapis: Record<string, string[]>;

  meta: {
    arefs: AnswerRefs;
    fields: MetaFields;
    groups: Record<string, {min: PositiveInteger; max: PositiveInteger}>;
    qdata: SelectData;
    staticSelectQuestions: Record<string, boolean>;
    qtypes: QuestionTypes;
  };

  naFieldValue: string;

  nonpersistent: string[];

  qgroup: Record<string, string>;

  rateSteps: boolean[];

  ratingPoll: {
    maxAttempts: number;
    indicationMaxAttempts: number;
    delaySeconds: number;
  };

  requiredFields: Record<StepName, Record<FieldName, boolean>>;

  secureFields: string[];

  stepIds: string[];

  steps: Step[];

  unlockable: boolean;

  version: string;

  whens: Data;

  readonly ineligibleLockCount: number;

  getId(): string;

  /**
   * Data API
   */
  dapi(
    step_id: PositiveInteger,
    name: string,
    bucket: StagingBucket,
    diff: Record<string, any>,
    cmatch: CmatchData,
    callback: (() => void) | null
  ): DataApiResult;

  /** Not explicitly listed on the program but is used in practice */
  dapimap?: Record<string, Record<string, string>>;

  initQuote(bucket: Bucket, store_only?: boolean): void;

  /**
   * Get known classifier fields
   */
  getClassifierKnownFields(): ClassificationResult;

  /**
   * Classify the given bucket data
   */
  classify(data: BucketData): ClassData;

  submit(
    step_id: PositiveInteger,
    bucket: QuoteDataBucket,
    cmatch: CmatchData,
    trigger_callback?: TriggerCallBack,
    decode_dangerous?: <T>(x: T) => T
  ): null | CommonObject;

  postSubmit(
    step_id: PositiveInteger,
    bucket: QuoteDataBucket,
    trigger_callback?: TriggerCallBack
  ): boolean;

  change(
    step_id: PositiveInteger,
    name: string,
    bucket: DelayedStagingBucket,
    diff: Record<string, any>,
    cmatch: CmatchData,
    trigger_callback: TriggerCallBack,
    decode_dangerous?: <T>(x: T) => T
  ): null | Record<string, unknown[]>;

  forward(
    step_id: PositiveInteger,
    bucket: QuoteDataBucket,
    cmatch: ClassData,
    trigger_callback: TriggerCallBack,
    decode_dangerous?: <T>(x: T) => T
  ): null | CommonObject;

  beforeLoadStep(
    step_id: PositiveInteger,
    bucket: QuoteDataBucket,
    trigger_callback: TriggerCallBack
  ): boolean;

  /**
   * Determine if a step is visible
   *
   * @param class_data - classifications
   * @param step_id    - of the step to get visibility for
   */
  isStepVisible(
    class_data: ClassificationResult,
    step_id: PositiveInteger
  ): boolean;

  /**
   * Determine if a field is applicable
   */
  hasNaField(field: string, classes: ClassData, index: number): boolean;

  /**
   * Determine whether question QTYPE is known
   */
  hasKnownType(name: string): boolean;

  /**
   * First valid step id
   */
  getFirstStepId(): PositiveInteger;

  processNaFields(data: Record<string, any>): void;

  /**
   * Get the next visible step after the supplied step
   *
   * @param class_data - classifications
   * @param step_id    - the starting step id
   *
   * @return The next visible step id or undefined if there isn't one
   */
  getNextVisibleStep(
    class_data: ClassData,
    step_id: PositiveInteger
  ): PositiveInteger | undefined;

  /**
   * Returns whether or not the given step is the manage quote
   * step.
   *
   * @param step_id - the step to check
   *
   * @return true if manage quote step, otherwise false
   */
  isManageQuoteStep(step_id: PositiveInteger): boolean;

  /**
   * Returns whether or not the given step is the last step
   *
   * @param class_data - classifications
   * @param step_id    - the step to check
   *
   * @return true if last step, otherwise false
   */
  isLastStep(class_data: ClassData, step_id: PositiveInteger): boolean;

  /**
   * Returns whether or not the given step is the indication quote
   * step.
   *
   * @param step_id - the step to check
   *
   * @return true if indication step, otherwise false
   */
  isIndicationQuoteStep(step_id: PositiveInteger): boolean;

  /** Names of fields under a given group without linked groups' fields */
  readonly groupExclusiveFields: Record<string, string[]>;

  readonly groupIndexField: Record<string, string>;

  /** Map of question ids to the steps that declare them */
  readonly qstep: Record<string, PositiveInteger>;

  /** Map of question ids to whether they are required when applicable */
  readonly qrequired: Record<string, boolean>;

  lockTimeout: {
    preRateExpiration: number;
    postRateExpiration: number;
    renewalExpiration: number;
    postRateGracePeriod: number;
    preRateGracePeriod: number;
    renewalEffDateGracePeriod: number;
  };
}
