/**
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {BucketData} from '../bucket/Bucket';
import {UserSession} from '../server/request/UserSession';
import {Program} from './Program';
import {readdirSync, statSync} from 'fs';

type ProgramName = string;

/**
 * A function to clean bucket data.
 */
export type CopyCleaner = (
  bucket_data: BucketData,
  session: UserSession,
  agency_number: string,
  request_bucket: BucketData,
  select_question_metadata: {
    static_select_questions: Program['meta']['staticSelectQuestions'];
    valid_select_values: Program['meta']['qdata'];
    question_defaults: Program['defaults'];
  }
) => BucketData;

/**
 * Load a program based on the program name
 *
 * This will be run for every request to the Dullahan service controller.
 */
export const programLoader = (program_name: ProgramName): (() => Program) => {
  if (!isValidProgramName(program_name)) {
    throw new Error('Invalid program name');
  }

  // The call to require should be cached so this should hopefully
  // not add a lot of overhead if called repeatedly.
  /* eslint-disable @typescript-eslint/no-var-requires */
  return require(`program/${program_name}/Program`);
  /* eslint-disable @typescript-eslint/no-var-requires */
};

/**
 * Load a program copy cleaner based on the program name
 *
 * This will be run for every request to the Dullahan service controller.
 */
export const programCopyCleanerLoader = (
  program_name: ProgramName
): CopyCleaner => {
  if (!isValidProgramName(program_name)) {
    throw new Error('Invalid program name');
  }

  // The call to require should be cached so this should hopefully
  // not add a lot of overhead if called repeatedly.
  /* eslint-disable @typescript-eslint/no-var-requires */
  const copy_config = require(`program/${program_name}/copy-cleaner`);
  /* eslint-disable @typescript-eslint/no-var-requires */
  return copy_config.cleanBucket;
};

const isValidProgramName = (x: unknown): x is ProgramName =>
  typeof x === 'string' && /^[A-Za-z0-9]+$/.test(x);

/**
 *  Imports all program definitions and constructs a list of program ids
 *  @param base_program_path - Base url to access program folder
 *
 *  @return                  - List of available programs
 */
export const getProgramIds = (base_program_path: string): string[] => {
  return readdirSync(base_program_path).filter(item => {
    return statSync(`${base_program_path}/${item}`).isDirectory();
  });
};
