/**
 * Contains program Quote class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @todo Use ``document'' terminology in place of ``quote''
 */

import {EventEmitter} from 'events';
import {PositiveInteger} from '../numeric';
import {Program} from '../program/Program';
import {Quote, QuoteId} from './Quote';
import {QuoteDataBucket} from '../bucket/QuoteDataBucket';
import {StagingBucket} from '../bucket/StagingBucket';
import {LayeredBucket} from '../bucket/layer/LayeredBucket';

export declare class BaseQuote extends EventEmitter implements Quote {
  /**
   * This is an easejs construct and can be removed once BaseQuote is converted
   * to typescript
   */
  __self: {
    /**
     * The message to display in the event of a save error
     */
    $: (key: 'SAVE_LOCK_MSG') => string;
  };

  constructor(
    quote_id: PositiveInteger,
    bucket: QuoteDataBucket,
    program: Program
  );

  /**
   * Retrieve Program associated with quote
   *
   * @return quote program
   */
  getProgram(): Program;

  /**
   * Returns the program id associated with the quote
   *
   * @return program id
   */
  getProgramId(): string;

  /**
   * Sets the id of the program
   *
   * @param id of program
   */
  setProgramId(program_id: string): this;

  /**
   * Returns the quote id
   *
   * The quote id is immutable. A different quote id would represent a
   * different quote, therefore a new object should be created with the
   * desired quote id.
   *
   * @return quote id
   */
  getId(): QuoteId;

  /**
   * Returns the id of the current step
   *
   * @return id of current step
   */
  getCurrentStepId(): PositiveInteger;

  /**
   * Sets the id of the current step
   *
   * @param id of current step
   */
  setCurrentStepId(step_id: number): this;

  /**
   * Sets an explicit lock, providing a reason for doing so
   *
   * @param reason - lock reason
   * @param step   - step that user may not navigate prior
   *
   * @return self
   */
  setExplicitLock(reason: string, step?: number): this;

  /**
   * Clears an explicit lock
   *
   * @return {Quote} self
   */
  clearExplicitLock(): this;

  /**
   * Set the date that the premium was calculated as a Unix timestamp
   *
   * @param timestamp - Unix timestamp representing premium date
   *
   * @return self
   */
  setLastPremiumDate(timestamp: UnixTimestamp): this;

  /**
   * Retrieve the last time the premium was calculated
   *
   * @return last calculated time or 0
   */
  getLastPremiumDate(): UnixTimestamp;

  /**
   * Retrieve the start date the document was created
   *
   * @return start date or 0
   */
  getStartDate(): UnixTimestamp;

  /**
   * Returns the layered bucket used to store the quote data
   *
   * @return the layered data bucket
   */
  getLayeredBucket(): LayeredBucket;

  /**
   * Returns the bucket used to store the quote form data
   *
   * @return the data bucket
   */
  getBucket(): QuoteDataBucket;

  /** Whether quote is locked, for any reason */
  isLocked(): boolean;

  /**
   * Retrieves the reason for an explicit lock
   *
   * @return lock reason
   */
  getExplicitLockReason(): string;

  /**
   * Returns the maximum step to which the explicit lock applies
   *
   * If no step restriction is set, then 0 will be returned.
   *
   * @return {number} locked max step or 0 if not applicable
   */
  getExplicitLockStep(): PositiveInteger;

  /**
   * Returns whether the quote has been imported
   *
   * @return true if imported, otherwise false
   */
  isImported(): boolean;

  /**
   * Returns whether the quote has been bound
   *
   * @return true if bound, otherwise false
   */
  isBound(): boolean;

  /**
   * Returns the id of the highest step the quote has reached
   *
   * @return top visited step id
   */
  getTopVisitedStepId(): PositiveInteger;

  /**
   * Returns the id of the highest step the quote has saved
   *
   * @return top saved step id
   */
  getTopSavedStepId(): PositiveInteger;

  /**
   * Sets the id of the highest step the quote has saved
   *
   * @param top saved step id
   */
  setTopSavedStepId(id: number): this;

  /**
   * Returns the quote's expiration date
   *
   * @return quote's expiration date
   */
  getExpirationDate(): number;

  /**
   * Sets username of the user who is accessing this quote
   *
   * @param {string} username
   */
  setUserName(username: string): this;

  /**
   * Returns the username of the user who is accessing this quote
   *
   * @return {string} username
   */
  getUserName(): string;

  /**
   * Sets roles of the user who is accessing this quote
   *
   * @param user roles
   *
   * @return {Quote} self
   */
  setUserRoles(roles: string[]): this;

  /**
   * Returns the roles of the user who is accessing this quote
   *
   * @return {string[]} roles
   */
  getUserRoles(): string[];

  /**
   * Sets the username of the user who created this quote
   *
   * @param {string} username
   */
  setCreatedByUserName(username: string): this;

  /**
   * Returns the username of the user who created this quote
   *
   * @return {string} username
   */
  getCreatedByUserName(): string;

  /**
   * Sets the username of the user who last updated this quote
   *
   * @param {string} username
   */
  setLastUpdatedByUserName(username: string): this;

  /**
   * Returns the username of the user who last updated this quote
   *
   * @return {string} username
   */
  getLastUpdatedByUserName(): string;

  /**
   * Returns the id of the agent that owns the quote
   *
   * @return {number} agent id
   */
  getAgentId(): number;

  /**
   * Returns the entity id of the agent that owns the quote
   *
   * @return {number} agent id
   */
  getAgentEntityId(): number;

  /**
   * Returns the name of the agent that owns the quote
   *
   * @return {string} agent name
   */
  getAgentName(): string;

  /**
   * Visits staging data
   */
  visitData(visitor: (bucket: StagingBucket) => void): void;

  /**
   * Determine whether or not a quote-wide error exists
   *
   * Use getError() to retrieve the actual error string.
   *
   * @return true if error exists, otherwise false
   */
  hasError(): boolean;

  /**
   * Retrieve quote-wide error
   *
   * @return quote-wide error, or empty string
   */
  getError(): string;

  /**
   * Sets whether or not there was an error saving the quote
   *
   * This error should lock the entire quote until it is manually cleared.
   *
   * @param value the new value
   */
  setSaveError(value: boolean): this;

  /**
   * Retrieve quote save error
   *
   * @return quote save error
   */
  getSaveError(): boolean;

  /**
   * Returns whether the quote has expired or not
   *
   * @param current_date current timestamp in seconds to determine
   *                              if expiration date has passed
   *
   * @return flag indicating if the quote has expired
   */
  hasExpired(current_date: UnixTimestamp): boolean;

  /**
   * Determine whether quote needs to be imported
   *
   * Bound quotes will never need importing.
   *
   * @param set flag value
   */
  needsImport(set: boolean): this;
  needsImport(set: undefined): boolean;
  needsImport(set?: boolean): boolean;

  /**
   * Returns the quote's initial rated date
   *
   * @return quote's initial rated date
   */
  getInitialRatedDate(): UnixTimestamp;

  /**
   * Returns quote data
   *
   * @param name - name of data to retrieve
   *
   * @return quote data
   */
  getDataByName(name: string): unknown[];

  /**
   * Sets quote data
   *
   * The data will be merged, not overwritten.
   *
   * @param data - data to set on the quote
   *
   * @return self
   */
  setData(data: CommonObject): this;

  /**
   * Sets the quote's initial rated date
   *
   * @param time - initial rated date as a Unix timestamp
   */
  setInitialRatedDate(time: UnixTimestamp): this;

  /**
   * Sets quote bound status
   *
   * Represents whether the quote has been bound
   *
   * @param {boolean} value true if bound, otherwise false
   *
   * @return {Quote} self
   */
  setBound(bound: boolean): this;

  /**
   * Sets the quote start date
   *
   * @param {number} time start date as a Unix timestamp
   *
   * @return {Quote} self
   */
  setStartDate(date: UnixTimestamp): this;

  /**
   * Sets quote imported status
   *
   * Represents whether the quote has been imported into our agency management
   * system.
   *
   * @param {boolean} value true if imported, otherwise false
   *
   * @return {Quote} self
   */
  setImported(imported: boolean): this;
}
