/**
 * Facade for vanilla document server
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {QuoteDataBucket} from '../bucket/QuoteDataBucket';
import {Store} from '../store/Store';
import {Collection} from 'mongodb';
import {TokenNamespace, generateTokenId} from './token/Token';
import {MongoTokenDao} from './token/MongoTokenDao';
import {createDapiManager} from './dapi/ServerDataApiManagerFactory';
import {DapiMetaSource} from './meta/DapiMetaSource';
import {DataProcessor} from './request/DataProcessor';
import {filter} from '../bucket/bucket_filter';
import {StagingBucket} from '../bucket/StagingBucket';
import {bucketClearer} from '../bucket/BucketClear';
import {DapiProcessor} from './dapi/DapiProcessor';
import {Server} from './Server';
import {JsonServerResponse} from './request/JsonServerResponse';
import {cleanDocument} from './quote/loader';
import {WebhookService} from './service/WebhookService';
import {PersistentTokenStore} from './token/store/PersistentTokenStore';
import {DefaultFeatureFlag} from '../system/flags/DefaultFeatureFlag';
import {SplitFeatureFlag} from '../system/flags/SplitFeatureFlag';
import {DocumentId} from '../document/Document';
import {PriorityLog} from './log/PriorityLog';
import {Bucket} from '../bucket/Bucket';
import {ServerDao} from './db/ServerDao';
import {Key} from '../../liza-crypto';

/**
 * Vanilla document server
 *
 * XXX: This is a mess, and it's only getting worse with dependencies
 * instantiated everywhere.  Everything should be instantiated in one place
 * rather than part of them being passed in here.  See controller.js.
 */
export class DocumentServer {
  /**
   * Create document server
   *
   * See above XXX.
   *
   * @param dao         - server DAO
   * @param logger      - log manager
   * @param enc_service - encryption service
   * @param origin_url  - HTTP_ORIGIN_URL
   * @param conf        - configuration store
   * @param collection  - database collection
   * @param ts_ctor     - timestamp constructor
   *
   * @return {Promise<Server>}
   */
  public async create(
    dao: ServerDao,
    logger: PriorityLog,
    enc_service: any,
    origin_url: string,
    conf: Store,
    collection: Collection,
    ts_ctor: () => UnixTimestamp,
    gcp_keyfile_path: string,
    liza_kek: Key
  ) {
    const dapi_conf = await conf.get('dapi');
    const feature_flag = await this.createFeatureFlag(conf);

    const token_dao = new MongoTokenDao(collection, 'dapitok', ts_ctor);

    const dapi_manager_ctor =
      createDapiManager(origin_url)(dapi_conf)(token_dao)(gcp_keyfile_path);

    const dapi_meta_source = new DapiMetaSource(() => new QuoteDataBucket());

    const data_processor = new DataProcessor(
      logger,
      filter,
      dapi_manager_ctor,
      dapi_meta_source,
      (bucket: Bucket) => new StagingBucket(bucket),
      bucketClearer('meta.')
    );

    const dapi_input_processor = new DapiProcessor(
      dao,
      logger,
      ts_ctor,
      data_processor
    );

    const srv = new Server(
      new JsonServerResponse(),
      dao,
      logger,
      enc_service,
      dapi_input_processor,
      ts_ctor,
      feature_flag,
      cleanDocument({liza_kek})
    );

    const webhook_svc = new WebhookService(
      logger,
      (quote_id: DocumentId, token_ns: TokenNamespace) =>
        new PersistentTokenStore(
          token_dao,
          quote_id,
          token_ns,
          generateTokenId
        ),
      () => new StagingBucket(new QuoteDataBucket()),
      dapi_manager_ctor,
      dapi_meta_source,
      dapi_input_processor
    );

    return {
      srv,
      dapi_input_processor,
      webhook_svc,
      gcp_keyfile_path,
    };
  }

  /**
   * Create a feature flag checker
   *
   * We will use split if it is configured and included as a dependency,
   * otherwise we will default to the flags defined in the config file
   *
   * @param {conf} feature_flag Feature Flag config
   *
   * @return {FeatureFlag} A feature flag checker
   */
  public createFeatureFlag(conf: Store) {
    return conf
      .get('services.featureFlag')
      .then(flag_conf =>
        flag_conf.reduce(
          (accum: Record<string, unknown>, value: unknown, key: string) => {
            accum[key] = value;
            return accum;
          },
          {}
        )
      )
      .then(flag_conf => {
        const flag = new DefaultFeatureFlag(flag_conf.flags);

        if (flag_conf.type === 'splitio') {
          const {SplitFactory} = require('@splitsoftware/splitio');
          const api_key = flag_conf.splitio.key;
          const host = flag_conf.splitio.redis.host;
          const port = flag_conf.splitio.redis.port;
          const password = flag_conf.splitio.redis.password;
          const redis_dsn =
            password.length > 0
              ? 'redis://:' + password + '@' + host + ':' + port + '/0'
              : 'redis://' + host + ':' + port + '/0';
          const factory = new SplitFactory({
            mode: 'consumer',
            core: {authorizationKey: api_key},
            storage: {
              type: 'REDIS',
              options: {
                url: redis_dsn,
              },
            },
          });

          return new SplitFeatureFlag(flag, {}, factory.client());
        }

        return flag;
      });
  }
}
