/*
 * Contains program Server class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @todo like Client and Ui, this mammoth did not evolve well and has too
 *       many responsibilities; refactor
 */

import {ClientActions} from '../client/action/ClientAction';
import {PositiveInteger, incrementPositiveInteger} from '../numeric';
import {
  ClassificationResult,
  EncryptedKeys,
  MetaFields,
  Program,
} from '../program/Program';
import {FeatureFlag} from '../system/flags/FeatureFlag';
import {DocumentData} from './db/MongoServerDao';
import {PriorityLog} from './log/PriorityLog';
import {ServerSideQuote} from './quote/ServerSideQuote';
import {JsonServerResponse} from './request/JsonServerResponse';
import {UserSession} from './request/UserSession';
import {UserRequest} from './request/UserRequest';
import {EventEmitter} from 'events';
import {initExistingDocument, QuoteNotFound} from './quote/loader';
import {forEachAggregatedClassSuffix} from '../util/classifications';
import {inspect} from 'util';
import {readFile} from 'fs';
import type {BucketData} from '../bucket/Bucket';
import {RateResult} from './rater/Rater';
import {DapiProcessor} from './dapi/DapiProcessor';
import {QuoteDataBucket} from '../bucket/QuoteDataBucket';
import {ServerDao} from './db/ServerDao';
import * as TE from 'fp-ts/lib/TaskEither';
import {ServiceSecretMap} from './service/KekService';
import {isObject} from '../type_safety';
import {FieldClassMatcher} from '../field/FieldClassMatcher';

/* eslint-disable @typescript-eslint/no-var-requires */
const {
  bucket: {
    BucketSiblingDescriptor,

    diff: {
      StdBucketDiffContext,
      GroupedBucketDiffContext,
      StdBucketDiffResult,
      GroupedBucketDiffResult,
      StdBucketDiff,
    },
  },

  server: {
    encsvc: {QuoteDataBucketCipher},
  },

  util: {ShallowArrayDiff},
  // eslint-disable-next-line node/no-unpublished-require
} = require('..');
/* eslint-enable @typescript-eslint/no-var-requires */

/**
 * Decrypt a single bucket field.
 *
 * For historical reasons, we unfortunately do not know for certain what we
 * are getting here, so the caller must accept any arbitrary data.  Most
 * likely we are receiving either a string or an array of strings.
 */
export type DecryptField = <T>(x: T) => T;

export class Server extends EventEmitter {
  /**
   * Holds bucket ciphers for each program
   * @type {Object.<string,QuoteDataBucketCipher>}
   */
  private _bucketCiphers: any = {};

  /**
   * Functions that will sanitizer data before it is sent to the client
   *
   * This allows the server to not be responsible for knowing what should be
   * filtered. Instead that responsibility is delegated to whichever subservice
   * wants data to be filtered
   */
  private _clientDataSanitizers: ((data: BucketData) => void)[] = [];

  /**
   * Step and program cache
   *
   * @type {Store}
   */
  private _cache: any = null;

  /**
   * A function to invoke rating
   */
  private _performRating?: (
    session: UserSession,
    quote: ServerSideQuote,
    suppliers: string[]
  ) => Promise<RateResult>;

  /**
   *
   * @param _response       - server Response
   * @param _dao            - server DAO
   * @param _logger         - log manager
   * @param _encService     - encryption service
   * @param _dapi_processor - data processor
   * @param _ts_ctor        - timestamp constructor
   * @param _feature_flag   - a feature flag object
   * @param _createCleaner  - quote cleaner for migration
   */
  constructor(
    private _response: JsonServerResponse,
    private _dao: ServerDao,
    private _logger: PriorityLog,
    private _encService: any,
    private _dapi_processor: DapiProcessor,
    private _ts_ctor: () => UnixTimestamp,
    private _feature_flag: FeatureFlag,
    private _createCleaner: (
      program: Program
    ) => (quote: ServerSideQuote) => any
  ) {
    super();
  }

  init(cache: any, rater: any): void {
    this._cache = cache;

    this._initDb();
    this.reload(rater);
  }

  /**
   * Register sanitizers to clean data before returning to the client
   *
   * @param sanitizers - an array of sanitizer functions
   */
  registerClientDataSanitizers(
    sanitizers: ((data: BucketData) => void)[]
  ): this {
    this._clientDataSanitizers.push(...sanitizers);

    return this;
  }

  reload(rater: any): void {
    rater.init(
      // log
      (msg: string) => {
        this._logger.log(PriorityLog.PRIORITY_IMPORTANT, msg);
      },

      // error
      (msg: string, stack: string) => {
        stack = stack || '<no stack trace>';

        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          '%s\n-!%s',
          msg,
          stack.replace(/\n/g, '\n-!')
        );
      }
    );
  }

  /**
   * Shut down any active connections and close gracefully
   */
  close(): void {
    this._feature_flag.close();
  }

  /**
   * Initializes the database (attempt DAO connection)
   */
  private _initDb(): void {
    // error listeners
    this._dao
      .on('saveQuoteError', (err, quote_id) => {
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          'Failed to save quote %d: %s',
          quote_id,
          err.message || ''
        );
      })
      .on('seqError', err => {
        this._logger.log(PriorityLog.PRIORITY_ERROR, 'Sequence error: %s', err);
      })
      .on('seqInit', seq => {
        this._logger.log(
          PriorityLog.PRIORITY_DB,
          'Initialized default sequence: %s',
          seq
        );
      })
      .on('ready', () => {
        this._logger.log(
          PriorityLog.PRIORITY_DB,
          '[Server] Connected to database; DAO ready'
        );
      });

    this._logger.log(
      PriorityLog.PRIORITY_DB,
      '[Server] Connecting to database...'
    );

    // Pass an empty function here, we are using events to log success
    this._dao.init(() => {});
  }

  publishQuote(
    request: UserRequest,
    quote: ServerSideQuote,
    is_internal: boolean
  ): void {
    if (!is_internal) {
      this.logAndSendResponse(
        request,
        quote,
        'not allowed for non-internal users'
      );
      return;
    }

    request.consumePostData((data: any) => {
      if (!data || !data.publish) {
        this.logAndSendResponse(request, quote, 'insufficient data received');
        return;
      }

      this._dao
        .publishQuote(quote, data.publish === '1')
        .then(() => this.logAndSendResponse(request, quote))
        .catch(err =>
          this.logAndSendResponse(request, quote, 'dao error: ' + err)
        );
    });
  }

  /**
   * Update document metadata using POST data
   *
   * @param request         - the user request
   * @param quote           - the quote to update
   * @param can_update_meta - whether the session can update meta based on role
   */
  metaUpdate(
    request: UserRequest,
    quote: ServerSideQuote,
    can_update_meta: boolean
  ): void {
    if (!can_update_meta) {
      this.logAndSendResponse(
        request,
        quote,
        'access denied',
        PriorityLog.PRIORITY_INFO
      );
      return;
    }

    request.consumePostData(data => {
      if (!isObject(data)) {
        this.logAndSendResponse(request, quote, 'no data received');
        return;
      }

      const allowed_keys: Record<string, RegExp> = {
        retail_agency: /^[a-z0-9]+$/i,
        liza_access_email: /^[^@]+@[^@]+$/,
      };

      // restrict update to specific meta keys
      const meta_key = data.key + '';
      const meta_value = data.value + '';
      if (
        !Object.keys(allowed_keys).includes(meta_key) ||
        !allowed_keys[meta_key].test(meta_value)
      ) {
        this.logAndSendResponse(request, quote, 'insufficient data received');
        return;
      }

      this._dao.saveQuoteMetaDirect(
        quote,
        {[meta_key]: [meta_value]},
        () => this.logAndSendResponse(request, quote),
        () => this.logAndSendResponse(request, quote, 'write error')
      );
    });
  }

  logAndSendResponse(
    request: UserRequest,
    quote: ServerSideQuote,
    err?: string,
    priority?: number
  ): void {
    const success = err === undefined;
    const msg = 'Quote %s access update: ' + (err || 'success');

    const level =
      priority ||
      (success ? PriorityLog.PRIORITY_INFO : PriorityLog.PRIORITY_ERROR);

    this._logger.log(level, msg, quote.getId());

    request.end(JSON.stringify([{success: +success}]));
  }

  sendResponse(
    request: UserRequest,
    quote: ServerSideQuote,
    data?: any,
    action?: ClientActions
  ): this {
    request.end(this._response.from(quote, data, action));
    return this;
  }

  sendError(
    request: UserRequest,
    message: string,
    action?: ClientActions,
    btn_caption?: string
  ): this {
    request.end(this._response.error(message, action, btn_caption));
    return this;
  }

  /**
   * Initializes a quote with any existing quote data
   *
   * @param quote    - id of the quote
   * @param program  - program that the quote will be a part of
   * @param session  - session to create quote
   * @param kekMap   - key encryption key map
   * @param callback - function to call when quote is ready
   * @param callback - function to call when an error occurs
   *
   * @return Server self to allow for method chaining
   */
  initQuote(
    quote: ServerSideQuote,
    program: Program,
    session: UserSession,
    kekMap: ServiceSecretMap,
    callback: (args?: any) => void,
    error_callback: (args?: any) => void
  ): this {
    // Note that `left` is transitional (we're not importing fp-ts here)
    initExistingDocument(this._dao)(program)(session)(this._feature_flag)(
      quote
    )(<DocumentData>{})(kekMap)().then(({left, right}: any) => {
      const leftArr: any = Array.isArray(left) ? left : [left];
      right = Array.isArray(right) ? right : [right];
      left
        ? error_callback.apply(
            this,
            left === QuoteNotFound ? undefined : leftArr
          )
        : callback.apply(this, right);
    });

    return this;
  }

  private _checkQuotePver(
    quote: ServerSideQuote,
    program: Program,
    callback: any
  ): void {
    // note that if program.version is not set, then something is likely
    // wrong with the build that generates it: always clean in this case to
    // be safe
    if (program.version && quote.getProgramVersion() === program.version) {
      callback(null, false);
      return;
    }

    // TODO: this is transitional; it'll go away
    if (!this._createCleaner) {
      // error out
      callback(Error('Missing cleaner'), false);
      return;
    }

    this._logger.log(
      PriorityLog.PRIORITY_INFO,
      'Quote %s program version change (%s -> %s); will be migrated',
      quote.getId(),
      quote.getProgramVersion(),
      program.version
    );

    // user may be kicked back during migration, so let's record where they
    // left off for later logging
    const pre_migrate_step_id = quote.getCurrentStepId();

    // trigger the event and let someone (hopefully) take care of this
    try {
      this._createCleaner(program)(quote)().then(({left}: any) => {
        if (left) {
          callback(left, true);
          return;
        }

        // Migration to this version of the Program has completed
        quote.setProgramVersion(program.version);

        this._logger.log(
          PriorityLog.PRIORITY_INFO,
          'Quote %s migrated to program version %s (kickback step %d -> %d)',
          quote.getId(),
          program.version,
          pre_migrate_step_id,
          quote.getCurrentStepId()
        );

        // Migration may have caused a kickback
        this._dao.saveQuoteState(quote, () => callback(left, true));
      });
    } catch (e) {
      let err_msg = 'Unknown error checking program version';

      if (e instanceof Error) {
        err_msg = e.message;
      }

      // this is an unhandled exception, as far as we're concerned;
      // re-throw so that we have a stack trace, but do so after we're
      // done processing
      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        'Quote %s scan failed (' + err_msg + ')',
        quote.getId()
      );

      callback(e, false);
    }
  }

  /**
   * Sends a new quote initialization request to the client
   *
   * @param request   - the user request
   * @param quote_new - function to create new quote
   * @param program   - the program to use
   */
  sendNewQuote(
    request: UserRequest,
    program: Program,
    initNewDocument: (
      flags: FeatureFlag
    ) => (doc_data: DocumentData) => TE.TaskEither<unknown, ServerSideQuote>
  ): this {
    const session = request.getSession();
    // XXX: this doesn't handle failures! (this code used to live in
    // #initQuote, and didn't handle failures there either)
    initNewDocument(this._feature_flag)(<DocumentData>{})().then(
      ({left, right: quote}: any) => {
        left
          ? this.sendError(request, 'Failed to create new quote: ' + left)
          : this._dao.saveQuote(
              quote,
              <ClassificationResult>{},
              () => {
                this._dao.saveQuoteMeta(quote, undefined, () => {
                  this.sendResponse(request, quote, {valid: false});
                });
              },
              () => {},
              {
                agentId: '' + session.agentId(),
                agentName: session.agentName(),
                agentEntityId: session.agentEntityId(),
                startDate: this._ts_ctor(),
                programId: quote.getProgramId(),
                initialRatedDate: 0,
                importedInd: quote.isImported() ? 1 : 0,
                boundInd: quote.isBound() ? 1 : 0,
                importDirty: 0,
                syncInd: 0,
                notifyInd: 0,
                syncDate: 0,
                lastPremDate: 0,
                internal: session.isInternal() ? 1 : 0,
                pver: program.version,
                explicitLock: quote.getExplicitLockReason(),
                explicitLockStepId: quote.getExplicitLockStep(),
              }
            );
      }
    );

    return this;
  }

  /**
   * Sends an init response to the client
   *
   * @param request          - the user request
   * @param quote            - the quote to process
   * @param program          - the program to use
   * @param quote_new        - a function to create a new quote
   * @param prev             - whether there is a current access lock
   * @param decryptQuoteData - a function to decrypt quote data
   */
  sendInit(
    request: UserRequest,
    quote: ServerSideQuote,
    program: Program,
    initNewDocument: (
      flags: FeatureFlag
    ) => (doc_data: DocumentData) => TE.TaskEither<unknown, ServerSideQuote>,
    prev: any,
    decryptQuoteData: (
      encrypted_keys: EncryptedKeys,
      encrypted_session_keys: string[],
      to_decrypt: BucketData
    ) => void
  ): this {
    // if no quote id was given, simply prompt for one for now
    if (quote.getId() === 0) {
      this.sendNewQuote(request, program, initNewDocument);
      return this;
    }

    if (quote.getProgramId() !== program.getId()) {
      // invalid program; change the program id
      this.sendResponse(
        request,
        quote,
        {
          valid: false,
        },
        [
          {
            action: 'setProgram',
            id: quote.getProgramId(),
            quoteId: quote.getId(),
          },
        ]
      );

      return this;
    }

    this._checkQuotePver(quote, program, (err: Error, mod: boolean) => {
      if (err) {
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          'Quote ' +
            quote.getId() +
            ' migration failed: ' +
            err.message +
            '; manual pver update and inspection may be required'
        );

        // return as fatal
        this.sendError(request, 'Quote migration failed: ' + err.message);
        return;
      }

      // save the quote updates (but only if it was modified)
      if (mod) {
        const error_cb = (
          quote: ServerSideQuote,
          type: 'save' | 'savemeta',
          err?: Error
        ) => {
          this._logger.log(
            PriorityLog.PRIORITY_ERROR,
            `Quote ${quote.getId()} migration failed to commit for '${type}': ${
              err?.message
            }`
          );

          this.sendError(request, 'Quote migration failed to commit');
        };

        this._dao.saveQuote(
          quote,
          <ClassificationResult>{},
          () => {
            this._dao.saveQuoteMeta(
              quote,
              undefined,
              () => {
                this._processInit(
                  request,
                  quote,
                  program,
                  prev,
                  decryptQuoteData
                ).catch(err =>
                  this.sendError(request, 'Quote init failed: ' + err.message)
                );
              },
              (quote: ServerSideQuote, err?: Error) =>
                error_cb(quote, 'savemeta', err)
            );
          },
          (quote: ServerSideQuote, err?: Error) => error_cb(quote, 'save', err),
          undefined,
          undefined,
          false
        );
      } else {
        this._processInit(
          request,
          quote,
          program,
          prev,
          decryptQuoteData
        ).catch(err =>
          this.sendError(request, 'Quote init failed: ' + err.message)
        );
      }
    });

    return this;
  }

  /**
   * Sends /init response
   *
   * @param request          - the user request
   * @param quote            - quote to process
   * @param program          - program to use
   * @param prev             - whether there is a current access lock
   * @param decryptQuoteData - a function to decrypt quote data
   *
   * @todo generate quote # rather than prompting
   */
  async _processInit(
    request: UserRequest,
    quote: ServerSideQuote,
    program: Program,
    prev: any,
    decryptQuoteData: (
      encrypted_keys: EncryptedKeys,
      encrypted_session_keys: string[],
      to_decrypt: BucketData
    ) => void
  ): Promise<this> {
    let actions: ClientActions = [];
    const program_id = program.getId(),
      session = request.getSession(),
      internal = session.isInternal();

    if (quote.hasError()) {
      this.sendError(request, quote.getError());
      return this;
    }
    // ensure that the user is authorized to access this quote
    else if (quote.isAuthorized(session) === false) {
      // todo: generate a new quote #
      // don't return any quote data if invalid - we don't want people spying
      // on the data!
      this.sendResponse(
        request,
        quote,
        {
          valid: false,
        },
        [{action: 'quotePrompt'}]
      );

      return this;
    }

    const data = quote.getBucket().getData() || {};

    /**
     * if we're not internal, filter out the internal questions from the
     * data array to ensure that they can't spy on our internal data
     *
     * TODO: this is not actually being used currently. It is filtered and then
     * ignored. Unfortunately the internals are currently being used in the
     * client so this cannot be changed at the moment
     * */
    if (internal === false) {
      for (const id in program.internal) {
        delete data[id];
      }
    }

    // Expire quote as needed
    if (quote.hasExpired(this._ts_ctor())) {
      if (program.newDocumentUrl) {
        const message =
          'This quote has expired and cannot be modified. Start a new submission by ' +
          '<a href="' +
          program.newDocumentUrl +
          '">' +
          'clicking here</a>' +
          ' or contact support with any questions.';

        quote.setExplicitLock(message);
      } else {
        let message =
          'This quote has expired and cannot be modified. ' +
          'Please contact support with any questions.';

        if (program.copyQuoteUrl && program.copyQuoteUrl !== '') {
          let url = program.copyQuoteUrl.replace(
            '{quote_id}',
            '' + quote.getId()
          );

          url = '<a href="' + url + '">' + 'clicking here</a>';

          message =
            'This quote has expired and cannot be modified. ' +
            'Create a copy by ' +
            url +
            ' or ' +
            'contact support with any questions.';
        }
        quote.setExplicitLock(message);
      }
    }

    let lock = quote.getExplicitLockReason();
    const bucket = quote.getBucket(),
      lock_step = quote.getExplicitLockStep();

    if (!lock && prev) {
      actions = [
        {
          action: 'warning',
          message:
            'Somebody else is currently viewing this quote; it ' +
            'has been locked and will be read-only until the ' +
            'other person is finished. Please try again later.' +
            (internal
              ? '<br /><br />N.B.: You are an internal user, so you ' +
                'may unlock the quote above, but be warned that ' +
                'concurrent writes may have negative affects on ' +
                'the integrity of the quote.' +
                '<br /><br />' +
                'Currently viewing: ' +
                prev
              : ''),
        },
      ];

      lock = 'Quote is locked due to concurrent access.';
    }

    actions.push({
      action: 'gostep',
      id: quote.getCurrentStepId(),
    });

    const autosave_flag = await this._feature_flag.isEnabled('liza_autosave', {
      user: program.id,
    });

    const encrypted_init_data = bucket.getData() || {};

    this._sanitizeClientData(encrypted_init_data);

    // Do not decrypt data for users authenticated with a session key
    const decrypted_data = request.has_skey()
      ? encrypted_init_data
      : decryptQuoteData(
          program.encryptedKeys,
          quote.getSessionKeys(),
          encrypted_init_data
        );

    // decrypt bucket contents, if necessary, and return
    this._getBucketCipher(program).decrypt(bucket, () => {
      this.sendResponse(
        request,
        quote,
        {
          valid: true,
          data: decrypted_data,

          meta: internal ? quote.getMetabucket().getData() : {},

          programId: program_id,

          currentStepId: quote.getCurrentStepId(),
          topVisitedStepId: quote.getTopVisitedStepId(),
          imported: quote.isImported(),
          bound: quote.isBound(),
          needsImport: quote.needsImport(),
          explicitLock: lock,
          explicitLockStepId: lock_step,
          agentId: quote.getAgentId(),
          agentName: quote.getAgentName(),
          agentEntityId: internal ? quote.getAgentEntityId() : 0,
          startDate: quote.getStartDate(),
          initialRatedDate: quote.getInitialRatedDate(),
          lastPremDate: quote.getLastPremiumDate(),
          autosave: autosave_flag,

          // set to undefined if not internal so it's not included in the
          // JSON response
          internal: internal === true ? true : undefined,
        },
        actions
      );
    });

    return this;
  }

  /**
   * Sanitize data that will be presented to the client
   *
   * Ideally this should delegate that responsibility to subsystems that have a
   * more specific idea of which data are sensitive
   *
   * @param data - the data to sanitize
   */
  private _sanitizeClientData(data: BucketData): void {
    this._clientDataSanitizers.forEach(fn => fn(data));
  }

  /**
   * Sends a step to the client
   *
   * TODO: step_id is not really a PositiveInteger, it is a string
   *
   * @param request - user request
   * @param quote   - quote to send a step for
   * @param program - program to use
   * @param step_id - id of the step
   *
   * @return void
   */
  sendStep(
    request: UserRequest,
    quote: ServerSideQuote,
    program: Program,
    step_id: PositiveInteger,
    session: UserSession
  ): void {
    const cur_id = quote.getCurrentStepId(),
      program_id = program.id;

    if (program.steps[step_id] === undefined) {
      this.sendError(
        request,
        'Invalid step request; step ' + step_id + ' does not exist.',
        [{action: 'gostep', id: cur_id}]
      );

      return;
    }

    const type = program.steps[step_id].type;

    // is this a management step? if so, we must be internal
    if (type === 'manage' && !session.isInternal()) {
      // we're not internal, so let's send them back to the first step
      this.sendResponse(request, quote, {}, [
        {action: 'gostep', id: program.getFirstStepId()},
      ]);

      return;
    }

    let top_allowed_step: PositiveInteger | undefined =
      program.getNextVisibleStep(
        program.classify(quote.getBucket().getData()),
        quote.getTopVisitedStepId()
      );

    if (!top_allowed_step) {
      top_allowed_step = <PositiveInteger>(quote.getTopVisitedStepId() + 1);
    }

    // are they permitted to navigate to this step?
    if (top_allowed_step && step_id > top_allowed_step) {
      // knock them back to the next step they're able to save
      const tostep_id = quote.getTopSavedStepId() + 1;

      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        'Quote ' +
          quote.getId() +
          ' has not yet reached step ' +
          step_id +
          '; forcing to step ' +
          tostep_id +
          '. Top allowed step is ' +
          top_allowed_step
      );

      this.sendError(
        request,
        'Failed to navigate to step: you have not yet reached ' +
          'the requested step.',
        [{action: 'gostep', id: tostep_id}]
      );

      return;
    }

    // perform forward-validations *on the current step* to ensure that they
    // cannot leave the quote and then return, requesting a future step (if
    // permitted), thereby evading client-side forward-validations
    if (step_id > cur_id) {
      const validated = this._forwardValidate(quote, program, cur_id, session);

      if (!validated) {
        this.sendError(
          request,
          'The previous step contains errors; please correct them ' +
            'before continuing.',
          [{action: 'gostep', id: cur_id}]
        );
        return;
      }
    }

    this._cache
      .get('step_html')
      .then((prog: any) => prog.get(program_id))
      .then((shtml: any) => shtml.get(step_id))
      .then((data: any) => {
        // send the step HTML to the client
        this.sendResponse(request, quote, {
          html: data,
        });
      })
      .catch((err: Error) => {
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          "Failed to load program '%s' step %d: %s",
          program_id,
          step_id,
          err.message
        );

        this.sendError(request, 'The step you requested could not be loaded.');

        throw err;
      });

    return;
  }

  /**
   * Step HTML cache miss function
   *
   * Load step HTML from disk.  This is intended to be used as a
   * miss function.
   *
   * TODO: Extract method
   *
   * @param program_id - program containing step
   * @param step_id    - step to load
   *
   * @return html for the requested step
   */
  loadStepHtml(program_id: string, step_id: string): Promise<string> {
    const step_filename =
      process.env.LV_ROOT_PATH +
      '/src/_gen/views/scripts/quote/' +
      program_id +
      '/steps/' +
      step_id +
      '.phtml';

    return new Promise((resolve, reject) => {
      readFile(
        step_filename,
        'utf-8',
        (err: NodeJS.ErrnoException | null, data: string) => {
          // we had a problem with the step
          if (err !== null) {
            reject(err);
            return;
          }

          resolve(data);
        }
      );
    });
  }

  /**
   * Perform forward-validations for a given quote and step
   *
   * This check is necessary to ensure that the client-side events are not
   * bypassed, which is realatively simple to do. For example, one could leave
   * the quote and return at a future step (so long as the operation is
   * otherwise permitted), preventing the `forward' event from triggering on
   * the client (as it is a relative event).
   *
   * @param _quote   - quote to forward-validate
   * @param _program - program to validate against
   * @param _step_id - id of current step (before navigation)
   * @param _session - user session
   *
   * @return validation success/failure
   */
  private _forwardValidate(
    _quote: ServerSideQuote,
    _program: Program,
    _step_id: PositiveInteger,
    _session: UserSession
  ): boolean {
    // TODO: we need cmatch data to pass to `forward'
    // see git history for previous implementation
    return true;
  }

  /**
   * Trigger any beforeLoad callbacks for a given step
   *
   * @param program - program to use
   * @param request - user request
   * @param quote   - quote to use
   * @param step_id - the step to check
   *
   * @returns whether a beforeLoad calllback was invoked
   */
  private _beforeLoad(
    program: Program,
    request: UserRequest,
    quote: ServerSideQuote,
    step_id: PositiveInteger
  ): boolean {
    return program.beforeLoadStep(
      step_id,
      quote.getBucket(),
      (event, _field, value, _index) => {
        switch (event) {
          case 'dapi':
            this._dapi_processor.triggerDapi(program, request, quote, value);
        }
      }
    );
  }

  /**
   * Handle a step visit
   *
   * @param step_id - the step we are visiting
   * @param request - user request
   * @param quote   - the current quote
   * @param program - the current program
   */
  async visitStep(
    step_id: PositiveInteger,
    request: UserRequest,
    quote: ServerSideQuote,
    program: Program
  ): Promise<this> {
    quote.refreshStepDelta(step_id);

    const classes = quote.classify();

    if (
      !program.isManageQuoteStep(step_id) &&
      !program.isStepVisible(classes, <PositiveInteger>+step_id)
    ) {
      this._logger.log(
        PriorityLog.PRIORITY_INFO,
        'Attempt to visit not applicable step %s for quote: %s',
        step_id,
        quote.getId()
      );

      this.sendResponse(request, quote, {}, [
        {action: 'gostep', id: program.getFirstStepId()},
      ]);

      return this;
    }

    let top_allowed_step = program.getNextVisibleStep(
      classes,
      quote.getTopVisitedStepId()
    );

    if (!top_allowed_step) {
      top_allowed_step = incrementPositiveInteger(quote.getTopVisitedStepId());
    }

    this._beforeLoad(program, request, quote, step_id);

    const {
      meta: {fields = {}},
    } = program;

    const eligible_dapis = this._getEligibleDapis(fields, classes);

    const {dapis, meta_update} = this._dapi_processor.processDiff(
      program,
      request,
      quote,
      {},
      undefined,
      eligible_dapis
    );

    this._dapi_processor.monitorMetadataPromise(
      quote,
      dapis,
      meta_update,
      request,
      program
    );

    await this._checkRatingTriggers(request.getSession(), quote, classes);

    // update the quote step, if valid
    if (step_id <= top_allowed_step) {
      quote.setCurrentStepId(step_id);
      this._dao.saveQuoteState(quote, () =>
        this.sendResponse(request, quote, {}, [])
      );
    } else {
      this.sendResponse(request, quote, {}, []);
    }

    return this;
  }

  /**
   * Check rating triggers and determine whether to invoke rating
   *
   * @param session - user session
   * @param quote   - current quote
   * @param classes - classification results
   *
   * @returns actions to send to the client
   */
  private async _checkRatingTriggers(
    session: UserSession,
    quote: ServerSideQuote,
    classes: ClassificationResult
  ): Promise<void> {
    const suppliers_to_rate: string[] = [];

    forEachAggregatedClassSuffix(classes, '--liza-trigger-rate', key =>
      suppliers_to_rate.push(key)
    );

    if (suppliers_to_rate.length > 0) {
      // We need to wait for rating to complete before returning the visit call
      // because otherwise we may call fetchRate too early and receive old
      // results
      await this._performRating?.(session, quote, suppliers_to_rate);
    }
  }

  /**
   * Determine which classification-based dapis are eligible
   *
   * @param fields  - a list of meta fields
   * @param classes - classification data
   *
   * @returns all classification-based dapis and whether they are eligible
   */
  private _getEligibleDapis(
    fields: MetaFields,
    classes: ClassificationResult
  ): Record<string, boolean> {
    return Object.fromEntries(
      Object.keys(fields)
        .filter(field => fields[field].dapi?.when)
        .map(field => [
          field,
          classes?.[fields[field]?.dapi?.when]?.is === true,
        ])
    );
  }

  sendProgramJs(request: UserRequest, program_id: string): this {
    this._cache
      .get('program_js')
      .then((progjs: any) => progjs.get(program_id))
      .then((data: any) => {
        // send the JS to the client
        request.setContentType('text/javascript').end(data);
      })
      .catch((err: Error) => {
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          "Failed to load program '%s' JS: %s",
          program_id,
          err
        );

        this.sendError(request, 'Unable to retrieve program data');

        throw err;
      });

    return this;
  }

  /**
   * Program JS cache miss function
   *
   * Loads program JS from disk.  This is intended to be used as a
   * miss function.
   *
   * TODO: Extract method
   *
   * @param program_id - program to load
   *
   * @return program file js
   */
  loadProgramFiles(program_id: string): Promise<string> {
    let retjs = '';
    const root_path =
        process.env.LV_ROOT_PATH + '/src/_gen/scripts/program/' + program_id,
      js_filename = root_path + '/Program.js',
      inc_filename = root_path + '/include.js';

    return new Promise((resolve, reject) => {
      // read both files
      readFile(
        js_filename,
        'utf8',
        (err: NodeJS.ErrnoException | null, data: string) => {
          if (err !== null) {
            reject(err);
            return;
          }

          // wrap in closure
          data =
            '(function(require,module){' +
            'var exports=module.exports={};' +
            data +
            "\n})(require,modules['program/" +
            program_id +
            "/Program']={});";

          retjs = data;

          // read include file
          readFile(
            inc_filename,
            'utf8',
            (err: NodeJS.ErrnoException | null, data: string) => {
              if (err === null) {
                retjs += data;
              }

              // we have all of our data; return the result
              resolve(retjs);
            }
          );
        }
      );
    });
  }

  /**
   * Handles a quote data post
   *
   * This function is called when an HTTP POST is made to save quote data.
   *
   * @param encryptQuoteData - function to encrypt quote data
   * @param step_id          - id of the step
   * @param request          - request object
   * @param quote            - instance of quote to operate on
   * @param program          - program associated with the quote
   * @param session          - user session
   * @param autosave         - identifies the POST as autosave
   */
  async handlePost(
    encryptQuoteData: (
      encrypted_keys: EncryptedKeys,
      encrypted_session_keys: string[],
      to_encrypt: BucketData
    ) => void,
    decryptField: DecryptField,
    step_id: PositiveInteger,
    request: UserRequest,
    quote: ServerSideQuote,
    program: Program,
    session: UserSession,
    autosave = false
  ): Promise<this> {
    // do not allow quote modification if locked unless logged in as an
    // internal user (FS#5772) and the program is unlockable
    if (
      (quote.isLocked() || step_id < quote.getExplicitLockStep()) &&
      !(session.isInternal() && program.unlockable)
    ) {
      this._logger.log(
        PriorityLog.PRIORITY_INFO,
        'Cannot save imported quote: %s',
        quote.getId()
      );

      this.sendError(
        request,
        'This quote has been locked and can no longer be modified.'
      );

      return this;
    }

    const class_data = quote.classify();

    let top_allowed_step: PositiveInteger | undefined =
      program.getNextVisibleStep(class_data, quote.getTopSavedStepId());

    if (!top_allowed_step) {
      top_allowed_step = <PositiveInteger>(quote.getTopSavedStepId() + 1);
    }

    // are they getting too far ahead of themselves?
    if (step_id > top_allowed_step && !autosave) {
      // knock back to next step that they're able to save
      const tostep_id = top_allowed_step;

      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        'Quote ' +
          quote.getId() +
          ' cannot yet save step ' +
          step_id +
          '; forcing to step ' +
          tostep_id
      );

      this.sendError(
        request,
        'Unable to save step: you have not yet reached ' +
          'the requested step.',
        [{action: 'gostep', id: tostep_id, title: 'Go Back'}]
      );

      // prohibit save
      return this;
    }

    let rdelta_data: Record<string, any> = {};

    request.consumePostData((post_data: any) => {
      // Initialize here so it can be used in quote save call even when
      // there is no post_data.data
      const forward_save = post_data?.forward_save === 'true';
      let second_rating_step_complete = false;

      // fill the quote with the posted data
      if (post_data.data) {
        try {
          // TODO: JSON.parse here is returning an any type. We should have
          //       some sort of validation here on the incoming payload
          const parsed_data = JSON.parse(post_data.data);

          encryptQuoteData(
            program.encryptedKeys || {},
            quote.getSessionKeys(),
            parsed_data
          );

          // This is inappropriate for the quote server. We should not be
          // referencing program specific questions like this. It breaks the
          // abstraction of the program object and causes complexity to build in
          // the system when we try to handle direct edge cases for program
          // specific logic.
          second_rating_step_complete =
            parsed_data.excess_prerate_rating_complete?.length > 0 &&
            parsed_data.excess_prerate_rating_complete[0] === '1';

          const concluding_save =
            forward_save &&
            program.isLastStep(class_data, <PositiveInteger>+step_id);

          const {
            meta: {fields = {}},
          } = program;

          const eligible_dapis =
            Object.keys(parsed_data).length > 0
              ? this._getEligibleDapis(fields, class_data)
              : {};

          const {dapis, meta_update, rdiff} = this._dapi_processor.processDiff(
            program,
            request,
            quote,
            parsed_data,
            undefined,
            eligible_dapis
          );

          // Leave rdelta_data undefined if rdiff is an empty object
          // but if there is a concluding save we still want to send
          // the delta
          if (concluding_save || Object.keys(rdiff).length > 0) {
            rdelta_data = {
              'rdelta.data': {
                data: rdiff,
                concluding_save: concluding_save,
                timestamp: this._ts_ctor(),
                step_id: +step_id,
              },
            };
          }

          // For indication step type, append ratedata to simulate
          // rate step without actually rating
          // Ignore autosave and posts going backwards in the quote
          if (
            program.isIndicationQuoteStep(<PositiveInteger>+step_id) &&
            !autosave &&
            forward_save
          ) {
            rdelta_data['rdelta.ratedata'] = {
              data: {},
              concluding_save: false,
              is_indication: true,
              timestamp: this._ts_ctor(),
              step_id: +step_id,
            };
          }

          if (concluding_save && program.lockOnImport) {
            quote.setImported(true);
          }

          this._dapi_processor.monitorMetadataPromise(
            quote,
            dapis,
            meta_update,
            request,
            program
          );
        } catch (e) {
          this._logger.log(
            PriorityLog.PRIORITY_ERROR,
            'Invalid POST data string (%s): %s',
            e instanceof Error
              ? e
              : new Error('Unknown error checking program version'),
            post_data.data
          );

          this.sendError(
            request,
            'There was an error saving your data. Please ' + 'try again.',
            [{action: 'gostep', id: step_id, title: 'Go Back'}]
          );
        }
      }

      // save the quote
      this._doQuoteSave(
        decryptField,
        step_id,
        request,
        quote,
        program,
        rdelta_data,
        autosave,
        class_data,
        forward_save,
        second_rating_step_complete
      );
    });

    return this;
  }

  /**
   * Provide the server with a function to perform rating
   *
   * @param performRating - the rating function
   */
  hookRating(
    performRating: (
      session: UserSession,
      quote: ServerSideQuote,
      suppliers: string[]
    ) => Promise<RateResult>
  ): this {
    this._performRating = performRating;

    return this;
  }

  private _doQuoteSave(
    decryptField: DecryptField,
    step_id: PositiveInteger,
    request: UserRequest,
    quote: ServerSideQuote,
    program: Program,
    rdelta_data: Record<string, any>,
    autosave: boolean,
    class_data: ClassificationResult,
    forward_save: boolean,
    second_rating_step_complete: boolean
  ): void {
    // whenever they save, we want to make sure we invalidate the premium,
    // unless this is a rating step or autosave
    if (!autosave) {
      const meta: Record<string, unknown> = {};
      const classes = quote.classify();

      forEachAggregatedClassSuffix(
        classes,
        '--liza-clear-rating-results',
        key => {
          const field = key + '_last_rated_date';
          const current = quote.getDataByName(field);
          const new_val = ['0'];

          // If we are changing the value then we must record it in a delta
          if (new_val?.[0] !== current?.[0]) {
            const delta_key = Object.keys(rdelta_data)?.[0] || 'rdelta.data';

            // Initialize the delta if necessary
            if (!rdelta_data[delta_key]) {
              rdelta_data[delta_key] = {
                data: {},
                concluding_save: false,
                timestamp: this._ts_ctor(),
                step_id,
              };
            }

            rdelta_data[delta_key].data[field] = quote.getDataByName(field);
          }

          quote.setData({[field]: new_val});
        }
      );

      const persisted_rates = Object.entries(classes)
        .filter(
          ([key, value]) =>
            /^--liza-clear-rating-results/.test(key) && value.is === false
        )
        .map(([key]) => key.replace(/^--liza-clear-rating-results/, ''))
        .filter(
          key => quote.getDataByName(`${key}_last_rated_date`)[0] !== '0'
        );

      // If all sub fields are cleared, clear the aggregate
      if (persisted_rates.length === 0) {
        quote.setData({last_rated_date: ['0']});
      }

      const is_rating_step = (program.rateSteps || [])[step_id];

      // Clear data on non-rate step changes or moving from one rate step to another.
      // - `rateStep` and `forward_save` only happen together when progressing
      //    to the next rating step.
      // - `excess_prereate_rating_complete` is only set on excess quote step,
      //    so we can use that when moving back to prev rate step.
      if (
        is_rating_step !== true ||
        (is_rating_step === true &&
          (forward_save || second_rating_step_complete))
      ) {
        meta.liza_timestamp_rate_request = [0];
        quote.setLastPremiumDate(<UnixTimestamp>0);
        quote.setRetryAttempts(0);
        quote.setMetadata(meta);
      }

      this._dao.saveQuoteMetaDirect(quote, meta);
    }

    let save_data: Record<string, any> | undefined = undefined;

    // If this is a concluding save and we are going to import, clear any
    // rates that are currently pending
    // TODO: This can probably be cleaned up because we no longer wait for
    // pending rates to finish before publishing rate deltas
    if (rdelta_data && rdelta_data.concluding_save === true) {
      const new_save_data: Record<string, any> = {};

      const quote_data = quote.getBucket().getData();

      Object.keys(quote_data).forEach(
        key => (new_save_data['data.' + key] = quote_data[key])
      );

      new_save_data['ratedata.__rate_pending'] = [];

      save_data = new_save_data;
    }

    this._quoteFill(
      decryptField,
      program,
      quote,
      step_id,
      request.getSession(),
      // success
      () => {
        // encrypt bucket
        const bucket = quote.getBucket();
        this._getBucketCipher(program).encrypt(bucket, () => {
          if (!autosave) {
            // If we are importing then we will set the top saved step to the
            // max so that there are no inhibited deltas
            quote.setTopSavedStepId(
              quote.isImported() ? program.steps.length - 1 : step_id
            );
          }

          // set current and top visited steps
          quote.setCurrentStepId(step_id);
          quote.setTopVisitedStepId(step_id);
          this._dao.saveQuoteState(quote);

          // only reset published indicator on a step save
          const force_publish = !autosave;

          this._dao.saveQuote(
            quote,
            class_data,
            // quote was saved successfully
            () => {
              if (autosave) {
                this.sendResponse(request, quote, {});
              } else {
                this._postSubmit(request, quote, step_id, program);
              }
            },
            // failed to save the quote
            () => {
              // todo: option to allow them to try again
              this.sendError(
                request,
                'There was a problem saving your quote. ' +
                  '<em>The previous step was not saved!</em>',
                [{action: 'lock', value: quote.__self.$('SAVE_LOCK_MSG')}]
              );
            },
            save_data,
            rdelta_data,
            force_publish
          );
        });
      },
      // failure
      (failures: any) => {
        // todo: detailed logging (this shouldn't happen)
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          'Server-side quote data validation failure for ' +
            'quote #%s, program %s, step %d:\n%s',
          quote.getId(),
          program.id,
          step_id,
          inspect(this._formatValidationFailures(failures))
        );

        this.sendError(
          request,
          'There was a problem with the data you entered. ' +
            'Please click "Go Back" below to go back to the ' +
            'previous step and correct the errors.',
          [
            {action: 'gostep', id: step_id},
            {action: 'invalidate', errors: failures},
          ],
          'Go Back'
        );
      },
      autosave
    );
  }

  /**
   * Format validation failures for encoded display
   *
   * That is, output data in a format that is useful for JSON-encoded display.
   *
   * @param failures - failure array per key
   *
   * @return formatted object
   */
  private _formatValidationFailures(failures: any): any {
    return Object.keys(failures).reduce((results: any, id) => {
      results[id] = failures[id].map((failure: any) => failure.toString());
      return results;
    }, {});
  }

  private _postSubmit(
    request: UserRequest,
    quote: ServerSideQuote,
    step_id: PositiveInteger,
    program: Program
  ): void {
    const bucket = quote.getBucket();

    const result = program.postSubmit(
      step_id,
      bucket,
      (event: string, _quote_id: string, value: string) =>
        this._handlePostSubmitEvent(
          request,
          quote,
          step_id,
          program,
          event,
          value
        )
    );

    // if there's no events, then just respond with a generic OK
    if (result === false) {
      this.sendResponse(request, quote);
    }
  }

  /**
   * Handle any post submit events and send response to client
   *
   * @param request  - client request
   * @param quote    - current quote for request
   * @param _step_id - id of submitted step
   * @param program  - associated with the quote
   * @param event    - post submit event to handle
   * @param value    - (optional) value supplied for event
   *
   * @todo: remove from this class
   */
  _handlePostSubmitEvent(
    request: UserRequest,
    quote: ServerSideQuote,
    _step_id: PositiveInteger,
    program: Program,
    event: string,
    value: string
  ): void {
    const internal = request.getSession().isInternal(),
      actions: ClientActions = [];

    const mergeAndFinish = () => {
      // clear any fields scheduled to be cleared on kickback
      const retdata = this._kbclear(program, quote);

      this._dao.mergeBucket(quote, retdata, () => {
        // if we're not internal, strip any potential
        // internal data from the response
        // XXX: maybe we should do this in
        // sendResponse() to ensure consistency
        if (internal === false) {
          for (const id in program.internal) {
            delete retdata[id];
          }
        }

        // don't send the response until the bucket
        // is saved; we don't want a race condition
        // if they're speeding through steps!
        finish(retdata);
      });
    };

    const finish = (data?: CommonObject) => {
      data = data || {};
      this.sendResponse(request, quote, data, actions);
    };

    const to_step_id = +value;

    switch (event) {
      // kick back to the given step, if they're already past it
      case 'kickBack':
        if (quote.getTopVisitedStepId() > to_step_id) {
          // knock them back to the step if they're currently
          // further
          if (quote.getCurrentStepId() > to_step_id) {
            quote.setCurrentStepId(to_step_id);
            actions.push({
              action: 'gostep',
              id: to_step_id,
            });
          }
        }

        mergeAndFinish();
        break;

      // Handle this case only to avoid logging errors
      case 'rate':
        break;

      default:
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          'Unknown postSubmit event: %s',
          event
        );

        finish();
        return;
    }
  }

  private _kbclear(
    program: Program,
    quote: ServerSideQuote
  ): Record<string, any> {
    const set: Record<string, any> = {};

    const class_data = program.classify(quote.getBucket().getData());

    for (const field in program.kbclear) {
      const data = quote.getDataByName(field);

      for (const i in data) {
        const val =
          program.clearNaFields && program.hasNaField(field, class_data, +i)
            ? program.naFieldValue
            : program.defaults[field] || '';

        data[i] = val;
      }

      set[field] = data;
    }

    quote.setData(set);

    // return the fields that have changed
    return set;
  }

  private _quoteFill(
    decryptField: DecryptField,
    program: Program,
    quote: ServerSideQuote,
    step_id: PositiveInteger,
    session: UserSession,
    success: any,
    failure: any,
    autosave: boolean
  ): this {
    const program_id = program.getId();

    // autosave does not need to perform assertions
    if (autosave) {
      success.call(this);
      return this;
    }

    // todo: unnecessary dependency
    const bucket_quote = quote.getBucket().getData(),
      bucket_tmp = new QuoteDataBucket(),
      data_tmp: Record<string, any> = {};

    // this actually takes only 1ms, even with a reasonably
    // sized bucket (tested with snowmobile) - both the copy and
    // setValues()
    for (const item in bucket_quote) {
      if (!Array.isArray(bucket_quote[item])) {
        // this is a problem (FS#5849)
        bucket_quote[item] = [];

        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          "Bucket item '%s' not an array for " +
            'quote id %s in program %s; set to empty',

          item,
          quote.getId(),
          program_id
        );
      }

      data_tmp[item] = bucket_quote[item].slice(0);
    }

    bucket_tmp.setValues(data_tmp);

    // Run all initialization stuff (e.g. calculated
    // values) on the bucket to prepare for
    // assertions. It's important to note that we
    // duplicate the bucket to ensure that none of the
    // calculated values are saved (the ones we want
    // to save are already in there).
    program.initQuote(bucket_tmp);

    const classdata = program.classify(bucket_tmp.getData());
    const cmatch = new FieldClassMatcher(program.whens).match(classdata);

    // WARNING: must set immediately before running
    // assertions, ensuring that stack doesn't clear
    program.isInternal = session.isInternal();

    const failures = program.submit(
      step_id,
      bucket_tmp,
      cmatch,
      undefined,
      decryptField
    );

    if (failures !== null) {
      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        'Server-side validation failure'
      );

      failure.call(this, failures);
      return this;
    }

    success.call(this);
    return this;
  }

  /**
   * Lazily loads and returns the requested Program object
   *
   * @param program_id String id of the program to retrieve
   *
   * @return Server self to allow for method chaining
   */
  getProgram(program_id: string): Promise<Program> {
    return this._cache
      .get('program')
      .then((pcache: any) => pcache.get(program_id))
      .catch((e: Error) => {
        // looks like it doesn't exist
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          "Program class '%s' could not be loaded: %s",
          program_id,
          e.message
        );

        throw e;
      });
  }

  /**
   * Program object cache miss function
   *
   * Instantiates program.  This is intended to be used as a miss
   * function.
   *
   * TODO: Extract method
   *
   * @param program_id - program to instantiate
   *
   * @return a program promise
   */
  loadProgram(program_id: string): Promise<Program> {
    return new Promise((resolve, reject) => {
      try {
        const program_path = 'program/' + program_id + '/Program';

        // node caches modules; make sure it's cleared
        delete require.cache[require.resolve(program_path)];

        // attempt to load the program class
        // eslint-disable-next-line @typescript-eslint/no-var-requires
        const program_module = require(program_path);
        const program: Program = program_module();

        // Clear cached classifier module
        delete require.cache[require.resolve(program.classifierModulePath)];

        // Assert that the program is in a valid state
        // if encryptedKeys is not set then something is wrong and we may be
        // erroneously saving data that should be encrypted
        if (!program.encryptedKeys) {
          throw new Error(`encryptedKeys not set for ${program_id}`);
        }

        resolve(program);
      } catch (e) {
        reject(e);
      }
    });
  }

  private _getBucketCipher(program: Program): any {
    return (
      this._bucketCiphers[program.id] ||
      (() => {
        // create a new bucket cipher
        const c = (this._bucketCiphers[program.id] = QuoteDataBucketCipher(
          this._encService,
          program.secureFields.slice(0) || []
        ));

        c.on('encrecover', (field: string, length: number) => {
          this._logger.log(
            PriorityLog.PRIORITY_ERROR,
            'Invalid encrypted field data (%s of length %d); cleared.',
            field,
            length
          );
        });

        return c;
      })()
    );
  }

  createRevision(request: UserRequest, quote: ServerSideQuote): void {
    this._dao.createRevision(quote, err => {
      if (err) {
        this._logger.log(
          PriorityLog.PRIORITY_DB,
          '[mkrev] failed to create revision: ' + err
        );

        this.sendError(request, 'Failed to create revision.');
        return;
      }

      this._logger.log(
        PriorityLog.PRIORITY_INFO,
        '[mkrev] created new revision for quote %s',
        quote.getId()
      );

      this.sendEmptyReply(request, quote);
    });
  }

  // TODO: currently only diffs against current revision (that is, the live
  // bucket)
  diffRevisionGroup(
    request: UserRequest,
    program: Program,
    quote: ServerSideQuote,
    gid: any,
    revid: any
  ): void {
    const progid = quote.getProgramId();

    // this really should not happen...unless we delete a program, I suppose
    if (program === undefined) {
      this.sendError(request, "Quote program id '" + progid + "' unknown");

      return;
    }

    // get all fields linked to this group---not just the exclusive fields
    const gfields = program.groupFields[gid];
    if (gfields === undefined) {
      this.sendError(
        request,
        "Unknown group '" + gid + "' for program '" + progid + "'"
      );

      return;
    }

    // do we have leaders?
    const lead_data: string | undefined = request.getGetData().leaders;
    if (!lead_data) {
      this.sendError(
        request,
        "No leaders provided for group '" +
          gid +
          "'; available " +
          'fields are: ' +
          gfields.join(', ')
      );

      return;
    }

    const leaders = lead_data.split(',');
    this._dao.getRevision(quote, revid, revdata => {
      if (!revdata) {
        this.sendError(
          request,
          'Revision ' + revid + ' not found for quote ' + quote.getId()
        );
        return;
      }

      const revbucket = new QuoteDataBucket().setValues(revdata.data);

      let diff;

      // XXX: tightly coupled; temporary impl
      try {
        const desc = BucketSiblingDescriptor()
          .defineGroup(gid, gfields)
          .markGroupLeaders(gid, leaders);

        diff = StdBucketDiff(
          ShallowArrayDiff(),
          (context: any, changes: any) => {
            return GroupedBucketDiffResult(
              StdBucketDiffResult(context, changes),
              context
            );
          }
        ).diff(
          GroupedBucketDiffContext(
            StdBucketDiffContext(quote.getBucket(), revbucket),
            desc,
            gid
          )
        );
      } catch (e) {
        let err_msg = 'Unknown revision diff error';

        if (e instanceof Error) {
          err_msg = e.message;
        }

        this.sendError(
          request,
          'An error occurred during processing: ' + err_msg
        );
        throw e;
      }

      this.sendResponse(request, quote, {
        map: diff.createIndexMap(),
        diff: diff.describeChangedValues(),
      });
    });
  }

  sendEmptyReply(request: UserRequest, quote: ServerSideQuote): void {
    this.sendResponse(request, quote, {});
  }

  /**
   * Handle a clear import request
   *
   * @param request - the user request
   * @param quote   - the quote to update
   */
  async clearImport(
    request: UserRequest,
    quote: ServerSideQuote
  ): Promise<this> {
    const session = request.getSession();

    if (!session.isInternal()) {
      this.logAndSendResponse(
        request,
        quote,
        'not allowed for non-internal users'
      );

      return this;
    }

    if (!quote.isImported()) {
      this.logAndSendResponse(
        request,
        quote,
        'cannot clear import on an imported quote'
      );

      return this;
    }

    this._dao.unimportQuote(
      quote,
      session.userName() ?? '',
      (quote: ServerSideQuote) => {
        this.sendResponse(request, quote, {});
      },
      (quote: ServerSideQuote) => {
        this.logAndSendResponse(
          request,
          quote,
          'could not save unlocking metadata'
        );
      }
    );

    return this;
  }
}
