/**
 * Server Factory
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Store} from '../store/Store';
import {ServerDao} from './db/ServerDao';
import {PriorityLog} from './log/PriorityLog';
import {UserSession} from './request/UserSession';
import {Collection} from 'mongodb';
import {DocumentServer} from './DocumentServer';
import {Key} from '../../liza-crypto';
/* eslint-disable @typescript-eslint/no-var-requires */
const MemcachedClient = require('memcached');
/* eslint-enable @typescript-eslint/no-var-requires */

export const createUserSession = (id: string): Promise<UserSession> => {
  return new Promise<UserSession>(resolve => {
    const session = new UserSession(id, createMemcached());

    session.on('ready', () => {
      resolve(session);
    });
  });
};

export const createMemcached = () =>
  new MemcachedClient(
    (process.env.MEMCACHE_HOST || 'localhost') +
      ':' +
      (process.env.MEMCACHE_PORT || 11211)
  );

export const createDocumentServer = (
  dao: ServerDao,
  logger: PriorityLog,
  enc_service: any,
  conf: Store,
  collection: Collection,
  ts_ctor: () => UnixTimestamp,
  gcp_keyfile_path: string,
  liza_kek: Key
) => {
  const origin_url = process.env.HTTP_ORIGIN_URL;

  if (!origin_url) {
    // this allows the system to work without configuration (e.g. for
    // local development), but is really bad
    logger.log(
      PriorityLog.PRIORITY_IMPORTANT,
      '*** HTTP_ORIGIN_URL environment variable not set; ' +
        'system will fall back to using the origin of HTTP requests, ' +
        'meaning an attacker can control where server-side requests go! ***'
    );

    throw new Error('HTTP_ORIGIN_URL not set');
  }

  return new DocumentServer().create(
    dao,
    logger,
    enc_service,
    origin_url,
    conf,
    collection,
    ts_ctor,
    gcp_keyfile_path,
    liza_kek
  );
};
