/**
 * Route controller
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @todo this is a mess of routing and glue code
 */

import {
  getClient,
  getQuotesCollection,
  getSeqCollection,
} from '../../system/db/MongoFactory';
import {BucketData} from '../../bucket/Bucket';
import {Store, storeRetrievalCtor} from '../../store/Store';
import {Server} from '../Server';
import {ServerDao} from '../db/ServerDao';
import {PriorityLog} from '../log/PriorityLog';
import {RatingService} from '../service/RatingService';
import {WebhookService} from '../service/WebhookService';
import {MongoServerDao} from '../db/MongoServerDao';
import {PositiveInteger, ts_ctor} from '../../numeric';
import {createDocumentServer} from '../ServerFactory';
import {WebhookBucketDataApi} from '../../dapi/WebhookBucketDataApi';
import {ServerSideQuote} from '../quote/ServerSideQuote';
import {createQuote, createQuoteQuick} from '../quote/QuoteFactory';
import {UserRequest} from '../request/UserRequest';
import {EncryptedKeys, Program} from '../../program/Program';
import {UnknownDapiError} from '../../error/UnknownDapiError';
import {InvalidTokenError} from '../../error/InvalidTokenError';
import {UnknownTokenError} from '../token/UnknownTokenError';
import {InvalidPayloadError} from '../../error/InvalidPayloadError';
import {makeTokenId, makeTokenNamespace} from '../token/Token';
import {DocumentProgramFormatter as DocumentProgramFormatterType} from '../../document/DocumentProgramFormatter';
import {DocumentIdLockManager} from '../lock/DocumentIdLockManager';
import {ConcurrentSessionManager} from '../request/ConcurrentSessionManager';
import {ProcessManager} from '../rater/ProcessManager';
import {DapiProcessor} from '../dapi/DapiProcessor';
import {KekService, ServiceSecretMap} from '../service/KekService';
import {
  decryptDeltaFieldData,
  decryptorFromKekMap,
  decryptQuoteData,
  encryptorFromKekMap,
  encryptQuoteData,
} from '../service/EncryptionService';
import {assertKeyLength} from '../../../liza-crypto';
import {initNewDocument} from '../quote/loader';
import {getProgramIds} from '../../program/ProgramLoader';
import {MemoryStore} from '../../store/MemoryStore';
import {MissLookup} from '../../store/MissLookup';
import {Cascading} from '../../store/Cascading';
import {UserResponse} from '../request/UserResponse';
import {FieldClassMatcher} from '../../field/FieldClassMatcher';
import {SecretStore} from '../service/SecretStore';
import {
  SecretStoreFactory,
  SecretStoreOptions,
} from '../service/SecretStoreFactory';
import {isArray, isSecretStoreOption, isString} from '../../types/guards';
const DocumentProgramFormatter = require('../../document/DocumentProgramFormatter');
const Semaphore = require('../lock/Semaphore');
/* eslint-enable @typescript-eslint/no-var-requires */

export class Controller {
  /**
   * A base regex for parsing requested routes
   */
  private regex_base =
    /^\/quote\/([a-z0-9-]+)\/?(?:\/(\d+)\/?(?:\/(.*))?|\/(program.js))?$/;

  /**
   * A regex for parsing step routes
   */
  private regex_step = /^step\/(\d+)\/?(?:\/(post|visit))?$/;

  /**
   * Create a new Controller instance
   *
   * @param rater       - process manager for fetching rater processes
   * @param logger      - logger
   * @param enc_service - bucket encryption service
   * @param conf        - config loader
   * @param env         - the environment name
   */
  public static async create(
    rater: ProcessManager,
    logger: PriorityLog,
    enc_service: any,
    conf: Store,
    env: string,
    program_path: string
  ): Promise<Controller> {
    const dbClient = await getClient(process.env);
    const db = await dbClient.db('program');
    const quote_coll = await getQuotesCollection(db);

    const dao = new MongoServerDao(
      quote_coll,
      await getSeqCollection(db),
      env,
      ts_ctor
    );

    const get = storeRetrievalCtor(conf);

    const [gcp_id, oauth_key_path, blacklisted_keks, secret_store_choice] =
      await Promise.all([
        get<string>('services.googleProjectId', isString),
        get<string>('oauth_key_path', isString),
        get<Array<any>>('services.blacklistedKeks', isArray),
        get<SecretStoreOptions>('services.secretStore', isSecretStoreOption),
      ]);

    const gcp_keyfile_path = process.env.LV_ROOT_PATH + '/' + oauth_key_path;

    const secret_store: SecretStore = new SecretStoreFactory(
      gcp_keyfile_path,
      gcp_id
    ).generateSecretStore(secret_store_choice);

    const kek_service = new KekService(
      secret_store,
      assertKeyLength,
      blacklisted_keks
    );
    const liza_kek = await kek_service.fetchInternalKek();

    const {srv, dapi_input_processor, webhook_svc} = await createDocumentServer(
      dao,
      logger,
      enc_service,
      conf,
      quote_coll,
      ts_ctor,
      gcp_keyfile_path,
      liza_kek
    );

    const server_cache = this._createCache(srv);
    srv.init(server_cache, rater);

    let services: string[] = [];
    await this.warmProgramCache(srv, program_path, program => {
      services = services.concat(program.downstreamServices);
    });
    const keks = await kek_service.fetchAll(services);

    return new Controller(
      rater,
      dao,
      srv,
      server_cache,
      new RatingService(logger, dao, rater, ts_ctor),
      webhook_svc,
      new DocumentIdLockManager(Semaphore(), Semaphore()).init(),
      new ConcurrentSessionManager(ts_ctor).init(),
      dapi_input_processor,
      encryptQuoteData((sess_keys: string[]) =>
        encryptorFromKekMap(keks, sess_keys)
      ),
      decryptQuoteData((sess_keys: string[]) =>
        decryptorFromKekMap(keks, sess_keys, KekService.INTERNAL_KEY)
      ),
      (sess_keys: string[]) =>
        decryptDeltaFieldData(
          decryptorFromKekMap(keks, sess_keys, KekService.INTERNAL_KEY)
        ),
      keks
    );
  }

  /**
   *  Using a defined program path, grabs all program ids and attempts to call each program
   *  loading into cache any program not already loaded.
   *
   *  @param srv          - Server object to grab program definitions
   *  @param program_path - Base url to access program folder
   *
   *  @return             - List of loaded programs
   */
  private static async warmProgramCache(
    srv: Server,
    program_path: string,
    cb?: (program: Program) => void
  ): Promise<Program[]> {
    return Promise.all(
      getProgramIds(program_path).map(async id => {
        const program = await srv.getProgram(id);
        cb && cb(program);
        return program;
      })
    );
  }

  /**
   * Create server cache
   *
   * @param rater                - process manager for fetching rater processes
   * @param dao                  - data access object
   * @param server               - server instance
   * @param server_cache         - server cache for looking up programs
   * @param rating_service       - rating service
   * @param webhook_service      - webhook service
   * @param lock_manager         - lock manager
   * @param session_manager      - session manager
   * @param dapi_input_processor - dapi processor
   * @param encryptQuoteData     - encryption function
   * @param decryptQuoteData     - decryption function
   * @param keks                 - key encrypting keys
   */
  constructor(
    private readonly rater: ProcessManager,
    private readonly dao: ServerDao,
    private readonly server: Server,
    private readonly server_cache: Store,
    private readonly rating_service: RatingService,
    private readonly webhook_service: WebhookService,
    private readonly lock_manager: DocumentIdLockManager,
    private readonly session_manager: ConcurrentSessionManager,
    private readonly dapi_input_processor: DapiProcessor,
    private readonly encryptQuoteData: (
      encrypted_keys: EncryptedKeys,
      encrypted_session_keys: string[],
      to_encrypt: BucketData
    ) => void,
    private readonly decryptQuoteData: (
      encrypted_keys: EncryptedKeys,
      encrypted_session_keys: string[],
      to_decrypt: BucketData
    ) => BucketData,
    private readonly decryptField: (sess_keys: string[]) => <T>(x: T) => T,
    private readonly keks: ServiceSecretMap
  ) {}

  /**
   * Initialize the controller
   *
   * TODO: There is async logic here but the caller is not waiting for it to
   *       complete. We should probably ensure that this is complete before
   *       serving any requests from http server.
   */
  public async init() {
    this.server.registerClientDataSanitizers([
      WebhookBucketDataApi.sanitizeClientData,
    ]);

    this.lock_manager.onLockAquired((id: number, request: UserRequest) => {
      // keep the quote session alive
      this.session_manager.touchSession(id, request.getSession());
    });

    this.dapi_input_processor.onDapiReturn(
      (quote_id: number, request: UserRequest, program: Program) => {
        return new Promise<void>((resolve, reject) => {
          // We will be calling this function from within a post call
          // which has its own write lock. This call should be made
          // asynchronously so we avoid deadlocks.
          //
          // It is also necessary that we free this lock manually
          // because the underlying request may have been resolved
          // already
          this.lock_manager.acquireWriteLock(
            quote_id,
            request,
            (free: () => void) => {
              createQuote(
                this.server,
                quote_id,
                program,
                request.getSession(),
                this.keks,
                (quote: ServerSideQuote) => {
                  this.dao
                    .ensurePendingRate(quote)
                    .then(suppliers => {
                      this.rating_service
                        .request(request.getSession(), quote, suppliers)
                        .then(() => {
                          free();
                          resolve();
                        });
                    })
                    .catch(e => {
                      free();
                      reject(e);
                    });
                },
                error => {
                  free();
                  reject(error);
                }
              );
            },
            true
          );
        });
      }
    );

    this.server.hookRating(
      this.rating_service.request.bind(this.rating_service)
    );

    return this;
  }

  close() {
    if (this.server && this.server.close instanceof Function) {
      this.server.close();
    }
  }

  /**
   * Create server cache
   *
   * TODO: This needs to be moved elsewhere; it is a stepping-stone
   * kluge.
   *
   * @param server - server containing miss methods
   */
  private static _createCache(server: Server): Store {
    const progjs_cache = new (MissLookup(
      MemoryStore,
      server.loadProgramFiles.bind(server)
    ))();

    const step_prog_cache = new (MissLookup(MemoryStore, (program_id: string) =>
      Promise.resolve(
        new (MissLookup(
          MemoryStore,
          server.loadStepHtml.bind(server, program_id)
        ))()
      )
    ))();

    const prog_cache = new (MissLookup(
      MemoryStore,
      server.loadProgram.bind(server)
    ))();

    const cache = new (Cascading(MemoryStore))();
    cache.add('program_js', progjs_cache);
    cache.add('step_html', step_prog_cache);
    cache.add('program', prog_cache);

    return cache;
  }

  /**
   * Reload the server and its cache
   */
  public reload() {
    // will cause all steps, programs, etc to be reloaded on demand
    this.server_cache.clear();

    this.server.reload(this.rater);
  }

  /**
   * Route a user request to the appropriate services
   *
   * @param request - user request
   */
  public route(request: UserRequest) {
    let data: RegExpMatchArray | null;

    if (!(data = request.getUri().match(this.regex_base))) {
      // we don't handle this URI
      return Promise.resolve(false);
    }

    // we don't want to cache the responses, as most of them change with each
    // request
    request.noCache();

    const program_id = data[1];

    return this.server.getProgram(program_id).then(program => {
      return new Promise((resolve, reject) => {
        this.doRoute(program, request, data, resolve, reject);
      });
    });
  }

  /**
   * Perform the routing for a request
   *
   * @param program - program class
   * @param request - user request
   * @param data    - data from parsing the route path
   * @param resolve - a function to call on success
   * @param reject  - a function to call on failure
   */
  private doRoute(
    program: Program,
    request: UserRequest,
    data: RegExpMatchArray | null,
    resolve: (value: boolean) => void,
    reject: (error: Error) => void
  ): void {
    // Expect that missing data will kill the process
    // This is only here because later on the same variable is reassigned
    // multiple times
    //
    // TODO: this is terrible, fix this
    if (!data) {
      throw new Error('Routing data not supplied');
    }

    // store our data in more sensible vars
    const program_id = data[1],
      quote_id = +data[2] || 0,
      cmd = data[3] || data[4] || '',
      session = request.getSession();

    // if we were unable to load the program class, that's a problem
    if (program === null) {
      this.server.sendError(
        request,
        'Internal error. Please contact our support team for ' +
          'support.' +
          '<br /><br />Your information has <em>not</em> been saved!'
      );

      resolve(true);
      return;
    }

    // is the user currently logged in?
    if (request.getSession().isLoggedIn() === false && !request.has_skey()) {
      // todo: this is temporary so we don't break our current setup; remove
      // this check once we can error out before we even get to this point
      // (PHP current handles the initial page load)
      if (cmd !== 'program.js' && !/^webhook\//.test(cmd)) {
        this.server.sendError(
          request,
          'Please <a href="/login">click here</a> to log in.'
        );

        resolve(true);
        return;
      }
    }

    // if the session key was provided, mark us as internal
    if (request.has_skey()) {
      request.getSession().setAgentId(900000);
    }

    // we'll be serving all our responses as plain text
    request.setContentType('text/plain');

    // create a quote to represent this request
    const handleRequest = (operation: (quote: ServerSideQuote) => void) => {
      createQuote(
        this.server,
        quote_id,
        program,
        request.getSession(),
        this.keks,
        operation,
        fatal => {
          // if fatal, notify the user and bail out
          if (fatal) {
            // an error occurred; quote invalid
            this.server.sendError(
              request,
              'There was a problem loading this quote; please contact ' +
                'our support team for assistance.'
            );

            return;
          }

          const initDoc = initNewDocument(this.dao)(id =>
            createQuoteQuick(id, program)
          )(program)(session)(this.keks);

          // otherwise, the given quote is invalid, but we can provide a new
          // one
          this.server.sendNewQuote(request, program, initDoc);
        }
      );
    };

    if ((data = cmd.match(this.regex_step))) {
      const step_id = data[1];
      const step_action = data[2] !== undefined ? data[2] : '';

      switch (step_action) {
        case 'post':
          this.lock_manager.acquireWriteLock(quote_id, request, () => {
            handleRequest(quote => {
              // TODO: fix this cast
              this.server.handlePost(
                this.encryptQuoteData,
                this.decryptField(quote.getSessionKeys()),
                <PositiveInteger>(<unknown>step_id),
                request,
                quote,
                program,
                session
              );
            });
          });
          break;

        case 'visit':
          this.lock_manager.acquireRwLock(quote_id, request, () => {
            handleRequest(quote => {
              // TODO: fix this cast
              this.server.visitStep(
                <PositiveInteger>parseInt(step_id),
                request,
                quote,
                program
              );
            });
          });
          break;

        default:
          // send the requested step to the client
          this.lock_manager.acquireRwLock(quote_id, request, () => {
            handleRequest(quote => {
              // TODO: fix this cast
              this.server.sendStep(
                request,
                quote,
                program,
                <PositiveInteger>(<unknown>step_id),
                session
              );
            });
          });
          break;
      }
    } else if (cmd === 'init') {
      const initDoc = initNewDocument(this.dao)(id =>
        createQuoteQuick(id, program)
      )(program)(session)(this.keks);

      this.lock_manager.acquireWriteLock(quote_id, request, () => {
        handleRequest(quote => {
          this.server.sendInit(
            request,
            quote,
            program,
            // for invalid quote requests
            initDoc,
            // concurrent access?
            this.session_manager.getConcurrentSessionUser(quote_id, session),
            this.decryptQuoteData
          );
        });
      });
    } else if (cmd === 'publish') {
      this.lock_manager.acquireWriteLock(quote_id, request, () => {
        handleRequest(quote => {
          this.server.publishQuote(request, quote, session.isInternal());
        });
      });
    } else if (cmd === 'meta') {
      this.lock_manager.acquireWriteLock(quote_id, request, () => {
        handleRequest(quote => {
          this.server.metaUpdate(request, quote, session.canUpdateMeta());
        });
      });
    } else if (cmd === 'clear-import') {
      this.lock_manager.acquireWriteLock(quote_id, request, () => {
        handleRequest(quote => {
          this.server.clearImport(request, quote);
        });
      });
    } else if (cmd === 'progdata') {
      this.lock_manager.acquireReadLock(quote_id, request, () => {
        handleRequest(quote => {
          const response = new UserResponse(request);
          const bucket = quote.getBucket();
          const class_matcher = new FieldClassMatcher(program.whens);

          const formatter: DocumentProgramFormatterType =
            DocumentProgramFormatter(program, class_matcher);

          formatter
            .format(bucket)
            .then(data => response.ok(data))
            .catch(e => response.internalError({}, e));
        });
      });
    } else if (cmd === 'mkrev') {
      // the database operation for this is atomic and disjoint from
      // anything else we're doing, so no need to acquire any sort of
      // lock
      handleRequest(quote => {
        this.server.createRevision(request, quote);
      });
    }
    // TODO: diff against other revisions as well
    else if ((data = cmd.match(/^revdiffgrp\/(.*?)\/(\d+)$/))) {
      // similar to above; no locking needed
      handleRequest(quote => {
        // Fail if we can't parse arguments
        if (!data) {
          request.error('Invalid arguments');
          return;
        }

        const gid = data[1] || '',
          revid = +data[2] || 0;

        this.server.diffRevisionGroup(request, program, quote, gid, revid);
      });
    } else if (cmd === 'program.js') {
      // no quote involved; just send the JS
      this.server.sendProgramJs(request, program_id);
      resolve(true);
      return;
    } else if (cmd === 'fetchRate') {
      this.lock_manager.acquireWriteLock(quote_id, request, () => {
        handleRequest(quote => {
          this.rating_service
            .fetchRate(quote)
            .then(result =>
              this.server.sendResponse(
                request,
                quote,
                result.content,
                result.actions
              )
            )
            .catch(err => {
              this.server.sendError(
                request,
                'There was a problem during the rating process. Unable to ' +
                  'continue. Please contact our support team for assistance.' +
                  // show details for internal users
                  (request.getSession().isInternal()
                    ? '<br /><br />[Internal] ' +
                      err.message +
                      '<br /><br />' +
                      '<hr />' +
                      (err.stack || '').replace(/\n/g, '<br />')
                    : '')
              );
            });
        });
      });
    } else if (/^worksheet\//.test(cmd)) {
      const wdata = cmd.match(/^worksheet\/(.+)\/([0-9]+)/);

      if (!wdata) {
        request.error('Invalid arguments');
        return;
      }

      const supplier = wdata[1],
        index = +wdata[2];

      handleRequest(quote => {
        // TODO: fix this casting
        this.dao
          .getWorksheet(quote.getId(), supplier, <PositiveInteger>index)
          .then(data => {
            this.server.sendResponse(request, quote, data);
          })
          .catch(err => reject(err));
      });
    } else if (cmd === 'autosave') {
      this.lock_manager.acquireWriteLock(quote_id, request, () => {
        handleRequest(quote => {
          this.server.handlePost(
            this.encryptQuoteData,
            this.decryptField(quote.getSessionKeys()),
            quote.getCurrentStepId(),
            request,
            quote,
            program,
            session,
            true
          );
        });
      });
    } else if (/^webhook\//.test(cmd)) {
      const args = cmd.match(/^webhook\/(.+)\/(.+)/);

      if (!args) {
        request.error('Invalid arguments');
        return;
      }

      const field = makeTokenNamespace(args[1]),
        token = makeTokenId(args[2]);

      request.consumePostData(post_data => {
        this.lock_manager.acquireWriteLock(quote_id, request, () => {
          handleRequest(quote => {
            this.webhook_service
              .fulfillWebhook(program, field, token, post_data, quote, request)
              .then(() => request.accepted())
              .catch(e => {
                const err_payload = JSON.stringify({
                  error: e.message,
                  data: {},
                });

                if (e instanceof UnknownDapiError) {
                  request.setResponseCode(404).end(err_payload);
                } else if (
                  e instanceof InvalidTokenError ||
                  e.error instanceof UnknownTokenError
                ) {
                  request.setResponseCode(403).end(err_payload);
                } else if (e instanceof InvalidPayloadError) {
                  request.setResponseCode(400).end(err_payload);
                } else {
                  request.error(e);
                }
              });
          });
        });
      });
    } else if (cmd === 'quicksave') {
      // TODO: we keep this route around only as a heartbeat, for now; the
      // original purpose of this route (to save staged data) has been
      // removed
      handleRequest(quote => {
        this.server.sendEmptyReply(request, quote);
      });

      this.session_manager.touchSession(quote_id, session);
    } else if (/^log\//.test(cmd)) {
      // the "log" URI currently does absolutely nothing; ideally, we'd be
      // able to post to this and log somewhere useful, but for now it
      // just appears in the logs
      handleRequest(quote => {
        this.server.sendEmptyReply(request, quote);
      });
    } else {
      resolve(false);
      return;
    }

    // we handled the request; don't do any additional routing
    resolve(true);
  }
}
