/**
 * Daemon class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

const clienterr = require('./clienterr');
const {ProcessManager} = require('../rater/ProcessManager');
const {UserRequest} = require('../request/UserRequest');
const {PriorityLog} = require('../log/PriorityLog');
const {AccessLog} = require('../log/AccessLog');

const AbstractClass = require('easejs').AbstractClass,
  UserSession = require('../request/UserSession').UserSession,
  {createMemcached} = require('../ServerFactory');

/**
 * Facade handling core logic for the daemon
 *
 * TODO: Factor out unrelated logic
 */
module.exports = AbstractClass('Daemon', {
  /**
   * System configuration
   * @type {Store}
   */
  'private _conf': null,

  /**
   * Server to accept HTTP requests
   * @type {HttpServer}
   */
  'private _httpServer': null,

  /**
   * Path to access log
   * @var {string}
   */
  'private _accessLogPath': '',

  /**
   * Path to debug log
   * @var {string}
   */
  'private _debugLogPath': '',

  /**
   * Access logger
   * @type {AccessLog}
   */
  'private _accessLog': null,

  /**
   * Debug logger
   * @type {DebugLog}
   */
  'private _debugLog': null,

  /**
   * Encryption service
   * @type {EncryptionService}
   */
  'private _encService': null,

  /**
   * Memcache client
   * @type {Object}
   */
  'private _memcache': null,

  /**
   * Routers to use to handle user requests, ordered from most likely to be
   * used to least for performance reasons
   *
   * @type {Array.<Object>}
   */
  'private _routers': null,

  /**
   * Rating service
   * @type {Object}
   */
  'private _rater': null,

  'public __construct': function (conf) {
    this._conf = conf;
  },

  /**
   * Starts initializing the daemon
   *
   * @return {undefined}
   */
  'public start'() {
    return Promise.all([
      this._createDebugLog(),
      this._createAccessLog(),
      this._conf.get('skey'),
    ])
      .then(async ([debug_log, access_log, skey]) => {
        this._debugLog = debug_log;
        this._accessLog = access_log;

        this._httpServer = this.getHttpServer();
        this._rater = new ProcessManager();
        this._encService = this.getEncryptionService();
        this._memcache = this.getMemcacheClient();
        this._routers = await this.getRouters();

        return skey;
      })
      .then(skey => this._startDaemon(skey));
  },

  'private _startDaemon'(skey) {
    this._debugLog.log(
      PriorityLog.PRIORITY_IMPORTANT,
      'Access log path: %s',
      this._accessLogPath
    );

    this._debugLog.log(
      PriorityLog.PRIORITY_IMPORTANT,
      'Debug log path: %s',
      this._debugLogPath
    );

    this._initSignalHandlers();
    this._testEncryptionService(() => {
      this._initMemoryLogger();

      this._initHttpServer(skey, () => {
        this._initUncaughtExceptionHandler();

        // ready to roll
        this._debugLog.log(
          PriorityLog.PRIORITY_INFO,
          'Daemon initialization complete.'
        );
      });
    });
  },

  'protected getDebugLog': function () {
    return this._debugLog;
  },

  'protected getHttpServer': function () {
    return require('./http_server');
  },

  /**
   * Get (and initialize) controller
   *
   * The controller will only be initialized with the session key SKEY
   *
   * @param {PriorityLog}       logger      logger
   * @param {EncryptionService} enc_service bucket encryption service
   * @param {Store}             conf        config loader
   * @param {string}            env         the environment name
   *
   * @return Promise<{Controller}> Promise for a controller
   */
  'protected getProgramController': function (
    logger,
    enc_service,
    conf,
    env,
    program_path
  ) {
    const {Controller} = require('./Controller');

    return Controller.create(
      this._rater,
      logger,
      enc_service,
      conf,
      env,
      program_path
    ).then(controller => controller.init());
  },

  'protected getScriptsController': function () {
    return require('./scripts');
  },

  'protected getClientErrorController': function () {
    return clienterr;
  },

  'protected getUserSession': function (sess_id, memcache) {
    return new UserSession(sess_id, memcache);
  },

  'protected getMemcacheClient': function () {
    return createMemcached()
      .on('issue', details => {
        this._debugLog.log(
          PriorityLog.PRIORITY_IMPORTANT,
          'Memcached error: %s',
          details.messages.join(', ')
        );
      })
      .on('failure', details => {
        this._debugLog.log(
          PriorityLog.PRIORITY_ERROR,
          'Failed to connect to memcached: %s',
          details.messages.join(', ')
        );
      })
      .on('reconnecting', () => {
        this._debugLog.log(
          PriorityLog.PRIORITY_IMPORTANT,
          'Attempting to reconnect to memcached server...'
        );
      })
      .on('reconnect', () => {
        this._debugLog.log(
          PriorityLog.PRIORITY_IMPORTANT,
          'Reconnected to memcached server.'
        );
      });
  },

  'abstract protected getEncryptionService': [],

  'protected getRouters': function () {
    return Promise.all([
      this.getProgramController(
        this._debugLog,
        this._encService,
        this._conf,
        process.env.NODE_ENV || 'local',
        process.env.RATER_PROGRAM_PATH
      ),
      Promise.resolve(this.getScriptsController()),
      Promise.resolve(this.getClientErrorController()),
    ]);
  },

  /**
   * Perform a graceful shutdown
   *
   * @param {string} signal the signal that caused the shutdown
   *
   * @return {undefined}
   */
  'protected shutdown': function (signal) {
    this._debugLog.log(
      PriorityLog.PRIORITY_IMPORTANT,
      'Received %s. Beginning graceful shutdown...',
      signal
    );

    this._debugLog.log(this._debugLog.PRIOIRTY_IMPORTANT, 'Closing routers...');
    this._routers.forEach(router => {
      if (router.close instanceof Function) {
        router.close();
      }
    });

    this._debugLog.log(
      this._debugLog.PRIOIRTY_IMPORTANT,
      'Closing HTTP server...'
    );
    this._httpServer.close();

    this._debugLog.log(
      PriorityLog.PRIORITY_IMPORTANT,
      'Shutdown complete. Exiting...'
    );

    // eslint-disable-next-line no-process-exit
    process.exit();
  },

  'private _createAccessLog': function () {
    return this._conf.get('log.access.path').then(log_path => {
      this._accessLogPath = log_path;
      return new AccessLog(this._accessLogPath);
    });
  },

  'private _createDebugLog': function () {
    return Promise.all([
      this._conf.get('log.priority'),
      this._conf.get('log.debug.path'),
    ]).then(([priority, debug_log_path]) => {
      this._debugLogPath = debug_log_path;

      return new PriorityLog(debug_log_path, priority);
    });
  },

  /**
   * Catches and logs uncaught exceptions to prevent early termination
   *
   * @return {undefined}
   */
  'private _initUncaughtExceptionHandler': function () {
    // chances are, we don't want to crash; we are, after all, a webserver
    process.on('uncaughtException', err => {
      this._debugLog.log(
        PriorityLog.PRIORITY_ERROR,
        'Uncaught exception: %s\n%s',
        err,
        err.stack ? err.stack : '(No stack trace)'
      );

      // should we terminate on uncaught exceptions?
      if (process.env.NODEJS_UCE_TERM) {
        this._debugLog.log(
          PriorityLog.PRIORITY_ERROR,
          'NODEJS_UCE_TERM set; terminating...'
        );

        // SIGINT
        process.kill(process.pid);
      }
    });

    // notify the user of the UCE_TERM flag is set
    if (process.env.NODEJS_UCE_TERM) {
      this._debugLog.log(
        1,
        'NODEJS_UCE_TERM set; ' + 'will terminate on uncaught exceptions.'
      );
    }
  },

  'private _initSignalHandlers': function () {
    // graceful shutdown on SIGINT and SIGTERM (cannot catch SIGKILL)
    try {
      process
        .on('SIGHUP', () => {
          this._debugLog.log(
            PriorityLog.PRIORITY_IMPORTANT,
            'SIGHUP received; requesting reload'
          );

          this._routers.forEach(router => {
            if (router.reload instanceof Function) {
              router.reload();
            }
          });
        })
        .on('SIGINT', () => {
          this.shutdown('SIGINT');
        })
        .on('SIGTERM', () => {
          this.shutdown('SIGTERM');
        });
    } catch (e) {
      console.log('note: signal handling unsupported on this OS');
    }
  },

  'private _testEncryptionService': function (callback) {
    const enc_test = 'test string';

    this._debugLog.log(
      PriorityLog.PRIORITY_INFO,
      'Performing encryption service sanity check...'
    );

    // encryption sanity check to ensure we won't end up working with data
    // that will only become corrupt
    this._encService.encrypt(enc_test, data => {
      this._encService.decrypt(data, data => {
        if (enc_test !== data.toString('ascii')) {
          this._debugLog.log(
            PriorityLog.PRIORITY_ERROR,
            'Encryption service is incompetant. Aborting.'
          );
          // eslint-disable-next-line no-process-exit
          process.exit(1);
        }

        this._debugLog.log(
          PriorityLog.PRIORITY_INFO,
          'Encryption service sanity check passed.'
        );

        callback();
      });
    });
  },

  'private _initMemoryLogger': function () {
    // log memory usage (every 15 min)
    setInterval(() => {
      this._debugLog.log(
        PriorityLog.PRIORITY_IMPORTANT,
        'Memory usage: %s MB (rss), %s/%s MB (V8 heap)',
        (process.memoryUsage().rss / 1024 / 1024).toFixed(2),
        (process.memoryUsage().heapUsed / 1024 / 1024).toFixed(2),
        (process.memoryUsage().heapTotal / 1024 / 1024).toFixed(2)
      );
    }, 900000);
  },

  'private _initHttpServer': function (skey, callback) {
    const _self = this;

    /**
     * Builds UserRequest from the provided request and response objects
     *
     * @param {HttpServerRequest} request
     * @param {HttpServerResponse} response
     *
     * @return {UserRequest} instance to represent current request
     */
    function request_builder(request, response) {
      return new UserRequest(request, response, skey, sess_id => {
        // build a new user session from the given session id
        return _self.getUserSession(sess_id, _self._memcache);
      });
    }

    // create the HTTP server and listen for connections
    try {
      this._httpServer = this.getHttpServer().create(
        this._routers,
        request_builder,
        this._accessLog,
        this._debugLog
      );

      this._conf
        .get('http.port')
        .then(port =>
          this._httpServer.listen(port, () => {
            this._debugLog.log(1, 'Server running on port %d', port);

            callback();
          })
        )
        .catch(e => this._httpError(e));
    } catch (e) {
      this._httpError(e);
    }
  },

  'private _httpError'(e) {
    this._debugLog.log(
      PriorityLog.PRIORITY_ERROR,
      'Unable to start HTTP server: %s',
      e
    );

    // TODO: use daemon-level promise and reject it
    // eslint-disable-next-line no-process-exit
    process.exit(1);
  },
});
