import {SavedDelta} from '../../bucket/delta';
import {NoPendingError} from '../../error/NoPendingError';
import {EmptyBucketData} from '../../bucket/Bucket';
import {BucketData} from '../../bucket/Bucket';
import {MissingParamError} from '../../error/MissingParamError';
import {Program} from '../../program/Program';
import {PriorityLog} from '../log/PriorityLog';
import {FieldData} from '../meta/DapiMetaSource';
import {ServerSideQuote} from '../quote/ServerSideQuote';
import {DataProcessor} from '../request/DataProcessor';
import {UserRequest} from '../request/UserRequest';
import {PendingDapiError} from '../token/PendingDapiError';
import {TriggerPendingIgnoredError} from '../token/TriggerPendingIgnoredError';
import {ServerDao} from '../db/ServerDao';

/* eslint-disable @typescript-eslint/no-var-requires */
const {Class} = require('easejs');

export class DapiProcessor {
  /**
   * A function to call when the dapi returns
   * @type {function}
   */
  private _postDapi: any = null;

  /**
   * Instantiate a new DapiProcessor
   *
   * @param dao            - server DAO
   * @param logger         - log manager
   * @param ts_ctor        - timestamp constructor
   * @param data_processor - data processor
   */
  constructor(
    private _dao: ServerDao,
    private _logger: PriorityLog,
    private _ts_ctor: () => UnixTimestamp,
    private _data_processor: DataProcessor
  ) {}

  public processDiff(
    program: Program,
    request: UserRequest,
    quote: ServerSideQuote,
    parsed_data: Record<string, any>,
    diff_override?: Record<string, any>,
    dapi_classifications: Record<string, boolean> = {}
  ): {
    filtered: BucketData;
    dapis: Promise<FieldData | void>[];
    meta_update: EmptyBucketData;
    rdiff: BucketData;
  } {
    return this._data_processor.processDiff(
      parsed_data,
      request,
      program,
      quote.getBucket(),
      quote,
      diff_override,
      e => {
        const msg =
          e.message || e.stack || `Error thrown with no message or stack: ${e}`;

        if (Class.isA(MissingParamError, e)) {
          this._logger.log(
            PriorityLog.PRIORITY_INFO,
            'Dapi missing params (quote id %d): %s',
            quote.getId(),
            msg
          );
        } else if (e instanceof PendingDapiError) {
          this._logger.log(
            PriorityLog.PRIORITY_INFO,
            'Server-side dapi is pending (quote id %d)',
            quote.getId()
          );
        } else if (e instanceof TriggerPendingIgnoredError) {
          this._logger.log(
            PriorityLog.PRIORITY_INFO,
            'Server-side classification-based dapi was triggered and ignored ' +
              'because it was not in a pending or accepted state (quote id %d)',
            quote.getId()
          );
        } else {
          this._logger.log(
            PriorityLog.PRIORITY_INFO,
            'Failed to save metadata (quote id %d): %s',
            quote.getId(),
            msg
          );
        }
      },
      dapi_classifications
    );
  }

  /**
   * Call the dapis and then perform post-call logic
   *
   * @param quote       - instance of quote to operate on
   * @param dapis       - dapis to call
   * @param meta_update - meta data values to update/clear
   * @param request     - request object
   * @param program     - program associated with the quote
   */
  monitorMetadataPromise(
    quote: ServerSideQuote,
    dapis: Promise<FieldData | void>[],
    meta_update: EmptyBucketData | BucketData,
    request: UserRequest,
    program: Program
  ): void {
    this._dao.mergeData(quote, meta_update, () =>
      dapis.map(promise =>
        promise
          .then((result: FieldData | void) => {
            if (!result) {
              return quote;
            }

            const {data, delta} = result;

            return new Promise<ServerSideQuote>((resolve, reject) => {
              this._dao.saveQuoteMeta(
                quote,
                data,
                _ => {
                  if (!delta) {
                    resolve(quote);

                    return;
                  }

                  const {save_data, rdelta} = DapiProcessor.constructSaveData(
                    quote,
                    delta,
                    this._ts_ctor
                  );

                  this._dao.saveQuote(quote, {}, resolve, reject, save_data, {
                    'rdelta.data': rdelta,
                  });
                },
                e => reject(e)
              );
            });
          })
          .then(quote => this._postDapi(quote.getId(), request, program))
          .catch(e => {
            if (Class.isA(NoPendingError, e)) {
              this._logger.log(
                PriorityLog.PRIORITY_INFO,
                'Did not rate on DAPI return (quote id %d): %s',
                quote.getId(),
                e.message
              );
            } else {
              this._logger.log(
                PriorityLog.PRIORITY_ERROR,
                'Uncaught Dapi Error on quote %s',
                quote.getId(),
                e.message
              );
            }
          })
      )
    );
  }

  /**
   * Build data needed to save the response
   *
   * @param quote   - quote object
   * @param delta   - delta generated by the dapi
   * @param ts_ctor - function to create a timestamp
   *
   * @returns save_data and a reverse delta
   */
  static constructSaveData(
    quote: ServerSideQuote,
    delta: BucketData,
    ts_ctor: () => UnixTimestamp
  ): {save_data: BucketData; rdelta: SavedDelta<BucketData>} {
    const save_data: BucketData = {};
    const rdelta: SavedDelta<BucketData> = {
      data: {},
      concluding_save: false,
      timestamp: ts_ctor(),
      step_id: quote.getCurrentStepId(),
    };

    // Loop through the delta and construct our save data and
    // corresponding reverse delta
    Object.keys(delta).forEach(key => {
      save_data['data.' + key] = delta[key];
      rdelta.data[key] = quote.getDataByName(key);
    });

    return {save_data, rdelta};
  }

  /**
   * Set a function to call when dapis return
   *
   * @param postDapi - a function which will return a promise
   */
  onDapiReturn(postDapi: any): this {
    this._postDapi = postDapi;

    return this;
  }

  /**
   * Trigger a specific dapi
   *
   * @param program   - program to use
   * @param request   - user request
   * @param quote     - quote to use
   * @param dapi_name - the dapi to trigger
   */
  triggerDapi(
    program: Program,
    request: UserRequest,
    quote: ServerSideQuote,
    dapi_name: string
  ): void {
    const meta_field = program.meta.fields[dapi_name];

    if (!meta_field) {
      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        `Cannot trigger unknown dapi ${dapi_name}`
      );

      return;
    }

    const first_mapped_field = Object.values(meta_field.dapi.mapsrc)?.[0];

    if (!first_mapped_field) {
      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        `Cannot trigger dapi ${dapi_name} because it has no defined sources`
      );

      return;
    }

    const {dapis, meta_update} = this.processDiff(
      program,
      request,
      quote,
      {},
      {
        [first_mapped_field]: quote.getDataByName(first_mapped_field),
      }
    );

    this.monitorMetadataPromise(quote, dapis, meta_update, request, program);
  }
}
