/**
 * Facilitates webhooks by setting bucket values
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {GoogleAuth} from 'google-auth-library';
import {TokenId} from '../token/Token';
import {
  DapiAction,
  DataApi,
  DataApiInput,
  DataApiResult,
} from '../../dapi/DataApi';

/**
 * Make an authenticated data api call to Google Cloud Platform (GCP)
 */
export class GcpDataApi implements DataApi {
  /**
   * Initialize Webhook Bucket Data API
   *
   * @param url - the url to call
   * @param api - a data api to decorate
   */
  constructor(
    private readonly _url: string,
    private readonly _auth: GoogleAuth
  ) {}

  /**
   * Request data from the service
   *
   * @param data     - request params
   * @param callback - server response callback
   * @param id       - unique dapi identifier
   * @param action   - the action for a dapi to take
   * @param token    - a token associated with a dapi call
   */
  public request(
    data: DataApiInput,
    callback: NodeCallback<DataApiResult>,
    _id?: string,
    _action?: DapiAction,
    _token?: TokenId
  ): this {
    this._auth
      .getIdTokenClient(this._url)
      .then(client => {
        client.request({url: this._url, data, method: 'POST'}, (err, res) => {
          if (err) {
            return callback(err, null);
          }

          if (!res) {
            return callback(new Error('No response received from GCP'), null);
          }

          // Send a null response indicating that the request was successful
          // but that no data was returned or expected to return
          callback(null, null);
        });
      })
      .catch(err => callback(err, null));

    return this;
  }
}
