/**
 * Facilitates decryption of data by augmenting the call with session keys
 *
 *  Copyright (C) 2010-2025 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {TokenId} from '../token/Token';
import {
  DapiAction,
  DataApi,
  DataApiInput,
  DataApiResult,
} from '../../dapi/DataApi';
import {ServerSideQuote} from '../quote/ServerSideQuote';

/**
 * Sets session keys in the request so that the endpoint that was called can
 * decrypt bucket data
 */
export class SessionKeyDataApi implements DataApi {
  /**
   * Initialize Session Key Data API
   *
   * @param quote              - the quote for the request
   * @param downstreamSystemId - the downstream system ID for the request
   * @param api                - a data api to decorate
   */
  constructor(
    private readonly _quote: ServerSideQuote,
    private readonly _downstreamSystemId: string,
    private readonly _api: DataApi
  ) {}

  /**
   * Request data from the service
   *
   * @param data     - request params
   * @param callback - server response callback
   * @param id       - unique dapi identifier
   * @param action   - the action for a dapi to take
   * @param token    - a token associated with a dapi call
   */
  public request(
    data: DataApiInput,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    action?: DapiAction,
    token?: TokenId
  ): this {
    if (typeof data !== 'object') {
      callback(
        new Error('SessionKey http dapi triggered without a valid data object'),
        null
      );

      return this;
    }

    // The downstream system ID is only required for oAuth DAPIs
    // where the DAPI contains encrypted fields
    if (this._downstreamSystemId) {
      const sessionKeys = this._quote.getDownstreamSessionKeys();

      if (!sessionKeys[this._downstreamSystemId]) {
        callback(
          new Error(
            `No session keys found for DataAPI '${id}' downstream system '${this._downstreamSystemId}'`
          ),
          null
        );

        return this;
      }

      data['downstream_session_keys'] = sessionKeys[this._downstreamSystemId];
    }

    this._api.request(data, callback, id, action, token);

    return this;
  }
}
