/**
 * Facilitates webhooks by setting bucket values
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {TokenId} from '../token/Token';
import {
  DapiAction,
  DataApi,
  DataApiInput,
  DataApiResult,
} from '../../dapi/DataApi';

/**
 * Set bucket values so that downstream systems are able to use these values for
 * triggering requests and responding to the quote server with webhooks
 */
export class WebhookHttpDataApi implements DataApi {
  /**
   * Initialize Webhook Bucket Data API
   *
   * @param api - a data api to decorate
   */
  constructor(private readonly _api: DataApi) {}

  /**
   * Request data from the service
   *
   * @param data     - request params
   * @param callback - server response callback
   * @param id       - unique dapi identifier
   * @param action   - the action for a dapi to take
   * @param token    - a token associated with a dapi call
   */
  public request(
    data: DataApiInput,
    callback: NodeCallback<DataApiResult>,
    id?: string,
    action?: DapiAction,
    token?: TokenId
  ): this {
    if (!token) {
      callback(
        new Error('Webhook http dapi triggered without a valid token'),
        null
      );

      return this;
    }

    if (typeof data !== 'object') {
      callback(
        new Error('Webhook http dapi triggered without a valid data object'),
        null
      );

      return this;
    }

    data[`webhook_trigger_${id}`] = '' + token;

    this._api.request(data, callback, id, action, token);

    return this;
  }
}
