/**
 * Mongo DB DAO for program server
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {
  Collection,
  FilterQuery,
  MongoCallback,
  MongoError,
  UpdateOneOptions,
  UpdateQuery,
  UpdateWriteOpResult,
} from 'mongodb';
import {DocumentId, QuoteId} from '../../document/Document';
import {EventEmitter} from 'events';
import {NoPendingError} from '../../error/NoPendingError';
import {PositiveInteger} from '../../numeric';
import {ServerDao, Callback, FailCallback} from './ServerDao';
import {ServerSideQuote, FieldState} from '../quote/ServerSideQuote';
import {WorksheetData, RateResult} from '../rater/Rater';
import {ClassificationResult} from '../../program/Program';

export type ErrorCallback = (err: NullableError) => void;
export type Revision = any;

/** Bucket data directly from the database */
export type RawBucketData = Record<string, any>;

/**
 * Document ("quote") record
 *
 * This structure was poorly planned and needs both cleanup (much of it is
 * unnecessary or does not belong) and restructuring.
 */
export type DocumentData = {
  agentEntityId?: number;
  agentId?: number;
  agentName?: string;
  boundInd?: 0 | 1;
  creditScoreRef?: string;
  currentStepId?: number;
  data?: RawBucketData;
  error?: string;
  explicitLock?: string;
  explicitLockStepId?: number;
  importDirty?: 0 | 1;
  importedInd?: 0 | 1;
  initialRatedDate?: UnixTimestamp;
  lastPremDate?: UnixTimestamp;
  meta?: RawBucketData;
  pver?: string;
  ratedata?: RateResult;
  retryAttempts?: number;
  startDate: UnixTimestamp;
  topSavedStepId?: number;
  topVisitedStepId?: number;
  fieldState?: FieldState;
  saveError?: boolean;
  programId?: string;
  sessionKeys?: Record<string, string[]>;
  latestKekVersion?: Record<string, number>;
};

export class ServerDaoSaveError extends Error {
  /**
   * @param message - error message
   * @param id      - data model's ID
   */
  constructor(message: string, public readonly id: number) {
    super(message);
    Object.setPrototypeOf(this, new.target.prototype);
    this.name = ServerDaoSaveError.name;
  }
}

/**
 * Convert a callback-based ServerDao to be a promise-based
 *
 * @param dao - relevant DAO
 *
 * @return promise-based API
 */
export type AsyncServerDao = ReturnType<typeof asyncServerDao>;
export const asyncServerDao = (dao: MongoServerDao) => {
  return {
    /**
     * Save the quote
     *
     * @param quote - quote
     * @param data  - saves this data instead of the bucket
     * @param push_data - quote data to push
     *
     * @return promise with quote
     */
    async saveQuote(
      quote: ServerSideQuote,
      data?: CommonObject,
      push_data?: CommonObject
    ): Promise<ServerSideQuote> {
      const on_failure = () => {
        throw new ServerDaoSaveError('Failed to save quote.', quote.getId());
      };

      return new Promise(resolve => {
        dao.saveQuote(quote, {}, resolve, on_failure, data, push_data);
      });
    },

    /**
     * Merges quote data with the existing (rather than overwriting)
     *
     * @param quote - quote
     * @param data  - saves this data instead of the bucket
     *
     * @return promise with quote
     */
    async mergeData(
      quote: ServerSideQuote,
      data: CommonObject
    ): Promise<ServerSideQuote> {
      const on_failure = () => {
        throw new ServerDaoSaveError('Failed to save quote.', quote.getId());
      };

      return new Promise(resolve => {
        dao.mergeData(quote, data, resolve, on_failure);
      });
    },

    /**
     * Save the quote metabucket
     *
     * @param quote    - quote
     * @param new_meta - saves this data instead of the metabucket
     *
     * @return promise with quote
     */
    async saveQuoteMeta(
      quote: ServerSideQuote,
      new_meta?: CommonObject
    ): Promise<ServerSideQuote> {
      const on_failure = () => {
        throw new ServerDaoSaveError(
          'Failed to save quote meta.',
          quote.getId()
        );
      };

      return new Promise(resolve => {
        dao.saveQuoteMeta(quote, new_meta, resolve, on_failure);
      });
    },

    /**
     * Save the quote metabucket directly
     *
     * @param quote    - quote
     * @param new_meta - saves this data instead of the metabucket
     *
     * @return promise with quote
     */
    async saveQuoteMetaDirect(
      quote: ServerSideQuote,
      new_meta?: Record<string, any>,
      single_string_as_empty = false
    ): Promise<ServerSideQuote> {
      const on_failure = () => {
        throw new ServerDaoSaveError(
          'Failed to save quote meta.',
          quote.getId()
        );
      };

      return new Promise(resolve => {
        dao.saveQuoteMetaDirect(
          quote,
          new_meta,
          resolve,
          on_failure,
          {},
          single_string_as_empty
        );
      });
    },

    async saveQuoteState(quote: ServerSideQuote): Promise<ServerSideQuote> {
      const on_failure = () => {
        throw new ServerDaoSaveError(
          'Failed to save quote state',
          quote.getId()
        );
      };

      return new Promise(resolve => {
        dao.saveQuoteState(quote, resolve, on_failure);
      });
    },
  };
};

/**
 * Uses MongoDB as a data store
 */
export class MongoServerDao extends EventEmitter implements ServerDao {
  /** Collection used to store quotes */
  readonly COLLECTION: string = 'quotes';

  /** Sequence (auto-increment) collection */
  readonly COLLECTION_SEQ: string = 'seq';

  /** Sequence key for quote ids */
  readonly SEQ_QUOTE_ID: string = 'quoteId';

  /** Sequence quoteId default */
  readonly SEQ_QUOTE_ID_DEFAULT: number = 200000;

  /** Whether the DAO is initialized and ready to be used */
  private _ready = true;

  /** The document fields to read */
  readonly PULL_FIELDS: Record<string, number> = {
    agentEntityId: 1,
    agentId: 1,
    agentName: 1,
    boundInd: 1,
    creditScoreRef: 1,
    currentStepId: 1,
    data: 1,
    error: 1,
    saveError: 1,
    explicitLock: 1,
    explicitLockStepId: 1,
    importDirty: 1,
    importedInd: 1,
    initialRatedDate: 1,
    lastPremDate: 1,
    meta: 1,
    pver: 1,
    ratedata: 1,
    retryAttempts: 1,
    startDate: 1,
    topSavedStepId: 1,
    topVisitedStepId: 1,
    fieldState: 1,
    programId: 1,
    sessionKeys: 1,
    latestKekVersion: 1,
  };

  /**
   * Initializes DAO
   *
   * @param _quote_collection - collection to save data to
   * @param _seq_collection   - collection to read sequences (auto-increments)
   * @param _env              - the name of the current environment
   * @param _ts_ctor          - a timestamp constructor
   */
  constructor(
    private readonly _quote_collection: Collection,
    private readonly _seq_collection: Collection,
    private readonly _env: string,
    private readonly _ts_ctor: () => UnixTimestamp
  ) {
    super();
  }

  /**
   * Initializes error events and attempts to connect to the database
   *
   * connectError event will be emitted on failure.
   *
   * @param success - function to call when connection is complete
   *                          (will not be called if connection fails)
   */
  init(success: () => void): this {
    this._seq_collection
      .find<any>({_id: this.SEQ_QUOTE_ID})
      .limit(1)
      .next((err: MongoError, data) => {
        if (err || !data) {
          this._initQuoteIdSeq(success);
          this.emit('ready');
          return;
        }

        if (success instanceof Function) {
          success();
        }
      });

    return this;
  }

  /**
   * Initialize the quote id sequence collection
   *
   * @param success - a function to call once it has been initialized
   */
  private _initQuoteIdSeq(success: () => void) {
    this._seq_collection.insertOne(
      {
        _id: this.SEQ_QUOTE_ID,
        val: this.SEQ_QUOTE_ID_DEFAULT,
      },
      (err: NullableError) => {
        if (err) {
          this.emit('seqError', err);
          return;
        }

        this.emit('seqInit', this.SEQ_QUOTE_ID);

        success();
      }
    );
  }

  /**
   * Saves a quote to the database
   *
   * A full save will include all metadata.  This should not cause any
   * problems with race conditions for pending Data API calls on meta
   * fields because those results write to individual indexes and do not
   * rely on existing data.
   *
   * @param quote         - the quote to save
   * @param class_data    - classification data
   * @param success       - function to call on success
   * @param failure       - function to call if save fails
   * @param save_data     - quote data to save (optional)
   * @param push_data     - quote data to push (optional)
   * @param force_publish - reset published indicator (optional)
   */
  saveQuote(
    quote: ServerSideQuote,
    class_data: ClassificationResult,
    success: Callback = () => {},
    failure: FailCallback = () => {},
    save_data?: any,
    push_data?: any,
    force_publish = true
  ): this {
    const id = quote.getId();

    // if we're not ready, then we can't save the quote!
    if (this._ready === false) {
      this.emit('saveQuoteError', {message: 'Database server not ready'}, id);

      failure.call(this, quote, new Error('Not ready'));
      return this;
    }

    if (save_data === undefined) {
      save_data = {};

      const quote_data = quote.getBucket().getData();

      Object.keys(quote_data).forEach(
        key => (save_data['data.' + key] = quote_data[key])
      );
    }

    const save_ts = this._ts_ctor();

    // some data should always be saved because the quote will be created if
    // it does not yet exist
    save_data.id = id;
    save_data.env = this._env;
    save_data.pver = quote.getProgramVersion();
    save_data.importDirty = 1;
    save_data.lastPremDate = quote.getLastPremiumDate();
    save_data.retryAttempts = quote.getRetryAttempts();
    save_data.initialRatedDate = quote.getRatedDate();
    save_data.explicitLock = quote.getExplicitLockReason();
    save_data.explicitLockStepId = quote.getExplicitLockStep();
    save_data.importedInd = +quote.isImported();
    save_data.boundInd = +quote.isBound();
    save_data.lastUpdate = save_ts;
    save_data.quoteSetId = id;

    // It really shouldn't be possible for this to be anything other than false
    // but it is here as a fallback in case we call saveQuote without checking
    // the lock status of the quote
    save_data.saveError = !!quote.getSaveError();

    // Attempt to set these fields if they have not been provided ad hoc
    save_data.startDate = save_data.startDate ?? quote.getStartDate();
    save_data.agentId = '' + (save_data.agentId ?? quote.getAgentId());
    save_data.agentName = save_data.agentName ?? quote.getAgentName();
    save_data.agentEntityId =
      save_data.agentEntityId ?? quote.getAgentEntityId();

    const exp_date = quote.getExpirationDate();
    save_data.quoteExpDate = exp_date === Infinity ? 0 : exp_date;

    // meta will eventually take over for much of the above data
    save_data['meta.liza_timestamp_initial_rated'] = [quote.getRatedDate()];
    const last_updated_by = quote.getUserName();
    if (last_updated_by !== '') {
      save_data['meta.last_updated_by_username'] = [last_updated_by];
      save_data['meta.last_updated_by_roles'] = quote.getUserRoles();
    }

    // save the stack so we can track this call via the oplog
    save_data._stack = new Error().stack;

    // only set created by on initial document insert
    const insert_only = {
      'meta.created_by_username': [quote.getUserName()],
      'meta.created_by_roles': quote.getUserRoles(),
      'meta.retail_agency': [quote.getAgencyNumber()],
    };

    // only set bind username on concluding save
    if (this.shouldUpdateBoundBy(quote, class_data)) {
      save_data['meta.bound_by_username'] = [quote.getUserName()];
      save_data['meta.bound_date_ts'] = save_ts;
      save_data['meta.bound_by_roles'] = quote.getUserRoles();
    }

    save_data['meta.unlocked_by_username'] = save_data[
      'meta.unlocked_by_username'
    ] ?? [''];
    save_data['meta.unlocked_date_ts'] = save_data['meta.unlocked_date_ts'] ?? [
      0,
    ];

    const inc_fields: Record<string, number> = {};

    if (force_publish) {
      /**
       * If we are publishing the quote for the delta processor to pick up:
       *
       * - Set publishResetTs simply as a reference for debugging
       *
       * - Set published to false so the delta processor will fetch this quote
       *
       * - Set publishResetCount so the delta processor can tell that there is
       *   potentially a new delta and it should not mark this quote published
       *   if it is processing it as we save
       */
      save_data.publishResetTs = save_ts;
      save_data.published = false;
      inc_fields['publishResetCount'] = 1;
    }

    save_data.latestKekVersion = quote.getLatestKeks();
    save_data.sessionKeys = {
      ...quote.getDownstreamSessionKeys(),
      liza: quote.getSessionKeys(),
    };

    save_data = this.filterNonpersistent(quote, save_data);

    // do not push empty objects
    const document: UpdateQuery<unknown> = {
      $set: save_data,
      $setOnInsert: insert_only,
    };

    if (Object.keys(inc_fields).length) {
      document['$inc'] = inc_fields;
    }

    if (push_data && Object.keys(push_data).length) {
      document['$push'] = push_data;
    }

    // update the quote data if it already exists (same id), otherwise
    // insert it
    this._performUpdate(
      id,
      {id: id},
      document,

      // create record if it does not yet exist
      {upsert: true},

      // on complete
      err => {
        // if an error occurred, then we cannot continue
        if (err) {
          this.emit('saveQuoteError', err, id);

          // let the caller handle the error
          if (failure instanceof Function) {
            failure(quote, err);
          }

          return;
        }

        // successful
        if (success instanceof Function) {
          success.call(this, quote);
        }
      }
    );

    return this;
  }

  /**
   * Merges quote data with the existing (rather than overwriting)
   *
   * @param quote   - quote to save
   * @param data    - quote data
   * @param success - successful callback
   * @param failure - failure callback
   * @param options - mongo options to specify
   */
  mergeData(
    quote: ServerSideQuote,
    data: Record<string, any>,
    success: Callback = () => {},
    failure: FailCallback = () => {},
    options: UpdateOneOptions = {}
  ): this {
    if (!data || !Object.keys(data).length) {
      success(quote);
      return this;
    }

    // we do not want to alter the original data; use it as a prototype
    const update = this.filterNonpersistent(quote, data);

    const quote_id = quote.getId();

    // save the stack so we can track this call via the oplog
    this._performUpdate(
      quote_id,
      {id: quote_id},
      {$set: update},
      options,

      (err: MongoError) => {
        if (err) {
          this.emit('saveQuoteError', err, quote_id);

          if (typeof failure === 'function') {
            failure(quote, err);
          }

          return;
        }

        if (typeof success === 'function') {
          success(quote);
        }
      }
    );

    return this;
  }

  /**
   * Filter out nonpersistent data
   *
   * @param data The bucket data to filter
   */
  filterNonpersistent(
    quote: ServerSideQuote,
    data: Record<string, any>
  ): Record<string, any> {
    for (const key of quote.getNonpersistent()) {
      delete data['data.' + key];
      delete data['data']?.[key];
    }

    return data;
  }

  /**
   * Merges bucket data with the existing bucket (rather than overwriting the
   * entire bucket)
   *
   * @param quote   - quote to save
   * @param data    - bucket data
   * @param success - successful callback
   * @param failure - failure callback
   */
  mergeBucket(
    quote: ServerSideQuote,
    data: Record<string, any>,
    success: Callback = () => {},
    failure: FailCallback = () => {}
  ): this {
    const update: Record<string, any> = {};

    for (const field in data) {
      if (!field) {
        continue;
      }

      update['data.' + field] = data[field];
    }

    return this.mergeData(quote, update, success, failure);
  }

  /**
   * Saves the quote state to the database
   *
   * The quote state includes the current step, the top visited step and the
   * explicit lock message.
   *
   * @param quote   - the quote to save
   * @param success - function to call on success
   * @param failure - function to call if save fails
   */
  saveQuoteState(
    quote: ServerSideQuote,
    success: Callback = () => {},
    failure: FailCallback = () => {}
  ): this {
    const update: Record<string, any> = {
      currentStepId: quote.getCurrentStepId(),
      topVisitedStepId: quote.getTopVisitedStepId(),
      topSavedStepId: quote.getTopSavedStepId(),
    };

    const field_state = quote.getFieldState();

    // Do not wipe out field state if it's not provided (having
    // getFieldState run the classifier automatically would be dangerous
    // since it is so expensive)
    if (field_state && Object.keys(field_state).length > 0) {
      update.fieldState = field_state;
    }

    return this.mergeData(quote, update, success, failure);
  }

  /**
   * Ensure the quote has pending rates
   *
   * @param quote - the quote to validate
   *
   * @returns a promise with the quote
   */
  ensurePendingRate(quote: ServerSideQuote): Promise<string[]> {
    return new Promise<string[]>((resolve, reject) => {
      this._quote_collection
        .find<any>({id: quote.getId()})
        .limit(1)
        .project({'ratedata.__rate_pending': 1})
        .next((err: MongoError, data) => {
          // was the quote found?
          if (!data || err) {
            // Return the quote unchanged
            reject(new Error('Could not find quote ' + quote.getId()));
            return;
          }

          const ratedata = data.ratedata;

          if (!ratedata) {
            reject(
              new NoPendingError('No prior rate for quote ' + quote.getId())
            );
            return;
          }

          if (
            !ratedata.__rate_pending ||
            !Array.isArray(ratedata.__rate_pending) ||
            ratedata.__rate_pending.length === 0 ||
            ratedata.__rate_pending?.[0] === 0
          ) {
            reject(
              new NoPendingError('No pending rates for quote ' + quote.getId())
            );
            return;
          }

          resolve(ratedata.__rate_pending);
        });
    });
  }

  /**
   * Sets whether the quote is published to the retail agent
   *
   * if published the liza_access_group is set to the retail agency on the quote
   * if unpublished the liza_access_group is cleared
   *
   * @param quote    - destination quote
   * @param publish  - true or false
   * @param fetch_agency - whether retail agency should be accessed from ServerSideQuote quote instance or from the existing mongo document
   */
  async publishQuote(
    quote: ServerSideQuote,
    publish: boolean,
    fetch_agency = false
  ): Promise<any> {
    const update: Record<string, any> = {};
    const save_ts = this._ts_ctor();

    update['meta.liza_timestamp_owner_vis_updated'] = [save_ts];
    update['publishResetTs'] = save_ts;
    update['published'] = false;

    // By default we should get the retail agency from the quote passed into the function itself,
    // but there may be a case when the invoking process of the publishQuote method does not have retail_agency
    // set on the in-memory quote, and we can assume it's already set on the Mongo record.
    const retail_agency = fetch_agency
      ? (await this._quote_collection.findOne({id: quote.getId()})).meta
          .retail_agency
      : quote.getMetabucket().getData().retail_agency;

    if (publish === true) {
      update['meta.liza_access_groups'] =
        retail_agency === undefined ? [] : retail_agency;
    } else {
      update['meta.liza_access_groups'] = [];
    }

    // Create a fake delta so that we make sure that this change is picked up
    // by the delta processor and propagated to downstream systems
    const delta = {
      'rdelta.data': {
        data: {},
        concluding_save: false,
        timestamp: save_ts,
        step_id: quote.getCurrentStepId(),
      },
    };

    return this._performUpdateAsync(
      quote.getId(),
      {id: quote.getId()},
      {$set: update, $push: delta, $inc: {publishResetCount: 1}}
    );
  }

  /**
   * Save document metadata (meta field on document)
   *
   * Only the provided indexes will be modified (that is---data will be
   * merged with what is already in the database).
   *
   * @param quote    - destination quote
   * @param new_meta - bucket-formatted data to write
   * @param success  - callback on success
   * @param failure  - callback on error
   * @param options  - mongo options to specify
   */
  saveQuoteMeta(
    quote: ServerSideQuote,
    new_meta?: Record<string, any>,
    success: Callback = () => {},
    failure: FailCallback = () => {},
    options: UpdateOneOptions = {}
  ): void {
    const update: Record<string, any> = {};

    new_meta = new_meta || quote.getMetabucket().getData();

    for (const key in new_meta) {
      const meta = new_meta[key];

      for (const i in meta) {
        update['meta.' + key + '.' + i] = new_meta[key][i];
      }
    }

    this.mergeData(quote, update, success, failure, options);
  }

  /**
   * Save document metadata directly
   *
   * This saves data directly to the top-level keys of the meta bucket and does
   * not iterate over any values that are iterable. These values will be saved
   * as-is.
   *
   * TODO: Replace saveQuoteMeta with this method when bugs related to not
   * iterating over inner-indexes are alleviated.
   *
   * @param quote    - destination quote
   * @param new_meta - bucket-formatted data to write
   * @param success  - callback on success
   * @param failure  - callback on error
   * @param options  - mongo options to specify
   */
  saveQuoteMetaDirect(
    quote: ServerSideQuote,
    new_meta?: Record<string, any>,
    success: Callback = () => {},
    failure: Callback = () => {},
    options: UpdateOneOptions = {},
    single_string_as_empty = false
  ): void {
    const update: typeof new_meta = {};

    new_meta = new_meta || quote.getMetabucket().getData();

    for (const key in new_meta) {
      // DEV-14039: Don't write empty values to avoid deleting valid data
      // Check for a save of [''] and consider it empty. This is so that we do
      // not overwrite webhook dapi data that may have been saved before this
      // call by a separate process
      if (
        !Array.isArray(new_meta[key]) ||
        new_meta[key].length === 0 ||
        (single_string_as_empty &&
          new_meta[key].length === 1 &&
          new_meta[key][0] === '')
      ) {
        continue;
      }

      update[`meta.${key}`] = new_meta[key];
    }

    this.mergeData(quote, update, success, failure, options);
  }

  /**
   * Saves the quote lock state to the database
   *
   * @param quote   - the quote to save
   * @param success - function to call on success
   * @param failure - function to call if save fails
   */
  saveQuoteLockState(
    quote: ServerSideQuote,
    success: Callback = () => {},
    failure: FailCallback = () => {}
  ): this {
    // lock state is saved by default
    return this.saveQuote(
      quote,
      <ClassificationResult>{},
      success,
      failure,
      {}
    );
  }

  /**
   * Pulls quote data from the database
   *
   * @param quote_id - id of quote
   */
  pullQuote(quote_id: DocumentId): Promise<DocumentData | null> {
    return new Promise((resolve, reject) => {
      this._quote_collection
        .find<DocumentData>({id: quote_id})
        .limit(1)
        .project(this.PULL_FIELDS)
        .next((err, data) => {
          if (err) {
            return reject(err);
          }

          // was the quote found?
          if (!data) {
            return resolve(null);
          }

          // return the quote data
          resolve(data);
        });
    });
  }

  /**
   * Returns the minimum allowable quote id
   *
   * @param callback - function to call with minimum quote id
   */
  getMinQuoteId(callback: (min_id: number) => void): this {
    // just in case it's asynchronous later on
    callback.call(this, this.SEQ_QUOTE_ID_DEFAULT);

    return this;
  }

  /**
   * Returns the current highest quote id
   *
   * @param callback - function to call with current highest quote id
   */
  getMaxQuoteId(callback: (max_id: number) => void): void {
    this._seq_collection
      .find({_id: this.SEQ_QUOTE_ID})
      .limit(1)
      .next((err, data) => {
        if (!data || err) {
          callback(0);
          return;
        }

        // return the max quote id
        callback(data.val);
      });
  }

  /**
   * Returns the next available quote id
   *
   * @param callback - function to call with next available quote id
   */
  getNextQuoteId(): Promise<number> {
    return new Promise<number>((accept, reject) => {
      this._seq_collection.findOneAndUpdate(
        {_id: this.SEQ_QUOTE_ID},
        {$inc: {val: 1}},
        {returnOriginal: false, sort: {val: -1}},
        (err: any, result: any) => {
          if (err || !result.value) {
            this.emit('seqError', err);

            reject(err);
            return;
          }

          // return the new id
          accept(result.value.val);
        }
      );
    });
  }

  /**
   * Create a new revision with the provided quote data
   *
   * The revision will contain the whole the quote. If space is a concern, we
   * can (in the future) calculate a delta instead (Mike recommends the Git
   * model of storing the deltas in previous revisions and the whole of the
   * bucket in the most recently created revision).
   *
   * @param quote    - the quote to create a revision with
   * @param callback - function to call on error
   */
  createRevision(quote: ServerSideQuote, callback: ErrorCallback): void {
    const quote_id = quote.getId();
    this._performUpdate(
      quote_id,
      {id: quote_id},
      {$push: {revisions: {data: quote.getBucket().getData()}}},
      {upsert: true},
      callback
    );
  }

  /**
   * Get a quote revision by its revision id
   *
   * @param quote    - the quote
   * @param revid    - the revision id
   * @param callback - a function to call with results
   */
  getRevision(
    quote: ServerSideQuote,
    revid: PositiveInteger,
    callback: (result: Revision | null) => void
  ): void {
    this._quote_collection
      .find({id: quote.getId()})
      .limit(1)
      .next((err, data) => {
        // was the quote found?
        if (err || !data || data.revisions.length < +revid + 1) {
          callback(null);
          return;
        }

        // return the quote data
        callback(data.revisions[+revid]);
      });
  }

  /**
   * Set worksheet data
   *
   * @param qid      - The quote id
   * @param data     - worksheet data
   * @param callback - a function to call on error
   */
  setWorksheets(
    qid: QuoteId,
    data: Record<string, any>,
    callback: ErrorCallback
  ): void {
    const update: Record<string, unknown> = {};

    for (const field in data) {
      update['worksheets.data.' + field] = data[field];
    }

    this._performUpdate(
      qid,
      {id: qid},
      {$set: update},
      {upsert: true},
      callback
    );
  }

  /**
   * Retrieve worksheet data for a given quote id and supplier
   *
   * @param qid      - the quote id
   * @param supplier - the supplier to retrieve the worksheet for
   * @param index    - the worksheet index
   *
   * @return Promise with worksheet data
   */
  getWorksheet(
    qid: QuoteId,
    supplier: string,
    index: PositiveInteger
  ): Promise<WorksheetData> {
    return new Promise<WorksheetData>((resolve, reject) => {
      this._quote_collection
        .find({id: qid})
        .limit(1)
        .project({'worksheets.data': 1})
        .next((err, result) => {
          if (err) {
            reject(err);
          }

          // was the quote found?
          if (
            !result ||
            !result.worksheets ||
            !result.worksheets.data ||
            !result.worksheets.data[supplier]
          ) {
            reject('Worksheet data not found');
            return;
          }

          resolve({
            data: result.worksheets.data[supplier][index],
          });
        });
    });
  }

  /**
   * Mark a quote as having a save failure
   *
   * @param quote_id - the quote to mark
   * @param cb       - a callback to execute on completion
   */
  private _markSaveFailure(quote_id: DocumentId, cb: () => void): void {
    this._quote_collection.updateOne(
      {id: quote_id},
      {$set: {saveError: true}},
      (err: MongoError) => {
        if (err) {
          this.emit('saveQuoteError', err, quote_id);
          cb();

          return;
        }

        cb();
      }
    );
  }

  /**
   * Mark a quote as having a save failure
   *
   * @param quote_id - the quote to mark
   */
  private _markSaveFailureAsync(
    quote_id: DocumentId
  ): Promise<UpdateWriteOpResult> {
    return this._quote_collection.updateOne(
      {id: quote_id},
      {$set: {saveError: true}}
    );
  }

  /**
   * Perform a mongo update
   *
   * This function is a thin wrapper for mongo's update function. It wraps the
   * update so that we can mark an error state on the quote if there was a save
   * failure
   *
   * @param quote_id - the quote to update
   * @param filter   - the filter used to select the document to update
   * @param update   - the update operations to be applied to the document
   * @param options  - settings to apply to the update operation
   * @param cb       - a callback to be executed on completion
   */
  private _performUpdate<T = unknown>(
    quote_id: DocumentId,
    filter: FilterQuery<T>,
    update: UpdateQuery<T> | Partial<T>,
    options: UpdateOneOptions,
    cb: MongoCallback<UpdateWriteOpResult>
  ) {
    this._quote_collection.updateOne(
      filter,
      update,
      options,
      (err: MongoError, data: UpdateWriteOpResult) => {
        if (err) {
          this.emit('saveQuoteError', err, quote_id);
          this._markSaveFailure(quote_id, () => cb(err, data));

          return;
        }

        cb(err, data);
      }
    );
  }

  /**
   * Perform a mongo update
   *
   * This function is a thin wrapper for mongo's update function. It wraps the
   * update so that we can mark an error state on the quote if there was a save
   * failure
   *
   * @param quote_id - the quote to update
   * @param filter   - the filter used to select the document to update
   * @param update   - the update operations to be applied to the document
   * @param options  - settings to apply to the update operation
   */
  private async _performUpdateAsync<T = any>(
    quote_id: DocumentId,
    filter: FilterQuery<T>,
    update: UpdateQuery<T> | Partial<T>,
    options?: UpdateOneOptions | undefined
  ): Promise<UpdateWriteOpResult> {
    return this._quote_collection
      .updateOne(filter, update, options)
      .catch(async err => {
        this.emit('saveQuoteError', err, quote_id);

        await this._markSaveFailureAsync(quote_id);

        throw new Error(err);
      });
  }

  /**
   * @inheritdoc
   */
  public unimportQuote(
    quote: ServerSideQuote,
    agent_name: string,
    success: Callback,
    failure: FailCallback
  ): this {
    quote.setImported(false);

    this.saveQuote(
      quote,
      <ClassificationResult>{},
      success,
      failure,
      {
        'meta.unlocked_by_username': [agent_name],
        'meta.unlocked_date_ts': [this._ts_ctor()],
      },
      {
        // Create a fake delta so that we make sure that this change is picked up
        // by the delta processor and propagated to downstream systems
        'rdelta.data': {
          data: {},
          concluding_save: false,
          timestamp: this._ts_ctor(),
          step_id: quote.getCurrentStepId(),
        },
      }
    );

    return this;
  }

  private shouldUpdateBoundBy(
    quote: ServerSideQuote,
    class_data: ClassificationResult
  ): boolean {
    const last_updated_by = quote.getUserName();

    if (!quote.isImported() || last_updated_by === '') {
      return false;
    }

    const meta_data = quote.getMetabucket().getData();
    const bound_by = meta_data?.['bound_by_username']?.[0] ?? '';
    const can_update =
      class_data?.['liza-can-update-bound-by-metadata']?.is ?? false;

    return can_update || bound_by === '';
  }
}
