/**
 * General server database operations
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This interface was created to satisfy MongoServerDao and may not be
 * sufficiently general for other database abstractions.
 */

import {WorksheetData} from '../rater/Rater';
import {PositiveInteger} from '../../numeric';
import {DocumentId, QuoteId} from '../../document/Document';
import {ServerSideQuote} from '../quote/ServerSideQuote';
import {ClassificationResult} from '../../program/Program';
import {DocumentData, ErrorCallback, Revision} from './MongoServerDao';
import {EventEmitter} from 'events';
import {UpdateOneOptions} from 'mongodb';

/** Success or failure callback */
export type Callback = (quote: ServerSideQuote, error?: Error) => void;
export type FailCallback = (quote: ServerSideQuote, error: Error) => void;

/**
 * Database abstraction
 */
export interface ServerDao extends EventEmitter {
  /**
   * Initializes error events and attempts to connect to the database
   *
   * connectError event will be emitted on failure.
   *
   * @param success - function to call when connection is complete
   *                          (will not be called if connection fails)
   */
  init(success: () => void): this;

  /**
   * Saves a quote to the database
   *
   * A full save will include all metadata.
   *
   * @param quote     - the quote to save
   * @param success   - function to call on success
   * @param failure   - function to call if save fails
   * @param save_data - quote data to save (optional)
   * @param push_data - quote data to push (optional)
   */
  saveQuote(
    quote: ServerSideQuote,
    class_data: ClassificationResult,
    success?: Callback,
    failure?: FailCallback,
    save_data?: Record<string, any>,
    push_data?: Record<string, any>,
    force_publish?: boolean
  ): this;

  /**
   * Merges quote data with the existing (rather than overwriting)
   *
   * @param quote   - quote to save
   * @param data    - quote data
   * @param success - successful callback
   * @param failure - failure callback
   */
  mergeData(
    quote: ServerSideQuote,
    data: any,
    success?: Callback,
    failure?: FailCallback
  ): this;

  /**
   * Merges bucket data with the existing bucket (rather than overwriting the
   * entire bucket)
   *
   * @param quote   - quote to save
   * @param data    - bucket data
   * @param success - successful callback
   * @param failure - failure callback
   */
  mergeBucket(
    quote: ServerSideQuote,
    data: Record<string, any>,
    success?: Callback,
    failure?: FailCallback
  ): this;

  /**
   * Saves the quote state to the database
   *
   * The quote state includes the current step, the top visited step and the
   * explicit lock message.
   *
   * @param quote   - the quote to save
   * @param success - function to call on success
   * @param failure - function to call if save fails
   */
  saveQuoteState(
    quote: ServerSideQuote,
    success?: Callback,
    failure?: FailCallback
  ): this;

  /**
   * Ensure the quote has pending rates
   *
   * @param quote - the quote to validate
   *
   * @returns a promise with the quote
   */
  ensurePendingRate(quote: ServerSideQuote): Promise<string[]>;

  /**
   * Save document metadata (meta field on document)
   *
   * Only the provided indexes will be modified (that is---data will be
   * merged with what is already in the database).
   *
   * @param quote    - destination quote
   * @param new_meta - bucket-formatted data to write
   * @param success  - callback on success
   * @param failure  - callback on error
   */
  saveQuoteMeta(
    quote: ServerSideQuote,
    new_meta?: Record<string, any>,
    success?: Callback,
    failure?: FailCallback
  ): void;

  /**
   * Saves the quote lock state to the database
   *
   * @param quote   - the quote to save
   * @param success - function to call on success
   * @param failure - function to call if save fails
   */
  saveQuoteLockState(
    quote: ServerSideQuote,
    success?: Callback,
    failure?: FailCallback
  ): this;

  /**
   * Unimports the quote
   *
   * @param quote   - the quote to save
   * @param success - function to call on success
   * @param failure - function to call if save fails
   */
  unimportQuote(
    quote: ServerSideQuote,
    agent_name: string,
    success?: Callback,
    failure?: FailCallback
  ): this;

  /**
   * Set worksheet data
   *
   * @param qid      - The quote id
   * @param data     - worksheet data
   * @param failure  - a function to call on error
   */
  setWorksheets(qid: QuoteId, data: any, failure?: NodeCallback<void>): void;

  /**
   * Retrieve worksheet data
   *
   * @param qid      - the quote id
   * @param supplier - the supplier to retrieve the worksheet for
   * @param index    - the worksheet index
   *
   * @return Promise with worksheet data
   */
  getWorksheet(
    qid: QuoteId,
    supplier: string,
    index: PositiveInteger
  ): Promise<WorksheetData>;

  /**
   * Sets whether the quote is published to the retail agent
   *
   * if published the liza_access_group is set to the retail agency on the quote
   * if unpublished the liza_access_group is cleared
   *
   * @param quote    - destination quote
   * @param publish  - true or false
   * @param fetch_agency - whether retail agency should be accessed from ServerSideQuote quote instance or from the existing mongo document
   */
  publishQuote(
    quote: ServerSideQuote,
    publish: boolean,
    fetch_agency?: boolean
  ): Promise<any>;

  /**
   * Save document metadata directly
   *
   * This saves data directly to the top-level keys of the meta bucket and does
   * not iterate over any values that are iterable. These values will be saved
   * as-is.
   *
   * TODO: Replace saveQuoteMeta with this method when bugs related to not
   * iterating over inner-indexes are alleviated.
   *
   * @param quote    - destination quote
   * @param new_meta - bucket-formatted data to write
   * @param success  - callback on success
   * @param failure  - callback on error
   * @param options  - mongo options to specify
   */
  saveQuoteMetaDirect(
    quote: ServerSideQuote,
    new_meta?: Record<string, any>,
    success?: Callback,
    failure?: Callback,
    options?: UpdateOneOptions,
    single_string_as_empty?: Boolean
  ): void;

  /**
   * Pulls quote data from the database
   *
   * @param quote_id - id of quote
   */
  pullQuote(quote_id: DocumentId): Promise<DocumentData | null>;

  /**
   * Returns the next available quote id
   *
   * @param callback - function to call with next available quote id
   */
  getNextQuoteId(): Promise<number>;

  /**
   * Create a new revision with the provided quote data
   *
   * The revision will contain the whole the quote. If space is a concern, we
   * can (in the future) calculate a delta instead (Mike recommends the Git
   * model of storing the deltas in previous revisions and the whole of the
   * bucket in the most recently created revision).
   *
   * @param quote    - the quote to create a revision with
   * @param callback - function to call on error
   */
  createRevision(quote: ServerSideQuote, callback: ErrorCallback): void;

  /**
   * Get a quote revision by its revision id
   *
   * @param quote    - the quote
   * @param revid    - the revision id
   * @param callback - a function to call with results
   */
  getRevision(
    quote: ServerSideQuote,
    revid: PositiveInteger,
    callback: (result: Revision | null) => void
  ): void;
}
