/**
 * Document ID Lock Manager
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Manage locking behavior for incoming requests
 */
import {Semaphore} from './Semaphore';
import {UserRequest} from '../request/UserRequest';
export class DocumentIdLockManager {
  /**
   * Interval at which to clear stale locks
   */
  public static CLEAR_STALE_INTERVAL_MS = 30e3;

  /**
   * Callback to be called when a lock is acquired
   */
  private onLockAquiredCb: (id: number, request: UserRequest) => void =
    () => {};

  /**
   * Whether we have initialized the manager
   */
  private initialized = false;

  /**
   * Construct a new lock manager
   */
  constructor(
    private readonly rlock: Semaphore,
    private readonly wlock: Semaphore
  ) {}

  /**
   * Initialize the lock manager
   *
   * in the unfortunate situation where a write lock hangs, for whatever
   * reason, this will ensure that it is eventually freed; note that, since
   * this shouldn't ever happen, this interval is 30s, meaning that a given
   * lock may exist for just under 60s
   */
  public init(): this {
    if (this.initialized) {
      return this;
    }

    setInterval(() => {
      // TODO: log properly and possibly kill the request
      // TODO: if the same quote repeatedly has stale locks, perhaps the
      // quote data is bad and should be locked
      this.wlock.freeStale(
        DocumentIdLockManager.CLEAR_STALE_INTERVAL_MS,
        (id: number) => {
          console.log('Freeing stale write lock: ' + id);
        }
      );

      this.rlock.freeStale(
        DocumentIdLockManager.CLEAR_STALE_INTERVAL_MS,
        (id: number) => {
          console.log('Freeing stale read lock: ' + id);
        }
      );
    }, DocumentIdLockManager.CLEAR_STALE_INTERVAL_MS);

    this.initialized = true;

    return this;
  }

  /**
   * Register a callback to be called when a lock is acquired
   *
   * Only one callback will be registered at a time, with newer ones overwriting
   * the prior one.
   *
   * @param cb a callback to be called when a lock is acquired
   */
  public onLockAquired(cb: (id: number, request: UserRequest) => void): void {
    this.onLockAquiredCb = cb;
  }

  /**
   * ALWAYS USE THIS FUNCTION WHEN TRYING TO ACQUIRE BOTH A READ AND A WRITE LOCK!
   * Otherwise, the possibility for a deadlock is introduced if something else is
   * attempting to acquire both locks in the opposite order!
   *
   * Acquire a read-write lock for a quote id
   *
   * @param id      - quote id
   * @param request - user request
   * @param c       - callback once the lock is acquired containing a function
   *                  to release the lock
   * @param manual  - whether or not to manually release the lock
   */
  public acquireRwLock(
    id: number,
    request: UserRequest,
    c: (free: () => void) => void,
    manual?: boolean
  ): void {
    this.acquireWriteLock(
      id,
      request,
      () => {
        this.acquireReadLock(id, request, c);
      },
      manual
    );
  }

  /**
   * Aquire a write lock for a document id
   *
   * @param id      - document id
   * @param request - user request
   * @param c       - callback once the lock is acquired containing a function
   *                  to release the lock
   * @param manual  - whether or not to manually release the lock
   */
  public acquireWriteLock(
    id: number,
    request: UserRequest,
    c: (free: () => void) => void,
    manual?: boolean
  ): void {
    this.acquireLock(this.wlock, id, request, c, manual);
  }

  /**
   * Aquire a read lock for a document id
   *
   * @param id      - document id
   * @param request - user request
   * @param c       - callback once the lock is acquired containing a function
   *                  to release the lock
   * @param manual  - whether or not to manually release the lock
   */
  public acquireReadLock(
    id: number,
    request: UserRequest,
    c: (free: () => void) => void,
    manual?: boolean
  ): void {
    this.acquireLock(this.rlock, id, request, c, manual);
  }

  /**
   * Acquire a semaphore for a quote id
   *
   * Note that, since this controller is single-threaded, we do not have to worry
   * about race conditions with regards to acquiring the lock.
   */
  private acquireLock(
    type: Semaphore,
    id: number,
    request: UserRequest,
    c: (free: () => void) => void,
    manual?: boolean
  ): void {
    type.acquire(id, (free: () => void) => {
      // automatically release the lock once the request completes (this is
      // also safer, as it is hopefully immune to exceptions before lock
      // release and will still work with the timeout system)
      if (!manual) {
        request.once('end', () => {
          free();
        });
      }

      // we're good!
      c(free);
    });

    this.onLockAquiredCb(id, request);
  }
}
