/**
 * Semaphore class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @todo this is actually a mutex; rename?
 */

/**
 * Provides the ability to acquire locks and automatically queues acquisition
 * requests to be executed when the lock becomes available.
 *
 * N.B. This implementation assumes a single-threaded implementation; it must be
 * modified to avoid lock race conditions should a multi-threaded server be
 * used.
 */
export declare class Semaphore {
  /**
   * Attempts to acquire the lock
   *
   * If the lock is available, immediately invokes the provided continuation.
   *
   * If the lock is not available, queues the continuation to be invoked as
   * soon as the lock is freed.
   *
   * Once the lock is successfully acquired, the provided continuation is
   * invoked with a single argument to another continuation that may be
   * invoked in order to free the lock.
   *
   * @param {string|number} id lock id
   *
   * @param {function(function())} c continuation to invoke when lock is
   *                               successfully acquired
   *
   * @return {Semaphore} self
   */
  acquire(id: number, c: (free: () => void) => void): void;

  /**
   * Frees locks that have been acquired for longer than the given amount of
   * time
   *
   * Note that this operation is potentially unsafe; only with sufficient time
   * intervals so as to determine that a request is unlikely to ever free the
   * lock, thus avoiding a deadlock.
   *
   * This operation is synchronous.
   *
   * @param {number} maxtime maximum number of time in milliseconds
   *
   * @param {function(string|number)} c continuation to invoke with id of
   *                                    freed lock(s), if any
   *
   * @return {Semaphore} self
   */
  freeStale(maxtime: number, c: (id: number) => void): this;

  isLocked(id: number): boolean;
}
