/**
 * HTTP access log
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {UserRequest} from '../request/UserRequest';
import {Log} from './Log';

/**
 * Logs HTTP access in Apache's combined log format
 */
export class AccessLog extends Log {
  /**
   * Monitors a user request for the end of the connection and adds an entry
   * to the access log
   *
   * @param user_request - User request to monitor
   */
  public attach(user_request: UserRequest): this {
    if (!(user_request instanceof UserRequest)) {
      throw new TypeError('UserRequest expected, ' + user_request + ' given');
    }

    // log when the request is complete
    user_request.on('end', () => {
      // determine the remote address (in case we're behind a proxy, look
      // at X-Forwarded-For)
      const remote_addr = user_request.getRemoteAddr(),
        username = user_request.getSession().agentId() || '-';

      // access log (apache combined log format)
      this.write(
        '%s %s - - [%s] "%s %s HTTP/%s" %d %d "%s" "%s"',
        remote_addr,
        username,
        new Date(),
        user_request.getMethod(),
        user_request.getUrl(),
        user_request.getHttpVersion(),
        user_request.getResponseCode(),
        user_request.getResponseLength(),
        user_request.getReferrer() || '-',
        user_request.getUserAgent() || '-'
      );
    });

    return this;
  }
}
