/**
 * Generic logger
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

const fs = require('fs'),
  utilFormat = require('util').format;

/**
 * Base logging class
 */
export class Log {
  /**
   * File descriptor (if any)
   */
  private _fd: number | null = null;

  /**
   * Name of file that is being logged to (if any)
   */
  private _filename = '';

  /**
   * Log to standard out
   */
  stdout = true;

  /**
   * Log to standard error
   */
  stderr = false;

  set filename(val) {
    if (val === null) {
      return;
    }

    // must be a string if not null
    const fn = '' + val;

    this._filename = fn;

    // open synchronously (logs are important, yo! - which is why we also
    // aren't going to enclose this in a try/catch block)
    this._fd = fs.openSync(fn, 'a');
  }

  get filename() {
    return this._filename;
  }

  /**
   * Initializes access log
   *
   * @param String filename file to log output to
   *
   * @return undefined
   */
  constructor(filename: string | undefined) {
    if (filename) {
      this.filename = '' + filename;
    }
  }

  /**
   * Writes to log in sprintf-style manner
   */
  public write(...args: unknown[]) {
    // convert arguments to an array
    const args_parsed = Array.prototype.slice.call(args);

    // log to standard out?
    if (this.stdout === true) {
      console.log.apply(this, args_parsed);
    }

    // log to file?
    if (this._fd !== null) {
      const buffer = Buffer.from(utilFormat.apply(this, args_parsed) + '\n');
      fs.writeSync(this._fd, buffer, 0, buffer.length, null);
    }

    return this;
  }
}
