/**
 * Priority log
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Log} from './Log';

/**
 * Logs messages only if they meet the requested priority (allowing for varying
 * levels of verbosity)
 */
export class PriorityLog extends Log {
  static PRIORITY_ERROR = 0;
  static PRIORITY_IMPORTANT = 1;
  static PRIORITY_DB = 2;
  static PRIORITY_INFO = 3;
  static PRIORITY_SOCKET = 5;

  /**
   * Highest priority to log
   * @var Integer
   */
  private _priority = 10;

  /**
   * Initialize logger with filename and priorities to log
   *
   * @param filename - file to log to
   * @param priority - highest priority to log
   */
  constructor(filename: string | undefined, priority: number) {
    // call the parent constructor
    super(filename);

    this._priority = +priority || this._priority;
  }

  /**
   * Write to the log at the given priority
   *
   * If the priority is less than or equal to the set priority for this
   * object, it will be logged. Otherwise, the message will be ignored.
   *
   * The first argument should be the priority. The remaining arguments should
   * be provided in a sprintf()-style fashion
   */
  public log(priority: number, ...args: unknown[]) {
    // don't log if the provided priority is outside the scope that was
    // requested
    if (priority > this._priority) {
      return this;
    }

    // if this was an error, prefix it with the error char (easy grepping)
    let status_char = ' ';
    switch (priority) {
      case PriorityLog.PRIORITY_IMPORTANT:
        status_char = '*';
        break;

      case PriorityLog.PRIORITY_ERROR:
        status_char = '!';
        break;

      case PriorityLog.PRIORITY_DB:
        status_char = '%';
        break;

      case PriorityLog.PRIORITY_SOCKET:
        status_char = '>';
        break;
    }

    // timestamp
    args[0] = status_char + '[' + new Date().toString() + '] ' + args[0];

    // forward to write method
    this.write(...args);

    return this;
  }
}
