/**
 * Data-API-based metadata population
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {PositiveInteger} from '../../numeric';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {DataApiManager} from '../../dapi/DataApiManager';
import {DapiAction, DataApiResult} from '../../dapi/DataApi';
import {ServerSideDapi} from '../../program/Program';
import {StagingBucket} from '../../bucket/StagingBucket';

export type FieldData = {
  field: string;
  index: PositiveInteger;
  data?: Record<string, any>;
  delta?: Record<string, any>;
};

/**
 * Retrieve data for meta field using Data API
 *
 * TODO: The reason this class exists at all is to encapsulate the horrid
 * API.  Once refactored, perhaps this class will no longer be necessary.
 */
export class DapiMetaSource {
  /**
   * Initialize with metabucket constructor
   */
  constructor(private readonly _bucketf: () => QuoteDataBucket) {}

  /**
   * Retrieve field data
   *
   * @param field        - field name
   * @param index        - field index
   * @param dapi_manager - manager for dapi calls
   * @param dapi         - dapi descriptor
   * @param data         - dapi input data
   *
   * @return object containing `field`, `index`, and return data
   */
  getFieldData(
    field: string,
    index: PositiveInteger,
    dapi_manager: DataApiManager,
    dapi: ServerSideDapi,
    data: Record<string, any>,
    action: DapiAction,
    bucket: StagingBucket,
    on_error: (err: Error, name?: string) => void
  ): Promise<FieldData> {
    return new Promise<FieldData>((resolve, reject) => {
      dapi_manager.getApiData(
        dapi.name,
        data,
        (err: Error, api_data: DataApiResult | null) => {
          if (err) {
            on_error(err, dapi.name);
          }

          this.mapDapiResponse(
            field,
            index,
            dapi,
            bucket,
            dapi_manager,
            api_data,
            resolve,
            reject
          );
        },
        field,
        index,
        bucket,
        reject,
        action
      );
    });
  }

  mapDapiResponse(
    field: string,
    index: PositiveInteger,
    dapi: ServerSideDapi,
    bucket: StagingBucket,
    dapi_manager: DataApiManager,
    api_data: DataApiResult | null,
    success_cb: (data: FieldData) => void,
    failure_cb: (reason?: string | Error) => void
  ): void {
    // Treat a null response as valid with no data
    if (api_data === null) {
      success_cb({
        field,
        index,
        data: {},
        delta: bucket.getDiff(),
      });

      return;
    }

    if (api_data?.length !== 1) {
      failure_cb(
        `Data API request for ${field} produced ${api_data?.length} results`
      );
      return;
    }

    const metabucket = this._bucketf();

    dapi_manager.setFieldData(
      dapi.name,
      index,
      api_data,
      dapi.value,
      '',
      false
    );

    dapi_manager.expandFieldData(
      dapi.name,
      index,
      metabucket,
      dapi.mapdest,
      true,
      {
        [dapi.name]: {
          [index]: api_data[0][dapi.value],
        },
      }
    );

    success_cb({
      field: field,
      index: index,
      data: metabucket.getData(),
      delta: bucket.getDiff(),
    });
  }
}
