/**
 * Quote Factory
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {createLayeredBucket} from '../../bucket/layer/createLayeredBucket';
import {Program} from '../../program/Program';
import {Server} from '../Server';
import {UserSession} from '../request/UserSession';
import {ServiceSecretMap} from '../service/KekService';
import {ServerSideQuote} from './ServerSideQuote';

/* eslint-disable @typescript-eslint/no-var-requires */
const ServerSideQuoteCtor = require('./ServerSideQuote');
/* eslint-enable @typescript-eslint/no-var-requires */

/**
 * Creates a new quote instance with the given quote id
 *
 * @param server         - server instance
 * @param quote_id       - id of the quote
 * @param program        - program that the quote will be a part of
 * @param session        - request to create quote
 * @param kekMap         - map of Key Encrypting Keys
 * @param callback       - function to call when quote is ready
 * @param error_callback - function to call when an error occurs
 *
 * @return undefined
 */
export const createQuote = (
  server: Server,
  quote_id: number,
  program: Program,
  session: UserSession,
  kekMap: ServiceSecretMap,
  callback: (quote: ServerSideQuote) => void,
  error_callback: (error?: Error | string) => void
) => {
  const quote = createQuoteQuick(quote_id, program);

  return server.initQuote(
    quote,
    program,
    session,
    kekMap,
    () => callback(quote),
    (error?: Error | string) => error_callback(error)
  );
};

export const createQuoteQuick = (
  id: number,
  program: Program,
  base_bucket?: QuoteDataBucket
) => {
  const bucket = base_bucket ?? new QuoteDataBucket();
  const layeredBucket = createLayeredBucket(bucket, program.bucketLayers);
  const metabucket = new QuoteDataBucket();
  const ratebucket = new QuoteDataBucket();
  const quote = ServerSideQuoteCtor(id, bucket, layeredBucket, program);

  quote.setMetabucket(metabucket);
  quote.setRateBucket(ratebucket);

  return quote;
};
