/**
 * Loads documents into memory
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import * as A from 'fp-ts/Array';
import * as B from 'fp-ts/Bounded';
import * as E from 'fp-ts/Either';
import * as IOE from 'fp-ts/IOEither';
import * as IO from 'fp-ts/IO';
import * as M from 'fp-ts/Monoid';
import * as O from 'fp-ts/Option';
import * as R from 'fp-ts/Record';
import * as T from 'fp-ts/Task';
import * as TE from 'fp-ts/TaskEither';

import {pipe, flow, constant, not} from 'fp-ts/function';
import {between, ordNumber, contramap} from 'fp-ts/Ord';

const {isArray} = Array;

import {DocumentId} from '../../document/Document';
import {Program} from '../../program/Program';
import {DocumentData, RawBucketData} from '../db/MongoServerDao';
import {ServerSideQuote, FieldState} from './ServerSideQuote';
import {ClassData} from '../../client/Cmatch';
import {invalidate} from '../../validate/invalid';
import {UserSession} from '../request/UserSession';
import {ts_ctor} from '../../numeric';
import {FeatureFlag} from '../../system/flags/FeatureFlag';
import {ServerDao} from '../db/ServerDao';
import {
  createSessionKey,
  decryptSessionKey,
  encryptorFromEncryptedSessionKeys,
  encryptSessionKey,
  isPossiblyEncrypted,
  Key,
  ProtectedString,
} from '../../../liza-crypto';
import {
  ServiceSecretMap,
  KekService,
  KekVersionMap,
  Service,
} from '../service/KekService';
import {BucketFieldValue} from '../../bucket/Bucket';
import {
  encryptorFromKekMap,
  maybeEncryptDeltaDatum,
} from '../service/EncryptionService';

type Session = string;
type SessionMap = Record<Service, Session[]>;

// Assigned to a constant because this string is checked to ignore
// as an error to init a new quote
export const QuoteNotFound = 'Quote not found';

/** Pull a document from a datasource using a DAO */
export const pullDocumentFromDao = (dao: ServerDao) =>
  flow(
    TE.tryCatchK(dao.pullQuote.bind(dao), String),
    TE.chainEitherKW(E.fromNullable(QuoteNotFound))
  );

export const applyDocumentDefaults =
  (program: Program) => (keks: ServiceSecretMap) => (doc_data: DocumentData) =>
    pipe(
      IOE.tryCatch<unknown, string[]>(
        () => KekService.getInternalSessionKeys(doc_data.sessionKeys),

        String
      ),
      IOE.chain(session_keys =>
        pipe(
          encryptorFromKekMap(keks, session_keys),
          enc =>
            applyBucketDefaults(program)(doc_data.data || {})((x: string) =>
              enc.encrypt(x)
            ),
          getBucketData => getBucketData(),
          bucket_data => () => {
            doc_data.data = bucket_data;
            return doc_data;
          },
          IOE.fromIO
        )
      )
    );

/**
 * Retrieve next available quote id from DB, with side-effects
 *
 * This will increment the counter in the database atomically.
 *
 * Other languages have conventions for side-effect-laden functions (e.g. a
 * '!' suffix in Scheme), which we don't have here, so this is instead
 * verbose to try to indicate what it's doing.
 */
export const incrementAndPersistNextQuoteId = (dao: ServerDao) =>
  TE.tryCatch(dao.getNextQuoteId.bind(dao), E.toError);

/**
 * Apply default bucket data for the given program
 *
 * This is a near-verbatim migration of what was `ProgramInit`; it still
 * needs to be converted into a more functional style.
 */
export const applyBucketDefaults =
  (program: Program) =>
  (data: RawBucketData) =>
  (encryptor: (x: string) => string): IO.IO<RawBucketData> =>
  () => {
    const {
      defaults = {},
      meta: {groups = {}},
    } = program;

    Object.keys(program.groupExclusiveFields).forEach(group => {
      let i = program.groupExclusiveFields[group].length;

      while (i--) {
        const field = program.groupExclusiveFields[group][i];

        // If the default value is in the list of encrypted keys
        // encrypt the value before setting
        const init_value = program.encryptedKeys[field]
          ? maybeEncryptDeltaDatum(encryptor)(defaults[field])
          : defaults[field];

        // generated questions with no types should never be part of
        // the bucket
        if (!program.hasKnownType(field)) {
          continue;
        }

        if (defaults[field] === undefined) {
          continue;
        }

        // If no document data, initialize with default value
        if (data[field] === undefined) {
          data[field] = [init_value];
        }

        // If min rows on the group is greater than the data
        // currently in the bucket, then populate the rest
        // of the data with the default data until the
        // arrays are the same length
        if (
          groups[group] !== undefined &&
          data[field].length < groups[group].min
        ) {
          let index = data[field].length;

          while (index < groups[group].min) {
            data[field][index] = init_value;
            index++;
          }
        }
      }
    });

    return data;
  };

/**
 * Populate a document (quote) with session data, giving precedence to data
 * already present on the document
 */
export const loadSessionIntoQuote =
  (session: UserSession) =>
  (quote: ServerSideQuote) =>
  (data: DocumentData): IO.IO<ServerSideQuote> =>
  () =>
    quote
      .setAgentId(data.agentId || session.agentId() || 0)
      .setUserName(session.userName() || '')
      .setUserRoles(session.userRoles() || [])
      .setAgencyNumber(session.agencyNumber() || '')
      .setAgentName(data.agentName || session.agentName() || '')
      .setData({
        __internal_user: [session.isInternal() ? '1' : '0'],
        __liza_access_roles: session.userRoles() || [],
      });

/**
 * Populate a document (quote) with data returned from the database
 */
export const loadDocumentIntoQuote =
  (quote: ServerSideQuote) =>
  (quote_data: DocumentData): IO.IO<ServerSideQuote> =>
  () =>
    quote
      .setProgramId(quote_data.programId || '')
      .setLastPersistedFieldState(quote_data.fieldState || undefined)
      .setMetadata(quote_data.meta || {})
      .setAgentEntityId(quote_data.agentEntityId || 0)
      .setInitialRatedDate(quote_data.initialRatedDate || 0)
      .setStartDate(quote_data.startDate || 0)
      .setImported(!!quote_data.importedInd)
      .setBound(!!quote_data.boundInd)
      .needsImport(!!quote_data.importDirty)
      .setCurrentStepId(quote_data.currentStepId || 0)
      .setTopVisitedStepId(quote_data.topVisitedStepId || 0)
      .setTopSavedStepId(quote_data.topSavedStepId || 0)
      .setProgramVersion(quote_data.pver || '')
      .setExplicitLock(
        quote_data.explicitLock || '',
        quote_data.explicitLockStepId || 0
      )
      .setError(quote_data.error || '')
      .setSaveError(quote_data.saveError || false)
      .setCreditScoreRef(quote_data.creditScoreRef || '')
      .setLastPremiumDate(quote_data.lastPremDate || <UnixTimestamp>0)
      .setRatedDate(quote_data.initialRatedDate || <UnixTimestamp>0)
      .setRatingData(quote_data.ratedata || {})
      .setRetryAttempts(quote_data.retryAttempts || 0)
      .setSessionKeys(quote_data.sessionKeys?.liza || [])
      .setLatestKeks(quote_data.latestKekVersion || {})
      .setDownstreamSessionKeys(quote_data.sessionKeys || {})
      .setData({
        ...(quote_data.data || {}),
        __liza_step_id_current: [(quote_data.currentStepId ?? 0) + ''],
        __liza_start_date: [quote.getStartDate() + ''],
      });

export const loadProgramIntoQuote =
  (program: Program) =>
  (feature_flag: FeatureFlag) =>
  (quote: ServerSideQuote): IO.IO<ServerSideQuote> =>
  () => {
    quote.setNonpersistent(program.nonpersistent || []);
    TE.tryCatch(
      () => feature_flag.isEnabled('save_program_version'),
      () => false
    )().then(active => {
      if (active._tag === 'Right' && active.right) {
        quote.setProgramVersion(program.version);
      }
    });

    return quote;
  };

const getFlagData =
  (session: UserSession) =>
  (feature_flag: FeatureFlag) =>
  (flag_name: string) =>
    pipe(
      TE.tryCatch(
        () => feature_flag.isEnabled(flag_name, {user: session.userName()}),
        () => ['0']
      ),
      TE.map(x => (x ? ['1'] : ['0'])),
      TE.getOrElse(T.of)
    );

const getFlags =
  (program: Program) => (session: UserSession) => (feature_flag: FeatureFlag) =>
    pipe(
      {...(program.flags || {})},
      R.map(getFlagData(session)(feature_flag)),
      R.sequence(T.task)
    );

export const mergeFlagData =
  (program: Program) =>
  (session: UserSession) =>
  (feature_flag: FeatureFlag) =>
  (data: DocumentData) =>
  async () => {
    data.data = {
      // This is ordered deliberately so that manually added flags are overridden
      // by bucket data. Otherwise the order would not matter because flags that
      // are defined in the bucket should not be able to collide with question ids
      ...(await getFlags(program)(session)(feature_flag)()),
      ...data.data,
    };

    return data;
  };

/**
 * Merge two sets of document data
 *
 * The second data set's members take priority over the first's.
 **/
export const mergeDocumentData =
  (sub_data: DocumentData) =>
  (super_data: DocumentData): DocumentData => {
    return {...sub_data, ...super_data};
  };

/** Load data into a document (quote) */
export const loadDocument =
  (pullDocument: (id: DocumentId) => TE.TaskEither<string, DocumentData>) =>
  (program: Program) =>
  (session: UserSession) =>
  (feature_flag: FeatureFlag) =>
  (quote: ServerSideQuote) =>
  (init_doc_data: DocumentData) =>
  (keks: ServiceSecretMap): TE.TaskEither<unknown, ServerSideQuote> =>
    pipe(
      quote.getId(),
      pullDocument,
      // Session keys must be set first in order to encrypt defaulted values below
      TE.chain(doc_data =>
        pipe(
          E.tryCatch(
            () =>
              loadSessionKeysIntoDocumentData(program.downstreamServices)(keks)(
                doc_data
              ),
            String
          ),
          TE.fromEither
        )
      ),
      TE.chain(flow(applyDocumentDefaults(program)(keks), TE.fromIOEither)),
      TE.chainFirst(flow(loadSessionIntoQuote(session)(quote), TE.fromIO)),
      TE.chain(
        flow(mergeFlagData(program)(session)(feature_flag), TE.fromTask)
      ),
      TE.chain(
        flow(
          mergeDocumentData(init_doc_data),
          loadDocumentIntoQuote(quote),
          TE.fromIO
        )
      ),
      TE.chain(flow(loadProgramIntoQuote(program)(feature_flag), TE.fromIO))
    );

/** Load and initialize an existing document */
export const initExistingDocument = flow(pullDocumentFromDao, loadDocument);
export type InitExistingDocument = typeof initExistingDocument;

const processNaFields = (program: Program) => (data: DocumentData) => () =>
  program.processNaFields(data.data || {});

/** Prepare initial data for a new document (quote) */
export const prepareNewDocument =
  (program: Program) => (keks: ServiceSecretMap) => (data: DocumentData) =>
    pipe(
      E.tryCatch(
        () =>
          loadSessionKeysIntoDocumentData(program.downstreamServices)(keks)(
            data
          ),
        String
      ),
      TE.fromEither,
      TE.chain(flow(applyDocumentDefaults(program)(keks), TE.fromIOEither)),
      TE.chainFirst(flow(processNaFields(program), TE.fromIO)),
      TE.mapLeft(String)
    );

/** Set system-level fields on the bucket */
const setSystemFields =
  (quote: ServerSideQuote): IO.IO<ServerSideQuote> =>
  () =>
    quote.setData({
      __quote_id: [quote.getId().toString()],
      __document_id: [quote.getId().toString()],
    });

/**
 * Attempts to decrypt a set of session keys, or defaults to creating a new one
 *
 * @param kek - Key encrypting key to decrypt session key with
 * @param session_keys - List of stings to attempt to decrypt
 * @returns - The active/new session key
 */
export const retrieveOrGenerateSessionKey =
  (session_keys: string[]) =>
  (kek: Key): ProtectedString =>
    pipe(
      session_keys,
      A.reduce(createSessionKey(), (prev, curr) =>
        pipe(
          O.tryCatch(() => decryptSessionKey(kek, curr)),
          O.getOrElse(constant(prev))
        )
      )
    );

/**
 * Updates document to hold latest version of each service,
 * returns a flag of whether a version changed
 *
 * @param doc_data - Document to check/update
 * @param keks - list of keks and versions to compare
 *
 * @returns - The updated Document, and a flag to indicate if changes occurred
 */
export const updateLatestVersions = (
  doc_data: DocumentData,
  keks: ServiceSecretMap
): boolean => {
  // Create latestKekVersion object if not already available
  if (!doc_data.latestKekVersion) {
    doc_data.latestKekVersion = {};
  }

  let version_differs = false;
  for (const service in keks) {
    const doc_latest_version = doc_data.latestKekVersion[service];
    const kek_latest_active_version = KekService.latestActive(keks, service)[0];

    if (doc_latest_version !== kek_latest_active_version) {
      doc_data.latestKekVersion[service] = kek_latest_active_version;
      version_differs = true;
    }
  }

  return version_differs;
};

/**
 * Sets liza session keys to the document if they do not exist
 * @param doc_data          - Document to set session keys to
 * @param keks              - Keks used to create a session if one does not exist
 * @returns ServerSideQuote
 */
export const loadSessionKeysIntoDocumentData =
  (downstream_services: string[]) =>
  (keks: ServiceSecretMap) =>
  (doc_data: DocumentData): DocumentData => {
    const provided_session_keys: string[] = pipe(
      doc_data.sessionKeys?.[KekService.INTERNAL_KEY],
      O.fromNullable,
      O.getOrElse(constant<string[]>([]))
    );

    // if the latest kek exists on the document, use that in place of latest
    const session_key: ProtectedString = pipe(
      doc_data?.latestKekVersion?.liza,
      O.fromNullable,
      O.chain(version =>
        pipe(
          keks.liza,
          O.fromNullable,
          O.mapNullable(kek => kek[version]?.key)
        )
      ),
      O.getOrElse(constant(KekService.getLatestInteralKek(keks))),
      retrieveOrGenerateSessionKey(provided_session_keys)
    );

    const requires_update = updateLatestVersions(doc_data, keks);
    if (!doc_data.sessionKeys || requires_update) {
      doc_data.sessionKeys = createSessions(
        session_key,
        keks,
        downstream_services
      );
    }

    return doc_data;
  };

/**
 * Generates a key value map of downstream services and generated session keys from
 * active keks
 *
 * @param {ProtectedString} session_key - session key to encrypt
 * @param {ServiceSecretMap} kek_map     - a map of systems to their respective keks
 * @example
 * {
 *  service_1: [ 'key-1', 'key-2', 'key-3' ],
 *  service_2: [ 'key-1', 'key-2', 'key-3' ],
 * }
 * @returns ServerSideQuote
 */
export const createSessions = (
  session_key: ProtectedString,
  kek_map: ServiceSecretMap,
  downstream_services: string[]
): Record<string, string[]> =>
  pipe(
    kek_map,
    R.filterWithIndex(
      (service, _) =>
        service === KekService.INTERNAL_KEY ||
        downstream_services.includes(service)
    ),
    Object.entries,
    A.reduce({}, (acc: SessionMap, curr: [Service, KekVersionMap]) => ({
      // This order is important so that we don't overwrite base keys
      [curr[0]]: Object.values(curr[1])
        .filter(x => x.is_active)
        .map(({key}) => encryptSessionKey(key, session_key)),
      ...acc,
    }))
  );

/** Initialize a new document that has not yet been persisted */
export const initNewDocument =
  (dao: ServerDao) =>
  (quote_new: (id: number) => ServerSideQuote) =>
  (program: Program) =>
  (session: UserSession) =>
  (keks: ServiceSecretMap) =>
  (feature_flag: FeatureFlag) =>
  (init_doc_data: DocumentData) =>
    pipe(
      incrementAndPersistNextQuoteId(dao),
      TE.map(quote_new),
      TE.chain(flow(setSystemFields, TE.fromIO)),
      TE.chain(quote =>
        loadDocument(() =>
          prepareNewDocument(program)(keks)(quoteBucketToDocData(quote))
        )(program)(session)(feature_flag)(quote)(init_doc_data)(keks)
      )
    );

/** Convert the underlying bucket data from a quote into DocumentData */
const quoteBucketToDocData = (quote: ServerSideQuote): DocumentData => ({
  startDate: ts_ctor(),
  data: quote.getBucket().getData(),
});

export const documentUsesProgram =
  (program: Program) => (quote: ServerSideQuote) =>
    quote.getProgramId() === program.getId();

/**
 * Field state as of when the user last interacted with the document
 *
 * If the feature is not enabled, this will return the _current_ field
 * state, which will effectively result in a noop if kickback logic is
 * run.  Kickback logic should _not_ be run, since the flag applies there as
 * well, but this value is safe in that, if there happens to be a bug where
 * it is run, we won't wipe out a ton of user data.
 *
 * Note that an empty object means that the user sees no fields.  This could
 * be a legitimate scenario (e.g. maybe a static message is shown), but it's
 * important that we do not act on it if it's not legitimate, since that
 * would cause all applicable fields to be set to an invalid state.
 *
 * If no prior field state can be found, that means that the quote predates
 * the existence of persisted field state.  Rather than failing, we attempt
 * to infer what the prior field state must have been the best we can.
 */
const lastPersistedFieldState =
  (program: Program) =>
  (quote: ServerSideQuote): E.Either<Error, FieldState> =>
    pipe(
      quote.getLastPersistedFieldState(),
      E.fromNullable(Error('Last field state is not available')),

      // TODO: This line (and associated functions) can be removed at some
      // point, at the expense of old documents without field state no longer
      // being accessible (the above error will be returned).
      E.orElse(() => inferPriorFieldState(program)(quote))
    );

/**
 * Infer prior state in absence of explicit field state
 *
 * This represents a best guess based on current bucket values and whether
 * fields are currently required.  For more information, see
 * `inferPriorFieldApplicable`.
 */
const inferPriorFieldState =
  (program: Program) =>
  (quote: ServerSideQuote): E.Either<never, FieldState> =>
    pipe(
      program.qstep,
      R.filter(between(ordNumber)(0, quote.getTopSavedStepId())),
      R.mapWithIndex(quote.getDataByName.bind(quote)),
      R.mapWithIndex(inferPriorFieldApplicable(program)),
      E.right
    );

/** Whether the provided question id is required for the given Program */
const isQuestionRequired = (program: Program) => (qid: string) =>
  program.qrequired[qid];

/** Whether the provided question id is internal for the given Program */
const isQuestionInternal = (program: Program) => (qid: string) =>
  !!program.internal[qid];

/**
 * Guess whether each index of a given field was previously applicable
 *
 * If a field is not required, then we don't know whether a user has
 * encountered it or simply neglected to answer, and so we must simply
 * assume it was applicable.
 *
 * Otherwise, if a required field has an empty value, we assume that the
 * user must not have been presented with it, since it would otherwise have
 * a value.
 */
const inferPriorFieldApplicable =
  (program: Program) => (qid: string, values: unknown[]) =>
    A.array.map(
      values,
      value => +(!isQuestionRequired(program)(qid) || value !== '')
    );

/** Current field state, as of this moment, with the current Program */
const currentFieldState = (quote: ServerSideQuote) =>
  pipe(
    quote.getFieldState(),
    E.fromNullable(Error('Current field state is not available'))
  );

/** `R.isEmpty` without loss of type information when used as a refinement  */
const isEmptyRecord: <A>(r: Record<string, A>) => boolean = R.isEmpty;

/**
 * Kick document back to the earliest field that is now applicable that was
 * not previously
 */
export const kickBackToNewlyApplicable =
  (program: Program) =>
  (quote: ServerSideQuote): IOE.IOEither<Error, ServerSideQuote> =>
  () =>
    pipe(
      // Determine which fields are now applicable that were not previously,
      // failing if one of the arguments is not available
      pipe(
        E.of<Error, typeof fieldsNowApplicable>(fieldsNowApplicable),
        E.ap(lastPersistedFieldState(program)(quote)),
        E.ap(currentFieldState(quote))
      ),

      // Determine kickback step, if any, not to exceed the current step
      E.map(
        flow(
          // We must only consider required questions, otherwise the
          // invalidation of the question would force it to be required.  This
          // also allows us to exclude questions from kickback consideration.
          R.filterWithIndex(isQuestionRequired(program)),

          // Kicking back based on internal fields causes problems, because
          // the question state differs based on what type of user is viewing
          // the document.  It's also affected by systems using the session
          // key, which could trigger kickbacks while a non-internal user is
          // actively modifying/viewing the document, and offer no way for the
          // user to correct it.
          R.filterWithIndex(not(isQuestionInternal(program))),

          R.mapWithIndex(fieldStep(program)),
          R.filter(betweenPermittedKickbackRange(quote)),

          // At this point we know whether there are any fields that will
          // trigger a kickback; let's reify this fact as an `Option` so that
          // we later know whether to overwrite any step values
          O.fromPredicate(not(isEmptyRecord)),
          O.map(
            flow(
              invalidateFields(quote),
              R.foldMap(meetBeforeCurrentStep(quote))(stepIdFromPair)
            )
          )
        )
      ),

      // Impose step as upper bound on quote, being sure to honor current
      // values if they happen to be less than the kickback step.  The
      // `Option` here will ensure that we take no action if no kickback
      // fields were found.
      E.map(
        O.fold(constant(quote), kickback_step_id =>
          quote
            .setCurrentStepId(
              Math.min(quote.getCurrentStepId(), kickback_step_id)
            )
            .setTopVisitedStepId(
              Math.min(quote.getTopVisitedStepId(), kickback_step_id)
            )
            .setTopSavedStepId(
              Math.min(quote.getTopSavedStepId(), kickback_step_id)
            )
        )
      )
    );

/**
 * Whether the given step id is within the range permissable for kickback
 *
 * A kickback will not occur if the feature flag is not set on the
 * document.  The will be removed for release.
 *
 * If a document it locked, it'll be kicked back no lower than the lock step
 * if one is set; if there is no lock step but the document is locked, then
 * it will not be kicked back at all.
 *
 * The document will never be kicked back prior to the current step.
 */
const betweenPermittedKickbackRange = (quote: ServerSideQuote) =>
  between(ordPairStepId)(
    pairFromStepId(
      quote.isLocked()
        ? quote.getExplicitLockStep() || quote.getTopSavedStepId()
        : -Infinity
    ),

    pairFromStepId(quote.getTopSavedStepId())
  );

/**
 * If only one of the provided values is an array, convert the other into an
 * array by broadcasting to an array of the same length as the other
 */
const normalizeIndexes = <T>(a: T | T[], b: T | T[]) =>
  isArray(a) && !isArray(b)
    ? [a, a.map(_ => b)]
    : !isArray(a) && isArray(b)
    ? [b.map(_ => a), b]
    : [a, b];

/**
 * Whether a given field has indexes that are applicable when they were not
 * previously
 *
 * The returned value, if `some`, has all indexes cleared except for those
 * indexes that have become applicable.
 */
const fieldNewlyApplicable = ([x, y]: any[]) =>
  pipe(
    isArray(x) ? x.map((v, i: number) => +(v && v !== y[i])) : +(x && x !== y),
    O.fromPredicate(x => (isArray(x) ? x.some(v => !!v) : !!x))
  );

/** List of fields that are now applicable but were not previously */
export const fieldsNowApplicable = (last: FieldState) => (cur: FieldState) =>
  pipe(
    cur,
    R.filterMapWithIndex((key, state) =>
      fieldNewlyApplicable(normalizeIndexes(state, last[key] || 0))
    )
  );

/** Step field is assigned to within the given program, otherwise Infinity **/
const fieldStep =
  (program: Program) =>
  (field: string, state: State): StateStepPair =>
    [state, program.qstep[field] ?? Infinity];

/** Meet (minimum) of the given array of values, defaulting to current step */
const meetBeforeCurrentStep = (quote: ServerSideQuote) =>
  M.getMeetMonoid({
    ...B.boundedNumber,
    top: quote.getTopSavedStepId(),
  });

type State = RecordValue<FieldState>;
type StepId = number;
type StateStepPair = [State, StepId];

const pairFromStepId = (id: StepId): StateStepPair => [[], id];
const stateFromPair = (pair: StateStepPair) => pair[0];
const stepIdFromPair = (pair: StateStepPair) => pair[1];
const ordPairStepId = contramap(stepIdFromPair)(ordNumber);

/**
 * Invalidate all newly applicable fields, by index
 *
 * All indexes that are set for a given field will be set to the invalid
 * value `invalidValue`, which will cause a validation failure and force the
 * user to change its value.
 */
const invalidateFields =
  (quote: ServerSideQuote) => (fields: Record<string, StateStepPair>) => {
    pipe(
      fields,
      R.map(stateFromPair),
      R.mapWithIndex(broadcastScalar(quote)),
      R.map(invalidDelta),
      quote.setData.bind(quote)
    );

    return fields;
  };

type InvalidDelta = undefined | string | InvalidDelta[];

/** Generate a value for an invalid delta from the given `State` */
const invalidDelta = (x: State): InvalidDelta =>
  isArray(x) ? x.map(invalidDelta) : [undefined, invalidate('')][x];

/** Broadcast a scalar value onto a vector */
const broadcastScalar = (quote: ServerSideQuote) => (field: string, x: State) =>
  isArray(x) ? x : quote.getDataByName(field).map(_ => x);

/** Quote cleaner configuration */
export type CleanDocumentConf = {
  liza_kek: Key;
};

/**
 * "Clean" quote, getting it into a stable state
 *
 * Quote cleaning will ensure that all group fields share at least the
 * same number of indexes as its leader, and that meta fields are
 * initialized.  This is useful when questions or meta fields are added.
 */
export const cleanDocument =
  (conf: CleanDocumentConf) => (program: Program) => (quote: ServerSideQuote) =>
    pipe(
      quote,

      // Consider it an error to attempt cleaning a quote with the incorrect
      // program, which would surely corrupt it [even further]
      TE.fromPredicate(
        documentUsesProgram(program),
        constant(Error('Quote/program mismatch'))
      ),

      TE.chainFirst(
        flow(
          flow(fixGroup(program), O.of),
          // TODO: get rid of this extra expensive classify, considering that
          // a previous part of the system probably already did it!
          O.ap(O.of(quote.classify())),
          O.map(f =>
            A.array.traverse(IO.io)(
              Object.keys(program.groupIndexField || {}),
              f
            )()
          ),
          TE.fromOption(constant(Error('fixGroup failure')))
        )
      ),

      TE.chainFirst(flow(fixMeta(program), TE.fromIO)),
      TE.chainFirst(flow(fixNonEncryptedFields(conf)(program), TE.fromIO)),

      // Kick user back to newly applicable fields
      TE.chainIOEitherKW(kickBackToNewlyApplicable(program))
    );

/**
 * Correct group fields to be at least the length of the leader
 *
 * If a group is part of a link, then its leader may be part of another
 * group, and the length of the fields of all linked groups will match
 * be at least the length of the leader.
 *
 * Unlike previous implementations, this _does not_ truncate fields,
 * since that risks data loss.  Instead, field length should be
 * validated on save.
 */
const fixGroup =
  (program: Program) =>
  (quote: ServerSideQuote) =>
  (class_data: ClassData) =>
  (group_id: string) =>
  () => {
    const length = groupLength(program)(quote)(group_id);

    // if we cannot accurately determine the length then it's too
    // dangerous to proceed and risk screwing up the data; abort
    // processing this group (this should never happen unless a program
    // is either not properly compiled or is out of date)
    if (isNaN(length)) {
      return;
    }

    const update: Record<string, string[]> = {};
    const group_fields = program.groupExclusiveFields[group_id];

    group_fields.forEach(field => {
      const flen = (quote.getDataByName(field) || []).length;

      // generated questions with no types should never be part of
      // the bucket
      if (!program.hasKnownType(field)) {
        return;
      }

      if (flen >= length) {
        return;
      }

      const data = [];
      const field_default = program.defaults[field] || '';

      for (let i = flen; i < length; i++) {
        data[i] =
          program.clearNaFields && program.hasNaField(field, class_data, i)
            ? program.naFieldValue
            : field_default;
      }

      update[field] = data;
    });

    quote.setData(update);
  };

/**
 * Determine length of group GROUP_ID
 *
 * The length of a group is the length of its leader, which may be part
 * of another group (if the group is linked).
 */
const groupLength =
  (program: Program) => (quote: ServerSideQuote) => (group_id: string) => {
    const index_field = program.groupIndexField[group_id];

    // we don't want to give the wrong answer, so just abort
    if (!index_field) {
      return NaN;
    }

    const data = quote.getDataByName(index_field);

    return Array.isArray(data) ? data.length : NaN;
  };

/**
 * Initialize missing metadata
 *
 * This is similar to bucket initialization, except there are no leaders
 * or default values---just empty arrays.  That may change in the future.
 */
const fixMeta = (program: Program) => (quote: ServerSideQuote) => () => {
  const {fields = {}} = program.meta;
  const metabucket = quote.getMetabucket();
  const metadata = metabucket.getData();

  Object.keys(fields).forEach(field_name => {
    if (Array.isArray(metadata[field_name])) {
      return;
    }

    metabucket.setValues({[field_name]: []});
  });
};

/**
 * Encrypt each unencrypted value for fields in `program.encryptedKeys`.
 *
 * This can happen when a field was _not_ marked as encrypted in previous
 * versions of `program`.
 */
export const fixNonEncryptedFields =
  ({liza_kek}: CleanDocumentConf) =>
  (program: Program) =>
  (quote: ServerSideQuote) =>
  () => {
    const encryptor = encryptorFromEncryptedSessionKeys(
      liza_kek,
      quote.getSessionKeys()
    );

    // TODO: It is not at all ideal that this implementation is able to vary
    // from that of delta encryption in `EncryptionService`/`Server`.  These
    // need to be unified at some point.  Until then, exercise caution when
    // making any changes to this part of the system and be sure that delta
    // encryption operates in the same manner as the cleaner.
    const maybeEncrypt = maybeEncryptDeltaDatum((x: string) =>
      encryptor.encrypt(x)
    );

    const bucket = quote.getBucket();
    const update: Record<string, BucketFieldValue> = {};
    Object.keys(program.encryptedKeys).forEach(key => {
      const data = bucket.getDataByName(key);
      if (!data) return;
      update[key] = data.map(x =>
        // Can only encrypt strings
        typeof x !== 'string' || isPossiblyEncrypted(x) ? x : maybeEncrypt(x)
      );
    });

    quote.setData(update);
  };
