/**
 * Handles rating with local, JS-compiled TAME-written raters
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {BucketData} from '../../bucket/Bucket';
import {
  SupplierRateResult,
  Rater,
  RateContext,
  ClassificationData,
} from './Rater';
import {EventEmitter} from 'events';
import {DslRaterContext} from './DslRaterContext';
import {ResultSet} from './ResultSet';

export class DslRater extends EventEmitter {
  /**
   * @param _raters    - list of raters
   * @param _ResultSet - ResultSet constructor (used in place of a factory)
   */
  constructor(
    private _raters: Rater[],
    private _ResultSet: (id: string) => ResultSet
  ) {
    super();
  }

  /**
   * Perform rating
   *
   * @param context             - the rating context
   * @param requested_suppliers - supplier(s) to rate
   */
  rate(
    context: DslRaterContext,
    requested_suppliers?: string | string[]
  ): void {
    if (this._raters.length === 0) {
      // TODO: this is a BS answer
      throw new Error('No raters are available at this time.');
    }

    this._doRate(this._raters.slice(0), context, requested_suppliers);
  }

  /**
   * Invoke the next rater in the queue
   *
   * @param queue               - the queue to operate on
   * @param context             - rating context
   * @param requested_suppliers - supplier(s) to rate
   */
  private _queueNext(
    queue: Rater[],
    context: DslRaterContext,
    requested_suppliers?: string | string[]
  ): void {
    process.nextTick(() => this._doRate(queue, context, requested_suppliers));
  }

  /**
   * Process next available supplier until complete
   *
   * @param queue               - a queue of raters
   * @param context             - the rating context
   * @param requested_suppliers - supplier(s) to rate
   */
  private _doRate(
    queue: Rater[],
    context: DslRaterContext,
    requested_suppliers?: string | string[]
  ) {
    const rater: Rater | undefined | null = queue.pop();

    if (rater === undefined) {
      this._complete(context);
      return;
    }

    // null means that the rater failed to load
    if (rater === null) {
      // TODO: log error
      this._queueNext(queue, context, requested_suppliers);
      return;
    }

    const supplier_name = rater.supplier;

    // A specific supplier was requested and this isn't it
    if (!this._rateSupplier(supplier_name, requested_suppliers)) {
      this._queueNext(queue, context, requested_suppliers);
      return;
    }

    const meta = rater.rater.meta;
    const result_set = this._ResultSet(supplier_name);

    // to be called by context when rating is complete for this rater
    const complete = () => {
      context.addResultSet(supplier_name, result_set);
      this._queueNext(queue, context, requested_suppliers);
    };

    // give context the chance to augment the data and determine how many
    // times we should rate with a given supplier; this continuation will be
    // called for each time that the context wishes to perform rating
    context.rate(
      supplier_name,
      meta,
      (data: BucketData, rcontext: RateContext) => {
        let single: SupplierRateResult;

        try {
          const can_term = context.canTerm();

          single = rater(data, can_term);

          // ensures that any previous eligibility errors are cleared out
          single.ineligible = '';
          single.submit = '';
        } catch (e) {
          // ineligible.
          single = {
            ineligible: e instanceof Error ? e.message : 'Unknown rating error',
            _unavailable: ['1'],
            _unavailable_all: ['1'],
            submit: '',
            premium: 0.0,
          };
        }

        // give the context to process the result of the rating before we
        // perform any of our own processing (that way they can trigger
        // submits, etc)
        single = context.processResult(single, meta, rcontext);

        this._processProhibits(rater, single, context, data)
          ._processSubmits(rater, single, data)
          ._flagEligibility(single);

        // purposely omitted third argument
        result_set.addResult(single, rcontext);
      },
      complete
    );
  }

  /**
   * Get submits from classes
   *
   * @param rater  - the rater to process
   * @param single - the rate results for a supplier
   * @param data   - the rate data
   */
  private _processSubmits(
    rater: Rater,
    single: SupplierRateResult,
    data: BucketData
  ): this {
    // ineligible results cannot submit, as they did not complete rating
    if (single.ineligible) {
      return this;
    }

    const c = single.__classes;

    if (!c || !c.submit) {
      return this;
    }

    const submits = [];
    for (const cname in c) {
      // Process submit classifications if they are *true*.  Classes
      // that are suffixed with a dash represent a generated rule that
      // should _not_ be displayed to the user
      if (/^submit-.*[^-]$/.test(cname) && c[cname]) {
        submits.push(this.getCdesc(cname, rater, data));
      }
    }

    single.submit = submits.join('; ');
    return this;
  }

  /**
   * Get prohibits from classes
   *
   * @param rater   - Supplier rater
   * @param single  - Supplier UI options
   * @param context - Rater context
   * @param data    - Bucket data
   */
  private _processProhibits(
    rater: Rater,
    single: SupplierRateResult,
    context: DslRaterContext,
    data: BucketData
  ): this {
    // Don't process prohibits if rating can naturally terminate
    if (context.canTerm()) {
      return this;
    }

    const classes = single.__classes;

    // An error was thrown during rating
    if (classes === undefined) {
      return this;
    }

    const prohibits = [];
    const regex = {
      prohibit: /^inelig-.+$/,
      assertion: /^-assert-.+$/,
      domainInvalid: /^domain-invalid-.+$/,
    };

    for (const className in classes) {
      const foundProhibit =
        regex.prohibit.test(className) ||
        regex.assertion.test(className) ||
        regex.domainInvalid.test(className);

      const classValue = classes[className];

      if (foundProhibit && classValue) {
        prohibits.push(this.getCdesc(className, rater, data));
      }
    }

    // Backup when an error is thrown but no prohibit is found (unlikely)
    if (this._isIneligible(classes, rater) && prohibits.length === 0) {
      prohibits.push('This supplier is ineligible.');
    }

    single.ineligible = this._filterProhibits(prohibits).join('; ');

    return this;
  }

  /**
   * Determine a suppliers eligibility
   *
   * @param classes - Classification data
   * @param rater   - Supplier rater
   *
   * @return true if the data contain an error
   */
  private _isIneligible(classes: ClassificationData, rater: Rater): boolean {
    // XXX: use elig-class when propagated by tameld;
    // this is a generated classification name
    const eligKey = '--elig-suppliers-' + rater.supplier;

    if (classes[eligKey] === undefined) {
      throw new Error('Missing supplier eligibility field: ' + eligKey);
    }

    return classes[eligKey] === false;
  }

  /**
   * Remove generic and duplicated prohibit reasons
   * Generic prohibits take the form of [{supplier-name} prohibit]
   *
   * @param prohibits - A list of prohibit reasons
   *
   * @return A filtered list of prohibit reasons
   */
  private _filterProhibits(prohibits: string[]): string[] {
    const unique = (value: string, index: number, array: string[]) =>
      array.indexOf(value) === index;

    const hasSpecificProhibit = (value: string) =>
      !/^\[.+\sprohibit\]$/.test(value);

    return prohibits.filter(unique).filter(hasSpecificProhibit);
  }

  /**
   * Mark supplier eligibility
   *
   * @param single - Single raters result
   */
  private _flagEligibility(single: SupplierRateResult): this {
    // from a broker's perspective
    single._unavailable = single.submit || single.ineligible ? '1' : '0';

    return this;
  }

  /**
   * Get the class descriptions
   *
   * @param cname - class name
   * @param rater - rater to get the description from
   * @param _data - bucket data
   *
   * @returns the class description
   */
  getCdesc(cname: string, rater: Rater, _data: BucketData): string {
    return rater.rater.classify.desc[cname] || cname;
  }

  /**
   * Let the context know that we're finished rating
   *
   * @param context - rating context
   */
  private _complete(context: DslRaterContext) {
    context.complete();
  }

  /**
   * Determine whether or not to rate a given supplier
   *
   * @param supplier_name - the supplier in question
   * @param requested     - suppliers that were requested when invoking rate
   */
  private _rateSupplier(
    supplier_name: string,
    requested?: string | string[]
  ): boolean {
    return (
      !requested ||
      (typeof requested === 'string' && supplier_name === requested) ||
      (Array.isArray(requested) && requested.indexOf(supplier_name) > -1)
    );
  }
}
