/**
 * Provides context-specific data to the DslRater
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {BucketData} from '../../bucket/Bucket';
import {EventEmitter} from 'events';
import {ResultSet} from './ResultSet';
import {SupplierRateResult, RateContext, RateResult} from './Rater';

export class DslRaterContext extends EventEmitter {
  /** Result sets */
  private _results: ResultSet[] = [];

  /**
   * @param _data     - quote data with which to rate
   * @param _can_term - whether to immediately terminate on assertion failure
   */
  constructor(
    private _data: BucketData | null = null,
    private _can_term: boolean = true
  ) {
    super();
    this.init();
  }

  /** XXX: return to protected (see commit message) **/
  getSourceData(): BucketData | null {
    return this._data;
  }

  /**
   * Return whether to terminate on prohibits/assertions
   */
  canTerm(): boolean {
    return this._can_term;
  }

  /**
   * Unimplemented here
   */
  protected init(): void {
    // may be implemented by subtypes
  }

  /**
   * Invoked rating
   *
   * @param _name     - supplier name
   * @param _meta     - meta data
   * @param rate      - rating function
   * @param _complete - completion callback
   */
  rate(
    _name: string,
    _meta: unknown,
    rate: (data: BucketData, rcontext: RateContext) => void,
    _complete: () => void
  ): void {
    rate(this._data ?? {}, undefined);
  }

  /**
   * Process the results
   *
   * this is unimplemented in this class but defined to allow child classes to
   * implement it
   *
   * @param result    - result
   * @param _meta     - meta data
   * @param _rcontext - rate context
   */
  processResult(
    result: SupplierRateResult,
    _meta: unknown,
    _rcontext: RateContext
  ): SupplierRateResult {
    return result;
  }

  /**
   * Add a completed ResultSet
   *
   * @param name - supplier/rater name
   * @param set  - completed rating result set
   */
  addResultSet(_name: string, set: ResultSet): this {
    this._results.push(set);

    return this;
  }

  /**
   * Manually complete the process
   */
  complete() {
    this._emitResults();
  }

  /**
   * Emit completion event
   *
   * @emits complete
   */
  private _emitResults(): void {
    this.emit('complete', this.postProcessResults(this._results));
  }

  /**
   * Post process the result set
   *
   * @param results - result set
   *
   * @return processed results
   */
  public postProcessResults(results: ResultSet[]): RateResult {
    const ret: Record<string, unknown> = {};

    results.forEach(set => {
      const id = set.getId(),
        merged = set.getMergedResults();

      // prefix each field with the result set id (TODO: this should
      // really be part of a decorator, not this class)
      for (const field in merged) {
        if (!field) {
          continue;
        }

        ret[id + '_' + field] = merged[field];
      }
    });

    return ret;
  }
}
