/**
 * Holds rating results and metadata
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {SupplierRateResult, RaterMeta} from './Rater';

export class ResultSet {
  /**
   * The result set id
   */
  private _id = '';

  /**
   * Individual results as separate objects
   */
  private _results: SupplierRateResult[] = [];

  /**
   * Metadata associated with individual results
   */
  private _meta: RaterMeta[] = [];

  /**
   * Current result index
   */
  private _resulti = 0;

  /**
   * Number of available results
   */
  private _availCount = 0;

  /**
   * @param id - the result set id
   */
  constructor(id: string) {
    this._id = '' + id;
  }

  /**
   * Add result to the set
   *
   * @param result - the result to add
   * @param meta   - rater metadata
   * @param index  - the index to add results at
   */
  public addResult(
    result: SupplierRateResult,
    meta: RaterMeta,
    index?: number
  ): void {
    if (index === undefined) {
      index = this._resulti++;
    } else if (index > this._resulti) {
      index = +index;
      this._resulti = index;
    }

    if (result._unavailable === '0') {
      this._availCount++;
    }

    this._results[index] = result;
    this._meta[index] = meta;
  }

  /**
   * Return the number of available suppliers
   */
  public getAvailableCount(): number {
    return this._availCount;
  }

  /**
   * Return the total number of results
   */
  public getResultCount(): number {
    return this._results.length;
  }

  /**
   * Return the result set id
   */
  public getId(): string {
    return this._id;
  }

  /**
   * Merge each result into a combined object
   *
   * Each field will contain an array with a value from each result. If a
   * result is missing a field, then the value for that particular index will
   * be defaulted to the empty string.
   *
   * @return merged results
   */
  getMergedResults(): SupplierRateResult {
    const set: Record<string, any> = {},
      fields: Record<string, any> = {};

    // merge all fields into a single object
    this._results.forEach((result, i) => {
      for (const field in result) {
        (set[field] = set[field] || [])[i] = result[field];

        fields[field] |= 0;
        fields[field]++;
      }
    });

    // fill in gaps in the array (if a field is not defined in a result)
    for (const field in fields) {
      const chk = set[field];

      // if the destination array is larger than the number of items it
      // contains, then we have holes to fill
      if (fields[field] === chk.length) {
        continue;
      }

      let i = chk.length;
      while (i--) {
        // default to an empty string
        chk[i] = chk[i] || '';
      }
    }

    return this._processSet(set);
  }

  /**
   * Process the set
   *
   * @param set - the set to process
   */
  private _processSet(set: any): SupplierRateResult {
    // combined availability
    set._unavailable_all = [this._availCount === 0 ? '1' : '0'];

    return set;
  }
}
