/**
 * Rating process
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This process permits both asynchronous rating (in the case of JS raters) as
 * well as hassle-free runtime reloading of raters withour risk of memory leaks
 * that may be caused by deleting entries from node's module cache.
 */
import fs = require('fs');
import {BucketData} from '../../bucket/Bucket';
import {RateResult, Rater} from './Rater';
import {ServerSideQuote} from '../quote/ServerSideQuote';
import {UserSession} from '../request/UserSession';
import {ClientActions} from '../../client/action/ClientAction';

export type RateRequestMessage = {
  cmd: 'rate';
  supplier: string;
  indv?: string[];
  rqid: string;
  quote: {
    id: string;
    agentId: string;
    agentName: string;
    creditScoreRef: string;
    data: BucketData;
  };
  agentId: string;
  internal: boolean;
};

export type RateReplyOkMessage = {
  cmd: 'rate-reply';
  rqid: string;
  status: 'ok';
  data: RateResult;
  actions: ClientActions;
  override: boolean;
  imported: boolean;
};

export type RateReplyErrorMessage = {
  cmd: 'rate-reply';
  rqid: string;
  status: 'error';
  msg: string;
};

export type RateReplyMessage = RateReplyOkMessage | RateReplyErrorMessage;

const map: Record<string, Rater> = {};
const rater_path = process.env.RATER_PROGRAM_PATH;

function forEachProgram(
  path: string,
  c: (filename: string, path: string) => void
) {
  fs.readdirSync(path).forEach(filename => {
    // ignore non-js files
    if (!filename.match(/\.js$/)) {
      return;
    }

    c(filename, rater_path + '/' + filename);
  });
}

// sent by parent to indicate that we should exit (we should never receive this
// signal from any sort of shell or anything)
try {
  process.on('SIGHUP', () => {
    // we're good
    // eslint-disable-next-line no-process-exit
    process.exit(0);
  });
} catch (e) {
  sendLog(
    process,
    'WARNING: OS does not support signal handlers; process will always ' +
      'appear to exit in error if reload is requested'
  );
}

// msg can really be anything and we should verify the type before using it
process.on('message', msg => {
  switch (msg.cmd) {
    case 'rate':
      doRate(process, map, msg);
      break;

    default:
      sendError(process, 'Unknown command from parent: ' + msg.cmd);
  }
});

process.on('uncaughtException', e => sendError(process, e));

if (!rater_path) {
  sendError(process, 'RATER_PROGRAM_PATH is not set');
  // eslint-disable-next-line no-process-exit
  process.exit(1);
}

// load raters
forEachProgram(rater_path, (filename: string, path: string) => {
  let rater = null;

  try {
    rater = require(path);
    sendLog(process, 'Loaded rater: ' + filename);
  } catch (e) {
    sendError(process, e instanceof Error ? e : new Error(e + ''));
  }

  const name = filename.replace(/\.js$/, '');

  map[name] = rater;
});

sendLog(process, 'Rating process ready (PID ' + process.pid + ').');

function doRate(
  node_process: NodeJS.Process,
  rater_map: Record<string, Rater>,
  data: RateRequestMessage
) {
  const quote = data.quote,
    rqid = data.rqid,
    rater = rater_map[data.supplier];

  if (!rater) {
    // This could be undefined depending on how the process was spawned
    if (!node_process.send) {
      throw new Error('Node process does not have a send method');
    }

    node_process.send({
      cmd: 'rate-reply',
      rqid: rqid,

      status: 'error',
      msg: 'Unknown supplier: ' + data.supplier,
    });

    return;
  }

  // TODO: temporary, until refactoring is complete
  let imported_flag = false;
  const rate_quote = {
    getId: function () {
      return quote.id;
    },

    getAgentId: function () {
      return quote.agentId;
    },

    getAgentName: function () {
      return quote.agentName;
    },

    getCreditScoreRef: function () {
      return quote.creditScoreRef;
    },

    getBucket: function () {
      return {
        getData: function () {
          return quote.data;
        },
      };
    },

    setImported: function (val?: boolean) {
      val = val === undefined ? true : !!val;
      imported_flag = val;
    },
  };

  const rate_session = {
    agentId: function () {
      return data.agentId;
    },

    isInternal: function () {
      return data.internal;
    },
  };

  // perform rating
  rater.rate(
    // Currently we are actually just using a stubbed out quote object
    <ServerSideQuote>(<unknown>rate_quote),
    // Currently we are actually just using a stubbed out session object
    <UserSession>(<unknown>rate_session),
    data.indv,
    (rdata, actions, override = false) => {
      // This could be undefined depending on how the process was spawned
      if (!node_process.send) {
        throw new Error('Node process does not have a send method');
      }

      node_process.send({
        cmd: 'rate-reply',
        rqid: rqid,

        status: 'ok',
        data: rdata,
        actions: actions,
        override: override,

        // not used by all raters
        imported: imported_flag,
      });
    },

    msg => {
      // This could be undefined depending on how the process was spawned
      if (!node_process.send) {
        throw new Error('Node process does not have a send method');
      }

      node_process.send({
        cmd: 'rate-reply',
        rqid: rqid,

        status: 'error',
        msg: msg,
      });
    }
  );
}

function sendLog(node_process: NodeJS.Process, msg: string) {
  // This could be undefined depending on how the process was spawned
  if (!node_process.send) {
    throw new Error('Node process does not have a send method');
  }

  node_process.send({cmd: 'log', msg: msg});
}

function sendError(node_process: NodeJS.Process, e: Error | string) {
  // This could be undefined depending on how the process was spawned
  if (!node_process.send) {
    throw new Error('Node process does not have a send method');
  }

  if (typeof e === 'string') {
    e = new Error(e);
  }

  node_process.send({
    cmd: 'error',
    msg: e.message,
    stack: e.stack,
  });
}
