/**
 * Concurrent Session Manager
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Mediate between different requests for the same document. Ensuring that we
 * do not have two different users modifying the document at the same time
 */

import {UserSession} from './UserSession';

export class ConcurrentSessionManager {
  /**
   * Interval at which to clear stale session flags
   *
   * set this to ~10s after the quicksave interval
   */
  public static CLEAR_INTERVAL_SEC = 70;

  /**
   * concurrent session flag
   */
  private sflag: Record<number, {userName?: string; time: number}> = {};

  /**
   * Whether we have initialized the manager
   */
  private initialized = false;

  /**
   * Initialize the session manager
   */
  constructor(private readonly ts_ctor: () => UnixTimestamp) {}

  /**
   * Initialize the session manager
   *
   * This involves starting an interval that will clear stale session flags
   *
   * @returns this
   */
  public init(): this {
    if (this.initialized) {
      return this;
    }

    setInterval(() => {
      const now = this.ts_ctor();

      for (const id in this.sflag) {
        // clear all session flags that have timed out
        if (
          now - this.sflag[id].time >
          ConcurrentSessionManager.CLEAR_INTERVAL_SEC
        ) {
          delete this.sflag[id];
        }
      }
    }, ConcurrentSessionManager.CLEAR_INTERVAL_SEC);

    this.initialized = true;

    return this;
  }

  /**
   * Keep a session alive for a user
   *
   * @param id      - document id
   * @param session - user session
   */
  public touchSession(id: number, session: UserSession): void {
    const cur = this.sflag[id];

    // do not allow touching the session if we're not the owner
    if (cur && cur.userName !== session.userName()) {
      return;
    }

    this.sflag[id] = {
      userName: session.userName(),
      time: this.ts_ctor(),
    };
  }

  /**
   * Get the user who has a concurrent session
   *
   * @param id      - document id
   * @param session - user session
   *
   * @return user name of the concurrent session
   */
  public getConcurrentSessionUser(
    id: number,
    session: UserSession
  ): string | undefined {
    const flag = this.sflag[id];

    // we have a flag; if we're the same user that created it (we check on name
    // because internally we all have the same agent id), then do not consider
    // this a concurrent access attempt
    if (!flag || flag.userName === session.userName()) {
      return '';
    }

    return flag.userName;
  }
}
