/**
 * Manages DataAPI requests and return data
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {
  Bucket,
  BucketData,
  BucketFieldValue,
  EmptyBucketData,
} from '../../bucket/Bucket';
import {BucketFilter} from '../../bucket/bucket_filter';
import {DapiMetaSource, FieldData} from '../meta/DapiMetaSource';
import {DataApiManager, DataApiConstructor} from '../../dapi/DataApiManager';
import {PositiveInteger} from '../../numeric';
import {MetaFields, Program} from '../../program/Program';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {ServerSideQuote} from '../quote/ServerSideQuote';
import {StagingBucket} from '../../bucket/StagingBucket';
import {UserRequest} from './UserRequest';
import {DapiAction} from '../../dapi/DataApi';
import {LayeredBucket} from '../../bucket/layer/LayeredBucket';
import {PriorityLog} from '../log/PriorityLog';

/**
 * Process data provided by the client
 *
 * TOOD: This contains Data API and bucket merging logic that is better done
 * elsewhere.
 */
export class DataProcessor {
  /**
   * Initialize processor
   *
   * The staging bucket constructor will be used to wrap the bucket for
   * diff-related operations.
   *
   * @param logger       - for making log entries
   * @param filter       - bucket filter
   * @param dapif        - data API constructor
   * @param meta_source  - metadata source
   * @param staging_ctor - staging bucket constructor
   * @param metaClearer  - converts DAPI fields into a meta update
   */
  constructor(
    private readonly _logger: PriorityLog,
    private readonly _filter: BucketFilter,
    private readonly _dapif: DataApiConstructor,
    private readonly _meta_source: DapiMetaSource,
    private readonly _stagingCtor: (bucket: Bucket) => StagingBucket,
    private readonly _metaClearer: (fields: BucketData) => EmptyBucketData
  ) {}

  /**
   * Process client-provided data diff
   *
   * This performs sanitization to ensure that we are storing only
   * "correct" data within our database. This also strips any unknown
   * bucket values, preventing users from using us as their own personal
   * database.
   *
   * @param data     - bucket diff data
   * @param request  - submitting request
   * @param program  - active program
   * @param bucket   - staging bucket
   * @param quote    - server side quote
   * @param on_error - error callback
   *
   * @return processed diff
   */
  processDiff(
    data: BucketData,
    request: UserRequest,
    program: Program,
    bucket: QuoteDataBucket,
    quote: ServerSideQuote,
    diff_override: BucketData | undefined,
    on_error: (err: Error, name?: string) => void,
    dapi_classifications: Record<string, boolean> = {}
  ): {
    filtered: BucketData;
    dapis: Promise<FieldData | void>[];
    meta_update: EmptyBucketData;
    rdiff: BucketData;
  } {
    const filtered = this._sanitizeDiff(data, request, program);

    const dapi_manager = this._dapif(program.apis, request, quote).on(
      'error',
      on_error
    );

    const staging = this._stagingCtor(bucket);

    // forbidBypass will force diff generation on initQuote
    staging.setValues(filtered);
    staging.forbidBypass();

    program.initQuote(staging, true);

    const diff = diff_override || staging.getDiff();

    const {
      mapis = {},
      meta: {fields = {}},
    } = program;

    const eligible_dapis = this._determineDapisToTrigger(mapis, diff);

    const dapis_to_request = this._addClassificationDapis(
      quote,
      fields,
      eligible_dapis,
      dapi_classifications
    );

    const dapis_to_trigger = this._determineDapiActions(
      dapis_to_request,
      dapi_classifications
    );

    const meta_update = this._getMetaUpdate(dapis_to_trigger);

    // array of promises for any dapi requests
    const dapis = this._triggerDapis(
      dapi_manager,
      fields,
      diff,
      staging,
      quote.getLayeredBucket(),
      dapis_to_trigger,
      on_error
    );

    const rdiff: BucketData = {};

    for (const diff_key in diff) {
      const original_val = staging.getOriginalDataByName(diff_key);

      // This should never happen but the bucket can technically return null
      // values so we should account for it. If it does then it's ok to omit
      // those in the downstream delta.
      if (!original_val) {
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          'Error calculating reverse delta, ' +
            `field ${diff_key} not found in bucket`
        );

        continue;
      }

      rdiff[diff_key] = original_val;
    }

    staging.commit();

    return {filtered, dapis, meta_update, rdiff};
  }

  /**
   * Sanitize client-provided data
   *
   * Internal fields will be stripped if the session is not
   * internal.  Following that, the filter provided via the ctor will be
   * applied.
   *
   * `permit_null` should be used only in the case of bucket diffs, which
   * contain nulls as terminators.
   *
   * @param data    - client-provided data
   * @param request - client request
   * @param program - active program
   *
   * @return filtered data
   */
  private _sanitizeDiff(
    data: BucketData,
    request: UserRequest,
    program: Program
  ): BucketData {
    if (!request.getSession().isInternal()) {
      this._cleanInternals(data, program);
    }

    return this._filter(data, program.meta.qtypes, {}, true);
  }

  /**
   * Strip internal fields from diff `data`
   *
   * Internal fields are defined by the program `program`.
   *
   * @param data    - bucket diff data
   * @param program - active program
   */
  private _cleanInternals(data: BucketData, program: Program): void {
    for (const id in program.internal) {
      delete data[id];
    }
  }

  private _getMetaUpdate(
    dapi_fields: Record<string, {indexes: string[]; action: DapiAction}>
  ): EmptyBucketData {
    /**
     * This ensures that stale data won't be accessible to systems while a
     * request hasn't yet completed.  For example, if performing a rate
     * lookup, it wouldn't be desirable to use an old rate even though data
     * used to retrieve it has since changed.
     */
    const field_indexes: Record<string, string[]> = {};

    Object.keys(dapi_fields).forEach(
      dapi_field =>
        (field_indexes[dapi_field] = dapi_fields[dapi_field].indexes)
    );

    return this._metaClearer(field_indexes);
  }

  /**
   * Trigger metadata Data API requests
   *
   * @param dapi_manager - dapi manager
   * @param fields       - meta field definitions
   * @param data         - client-provided data
   * @param bucket       - active bucket
   * @param dapi_fields  - dapi fields to trigger
   * @param on_error     - error callback
   *
   * @return an array containing the dapis and cleared meta values
   */
  private _triggerDapis(
    dapi_manager: DataApiManager,
    fields: MetaFields,
    data: BucketData,
    bucket: StagingBucket,
    layered_bucket: LayeredBucket,
    dapi_fields: Record<string, {indexes: string[]; action: DapiAction}>,
    on_error: (err: Error, name?: string) => void
  ): Promise<FieldData | void>[] {
    const dapis = Object.keys(dapi_fields)
      .map(field => {
        const {dapi} = fields[field];
        const {indexes, action} = dapi_fields[field];
        const {mapsrc} = dapi;

        return indexes.map(i =>
          this._meta_source
            .getFieldData(
              field,
              <PositiveInteger>+i,
              dapi_manager,
              dapi,
              this._bucketToDapiInputs(
                mapsrc,
                layered_bucket,
                <PositiveInteger>+i,
                data
              ),
              action,
              // This bucket is needed for bucket dapis and webhook dapis,
              // neither should need layered data but we should still support.
              // it. It is also required to be a staging bucket so that we can
              // use that to calculate a reverse delta.
              bucket,
              on_error
            )
            .catch(e =>
              on_error(new Error(`Field "${field}" error: ${e.message ?? e}`))
            )
        );
      })
      .reduce((result, x) => result.concat(x), []);

    return dapis;
  }

  /**
   * Determine which fields require a Data API to be triggered
   *
   * Note: this is determining which dapis should be attempted to be called.
   * When we actually invoke the dapis later on, they made be ended early
   * because all of the inputs are not satisfied or there is a new token, etc.
   *
   * @param mapis - metadata dapi descriptors
   * @param data  - client-provided data
   *
   * @return fields with indexes in need of dapi calls
   */
  private _determineDapisToTrigger(
    mapis: Record<string, string[]>,
    data: BucketData
  ): Record<string, string[]> {
    return Object.keys(mapis).reduce(
      (result: Record<string, string[]>, src_field: string) => {
        const source_field_data = data[src_field];

        if (source_field_data === undefined) {
          return result;
        }

        const dapi_fields = mapis[src_field];

        // get each index that changed
        dapi_fields.forEach((field: string) => {
          result[field] = result[field] || [];
          Object.keys(source_field_data).forEach(i => {
            if (
              source_field_data[+i] === undefined ||
              source_field_data[+i] === null
            ) {
              return;
            }

            result[field][+i] = i;
          });
        });

        return result;
      },
      {}
    );
  }

  /**
   * Add dapis that have a true "when" classification
   *
   * If a classification is true, it may not register as changed and be
   * picked up by #_determineDapisToTrigger. In this case we augment data
   * and then check for duplicates later
   *
   * @param quote                - server side quote
   * @param fields               - meta field definitions
   * @param eligible_dapis       - dapis that have already been added
   * @param dapi_classifications - classification values for each dapi
   * @returns
   */
  private _addClassificationDapis(
    quote: ServerSideQuote,
    fields: MetaFields,
    eligible_dapis: Record<string, string[]>,
    dapi_classifications: Record<string, boolean>
  ): Record<string, string[]> {
    Object.keys(dapi_classifications).forEach(dapi => {
      if (dapi_classifications[dapi] === false || eligible_dapis[dapi]) {
        return;
      }

      const first_mapped_field = Object.values(fields[dapi].dapi.mapsrc)?.[0];
      const first_mapped_data = quote.getDataByName(first_mapped_field);

      eligible_dapis[dapi] = first_mapped_data.map((_, index) => '' + index);
    });

    return eligible_dapis;
  }

  /**
   * Assign actions to dapis so that they can be triggered appropriately
   *
   * @param eligible_dapis       - dapis that will be triggered
   * @param dapi_classifications - classification values for each dapi
   *
   * @returns information about which dapi index to trigger and how
   */
  private _determineDapiActions(
    eligible_dapis: Record<string, string[]>,
    dapi_classifications: Record<string, boolean>
  ) {
    const dapis_to_request: Record<
      string,
      {indexes: string[]; action: DapiAction}
    > = {};

    Object.keys(eligible_dapis).forEach(dapi => {
      let action = DapiAction.TRIGGER;

      if (dapi_classifications[dapi] === false) {
        action = DapiAction.PEND;
      }

      if (dapi_classifications[dapi] === true && !dapis_to_request[dapi]) {
        action = DapiAction.TRIGGER_PENDING;
      }

      dapis_to_request[dapi] = {indexes: eligible_dapis[dapi], action};
    });

    return dapis_to_request;
  }

  /**
   * Map data from bucket to dapi inputs
   *
   * @param mapsrc    - mapping of dapi inputs to bucket field names
   * @param bucket    - a layered bucket to read data from
   * @param index     - field index
   * @param diff_data - client-provided data
   *
   * @return key/value dapi input data
   */
  private _bucketToDapiInputs(
    mapsrc: Record<string, string>,
    bucket: LayeredBucket,
    index: PositiveInteger,
    diff_data: BucketData
  ): BucketData {
    return Object.keys(mapsrc).reduce<BucketData>((result, srcid) => {
      const bucketid = mapsrc[srcid];

      const bdata =
        (diff_data[bucketid] || [])[index] ||
        (bucket.getDataByName(bucketid) || [])[index];

      if (typeof bdata === 'undefined') {
        // This might be better as an empty string, but we will
        // need to look into it more before we change it. It has
        // been an empty array for a long time and it might cause
        // issues to change it now.
        result[srcid] = [];
      } else {
        result[srcid] = <BucketFieldValue>bdata;
      }

      return result;
    }, {});
  }
}
