/**
 * UserRequest class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EventEmitter} from 'events';
import {UserSession} from './UserSession';
import {IncomingMessage, ServerResponse} from 'http';

/**
 * Encapsulates user request and response data in an easy-to-use class
 *
 * This class doesn't really add any new functionality. It just makes working
 * with requests and responses a bit easier.
 */
export class UserRequest extends EventEmitter {
  /**
   * Request timeout in seconds
   *
   * In the future, we can make this configurable. Currently, it's
   * unnecessary. Feel free to add such a feature if you have need for it.
   */
  private TIMEOUT_SEC = 120;

  /**
   * Requested URI
   */
  private uri = '';

  /**
   * Query (GET vars)
   */
  private query: Record<string, string> = {};

  /**
   * Contains post data, when available
   */
  private postData: undefined | Record<string, unknown> = undefined;

  /**
   * Functions to call when post data is available
   */
  private postDataCallback: ((postData: unknown) => void) | null = null;

  /**
   * HTTP status code to respond with (default 200 OK)
   */
  private responseCode = 200;

  /**
   * HTTP MIME Content type (text/html by default)
   */
  private contentType = 'text/html; charset=utf-8';

  /**
   * Headers to send
   */
  private headers: Record<string, string> = {};

  /**
   * Whether the headers have been sent
   */
  private headersSent = false;

  /**
   * Length of response
   */
  private responseLength = 0;

  /**
   * User's session
   */
  private session: UserSession;

  /**
   * Timer that will cause a timeout after TIMEOUT seconds
   */
  private _timeout: null | NodeJS.Timeout = null;

  /**
   * Constructor
   *
   * @param request        - http.ServerRequest
   * @param response       - http.ServerResponse
   * @param skey           - shared session key used to authenticate instead
   *                         of a session
   * @param session_builder - function to build a session
   *
   * @return undefined
   */
  constructor(
    private readonly request: IncomingMessage,
    private readonly response: ServerResponse,
    private readonly _skey: string,
    session_builder: (sessionId: string) => UserSession
  ) {
    super();

    // eslint-disable-next-line node/no-deprecated-api
    const request_data = require('url').parse(request.url, true);

    this.uri = request_data.pathname;
    this.query = request_data.query || {};

    this._initRequestPost();

    // "session key" used internally for certain scripts
    const request_skey = this.query.skey;

    // initialize session
    this.session = session_builder(this._getCookies()['PHPSESSID']).on(
      'ready',
      data => {
        // if no data is available, then the session could not be
        // initialized; abort the request :( (this should be ignored if
        // a session key is provided, since in that case we do not care
        // about a session)
        if (data === null && !request_skey) {
          this.setResponseCode(500);
          this.end('Session initialization failure.');

          return;
        }

        // now we're ready to roll
        this.emit('ready');
      }
    );

    // set timeout in the event we fail to respond due to some bug/uncaught
    // exception/etc
    this._timeout = setTimeout(() => {
      this.setResponseCode(408);
      this.end('Request timed out.');
    }, this.TIMEOUT_SEC * 1000);
  }

  /**
   * String representation of object
   */
  toString() {
    return '[object UserRequest]';
  }

  /**
   * Watches for post data and returns the data to any waiting callbacks
   */
  private _initRequestPost() {
    const querystring = require('querystring');

    let post_raw = '';

    this.request
      .addListener('data', data => {
        post_raw += data;
      })
      .addListener('end', () => {
        if (this.request.headers['content-type'] === 'application/json') {
          this.postData = JSON.parse(post_raw);
        } else {
          this.postData = querystring.parse(post_raw);
        }

        if (this.postDataCallback) {
          this.postDataCallback(this.postData);
          delete this.postData;
        }
      });
  }

  /**
   * Check whether the proper skey (session key) was provided
   *
   * This is a basic authentication token that allows bypassing authentication
   * for internal tasks (like creating quotes).
   *
   * XXX: A single shared secret is a terrible idea; this was intended to
   * be a temporary solution.  Fix this crap in favor of proper authentication
   * between services.
   */
  has_skey(): boolean {
    if (!this._skey) {
      return false;
    }

    return this.getGetData().skey === this._skey;
  }

  /**
   * Returns the requested URI
   */
  getUri(): string {
    return this.uri;
  }

  /**
   * Returns query (GET) data
   */
  getGetData(): Record<string, string> {
    return this.query;
  }

  /**
   * Requests the post data
   *
   * This is asynchronous. If the data is already available, the callback will
   * be called immediately. If the data is not yet available, it will be
   * called as soon as it becomes available.
   *
   * WARNING: this may return postentially sensitive unencrypted data
   *
   * @param callback - function to call when data is available
   */
  consumePostData(callback: (postData: unknown) => void): this {
    // if we already have the post data, give it to them immediately
    if (this.postData !== undefined) {
      callback.call(this, this.postData);
      delete this.postData;

      return this;
    }

    if (this.postDataCallback) {
      throw new Error('Post data cannot be consumed twice');
    }

    // otherwise, we need to call the callback when the data is available
    this.postDataCallback = callback;

    return this;
  }

  /**
   * Sets the HTTP status code to respond with
   *
   * @param code - HTTP status code
   */
  setResponseCode(code: number): this {
    if (this.headersSent === true) {
      console.error('Headers already sent; response code not set');
      return this;
    }

    this.responseCode = +code;
    return this;
  }

  /**
   * Returns the HTTP status code sent to the client
   *
   * @return HTTP status code
   */
  getResponseCode(): number {
    return this.responseCode;
  }

  /**
   * Sets the content type
   *
   * @param type - content type
   */
  setContentType(type: string): this {
    this.contentType = '' + type;
    return this;
  }

  /**
   * Tells the client not to cache the response
   */
  noCache(): this {
    // the first two are for IE6, the others are HTTP/1.0
    this.headers['Cache-Control'] =
      'private, max-age=0, no-store, no-cache, must-revalidate, ' +
      'post-check=0, pre-check=0';

    return this;
  }

  /**
   * Send headers to the client
   */
  private _sendHeaders(): void {
    this.headers['Content-Type'] = this.contentType;

    this.response.writeHead(this.responseCode, this.headers);
    this.headersSent = true;

    // we don't need this function anymore
    this._sendHeaders = function () {};
  }

  /**
   * Write data to the client
   *
   * @param chunk    - data to write
   * @param encoding - encoding to use
   */
  write(chunk: string, encoding = 'utf8'): this {
    this.responseLength += chunk.length;

    this._sendHeaders();
    this.response.write(chunk, encoding);

    return this;
  }

  /**
   * Send an error response to the client
   *
   * @param data     - error data
   * @param encoding - encoding to use
   */
  error(data: string | Error, encoding?: string): void {
    this.setResponseCode(503);
    this.end(
      JSON.stringify({
        error: {},
        data: data,
      }),
      encoding
    );
  }

  /**
   * Respond with a 200 OK status
   *
   * @param data - optional message to respond with
   */
  ok(data?: string): void {
    this.setResponseCode(200);
    this.end(
      JSON.stringify({
        error: null,
        data: data,
      })
    );
  }

  /**
   * Respond with a 202 Accepted status
   *
   * @param data - optional message to respond with
   */
  accepted(data?: string): void {
    this.setResponseCode(202);
    this.end(
      JSON.stringify({
        error: null,
        data: data,
      })
    );
  }

  /**
   * End client response
   *
   * @param chunk    - data to write
   * @param encoding - encoding to use
   */
  end(data?: string, encoding = 'utf8'): this {
    data = data || '';

    if (this._timeout) {
      clearTimeout(this._timeout);
    }

    this._timeout = null;

    this.responseLength += data.length;

    this._sendHeaders();
    this.response.end(data, encoding);

    this.emit('end');

    return this;
  }

  /**
   * Returns the length of the response
   *
   * Note that this is not very accurate, as this doesn't take into account
   * multi-byte characters.
   *
   * @return response length
   *
   * @todo multibyte
   */
  getResponseLength(): number {
    return this.responseLength;
  }

  /**
   * Returns request cookies as an object
   *
   * @return request cookies
   */
  private _getCookies(): Record<string, string> {
    const cookies: Record<string, string> = {},
      cookie_data = this.request.headers.cookie;

    if (cookie_data === undefined) {
      return {};
    }

    // parse the cookies into an easily accessible object
    cookie_data.split(';').forEach(val => {
      const data = val.split('=', 2);
      cookies[data[0].trim()] = data[1];
    });

    // any future calls to this function will simply return the already
    // generated cookies array to save us some time
    this._getCookies = function () {
      return cookies;
    };

    return cookies;
  }

  /**
   * Returns the current session
   *
   * @return UserSession
   */
  getSession() {
    return this.session;
  }

  getRemoteAddr() {
    // since we may be proxied, let the proxy forward header take precedence
    return (
      this.request.headers['x-forwarded-for'] ||
      this.request.connection.remoteAddress
    );
  }

  getHostAddr() {
    return (
      this.request.headers['x-forwarded-host'] || this.request.headers.host
    );
  }

  getUserAgent() {
    return this.request.headers['user-agent'];
  }

  /**
   * Get the referrer for the request
   *
   * Note: intentionally misspelled, see:
   * https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Referer
   */
  getReferrer() {
    return this.request.headers['referer'];
  }

  getSessionId() {
    return this._getCookies().PHPSESSID || null;
  }

  getHttpVersion() {
    return this.request.httpVersion;
  }

  getUrl() {
    return this.request.url;
  }

  getMethod() {
    return this.request.method;
  }

  getSessionIdName() {
    return 'PHPSESSID';
  }
}
