/**
 * Encryption Service
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {EncryptedKeys} from '../../program/Program';
import {
  Decryptor,
  decryptorFromEncryptedSessionKeys,
  EncryptedSessionKey,
  Encryptor,
  encryptorFromEncryptedSessionKeys,
} from '../../../liza-crypto';
import {BucketData, BucketFieldDataMap} from '../../bucket/Bucket';
import {isInvalid} from '../../validate/invalid';
import {KekService, ServiceSecretMap} from './KekService';

/**
 * Whether a datum ought to be encrypted.
 *
 * If the datum is `null`, then it indicates either an unchanged value or a
 * truncation point and will be left unmodified.
 *
 * If a value is the empty string, it is assumed that
 *   (a) the value is not applicable; and
 *   (b) that the system is configured to prevent the empty string as a
 *       valid applicable value.
 *
 * XXX: Upholding these invariants is a fool's errand, and the system must
 * be able to enforce its own invariants.  Until then, this system is
 * fragile.  It is important that it fail securely: the current predicates
 * do ensure that, if encryption does not take place, that they leak only a
 * little bit of information (missing data or an empty value).
 *
 * TODO: The delta itself should provide an abstraction that provides a
 * mapping operation to identify data that ought to be mapped over (skip
 * nulls, for example).
 */
const shouldEncryptDatum = (value: unknown): boolean =>
  !(value === null || value === '' || isInvalid(value));

/**
 * Whether we should attempt to decrypt a value.
 *
 * The expectation is that the system should not attempt to decrypt data
 * unless it is expecting that decryption will succeed; this allows us to
 * fail safely (on decryption failure) if the system is expecting encrypted
 * data but receives something else, which could indicate a data leak.
 */
const shouldDecryptDatum = (value: unknown): value is string =>
  typeof value === 'string' && value !== '' && !isInvalid(value);

/**
 * Possibly encrypt a delta datum.
 *
 * See `shouldEncryptDatum` for whether a datum will be encrypted.
 */
export const maybeEncryptDeltaDatum =
  (enc: (x: string) => string) =>
  <T>(datum: T): string | T =>
    shouldEncryptDatum(datum) ? enc('' + datum) : datum;

/**
 * Decrypt data for a single field
 *
 * Contrary to `decryptQuoteData`, this function must only be called on
 * fields that are expected to be encrypted—attempting to decrypt a plaintext
 * field will result in an error.
 *
 * This function is suitable for use in situations where either an entire
 * field (array of values) or a single value (scalar) requires decryption.
 * If an array is provied, all values matching `shouldDecryptDatum` must be
 * ciphertexts.
 *
 * Any other value will be returned unmodified.
 */
export const decryptDeltaFieldData =
  (dec: Decryptor): BucketFieldDataMap =>
  <T>(data: T) => {
    if (Array.isArray(data)) {
      return data.map(decryptDeltaFieldData(dec));
    } else if (shouldDecryptDatum(data)) {
      return dec.decrypt(data).getDecryptedText();
    } else {
      return data;
    }
  };

/**
 * Encrypt quote data
 *
 * This is an in place mutation so that unencrypted data is overwritten in the
 * original object. This is done to limit the chances of the unencrypted data
 * being exposed.
 *
 * @param encryptor_ctor         - a function to create an encryptor
 * @param encrypted_keys         - the keys of fields that are to be encrypted
 * @param encrypted_session_keys - encrypted session keys
 * @param to_encrypt             - the data to encrypt
 */
export const encryptQuoteData =
  (encryptor_ctor: (keys: string[]) => Encryptor) =>
  (
    encrypted_keys: EncryptedKeys,
    encrypted_session_keys: string[],
    to_encrypt: BucketData
  ): void => {
    const encryptor = encryptor_ctor(encrypted_session_keys);

    Object.entries(to_encrypt).forEach(([key, values]) => {
      // If the key has a value in the encrypt keys object and val is set
      if (encrypted_keys[key]) {
        values = values.map(
          maybeEncryptDeltaDatum((x: string) => encryptor.encrypt(x))
        );
      }

      to_encrypt[key] = values;
    });
  };

/**
 * Decrypt quote data
 *
 * This is returning a new object so that we do not place unencrypted data in
 * an object that may be used elsewhere. This function creates a new object by
 * destructuring the original object and adding the decrypted data to it.
 *
 * @param decryptor_ctor         - a function to create an decryptor
 * @param encrypted_keys         - the keys of fields that are to be decrypted
 * @param encrypted_session_keys - encrypted session keys
 * @param to_decrypt             - the data to decrypt
 *
 * @returns the decrypted data
 */
export const decryptQuoteData =
  (decryptor_ctor: (keys: string[]) => Decryptor) =>
  (
    encrypted_keys: EncryptedKeys,
    encrypted_session_keys: string[],
    to_decrypt: BucketData
  ): BucketData => {
    const decryptor = decryptor_ctor(encrypted_session_keys);
    const decrypt = decryptDeltaFieldData(decryptor);

    const decrypted_data: BucketData = {};
    const encrypted_fields = Object.keys(encrypted_keys);

    for (const key of encrypted_fields) {
      if (to_decrypt[key]) {
        decrypted_data[key] = to_decrypt[key].map(decrypt);
      }
    }

    return {...to_decrypt, ...decrypted_data};
  };

/**
 * Generates an encryptor from liza's session keys
 *
 * @param kek_map - A mapping of services/versions to keys
 * @param encrypted_session_keys - a list of session keys
 * @returns - An encryptor
 */
export const encryptorFromKekMap = (
  kek_map: ServiceSecretMap,
  encrypted_session_keys: EncryptedSessionKey[]
): Encryptor => {
  const kek = KekService.latestActive(kek_map, KekService.INTERNAL_KEY)[1].key;
  return encryptorFromEncryptedSessionKeys(kek, encrypted_session_keys);
};

/**
 * Generates a decryptor for a specified service
 *
 * @param kek_map - A mapping of services/versions to keys
 * @param encrypted_session_keys - a list of session keys
 * @param service - Service to retrieve decryptor for
 * @returns - A decryptor for a specified service
 */
export const decryptorFromKekMap = (
  kek_map: ServiceSecretMap,
  encrypted_session_keys: EncryptedSessionKey[],
  service: string
): Decryptor => {
  const kek = KekService.latestActive(kek_map, service)[1].key;
  return decryptorFromEncryptedSessionKeys(kek, encrypted_session_keys);
};
