import {SecretManagerServiceClient} from '@google-cloud/secret-manager';
import {Secret, SecretStore} from './SecretStore';
import {Key} from '../../../liza-crypto';

export class GcpSecretStore implements SecretStore {
  constructor(
    private readonly project_id: string,
    private readonly client: SecretManagerServiceClient
  ) {}

  async getSecret(versioned_secret: string): Promise<Key> {
    const response = await this.client.accessSecretVersion({
      name: versioned_secret,
    });
    const key = response?.[0]?.payload?.data?.toString();
    if (!key) {
      throw new Error(`No kek found for ${versioned_secret}`);
    }

    return Buffer.from(key, 'base64');
  }

  async listSecretVersions(service: string): Promise<Secret[]> {
    const versions = this.client.listSecretVersionsAsync({
      filter: 'state = ENABLED',
      parent: this.getSecretName(service),
    });

    const secrets = [];
    for await (const version of versions) {
      if (!version.name) {
        throw new Error('Secret version is un-named and cannot be retrieved');
      }

      const string_version = version.name.match(/.*versions\/(\d+)/);
      // First match is at index 1
      const version_number = parseInt(string_version?.[1] ?? '0');
      if (version_number < 1) {
        throw new Error(
          `Version number is less than one, tried to convert: ${string_version}`
        );
      }
      secrets.push({version: version_number, name: version.name});
    }

    return secrets;
  }

  /**
   * Construct the name for the secret we're fetching
   *
   * @param service - name of the service to fetch the secret for
   */
  private getSecretName(service: string): string {
    return `projects/${this.project_id}/secrets/${service}-kek`;
  }
}
