/**
 * Service for handling KEKs (Key Encrypting Key)
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {Key} from '../../../liza-crypto/';
import {SecretStore} from './SecretStore';

export type Service = string;
export type Version = number;
export type KeyData = {
  key: Key;
  is_active: boolean;
};
/**
 * @example
 * {
 *    1: { key: 'some-key', is_active: true },
 *    2: { key: 'some-key', is_active: true }
 * }
 */
export type KekVersionMap = Record<Version, KeyData>;
/**
 * @example
 * {
 *  some_service: {
 *    1: { key: 'some-key', is_active: true },
 *    2: { key: 'some-key', is_active: true },
 *  }
 * }
 */
export type ServiceSecretMap = Record<Service, KekVersionMap>;

export class KekService {
  public static readonly INTERNAL_KEY: string = 'liza';

  /**
   * Returns interal services session keys
   *
   * @param session_keys - A list of session keys and various services
   *
   * @returns - an array of session keys associated with this internal service
   */
  public static getInternalSessionKeys(
    session_keys: Record<string, string[]> | undefined
  ): string[] {
    const liza_session_keys = session_keys?.[KekService.INTERNAL_KEY];
    if (!liza_session_keys) {
      throw new Error('Missing liza session key');
    }
    return liza_session_keys;
  }

  /**
   * Returns interal key encrypting key (KEK)
   *
   * @param keks - a mapping of keks and their secret versions
   *
   * @returns - the latest active internal service KEK
   */
  public static getLatestInteralKek(keks: ServiceSecretMap): Key {
    return this.latestActive(keks, KekService.INTERNAL_KEY)[1].key;
  }

  /**
   * Given a map of keks and a given service, returns a tuple of the version and key for
   * said service.
   *
   * @param kek_map - a list of keks and their versioned keys
   * @param service - a service to look up
   * @returns {[Version, KeyData]}
   */
  public static sortKeks(
    kek_map: ServiceSecretMap,
    service: string
  ): [Version, KeyData][] {
    return Object.entries(kek_map[service] ?? {})
      .map<[Version, KeyData]>(([v, a]) => [Number(v), a])
      .sort(([v_1, _a_1], [v_2, _a_2]) => (v_1 < v_2 ? 1 : -1));
  }

  /**
   * Latest key, regardless of active status
   * @returns {Key}
   */
  public static latestActive(
    kek_map: ServiceSecretMap,
    service: string
  ): [Version, KeyData] {
    const latest = this.sortKeks(kek_map, service).find(
      ([_v, a]) => a.is_active
    );
    if (!latest) {
      throw new Error(`Missing ${service} kek`);
    }
    return latest;
  }

  /**
   * Instantiate a new KEK service
   *
   * @param client     - a secret manager client
   * @param project_id - the project id
   */
  constructor(
    private readonly client: SecretStore,
    private readonly assertKeyLength: (key: Key) => void,
    private readonly blacklisted_keks: string[]
  ) {}

  /**
   * Fetches and returns the most recent Liza kek
   *
   * @returns Latest Kek for Liza
   */
  public async fetchInternalKek(): Promise<Key> {
    const [keks, liza_service] = await this.fetch(KekService.INTERNAL_KEY);
    return KekService.getLatestInteralKek({[liza_service]: keks});
  }

  /**
   * Given a list of services, fetches all secret versions for each services
   *
   * @param services - A list of services to query GCP for. Liza is always included
   *
   * @returns - A record of versions and their secrets, assigned to their respective services
   * @example
   * {
   *  liza: {
   *    '1': { key: 'key-1', is_active: true },
   *    '2': { key: 'key-2', is_active: true }
   *  },
   *  a_new_service: {
   *    '1': ...
   *  },
   *  ...
   * }
   */
  public async fetchAll(services?: string[]): Promise<ServiceSecretMap> {
    return (
      await Promise.all(
        // Always add Liza's internal key
        [KekService.INTERNAL_KEY, ...(services ?? [])].map(service =>
          this.fetch(service)
        )
      )
    ).reduce((acc, [keks, service]) => ({[service]: keks, ...acc}), {});
  }

  /**
   * Fetches active secret versions for an individual service
   *
   * @param service - specified service to fetch from GCP
   *
   * @returns - a list of Keks for the service, marked by their version numbers and active status
   * @example
   * [
   *   {
   *     '1': { key: 'key-1', is_active: false },
   *     '2': { key: 'key-2', is_active: true },
   *     '3': { key: 'key-3', is_active: true },
   *     ...
   *   },
   *   'liza'
   * ]
   */
  public async fetch(service: string): Promise<[KekVersionMap, Service]> {
    const secret_list = await this.client.listSecretVersions(service);

    const secrets: KekVersionMap = {};
    let has_active = false;
    for (const secret of secret_list) {
      const key = await this.fetchVersionedSecret(secret.name);

      // If not found in black-list assume its active
      const is_blacklisted =
        this.blacklisted_keks.indexOf(`${service}_kek:${secret.version}`) !==
        -1;

      if (!is_blacklisted) has_active = true;

      secrets[secret.version] = {
        key: key,
        is_active: !is_blacklisted,
      };
    }

    // Toss out any invalid keys obtained, but error if NONE are applicable
    if (Object.keys(secrets).length < 1) {
      throw new Error(`No valid keks found for service: ${service}`);
    }
    if (!has_active) {
      throw new Error(`No active keks found for service: ${service}`);
    }

    return [secrets, service];
  }

  /**
   * Fetches details of a specified version of a secret
   *
   * @param secret_name - Full name for the secret, including project and version number
   *
   * @returns - a key encrypting key, in the form of a Buffer.
   */
  private async fetchVersionedSecret(versioned_secret: string): Promise<Key> {
    const key = await this.client.getSecret(versioned_secret);

    try {
      this.assertKeyLength(key);
      return key;
    } catch (e) {
      throw new Error(e instanceof Error ? e.message : `${e}`);
    }
  }
}
