/**
 * Rating service
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import type {
  ClassificationData,
  RateResult,
  WorksheetData,
} from '../rater/Rater';
import type {ClientActions} from '../../client/action/ClientAction';
import type {PositiveInteger} from '../../numeric';
import type {ProcessManager} from '../rater/ProcessManager';
import type {QuoteId} from '../../quote/Quote';
import type {ServerDao} from '../db/ServerDao';
import type {ServerSideQuote} from '../quote/ServerSideQuote';
import type {UserSession} from '../request/UserSession';
import {PriorityLog} from '../log/PriorityLog';
import {getPrefixes, getPrefixesTrue} from '../../bucket/util/commonSuffix';
import {ClassificationResult} from '../../program/Program';

type RequestCallback = () => void;

/** Content result of rating */
export type RateRequestContent = {
  data: RateResult;
  initialRatedDate: UnixTimestamp;
  lastRatedDate: UnixTimestamp;
};

/** Complete Rating Results */
export type RateRequestResult = {
  content: RateRequestContent;
  actions: ClientActions;
};

/**
 * Handle rating requests
 *
 * XXX: This class was extracted from Server and needs additional
 * refactoring, testing, and cleanup.
 *
 * TODO: Logging should be implemented by observers
 */
export class RatingService {
  /** The maximum amount of retries to attempt */
  readonly RETRY_MAX_ATTEMPTS_DEFAULT: PositiveInteger = <PositiveInteger>24;

  /** Seconds to wait between retries */
  readonly RETRY_DELAY_DEFAULT: PositiveInteger = <PositiveInteger>5;

  /**
   * Initialize rating service
   *
   * @param _logger        - logging system
   * @param _dao           - database connection
   * @param _rater_manager - rating manager
   * @param _ts_ctor       - a timestamp constructor
   */
  constructor(
    private readonly _logger: PriorityLog,
    private readonly _dao: ServerDao,
    private readonly _rater_manager: ProcessManager,
    private readonly _ts_ctor: () => UnixTimestamp
  ) {}

  fetchRate(quote: ServerSideQuote): Promise<RateRequestResult> {
    return new Promise<RateRequestResult>((resolve, reject) => {
      const actions: ClientActions = [];
      const data = quote.getRatingData();

      this._processRetries(quote, data, actions);

      // Merge the rate data back into the bucket
      this._dao.mergeBucket(
        quote,
        data,
        () =>
          resolve({
            // send last rated data
            content: {
              data,
              initialRatedDate: quote.getRatedDate(),
              lastRatedDate: quote.getLastPremiumDate(),
            },
            actions: actions,
          }),
        () => reject('Failed merging rate data')
      );
    }).catch(err => {
      this._logRatingError(quote, err);
      throw err;
    });
  }

  /**
   * Sends rates to the client
   *
   * Note that the promise will be resolved after all data saving is
   * complete; the request will be sent back to the client before then.
   *
   * @param session   - user session
   * @param quote     - quote to export
   * @param suppliers - suppliers to rate (or empty)
   * @param force     - (optional) force a rate, overriding quote valid checks
   *
   * @return result promise
   */
  request(
    session: UserSession,
    quote: ServerSideQuote,
    suppliers?: string[]
  ): Promise<RateResult> {
    return new Promise<RateResult>(resolve => {
      resolve(this._performRating(session, quote, suppliers));
    }).catch(err => {
      this._logRatingError(quote, err);
      throw err;
    });
  }

  /**
   * Perform rating and process result
   *
   * @param session   - user session
   * @param quote     - quote to process
   * @param suppliers - suppliers to rate (or empty)
   *
   * @return promise for results of rating
   */
  private _performRating(
    session: UserSession,
    quote: ServerSideQuote,
    suppliers?: string[]
  ): Promise<RateResult> {
    return new Promise<RateResult>((resolve, reject) => {
      const rater = this._rater_manager.byId(quote.getProgramId());

      this._logger.log(
        PriorityLog.PRIORITY_INFO,
        "Performing '%s' rating for%s quote #%s",
        quote.getProgramId(),
        suppliers ? ` suppliers ${suppliers.join(', ')} on` : '',
        quote.getId()
      );

      // We conditionally call classify here if it has not been called yet.
      // Field state is only set on a quote if it has been classified.
      // Classification data is needed for the applicability of layered buckets.
      //
      // It is expected that the initial rating triggered by moving to the rate
      // step will have a quote with classification data. If we are triggering
      // rating from a server-side dapi response, classify is expected to have
      // not been run yet.
      if (!quote.getFieldState()) {
        quote.classify();
      }

      // Only update the rate request timestamp on the first request made
      if (quote.getRetryAttempts() === 0) {
        const ts = this._ts_ctor();
        const meta = {liza_timestamp_rate_request: [ts]};

        quote.setMetadata(meta);
        this._dao.saveQuoteMeta(quote, meta);
      }

      rater.rate(
        quote,
        session,
        suppliers,
        (rate_data: RateResult, actions: ClientActions, override = false) => {
          actions = actions || [];

          this.postProcessRaterData(rate_data, actions, quote);
          quote.setRatingData(rate_data);

          const rate_data_effective = quote.getRatingData();

          const class_dest = {};

          this._cleanRateData(rate_data_effective, class_dest);

          // save all data server-side (important: do after
          // post-processing); async
          this._saveRatingData(
            quote,
            rate_data,
            rate_data_effective,
            class_dest,
            override,
            () => resolve(rate_data_effective)
          );
        },
        (message: string) => {
          this._logRatingError(quote, Error(message));

          reject(Error(message));
        }
      );
    });
  }

  /**
   * Saves rating data
   *
   * Data will be merged with existing bucket data and saved. The idea behind
   * this is to allow us to reference the data (e.g. for reporting) even if
   * the client does not save it.
   *
   * @param quote          - quote to save data to
   * @param data           - rating data
   * @param data_effective - rating data to use for saving
   * @param override       - rating was overridden and results are custom
   * @param c              - callback
   */
  private _saveRatingData(
    quote: ServerSideQuote,
    data: RateResult,
    data_effective: RateResult,
    classes: Record<string, any>,
    override: boolean,
    c: RequestCallback
  ): void {
    // only update the last premium calc date on the initial request
    const cur_date = this._ts_ctor();

    // If we overrode the rating process we do not want to mark it as a
    // valid rate
    quote.setLastPremiumDate(override ? <UnixTimestamp>0 : cur_date);
    quote.setRatedDate(cur_date);

    const save_data: Record<string, any> = {ratedata: data_effective};
    const rdelta_data = {
      'rdelta.ratedata': {
        data: {},
        concluding_save: false,
        timestamp: cur_date,
      },
    };

    // Save quote classes
    for (const key in classes) {
      save_data['classData.' + key] = classes[key];
    }

    this._logger.log(
      PriorityLog.PRIORITY_INFO,
      "'%s' rating with single mongo call for quote #%s; ",
      quote.getProgramId(),
      quote.getId()
    );

    // Merge bucket data
    for (const field in data) {
      if (!field) {
        continue;
      }

      save_data['data.' + field] = data[field];
    }

    this._dao.saveQuote(
      quote,
      <ClassificationResult>{},
      c,
      c,
      save_data,
      rdelta_data
    );
  }

  /**
   * Process rater data returned from a rater
   *
   * @param data     - rating data returned
   * @param actions  - actions to send to client
   * @param quote    - quote used for rating
   */
  protected postProcessRaterData(
    data: RateResult,
    actions: ClientActions,
    quote: ServerSideQuote
  ): void {
    const meta = data._cmpdata || {};

    // the metadata will not be provided to the client
    delete data._cmpdata;

    // rating worksheets are returned as metadata
    this._processWorksheetData(quote.getId(), data);
    this._processRetries(quote, data, actions);

    const program = quote.getProgram();

    if (
      program.ineligibleLockCount > 0 &&
      +meta.count_ineligible >= program.ineligibleLockCount
    ) {
      // lock the quote client-side (we don't send them the reason; they
      // don't need it) to the current step
      actions.push({action: 'lock'});

      const lock_reason = 'Supplier ineligibility restriction';
      const lock_step = quote.getCurrentStepId();

      // the next step is the step that displays the rating results
      quote.setExplicitLock(lock_reason, lock_step + 1);

      // important: only save the lock state, not the step states, as we
      // have a race condition with async. rating (the /visit request may
      // be made while we're rating, and when we come back we would then
      // update the step id with a prior, incorrect step)
      this._dao.saveQuoteLockState(quote);
    }

    // if any have been deferred, instruct the client to request them
    // individually
    if (Array.isArray(meta.deferred) && meta.deferred.length > 0) {
      const torate: string[] = [];

      meta.deferred.forEach((alias: string) => {
        actions.push({action: 'indvRate', id: alias});
        torate.push(alias);
      });

      // we log that we're performing rating, so we should also log when
      // it is deferred (otherwise the logs will be rather confusing)
      this._logger.log(
        PriorityLog.PRIORITY_INFO,
        "'%s' rating deferred for quote #%s; will rate: %s",
        quote.getProgramId(),
        quote.getId(),
        torate.join(',')
      );
    }
  }

  /**
   * Send rating error to user and log
   *
   * @param quote   - problem quote
   * @param err     - error
   */
  private _logRatingError(quote: ServerSideQuote, err: Error): void {
    // well that's no good
    this._logger.log(
      PriorityLog.PRIORITY_ERROR,
      'Rating for quote %d (program %s) failed: %s',
      quote.getId(),
      quote.getProgramId(),
      err.message + '\n-!' + (err.stack || '').replace(/\n/g, '\n-!')
    );
  }

  /**
   * Clear all pending retry attempts
   *
   * @param data Rating results
   *
   * @return only retry attempt fields
   */
  private _clearRetries(data: RateResult): RateResult {
    const cleared: RateResult = {};

    const result_ids = getPrefixes('__unavailable')(data);

    for (const field of result_ids) {
      // Reset the field to zero
      data[field + '___retry'] = [0];
      cleared[field + '___retry'] = [0];
    }

    data['__rate_pending'] = [];
    data['suppliers_pending_rate'] = [];
    cleared['__rate_pending'] = [];

    return cleared;
  }

  /**
   * Process and save worksheet data from rating results
   *
   * @param qid  - quote id
   * @param data - rating result
   */
  private _processWorksheetData(qid: QuoteId, data: RateResult): void {
    // TODO: this should be done earlier on, so that this is not necessary
    const wre = /^(.+)___worksheet$/;

    const worksheets: Record<string, WorksheetData> = {};

    // extract worksheets for each supplier
    for (const field in data) {
      let match;
      if ((match = field.match(wre))) {
        const name = match[1];

        worksheets[name] = data[field];
        delete data[field];
      }
    }

    this._dao.setWorksheets(qid, worksheets, (err: NullableError) => {
      if (err) {
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          'Failed to save rating worksheets for quote %d',
          qid,
          err.message + '\n-!' + (err.stack || '').replace(/\n/g, '\n-!')
        );
      }
    });
  }

  /**
   * Prepares rate data to be sent back to the client
   *
   * There are certain data saved server-side that there is no use serving to
   * the client.
   *
   * @param data    - rate data
   * @param classes - classification data
   *
   * @return modified rate data
   */
  private _cleanRateData(data: RateResult, classes: ClassificationData) {
    // clear class data
    for (const key in data) {
      let mdata;

      // supplier___classes
      if ((mdata = key.match(/^(.*)___classes$/))) {
        classes[mdata[1]] = data[key];
        continue;
      }
    }
  }

  /**
   * Process retry logic
   *
   * @param quote    - quote used for rating
   * @param data     - rating data returned
   * @param actions  - actions to sent to the client
   */
  private _processRetries(
    quote: ServerSideQuote,
    data: RateResult,
    actions: ClientActions
  ): void {
    // Gather determinant factors
    const supplier_total = getPrefixes('__unavailable')(data);
    const retry_total = getPrefixes('___retry')(data);
    const retrying_suppliers = getPrefixesTrue('___retry')(data)(retry_total);

    const retry_attempts = quote.getRetryAttempts();

    // Make the assumption here that if there are no retry fields at all then
    // we do not want deferred rating
    const missing_retries =
      retry_total.length === 0 ? 0 : supplier_total.length - retry_total.length;

    const program = quote.getProgram();
    const cur_step = quote.getCurrentStepId();
    let maxAttemptsByStep = program?.ratingPoll?.maxAttempts;
    if (program.isIndicationQuoteStep(cur_step)) {
      maxAttemptsByStep = program?.ratingPoll?.indicationMaxAttempts;
    }

    const max_attempts = maxAttemptsByStep || this.RETRY_MAX_ATTEMPTS_DEFAULT;

    // Make determinations
    const max_attempts_reached = retry_attempts >= max_attempts;
    const has_pending = missing_retries + retrying_suppliers.length > 0;

    // Clear retry attempts when we have no more pending rates
    quote.setRetryAttempts(has_pending ? retry_attempts + 1 : 0);

    data['__rate_pending'] = retrying_suppliers;
    this._dao.mergeData(quote, {
      'ratedata.__rate_pending': retrying_suppliers,
      retryAttempts: quote.getRetryAttempts(),
    });

    // Provide values for downstream systems to use. This is done separately
    // so that these values can be used and modified without modifying the
    // deferred rating system.
    data['suppliers_rated'] = supplier_total;
    data['suppliers_pending_rate'] = retrying_suppliers;

    const retry_delay =
      quote.getProgram()?.ratingPoll?.delaySeconds || this.RETRY_DELAY_DEFAULT;

    if (has_pending && !max_attempts_reached) {
      // Set rate event value to -1 so that it will be in the background
      actions.push({
        action: 'delay',
        seconds: retry_delay,
        then: {
          action: 'fetchRate',
          value: -1,
        },
      });
    } else if (has_pending && max_attempts_reached) {
      this._logger.log(
        PriorityLog.PRIORITY_ERROR,
        "Rating timeout reached for program '%s' quote #%s step %s and supplier(s) %s",
        quote.getProgramId(),
        quote.getId(),
        cur_step,
        retrying_suppliers.join(', ')
      );

      const clear_data = this._clearRetries(data);
      const save_data = <RateResult>{};

      // Update quote ratedata
      quote.setRatingData(data);

      // do not overwrite data that we still want
      Object.keys(clear_data).forEach(key => {
        save_data['ratedata.' + key] = clear_data[key];
        save_data['data.' + key] = clear_data[key];
      });

      this._dao.saveQuote(quote, {}, () => {}, this._logRatingError, save_data);
    }
  }
}
