import {Program, ServerSideDapi} from '../../program/Program';
import {DataApiConstructor} from '../../dapi/DataApiManager';
import {DapiMetaSource} from '../meta/DapiMetaSource';
import {UnknownDapiError} from '../../error/UnknownDapiError';
import {StagingBucket} from '../../bucket/StagingBucket';
import {DapiProcessor} from '../dapi/DapiProcessor';
import {ServerSideQuote} from '../quote/ServerSideQuote';
import {UserRequest} from '../request/UserRequest';
import {isPositiveInteger, PositiveInteger} from '../../numeric';
import {DataApiResultItem} from '../../dapi/DataApi';
import {isObject} from '../../type_safety';
import {PriorityLog} from '../log/PriorityLog';
import {TokenStore} from '../token/store/TokenStore';
import {TokenId, TokenNamespace} from '../token/Token';
import {DocumentId} from '../../document/Document';
import {InvalidPayloadError} from '../../error/InvalidPayloadError';
import {InvalidTokenError} from '../../error/InvalidTokenError';
import {WebhookBucketDataApi} from '../../dapi/WebhookBucketDataApi';
import {BucketData} from '../../bucket/Bucket';

export class WebhookService {
  constructor(
    private readonly _logger: PriorityLog,
    private readonly _store_ctor: (
      quote_id: DocumentId,
      token_ns: TokenNamespace
    ) => TokenStore,
    private readonly _bucket_ctor: () => StagingBucket,
    private readonly _dapif: DataApiConstructor,
    private readonly _meta_source: DapiMetaSource,
    private readonly _dapi_input_processor: DapiProcessor
  ) {}

  /**
   * Handle an incoming webhook request
   *
   * @param program - the program
   * @param field   - the requested webhook field
   * @param token   - a token to verify the incoming webhook
   * @param data    - the posted payload
   * @param quote   - the quote
   * @param request - incoming request
   */
  async fulfillWebhook(
    program: Program,
    field: TokenNamespace,
    token: TokenId,
    data: unknown,
    quote: ServerSideQuote,
    request: UserRequest
  ): Promise<void> {
    const {name, index, dapi_definition} = this._parseMetaField(field, program);
    const parsed_data = this._validatePayload(data);
    const token_store = this._store_ctor(quote.getId(), field);

    // This will throw an exception if a token cannot be found, it will be
    // handled by the caller
    const token_found = await token_store.lookupToken(token);

    if (!token_found.last_created || !token_store.isDoneable(token_found)) {
      throw new InvalidTokenError();
    }

    const dapi_manager = this._dapif(program.apis, request, quote).on(
      'error',
      err =>
        this._logger.log(
          PriorityLog.PRIORITY_ERROR,
          `Dapi Constructor Error: ${err.message}`
        )
    );

    const bucket = this._bucket_ctor();
    const save_data = this._getDataToSave(quote, field);

    bucket.setValues(save_data);

    return new Promise<void>((resolve, reject) => {
      this._meta_source.mapDapiResponse(
        name,
        index,
        dapi_definition,
        bucket,
        dapi_manager,
        [parsed_data],
        mapped_data => {
          this._dapi_input_processor.monitorMetadataPromise(
            quote,
            [Promise.resolve(mapped_data)],
            Object.fromEntries(
              Object.keys(save_data).map(key => ['data.' + key, save_data[key]])
            ),
            request,
            program
          );

          const token_data = JSON.stringify([parsed_data]);

          // Complete token and accept immediately
          token_store
            .completeToken(token_found, token_data)
            .then(token_done => token_store.acceptToken(token_done, token_data))
            .then(() => resolve())
            .catch(reject);
        },
        reject
      );
    });
  }

  /**
   * Parse the field, index, and dapi that was requested
   *
   * @param field   - the request field
   * @param program - the program
   *
   * @returns an object containing the parsed field, index, and dapi definition
   */
  private _parseMetaField(
    field: string,
    program: Program
  ): {name: string; index: PositiveInteger; dapi_definition: ServerSideDapi} {
    const split_point = field.lastIndexOf('_');
    const name = field.substring(0, split_point);
    const index = +field.substring(split_point + 1);

    if (!isPositiveInteger(index)) {
      throw new Error('Could not parse field: bad input');
    }

    const dapi_definition = program.meta.fields[name]?.dapi;

    if (!dapi_definition) {
      throw new UnknownDapiError();
    }

    return {name, index, dapi_definition};
  }

  /**
   * Ensure that the incoming payload matches the format we expect
   *
   * @param data - the payload to validate
   *
   * @returns a validated payload
   */
  private _validatePayload(data: unknown): DataApiResultItem {
    if (!isObject(data)) {
      throw new InvalidPayloadError();
    }

    if (Array.isArray(data)) {
      throw new InvalidPayloadError('Array not expected');
    }

    return data;
  }

  /**
   * Get data to save to the data bucket
   *
   * @param quote - quote object
   * @param field - the field we are saving
   *
   * @returns bucket data that represents a completion of the webhook request
   */
  private _getDataToSave(quote: ServerSideQuote, field: string): BucketData {
    return {
      [`webhook_trigger_${field}`]: [''],
      [WebhookBucketDataApi.WEBHOOK_AGG_KEY]: quote
        .getDataByName(WebhookBucketDataApi.WEBHOOK_AGG_KEY)
        .filter(val => val !== `webhook_trigger_${field}`),
    };
  }
}
