/**
 * Event Meditator
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Hook events and log them
 */

import {EventEmitter} from 'events';
import {PsrLogger} from './PsrLogger';
import {hasContext} from '../error/ContextError';

export class EventMediator {
  /**
   * Initialize mediator
   *
   * @param _log     - A PSR-3 style logger
   * @param _emitter - An event emitter
   */
  constructor(
    private readonly _log: PsrLogger,
    private readonly _emitter: EventEmitter
  ) {
    this._emitter.on('delta-publish', msg =>
      this._log.notice('Published delta to exchange', msg)
    );

    this._emitter.on('document-processed', msg =>
      this._log.notice(
        'Deltas on document processed successfully. Document has been ' +
          'marked as completely processed.',
        msg
      )
    );

    this._emitter.on('amqp-conn-warn', msg =>
      this._log.warning('AMQP Connection Error', msg)
    );

    this._emitter.on('amqp-reconnect', () =>
      this._log.notice('...attempting to re-establish AMQP connection')
    );

    this._emitter.on('amqp-reconnected', () =>
      this._log.notice('AMQP re-connected')
    );

    this._emitter.on('request-received', (ctx = {}) =>
      this._log.notice('Request Received', ctx)
    );

    this._emitter.on('request-failure', (ctx = {}) =>
      this._log.notice('Request Failed', ctx)
    );

    this._emitter.on('response-sent', (ctx = {}) =>
      this._log.notice('Response Sent', ctx)
    );

    this._emitter.on('callback-sent', (ctx = {}) =>
      this._log.notice('Callback Sent', ctx)
    );

    this._emitter.on('callback-success', (ctx = {}) =>
      this._log.notice('Callback Success', ctx)
    );

    this._emitter.on('callback-failure', (ctx = {}) =>
      this._log.notice('Callback Failure', ctx)
    );

    this._emitter.on('dapi-failure', (ctx = {}) => {
      let err = 'unknown';

      if (Object.prototype.hasOwnProperty.call(ctx, 'error')) {
        if (typeof ctx.error === 'string') {
          err = ctx.error;
        } else if (ctx.error instanceof Error) {
          err = ctx.error.message;
        }
      }

      this._log.notice(`DAPI Failure: ${err}`, ctx);
    });

    this._emitter.on('rate-error', (ctx = {}) =>
      this._log.error('Rate Error', ctx)
    );

    this._emitter.on('quote-save-success', (ctx = {}) =>
      this._log.notice('Quote Save Success', ctx)
    );

    this._emitter.on('quote-save-failure', (ctx = {}) =>
      this._log.error('Quote Save Failure', ctx)
    );

    this._emitter.on('quote-validation-failure', (ctx = {}) =>
      this._log.error('Quote Validation Failure', ctx)
    );

    this._emitter.on('quote-validation-warning', (ctx = {}) =>
      this._log.warning('Quote Validation Warning', ctx)
    );

    this._emitter.on('rate-polling-begin', (ctx = {}) =>
      this._log.debug('Starting to poll for rated suppliers', ctx)
    );

    this._emitter.on('pending-rate', (ctx = {}) =>
      this._log.debug('A supplier is still pending', ctx)
    );

    this._emitter.on('rating-completed', (ctx = {}) =>
      this._log.debug('All suppliers have finished rating', ctx)
    );

    this._emitter.on('rating-timeout', (ctx = {}) =>
      this._log.warning('The rate polling process has timed out', ctx)
    );

    this._emitter.on('copy-request-received', (ctx = {}) =>
      this._log.notice('Received copy request', ctx)
    );

    this._emitter.on('copy-request-successful', (ctx = {}) =>
      this._log.notice('Successful copy request', ctx)
    );

    this._emitter.on('copy-request-unsuccessful', (ctx = {}) =>
      this._log.warning('Unsuccessful copy request', ctx)
    );

    this._emitter.on('init-request-received', (ctx = {}) =>
      this._log.notice('Received init request', ctx)
    );

    this._emitter.on('init-request-successful', (ctx = {}) =>
      this._log.notice('Successful init request', ctx)
    );

    this._emitter.on('init-request-unsuccessful', (ctx = {}) =>
      this._log.warning('Unsuccessful init request', ctx)
    );

    this._emitter.on('error', arg => this._handleError(arg));
  }

  /**
   * Handle an error event
   *
   * @param e - any
   */
  private _handleError(e: any): void {
    let msg = '';
    let context: Record<string, any> = {};

    if (e instanceof Error) {
      msg = e.message;

      if (hasContext(e)) {
        context = e.context;
      }

      context.stack = e.stack;
    }

    this._log.error(msg, context);
  }
}
