import {NodeTracerProvider} from '@opentelemetry/node';
import {BatchSpanProcessor} from '@opentelemetry/tracing';
import {registerInstrumentations} from '@opentelemetry/instrumentation';
import {ExpressInstrumentation} from '@opentelemetry/instrumentation-express';
import {HttpInstrumentation} from '@opentelemetry/instrumentation-http';
import {MongoDBInstrumentation} from '@opentelemetry/instrumentation-mongodb';
import {OTLPTraceExporter} from '@opentelemetry/exporter-trace-otlp-http';
import {IncomingMessage} from 'http';
import {RequestOptions} from 'https';

/**
 * Initialize the tracing
 */
export function init(
  provider: NodeTracerProvider,
  env: NodeJS.ProcessEnv
): void {
  if (!env.OPENTELEMETRY_COLLECTOR_EXPORTER_ENABLED) {
    return;
  }

  const collectorOptions = {
    serviceName: env.OPENTELEMETRY_SERVICE_NAME ?? 'unknown-liza-service',
    url: env.OPENTELEMETRY_URL ?? 'http://localhost:4318/v1/traces',
  };

  const exporter = new OTLPTraceExporter(collectorOptions);
  const processor = new BatchSpanProcessor(exporter);
  provider.addSpanProcessor(processor);
  provider.register();

  registerInstrumentations({
    tracerProvider: provider,
    instrumentations: [
      new ExpressInstrumentation(),
      new MongoDBInstrumentation(),
      new HttpInstrumentation({
        ignoreIncomingRequestHook: (request: IncomingMessage) =>
          request.url?.match(/^\/metrics|healthcheck/) !== null,
        ignoreOutgoingRequestHook: (request: RequestOptions) =>
          [
            'pushgateway.service.consul:9091',
            'dmz2docker01.rsgcorp.local:9091',
          ].includes(request?.host ?? ''),
      }),
    ],
  });
}
