/**
 * Liza Fields Store
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {QuestionTypes, SelectData} from '../../program/Program';

/**
 * Store of fields data that can be accessed by
 * any field context or field styler from any step
 */
export class FieldsStore {
  /**
   * Initialize FieldsStore
   *
   * @param _qtypes      - question types
   * @param _select_data - select data
   */
  constructor(
    private readonly _qtypes: QuestionTypes,
    private readonly _select_data: SelectData
  ) {}

  /**
   * Set the label of a select by name and value
   *
   * Select labels needs to be accessible
   * by displays that could be on a different group or step,
   * and those labels can be modified by DAPIs
   *
   * @param name - name (ref id) of field
   * @param value - value
   * @param label - label
   */
  setSelectLabel(name: string, value: string, label: string): void {
    if (this._select_data[name] !== undefined) {
      this._select_data[name][value] = label;
    }
  }

  /**
   * Return the string associated with the given value
   * (the text for the option), or the given value if it does not exist
   *
   * @param name - name (ref id) of field
   * @param value - value of field
   */
  getSelectLabel(name: string, value: string): string {
    const val = this._select_data[name][value];
    return val === undefined ? value : val;
  }

  /**
   * Return the question type
   *
   * @param name - name (ref id) of field
   */
  getQuestionType(name: string): string {
    return this._qtypes[name]?.type || this._qtypes[name];
  }
}
