/**
 * Liza Date Picker Field Styler
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {PositiveInteger} from '../../../numeric';
import {ContextContent} from '../FieldContext';
import {FieldStyler} from './FieldStyler';

/**
 * Styles Date picker fields
 */
export class DatePickerFieldStyler implements FieldStyler {
  /**
   * Field element ID prefix
   */
  private _field_id_prefix = 'q_';

  /**
   * Initialize DatePickerFieldStyler
   *
   * @param name field name
   * @param index field index
   */
  constructor(
    protected readonly name: string,
    protected readonly index: PositiveInteger
  ) {}

  /**
   * Set value of the field
   *
   * @param content - field content
   * @param value - value to set
   */
  setValue(content: ContextContent, value: string): void {
    const element = this._getElement(content);

    let formattedDate = '';

    // The value for a datePicker field is formatted as m/d/YYYY
    // HTML5 date value must be formatted as yyyy-MM-dd format
    const dateObject = new Date(value);

    // Validate this is a date before parsing
    if (dateObject.getDate()) {
      formattedDate = dateObject.toISOString().split('T')[0];
    }

    if (element !== null && formattedDate !== '') {
      element.value = '' + formattedDate;
    }
  }

  /**
   * Disable field
   *
   * @param content - field content
   * @param disable - boolean
   */
  disable(content: ContextContent, disable: boolean): void {
    const element = this._getElement(content);

    if (element !== null) {
      element.disabled = disable;
    }
  }

  /**
   * Get field element
   *
   * @param content - field content
   */
  private _getElement(content: ContextContent): HTMLInputElement | null {
    const id = this._field_id_prefix + this.name + '_' + this.index;

    const element = content.querySelector('#' + id);

    if (element instanceof HTMLInputElement) {
      return element;
    }

    // In case the element is not found by id, we can query by data-field-name.
    // The index is not needed because the content is at the field level
    const elem_by_field = content.querySelector(
      `[data-field-name="${this.name}"]`
    );

    if (elem_by_field instanceof HTMLInputElement) {
      return elem_by_field;
    }

    return null;
  }
}
