/**
 * Liza Live Search Field Styler
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {PositiveInteger} from '../../../numeric';
import {ContextContent} from '../FieldContext';
import {FieldStyler} from './FieldStyler';
import {LiveSearch, LiveSearchProps} from '../../question/LiveSearch';

/**
 * Styles live search DOM fields
 */
export class LiveSearchFieldStyler implements FieldStyler {
  /**
   * Field is disabled
   */
  private _is_disabled = false;

  /**
   * Initialize LiveSearchFieldStyler
   *
   * @param _name          - field name
   * @param _index         - field index
   * @param _doc           - dom
   * @param _createElement - a function to create a react element
   * @param _render        - a function to render a react element on the dom
   */
  constructor(
    protected readonly _name: string,
    protected readonly _index: PositiveInteger,
    protected readonly _doc: Document,
    protected readonly _createElement: CreateElement<
      LiveSearchProps,
      LiveSearch
    >,
    protected readonly _render: ReactDOM.Renderer
  ) {}

  /**
   * Set value of the live search field
   *
   * @param content - field content
   * @param value   - value to set
   */
  setValue(content: ContextContent, _value: string): void {
    this._createLiveSearch(content);
  }

  /**
   * Disable input and select field
   *
   * @param content - field content
   * @param disable - boolean
   */
  disable(content: ContextContent, disable: boolean): void {
    this._is_disabled = disable;
    this._createLiveSearch(content);
  }

  /**
   * Create Live Search component
   *
   * In preprocessing, two additional fields are generated with the live search
   * question. They handle the input and select logic for the dapi. The search
   * component monitors these two hidden questions so we will need to find them
   * and pass them in as properties.
   *
   * @param content - field content
   */
  private _createLiveSearch(content: ContextContent): void {
    const anchor = content.querySelector(`[data-field-name="${this._name}"]`);
    const cleaned_name = this._name.replace('_anchor', '');
    const parent = content.parentNode;

    const input = parent?.querySelector(
      `[id='q_${cleaned_name}_input_${this._index}']`
    );

    const select = parent?.querySelector(
      `[id='q_${cleaned_name}_${this._index}']`
    );

    if (!input || !select) {
      return;
    }

    const live_search = this._createElement(
      LiveSearch,
      {
        input: <HTMLInputElement>input,
        select: <HTMLSelectElement>select,
        document: this._doc,
        disabled: this._is_disabled,
      },
      null
    );

    this._render(live_search, anchor);
  }
}
