/**
 * Creates DomField
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {BucketField} from '../../field/BucketField';
import {ElementStyler} from '../ElementStyler';
import {DomField} from './DomField';

export class DomFieldFactory {
  constructor(
    private readonly _element_styler: ElementStyler,
    private readonly _console: Console
  ) {}

  /**
   * Create a DomField from the given field description
   *
   * The provided DomField will wait to access the DOM until an operation
   * requires it.
   *
   * @param name  - field name
   * @param index - field index
   * @param root  - root element containing the field (optionally lazy)
   *
   * @return generated field
   */
  public create(
    name: string,
    index: number,
    root: ((cb: (element: HTMLElement) => void) => void) | HTMLElement
  ): DomField {
    return new DomField(
      new BucketField(name, index),

      // lazy load on first access
      callback => {
        const c = (root_cur: HTMLElement) => {
          callback(
            this._element_styler.getElementByNameLax(
              name,
              index,
              null,
              root_cur
            )[0]
          );
        };

        // are we lazy?
        if (typeof root === 'function') {
          // wait to fulfill this request until after the element
          // becomes available
          root(result => {
            root = result;
            c(root);
          });

          return;
        }

        // not lazy; continue immediately
        c(root);
      },
      this._console
    );
  }
}
