/**
 * Flat UI group with accordion display
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {isPositiveInteger} from '../../numeric';
import {ClientQuote} from '../../client/quote/ClientQuote';
import {GroupUi} from './GroupUi';
import {Group} from '../../group/Group';
import {GroupContext} from '../context/GroupContext';
import {ElementStyler} from '../ElementStyler';
import {GroupStateManager} from './GroupStateManager';
import jQuery = require('jquery');

/**
 * Collapsable accordion group
 *
 * This group is similar to a stacked group,
 * but the HTML content is specific to accordion
 */
declare type jQuery = typeof jQuery;

export class AccordionGroupUi extends GroupUi {
  /**
   * Containing group DOM element
   */
  private _container: HTMLElement;

  /**
   * Container for the initial row
   */
  private _row_container: HTMLElement;

  /**
   * Add Button for accordion
   */
  private _add_button: HTMLElement;

  constructor(
    group: Group,
    content: HTMLElement,
    styler: ElementStyler,
    jquery: jQuery,
    context: GroupContext,
    protected readonly _state_manager: GroupStateManager
  ) {
    super(group, content, styler, jquery, context, _state_manager);
    const container: HTMLElement | null = this.content.querySelector(
      'div.liza-accordion-group'
    );
    if (!container) throw Error('Container not initialized');
    this._container = container;
    const rowContainer: HTMLElement | null = this._container.querySelector(
      '.liza-accordion-row-container'
    );
    if (!rowContainer) throw Error('No initial row copied');
    this._row_container = rowContainer;
    const addButton: HTMLElement | null = this._container.querySelector(
      '.liza-add-accordion'
    );
    if (!addButton) throw Error('No addition button found');
    this._add_button = addButton;
  }

  /**
   * Process group before initial display
   *
   * @param quote - active quote
   */
  protected processContent(_quote?: ClientQuote): void {
    // determine if we should lock this group down
    if (this._container.classList.contains('locked')) {
      this.group.locked(true);
    }
    this._row_container.parentNode?.removeChild(this._row_container);
    this._attachAddAccordionHandlers();
    this.watchFirstElement(this._row_container, _quote);
  }

  /**
   * Get remove accordion buttons for hiding and display
   */
  private _getRemoveButtons(): NodeListOf<HTMLElement> {
    return this._container.querySelectorAll('.liza-remove-accordion');
  }

  /**
   * add onclick handler events, also hide delete if first element
   */
  private _attachAddAccordionHandlers(): void {
    if (
      this.group.locked() ||
      this.getCurrentIndex() === this.group.maxRows() - 1
    ) {
      this._add_button.style.display = 'none';
    }
    this._add_button.onclick = () => this.initIndex();
  }

  /**
   * Add index (stacked item)
   *
   * @param index - index that has been added (0-indexed)
   */
  protected addIndex(index: number): this {
    super.addIndex(index);
    this._addAccordion();

    if (!isPositiveInteger(index)) {
      throw new Error('Could not parse index: bad input');
    }

    // if we reach max row count, hide add button
    if (index === this.group.maxRows() - 1) {
      this._add_button.style.display = 'none';
    }

    return this;
  }

  /**
   * on emmited event for lock. hide and remove an customizable elements
   */
  public hideAddRemove(value: boolean): this {
    const removeButtons = this._getRemoveButtons();
    const max = this.group.maxRows();
    if (value) {
      this._add_button.style.display = 'none';
      removeButtons.forEach(
        (button: HTMLElement) => (button.style.display = 'none')
      );
    } else {
      if (max === 0 || removeButtons.length < max) {
        this._add_button.style.display = 'block';
      }
      removeButtons.forEach(
        (button: HTMLElement) => (button.style.display = 'block')
      );
    }
    return this;
  }

  /**
   * Creates a new accordion to append to group
   */
  private _addAccordion(): void {
    const index = this.getCurrentIndex();
    if (!isPositiveInteger(index)) return;

    const item = this._row_container.cloneNode(true);

    if (!(item instanceof HTMLElement)) return;
    item.id = `${item.id}_${index}`;

    // properly name the elements to prevent id conflicts
    this.setElementIdIndexes(item.getElementsByTagName('*'), index);

    const item_dl: HTMLElement | null = item.querySelector('dl');

    // add the index to the row title
    const label: HTMLElement | null = item.querySelector(
      '.liza-accordion-head'
    );

    if (label) {
      const currentLabelText = label.innerText;
      label.innerText = currentLabelText + ' ' + (index + 1);
      label.onclick = (e: MouseEvent) => this._handleChange(e);
      label.click();
    }

    // Set field content parent for this index
    if (item_dl) {
      this.fieldContentParent[index] = item_dl;
    }

    this.context.addIndex(index, this.fieldContentParent[index]);

    this._container.appendChild(item);

    // Todo: Transitional step to remove jQuery
    const $item = this.jquery(item);

    this.styler.apply($item);
    this.postAddRow($item, index);

    const removeButton = item.querySelector('.liza-remove-accordion');
    if (removeButton instanceof HTMLElement) {
      if (this.group.locked()) {
        removeButton.style.display = 'none';
      }
      removeButton.onclick = () => {
        this.destroyIndex(index);
      };
    }
  }

  /**
   * Handle change to any liza-accordion-head
   *
   * Change the height for the accordion effect
   * by adding/removing CSS class
   */
  private _handleChange(e?: MouseEvent): void {
    const contentElements = this._container.querySelectorAll(
      '.liza-accordion-content'
    );
    const arrowElements = this._container.querySelectorAll(
      '.liza-accordion-head'
    );

    contentElements.forEach((value: Element) => {
      value.classList.remove('liza-accordion-expand');
    });
    arrowElements.forEach((value: Element) => {
      value.classList.remove('liza-accordion-head-expanded');
    });

    if (!e) {
      const row = this._container.querySelector(
        '.liza-accordion-row-container'
      );
      const element = row?.querySelector('.liza-accordion-content');
      const head = row?.querySelector('.liza-accordion-head');
      element?.classList.add('liza-accordion-expand');
      head?.classList.add('liza-accordion-head-expanded');
      return;
    }

    const target = e.target;
    if (target instanceof HTMLElement) {
      const expandElements = target.parentNode?.querySelectorAll(
        '.liza-accordion-content'
      );
      const changeArrow = target.parentNode?.querySelector(
        '.liza-accordion-head'
      );

      if (expandElements && changeArrow) {
        expandElements.forEach((value: Element) => {
          value.classList.add('liza-accordion-expand');
        });
        changeArrow.classList.add('liza-accordion-head-expanded');
      }
    }
  }

  /**
   * Remove index (accordion item/row)
   *
   * @param index - index that has been removed (0-indexed)
   */
  public removeIndex(index: number): this {
    if (this.getCurrentIndexCount() === 1) return this;
    this.contextRemoveIndex(index);
    this._removeAccordion();
    return super.removeIndex(index);
  }

  /**
   * Remove accordion and include addbutton if not already shown
   */
  private _removeAccordion(): this {
    const index = this.getCurrentIndex();
    const items = this._container.querySelectorAll(
      '.liza-accordion-row-container'
    );

    this.styler.remove(this.jquery(items[index]));
    this._add_button.style.display = 'block';

    if (
      this._container.querySelectorAll('.liza-accordion-expand').length === 0
    ) {
      const target = items[index - 1].querySelector('.liza-accordion-head');
      if (target instanceof HTMLElement) target.click();
    }

    return this;
  }

  /**
   * Hide the header if there are no visible fields
   *
   * @param field
   * @param index
   */
  public hideField(field: string, index: number) {
    super.hideField(field, index);

    if (!this.hasVisibleField(index)) {
      // This possible effects of ignoring these errors has been
      // deemed less severe than the current effects when it fails
      // (i.e. showing all the questions).
      // This will also eventually be changed to have the elements
      // be hidden by default and show when appropriate. This will
      // likely fix the current issue and reduce the chances of
      // something like this happening.
      try {
        const header = this._container.querySelectorAll(
          '.liza-accordion-row-container'
        )[index];

        if (header) {
          header.classList.add('hidden');
        }
      } catch (e) {
        console.warn(e);
      }
    }
  }

  /**
   * Show the header if there are visible fields
   *
   * @param field
   * @param index
   */
  public showField(field: string, index: number) {
    super.showField(field, index);

    if (this.hasVisibleField(index)) {
      const header = this._container.querySelectorAll(
        '.liza-accordion-row-container'
      )[index];

      if (header) {
        header.classList.remove('hidden');
      }
    }
  }

  /**
   * Display the requested field
   *
   * If field needs to be displayed to the user, ensure
   * the accordion group it belongs to is expanded
   *
   * @param field_name - name of field to display
   * @param i          - index of field
   */
  public displayField(field: string, index: number): this {
    const element = this.styler.getWidgetByName(field, index);

    // if we were unable to locate it, then don't worry about it
    if (element.length === 0) {
      console.log(
        `displayField: Element ${field} with index ${index} is not found`
      );
      return this;
    }

    // locate the accordion input for this index
    const input = this._container.querySelectorAll('.liza-accordion-head')[
      index
    ];

    if (input instanceof HTMLElement) {
      input.click();
    }

    return this;
  }
}
