/* TODO auto-generated eslint ignore, please fix! */
/* eslint @typescript-eslint/no-var-requires: "off", @typescript-eslint/no-this-alias: "off", @typescript-eslint/no-inferrable-types: "off", no-var: 'off', prefer-arrow-callback: 'off', no-unused-vars: 'off', prefer-const: 'off', no-undef: 'off' */
/**
 * General UI logic for groups
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @needsLove
 *   - References to "quote" should be replaced with generic terminology
 *     representing a document.
 *   - Dependencies need to be liberated: Styler; Group.
 *   - This class is doing too much.
 * @end needsLove
 */

import {Group} from '../../group/Group';
import {ElementStyler} from '../ElementStyler';
import {GroupContext} from '../context/GroupContext';
import {GroupStateManager} from './GroupStateManager';
import {ClientQuote} from '../../client/quote/ClientQuote';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {PositiveInteger} from '../../numeric';
import {EventEmitter} from 'events';

declare type jQuery = any;

/**
 * Styles a group for display in the UI
 */
export class GroupUi extends EventEmitter {
  /**
   * Raised when an index is added to the group (e.g. row addition)
   */
  readonly EVENT_INDEX_ADD: string = 'indexAdd';

  /**
   * Raised when an index is removed from the group (e.g. row deletion)
   */
  readonly EVENT_INDEX_REMOVE: string = 'indexRemove';

  /**
   * Raised when an index is reset rather than removed
   */
  readonly EVENT_INDEX_RESET: string = 'indexReset';

  /**
   * Emitted when a row/tab/etc is added to a group
   */
  readonly EVENT_POST_ADD_ROW: string = 'postAddRow';

  /**
   * An action taken by the user
   */
  readonly EVENT_ACTION: string = 'action';

  /**
   * Group being styled
   */
  group: Group;

  /**
   * Group content
   */
  $content: jQuery;

  /**
   * Styler used to style elements
   */
  styler: ElementStyler;

  /**
   * Functions to call when group is invalidated
   * @type {Array.<Function>}
   */
  invalidateHooks: any[] = [];

  /**
   * Whether the group is visible
   */
  private _visible: boolean = true;

  /**
   * Number of indexes (1-based)
   */
  private _indexCount: number = 0;

  /**
   * Whether the group is active (available for the user to interact with)
   */
  private _active: boolean = false;

  /**
   * Continuation to perform on step visit
   * @type {function()}
   */
  private _emptyOnVisit: any = null;

  /**
   * Field visibility cache by field id (reduces DOM lookups)
   */
  private _visCache: Record<string, boolean[]> = {};

  /**
   * Number of visibile fields per index
   */
  private _visCount: any = [];

  /**
   * Number of visibile fields, including multiple indexes
   */
  private _visCountTotal: number = 0;

  /**
   * Number of unique fields, disregarding indexes
   */
  private _rawFieldCount: number = 0;

  /**
   * An array of classifications with css classes we
   * would like to bind to them
   *
   * Structured: classification => css class
   */
  private _bind_classes: any[] = [];

  /**
   * DOM group context
   */
  protected context: GroupContext;

  /**
   * Root DOM context (deprecated)
   * @type {DomContext}
   */
  protected rcontext: any = null;

  /**
   * Group context
   */
  protected content: HTMLElement;

  /**
   * Array of direct parent of field content per index
   */
  protected fieldContentParent: HTMLElement[] = [];

  /**
   * jQuery object
   */
  protected jquery: jQuery = null;

  /**
   * Whether or not we will determine visibility with classifications
   */
  protected _use_vis_class: boolean = true;

  /**
   * Initializes GroupUi
   *
   * @param group         - group to style
   * @param content       - the group content
   * @param styler        - styler to use to style elements
   * @param jquery        - jQuery-compatible object
   * @param context       - group context
   * @param state_manager - state manager for the group
   */
  constructor(
    group: Group,
    content: HTMLElement,
    styler: ElementStyler,
    jquery: jQuery,
    context: GroupContext,
    protected readonly _state_manager: GroupStateManager
  ) {
    super();

    this.group = group;
    this.content = content;
    this.styler = styler;
    this.jquery = jquery;
    this.context = context;

    // Todo: Transition away from jQuery
    this.$content = this.jquery(content);
  }

  public init(quote: ClientQuote) {
    const fields = this.group.getExclusiveFieldNames();
    const cmatch_fields = this.group.getExclusiveCmatchFieldNames();

    this.context.init(
      fields,
      cmatch_fields,
      this.content,
      this.group.isInternal()
    );

    const detach_fields = this.supportsMultipleIndex() ? fields : cmatch_fields;
    this.context.detachStoreContent(detach_fields);

    this._initActions();
    this._monitorIndexChange(quote);
    this.processContent(quote);

    // in an attempt to prevent memory leaks
    this._emptyOnVisit = null;

    // get the number of unique fields in the group
    this._rawFieldCount = this.group.getUserFieldNames().length;

    if (this.group.getWhenFieldName() === '') {
      this._use_vis_class = false;
    }

    this._bindClasses(quote);

    return this;
  }

  private _monitorIndexChange(quote: ClientQuote) {
    var _self = this,
      first = this.getFirstElementName();

    quote.on('preDataUpdate', function (data) {
      // ignore if the data has not changed
      if (data[first] === undefined) {
        return;
      }

      // get the data from the bucket (the staging data is not useful for
      // determining the length, since it only includes what has changed,
      // which will be void of any higher, unchanged indexes)
      var blen = quote.getDataByName(first).length,
        dlen = _self._stripRm(data[first]).length,
        len = 0,
        rm: any[] = [];

      // did we remove an index? if so, then this represents the correct
      // length.
      if (dlen < data[first].length) {
        len = dlen;

        // keep a record of which indexes were removed
        var df = data[first];
        for (var i = 0, l = df.length; i < l; i++) {
          if (df[i] === null) {
            rm.push(i);
          }
        }
      } else {
        // otherwise, it's a change; take the longer length
        len = blen > dlen ? blen : dlen;
      }

      if (len === _self.getCurrentIndexCount()) {
        return;
      }

      function doempty() {
        // TODO: knock it off.
        quote.visitData(function (_: any) {
          // we cannot call preEmptyBucket because the bucket, at this
          // point, has not yet been modified with the new data
          _self._quickIndexChange(true, len, rm);
        });
      }

      // if we're not an active group, then theres little use in updating
      // ourselves; next time they visit the step, we can update ourselves
      // (not only does this help with performance, but it also eliminates
      // some problems that may occur due to us not being attached to the
      // DOM)
      if (!_self._active) {
        // store the latest continuation
        _self._emptyOnVisit = doempty;
        return;
      }

      doempty();
    });
  }

  /**
   * Strip removes from a data diff
   *
   * Removes are represented by nulls
   *
   * @param data - data diff
   *
   * @return data with nulls stripped
   */
  private _stripRm(data: any): any {
    for (var i in data) {
      if (data[i] === null) {
        // null marks the end of the data
        return data.slice(0, i);
      }
    }

    return data;
  }

  /**
   * Performs any necessary processing on the content before its displayed
   *
   * Subtypes may override this for custom functionality
   */
  protected processContent(_quote?: ClientQuote) {}

  /**
   * Group types that can support multiple index
   */
  protected supportsMultipleIndex(): boolean {
    return true;
  }

  /**
   * Trigger events on action interaction
   */
  private _initActions() {
    var _self = this;

    this.$content.find('.action').live('click', function (e: any) {
      e.preventDefault();

      // TODO: index
      var $this = _self.jquery(e.target),
        ref = $this.attr('data-ref'),
        type = $this.attr('data-type'),
        index = +$this.attr('data-index') || 0;

      _self.emit(_self.EVENT_ACTION, type, ref, index);
    });
  }

  /**
   * Attempts to return the group id of the group containing the given element
   *
   * This method will cache the value of the id upon the first request by
   * replacing itself with a function that returns only the value.
   *
   * @return group id or undefined if no group/id exists
   */
  public getGroupId(): string | undefined {
    var id = this.$content.attr('id');

    return id;
  }

  /**
   * Sets the index (for the name attribute) of all given elements
   *
   * The name format is expected to be: name_i, where i is the index.
   *
   * @param elements - elements to set index on
   * @param index    - index to set
   */
  protected setElementIdIndexes(
    elements: HTMLCollectionOf<Element>,
    index: number
  ) {
    for (let i = 0; i < elements.length; i++) {
      const element = elements[i];
      const id: string = element.getAttribute('id') || '';
      let element_data: string[] | null;
      let element_label_data: string[] | null;

      // grab the index from the id if found
      if ((element_data = id.match(/^(.*?)(\d+)$/))) {
        // regenerate the id
        element.setAttribute('id', element_data[1] + index);
      }

      // For input labels, also update "for" attributes
      // (currently needed for Accordion group)
      const id_for: string | null = element.getAttribute('for');
      // grab the index from the id if found
      if (id_for && (element_label_data = id_for.match(/^(.*?)(\d+)$/))) {
        // regenerate the id
        element.setAttribute('for', element_label_data[1] + index);
      }

      element.setAttribute('data-index', index.toString());
    }
  }

  /**
   * Watches the first element for changes and invalidates the group when it
   * does
   *
   * This is used when groups base their row/tab/whatever count on the first
   * element to ensure that they are properly regenerated when the count
   * changes.
   */
  protected watchFirstElement($base: any, quote: any) {
    var group = this,
      first_name = this.getFirstElementName($base);

    if (first_name) {
      quote.on('dataCommit', function (data: any) {
        var first_data = data[first_name];
        if (first_data === undefined) {
          return;
        }

        group.invalidate();
      });
    }
  }

  /**
   * Retireve the current index count
   *
   * This should be one more than the current 0-based index (like an array
   * length). Subtypes may override this if they do not wish to use the
   * built-in index tracking.
   *
   * @return index count
   */
  public getCurrentIndexCount() {
    return this._indexCount;
  }

  public getCurrentIndex() {
    return this.getCurrentIndexCount() - 1;
  }

  /**
   * Allows groups to do any necessary processing before a bucket is emptied
   *
   * For example, a group may need to recreate rows in order to make room for
   * the values stored in the bucket.
   *
   * Subtypes are free to override this if the default functionality is
   * insufficient. Note that there are a number of methods used by this one
   * that too may be overridden to alter its functionality without overriding
   * this method.
   *
   * @param bucket  - bucket
   * @param updated - whether this is an update (rather than inital
   *                          append)
   */
  public preEmptyBucket(bucket: QuoteDataBucket, updated: boolean): this {
    const first = this.getFirstElementName();

    // getDataByName should never be returning null here but if it does, a
    // length of 1 seems like and appropriate default because that is the
    // default length of a field in the bucket. Ideally we would log when this
    // happens so we know.
    const flen: number = bucket.getDataByName(first)?.length ?? 1;

    this._quickIndexChange(updated, flen);

    return this;
  }

  private _quickIndexChange(
    updated: boolean,
    len: number,
    rm: any[] = []
  ): void {
    var _self = this,
      curlen = this.getCurrentIndexCount();

    this.handleIndexChange(
      len,
      curlen,
      function __add(n: number) {
        do {
          _self.addIndex(len - n);
        } while (--n);
      },

      function __rm(n: number) {
        while (n--) {
          _self.removeIndex(rm.pop());
        }
      }
    );

    this.postPreEmptyBucket();

    if (!updated) {
      this.postPreEmptyBucketFirst();
    }
  }

  /**
   * Indicates that a tab/row/column/etc (representing an index) should be
   * added
   *
   * Subtypes may override this for custom functionality.
   *
   * @param index - index that has been added
   */
  protected addIndex(index: number): this {
    this._indexCount++;
    this._recalcFieldCount(1, index);

    return this;
  }

  /**
   * Indicates that a tab/row/column/etc (representing an index) should be
   * removed
   *
   * The last index in the group should be removed.
   *
   * Subtypes may override this for custom functionality.
   *
   * @param index - index that has been removed
   */
  protected removeIndex(index: number): this {
    var fields = this.group.getExclusiveFieldNames();

    for (var field in fields) {
      var cached = this._visCache[fields[field]];

      cached !== undefined && cached.pop();
    }

    this._indexCount--;
    this._recalcFieldCount(-1, index);

    return this;
  }

  /**
   * Allows for processing after preEmptyBucket() has been run without
   * overriding preEmptyBucket() itself
   */
  protected postPreEmptyBucket(): this {
    return this;
  }

  protected postPreEmptyBucketFirst(): this {
    return this;
  }

  /**
   * Invokes add/remove procedures based on what must be done given a number
   * of indexes and the current number of indexes
   *
   * If any indexes need to be added/removed (subject to min/max limits), the
   * appropriate callback will be called with the number of adds/removes to be
   * performed respectively. Otherwise, no callback will be called.
   *
   * @param n   - desired index count
   * @param cur - current index count
   * @param ca  - add continuation
   * @param crm - remove continuation
   */
  protected handleIndexChange(
    n: number,
    cur: number,
    ca: (_: number) => void,
    crm: (_: number) => void
  ): this {
    var min = this.group.minRows(),
      max = this.group.maxRows(),
      diff = n - cur,
      len = cur + diff;

    // restrict to min/max
    if (len < min) {
      len = min;
    } else if (max > 0 && len > max) {
      len = max;
    }

    // now that we have the desired length, calculate that difference (may
    // not differ from diff unless a max/min is hit)
    var truediff = len - cur;

    // we have nothing to do if the diff is 0
    if (truediff === 0) {
      return this;
    }

    // call the add/remove callback depending on our diff
    (truediff < 0 ? crm : ca)(Math.abs(truediff));

    return this;
  }

  /**
   * Initialize an index to be added to the group
   *
   * This action simply raises an event that hooks may properly handle---that
   * is, we're merely indicating our desire to add an index. Whether or not
   * it actually happens [correctly] is beyond our control.
   *
   * If properly handled, presumably the bucket will be updated with the new
   * index and that, in turn, will kick off the hooks to add the necessary UI
   * elements to reflect the addition.
   */
  protected initIndex(): this {
    this.emit(this.EVENT_INDEX_ADD, this.getCurrentIndexCount());

    return this;
  }

  /**
   * Initialize an index to be removed from the group
   *
   * This action simply raises an event that hooks may properly handle---that
   * is, were merely indicating our desire to remove an index. Whether or not
   * it actually happens [correctly] is beyond our control.
   *
   * If properly handled, presumably the bucket will be updated to reflect the
   * deletion and the hooks will then kick off the necessary UI updates.
   */
  protected destroyIndex(index: number): this {
    index = index === undefined ? this.getCurrentIndex() : +index;

    this.emit(this.EVENT_INDEX_REMOVE, index);

    return this;
  }

  /**
   * Call GroupContext to remove an index
   *
   * @param index - index that has been removed
   */
  protected contextRemoveIndex(index: number): void {
    var fields = this.group.getExclusiveFieldNames();
    this.context.removeIndex(
      fields,
      <PositiveInteger>index,
      <PositiveInteger>this.getCurrentIndexCount()
    );
  }

  /**
   * Allows group to perform any necessary operations before an element is
   * scrolled to
   *
   * @param field_name - name of field to display
   * @param i          - index of field
   */
  public preScrollTo(field_name: string, i: number): this {
    // do not do anything if this group does not contain the requested field
    if (!this.group.hasExclusiveField(field_name)) {
      return this;
    }

    // let subtypes display the element in whatever manner makes sense
    this.displayField(field_name, i);

    return this;
  }

  /**
   * Display the requested field
   *
   * The field should not be given focus; it should just be brought to the
   * foreground.
   *
   * This is intended for groups that may conceal fields from the user (e.g.
   * tabbed groups).
   *
   * @param field_name - name of field to display
   * @param i          - index of field
   */
  protected displayField(_field_name: string, _i: number): this {
    // subtypes must override this method if they have the ability to
    // conceal fields from the user (e.g. tabs)
    return this;
  }

  public getId(): undefined | string {
    // return all but the beginning 'group_'
    return this.$content.attr('id').substring(6);
  }

  /**
   * Bind css classes to classifications
   *
   * @param quote - the quote to listen on
   */
  private _bindClasses(quote: ClientQuote): void {
    // Get the css classes that we would like to bind to classifications
    this._bind_classes = this._getBindClasses();

    const self = this;

    quote.onClassifyAndNow(function (classes: any) {
      for (let bind_class in self._bind_classes) {
        let css_class = self._bind_classes[bind_class];

        if (classes[bind_class] && classes[bind_class].is === true) {
          self.content.classList.add(css_class);
        } else {
          self.content.classList.remove(css_class);
        }
      }

      if (self._use_vis_class) {
        classes[self.group.getWhenFieldName()]?.is === true
          ? self.show()
          : self.hide();
      }
      self.onClassify();
    });
  }

  /**
   * Called on quote classification
   *
   * Can be overridden in child classes
   *
   */
  protected onClassify(): void {}

  /**
   * Get the class attributes and their classifiers
   *
   * @return bound classes and their conditional classifications
   */
  private _getBindClasses(): any[] {
    // return all but the beginning group_
    const class_str = this.content.getAttribute('data-class-bind');

    if (!class_str) {
      return [];
    }

    const classes: any[] = [];

    class_str.split(' ').forEach(datum => {
      const kv = datum.split(':');

      if (!kv[0] || !kv[1]) {
        return;
      }

      // Add the classifications as the keys and the css class as value
      classes[<any>kv[1]] = kv[0];
    });

    return classes;
  }

  /**
   * Gets the name of the first question or answer element available
   *
   * This method exists because answer elements cannot have name attributes.
   * Instead they store the reference name in a 'ref_id' data key.
   *
   * @param $parent - parent element to search
   *
   * @return String name, otherwise empty string
   */
  public getFirstElementName(_$parent?: jQuery): string {
    return this.group.getIndexFieldName();
  }

  /**
   * Hooks the event to be triggered when a row is added to a group
   *
   * Children should call this method when they add a new row.
   *
   * @param $element - element that was added
   * @param index    - index of the added element
   */
  protected postAddRow(_$element: any, index: number): this {
    this.emit(this.EVENT_POST_ADD_ROW, index);

    return this;
  }

  public hide(): this {
    this.$content.hide();
    this._visible = false;

    return this;
  }

  public show(): this {
    this.$content.show();
    this._visible = true;

    return this;
  }

  /**
   * Hides the field based on field name and index
   *
   * @param field - field name
   * @param index - field index
   */
  public hideField(field: string, index: number): void {
    if (this.isFieldVisible(field, index) === false) {
      // nothing to do?
      return;
    }

    this._setFieldVisible(field, index, false);

    // can be overridden by subtypes
    this.doHideField(field, index);
  }

  protected doHideField(field: string, index: number) {
    this.context.hide(field, <PositiveInteger>index);
  }

  /**
   * Returns a boolean depending on if there are visible fields
   * based off of the visCount
   *
   * @param index
   */
  public hasVisibleField(index: number): boolean {
    return this._visCount[index] > 0 ? true : false;
  }

  /**
   * Shows the field based on field name and index
   *
   * @param field
   * @param index
   */
  public showField(field: string, index: number): void {
    if (this.isFieldVisible(field, index) === true) {
      // nothing to do
      return;
    }

    this._setFieldVisible(field, index, true);

    // can be overridden by subtypes
    this.doShowField(field, index);
  }

  protected doShowField(field: string, index: number): void {
    this.context.show(
      field,
      <PositiveInteger>index,
      this.fieldContentParent[index]
    );
  }

  public isFieldVisible(id: string, index?: number): undefined | boolean {
    if (<number>index > this.getCurrentIndex()) {
      return false;
    }

    this._visCache[id] = this._visCache[id] || [];

    // if no index was provided, then determine if *any* of the indexes are
    // available
    if (index === undefined) {
      var result: undefined | boolean = false,
        i = this.getCurrentIndexCount();

      while (i--) {
        result = result || this.isFieldVisible(id, i);
      }

      return result;
    }

    // may be undefined
    return this._visCache[id][index];
  }

  private _setFieldVisible(id: string, index: number, visible: boolean) {
    var old = this._visCache[id][index];

    // should only ever be called after isFieldVisible(); if this is false,
    // initialize
    this._visCache[id][index] = !!visible;

    // we assume every field to be visible by default count-wise; as such,
    // if the original cache value is undefined (meaning we aren't sure if
    // it is visible or not), we do not want to increase the visibility
    // count
    this._incVis(index, visible ? (old === undefined ? 0 : 1) : -1);

    this._checkVisCount();
  }

  private _incVis(index: number, by: number) {
    if (this._visCount[index] === undefined) {
      this._visCount[index] = 0;
    }

    this._visCount[index] += by;
    this._visCountTotal += by;
  }

  private _recalcFieldCount(change: number, index: number) {
    var raw = this._rawFieldCount;

    var count = 0;
    if (change === -1) {
      count = this._visCount[index];
    } else {
      count = raw;
    }

    this._incVis(index, change * count);

    this._checkVisCount();
  }

  private _checkVisCount(): void {
    // if we have no fields, then ignore visibility checks
    if (this._rawFieldCount === 0) {
      return;
    }

    // XXX: ideally, < should never occur, but until this implementation is
    // finalized and all the bugs are worked out, it can :x
    if (this._visCountTotal <= 0 && this._visible) {
      this.onAllFieldsHidden();
    } else if (!this._visible && this._visCountTotal > 0) {
      this.onFieldsVisible();
    }
  }

  protected onAllFieldsHidden(): void {
    if (this._use_vis_class) {
      return;
    }

    // default action is to hide self
    this.hide();
  }

  protected onFieldsVisible(): void {
    if (this._use_vis_class) {
      return;
    }

    // default action is to show self
    this.show();
  }

  public invalidate(hook?: any): this {
    if (hook instanceof Function) {
      this.invalidateHooks.push(hook);
      return this;
    }

    // call the hooks
    for (var i = 0, len = this.invalidateHooks.length; i < len; i++) {
      this.invalidateHooks[i]();
    }

    return this;
  }

  /**
   * Returns group object being styled
   */
  public getGroup(): Group {
    return this.group;
  }

  public preRender(): this {
    if (this._emptyOnVisit === null) {
      return this;
    }

    // invoke the continuation
    this._emptyOnVisit();

    this._emptyOnVisit = null;

    return this;
  }

  public visit(): this {
    return this;
  }

  /**
   * Returns whether the given name is a field
   *
   * If not a field, it could be another non-input element, such as a static
   * element.
   *
   * @param name - inquiry
   *
   * @return true if field, otherwise false
   */
  public isAField(name: string) {
    return this.styler.isAField(name);
  }

  public getContentByIndex(_name: string, _index: number) {
    return this.$content;
  }

  /**
   * Sets element value given a name and index
   *
   * This has the performance benefit of searching *only* within the group
   * rather than scanning the entire DOM (or a much larger subset)
   *
   * @param name  - element name
   * @param index - index to set
   * @param value - value to set
   */
  public setValueByName(name: string, index: number, value: string) {
    this.context.setValueByName(name, <PositiveInteger>index, value);
    return this;
  }

  public setActive(active: boolean) {
    active = active === undefined ? true : !!active;

    this._active = active;
    return this;
  }

  public disableField(field: string, disable: boolean) {
    disable = disable === undefined ? true : !!disable;
    // Disable field for every index
    const curlen = this.getCurrentIndexCount();
    for (let index = 0; index < curlen; index++) {
      this.context.disable(field, <PositiveInteger>index, disable);
    }
    return this;
  }

  public setOptions(name: string, index: number, options: any[], val?: string) {
    this.context.setOptions(name, <PositiveInteger>index, options, val);
    return this;
  }

  public clearOptions(name: string, index: number) {
    this.setOptions(name, index, []);
    return this;
  }
}
