/**
 * LiveSearch class
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import * as React from 'react';
import Select, {
  ActionMeta,
  InputActionMeta,
  OptionTypeBase,
  StylesConfig,
  ValueType,
} from 'react-select';

declare type LiveSearchOption = OptionTypeBase;

export interface LiveSearchProps extends Record<string, unknown> {
  input: HTMLInputElement;
  select: HTMLSelectElement;
  document: Document;
  disabled: boolean;
}

declare type LiveSearchState = {
  input_value: ValueType<LiveSearchOption, false>;
  options: LiveSearchOption[];
  selected_option: LiveSearchOption | null;
};

export class LiveSearch extends React.Component<LiveSearchProps, LiveSearchState> {
  private _placeholder: string;

  private _is_disabled: boolean = false;

  private _styles: StylesConfig<LiveSearchOption, false> = {
    singleValue: base => {
      const width = '100%';
      return {...base, width};
    },
    valueContainer: base => {
      const height = '64px';
      return {...base, height};
    },
  };

  constructor(props: LiveSearchProps) {
    super(props);

    const placeholder_attr = this.props.select.getAttribute('data-placeholder');

    this._placeholder = placeholder_attr || 'Begin typing to search...';

    this._is_disabled = this.props.disabled;
  }

  componentDidMount() {
    const observer = new window.MutationObserver(
      (mutationList: Array<MutationRecord>) =>
        mutationList
          .filter((m: MutationRecord) => m.type === 'childList')
          .forEach((m: MutationRecord) => {
            m.addedNodes.forEach(this.updateOptions.bind(this));
          })
    );
    observer.observe(this.props.select, {childList: true, subtree: true});
  }

  /**
   * Update the options
   *
   * Pull available options from the options assigned to the generated select
   */
  updateOptions(): void {
    const options: LiveSearchOption[] = [];
    let selected_option: LiveSearchOption | null = null;

    Array.prototype.forEach.call(
      this.props.select,
      (option: HTMLOptionElement) => {
        if (option.value === undefined || !option.text) {
          return;
        }

        const new_option = {
          value: option.value,
          label: option.text,
        };

        if (option.selected) {
          selected_option = new_option;
        }

        options.push(new_option);
      }
    );

    this.setState({
      ...this.state,
      options: options,
      selected_option: selected_option,
    });
  }

  /**
   * Format the option label to allow html labels to be rendered properly
   *
   * @param data - the option to format a label for
   * @returns A formatted react node
   */
  formatOptionLabel(data: LiveSearchOption): React.ReactNode {
    return <div dangerouslySetInnerHTML={{__html: data.label}} />;
  }

  /**
   * Handle input changes
   *
   * Take the component input and set the value on a generated input which
   * will trigger the dapi call
   *
   * @param value  - the new input
   * @param action - the action that prompted a new input
   */
  onInputChange(value: string, action: InputActionMeta): void {
    if (action.action !== 'input-change') {
      return;
    }

    this.setElementValue(this.props.input, value);
  }

  /**
   * Handle change events
   *
   * @param option - the new value after the change
   * @param action - the action that prompted the change
   */
  onChange(
    option: LiveSearchOption | null,
    action: ActionMeta<OptionTypeBase>
  ): void {
    let affected_option: LiveSearchOption | null = option;

    switch (action.action) {
      case 'select-option':
        this.setElementValue(this.props.select, affected_option?.value);
        break;

      case 'clear':
        affected_option = null;

        this.setElementValue(this.props.input, '');
        break;
    }

    this.setState({
      ...this.state,
      selected_option: affected_option,
    });
  }

  /**
   * Set a value for a given element
   *
   * We need to also dispatch an event to let external systems know that the
   * value has been updated programmatically
   *
   * @param target - the element to set a value for
   * @param value  - the value to set
   */
  setElementValue(
    target: HTMLInputElement | HTMLSelectElement,
    value: string
  ): void {
    target.value = value;

    const event = this.props.document.createEvent('Event');

    event.initEvent('change', true, true);
    target.dispatchEvent(event);
  }

  /**
   * Do not filter any options; this overrides
   * default component filtering
   *
   * @param _option - option to filter
   * @param _input - search query
   */
  filterOption(_option: LiveSearchOption, _input: any): boolean {
    return true;
  }

  /**
   * Render the component
   */
  render(): JSX.Element {
    return (
      <Select
        value={this.state?.selected_option}
        options={this.state?.options || []}
        onInputChange={this.onInputChange.bind(this)}
        onChange={this.onChange.bind(this)}
        formatOptionLabel={this.formatOptionLabel.bind(this)}
        placeholder={this._placeholder}
        isClearable={true}
        isDisabled={this._is_disabled}
        styles={this._styles}
        filterOption={this.filterOption}
      />
    );
  }
}
