/**
 * General UI logic for steps
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @needsLove
 *   - References to "quote" should be replaced with generic terminology
 *     representing a document.
 *   - Dependencies need to be liberated:
 *       - BucketDataValidator.
 *   - Global references (e.g. jQuery) must be removed.
 *   - jQuery must be eliminated.
 *     - The public API now accepts and returns vanilla DOM content, so at
 *       least it's encapsulated now.
 *   - Checkbox-specific logic must be extracted.
 *   - This class is doing too much.
 * @end needsLove
 */

import {Collection} from './Collection';
import {ElementStyler} from '../ElementStyler';
import {GroupUi} from '../group/GroupUi';
import {Step} from '../../step/Step';
import {StepUi} from './StepUi';
import {QuoteDataBucket} from '../../bucket/QuoteDataBucket';
import {EventEmitter} from 'events';

/**
 * Handles display of a step
 *
 * @return {StepUi}
 */
export class GeneralStepUi extends EventEmitter implements StepUi {
  /**
   * Called after step data is processed
   */
  readonly EVENT_POST_PROCESS = 'postProcess';

  /**
   * Called after step is appended to the DOM
   */
  readonly EVENT_POST_APPEND = 'postAppend';

  /**
   * Called when data is changed (question value changed)
   */
  readonly EVENT_DATA_CHANGE = 'dataChange';

  /**
   * Raised when an index is added to a group (e.g. row addition)
   */
  readonly EVENT_INDEX_ADD = 'indexAdd';

  /**
   * Raised when an index is reset in a group (rather than removed)
   */
  readonly EVENT_INDEX_RESET = 'indexReset';

  /**
   * Raised when an index is removed from a group (e.g. row deletion)
   */
  readonly EVENT_INDEX_REMOVE = 'indexRemove';

  /**
   * Represents an action trigger
   */
  readonly EVENT_ACTION = 'action';

  /**
   * Triggered when the step is active
   */
  readonly EVENT_ACTIVE = 'active';

  /**
   * jQuery instance
   */
  private _jquery: any = null;

  /**
   * Step data (DOM representation)
   */
  $content: any = null;

  /**
   * Whether the step should be repopulated with bucket data upon display
   */
  invalid = false;

  /**
   * Stores group objects representing each group
   */
  groups: Record<any, any> = {};

  /**
   * Format bucket data for display
   */
  private _formatter: any = null;

  /**
   * Stores references to which group fields belong to
   */
  private _fieldGroup: any = {};

  /**
   * Hash of answer contexts (jQuery) for quick lookup
   */
  private _answerContext: any = {};

  /**
   * Answer field references for quick lookup
   */
  private _answerFieldRefs: any = {};

  /**
   * Hash of static answer indexes, if applicable
   */
  private _answerStaticIndex: any = {};

  /**
   * Whether the step is the currently active (visible) step
   */
  private _active = false;

  /**
   * Collections of groups
   */
  private _collections: Collection[] = [];

  /**
   * Whether the step is locked (all elements disabled)
   */
  private _locked = false;

  /**
   * A function to call to update bucket data
   */
  private _forceAnswerUpdate: any = null;

  /**
   * Initializes StepUi object
   *
   * The data_get function is used to retrieve the step data, allowing the
   * logic to be abstracted from the Step implementation. It must accept two
   * arguments: the id of the step to load, and a callback function, as the
   * operation is likely to be asynchronous.
   *
   * A callback function is used for when the step is ready to be used. This
   * is done because the loading of the data is (ideally_ an asynchronous
   * operation.  This operation is performed in the constructor, to ensure
   * that each instance of a Step class has data associated with it.
   * Therefore, the object will be instantiated, but the data_get function
   * will still be running in the background. The step should not be used
   * until the data loading is complete.  That is when the callback will be
   * triggered.
   *
   * TODO: Remove jQuery
   *
   * @param step          - step to render
   * @param styler        - element styler
   * @param formatter     - validator/formatter
   * @param document      - Document object
   * @param jquery        - jQuery instance
   */
  constructor(
    private step: Step,
    private readonly _styler: ElementStyler,
    formatter: any,
    protected readonly document: Document,
    jquery: any
  ) {
    super();
    this._formatter = formatter;

    // TODO: transition code; remove default
    this._jquery = jquery || (typeof $ !== 'undefined' ? $ : undefined);
  }

  /**
   * Initializes step
   */
  init(): this {
    this.step.on('updateQuote', (bucket: QuoteDataBucket) => {
      this._hookBucket(bucket);
      this._processAnswerFields();
      this.invalidate();
    });

    return this;
  }

  /**
   * Initialize group field data
   */
  initGroupFieldData(): void {
    for (const group in this.groups) {
      const groupui = this.groups[group],
        fields = groupui.group.getExclusiveFieldNames();

      for (const i in fields) {
        this._fieldGroup[fields[i]] = groupui;
      }
    }
  }

  /**
   * Sets content to be displayed
   *
   * @param content - content to display
   */
  setContent(content: HTMLElement): this {
    // TODO: transition away from jQuery
    this.$content = this._jquery(content);

    this._processAnswerFields();

    return this;
  }

  /**
   * Returns the step that this object is styling
   *
   * @return lovullo.program.Step
   */
  getStep(): Step {
    return this.step;
  }

  /**
   * Returns the generated step content as a jQuery object
   *
   * @return generated step content
   */
  getContent(): HTMLElement {
    return this.$content[0];
  }

  /**
   * Set collections on this step
   *
   * @param collections - collections to set
   */
  set collections(collections: Collection[]) {
    this._collections = collections;
  }

  /**
   * Monitor the dom for changes and stage updates
   */
  monitorDom(): void {
    this.$content.bind('change.program', (event: any) => {
      this._stageData(event);
    });

    this.document.addEventListener('keyup', (event: any) => {
      if (
        event.target.hasAttribute('data-stage-event') &&
        event.target.dataset.stageEvent === 'keyup'
      ) {
        this._stageData(event);
      }
    });
  }

  /**
   * Stage data to staging bucket and emit data change event
   *
   * @param event - that triggered data change
   */
  private _stageData(event: any): void {
    // do nothing if the step is locked
    if (this._locked) {
      return;
    }

    // get the name of the altered element
    const $element = this._styler.getNameElement(this._jquery(event.target));

    let val = $element.val(),
      name = $element.attr('name');

    if (!name) {
      // rogue field not handled by the framework!
      return;
    }

    // remove the trailing square brackets from the name
    name = name.substring(0, name.length - 2);

    const element = $element[0];
    let index = element.getAttribute('data-index');

    // For single index groups, index should be 0
    if (index === undefined || index === null) {
      index = 0;
    }

    if (
      element.getAttribute('type') === 'radio' ||
      element.getAttribute('type') === 'checkbox'
    ) {
      // 2 in this instance is the yes/no group length.
      const group_length = element.getAttribute('data-question-length')
        ? +element.getAttribute('data-question-length')
        : 2;

      const is_checked = element.checked;

      // if it's not checked and in a group, then this isn't
      // the radio we're interested in. Sorry!
      if (!is_checked && group_length !== 1) {
        element.checked = true;

        return;
      }

      // if this is a lonely checkbox, we only want the value if it is checked
      if (group_length === 1 && !is_checked) {
        val = 0;
      }
    }

    const values: any = {};
    values[name] = [];
    values[name][index] = val;

    // update our bucket with this new data
    this.emit(this.EVENT_DATA_CHANGE, values);
  }

  /**
   * Prepares answer fields
   *
   * This method will populate the answer fields with values already in the
   * bucket and hook the bucket so that future updates will also be reflected.
   */
  _processAnswerFields(): void {
    this._prepareAnswerContexts();

    // perform initial update for the step when we are first created, then
    // hook everything else (we do not need the hooks before then, as we
    // will be forcefully updating the step with values)
    this.once('postAppend', (bucket: QuoteDataBucket) => {
      let forceupdate = false;

      // when the value we're watching is updated in the bucket, update
      // the displayed value
      let doUpdate: any;
      bucket.on(
        'stagingUpdate',
        (doUpdate = (data: any) => {
          // defer updates unless we're active
          if (!this._active) {
            if (forceupdate === false) {
              forceupdate = true;

              this.once(this.EVENT_ACTIVE, () => {
                doUpdate(bucket.getData());
                forceupdate = false;
              });
            }

            return;
          }

          this.answerDataUpdate(data);
        })
      );

      doUpdate(bucket.getData());

      // set the values when a row is added
      this.on('postAddRow', (index: number) => {
        const data = bucket.getData();

        for (const name in this._answerContext) {
          const value = (data[name] || {})[index];

          if (value === undefined) {
            continue;
          }

          this._setValueByName(name, index, <string>value);
        }
      });

      this._forceAnswerUpdate = doUpdate;
    });
  }

  /**
   * Update and style answer field data
   *
   * @param data - key-value diff
   */
  answerDataUpdate(data: any): void {
    const data_fmt = this._formatter.format(data);

    // give the UI a chance to update the DOM; otherwise, the
    // answer elements we update may no longer be used (this also
    // has performance benefits since it allows repainting before
    // potentially heavy processing)
    setTimeout(() => {
      this._updateAnswerFieldData(data_fmt);
    }, 25);
  }

  /**
   * Retrieve answer DOM context
   *
   * @param name - field name
   *
   * @return DOM context
   */
  getAnswerContext(name: string): any {
    return this._answerContext[name];
  }

  /**
   * Retrieve field references for answers
   *
   * @param name - field name
   *
   * @return data-field-name references from DataSet
   */
  getAnswerFieldRefs(name: string): string[] {
    return this._answerFieldRefs[name] || [];
  }

  /**
   * Set the answer value
   *
   * @param name  - field name
   * @param index - field index
   * @param value - value to set
   */
  private _setValueByName(name: string, index: number, value: string): void {
    const field_names = this.getAnswerFieldRefs(name);

    // set the value for each answer/display referenced for this field
    field_names.forEach((field_name: string) => {
      // fall back to element styler when group is unknown
      const group = this.getElementGroup(field_name);
      group
        ? group.setValueByName(field_name, index, value)
        : this._updateAnswer(name, index, value);
    });
  }

  /**
   * Update DOM answer fields with respective datum in diff DATA
   *
   * Only watched answer fields are updated.  The update is performed on
   * the discovered context during step initialization.
   *
   * @param data - bucket diff
   */
  private _updateAnswerFieldData(data: any): void {
    // return if step is not active
    if (!this._active) {
      return;
    }

    // we only care if the data we're watching has been
    // changed
    for (const name in data) {
      if (!this.getAnswerContext(name)) {
        continue;
      }

      const curdata = data[name],
        si = this._answerStaticIndex[name];

      let i = curdata.length;

      // static index override
      if (!isNaN(si)) {
        // update every index on the DOM
        i = this._styler.getAnswerElementByName(
          name,
          undefined,
          undefined,
          this.getAnswerContext(name)
        ).length;
      }

      while (i--) {
        const index = isNaN(si) ? i : si,
          value = curdata[index];

        // take into account diff; note that if one of
        // them is null, that means it has been removed
        // (and will therefore not be displayed), so we
        // don't have to worry about clearing out a value
        if (value === undefined || value === null) {
          continue;
        }

        this._setValueByName(name, i, curdata[index]);
      }
    }
  }

  /**
   * Prepare answer contexts
   */
  private _prepareAnswerContexts(): void {
    // get a list of all the answer elements
    this.$content.find('span.answer').each((_: any, elem: any) => {
      const $this = <any>$(elem),
        ref_id = <string>$this.attr('data-answer-ref'),
        index = $this.attr('data-answer-static-index');

      // clear the value (which by default contains the name of the answer
      // field)
      $this.text('');

      // if we've already found an element for this ref, then it is
      // referenced in multiple places; simply store the context as the
      // entire step
      if (this.getAnswerContext(ref_id)) {
        this._answerContext[ref_id] = this.$content;
        // Typescript can't verify the type from jquery so we want to keep this
        // eslint-disable-next-line no-extra-boolean-cast
        if (!!$this.context.dataset.fieldName) {
          this._answerFieldRefs[ref_id].push($this.context.dataset.fieldName);
        }
        return;
      }

      // store the parent fieldset as our context to make DOM lookups a
      // bit more performant
      const fieldset = $this.parents('fieldset');
      this._answerContext[ref_id] = fieldset;

      // Typescript can't verify the type from jquery so we want to keep this
      // eslint-disable-next-line no-extra-boolean-cast
      if (!!fieldset.context.dataset.fieldName) {
        this._answerFieldRefs[ref_id] = [fieldset.context.dataset.fieldName];
      }

      this._answerStaticIndex[ref_id] = index ? +index : NaN;
    });
  }

  /**
   * Update the display of an answer field
   *
   * The value will be styled before display.
   *
   * @param name  - field name
   * @param index - index to update
   * @param value - answer value (unstyled)
   */
  private _updateAnswer(name: string, index: number, value: string): void {
    const $element = this._styler.getAnswerElementByName(
      name,
      index,
      null,
      this.getAnswerContext(name) || this.$content
    );

    let i = $element.length;
    if (i > 0) {
      while (i--) {
        const styled = this._styler.styleAnswer(name, value);

        $element.text(styled);

        const id = $element[i].attributes['data-field-name'];
        if (!id) {
          continue;
        }

        this.emit('displayChanged', id.value, index, value);
      }
    }
  }

  /**
   * Monitors the bucket for data changes and updates the elements accordingly
   *
   * @param bucket - a quote data bucket
   */
  _hookBucket(bucket: QuoteDataBucket): void {
    // when the bucket data is updated, update the element to reflect the
    // value
    bucket.on('stagingUpdate', (data: any) => {
      const data_fmt = this._formatter.format(data);

      for (const name in this.step.getExclusiveFieldNames()) {
        // if this data hasn't changed, then ignore the element
        if (data_fmt[name] === undefined) {
          continue;
        }

        // update each of the elements (it is important to update the
        // number of elements on the screen, not the number of elements
        // in the data array, since the array is a diff and will contain
        // information regarding removed elements)
        const data_len = data_fmt[name].length;

        for (let index = 0; index < data_len; index++) {
          const val = data_fmt[name][index];

          // if the value is not set or has been removed (remember,
          // we're dealing with a diff), then ignore it
          if (val === undefined || val === null) {
            continue;
          }

          // set the value of the element using the appropriate group
          // (for performance reasons, so we don't scan the whole DOM
          // for the element)
          this.getElementGroup(name)?.setValueByName(name, index, val);
        }
      }
    });
  }

  /**
   * Called after the step is appended to the DOM
   *
   * This method will simply loop through all the groups that are a part of
   * this step and call their postAppend() methods. If the group does not have
   * an element id, it will not function properly.
   */
  postAppend(bucket: QuoteDataBucket): this {
    // let the styler do any final styling
    this._styler.postAppend(this.$content.parent());

    // If we have data in the bucket (probably loaded from the server), show
    // it. We use a delay to ensure that the UI is ready for the update. In
    // certain cases (such as with tabs), the UI may not have rendered all
    // the elements.
    this.emptyBucket(null, true);

    // monitor bucket changes and update the elements accordingly
    this._hookBucket(bucket);

    this.emit(this.EVENT_POST_APPEND, bucket);

    return this;
  }

  /**
   * Empties the bucket into the step (filling the fields with its values)
   *
   * @param callback - function to call when bucket has been emptied
   * @param delay    - whether to execute immediately or set a timer
   */
  emptyBucket(callback?: any, delay?: boolean): this {
    delay = delay === undefined ? false : true;

    const bucket = this.getStep().getBucket();

    // first, clear all the elements
    for (const group in this.groups) {
      this.groups[group].preEmptyBucket(bucket);
    }

    // then update all the elements with the form values in the bucket
    // (using setTimeout allows the browser UI thread to process repaints,
    // added elements, etc, which will ensure that the elements will be
    // available to empty into)
    const empty = () => {
      const data: Record<string, any> = {};

      for (const name in this.step.getExclusiveFieldNames()) {
        data[name] = bucket.getDataByName(name);
      }

      // format the data (in-place, since we're the only ones using this
      // object)
      this._formatter.format(data, true);

      for (const name in data) {
        const values = data[name];
        let i = values.length;

        while (i--) {
          let value = values[i];

          // set the data and do /not/ trigger the change event
          const group = this.getElementGroup(name);
          if (!group) {
            // This should not happen (see FS#13653); emit an error
            // and continue processing in the hopes that we can
            // display most of the data
            this.emit(
              'error',
              Error('Unable to locate group for field `' + name + "'")
            );

            continue;
          }

          // Note the broader context here: we are emptying the
          // bucket.  This means that any undefined value has surely come
          // from the formatter, which means that the provided value in the
          // bucket is now invalid.  Since the DOM may be populated with a
          // default value (which does not actually reflect the underlying
          // bucket state), we should forcefully clear this field so that
          // the user can re-enter a valid value.
          if (value === undefined) {
            value = '';
          }

          this.getElementGroup(name)?.setValueByName(name, i, value);
        }
      }

      // answers are normally only updated on bucket change
      this._forceAnswerUpdate(bucket.getData());

      if (callback instanceof Function) {
        callback.call(this);
      }
    };

    // either execute immediately or set a timer (allowing the UI to update)
    // if a delay was requested
    if (delay) {
      setTimeout(empty, 25);
    } else {
      empty();
    }

    return this;
  }

  /**
   * Resets a step to its previous state or hooks the event
   *
   * @param callback - function to call when reset is complete
   */
  reset(callback: any): this {
    this.getStep().getBucket().revert();

    if (typeof callback === 'function') {
      callback.call(this);
    }

    // clear invalidation flag
    this.invalid = false;

    return this;
  }

  /**
   * Returns whether all the elements in the step contain valid data
   *
   * @param cmatch - cmatch data
   *
   * @return true if all elements are valid, otherwise false
   */
  isValid(cmatch: any): boolean {
    return this.step.isValid(cmatch);
  }

  /**
   * Returns the id of the first failed field if isValid() failed
   *
   * Note that the returned element may not be visible. Visible elements will
   * take precidence --- that is, invisible elements will be returned only if
   * there are no more invalid visible elements, except in the case of
   * required fields.
   *
   * @param cmatch - cmatch data
   *
   * @return id of element, or empty string
   */
  getFirstInvalidField(cmatch: any): Array<string | number | boolean> | null {
    const element = this.getContent().querySelectorAll('.invalid_field')[0];

    // no invalid fields, so what about missing required fields?
    if (!element) {
      // append 'true' indiciating that this is a required field check
      const result: Array<string | boolean> | null =
        this.step.getRequired(cmatch);
      if (result !== null) {
        result.push(true);
      }

      return result;
    }

    // grab index from data set
    const index = element.getAttribute('data-index');
    const name = element.getAttribute('data-field-name');
    if (typeof index === 'string' && typeof name === 'string') {
      // return the element name and index
      return [
        name,
        index,
        // not a required field failure
        false,
      ];
    }
    return null;
  }

  /**
   * Scrolls to the element identified by the given id
   *
   * @param field        - name of field to scroll to
   * @param i            - index of field to scroll to
   * @param show_message - whether to show the tooltip
   * @param message      - tooltip message to display
   */
  scrollTo(
    field: string,
    i: number,
    show_message: boolean,
    message: string
  ): this {
    show_message = show_message === undefined ? true : !!show_message;

    if (!field || i < 0 || i === undefined) {
      // cause may be empty
      const cause = this.step.getValidCause();

      this.emit(
        'error',
        Error(
          'Could not scroll: no field/index provided' +
            (cause ? ' (cause: ' + cause + ')' : '')
        )
      );

      return this;
    }

    const $element = this._styler.getWidgetByName(field, i);

    // if the element couldn't be found, then this is useless
    if ($element.length === 0) {
      this.emit(
        'error',
        Error('Could not scroll: could not locate ' + field + '[' + i + ']')
      );

      return this;
    }

    // allow the groups to preprocess the scrolling
    for (const group in this.groups) {
      this.groups[group].preScrollTo(field, i);
    }

    // is the element visible now that we've given the groups a chance to
    // display it?
    if ($element.is(':visible') !== true) {
      // fail; don't bother scrolling
      this.emit(
        'error',
        Error(
          'Could not scroll: element ' +
            field +
            '[' +
            i +
            '] ' +
            'is not visible'
        )
      );

      return this;
    }

    // scroll to just above the first invalid question so that it
    // may be fixed
    this.$content.parent().scrollTo($element, 100, {
      offset: {top: -150},
      onAfter: () => {
        // focus on the element and display the tooltip
        this._styler.focus($element, show_message, message);
      },
    });

    return this;
  }

  /**
   * Invalidates the step, stating that it should be reset next time it is
   * displayed
   *
   * Resetting the step will clear the invalidation flag.
   */
  invalidate(): void {
    this.invalid = true;
  }

  /**
   * Returns whether the step has been invalidated
   *
   * @return true if step has been invalidated, otherwise false
   */
  isInvalid(): boolean {
    return this.invalid;
  }

  /**
   * Returns the GroupUi object associated with the given element name, if
   * known
   *
   * @param name - element name
   *
   * @return group if known, otherwise null
   */
  getElementGroup(name: string): GroupUi | null {
    return this._fieldGroup[name] || null;
  }

  /**
   * Forwards add/remove hiding requests to groups
   *
   * @param value - whether to hide (default: true)
   */
  hideAddRemove(value: boolean): this {
    value = value !== undefined ? !!value : true;

    for (const group in this.groups) {
      const groupui = this.groups[group];
      if (groupui.hideAddRemove instanceof Function) {
        groupui.hideAddRemove(value);
      }
    }

    return this;
  }

  /**
   * Call prerender for all groups
   */
  preRender(): this {
    for (const group in this.groups) {
      this.groups[group].preRender();
    }

    return this;
  }

  /**
   * Visit all groups
   *
   * @param callback - A function to call on completion
   */
  visit(callback: any): this {
    // "invalid" means that the displayed data is not up-to-date
    if (this.invalid) {
      this.emptyBucket();
      this.invalid = false;
    }

    for (const group in this.groups) {
      this.groups[group].visit();
    }

    let cn = 0;

    // we perform async. processing, so ideally the caller should know
    // when we're actually complete
    const c = function () {
      if (--cn === 0) {
        callback();
      }
    };

    this.step.eachSortedGroupSet((ids: string[]) => {
      cn++;
      this._sortGroups(ids, c);
    });

    if (cn === 0) {
      callback && callback();
    }

    this._collections.forEach(collection => collection.visit());

    return this;
  }

  /**
   * Sort groups
   *
   * @param ids      - ids to sort
   * @param callback - a function to call on completion
   */
  private _sortGroups(ids: string[], callback: any): void {
    // detach them all (TODO: a more efficient method could be to detach
    // only the ones that aren not already in order, or ignore ones that
    // would be hidden..etc)
    const len = ids.length;

    // if one group (or none), nothing to sort
    if (len <= 1) {
      return;
    }

    function getGroup(name: string) {
      return document.getElementById('group_' + name);
    }

    const nodes: Record<string, HTMLElement | null> = {};
    for (const i in ids) {
      nodes[i] = getGroup(ids[i]);
    }

    let prev = nodes[0];
    if (!(prev && prev.parentNode)) {
      return;
    }

    // TODO: this makes the assumption that there is a parent node; we
    // should not be concerned with that, and should find some other way
    // of hiding the entire step while sorting (which the Ui handles)
    const step_node = this.$content[0].parentNode,
      container = step_node.parentNode,
      sibling = step_node.nextSibling,
      parent = prev.parentNode;

    let i = len - 1;

    if (!container) {
      return;
    }

    // to prevent DOM updates for each and every group move, detach the node
    // that contains all the groups from the DOM; we'll re-add it after
    // we're done
    container.removeChild(step_node);

    // we can sort the groups in place without screwing up the DOM by simply
    // starting with the last node and progressively inserting nodes
    // before that element; we start at the end simply because there is
    // Node#insertBefore, but no Node#insertAfter
    setTimeout(() => {
      let group;

      try {
        do {
          group = nodes[i];

          if (!group || !prev) {
            continue;
          }

          // remove from DOM and reposition, unless we are already in
          // position
          if (prev.previousSibling !== group) {
            parent.removeChild(group);
            parent.insertBefore(group, prev);
          }

          prev = group;
        } while (i--);
      } catch (e) {
        // we need to make sure we re-attach the container, so don't blow up
        // if sorting fails
        console.error && console.error(e, group, prev);
      }

      // now that sorting is complete, re-add the groups in one large DOM
      // update, maintaining element order
      if (sibling) {
        container.insertBefore(step_node, sibling);
      } else {
        container.appendChild(step_node);
      }

      callback();
    }, 25);
  }

  /**
   * Marks a step as active (or inactive)
   *
   * A step should be marked as active when it is the step that is currently
   * accessible to the user.
   *
   * @param active - whether step is active
   */
  setActive(active: boolean): this {
    active = active === undefined ? true : !!active;

    this._active = active;

    // notify each individual group of whether or not they are now active
    for (const id in this.groups) {
      this.groups[id].setActive(active);
    }

    if (active) {
      this.emit(this.EVENT_ACTIVE);
    }

    return this;
  }

  /**
   * Lock/unlock a step (preventing modifications)
   *
   * If the lock status has changed, the elements on the step will be
   * disabled/enabled respectively.
   *
   * @param lock - whether step should be locked
   */
  lock(lock: boolean): this {
    lock = lock === undefined ? true : !!lock;

    // if the lock has changed, then alter the elements
    if (lock !== this._locked) {
      for (const name in this.step.getExclusiveFieldNames()) {
        this.getElementGroup(name)?.disableField(name, lock);
      }

      this._collections.forEach(collection => collection.lock(lock));
    }

    this._locked = lock;
    return this;
  }
}
