/**
 * Error condition field styler
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {DomField} from '../field/DomField';
import {FieldStyler} from './FieldStyler';

/**
 * Style field to indicate an error and displays an error message
 */
export class ErrorFieldStyler extends FieldStyler {
  public getId() {
    return 'error';
  }

  /**
   * Determines whether the field has been styled
   *
   * Having this predicate on the styler rather than the field ensures
   * that, even if the two somehow get out of sync (or styles are applied
   * elsewhere), application/revocation will function sanely.
   *
   * @param field   - field to style
   * @param element - DOM element to style
   *
   * @return {boolean} whether FIELD has been styled by this styler
   */
  public isApplied(
    _field: DomField,
    element: HTMLElement,
    row: (HTMLElement | null)[],
    msg?: string
  ) {
    if (msg && this._getMessage(row[0]) !== msg) {
      return false;
    }

    return /\binvalid_field\b/.test(element.className);
  }

  public applyStyle(
    field: DomField,
    element: HTMLElement,
    row: (HTMLElement | null)[],
    msg: string
  ): this {
    // style the row containing the element
    for (const i in row) {
      this.addClass(row[i], 'invalid');
    }

    // TODO: legacy; remove
    this.addClass(element, 'invalid_field');

    // display the error message
    this._createOrUpdateMessage(field.getName(), msg, row[0], row[1]);

    return this;
  }

  public revokeStyle(
    _field: DomField,
    element: HTMLElement,
    row: (HTMLElement | null)[]
  ): this {
    // un-style the row containing the element
    // style the row containing the element
    for (const i in row) {
      this.removeClass(row[i], 'invalid');
    }

    // TODO: legacy; remove
    this.removeClass(element, 'invalid_field');

    this._destroyMessage(row[0], row[1]);

    return this;
  }

  /**
   * Get the error message if present
   *
   * @param dd - the node to search on
   *
   * @returns the error message or null if it wasn't found
   */
  private _getMessage(dd: HTMLElement | null): string | null {
    if (!dd) {
      return null;
    }

    const node = this._getErrorNode(dd);

    return node && node.innerHTML;
  }

  private _createOrUpdateMessage(
    _name: string,
    message: string,
    dd: HTMLElement | null,
    dt: HTMLElement | null
  ) {
    // we can only generate the message if the parent row is available
    if (!(dd && dt)) {
      return;
    }

    const existing = this._getErrorNode(dd);
    if (existing) {
      existing.innerHTML = message;

      return;
    }

    const msg = document.createElement('div');
    msg.className = 'errmsg';
    msg.innerHTML = message;

    // append to dd
    dd.appendChild(msg);

    const height = msg.offsetTop + msg.offsetHeight;

    // element does not have height until added to DOM
    // set a default to ensure it appears to user
    const heightPixels = height === 0 ? 45 + 'px' : height + 10 + 'px';

    dd.style.height = heightPixels;
    dt.style.height = heightPixels;
  }

  private _destroyMessage(dd: HTMLElement | null, dt: HTMLElement | null) {
    if (!dd) {
      return;
    }

    dd.style.height = '';

    // note that dt may not actually exist (in fact, dd may not even be a
    // dd; we should rename these variables)
    dt && (dt.style.height = '');

    const node = this._getErrorNode(dd);

    // if we found it, then remove it
    node && dd.removeChild(node);
  }

  /**
   * Get the error node from the dd element
   *
   * @param dd - the element to search on
   *
   * @returns the error node if found, otherwise null
   */
  private _getErrorNode(dd: HTMLElement): HTMLElement | null {
    let node;

    // search for the message node, starting with the last element (since
    // the error message was appended, we're likely to find it on our first
    // try)
    for (
      node = dd.lastChild;
      node && node instanceof HTMLElement && node.className !== 'errmsg';
      node = node.previousSibling
    );

    return node instanceof HTMLElement ? node : null;
  }
}
