/**
 * Classification utility functions
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {ClassificationResult} from '../program/Program';

/**
 * Evaluate an aggregated classification and return true keys
 *
 * The first check is for performance reasons, it is assumed that the class
 * without a suffix will be true if any suffixed classes are true. This way we
 * can do an inexpensive check before needing to iterate on all of the classes
 *
 * @param classes - the classes to check
 * @param name    - the name of the aggregate class
 * @param cb      - a callback to invoke with true class values
 */
export function forEachAggregatedClass(
  classes: ClassificationResult,
  name: string,
  cb: (class_name: string) => void
): void {
  classes[name]?.is &&
    Object.keys(classes)
      .filter(key => key.indexOf(`${name}-`) !== -1 && classes[key].is)
      .forEach(key => cb(key));
}

/**
 * Evaluate an aggregated classification and return the suffix of true keys
 *
 * The first check is for performance reasons, it is assumed that the class
 * without a suffix will be true if any suffixed classes are true. This way we
 * can do an inexpensive check before needing to iterate on all of the classes
 *
 * @param classes - the classes to check
 * @param name    - the name of the aggregate class
 * @param cb      - a callback to invoke with true class values
 */
export function forEachAggregatedClassSuffix(
  classes: ClassificationResult,
  name: string,
  cb: (class_name: string) => void
): void {
  forEachAggregatedClass(classes, name, (class_name: string) =>
    cb(class_name.replace(`${name}-`, ''))
  );
}
