/**
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

export type AlphaNumeric = string;
type NonEmptyString = string;
type BucketKey = NonEmptyString;
type BucketValue = string[];
export type Bucket = Record<BucketKey, BucketValue>;

export const isString = (x: unknown): x is string => typeof x === 'string';

/**
 * Check if an unknown is an alphanumeric string
 */
export const isAlphaNumeric = (x: unknown): x is AlphaNumeric =>
  isString(x) && /^[A-Za-z0-9]*$/.test(x);

export const isNonEmptyString = (x: unknown): x is NonEmptyString =>
  isString(x) && x.length > 0;

export const isBucketKey = (key: unknown): key is BucketKey =>
  isNonEmptyString(key) && /^[A-Za-z0-9_-]+$/.test(key);

export const isNonEmptyArray = (arr: unknown): arr is unknown[] =>
  Array.isArray(arr) && arr.length > 0;

export const isBucketValue = (val: unknown): val is BucketValue =>
  isNonEmptyArray(val) && val.every(isString);

export const isBucket = (obj: unknown): obj is Bucket =>
  typeof obj === 'object' &&
  obj !== null &&
  Object.keys(obj).every(isBucketKey) &&
  Object.values(obj).every(isBucketValue);
