/**
 * Validation failure
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Field, instanceofField} from '../field/Field';

/**
 * Represents a field validation failure
 */
export class Failure {
  /**
   * Fields that caused the error
   */
  private _causes: Field[];

  /**
   * Create failure with optional reason and causes
   *
   * The field FIELD is the target of the failure, which might have
   * been caused by another field CAUSES.  If cause is omitted, it is
   * assumed to be FIELD.  The string REASON describes the failure.
   *
   * @param field         - failed field
   * @param reason        - description of failure
   * @param causes        - field that triggered the failure
   * @param renderMessage - message to show when rendering
   */
  constructor(
    private readonly _field: Field,
    private readonly _reason: string = '',
    causes?: Field[],
    private readonly _renderMessage?: string
  ) {
    if (!instanceofField(_field)) {
      throw TypeError('Field expected');
    }

    if (causes !== undefined) {
      this._checkCauses(causes);
    }

    this._causes = causes || [this._field];
  }

  /**
   * Validate cause data types
   *
   * Ensures that CAUSES is an array of Field objects; otherwise, throws
   * a TypeError.
   *
   * @param causes - failure causes
   */
  private _checkCauses(causes: Field[]) {
    if (Object.prototype.toString.call(causes) !== '[object Array]') {
      throw TypeError('Array of causes expected');
    }

    for (const i in causes) {
      if (!instanceofField(causes[i])) {
        throw TypeError('Field expected for causes');
      }
    }
  }

  /**
   * Retrieve target of the failure
   *
   * @return target field
   */
  public getField(): Field {
    return this._field;
  }

  /**
   * Retrieve a description of the failure, or the empty string
   *
   * @return {string} failure description
   */
  public getReason(): string {
    return this._reason;
  }

  /**
   * Retrieve field that caused the failure
   *
   * Unless a separate causes was provided during instantiation, the
   * failure is assumed to have been caused by the target field itself.
   *
   * @return causes of failure
   */
  public getCauses(): Field[] {
    return this._causes;
  }

  /**
   * Return a message to show when rendering the failure
   *
   * @return {?string} message to show when rendering the failure
   */
  public getRenderMessage(): string | undefined {
    return this._renderMessage;
  }

  /**
   * Produce failure reason when converted to a string
   *
   * This allows the failure to be used in place of the traditional system
   * of error strings.
   */
  public toString() {
    return this._reason;
  }

  /**
   * Produce a new Failure merging two failures' messages and causes
   */
  public merge(failure: Failure): Failure {
    if (!(failure instanceof Failure)) {
      throw TypeError('Cannot merge with non-Failure ' + failure);
    }

    const all_reasons = this._reason + '; ' + failure.getReason(),
      all_causes = this._causes.concat(failure.getCauses());

    // Prefer the newer failure message
    const renderMessage = failure.getRenderMessage() ?? this._renderMessage;

    return new Failure(this._field, all_reasons, all_causes, renderMessage);
  }
}
