/**
 * VIN (Vehicle Identification Number) formatter
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {DisplayableError} from '../../error/DisplayableError';
import {ValidatorFormatter} from '../ValidatorFormatter';

/**
 * Formats and validates a VIN
 */
export class VinFormatter implements ValidatorFormatter {
  /**
   * Parse a VIN
   *
   * @param data - data to parse
   *
   * @return formatted VIN
   */
  parse(data: string): string {
    const capitalized = data.toUpperCase();
    const cleaned = capitalized.replace(/\s/g, '');

    if (cleaned.match(/[^A-Z0-9]/i)) {
      throw new DisplayableError(
        'VIN should only contain letters and numbers. Verify that the VIN entered is correct.'
      );
    }

    if (cleaned.match(/[IOQ]/)) {
      throw new DisplayableError(
        'VIN should not contain I, O, or Q. Verify that the VIN entered is correct.'
      );
    }

    if (cleaned.length !== 17) {
      throw new DisplayableError(
        'VIN must be exactly 17 characters long. Verify that the VIN entered is correct.'
      );
    }

    if (cleaned.substring(14).replace(/[^0-9]/g, '').length !== 3) {
      throw new DisplayableError(
        'The last 3 characters must be numbers. Verify that the VIN entered is correct.'
      );
    }

    if (!this.checkDigitValid(cleaned)) {
      throw new DisplayableError(
        'Invalid VIN. Verify that the VIN entered is correct.'
      );
    }

    return cleaned;
  }

  /**
   * Format a VIN
   *
   * @param data - data to format
   *
   * @return formatted VIN
   */
  retrieve(data: string): string {
    return data;
  }

  /**
   * Validate a VIN's check digit
   *
   * @param vin - vin to validate
   *
   * @return true if valid, false otherwise
   */
  private checkDigitValid(vin: string): boolean {
    // A map used to handle the representation of X as 10
    const map = '0123456789X';

    // A weight assigned to each position
    const weights = '8765432X098765432';

    let sum = 0;
    for (let i = 0; i < 17; i++) {
      sum += this.transliterate(vin[i]) * map.indexOf(weights[i]);
    }

    return vin[8] === map[sum % 11];
  }

  /**
   * Map a VIN character to a number
   *
   * This is based on Wikipedia's table of transliterations
   *
   * https://en.wikipedia.org/wiki/Vehicle_identification_number#Transliterating_the_numbers
   *
   * @param char - a character to transliterate to a number
   *
   * @returns the transliterated number
   */
  private transliterate(char: string): number {
    let mapString = '';

    // The string is constructed this way to illustrate it as a table where
    // we get a numeric value from the column position of the character.
    //
    // Dots represent N/A values
    mapString += '0123456789';
    mapString += '.ABCDEFGH.';
    mapString += '.JKLMN.P.R';
    mapString += '..STUVWXYZ';

    return mapString.indexOf(char) % 10;
  }
}
