/**
 * Tests BucketFilter
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {expect} from 'chai';
import {filter as Sut} from '../../src/bucket/bucket_filter';

describe('bucket_filter', () => {
  describe('filter', () => {
    it('keeps datum that passes the regex test', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {};
      const result = Sut(data, key_types, ignore_types);
      const expected = {foo: ['bar', 'baz']};

      expect(result).to.deep.equal(expected);
    });

    it('corrects invalid datum', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {};
      data.foo = ['foo123bar&&'];

      const result = Sut(data, key_types, ignore_types);
      const expected = {foo: ['foobar']};

      expect(result).to.deep.equal(expected);
    });

    // Is this expected behavior?
    it('serializes null values by default', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {};
      data.foo = [null];

      const result = Sut(data, key_types, ignore_types);
      const expected = {foo: ['null']};

      expect(result).to.deep.equal(expected);
    });

    it('allows null values when permitted', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {};
      const permit_null = true;
      data.foo = [null];

      const result = Sut(data, key_types, ignore_types, permit_null);
      const expected = {foo: [null]};

      expect(result).to.deep.equal(expected);
    });

    it('deletes data that has an undefined type', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {};
      key_types.foo = undefined;

      const result = Sut(data, key_types, ignore_types);
      const expected = {};

      expect(result).to.deep.equal(expected);
    });

    it('deletes data for an ignored type', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {foo: true};

      const result = Sut(data, key_types, ignore_types);
      const expected = {};

      expect(result).to.deep.equal(expected);
    });

    it('replaces null data with an empty array', () => {
      const data = getData();
      const key_types = getKeyTypes();
      const ignore_types = {};
      data.foo = null;

      const result = Sut(data, key_types, ignore_types);
      const expected = {foo: []};

      expect(result).to.deep.equal(expected);
    });

    it('does not filter data on default type', () => {
      const key_types = {foo: {type: 'does not exist', dim: 1}};
      const value = 'Nothing is stripped, not even //';

      const result = Sut({foo: [value]}, key_types, {});
      const expected = {foo: [value]};

      expect(result).to.deep.equal(expected);
    });
  });
});

const getData = (): Record<string, any> => {
  return {foo: ['bar', 'baz']};
};

const getKeyTypes = (): Record<string, any> => {
  return {
    foo: {type: 'name', dim: 1},
  };
};
