import {expect} from 'chai';
import {ClassData} from '../../../src/client/Cmatch';
import {BucketLayerData} from '../../../src/program/Program';
import {LayeredBucket} from '../../../src/bucket/layer/LayeredBucket';
import {BaseLayer} from '../../../src/bucket/layer/BaseLayer';
import {createLayeredBucket} from '../../../src/bucket/layer/createLayeredBucket';
import {BucketData} from '../../../src/bucket/Bucket';
import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';

describe('DefaultingLayeredBucket', () => {
  describe('#getData', () => {
    it('Should add any new data found in additional layers', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
        layer_2: {
          is: true,
        },
        layer_3: {
          is: true,
        },
      });

      const base: LayeredBucket = getBaseLayer({d: ['test d']});

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'test a',
          },
          id: 'test 1',
          when: 'layer_1',
        },
        {
          defaults: {
            b: 'test b',
          },
          id: 'test 2',
          when: 'layer_2',
        },
        {
          defaults: {
            c: 'test c',
          },
          id: 'test 3',
          when: 'layer_3',
        },
      ];

      const expected = {
        a: ['test a'],
        b: ['test b'],
        c: ['test c'],
        d: ['test d'],
      };

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getData()).to.deep.equal(expected);
    });

    it('Should only apply layers that are active', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
        layer_2: {
          is: false,
        },
        layer_3: {
          is: false,
        },
      });
      const base: LayeredBucket = getBaseLayer();

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'test a',
          },
          id: 'test 1',
          when: 'layer_1',
        },
        {
          defaults: {
            b: 'test b',
          },
          id: 'test 2',
          when: 'layer_2',
        },
        {
          defaults: {
            c: 'test c',
          },
          id: 'test 3',
          when: 'layer_3',
        },
      ];

      const expected = {
        a: ['test a'],
      };

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getData()).to.deep.equal(expected);
    });

    it('Should override values with the deepest layer found', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
        layer_2: {
          is: true,
        },
        layer_3: {
          is: true,
        },
      });
      const base: LayeredBucket = getBaseLayer({a: ['Base']});

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'overridden',
            b: 'overridden',
            c: 'overridden',
          },
          id: 'test 1',
          when: 'layer_1',
        },
        {
          defaults: {
            a: 'overridden',
            b: 'Layer 2',
            c: 'overridden',
          },
          id: 'test 2',
          when: 'layer_2',
        },
        {
          defaults: {
            a: 'overridden',
            c: 'Layer 3',
          },
          id: 'test 3',
          when: 'layer_3',
        },
      ];

      const expected = {
        a: ['Base'],
        b: ['Layer 2'],
        c: ['Layer 3'],
      };

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getData()).to.deep.equal(expected);
    });

    it('Aligns indexes with default data', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
      });

      const base: LayeredBucket = getBaseLayer({
        a: ['base data', '', 'base data'],
      });

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'default data',
          },
          id: 'test 1',
          when: 'layer_1',
        },
      ];

      const expected = {a: ['base data', 'default data', 'base data']};

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getData()).to.deep.equal(expected);
    });

    it('Defaults to false if expected classification is not found', () => {
      const classData = <ClassData>(<unknown>{});

      const base: LayeredBucket = getBaseLayer({
        a: ['base data'],
      });

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'layered data',
          },
          id: 'test 1',
          when: 'layer_1',
        },
      ];

      const expected = {a: ['base data']};

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getData()).to.deep.equal(expected);
    });
  });

  describe('#getDataByName', () => {
    it('Should return the innermost layer`s data', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
        layer_2: {
          is: true,
        },
        layer_3: {
          is: true,
        },
      });

      const base: LayeredBucket = getBaseLayer({});

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'overridden',
          },
          id: 'test 1',
          when: 'layer_1',
        },
        {
          defaults: {
            a: 'overridden',
          },
          id: 'test 2',
          when: 'layer_2',
        },
        {
          defaults: {
            a: 'this is layer 3!',
          },
          id: 'test 3',
          when: 'layer_3',
        },
      ];

      const expected = ['this is layer 3!'];

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getDataByName('a')).to.deep.equal(expected);
    });

    it('Should always prioritize bucket data when it is found', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
        layer_2: {
          is: true,
        },
      });
      const base: LayeredBucket = getBaseLayer({
        a: ['this is the base layer'],
      });

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'overridden',
          },
          id: 'test 1',
          when: 'layer_1',
        },
        {
          defaults: {
            a: 'overridden',
          },
          id: 'test 2',
          when: 'layer_2',
        },
      ];

      const expected = ['this is the base layer'];

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getDataByName('a')).to.deep.equal(expected);
    });

    it('Should grab a value even if its not found in the first layers', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
        layer_2: {
          is: true,
        },
        layer_3: {
          is: true,
        },
      });

      const base: LayeredBucket = getBaseLayer();

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'overridden',
          },
          id: 'test 1',
          when: 'layer_1',
        },
        {
          defaults: {
            b: 'overridden',
          },
          id: 'test 2',
          when: 'layer_2',
        },
        {
          defaults: {
            c: 'Found in layer 3',
          },
          id: 'test 3',
          when: 'layer_3',
        },
      ];

      const expected = ['Found in layer 3'];

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getDataByName('c')).to.deep.equal(expected);
    });

    it('Should skip inactive layers', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: false,
        },
      });

      const base: LayeredBucket = getBaseLayer({a: ['base data']});

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'not returned',
          },
          id: 'test 1',
          when: 'layer_1',
        },
      ];

      const expected = ['base data'];

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getDataByName('a')).to.deep.equal(expected);
    });

    it('Aligns indexes with default data', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
      });

      const base: LayeredBucket = getBaseLayer({
        a: ['base data', '', 'base data'],
      });

      const layers: BucketLayerData[] = [
        {
          defaults: {
            a: 'default data',
          },
          id: 'test 1',
          when: 'layer_1',
        },
      ];

      const expected = ['base data', 'default data', 'base data'];

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getDataByName('a')).to.deep.equal(expected);
    });

    it('Field value with empty string should return empty string when there is no default', () => {
      const classData = <ClassData>(<unknown>{
        layer_1: {
          is: true,
        },
      });

      const base: LayeredBucket = getBaseLayer({
        a: [''],
      });

      const layers: BucketLayerData[] = [
        {
          defaults: {
            not_a: 'default data',
          },
          id: 'test 1',
          when: 'layer_1',
        },
      ];

      const expected = [''];

      const sut = createLayeredBucket(base, layers);

      sut.updateClassData(classData);

      expect(sut.getDataByName('a')).to.deep.equal(expected);
    });
  });
});

const getBaseLayer = (data: BucketData = {}): BaseLayer => {
  const bucket = <QuoteDataBucket>{
    getDataByName: (name: string) => {
      return data[name];
    },

    getData() {
      return data;
    },
  };

  return new BaseLayer(bucket);
};
