/**
 * Test the bucket utility functions
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {
  getPrefixes,
  getPrefixesTrue,
} from '../../../src/bucket/util/commonSuffix';

import {expect} from 'chai';
import {BucketData} from '../../../src/bucket/Bucket';

describe('#getPrefixes', () => {
  it('returns prefixes from keys with a common suffix', () => {
    const input = {
      foo_with_suffix: [],
      bar_without_suffix: [],
      baz_with_suffix: [],
      baq_with_suffix_in_middle: [],
      _with_suffix: [],
    };

    const expected = ['foo', 'baz'];
    const given = getPrefixes('_with_suffix')(input);

    expect(given).to.deep.equal(expected);
  });

  it('returns empty array on undefined data', () => {
    const input = <Record<string, any>>(<unknown>undefined);

    const expected: string[] = [];
    const given = getPrefixes('_with_suffix')(input);

    expect(given).to.deep.equal(expected);
  });

  it('returns empty array on empty data object', () => {
    const input = {};

    const expected: string[] = [];
    const given = getPrefixes('_with_suffix')(input);

    expect(given).to.deep.equal(expected);
  });
});

describe('#getPrefixesTrueAndTotal', () => {
  type getPrefixesTrueAndTotalTestCase = [string, BucketData, string[]];

  (<getPrefixesTrueAndTotalTestCase[]>[
    [
      'The retry counts are found',
      {
        'supplier-a__retry': [0],
        'supplier-b__retry': [1],
        'supplier-c__retry': [1],
        'supplier-d__retry': [0],
      },
      ['supplier-b', 'supplier-c'],
    ],
    [
      'Nested arrays are handled',
      {
        'supplier-a__retry': [0],
        'supplier-b__retry': [0],
        'supplier-c__retry': [0],
        'supplier-d__retry': [[0]],
      },
      [],
    ],
    [
      'Only counts retries and not other data',
      {
        'supplier-a__retry': [0],
        'supplier-b__retry': [0],
        'supplier-c__reetry': [1],
        'supplier-d__retry': [1],
      },
      ['supplier-d'],
      ,
    ],
  ]).forEach(([label, data, expected]) => {
    it(label, () => {
      expect(
        getPrefixesTrue('__retry')(data)([
          'supplier-a',
          'supplier-b',
          'supplier-c',
          'supplier-d',
        ])
      ).to.deep.equal(expected);
    });
  });
});
