/**
 * Tests Calc
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* eslint-disable no-undef */

const liza = require('../..'),
  relativeDate = liza.calc.Calc.relativeDate,
  uniq = liza.calc.Calc.uniq,
  lower = liza.calc.Calc.lower,
  upper = liza.calc.Calc.upper,
  isIn = liza.calc.Calc.isIn,
  numericIsIn = liza.calc.Calc.numericIsIn,
  expect = require('chai').expect;

describe('relativeDate', () => {
  //returns a date relative to another date by adding a given value

  it("10 days is added, the date's month changes", () => {
    expect(relativeDate(['2019-11-25'], ['10d'])).to.have.all.members([
      '2019-12-05',
    ]);
  });

  it("10 days is added, the date's month and year changes", () => {
    expect(relativeDate(['2019-12-28'], ['10d'])).to.have.all.members([
      '2020-01-07',
    ]);
  });

  it('edge case 6 months is added, date does not go into the 7th month', () => {
    expect(relativeDate(['2019-12-31'], ['6m'])).to.have.all.members([
      '2020-06-30',
    ]);
  });

  it('edge case 3 months is added, date does not go into the 4th month', () => {
    expect(relativeDate(['2019-08-31'], ['3m'])).to.have.all.members([
      '2019-11-30',
    ]);
  });

  it('2 years is added', () => {
    expect(relativeDate(['2019-12-31'], ['2y'])).to.have.all.members([
      '2021-12-31',
    ]);
  });

  it('edge case february', () => {
    expect(relativeDate(['2018-12-31'], ['2m'])).to.have.all.members([
      '2019-02-28',
    ]);
  });

  it('edge case february leap year', () => {
    expect(relativeDate(['2019-12-31'], ['2m'])).to.have.all.members([
      '2020-02-29',
    ]);
  });
});

describe('uniq', () => {
  //returns an array of unique values in same order they are given

  it('returns same element', () => {
    expect(uniq(['AL'])).to.deep.equal(['AL']);
  });

  it('returns unique element', () => {
    expect(uniq(['AL', 'AL'])).to.deep.equal(['AL']);
  });

  it('returns unique element with three duplicates', () => {
    expect(uniq(['NY', 'NY', 'NY'])).to.deep.equal(['NY']);
  });

  it('returns unique elements, one duplicate', () => {
    expect(uniq(['TX', 'NY', 'NY'])).to.deep.equal(['TX', 'NY']);
  });

  it('returns unique elements, two duplicates', () => {
    expect(uniq(['TX', 'NY', 'NY', 'TX'])).to.deep.equal(['TX', 'NY']);
  });

  it('returns unique with mixed duplicates', () => {
    expect(
      uniq(['CA', 'NY', 'TX', 'OH', 'TX', 'NY', 'AL', 'LA', 'LA', 'CA'])
    ).to.deep.equal(['CA', 'NY', 'TX', 'OH', 'AL', 'LA']);
  });
});

describe('lower', () => {
  it('converts all elements', () => {
    expect(lower(['Foo', 'bar', 'BAZ', '', null])).to.deep.equal([
      'foo',
      'bar',
      'baz',
      '',
    ]);
  });
});

describe('upper', () => {
  it('converts all elements', () => {
    expect(upper(['Foo', 'bar', 'BAZ', '', null])).to.deep.equal([
      'FOO',
      'BAR',
      'BAZ',
      '',
    ]);
  });
});

describe('numericIsIn', () => {
  describe('true conditions', () => {
    [
      {
        label: 'single exact match',
        data: [1],
        value: [1],
      },
      {
        label: 'single match present at start',
        data: [1, 2],
        value: [1],
      },
      {
        label: 'single match present at end',
        data: [1, 2],
        value: [2],
      },
      {
        label: 'duplicate matches',
        data: [1, 2],
        value: [2, 2],
      },
      {
        label: 'multiple matches',
        data: [1, 2, 3],
        value: [1, 2],
      },
      {
        label: 'multiple exact matches',
        data: [1, 2],
        value: [1, 2],
      },
    ].forEach(({label, data, value}) => {
      it(label, () => expect(numericIsIn(data, value)).to.deep.equal(['1']));
    });
  });

  describe('false conditions', () => {
    [
      {
        label: 'no match',
        data: [1, 2],
        value: [0],
      },
      {
        label: 'not all values match',
        data: [1, 2],
        value: [2, 3],
      },
      {
        label: 'empty value',
        data: [1, 2],
        value: [],
      },
      {
        label: 'empty data',
        data: [],
        value: [1],
      },
    ].forEach(({label, data, value}) => {
      it(label, () => expect(numericIsIn(data, value)).to.deep.equal(['0']));
    });
  });
});

describe('isIn', () => {
  describe('true conditions', () => {
    [
      {
        label: 'single exact match',
        data: ['a'],
        value: ['a'],
      },
      {
        label: 'single match present at start',
        data: ['a', 'b'],
        value: ['a'],
      },
      {
        label: 'single match present at end',
        data: ['a', 'b'],
        value: ['b'],
      },
      {
        label: 'duplicate matches',
        data: ['a', 'b'],
        value: ['b', 'b'],
      },
      {
        label: 'multiple matches',
        data: ['a', 'b', 'c'],
        value: ['a', 'b'],
      },
      {
        label: 'multiple exact matches',
        data: ['a', 'b'],
        value: ['a', 'b'],
      },
    ].forEach(({label, data, value}) => {
      it(label, () => expect(isIn(data, value)).to.deep.equal(['1']));
    });
  });

  describe('false conditions', () => {
    [
      {
        label: 'no match',
        data: ['a', 'b'],
        value: ['0'],
      },
      {
        label: 'not all values match',
        data: ['a', 'b'],
        value: ['b', 'c'],
      },
      {
        label: 'empty value',
        data: ['a', 'b'],
        value: [],
      },
      {
        label: 'empty data',
        data: [],
        value: ['a'],
      },
    ].forEach(({label, data, value}) => {
      it(label, () => expect(isIn(data, value)).to.deep.equal(['0']));
    });
  });
});
