/**
 * Tests NavigateEventHandler
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {expect} from 'chai';
import {NavigateEventHandler as Sut} from '../../../src/client/event/NavigateEventHandler';
import {Client} from '../../../src/client/Client';
import {Nav} from '../../../src/client/nav/Nav';
import {Ui} from '../../../src/ui/Ui';

describe('Handle navigate event', () => {
  [
    {
      label:
        'does not navigate or redraw nav if step id is not found in program',
      step_ids: ['step_one', 'step_two'],
      target_step: 'foo',
      expected_nav_called: false,
      expected_redraw_called: false,
    },
    {
      label: 'does not navigate or redraw nav if step id is zero',
      step_ids: ['step_one', 'step_two'],
      target_step: 'step_one',
      expected_nav_called: false,
      expected_redraw_called: false,
    },
    {
      label: 'navigates and redraws if step id is found in program',
      step_ids: ['step_one', 'step_two'],
      target_step: 'step_two',
      expected_nav_called: true,
      expected_redraw_called: true,
    },
  ].forEach(
    ({
      label,
      step_ids,
      target_step,
      expected_redraw_called,
      expected_nav_called,
    }) => {
      it(label, done => {
        const {nav, client, ui} = createStubs(step_ids);

        let navigate_to_step_called = false;
        nav.navigateToStep = (_step_id: any) => {
          navigate_to_step_called = true;
        };

        let redraw_called = false;
        ui.redrawNav = () => {
          redraw_called = true;
          return <Ui>{};
        };

        const sut = new Sut(client);
        sut.handle('navigate', done, {
          action: 'navigate',
          value: target_step,
        });

        expect(navigate_to_step_called).to.equal(expected_nav_called);
        expect(redraw_called).to.equal(expected_redraw_called);
      });
    }
  );
});

function createStubNavigation() {
  return <Nav>(<unknown>{
    navigateToStep: (_step_id: any) => {},
  });
}

function createStubUi() {
  return <Ui>(<unknown>{
    redrawNav: () => {},
  });
}

function createStubClient(nav: Nav, ui: Ui, step_ids: any[]) {
  return <Client>(<unknown>{
    nav: nav,
    program: {
      stepIds: step_ids,
    },
    getUi: () => ui,
  });
}

function createStubs(step_ids: any[]) {
  const nav = createStubNavigation();
  const ui = createStubUi();
  const client = createStubClient(nav, ui, step_ids);

  return {
    client: client,
    nav: nav,
    ui: ui,
  };
}
