import {
  ValueSetEventHandler as Sut,
  ClientActionSetData,
} from '../../../src/client/event/ValueSetEventHandler';
import {ClientQuote} from '../../../src/client/quote/ClientQuote';
import {Client} from '../../../src/client/Client';
import {DelayedStagingBucket} from '../../../src/bucket/DelayedStagingBucket';
import {StagingBucket} from '../../../src/bucket/StagingBucket';

import {expect} from 'chai';

describe('ValueSetEventHandler', () => {
  describe('#handle', () => {
    it('sets relative indexes up until last source index', done => {
      const expected = {
        foo: ['zero', 'one', 'one'],
      };

      let qcalled = false;
      const quote = <ClientQuote>(<unknown>{
        setData(data: ClientActionSetData) {
          expect(data).to.deep.equal(expected);
          qcalled = true;
        },
        visitData(visitor: (bucket: StagingBucket) => void): void {
          visitor(<DelayedStagingBucket>{processValues: () => {}});
        },
      });

      const callback = () => {
        expect(qcalled).to.be.true;
        done();
      };

      const sut = getSut(quote);

      sut.handle('set', callback, {
        elementName: 'foo',
        indexes: [0, 1, 2],
        value: ['zero', 'one'],
      });
    });

    it('sets only given indexes', done => {
      const expected = {
        bar: [, 'set1', 'set2'],
      };

      let qcalled = false;
      let pvCalled = false;
      const quote = <ClientQuote>(<unknown>{
        setData(data: ClientActionSetData) {
          expect(data).to.deep.equal(expected);
          qcalled = true;
        },
        visitData(visitor: (bucket: StagingBucket) => void): void {
          visitor(<DelayedStagingBucket>{
            processValues: () => {
              pvCalled = true;
            },
          });
        },
      });

      const callback = () => {
        expect(qcalled).to.be.true;
        expect(pvCalled).to.be.true;
        done();
      };

      const sut = getSut(quote);

      sut.handle('set', callback, {
        elementName: 'bar',
        indexes: [1, 2],
        value: ['ignore', 'set1', 'set2', 'extra'],
      });
    });

    it('handles string values and sets data on first index', done => {
      const expected = {
        bar: ['foo'],
      };

      let qcalled = false;
      let pvCalled = false;
      const quote = <ClientQuote>(<unknown>{
        setData(data: ClientActionSetData) {
          expect(data).to.deep.equal(expected);
          qcalled = true;
        },
        visitData(visitor: (bucket: StagingBucket) => void): void {
          visitor(<DelayedStagingBucket>{
            processValues: () => {
              pvCalled = true;
            },
          });
        },
      });

      const callback = () => {
        expect(qcalled).to.be.true;
        expect(pvCalled).to.be.true;
        done();
      };

      const sut = getSut(quote);

      sut.handle('set', callback, {
        elementName: 'bar',
        indexes: [0],
        value: 'foo',
      });
    });

    it('string value is set on all indexes of destination field', done => {
      const expected = {
        bar: ['foo', 'foo', 'foo'],
      };

      let qcalled = false;
      let pvCalled = false;
      const quote = <ClientQuote>(<unknown>{
        setData(data: ClientActionSetData) {
          expect(data).to.deep.equal(expected);
          qcalled = true;
        },
        visitData(visitor: (bucket: StagingBucket) => void): void {
          visitor(<DelayedStagingBucket>{
            processValues: () => {
              pvCalled = true;
            },
          });
        },
      });

      const callback = () => {
        expect(qcalled).to.be.true;
        expect(pvCalled).to.be.true;
        done();
      };

      const sut = getSut(quote);

      sut.handle('set', callback, {
        elementName: 'bar',
        indexes: [0, 1, 2],
        value: 'foo',
      });
    });
  });
});

const getSut = (quote: ClientQuote) => {
  return new Sut(getClient(quote));
};

const getClient = (quote: ClientQuote) => {
  return <Client>(<unknown>{
    getQuote: () => quote,
  });
};
