/**
 * Test case for webhook dapi
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {expect} from 'chai';
import {TokenId} from '../../src/server/token/Token';
import {Bucket, BucketData} from '../../src/bucket/Bucket';
import {WebhookBucketDataApi as Sut} from '../../src/dapi/WebhookBucketDataApi';

describe('WebhookBucketDataApi', () => {
  it('Sets token values in the bucket', () => {
    const bucket = getMockBucket();
    const sut = new Sut(bucket);

    let setValues_called = false;

    bucket.getDataByName = () => [];

    bucket.setValues = vals => {
      expect(vals).to.deep.equal({
        ['webhook_trigger_foobar']: ['123123'],
        ['webhook_trigger_aggregate']: ['webhook_trigger_foobar'],
      });

      setValues_called = true;

      return bucket;
    };

    sut.request('', () => {}, 'foobar', undefined, <TokenId>'123123');

    expect(setValues_called).to.be.true;
  });

  it('Throws error on missing token', () => {
    const bucket = getMockBucket();
    const sut = new Sut(bucket);

    let error;

    try {
      sut.request('', () => {}, 'foobar', undefined, undefined);
    } catch (e) {
      if (e instanceof Error) {
        error = e;
      }
    }

    expect(error?.message).to.equal(
      'Webhook bucket dapi triggered without a valid token'
    );
  });

  it('Cleans tokens from bucket', () => {
    const data: BucketData = {
      webhook_trigger_aggregate: ['webhook_trigger_foobar'],
      webhook_trigger_foobar: ['234'],
      other_data: ['456'],
    };

    Sut.sanitizeClientData(data);

    expect(data).to.deep.equal({other_data: ['456']});
  });
});

const getMockBucket = () => <Bucket>(<unknown>{
    setValues: () => {},
    getDataByName: () => {},
  });
