/**
 * Test case for data transmission over HTTP(S)
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {HttpImpl} from '../../../src/dapi/http/HttpImpl';
import {HttpDataApi as Sut} from '../../../src/dapi/http/HttpDataApi';

const expect = require('chai').expect,
  dummy_url = 'http://foo';

const dummy_impl = <HttpImpl>{
  requestData: (
    _: string,
    __: string,
    ___: string,
    ____: NodeCallback<string>
  ) => dummy_impl,
  setOptions: () => {},
};

describe('HttpDataApi', () => {
  it('permits RFC 2616 HTTP methods', () => {
    const m = ['GET', 'POST', 'PUT', 'HEAD', 'OPTIONS', 'DELETE', 'TRACE'];

    m.forEach(method => {
      expect(() => {
        new Sut(dummy_url, method, dummy_impl);
      }).to.not.throw(Error);
    });
  });

  it('does not permit non-RFC-2616 HTTP methods', () => {
    expect(() => {
      new Sut(dummy_url, 'FOO', dummy_impl);
    }).to.throw(Error, 'FOO');
  });

  describe('.request', () => {
    const impl = new (class implements HttpImpl {
      public provided: any[] = [];
      public data: any = '';
      public err: any = null;

      requestData(
        url: string,
        method: string,
        data: string,
        c: NodeCallback<string>
      ) {
        this.provided[0] = url;
        this.provided[1] = method;
        this.provided[2] = data;
        this.provided[3] = c;
        c(this.err, this.data);

        return this;
      }

      setOptions() {}

      reset() {
        this.data = '';
        this.err = null;
      }
    })();

    /**
     * The actual request is performed by some underling implementation.
     * This additional level of indirection allows the general concept of an
     * "HTTP Data API" to vary from an underyling HTTP protocol
     * implementation; they are separate concerns, although the distinction
     * may seem subtle.
     */
    it('delegates to provided HTTP implementation', () => {
      const method = 'GET',
        data = 'ribbit',
        c = function () {};

      impl.reset();

      new Sut(dummy_url, method, impl).request(data, c);

      const provided = impl.provided;
      expect(provided[0]).to.equal(dummy_url);
      expect(provided[1]).to.equal(method);
      expect(provided[2]).to.equal(data);
    });

    [
      // default is urlencoded
      {
        enctype: '',
        method: 'POST',
        data: {foo: 'bar=baz', '&bar': 'moo%cow'},
        expected:
          'foo=' +
          encodeURIComponent('bar=baz') +
          '&' +
          encodeURIComponent('&bar') +
          '=' +
          encodeURIComponent('moo%cow'),
      },

      // same as above
      {
        enctype: 'application/x-www-form-urlencoded',
        method: 'POST',
        data: {foo: 'bar=baz', '&bar': 'moo%cow'},
        expected:
          'foo=' +
          encodeURIComponent('bar=baz') +
          '&' +
          encodeURIComponent('&bar') +
          '=' +
          encodeURIComponent('moo%cow'),
      },

      // empty string
      {
        enctype: 'application/x-www-form-urlencoded',
        method: 'POST',
        data: {},
        expected: '',
      },

      // json
      {
        enctype: 'application/json',
        method: 'POST',
        data: {foo: 'bar'},
        expected: '{"foo":"bar"}',
      },

      // ignored if GET
      {
        enctype: 'application/json',
        method: 'GET',
        data: {foo: 'bar'},
        expected: 'foo=bar',
      },
    ].forEach(({enctype, method, data, expected}, i) => {
      it(`${method} encodes (${i})`, () => {
        impl.reset();

        new Sut(dummy_url, method, impl, enctype).request(<any>data, _ => {});

        expect(impl.provided[2]).to.equal(expected);
      });
    });

    [
      {
        label: 'urls without brackets are passed through unchanged',
        url: '/unchanged',
        data: {},
        expected: '/unchanged',
      },
      {
        label: 'urls with brackets have fields substituted',
        url: '/this/has/{val}/changed',
        data: {
          val: 'been',
        },
        expected: '/this/has/been/changed',
      },
      {
        label: 'fields will not be substituted if it is not a string',
        url: '/not/{val}/subbed',
        data: {
          val: {
            foo: 'bar',
          },
        },
        expected: '/not/{val}/subbed',
      },
      {
        label: 'fields will not be substituted if data does not have the key',
        url: '/not/{val}/subbed',
        data: {},
        expected: '/not/{val}/subbed',
      },
      {
        label: 'multiple values can be substituted in url',
        url: '/{one}/{two}',
        data: {
          one: 'foo',
          two: 'bar',
        },
        expected: '/foo/bar',
      },
    ].forEach(({label, url, data, expected}) => {
      it(label, () => {
        impl.reset();

        new Sut(url, 'POST', impl, 'application/json').request(
          <any>data,
          _ => {}
        );

        expect(impl.provided[0]).to.equal(expected);
      });
    });

    /**
     * Method chaining
     */
    it('returns self', () => {
      impl.reset();

      const sut = new Sut(dummy_url, 'GET', impl),
        ret = sut.request('', () => {});

      expect(ret).to.equal(sut);
    });

    /**
     * String requests are intended to be raw messages, whereas objects are
     * treated as key-value params.
     */
    it('accepts string and object data', () => {
      impl.reset();

      expect(() => {
        new Sut(dummy_url, 'GET', impl)
          .request('', () => {}) // string
          .request({}, () => {}); // object
      }).to.not.throw(Error);
    });

    it('accepts null data, converting to empty string', () => {
      impl.reset();

      expect(() => {
        new Sut(dummy_url, 'GET', impl).request(<any>null, () => {
          expect(impl.provided[2]).to.equal('');
        });
      }).to.not.throw(Error);
    });

    it('rejects all other data types', () => {
      impl.reset();

      const sut = new Sut(dummy_url, 'GET', impl);

      [123, Infinity, undefined, NaN, function () {}].forEach(data => {
        expect(() => {
          sut.request(<any>data, () => {});
        }).to.throw(TypeError);
      });
    });

    it('returns response provided by HTTP implementation', done => {
      impl.reset();

      impl.data = '{}';
      new Sut(dummy_url, 'GET', impl).request('', (_err, resp) => {
        expect(resp).to.deep.equal(JSON.parse(impl.data));
        done();
      });
    });

    it('converts response to JSON', done => {
      impl.reset();

      impl.data = '{"foo": "bar"}';

      new Sut(dummy_url, 'GET', impl).request('', (err, data) => {
        // should have been converted to JSON
        expect(data).to.deep.equal({foo: 'bar'});
        expect(err).to.equal(null);
        done();
      });
    });

    describe('when JSON parsing fails', () => {
      it('returns error', done => {
        impl.reset();

        new Sut(dummy_url, 'GET', impl).request('<?a>', (err, _data) => {
          expect(err).to.be.instanceOf(Error);
          done();
        });
      });
    });

    describe('on request error from supertype', () => {
      it('attempts to format output as JSON', done => {
        impl.reset();

        impl.data = '{"foo": "bar"}';

        new Sut(dummy_url, 'GET', impl).request('', (_, data: any) => {
          expect(data).to.be.a('object');
          expect(data.foo).to.equal('bar');
          done();
        });
      });

      it('produces both errors on bad JSON output', done => {
        impl.reset();

        const error_message_expected =
          'Multiple errors occurred; Original: "' +
          'Error: foobar' +
          '"; Parsed: "' +
          'SyntaxError: Unexpected end of JSON input' +
          '"';
        impl.err = new Error('foobar');

        new Sut(dummy_url, 'GET', impl).request('Bad JSON', (err, _) => {
          // the main error should indicate both
          expect(err).to.be.instanceOf(Error);

          // and we should provide references to both
          expect(err?.message).to.equal(error_message_expected);

          done();
        });
      });
    });
  });
});
