/**
 *  Tests for IndicationCreationManager
 *
 *  Copyright (C) 2010-2023 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {EventEmitter} from 'events';
import {Either} from 'fp-ts/lib/Either';
import {pollForRates as sut} from '../../../src/dullahan/client/RatePoller';
import {RateResult} from '../../../src/server/rater/Rater';
import {RatingService} from '../../../src/server/service/RatingService';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';

import {expect} from 'chai';

let resolve_timeout = () => {};

beforeEach(() => {
  resolve_timeout = () => {};
});

describe('RatePoller', () => {
  it('Waits until fetchRate returns with nothing to proceed', async () => {
    let request_rate_calls = 0;

    const timeout_ms = 123;

    const mockRatingService = <RatingService>(<unknown>{
      fetchRate: async () => {
        if (request_rate_calls++ === 0) {
          return {
            content: {
              data: {suppliers_pending_rate: ['foo']},
              initialRatedDate: <UnixTimestamp>123,
              lastRatedDate: <UnixTimestamp>123,
            },
            actions: [{action: 'delay', then: {action: 'fetchRate'}}],
          };
        }

        return {
          content: {
            data: {suppliers_pending_rate: ['foo']},
            initialRatedDate: <UnixTimestamp>123,
            lastRatedDate: <UnixTimestamp>123,
          },
          actions: [],
        };
      },
    });

    const emitted_events: string[] = [];
    const emitter = createEmitter();

    emitter.emit = (event: any) => {
      emitted_events.push(event);

      return true;
    };

    const sutFn = sut(
      emitter,
      createMockDelay(timeout_ms),
      mockRatingService,
      mockGetPollingSettings(timeout_ms)
    );

    const quote = createQuote();

    await sutFn(
      234,
      () => Promise.resolve(<Either<unknown, ServerSideQuote>>{right: quote}),
      {}
    );

    expect(emitted_events).to.deep.equal([
      'rate-polling-begin',
      'pending-rate',
      'rating-completed',
    ]);
  });

  it('Reloads the quote each iteration', async () => {
    let request_rate_calls = 0;
    let reload_quote_calls = 0;

    const timeout_ms = 123;

    const mockRatingService = <RatingService>(<unknown>{
      fetchRate: async () => {
        if (request_rate_calls++ === 0) {
          return {
            content: {
              data: {suppliers_pending_rate: ['foo']},
              initialRatedDate: <UnixTimestamp>123,
              lastRatedDate: <UnixTimestamp>123,
            },
            actions: [{action: 'delay', then: {action: 'fetchRate'}}],
          };
        }

        return {
          content: {
            data: {suppliers_pending_rate: ['foo']},
            initialRatedDate: <UnixTimestamp>123,
            lastRatedDate: <UnixTimestamp>123,
          },
          actions: [],
        };
      },
    });

    const sutFn = sut(
      createEmitter(),
      createMockDelay(timeout_ms),
      mockRatingService,
      mockGetPollingSettings(timeout_ms)
    );

    const quote = createQuote();

    await sutFn(
      234,
      () => {
        reload_quote_calls++;

        return Promise.resolve(<Either<unknown, ServerSideQuote>>{
          right: quote,
        });
      },
      {}
    );

    expect(reload_quote_calls).to.deep.equal(2);
  });

  it('Returns new rate results', async () => {
    const timeout_ms = 123;

    const rate_results_expected = {foo: ['bar']};

    const mockRatingService = <RatingService>(<unknown>{
      fetchRate: async () => ({
        content: {
          data: rate_results_expected,
          initialRatedDate: <UnixTimestamp>123,
          lastRatedDate: <UnixTimestamp>123,
        },
        actions: [],
      }),
    });

    const sutFn = sut(
      createEmitter(),
      createMockDelay(timeout_ms),
      mockRatingService,
      mockGetPollingSettings(timeout_ms)
    );

    const quote = createQuote();

    const rate_result_actual = await sutFn(
      234,
      async () =>
        <Either<unknown, ServerSideQuote>>{
          right: quote,
        },
      {}
    );

    expect(rate_result_actual).to.deep.equal(rate_results_expected);
  });

  it('Returns original rate results on immediate timeout', async () => {
    const timeout_ms = 123;

    // const rate_results_from_reload = {foo: ['bar']};
    const rate_results_expected = {baz: ['moo']};

    const mockRatingService = <RatingService>(<unknown>{
      fetchRate: async () => new Promise<RateResult>(() => {}),
    });

    const emitted_events: string[] = [];
    const emitter = createEmitter();

    emitter.emit = (event: any) => {
      emitted_events.push(event);

      return true;
    };

    const sutFn = sut(
      emitter,
      createMockDelay(timeout_ms, false),
      mockRatingService,
      mockGetPollingSettings(timeout_ms)
    );

    const quote = createQuote();

    const rate_results_actual = await sutFn(
      234,
      () => {
        resolve_timeout();

        return Promise.resolve(<Either<unknown, ServerSideQuote>>{
          right: quote,
        });
      },
      rate_results_expected
    );

    expect(rate_results_actual).to.deep.equal(rate_results_expected);
  });

  it('Times out properly', async () => {
    let reload_quote_calls = 0;

    const timeout_ms = 123;

    const mockRatingService = <RatingService>(<unknown>{
      fetchRate: async () => {
        return {
          content: {
            data: {suppliers_pending_rate: ['foo']},
            initialRatedDate: <UnixTimestamp>123,
            lastRatedDate: <UnixTimestamp>123,
          },
          actions: [{action: 'delay', then: {action: 'fetchRate'}}],
        };
      },
    });

    const emitted_events: string[] = [];
    const emitter = createEmitter();

    emitter.emit = (event: any) => {
      emitted_events.push(event);

      return true;
    };

    const sutFn = sut(
      emitter,
      createMockDelay(timeout_ms, false),
      mockRatingService,
      mockGetPollingSettings(timeout_ms)
    );

    const quote = createQuote();

    await sutFn(
      234,
      () => {
        if (++reload_quote_calls > 1) {
          resolve_timeout();
        }

        return Promise.resolve(<Either<unknown, ServerSideQuote>>{
          right: quote,
        });
      },
      {}
    );

    expect(reload_quote_calls).to.deep.equal(1);
    expect(emitted_events).to.deep.equal([
      'rate-polling-begin',
      'pending-rate',
    ]);
  });
});

const createMockDelay =
  (timeout_ms: number, resolve_on_interval = true) =>
  async (ms: number) => {
    const is_interval = ms !== timeout_ms;

    if (is_interval === resolve_on_interval) {
      return Promise.resolve();
    }

    return new Promise<void>(resolve => (resolve_timeout = resolve));
  };

// Delay relies on the fact that these numbers cannot be the same
const mockGetPollingSettings = (timeout_ms: number) => (_: number) => ({
  timeout_ms,
  interval_ms: ++timeout_ms,
});

const createQuote = () => ({
  getId: () => 0,
});

const createEmitter = () => <EventEmitter>(<unknown>{
    emit: () => {},
    on: () => {},
  });
