/**
 *  Tests for Indication Controller
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {expect, use as chai_use, assert} from 'chai';
chai_use(require('chai-as-promised'));

import * as TE from 'fp-ts/TaskEither';
import * as E from 'fp-ts/Either';
import {AsyncQuoteFactory} from '../../../src/dullahan/quote/QuoteFactory';
import {Bucket} from '../../../src/bucket/Bucket';
import {ClientQuote} from '../../../src/client/quote/ClientQuote';
import {Cmatch, CmatchData} from '../../../src/client/Cmatch';
import {DataApiManager} from '../../../src/dapi/DataApiManager';
import {DataValidator} from '../../../src/validate/DataValidator';
import {DelayedStagingBucket} from '../../../src/bucket/DelayedStagingBucket';
import {EventEmitter} from 'events';
import {FeatureFlag} from '../../../src/system/flags/FeatureFlag';
import {IndicationCreateDeps} from '../../../src/dullahan/DullahanFactory';
import {IndicationCreationManager} from '../../../src/dullahan/indication/IndicationCreationManager';
import {MongoServerDao} from '../../../src/server/db/MongoServerDao';
import {PositiveInteger} from '../../../src/numeric';
import {ProgramFactory} from '../../../src/dullahan/program/ProgramFactory';
import {Program} from '../../../src/program/Program';
import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';
import {RateResult} from '../../../src/server/rater/Rater';
import {RatingService} from '../../../src/server/service/RatingService';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {StepDataSetter} from '../../../src/dullahan/step/StepDataSetter';
import {StepSimulator} from '../../../src/dullahan/step/StepSimulator';
import {UserSession} from '../../../src/server/request/UserSession';
import {indication as Sut} from '../../../src/dullahan/controllers/IndicationController';
import {mockReq, mockRes} from 'sinon-express-mock';
import {
  DataApiMediator,
  UiFieldData,
} from '../../../src/client/dapi/DataApiMediator';
import {
  Validator,
  ValidatorFactory,
} from '../../../src/dullahan/validate/Validator';
import {
  HttpClient,
  InvalidUrlError,
} from '../../../src/system/network/HttpClient';
import {InitExistingDocument} from '../../../src/server/quote/loader';
import {randomBytes} from 'crypto';

describe('IndicationController', () => {
  describe('create', () => {
    it('fails when an invalid webhook is provided', done => {
      const http_client = createHttpClient();

      http_client.validateUrl = () => {
        throw new InvalidUrlError('');
      };

      const sut = createSutCreate({http_client});
      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      sut(req, res).then(_ => {
        expect(res.status.withArgs(422).callCount).to.equal(1);
        expect(
          res.send.withArgs({
            message: 'Request must contain a valid webhook URL.',
          }).callCount
        ).to.equal(1);
        done();
      });
    });

    it('calls webhook when a valid request is processed', done => {
      const tests = [
        'https://some.websitethatdoesnotexist.joe',
        'https%3A%2F%2Fsome.websitethatdoesnotexist.joe',
      ].map(webhook => {
        const sut = createSutCreate();
        const req = mockReq(createValidRequestPayload());
        const res = mockRes({});

        req.query.callback = webhook;

        return sut(req, res).then(() => {
          expect(res.status.withArgs(202).callCount).to.equal(1);
        });
      });

      Promise.all(tests).then(_ => {
        done();
      });
    });

    it('fails when invalid session data are provided', done => {
      const http_client = createHttpClient();
      const given = 'https://some.websitethatdoesnotexist.joe';

      const sut = createSutCreate({http_client});
      const req = mockReq({body: {session: {}}});
      const res = mockRes({});

      req.query.callback = given;

      sut(req, res).then(_ => {
        expect(res.status.withArgs(422).callCount).to.equal(1);
        expect(
          res.send.withArgs({
            message:
              'Missing session data for the following fields: agentId, agentName, agentEntityId, agencyNumber',
          }).callCount
        ).to.equal(1);
        done();
      });
    });

    it('emits validation failures', () => {
      let given_failure = {};
      const expected_failure = {
        quote_id: 123,
        missing_required: ['something bad happened'],
        failed_assertions: [],
      };

      const deps = {
        createValidator: getCreateValidator(),
        emitter: createEventEmitter((event: any, data: any) => {
          if (event === 'quote-validation-failure') {
            given_failure = data;
          }
        }),
        missing_required: ['something bad happened'],
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      return createSutCreate(deps)(req, res).then(_ => {
        expect(given_failure).to.deep.equal(expected_failure);
      });
    });

    it(`sets the topSavedStepId of the quote to the first step id of the program`, () => {
      const {quote_factory, quote} = createQuoteFactoryWithQuote({});
      let saved_quote: undefined | ServerSideQuote;
      const dao = <MongoServerDao>(<unknown>{
        init(_callback: Function) {},
        getNextQuoteId() {
          return new Promise(resolve => resolve(1));
        },

        saveQuote(_quote: unknown, _class_data: unknown, success: any) {
          return success();
        },

        saveQuoteMetaDirect(_quote: unknown, _new_meta: unknown, success: any) {
          return success();
        },

        saveQuoteMeta(_quote: unknown, _new_meta: unknown, success: any) {
          return success();
        },

        publishQuote(_quote: unknown) {
          return Promise.resolve();
        },

        saveQuoteState(_quote: ServerSideQuote, success: any) {
          saved_quote = _quote;
          return success();
        },
      });
      const deps = {
        quote_factory,
        dao,
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      return createSutCreate(deps)(req, res).then(_ => {
        expect(quote.getTopSavedStepId()).to.equal(2);
        expect(saved_quote).to.equal(quote);
      });
    });

    it('fails from validation failures', () => {
      let given_failure: Record<string, unknown[]> = {
        dapi_validation_failures: [],
      };

      const output_validator =
        (_bucket: Bucket) => (_program: Program) => (error_log: string[]) => {
          error_log.push('foo');
          return (
            _name: string,
            _i: number | string,
            _field_data: UiFieldData[]
          ) => {
            error_log.push('foo');
          };
        };

      const deps = {
        createDapiOutputValidator: output_validator,
        createClientDapiManager: function (_: UserSession) {
          return {
            on: function (event: string, cb: any) {
              if (event === 'updateFieldData') {
                cb();
              }

              return this;
            },
          };
        },
        // NB: When DAPIs are fully enabled we will want to throw an error

        // emitter: createEventEmitter((event: any, data: Error) => {
        //   if (event === 'error') {
        //     given_failure = data.message;
        //   }
        // }),
        emitter: createEventEmitter((event: any, ctx: any) => {
          if (event === 'quote-validation-warning') {
            given_failure = ctx;
          }
        }),
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      return createSutCreate(deps)(req, res).then(_ => {
        expect(given_failure.dapi_validation_failures.length).to.be.gt(0);
      });
    });

    it('invokes client APIs only once', () => {
      const results: any = [];
      const expected = [{field: 'foo', id: 1}];

      const program_factory = createProgramFactory();
      const program = createProgram();

      program.change = (id, field): Record<string, unknown[]> => {
        results.push({id, field});
        return {};
      };

      program_factory.createProgram = () => ({
        bucket: createBucket(),
        program: program,
      });

      const deps = {program_factory};
      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      return createSutCreate(deps)(req, res).then(_ => {
        expect(results).to.deep.equal(expected);
      });
    });

    it('clears internal fields for a non-internal user', () => {
      let given_post_data = {};
      const expected_post_data = {
        bucket: {other_field: ['']},
        rdata: {},
        validation_errors: [],
        sso: false,
      };

      const deps = {
        http_client: createHttpClient((data: any) => {
          given_post_data = data;
        }),
        createValidator: getCreateValidator([]),
        is_internal: false,
        quote_bucket_data: {test_quote_ind: [''], other_field: ['']},
        program_internal: {test_quote_ind: true},
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      return createSutCreate(deps)(req, res).then(_ => {
        expect(given_post_data).to.deep.equal(expected_post_data);
      });
    });

    it('does not clear internal fields for an internal user', async () => {
      let given_post_data = {};
      const expected_post_data = {
        bucket: {test_quote_ind: [''], other_field: ['']},
        rdata: {},
        validation_errors: [],
        sso: true,
      };

      const deps = {
        http_client: createHttpClient((data: any) => {
          given_post_data = data;
        }),
        createValidator: getCreateValidator([]),
        is_internal: true,
        quote_bucket_data: {test_quote_ind: [''], other_field: ['']},
        program_internal: {test_quote_ind: true},
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      await createSutCreate(deps)(req, res);
      expect(given_post_data).to.deep.equal(expected_post_data);
    });

    it('waits for pending rates', async () => {
      let pollForRates_called = false;

      const deps = {
        http_client: createHttpClient(),
        createValidator: getCreateValidator([]),
        is_internal: true,
        quote_bucket_data: {test_quote_ind: [''], other_field: ['']},
        program_internal: {test_quote_ind: true},
        pollForRates: () => {
          pollForRates_called = true;
        },
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      await createSutCreate(deps)(req, res);

      expect(pollForRates_called).to.be.true;
    });

    it('stops processing on missing required fields', async () => {
      let given_post_data = {};
      const expected_post_data = {
        bucket: {},
        rdata: {},
        validation_errors: ['foo'],
        sso: true,
      };

      const manager = createManager();

      manager.invokeServerDapis =
        () => () => () => () => () => () => () => () => {
          assert.fail();

          return Promise.reject();
        };

      const deps = {
        http_client: createHttpClient((data: any) => {
          given_post_data = data;
        }),
        createValidator: getCreateValidator(),
        manager,
        missing_required: ['foo'],
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      await createSutCreate(deps)(req, res);

      expect(given_post_data).to.deep.equal(expected_post_data);
    });

    it('stops processing on failed assertions', async () => {
      let given_post_data = {};
      const expected_post_data = {
        bucket: {},
        rdata: {},
        validation_errors: ['foo'],
        sso: true,
      };

      const manager = createManager();

      manager.invokeServerDapis =
        () => () => () => () => () => () => () => () => {
          assert.fail();

          return Promise.reject();
        };

      const deps = {
        http_client: createHttpClient((data: any) => {
          given_post_data = data;
        }),
        createValidator: getCreateValidator(),
        feature_flag: createFeatureFlag(true),
        manager,
        missing_required: ['foo'],
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      await createSutCreate(deps)(req, res);

      expect(given_post_data).to.deep.equal(expected_post_data);
    });

    it('use rate data returned from rate poller', async () => {
      let pollForRates_called = false;
      let given_post_data = {};

      const rate_data_expected = {foo: 'bar'};
      const data_expected = {other_field: ['']};

      const post_data_expected = {
        bucket: data_expected,
        rdata: rate_data_expected,
        validation_errors: [],
        sso: true,
      };

      const deps = {
        http_client: createHttpClient((data: any) => {
          given_post_data = data;
        }),
        createValidator: getCreateValidator([]),
        is_internal: true,
        quote_bucket_data: data_expected,
        program_internal: {test_quote_ind: true},
        pollForRates: () => {
          pollForRates_called = true;

          return rate_data_expected;
        },
      };

      const req = mockReq(createValidRequestPayload());
      const res = mockRes({});

      await createSutCreate(deps)(req, res).then(_ => {
        expect(given_post_data).to.deep.equal(post_data_expected);
      });

      expect(pollForRates_called).to.be.true;
    });
  });
});

const createEventEmitter = (event_cb = (_: any, __: any) => {}) => {
  return <EventEmitter>(<unknown>{
    emit(event: any, data: any) {
      event_cb(event, data);
    },
  });
};

const createHttpClient = (data_cb = (_: any) => {}) => {
  return <HttpClient>(<unknown>{
    async post(_: any, data: any) {
      data_cb(data);

      return {};
    },

    validateUrl() {
      return true;
    },
  });
};

const createProgramFactory = (program_internal = {}) => {
  return <ProgramFactory>(<unknown>{
    createProgram() {
      return {
        bucket: createBucket(),
        program: createProgram(program_internal),
      };
    },
  });
};

const createQuoteFactory = (quote_bucket_data: any) => {
  const quoteFactory = () => () => () => () => () => () => () =>
    Promise.resolve(createQuote(quote_bucket_data));

  return <AsyncQuoteFactory>(<unknown>quoteFactory);
};

const createQuoteFactoryWithQuote = (quote_bucket_data: any) => {
  const quote = createQuote(quote_bucket_data);
  const quoteFactory = () => () => () => () => () => () => () =>
    Promise.resolve(quote);

  return {
    quote_factory: <AsyncQuoteFactory>(<unknown>quoteFactory),
    quote,
  };
};

const createBucket = () => {
  return <QuoteDataBucket>(<unknown>{
    on() {},
    getData() {
      return {};
    },
    getDataByName: () => 'foo',
  });
};

const createProgram = (internal = {}) => {
  return <Program>(<unknown>{
    initQuote() {},
    change() {},
    classify(_data: CommonObject) {},
    getId() {
      return 1;
    },
    getFirstStepId() {
      return 2;
    },
    meta: {
      groups: {},
    },
    internal,
  });
};

const createQuote = (quote_bucket_data = {}) => {
  return <ServerSideQuote>(<unknown>{
    setData(_data: any) {},
    getId() {
      return 123;
    },
    setMetadata: () => {},
    getBucket: () => {
      return {
        getData: () => {
          return quote_bucket_data;
        },
      };
    },
    getFieldState: () => {
      return {};
    },
    getStartDate: () => 123,
    setTopSavedStepId: (_id: number) => {
      (this || ({} as {topSavedStepId: number})).topSavedStepId = _id;
    },
    getTopSavedStepId: () => {
      return (this || ({} as {topSavedStepId: number})).topSavedStepId;
    },
    getDataByName: (_field_name: string) => {
      if (_field_name === 'agent_email') {
        return ['test_agent_email@test.com'];
      } else {
        return [];
      }
    },
    classify: () => {},
  });
};

const createDao = () => {
  const dao = <MongoServerDao>(<unknown>{
    init(_callback: Function) {},
    getNextQuoteId() {
      return new Promise(resolve => resolve(1));
    },

    saveQuote(_quote: unknown, _class_data: unknown, success: any) {
      return success();
    },

    saveQuoteMetaDirect(_quote: unknown, _new_meta: unknown, success: any) {
      return success();
    },

    saveQuoteMeta(_quote: unknown, _new_meta: unknown, success: any) {
      return success();
    },

    publishQuote(_quote: unknown) {
      return Promise.resolve();
    },

    saveQuoteState(_quote: ServerSideQuote, success: any) {
      return success();
    },
  });

  return dao;
};

const getCreateValidator = (
  missing_required: any[] = [],
  failed_assertions: any[] = [],
  triggered_events: any[] = []
) => {
  return (program: Program, _step_simulator: StepSimulator) => {
    return <Validator>(<unknown>{
      validate() {
        program.change(
          <PositiveInteger>1,
          'foo',
          <DelayedStagingBucket>{},
          <CmatchData>{},
          {},
          () => {}
        );

        return {missing_required, failed_assertions, triggered_events};
      },

      validateField() {
        return {missing_required, failed_assertions, triggered_events};
      },

      validateStep() {
        return {missing_required, failed_assertions, triggered_events};
      },
    });
  };
};

const createRatingService = () => {
  return <RatingService>(<unknown>{
    request: () => {
      return {};
    },
  });
};

const createManager = () => {
  return <IndicationCreationManager>(<unknown>{
    quoteMaker: () => ({}),
    initialQuoteSave: () => () => () => () => () => () => {},
    invokeServerDapis: () => () => () => () => () => () => () => () => {},
  });
};

const createFeatureFlag = (enabled = false) => {
  return <FeatureFlag>(<unknown>{
    isEnabled: () => Promise.resolve(enabled),
  });
};

const createUserSession =
  (is_internal = true) =>
  (_ = ''): Promise<UserSession> => {
    return Promise.resolve(<UserSession>(<unknown>{
      agentId: () => '',
      agentName: () => '',
      agentEntityId: () => '',
      agencyNumber: () => '',
      isInternal: () => is_internal,
      userName: () => 'foo',
    }));
  };

const createStepSimulator = () => {
  return <StepSimulator>(<unknown>{
    each: () => {},
  });
};

const createStepDataSetter =
  (missing_required: string[] = [], failed_assertions: string[] = []) =>
  (_p: Program) =>
    <StepDataSetter>(<unknown>{
      setDataByStep: () => {
        _p.change(
          <PositiveInteger>1,
          'foo',
          <DelayedStagingBucket>{},
          <CmatchData>{},
          {},
          () => {}
        );

        return {
          missing_required,
          failed_assertions,
          triggered_events: [],
        };
      },
    });

const createClientQuote = () => <ClientQuote>(<unknown>{
    on: () => {},
    forceClassify: () => {},
  });

const getCreateClientDapiManager = () => <DataApiManager>(<unknown>{
    on: () => {
      return getCreateClientDapiManager();
    },
  });

const createDapiMediator = () => <DataApiMediator>(<unknown>{
    monitor: () => {},
  });

const createDataValidator = (_p: Program) => <DataValidator>(<unknown>{
    validate: () => {},
  });

const getCreateDapiOutputValidator =
  (_bucket: Bucket) =>
  (_error_log: string[]) =>
  (_name: string, _i: number | string, _field_data: UiFieldData[]) => {};

const createCmatch = () => () => <Cmatch>(<unknown>{
    hookClassifier: () => {},
  });

const mockPollForRates = (
  _start_timestamp: number,
  _reloadQuote: () => Promise<E.Either<unknown, ServerSideQuote>>
) => Promise.resolve({});

const createSutCreate = (dependencies: CommonObject = {}) => {
  const {
    createClientDapiManager,
    createDapiOutputValidator,
    createSession,
    createValidator,
    dao,
    emitter,
    feature_flag,
    http_client,
    manager,
    missing_required,
    failed_assertions,
    program_factory,
    quote_factory,
    rating_service,
    is_internal,
    quote_bucket_data,
    program_internal,
    pollForRates,
  } = dependencies;

  const creation_factory: IndicationCreateDeps = {
    createValidator: (p: Program, step_simulator: StepSimulator) => {
      return createValidator
        ? (<ValidatorFactory>createValidator)(p, step_simulator)
        : getCreateValidator()(p, step_simulator);
    },

    createClientQuote,
    createClientDapiManager:
      <any>createClientDapiManager ?? getCreateClientDapiManager,
    createDapiMediator,
    createStepDataSetter: createStepDataSetter(
      <string[]>missing_required,
      <string[]>failed_assertions
    ),
    createStepSimulator,
    createCmatch,
    createDataValidator,
    createDapiOutputValidator:
      <any>createDapiOutputValidator ?? getCreateDapiOutputValidator,
  };

  return Sut.create(
    <IndicationCreationManager>manager ?? createManager(),
    <EventEmitter>emitter ?? createEventEmitter(),
    <HttpClient>http_client ?? createHttpClient(),
    <FeatureFlag>feature_flag ?? createFeatureFlag(),
    <(id: string) => Promise<UserSession>>createSession ??
      createUserSession(is_internal as boolean),
    creation_factory,
    <ProgramFactory>program_factory ??
      createProgramFactory(program_internal as any),
    <AsyncQuoteFactory>quote_factory ?? createQuoteFactory(quote_bucket_data),
    <MongoServerDao>dao ?? createDao(),
    <RatingService>rating_service ?? createRatingService(),
    <InitExistingDocument>(
      (<unknown>(
        (() => () => () => () => (quote: ServerSideQuote) => () => () =>
          TE.tryCatch<Error, ServerSideQuote>(
            () => Promise.resolve(quote),
            E.toError
          ))
      ))
    ),
    <
      (
        _start_timestamp: number,
        _reloadQuote: () => Promise<E.Either<unknown, ServerSideQuote>>
      ) => Promise<RateResult>
    >pollForRates ?? mockPollForRates,
    '',
    {
      liza: {
        '1': {
          is_active: true,
          key: Buffer.from(randomBytes(32).toString('base64')),
        },
      },
    }
  );
};

const createValidRequestPayload = () => {
  return {
    body: {
      bucket: {},
      session: {
        agentId: <PositiveInteger>900000,
        agentName: 'foo',
        agentEntityId: <PositiveInteger>1234567,
        agencyNumber: 'AGT12345',
      },
    },
    query: {
      callback: 'https://foo.com',
    },
  };
};
