/**
 *  Tests for Service Controller
 *
 *  Copyright (C) 2010-2023 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {EventEmitter} from 'events';
import {expect} from 'chai';
import {mockReq, mockRes} from 'sinon-express-mock';
import {ServiceController} from '../../../src/dullahan/controllers/ServiceController';
import {createUserSession} from '../../../src/server/ServerFactory';
import {UserSession} from '../../../src/server/request/UserSession';
import {Program, QuestionTypes} from '../../../src/program/Program';
import {MongoServerDao} from '../../../src/server/db/MongoServerDao';
import {randomBytes} from 'crypto';
import {BucketDataValidator} from '../../../src/validate/BucketDataValidator';
import SessionError from '../../../src/error/SessionError';
import {CopyRequestBody} from '../../../src/dullahan/validate/copy_validator';
import {PositiveInteger} from '../../../src/numeric';
import {
  initExistingDocument,
  initNewDocument,
} from '../../../src/server/quote/loader';
import * as TE from 'fp-ts/TaskEither';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {CopyCleaner} from '../../../src/program/ProgramLoader';
import {InitRequestBody} from '../../../src/dullahan/validate/init_validator';
import {InitRateRequestBody} from '../../../src/dullahan/validate/init_rate_validator';
import {BucketData} from '../../../src/bucket/Bucket';
import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';
import {FeatureFlag} from '../../../src/system/flags/FeatureFlag';

const program_qtypes = '';
const program_static_select_questions = {foo: true};
const program_qdata = {foo: {'1': 'Yes'}};
const program_defaults = {foo: '1'};

describe('ServiceController', () => {
  const fake_keks = {
    liza: {
      '1': {
        is_active: true,
        key: Buffer.from(randomBytes(32).toString('base64')),
      },
    },
  };
  const fakeFeatureFlag: FeatureFlag = {
    isEnabled: (_flag: string) => Promise.resolve(true),
    close: () => {},
  };
  describe('init', () => {
    it('Should set bucket values to a cleaned quote', async () => {
      const program = 'somebodyoncetoldme';
      const agency_number = 'theworldisgonnarollme';
      const request_bucket = {
        foo: ['bar'],
        baz: ['foobarbaz'],
      };
      const req = mockReq({
        body: createValidInitQuote({
          program,
          bucket: request_bucket,
        }),
      });
      const res = mockRes({});

      const clean_data = {clean: ['some-data']};
      const initial_data = {unclean: ['other-data']};
      // Ensure we're calling setValues with the clean_data, not the initial data sent in
      let set_values_called = false;
      const server_side_quote = createServerSideQuote(program);
      server_side_quote.getBucket = () => <QuoteDataBucket>(<unknown>{
          getData: () => initial_data,
          clear: () => {},
          setValues: (data: BucketData) => {
            set_values_called = true;
            expect(data).to.deep.equal({...clean_data, ...request_bucket});
          },
        });
      server_side_quote.getAgencyNumber = () => agency_number;

      const createNewDocument: typeof initNewDocument =
        () => () => () => () => () => () => () =>
          TE.of(server_side_quote);

      // Copy cleaner sets the data
      let called_copy_cleaner = false;
      const copyCleaner =
        (): CopyCleaner =>
        (bucket_data, session, agency, request, select_question_metadata) => {
          expect(bucket_data).to.deep.equal(initial_data);
          expect(request).to.deep.equal(request_bucket);
          expect(session.isLoggedIn()).to.be.true;
          expect(agency).to.equal(agency_number);
          expect(select_question_metadata.question_defaults).to.deep.equal(
            program_defaults
          );
          expect(
            select_question_metadata.static_select_questions
          ).to.deep.equal(program_static_select_questions);
          expect(select_question_metadata.valid_select_values).to.deep.equal(
            program_qdata
          );
          called_copy_cleaner = true;
          return clean_data;
        };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        copyCleaner,
        createNewDocument,
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.init(req, res);
      // We mock the function that would return a 200
      // so check that we're not returning anything
      expect(res.status.notCalled).to.be.true;
      expect(called_copy_cleaner).to.be.true;
      expect(set_values_called).to.be.true;
    });

    it('Should exit gracefully on a successful init', async () => {
      const program = 'somebodyoncetoldme';
      const req = mockReq({body: createValidInitQuote({program})});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor(program),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.init(req, res);
      // We mock the function that would return a 200
      // so check that we're not returning anything
      expect(res.status.notCalled).to.be.true;
    });

    it('response with 422 code when body not valid', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.init(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(422)).to.be.true;
    });

    it('response with 401 code when a Session Error is thrown', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const createErrorSession = () => {
        throw new SessionError('foo');
      };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        createErrorSession,
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.init(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(401)).to.be.true;
    });

    it('response with 500 code when general Error is thrown', async () => {
      const program = 'foo';
      const req = mockReq({body: createValidInitQuote({program})});
      const res = mockRes({});

      // creating invalid function to throw general error:
      const getBrokenSessionBuilder = () => () => {
        throw new Error('foo');
      };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getBrokenSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor(program),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.init(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(500)).to.be.true;
    });

    it('response with 500 code when program cannot be loaded', async () => {
      const req = mockReq({body: createValidInitQuote()});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        () => {
          throw new Error('foo');
        },
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.init(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(500)).to.be.true;
    });
  });

  describe('copy', () => {
    it('Should set bucket values to a cleaned quote', async () => {
      const program = 'somebodyoncetoldme';
      const agency_number = 'theworldisgonnarollme';
      const request_bucket = {
        foo: ['bar'],
        baz: ['foobarbaz'],
      };
      const req = mockReq({
        body: createValidCopyQuote({
          program,
          bucket: request_bucket,
        }),
      });
      const res = mockRes({});

      const clean_data = {clean: ['some-data']};
      const initial_data = {unclean: ['other-data']};
      // Ensure we're calling setValues with the clean_data, not the initial data sent in
      let set_values_called = false;
      const server_side_quote = createServerSideQuote(program);
      server_side_quote.getBucket = () => <QuoteDataBucket>(<unknown>{
          getData: () => initial_data,
          clear: () => {},
          setValues: (data: BucketData) => {
            set_values_called = true;
            expect(data).to.deep.equal({...clean_data, ...request_bucket});
          },
        });
      server_side_quote.getAgencyNumber = () => agency_number;

      const createNewDocument: typeof initNewDocument =
        () => () => () => () => () => () => () =>
          TE.of(server_side_quote);

      const createExistingDocument: typeof initExistingDocument =
        () => () => () => () => () => () => () =>
          TE.of(server_side_quote);

      // Copy cleaner sets the data
      let called_copy_cleaner = false;
      const copyCleaner =
        (): CopyCleaner =>
        (bucket_data, session, agency, request, select_question_metadata) => {
          expect(bucket_data).to.deep.equal(initial_data);
          expect(request).to.deep.equal(request_bucket);
          expect(session.isLoggedIn()).to.be.true;
          expect(agency).to.equal(agency_number);
          expect(select_question_metadata.question_defaults).to.deep.equal(
            program_defaults
          );
          expect(
            select_question_metadata.static_select_questions
          ).to.deep.equal(program_static_select_questions);
          expect(select_question_metadata.valid_select_values).to.deep.equal(
            program_qdata
          );
          called_copy_cleaner = true;
          return clean_data;
        };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        copyCleaner,
        createNewDocument,
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.copy(createExistingDocument)(req, res);
      // We mock the function that would return a 200
      // so check that we're not returning anything
      expect(res.status.notCalled).to.be.true;
      expect(called_copy_cleaner).to.be.true;
      expect(set_values_called).to.be.true;
    });

    it('Should exit gracefully on a successful copy', async () => {
      const program = 'somebodyoncetoldme';
      const req = mockReq({body: createValidCopyQuote({program})});
      const res = mockRes({});
      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor(program),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.copy(createExistingQuoteCtor(program))(req, res);
      // We mock the function that would return a 200
      // so check that we're not returning anything
      expect(res.status.notCalled).to.be.true;
    });

    it('response with 422 code when body not valid', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.copy(createExistingQuoteCtor('bar'))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(422)).to.be.true;
    });

    it('response with 401 code when a Session Error is thrown', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const createErrorSession = () => {
        throw new SessionError('foo');
      };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        createErrorSession,
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.copy(createExistingQuoteCtor('bar'))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(401)).to.be.true;
    });

    it('response with 500 code when general Error is thrown', async () => {
      const program = 'foo';
      const req = mockReq({
        body: createValidCopyQuote({program}),
      });
      const res = mockRes({});

      // creating invalid function to throw general error:
      const getBrokenSessionBuilder = () => () => {
        throw new Error('foo');
      };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getBrokenSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor(program),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.copy(createExistingQuoteCtor(program))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(500)).to.be.true;
    });

    it('response with 500 code when program cannot be loaded', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        () => {
          throw new Error('foo');
        },
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.copy(createExistingQuoteCtor('foo'))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(500)).to.be.true;
    });
  });

  describe('initRate', () => {
    it('Should set bucket values to a cleaned quote', async () => {
      const program = 'somebodyoncetoldme';
      const agency_number = 'theworldisgonnarollme';
      const request_bucket = {
        foo: ['bar'],
        baz: ['foobarbaz'],
      };
      const req = mockReq({
        body: createValidInitRateQuote({
          program,
          bucket: request_bucket,
        }),
      });
      const res = mockRes({});

      const clean_data = {clean: ['some-data']};
      const initial_data = {unclean: ['other-data']};
      // Ensure we're calling setValues with the clean_data, not the initial data sent in
      let set_values_called = false;
      const server_side_quote = createServerSideQuote(program);
      server_side_quote.getBucket = () => <QuoteDataBucket>(<unknown>{
          getData: () => initial_data,
          clear: () => {},
          setValues: (data: BucketData) => {
            set_values_called = true;
            expect(data).to.deep.equal({...clean_data, ...request_bucket});
          },
        });
      server_side_quote.getAgencyNumber = () => agency_number;

      const createNewDocument: typeof initNewDocument =
        () => () => () => () => () => () => () =>
          TE.of(server_side_quote);

      const createExistingDocument: typeof initExistingDocument =
        () => () => () => () => () => () => () =>
          TE.of(server_side_quote);

      // Copy cleaner sets the data
      let called_copy_cleaner = false;
      const copyCleaner =
        (): CopyCleaner =>
        (bucket_data, session, agency, request, select_question_metadata) => {
          expect(bucket_data).to.deep.equal(initial_data);
          expect(request).to.deep.equal(request_bucket);
          expect(session.isLoggedIn()).to.be.true;
          expect(agency).to.equal(agency_number);
          expect(select_question_metadata.question_defaults).to.deep.equal(
            program_defaults
          );
          expect(
            select_question_metadata.static_select_questions
          ).to.deep.equal(program_static_select_questions);
          expect(select_question_metadata.valid_select_values).to.deep.equal(
            program_qdata
          );
          called_copy_cleaner = true;
          return clean_data;
        };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        copyCleaner,
        createNewDocument,
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.initRate(createExistingDocument)(req, res);
      // We mock the function that would return a 200
      // so check that we're not returning anything
      expect(res.status.notCalled).to.be.true;
      expect(called_copy_cleaner).to.be.true;
      expect(set_values_called).to.be.true;
    });

    it('Should exit gracefully on a successful copy', async () => {
      const program = 'somebodyoncetoldme';
      const req = mockReq({body: createValidInitRateQuote({program})});
      const res = mockRes({});
      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor(program),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.initRate(createExistingQuoteCtor(program))(req, res);
      // We mock the function that would return a 200
      // so check that we're not returning anything
      expect(res.status.notCalled).to.be.true;
    });

    it('response with 422 code when body not valid', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.initRate(createExistingQuoteCtor('bar'))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(422)).to.be.true;
    });

    it('response with 401 code when a Session Error is thrown', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const createErrorSession = () => {
        throw new SessionError('foo');
      };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        createErrorSession,
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.initRate(createExistingQuoteCtor('bar'))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(401)).to.be.true;
    });

    it('response with 500 code when general Error is thrown', async () => {
      const program = 'foo';
      const req = mockReq({
        body: createValidInitRateQuote({program}),
      });
      const res = mockRes({});

      // creating invalid function to throw general error:
      const getBrokenSessionBuilder = () => () => {
        throw new Error('foo');
      };

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getBrokenSessionBuilder(),
        getProgramLoader(),
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor(program),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.initRate(createExistingQuoteCtor(program))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(500)).to.be.true;
    });

    it('response with 500 code when program cannot be loaded', async () => {
      const req = mockReq({});
      const res = mockRes({});

      const Sut = new ServiceController(
        dataValidator(),
        getEventEmitter(),
        getSessionBuilder(),
        () => {
          throw new Error('foo');
        },
        getProgramCopyCleanerLoader(),
        createNewDocumentCtor('foo'),
        getDao(),
        fakeFeatureFlag,
        fake_keks
      );

      await Sut.initRate(createExistingQuoteCtor('foo'))(req, res);

      expect(res.status.callCount).to.equal(1);
      expect(res.status.calledWith(500)).to.be.true;
    });
  });
});

const getSessionBuilder = (logged_in: boolean = true) =>
  <typeof createUserSession>(
    (<unknown>(
      (async (_: string) => <Promise<UserSession>>(<unknown>Promise.resolve({
          isLoggedIn: () => logged_in,
          userName: () => '',
          agencyNumber: () => '',
          userRoles: () => '',
          isInternal: () => true,
          agentEntityId: () => '',
        })))
    ))
  );

const getEventEmitter = () => <EventEmitter>(<unknown>{
    emit: () => {},
  });

const getProgramLoader = () => () => () => <Program>(<unknown>{
    meta: {
      qtypes: program_qtypes,
      staticSelectQuestions: program_static_select_questions,
      qdata: program_qdata,
    },
    defaults: program_defaults,
  });

const getProgramCopyCleanerLoader =
  (bucket_data: BucketData = {}) =>
  (): CopyCleaner =>
  () =>
    bucket_data;

const getDao = () =>
  <MongoServerDao>(<unknown>{saveQuote: () => {}, pullQuote: () => {}});

const dataValidator = () => (_: QuestionTypes) =>
  <BucketDataValidator>{
    validate: (data: Record<string, any>) => data,
    format: (data: Record<string, any>) => data,
  };

const createValidCopyQuote = (
  override: Partial<CopyRequestBody> = {}
): CopyRequestBody => ({
  quote_id: 1 as PositiveInteger,
  policy_id: 'foo',
  session_id: 'bar',
  program: 'baz',
  bucket: {},
  origin: 'foo2',
  ...override,
});

const createValidInitRateQuote = (
  override: Partial<InitRateRequestBody> = {}
): InitRateRequestBody => ({
  quote_id: 1 as PositiveInteger,
  policy_id: 'foo',
  session_id: 'bar',
  program: 'baz',
  bucket: {},
  origin: 'foo2',
  ...override,
});

const createValidInitQuote = (
  override: Partial<InitRequestBody> = {}
): InitRequestBody => ({
  policy_id: 'foo',
  session_id: 'bar',
  program: 'baz',
  bucket: {},
  origin: 'foo2',
  ...override,
});

const createServerSideQuote = (program_id: string) => <ServerSideQuote>(<
    unknown
  >{
    isAuthorized: () => true,
    getProgramId: () => program_id,
    getId: () => 123,
    getBucket: () => ({
      getData: () => {},
      clear: () => {},
      setValues: (data: BucketData) => data,
    }),
    getAgencyNumber: () => 1,
    getProgram: () => getProgramLoader()()(),
  });

const createNewDocumentCtor =
  (program_id: string): typeof initNewDocument =>
  () =>
  () =>
  () =>
  () =>
  () =>
  () =>
  () =>
    TE.of(createServerSideQuote(program_id));

const createExistingQuoteCtor =
  (program_id: string): typeof initExistingDocument =>
  () =>
  () =>
  () =>
  () =>
  () =>
  () =>
  () =>
    TE.of(createServerSideQuote(program_id));
