/**
 *  Tests for System Controller
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {Db} from 'mongodb';
import {EventEmitter} from 'events';
import {expect} from 'chai';
import {mockReq, mockRes} from 'sinon-express-mock';
import {system as Sut} from '../../../src/dullahan/controllers/SystemController';

describe('SystemController', () => {
  describe('healthcheck', () => {
    it('response with 204 code when healthy', () => {
      const req = mockReq({});
      const res = mockRes({});

      return Sut.healthcheck(getEventEmitter())(getDb())(req, res).then(_ => {
        expect(res.status.withArgs(204).callCount).to.equal(1);
        expect(res.status.withArgs(503).callCount).to.equal(0);
      });
    });

    it('fails when the db is not ok', () => {
      const req = mockReq({});
      const res = mockRes({});
      const db = getDb();
      db.command = () => new Promise(res => res({ok: 0}));

      return Sut.healthcheck(getEventEmitter())(db)(req, res).then(_ => {
        expect(res.status.withArgs(204).callCount).to.equal(0);
        expect(res.status.withArgs(503).callCount).to.equal(1);
      });
    });

    it('fails when a misc error occurs', () => {
      const req = mockReq({});
      const res = mockRes({});
      const db = getDb();
      db.command = () =>
        new Promise(_ => {
          throw new Error();
        });

      return Sut.healthcheck(getEventEmitter())(db)(req, res).then(_ => {
        expect(res.status.withArgs(204).callCount).to.equal(0);
        expect(res.status.withArgs(503).callCount).to.equal(1);
      });
    });
  });
});

const getEventEmitter = () => <EventEmitter>(<unknown>{
    emit: () => {},
  });

const getDb = () => <Db>(<unknown>{
    command: () => new Promise(res => res({ok: 1})),
  });
