/**
 *  Tests for IndicationCreationManager
 *
 *  Copyright (C) 2010-2023 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {IndicationCreationManager as Sut} from '../../../src/dullahan/indication/IndicationCreationManager';
import {DapiFactory} from '../../../src/dullahan/dapi/DapiFactory';
import {Program} from '../../../src/program/Program';
import {SavedDelta} from '../../../src/bucket/delta';
import {StagingBucket} from '../../../src/bucket/StagingBucket';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {UserSession} from '../../../src/server/request/UserSession';
import {Bucket, BucketData, EmptyBucketData} from '../../../src/bucket/Bucket';
import {EventEmitter} from 'events';
import {MongoServerDao} from '../../../src/server/db/MongoServerDao';
import {PriorityLog} from '../../../src/server/log/PriorityLog';
import {expect, use as chai_use} from 'chai';

chai_use(require('chai-as-promised'));

let data: CommonObject | null = null;
let quoteSetMetaDataCount = 0;

beforeEach(function () {
  quoteSetMetaDataCount = 0;
  data = null;
});

describe('IndicationCreationManager', () => {
  describe('constructor', () => {
    it('returns an IndicationCreationManager', () => {
      const sut = new Sut(
        createStubLogger(),
        createDapiFactory(),
        mockConstructSaveData
      );
      expect(sut).to.be.instanceof(Sut);
    });
  });

  describe('invokeServerDapis', () => {
    it('no dapis do no work', async () => {
      const sut = new Sut(
        createStubLogger(),
        createDapiFactory(),
        mockConstructSaveData
      );

      // failed DAPIs emit an event
      const emitter = createEmitter();
      emitter.on('dapi-failure', () =>
        expect.fail('Should not have emitted event')
      );

      await sut.invokeServerDapis(createProgram())(createQuote())(
        createSession()
      )('sess_id')(createBucket())(emitter)(createDao())('');
      // 1 for the last setMetaData call with meta_update
      expect(quoteSetMetaDataCount).to.equal(1);
    });

    (
      [
        [
          'one dapi that fails',
          () => ({
            dapis: [Promise.reject()],
            meta_update: {},
          }),
          0,
          1,
        ],
        [
          'many dapis that fail',
          () => ({
            dapis: Array(10).fill(new Promise((_, reject) => reject())),
            meta_update: {},
          }),
          0,
          10,
        ],
        [
          'one dapi that succeeds',
          () => ({
            dapis: [Promise.resolve({data: {foo: ['1']}, delta: {bar: ['0']}})],
            meta_update: {},
          }),
          1,
          0,
        ],
        [
          'many dapis that succeed',
          () => ({
            dapis: Array(10).fill(
              Promise.resolve({data: {foo: ['1']}, delta: {bar: ['0']}})
            ),
            meta_update: {},
          }),
          10,
          0,
        ],
        [
          'mix of successful and failed dapis',
          () => ({
            dapis: [
              new Promise((_, reject) => reject()),
              Promise.resolve({data: {foo: ['1']}, delta: {bar: ['0']}}),
              new Promise((_, reject) => reject()),
              Promise.resolve({data: {foo: ['1']}, delta: {bar: ['0']}}),
            ],
            meta_update: {},
          }),
          2,
          2,
        ],
      ] as [
        string,
        () => {dapis: any[]; meta_update: EmptyBucketData},
        number,
        number
      ][]
    ).forEach(([label, dapi_results, successes, failures]) => {
      it(label as string, async () => {
        const sut = new Sut(
          createStubLogger(),
          createDapiFactory(dapi_results),
          mockConstructSaveData
        );

        let saveCount = 0;
        let eventCount = 0;

        const dao = createDao(() => saveCount++);

        // failed DAPIs emit an event
        const emitter = createEmitter();
        emitter.on('dapi-failure', () => eventCount++);

        await sut.invokeServerDapis(createProgram())(createQuote())(
          createSession()
        )('sess_id')(createBucket())(emitter)(dao)('');
        // The +1 is for the meta_update save at the end
        expect(quoteSetMetaDataCount).to.equal(successes + 1);
        expect(saveCount).to.equal(successes);
        expect(eventCount).to.equal(failures);
      });
    });

    it('rdelta is generated then saved', async () => {
      const delta_given = {foo: ['1']};
      const rdelta_given = <SavedDelta<BucketData>>(<unknown>{foo: ['0']});
      const save_data_given = {baz: ['1']};
      const dapis = [Promise.resolve({data: {foo: ['1']}, delta: delta_given})];

      let construct_save_data_called = false;
      let save_data_called = false;
      let merge_data_called = false;

      const meta_given: EmptyBucketData = {'meta.baz': ['']};
      const meta_dao_expected: EmptyBucketData = {'meta.baz': ['']};
      const meta_quote_expected: EmptyBucketData = {baz: ['']};

      const sut = new Sut(
        createStubLogger(),
        createDapiFactory(() => ({dapis, meta_update: meta_given})),
        (_: ServerSideQuote, delta: BucketData) => {
          expect(delta).to.deep.equal(delta_given);

          construct_save_data_called = true;

          return {
            save_data: save_data_given,
            rdelta: rdelta_given,
          };
        }
      );

      const quote = createQuote();
      let metaDataCallCount = 0;
      let final_setMetadata_called = false;

      quote.setMetadata = (meta: EmptyBucketData) => {
        metaDataCallCount++;

        if (metaDataCallCount === 1) {
          expect(meta).to.deep.equal(meta_quote_expected);

          final_setMetadata_called = true;
        }

        return quote;
      };

      const dao = createDao(
        (save_data: any, push_data: any) => {
          expect(save_data).to.deep.equal(save_data_given);
          expect(push_data['rdelta.data']).to.deep.equal(rdelta_given);

          save_data_called = true;
        },
        (data: any) => {
          expect(data).to.deep.equal(meta_dao_expected);

          merge_data_called = true;
        }
      );

      await sut.invokeServerDapis(createProgram())(quote)(createSession())(
        'sess_id'
      )(createBucket())(createEmitter())(dao)('');
      expect(construct_save_data_called, 'save data constructed').to.be.true;
      expect(save_data_called, 'data saved to mongo').to.be.true;
      expect(merge_data_called, 'meta_update saved to mongo').to.be.true;
      expect(final_setMetadata_called, 'final setMetadata call made').to.be
        .true;
    });
  });
});

const createDapiFactory = (
  ret: () => {dapis: any[]; meta_update: EmptyBucketData} = () => ({
    dapis: [],
    meta_update: {},
  })
) => <DapiFactory>(<unknown>{
    getDapis: () => () => () => () => () => () => ret,
  });

const mockConstructSaveData = (_: ServerSideQuote, __: BucketData) => ({
  save_data: {},
  rdelta: <SavedDelta<BucketData>>{},
});

const createProgram = () => {
  return <Program>{
    initQuote(_bucket: StagingBucket, _store_only: boolean): void {},
    meta: {groups: {}},
    groupExclusiveFields: {},
  };
};

const createQuote = () => <ServerSideQuote>(<unknown>{
    setMetadata: (_: unknown) => {
      quoteSetMetaDataCount++;
    },
    getId: () => 123,
  });
const createSession = () => <UserSession>{};

const createEmitter = () => new EventEmitter();

const createBucket = () => {
  return <Bucket>{
    setValues(_data: Record<string, any>) {
      data = _data;
      return this;
    },

    getData(): Record<string, any> | null {
      return data;
    },
  };
};

const createDao = (
  saveQuoteCallback: Function = () => {},
  mergeDataCallback: Function = () => {}
) => {
  const dao = <MongoServerDao>(<unknown>{
    init(_callback: Function) {},
    getNextQuoteId() {
      return new Promise(resolve => resolve(1));
    },

    saveQuote(
      _quote: unknown,
      _class_data: unknown,
      success: any,
      _failure: any,
      save_data: any,
      push_data: any
    ) {
      saveQuoteCallback(save_data, push_data);
      return success();
    },

    mergeData(_quote: unknown, data: unknown, success: any, _failure: any) {
      mergeDataCallback(data);
      return success();
    },

    saveQuoteMetaDirect(_quote: unknown, _new_meta: unknown, success: any) {
      return success();
    },

    saveQuoteMeta(_quote: unknown, _new_meta: unknown, success: any) {
      return success();
    },

    publishQuote(_quote: unknown) {
      return Promise.resolve();
    },

    saveQuoteState(_quote: ServerSideQuote, success: any) {
      return success();
    },
  });

  return dao;
};

function createStubLogger(): PriorityLog {
  const logger = {
    log: () => logger,
  };

  return <PriorityLog>(<unknown>logger);
}
