/**
 *  Tests for Step Simulator
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {Cmatch} from '../../../src/client/Cmatch';
import {DataApiManager, PendingDapi} from '../../../src/dapi/DataApiManager';
import {DelayedStagingBucket} from '../../../src/bucket/DelayedStagingBucket';
import {PositiveInteger, ts_ctor} from '../../../src/numeric';
import {Program} from '../../../src/program/Program';
import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {StepDataSetter as Sut} from '../../../src/dullahan/step/StepDataSetter';
import {Validator} from '../../../src/dullahan/validate/Validator';
import {expect} from 'chai';
import {
  StepSimulator,
  StepSummary,
} from '../../../src/dullahan/step/StepSimulator';

describe('StepDataSetter', () => {
  describe('#setDataByStep', () => {
    it('records current step id', () => {
      let result: unknown[] = [];
      const expected: unknown = [
        {id: 1, field: 'foo'},
        {id: 1, type: 'program#submit'},
        {id: 1, type: 'program#forward'},
        {id: 2, field: 'bar'},
        {id: 2, type: 'program#submit'},
        {id: 2, type: 'program#forward'},
      ];

      const sut = getSut();

      expect(sut.getCurrentStep()).to.eq(0);
      const program = getProgram(result);

      return sut
        .setDataByStep(
          getSourceBucket(),
          getStagingBucket(),
          getQuote(),
          getStepSimulator(),
          getDapiManager(),
          getValidator(program),
          getCmatch(),
          ts_ctor
        )
        .then(() => {
          expect(result).to.deep.eq(expected);
        });
    });

    it('sets data for each step', () => {
      let result: unknown[] = [];
      const sut = getSut();
      const program = getProgram(result);

      return sut
        .setDataByStep(
          getSourceBucket(),
          getStagingBucket(),
          getQuote(),
          getStepSimulator(),
          getDapiManager(),
          getValidator(program),
          getCmatch(),
          ts_ctor
        )
        .then(() => {
          expect(sut.getCurrentStep()).to.eq(2);
        });
    });

    it('only sets data for applicable steps', () => {
      let result: unknown[] = [];
      const expected: unknown = [
        {id: 1, field: 'foo'},
        {id: 1, type: 'program#submit'},
        {id: 1, type: 'program#forward'},
        {id: 3, field: 'baz'},
        {id: 3, type: 'program#submit'},
        {id: 3, type: 'program#forward'},
      ];

      const cmatch = getCmatch();
      cmatch.getMatches = () => ({
        foo: {any: true},
        bar: {any: false},
        baz: {any: true},
      });

      const sut = getSut();

      const step_simulator = getStepSimulator();
      const program = getProgram(result);

      step_simulator.getSteps = () => {
        return <Promise<StepSummary[]>>(<unknown>Promise.resolve([
          {id: 1, fields: ['foo']},
          {id: 3, fields: ['baz']},
        ]));
      };

      return sut
        .setDataByStep(
          getSourceBucket(),
          getStagingBucket(),
          getQuote(),
          step_simulator,
          getDapiManager(),
          getValidator(program),
          cmatch,
          ts_ctor
        )
        .then(() => {
          expect(result).to.deep.eq(expected);
        });
    });

    it('hits timeout', () => {
      const sut = getSut();
      let count = 0;
      const expected_count = 2;
      const manager = getDapiManager();

      const stubbed_ts_ctor = () => {
        return <UnixTimestamp>(count ? 20000 : 0);
      };

      manager.getPendingApiCalls = () => {
        count++;

        return {
          foo: {
            uid: <UnixTimestamp>10000,
            name: 'foo',
            index: '0',
          },
        };
      };

      const program = getProgram([]);

      return sut
        .setDataByStep(
          getSourceBucket(),
          getStagingBucket(),
          getQuote(),
          getStepSimulator(),
          manager,
          getValidator(program),
          getCmatch(),
          stubbed_ts_ctor
        )
        .then(() => {
          expect(count).to.eq(expected_count);
        });
    });

    it('defers stepping until dapis return', () => {
      let result: unknown[] = [];

      const pending_dapis: Record<string, number> = {};

      const expected = [
        'program change for foo',
        'pending dapi for foo',
        'pending dapi for foo',
        'pending dapi for foo',
        {id: 1, type: 'program#submit'},
        {id: 1, type: 'program#forward'},
        'program change for bar',
        'pending dapi for bar',
        {id: 2, type: 'program#submit'},
        {id: 2, type: 'program#forward'},
      ];

      const program = getProgram(result);
      const manager = getDapiManager();

      program.change = (_id: unknown, field: string, _, __, ___, ____) => {
        pending_dapis[field] = 0;

        result.push(`program change for ${field}`);
        return null;
      };

      manager.getPendingApiCalls = () => {
        const retries: Record<string, number> = {
          foo: 3,
          bar: 1,
        };

        // Clear completed dapis
        Object.keys(pending_dapis).forEach(name => {
          if (pending_dapis[name] >= retries[name]) {
            delete pending_dapis[name];
          }
        });

        if (Object.keys(pending_dapis).length > 0) {
          result.push(
            Object.keys(pending_dapis).map(
              name => `pending dapi for ${name}`
            )[0]
          );
        }

        const pending: Record<string, PendingDapi> = {};

        for (let name in pending_dapis) {
          pending_dapis[name]++;
          pending[name] = {
            uid: <UnixTimestamp>new Date().valueOf(),
            name,
            index: '0',
          };
        }

        return pending;
      };

      const sut = new Sut(() => Promise.resolve());

      return sut
        .setDataByStep(
          getSourceBucket(),
          getStagingBucket(),
          getQuote(),
          getStepSimulator(),
          manager,
          getValidator(program),
          getCmatch(),
          ts_ctor
        )
        .then(() => {
          expect(result).to.deep.eq(expected);
        });
    });
  });
});

const getSourceBucket = () => <QuoteDataBucket>(<unknown>{
    getDataByName: (data: string) => data,
  });

const getStagingBucket = () => <DelayedStagingBucket>(<unknown>{commit() {}});

const getDapiManager = () => <DataApiManager>(<unknown>{
    commit() {},
    getPendingApiCalls() {
      return {};
    },
  });

const getQuote = () => <ServerSideQuote>(<unknown>{});

const getValidator = (p: Program) => <Validator>(<unknown>{
    validateField: (_: any, __: any, id: PositiveInteger, field: string) => {
      p.change(id, field, <any>{}, {}, {}, <any>{});
    },
    validateStep: (_: any, __: any, ___: any, id: PositiveInteger) => {
      p.submit(id, <any>{}, {});
      p.forward(id, <any>{}, {}, <any>{});
    },
  });

const getStepSimulator = () => <StepSimulator>(<unknown>{
    getSteps: () => {
      return Promise.resolve([
        {id: 1, fields: ['foo']},
        {id: 2, fields: ['bar']},
      ]);
    },
  });

const getProgram = (result: unknown[]) => <Program>(<unknown>{
    getFirstStepId: () => 1,
    change: (id: unknown, field: unknown) => {
      result.push({id, field});
    },
    submit: (id: unknown) => {
      result.push({id, type: 'program#submit'});
    },
    forward: (id: unknown) => {
      result.push({id, type: 'program#forward'});
    },
  });

const getCmatch = () => <Cmatch>(<unknown>{
    getMatches: () => ({foo: {any: true}, bar: {any: true}}),
  });

const getSut = () => new Sut(() => Promise.resolve());
