/* eslint-disable no-sparse-arrays */
import {DelayedStagingBucket} from '../../../src/bucket/DelayedStagingBucket';
import {Program} from '../../../src/program/Program';
import {expect} from 'chai';
import {
  DapiUpdateError,
  validateDapiOutput,
} from '../../../src/dullahan/validate/DapiValidator';

describe('DapiValidator', () => {
  describe('validateDapiOutput', () => {
    const name = 'foo';

    [
      {
        index: 0,
        bucket: {},
        expected: {foo: ['Foo City']},
      },
      {
        index: 1,
        bucket: {},
        expected: {foo: [, 'Foo City']},
      },
    ].forEach(({index, expected, bucket}) => {
      it(`fills an undefined bucket value, single result - ${index}`, () => {
        const b = getBucket(bucket);

        validateDapiOutput(b)(getProgram())(getErrorLog())(
          name,
          index,
          getSingleResult(),
          {}
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        index: 0,
        bucket: {foo: ['FOO CITY']},
        expected: {foo: ['Foo City']},
      },
      {
        index: 1,
        bucket: {foo: ['0', 'invalid input']},
        expected: {foo: ['0', 'Foo City']},
      },
    ].forEach(({index, bucket, expected}) => {
      it(`overwrites a defined bucket value, single result - ${index}`, () => {
        const b = getBucket(bucket);

        validateDapiOutput(b)(getProgram())(getErrorLog())(
          name,
          index,
          getSingleResult(),
          {}
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        index: 0,
        bucket: {foo: ['foo city']},
        expected: {foo: ['Foo City']},
      },
      {
        index: 1,
        bucket: {foo: ['0', 'FOO City']},
        expected: {foo: ['0', 'Foo City']},
      },
      {
        index: 1,
        bucket: {foo: ['0', 'Foo City']},
        expected: {foo: ['0', 'Foo City']},
      },
    ].forEach(({index, bucket, expected}) => {
      it(`overwrites a defined bucket value, multiple result - ${index} - ${bucket.foo[0]}`, () => {
        const b = getBucket(bucket);

        validateDapiOutput(b)(getProgram())(getErrorLog())(
          name,
          index,
          getMultipleResults(),
          {}
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        name: 'foo',
        index: 0,
        bucket: {foo: ['foo city']},
        expected: {
          foo: ['Foo City'],
          my_value: ['123 foo'],
          your_value: ['foo'],
        },
      },
    ].forEach(({name, index, bucket, expected}) => {
      it(`fills mapped values - ${name}`, () => {
        const b = getBucket(bucket);
        const map = {my_value: 'other_field', your_value: 'other_field_2'};

        validateDapiOutput(b)(getProgram())(getErrorLog())(
          name,
          index,
          getSingleResult(),
          map
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        name: 'foo',
        index: 0,
        bucket: {foo: ['foo city']},
        expected: {
          foo: ['Foo City'],
          my_value: ['123 foo'],
          your_value: ['foo'],
        },
      },
      {
        name: 'bar',
        index: 1,
        bucket: {bar: ['0', 'bar City'], my_value: ['4', '234 baR']},
        expected: {
          bar: ['0', 'Bar City'],
          my_value: ['4', '234 bar'],
          your_value: [, 'bar'],
        },
      },
      {
        name: 'baz',
        index: 1,
        bucket: {
          baz: ['0', 'BAz City'],
          my_value: [, '345 bAz'],
          your_value: [, 'BAZ'],
        },
        expected: {
          baz: ['0', 'Baz City'],
          my_value: [, '345 baz'],
          your_value: [, 'baz'],
        },
      },
    ].forEach(({name, index, bucket, expected}) => {
      it(`overwrites mapped values - ${name} - ${index} - ignores case`, () => {
        const b = getBucket(bucket);
        const map = {
          my_value: 'other_field',
          your_value: 'other_field_2',
        };

        validateDapiOutput(b)(getProgram())(getErrorLog())(
          name,
          index,
          getMultipleResults(),
          map
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        name: 'foo',
        index: 0,
        bucket: {foo: ['foo city'], my_value: ['0']},
        expected: {
          foo: ['Foo City'],
          my_value: ['123 foo'],
          your_value: ['foo'],
        },
      },
    ].forEach(({name, index, bucket, expected}) => {
      it('overwrites defaulted mapped values - ${name} - ${index}', () => {
        const b = getBucket(bucket);
        const map = {my_value: 'other_field', your_value: 'other_field_2'};

        const program = <Program>(<unknown>{
          defaults: {
            my_value: '0',
          },
        });
        validateDapiOutput(b)(program)(getErrorLog())(
          name,
          index,
          getMultipleResults(),
          map
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        desc: 'empty bucket',
        index: 0,
        bucket: {},
        expected: {foo: ['Foo City']},
      },
      {
        desc: 'empty foo in bucket',
        index: 0,
        bucket: {foo: []},
        expected: {foo: ['Foo City']},
      },
      {
        desc: 'foo = FooCity',
        index: 0,
        bucket: {foo: ['FooCity']},
        expected: {foo: ['Foo City']},
      },
      {
        desc: 'foo = Fo City',
        index: 1,
        bucket: {foo: ['foo', 'Fo City']},
        expected: {foo: ['foo', 'Foo City']},
      },
    ].forEach(({desc, index, bucket, expected}) => {
      it(`errors when it can't resolve multiple results - ${index} - ${desc}`, () => {
        const b = getBucket(bucket);

        const error_log = getErrorLog();

        validateDapiOutput(b)(getProgram())(error_log)(
          name,
          index,
          getMultipleResults(),
          {}
        );

        expect(error_log).to.be.empty;
        expect(b.getData()).to.deep.equal(expected);
      });
    });

    [
      {
        name: 'foo',
        index: 0,
        bucket: {foo: ['foo city'], my_value: ['invalid']},
        expected: [
          {
            current_value: 'invalid',
            field: 'my_value',
            message: 'Current value prevents new value from being set',
            new_value: '123 foo',
          },
        ],
      },
      {
        name: 'bar',
        index: 1,
        bucket: {
          bar: ['0', 'bar City'],
          my_value: ['0', 'invalid'],
          you_value: ['0', 'invalid'],
        },
        expected: [
          {
            current_value: 'invalid',
            field: 'my_value',
            message: 'Current value prevents new value from being set',
            new_value: '234 bar',
          },
        ],
      },
    ].forEach(({name, index, bucket, expected}) => {
      it(`errors when it can't resolve dapi_map results - ${name} - ${index}`, () => {
        const b = getBucket(bucket);

        const error_log = getErrorLog();
        const map = {my_value: 'other_field', your_value: 'other_field_2'};

        validateDapiOutput(b)(getProgram())(error_log)(
          name,
          index,
          getMultipleResults(),
          map
        );

        expect(error_log).to.deep.equal(expected);
      });
    });

    [
      {
        name: 'foo',
        index: -1,
        bucket: {
          foo: ['foo city', 'foo CitY'],
        },
        expected: {foo: ['Foo City', 'Foo City']},
        map: <Record<string, string>>{},
      },
      {
        name: 'baz',
        index: -1,
        bucket: {
          baz: ['baz city', 'BAz CitY', 'Baz CitY'],
          untouched: [1, 2, 3],
          my_value: ['345 baz', '345 baz', '345 bAz'],
          your_value: ['BAZ', 'baZ', 'BAz'],
        },
        expected: {
          baz: ['Baz City', 'Baz City', 'Baz City'],
          untouched: [1, 2, 3],
          my_value: ['345 baz', '345 baz', '345 baz'],
          your_value: ['baz', 'baz', 'baz'],
        },
        map: {
          my_value: 'other_field',
          your_value: 'other_field_2',
        },
      },
    ].forEach(({name, index, bucket, expected, map}) => {
      it(`handles all indexes - ${name} - ${index}`, () => {
        const b = getBucket(bucket);

        validateDapiOutput(b)(getProgram())(getErrorLog())(
          name,
          index,
          getMultipleResults(),
          map
        );

        expect(b.getData()).to.deep.equal(expected);
      });
    });
  });
});

const getBucket = (data: CommonObject = {}) => {
  return <DelayedStagingBucket>(<unknown>{
    getData: () => data,
    getDataByName: (key: string) => data[key],
    processValues: () => {},
    setValues: (values: CommonObject) => (data = {...data, ...values}),
    commit: () => {},
  });
};

const getSingleResult = () => ({
  'Foo City': {id: 'Foo City', other_field: '123 foo', other_field_2: 'foo'},
  ___empty: false,
});

const getMultipleResults = () => ({
  '': {id: '', other_field: 'Select a value'},
  'Foo City': {id: 'Foo City', other_field: '123 foo', other_field_2: 'foo'},
  'Bar City': {id: 'Bar City', other_field: '234 bar', other_field_2: 'bar'},
  'Baz City': {id: 'Baz City', other_field: '345 baz', other_field_2: 'baz'},
  ___empty: false,
});

const getErrorLog = (): DapiUpdateError[] => {
  return [];
};

const getProgram = () => <Program>{defaults: {}};
