/**
 *  Tests for Indication Validator
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {ClientQuote} from '../../../src/client/quote/ClientQuote';
import {Cmatch} from '../../../src/client/Cmatch';
import {DelayedStagingBucket} from '../../../src/bucket/DelayedStagingBucket';
import {PositiveInteger} from '../../../src/numeric';
import {Program} from '../../../src/program/Program';
import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {StepSimulator} from '../../../src/dullahan/step/StepSimulator';
import {Step} from '../../../src/step/Step';
import {Validator as Sut} from '../../../src/dullahan/validate/Validator';
import {expect} from 'chai';

describe('Validator', () => {
  describe('#validate', () => {
    it('Only iterates first step to highest requested/applicable step', () => {
      const getStep = createMockGetStep();
      const program = createMockProgram();
      const quote = createMockQuote();

      const steps_expected = [1];
      const steps_submitted: any = [];

      program.submit = (id: any, _: any, __: any) => {
        // convert this from a PositiveInteger
        steps_submitted.push(+id);

        return null;
      };

      return getSut({
        program,
        getStep,
        stepSimulator: createMockStepSimulator(program, ['foo', 'bar']),
      })
        .validate(quote, getCreateClientQuote(), getCreateCmatch())
        .then(() => {
          expect(steps_submitted).to.deep.equal(steps_expected);
        });
    });

    it('Returns failed assertions', () => {
      const expected_failures = [
        'Step 1: foo: change_error',
        'Step 1: foo: submit_error',
        'Step 1: foo: forward_error',
      ];

      const program = createMockProgram({});
      const quote = createMockQuote();
      const getStep = createMockGetStep();

      program.change = (_id, name): Record<string, unknown[]> => {
        return name === 'foo' ? {foo: ['change_error']} : {};
      };
      program.forward = () => ({foo: ['forward_error']});
      program.submit = () => ({foo: ['submit_error']});

      return getSut({
        program,
        getStep,
        stepSimulator: createMockStepSimulator(program, ['foo', 'bar']),
      })
        .validate(quote, getCreateClientQuote(), getCreateCmatch())
        .then(response => {
          expect(response.failed_assertions).to.deep.equal(expected_failures);
        });
    });

    it('Returns missing required fields', () => {
      const given_missing = [['some_field', '0']];
      const expected_missing = ['Step 1: some_field missing index 0'];

      const getStep = createMockGetStep(given_missing);
      const quote = createMockQuote();
      const program = createMockProgram();

      return getSut({
        getStep,
        stepSimulator: createMockStepSimulator(program, ['foo', 'bar']),
      })
        .validate(quote, getCreateClientQuote(), getCreateCmatch())
        .then(response => {
          expect(response.missing_required).to.deep.equal(expected_missing);
        });
    });

    it('forwards program per step', () => {
      const given_missing = [['some_field', '0']];
      let forwards = 0;
      const expected = 1;

      const getStep = createMockGetStep(given_missing);
      const quote = createMockQuote();
      const program = createMockProgram();
      program.forward = (_a, _b, _c, _d) => {
        forwards++;
        return null;
      };

      return getSut({
        getStep,
        program,
        stepSimulator: createMockStepSimulator(program, ['foo', 'bar']),
      })
        .validate(quote, getCreateClientQuote(), getCreateCmatch())
        .then(() => {
          expect(forwards).to.eq(expected);
        });
    });
  });

  describe('validateField', () => {
    it('Returns failed assertions', () => {
      const given_failures = {some_field: 'Something is wrong'};
      const expected_failures = ['Step 1: foo: error'];

      const program = createMockProgram(given_failures);
      const getStep = createMockGetStep();

      program.change = () => ({
        foo: ['error'],
      });

      const results = getSut({
        program,
        getStep,
        stepSimulator: createMockStepSimulator(program, ['foo', 'bar']),
      }).validateField(
        createMockDelayedStagingBucket(),
        getCreateCmatch(),
        <PositiveInteger>1,
        'foo'
      );

      expect(results.failed_assertions).to.deep.equal(expected_failures);
    });
  });
});

describe('validateStep', () => {
  it('Returns failed assertions', () => {
    const expected_failures = [
      'Step 1: foo: submit_error',
      'Step 1: foo: forward_error',
    ];

    const program = createMockProgram({});
    const getStep = createMockGetStep();

    program.forward = () => ({foo: ['forward_error']});
    program.submit = () => ({foo: ['submit_error']});

    const results = getSut({
      program,
      getStep,
      stepSimulator: createMockStepSimulator(program, ['foo', 'bar']),
    }).validateStep(
      createMockBucket(),
      createMockQuote(),
      getCreateCmatch(),
      <PositiveInteger>1
    );

    expect(results.failed_assertions).to.deep.equal(expected_failures);
  });
});

function getSut(dependencies: CommonObject = {}) {
  const {program, getStep, stepSimulator} = dependencies;

  const p = <Program>program ?? createMockProgram();
  return new Sut(
    p,
    <(_: any, __: any) => Step>getStep ?? createMockGetStep(),
    <StepSimulator>stepSimulator
  );
}

const getCreateClientQuote = () => {
  return <ClientQuote>(<unknown>{
    forceClassify: () => {},
    visitData: (cb: any) => cb({getDataByName: () => {}}),
  });
};

const getCreateCmatch = () => {
  return <Cmatch>(<unknown>{
    hookClassifier: () => {},
    getMatches: () => {
      return {};
    },
  });
};

function createMockGetStep(
  missing_required: any[] = []
): (id: any, quote: any) => Step {
  return (_id: any, _quote: any) => {
    const step = <Step>(<unknown>{
      setRequiredFieldNames: () => step,
      getRequired: () => {
        return missing_required;
      },
    });

    return step;
  };
}

function createMockProgram(failures = {}): Program {
  return <Program>(<unknown>{
    submit: () => failures,
    getFirstStepId: () => 1,
    steps: [{}, {}, {}, {}],
    rateSteps: [false, false, true, false],
    requiredFields: [{}, {}, {}, {}],
    qstep: {foo: 1, bar: 1, baz: 1},
    isStepVisible: () => true,
    change: () => {},
    forward: () => {},
  });
}

function createMockBucket() {
  return <QuoteDataBucket>(<unknown>{});
}

function createMockDelayedStagingBucket() {
  return <DelayedStagingBucket>(<unknown>{});
}

function createMockStepSimulator(program: Program, _request_fields: string[]) {
  return <StepSimulator>(<unknown>{
    each: (cb: Function) => {
      [program.getFirstStepId()].forEach(id => {
        cb(id, ['foo', 'bar', 'baz']);
      });
    },
  });
}

function createMockQuote(): ServerSideQuote {
  return <ServerSideQuote>(<unknown>{
    classify: () => {
      return {
        '--vis-foo': {},
      };
    },
    getBucket: () => {
      return {};
    },
    getFieldState: () => {
      return {
        foo: 1,
        bar: 1,
        baz: 1,
      };
    },
  });
}
