import * as Sut from '../../../src/dullahan/validate/copy_validator';
import CopyRequestAssertionError from '../../../src/error/CopyRequestAssertionError';
import {expect} from 'chai';

describe('assertCopyRequest', () => {
  it('should not assert with valid input', () => {
    expect(() =>
      Sut.assertCopyRequest({
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo123',
        bucket: {foo: ['bar']},
        origin: 'bar',
      })
    ).to.not.throw;
  });

  [
    ['number', 123, 'object'],
    ['string', 'abc', 'object'],
    [
      'missing quote_id',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'quote ID must be a positive integer',
    ],
    [
      'string quote_id',
      {
        quote_id: 'aaa',
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'quote ID must be a positive integer',
    ],
    [
      'quote_id is 0',
      {
        quote_id: 0,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'quote ID must be a positive integer',
    ],
    [
      'quote_id is negative',
      {
        quote_id: -100,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'quote ID must be a positive integer',
    ],
    [
      'missing session_id',
      {
        quote_id: 123,
        policy_id: '',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'number session_id',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 345,
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'empty session_id',
      {
        quote_id: 123,
        policy_id: '',
        session_id: '',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'missing policy_id',
      {
        quote_id: 123,
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'policy ID must be a string',
    ],
    [
      'number policy_id',
      {
        quote_id: 123,
        policy_id: 123,
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'policy ID must be a string',
    ],
    [
      'empty session_id',
      {
        quote_id: 123,
        policy_id: '',
        session_id: '',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'missing program',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'number program',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 321,
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'empty program',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: '',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'invalid char in program',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'abc../%#$',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'missing bucket',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'string bucket',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: '',
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'null bucket',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: null,
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket key is invalid',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {'a%%%': ['b'], c: ['d']},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket value is not an array',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {a: ['b'], c: 'd'},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket value has an empty array',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {a: ['b'], c: []},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket value has an non-string in array',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {a: ['b'], c: [1]},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'missing origin',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
    [
      'null origin',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: null,
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
    [
      'origin is invalid',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: 'a%%%',
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
    [
      'empty origin',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: '',
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
  ].forEach(([label, given, expected]) => {
    it(`should assert with invalid input - ${label}`, () => {
      expect(() => Sut.assertCopyRequest(given)).to.throw(
        CopyRequestAssertionError,
        expected as string
      );
    });
  });
});
