/**
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import * as Sut from '../../../src/dullahan/validate/init_validator';
import InitRequestAssertionError from '../../../src/error/InitRequestAssertionError';
import {expect} from 'chai';

describe('assertInitRequest', () => {
  it('should not assert with valid input', () => {
    expect(() =>
      Sut.assertInitRequest({
        policy_id: '',
        session_id: 'abc',
        program: 'foo123',
        bucket: {foo: ['bar']},
        origin: 'bar',
      })
    ).to.not.throw;
  });

  it('should not assert with valid input - null quote id', () => {
    expect(() =>
      Sut.assertInitRequest({
        quote_id: null,
        policy_id: '',
        session_id: 'abc',
        program: 'foo123',
        bucket: {foo: ['bar']},
        origin: 'bar',
      })
    ).to.not.throw;
  });

  [
    ['number', 123, 'object'],
    ['string', 'abc', 'object'],
    [
      'has quote_id',
      {
        quote_id: 123,
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'Init must not have a quote_id',
    ],
    [
      'missing session_id',
      {
        policy_id: '',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'number session_id',
      {
        policy_id: '',
        session_id: 345,
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'empty session_id',
      {
        policy_id: '',
        session_id: '',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'missing policy_id',
      {
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'policy ID must be a string',
    ],
    [
      'number policy_id',
      {
        policy_id: 123,
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'policy ID must be a string',
    ],
    [
      'empty session_id',
      {
        policy_id: '',
        session_id: '',
        program: 'foo',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'session ID must be a non-empty string',
    ],
    [
      'missing program',
      {
        policy_id: '',
        session_id: 'abc',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'number program',
      {
        policy_id: '',
        session_id: 'abc',
        program: 321,
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'empty program',
      {
        policy_id: '',
        session_id: 'abc',
        program: '',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'invalid char in program',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'abc../%#$',
        bucket: {foo: ['bar']},
        origin: 'bar',
      },
      'program must be a non-empty string',
    ],
    [
      'missing bucket',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'string bucket',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: '',
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'null bucket',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: null,
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket key is invalid',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {'a%%%': ['b'], c: ['d']},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket value is not an array',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {a: ['b'], c: 'd'},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket value has an empty array',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {a: ['b'], c: []},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'bucket value has an non-string in array',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {a: ['b'], c: [1]},
        origin: 'bar',
      },
      'bucket invalid',
    ],
    [
      'missing origin',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
    [
      'null origin',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: null,
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
    [
      'origin is invalid',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: 'a%%%',
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
    [
      'origin is empty',
      {
        policy_id: '',
        session_id: 'abc',
        program: 'foo',
        origin: '',
        bucket: {foo: ['bar']},
      },
      'origin must be a non-empty string',
    ],
  ].forEach(([label, given, expected]) => {
    it(`should assert with invalid input - ${label}`, () => {
      expect(() => Sut.assertInitRequest(given)).to.throw(
        InitRequestAssertionError,
        expected as string
      );
    });
  });
});
