/**
 * Tests Server
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {expect, assert} from 'chai';
import {createStubInstance} from 'sinon';
import {DocumentData, MongoServerDao} from '../../src/server/db/MongoServerDao';
import {FeatureFlag} from '../../src/system/flags/FeatureFlag';
import {JsonServerResponse} from '../../src/server/request/JsonServerResponse';
import {PositiveInteger} from '../../src/numeric';
import {PriorityLog} from '../../src/server/log/PriorityLog';
import {EncryptedKeys, Program} from '../../src/program/Program';
import {QuoteDataBucket} from '../../src/bucket/QuoteDataBucket';
import {Server as Sut} from '../../src/server/Server';
import {ServerSideQuote} from '../../src/server/quote/ServerSideQuote';
import {UserRequest} from '../../src/server/request/UserRequest';
import {UserSession} from '../../src/server/request/UserSession';
import {Callback, FailCallback} from '../../src/server/db/ServerDao';
import {BucketData} from '../../src/bucket/Bucket';
import {DapiProcessor} from '../../src/server/dapi/DapiProcessor';
import {ClassificationResult} from '../../src/program/Program';
import {randomBytes} from 'crypto';
import {KekService} from '../../src/server/service/KekService';
import {QuoteNotFound} from '../../src/server/quote/loader';

import * as TE from 'fp-ts/lib/TaskEither';

const identity = <T>(x: T) => x;

describe('Server#sendStep', () => {
  [
    {
      label: 'Sends error when navigating to step that has not been reached',
      step_id: 6,
      first_step_id: 2,
      current_step_id: 3,
      next_visible_step_id: 5,
      visited_step_id: 4,
      saved_step_id: 3,
      internal: false,
      expected_failure: true,
      expected_action: [{action: 'gostep', id: 4}],
    },
    {
      label: 'Non-internal user cannot visit manage step',
      step_id: 1,
      first_step_id: 2,
      current_step_id: 2,
      next_visible_step_id: 3,
      visited_step_id: 2,
      saved_step_id: 2,
      internal: false,
      expected_failure: false,
      expected_action: [{action: 'gostep', id: 2}],
    },
    {
      label: 'Internal user can visit manage step',
      step_id: 1,
      first_step_id: 2,
      current_step_id: 2,
      next_visible_step_id: 3,
      visited_step_id: 2,
      saved_step_id: 2,
      internal: true,
      expected_failure: false,
      expected_action: undefined,
    },
    {
      label: 'Send step when navigating to a previous step',
      step_id: 2,
      first_step_id: 2,
      current_step_id: 3,
      next_visible_step_id: 4,
      visited_step_id: 4,
      saved_step_id: 3,
      internal: false,
      expected_failure: false,
      expected_action: undefined,
    },
    {
      label: 'Send step when navigating to the next valid step (top saved + 1)',
      step_id: 99,
      first_step_id: 2,
      current_step_id: 2,
      next_visible_step_id: 3,
      visited_step_id: 3,
      saved_step_id: 2,
      internal: false,
      expected_failure: true,
      expected_action: [{action: 'gostep', id: 2}],
    },
    {
      label:
        'Send step when navigating to the next visible step (top saved + 1) skip hidden steps',
      step_id: 5,
      first_step_id: 2,
      current_step_id: 3,
      next_visible_step_id: 5,
      visited_step_id: 3,
      saved_step_id: 3,
      internal: false,
      expected_failure: false,
      expected_action: undefined,
    },
  ].forEach(
    ({
      label,
      step_id,
      first_step_id,
      current_step_id,
      next_visible_step_id,
      visited_step_id,
      saved_step_id,
      internal,
      expected_failure,
      expected_action,
    }) => {
      it(label, done => {
        const response = createMockResponse();

        let error_is_sent = false;
        let response_is_sent = false;

        response.from = (_, __, action) => {
          response_is_sent = true;
          expect(expected_failure).to.be.false;
          expect(error_is_sent).to.be.false;
          expect(response_is_sent).to.equal(true);
          expect(action).to.deep.equal(expected_action);
          done();

          return '';
        };
        response.error = (_, action, __) => {
          error_is_sent = true;
          expect(expected_failure).to.be.true;
          expect(error_is_sent).to.be.true;
          expect(action).to.deep.equal(expected_action);
          expect(response_is_sent).to.be.false;
          done();

          return '';
        };

        const sut = getSut(response);
        sut.init(createMockCache(), createMockRater());
        const quote = createMockQuote(
          current_step_id,
          saved_step_id,
          visited_step_id
        );
        const program = createMockProgram({
          first_step_id,
          next_visible_step_id,
        });
        const session = createMockSession(internal);
        const request = createMockRequest();

        sut.sendStep(
          request,
          quote,
          program,
          <PositiveInteger>step_id,
          session
        );
      });
    }
  );
});

describe('Server#visitStep', () => {
  [
    {
      label:
        'Valid step visit for previous step sets current step and sends response',
      step_to_visit: 3,
      current_step: 5,
      top_saved_step: 5,
      top_visited_step: 5,
      next_visible_step: 5,
      expected_current_step: 3,
    },
    {
      label:
        'Valid step visit for next step sets current step and sends response',
      step_to_visit: 5,
      current_step: 4,
      top_saved_step: 4,
      top_visited_step: 4,
      next_visible_step: 5,
      expected_current_step: 5,
    },
    {
      label:
        'Valid step visit for next step (skip hidden) sets current step and sends response',
      step_to_visit: 5,
      current_step: 3,
      top_saved_step: 3,
      top_visited_step: 3,
      next_visible_step: 5,
      expected_current_step: 5,
    },
    {
      label:
        'Valid step visit from current step sets current step and sends response',
      step_to_visit: 3,
      current_step: 3,
      top_saved_step: 3,
      top_visited_step: 3,
      next_visible_step: 5,
      expected_current_step: 3,
    },
  ].forEach(
    ({
      label,
      step_to_visit,
      current_step,
      top_saved_step,
      top_visited_step,
      next_visible_step,
      expected_current_step,
    }) => {
      it(label, done => {
        let error_is_sent = false;

        const quote = createMockQuote(
          current_step,
          top_saved_step,
          top_visited_step
        );

        quote.getBucket = () =>
          <any>{
            getData: () => {},
            getDataByName: () => [current_step],
          };

        const program = createMockProgram();
        const request = createMockRequest();
        const response = createMockResponse();
        const dao = createMockDao();
        const logger = createMockLogger();
        const feature_flag = createMockFeatureFlag(true);

        const sut = getSut(response, dao, logger, undefined, feature_flag);

        // check that a response is sent to user w/no actions (do not kick back)
        response.from = (_, __, action) => {
          expect(error_is_sent).to.be.false;
          expect(action).to.deep.equal([]);
          expect(given_cur_step_id).to.be.equal(expected_current_step);
          expect(save_quote_state_is_called).to.be.true;
          done();

          return '';
        };

        program.getNextVisibleStep = () => <PositiveInteger>next_visible_step;

        // make sure current step is set
        let given_cur_step_id: any = undefined;
        quote.setCurrentStepId = step_id => {
          given_cur_step_id = step_id;

          return quote;
        };

        // current and top step ids are persisted
        let save_quote_state_is_called = false;
        dao.saveQuoteState = (quote, success_cb) => {
          save_quote_state_is_called = true;
          success_cb && success_cb(quote);

          return dao;
        };

        response.error = (_, __, ___) => {
          error_is_sent = true;

          return '';
        };

        sut.visitStep(<PositiveInteger>step_to_visit, request, quote, program);
      });
    }
  );

  interface CheckRatingTriggersTestCase {
    label: string;
    rates_triggered: string[];
    classes: Record<string, {is: boolean}>;
  }

  (<CheckRatingTriggersTestCase[]>[
    {
      label: 'No rates are triggered when trigger classifications are missing',
      rates_triggered: [],
      classes: {foo: {is: true}},
    },
    {
      label: 'Rates are triggered when classifications are true and present',
      rates_triggered: ['foo'],
      classes: {
        '--liza-trigger-rate': {is: true},
        '--liza-trigger-rate-foo': {is: true},
      },
    },
    {
      label: 'Rates are not triggered when classifications are false',
      rates_triggered: [],
      classes: {
        '--liza-trigger-rate': {is: true},
        '--liza-trigger-rate-foo': {is: false},
      },
    },
    {
      label: 'Only one rate action is sent at a time',
      rates_triggered: ['foo', 'bar'],
      classes: {
        '--liza-trigger-rate': {is: true},
        '--liza-trigger-rate-foo': {is: true},
        '--liza-trigger-rate-bar': {is: true},
      },
    },
    {
      label: 'If the aggregate classification is false, skip individual checks',
      rates_triggered: [],
      classes: {
        '--liza-trigger-rate': {is: false},
        '--liza-trigger-rate-foo': {is: true},
        '--liza-trigger-rate-bar': {is: true},
      },
    },
  ]).forEach(({label, rates_triggered, classes}) => {
    it(label, () => {
      const quote = createMockQuote();

      quote.classify = () => classes;
      quote.getDataByName = (name: string): unknown[] => {
        switch (name) {
          case '__liza_step_id_current':
            return [1];
          default:
            return [];
        }
      };

      const sut = getSut(
        undefined,
        undefined,
        undefined,
        undefined,
        createMockFeatureFlag(true)
      );

      sut.hookRating((_: any, __: any, suppliers) => {
        expect(suppliers).to.deep.equal(rates_triggered);

        return Promise.resolve({});
      });

      return sut.visitStep(
        <PositiveInteger>1,
        createMockRequest(),
        quote,
        createMockProgram()
      );
    });
  });

  it('Invisible step kicks user back to first step', done => {
    const step_to_visit = 3;
    const current_step = 2;
    const top_saved_step = 5;
    const top_visited_step = 5;
    const first_step_id = 1;
    let error_is_sent = false;

    const quote = createMockQuote(
      current_step,
      top_saved_step,
      top_visited_step
    );
    const program = createMockProgram({first_step_id});
    const session = createMockSession();
    const request = createMockRequest(session);
    const response = createMockResponse();
    const logger = createMockLogger();
    const feature_flag = createMockFeatureFlag();

    const sut = getSut(response, undefined, logger, undefined, feature_flag);

    // make this step invisible
    program.isStepVisible = () => false;

    let log_is_called = false;
    logger.log = (level, message, step_id, __) => {
      log_is_called = true;
      expect(level).to.be.equal(PriorityLog.PRIORITY_INFO);
      expect(step_id).to.be.equal(step_to_visit);
      expect(message).to.contain('not applicable step');
      expect(log_is_called).to.be.true;

      return logger;
    };

    // check that a response is sent to user w/kickback action
    response.from = (_, __, action) => {
      expect(error_is_sent).to.be.false;
      expect(log_is_called).to.be.true;
      expect(action).to.deep.equal([{action: 'gostep', id: first_step_id}]);
      done();

      return '';
    };

    response.error = (_, __, ___) => {
      error_is_sent = true;

      return '';
    };

    sut.visitStep(<PositiveInteger>step_to_visit, request, quote, program);
  });

  it('allowed to visit the Manage step', done => {
    const step_to_visit = 1;
    const current_step = 2;
    const top_saved_step = 2;
    const top_visited_step = 2;
    const first_step_id = 1;
    let error_is_sent = false;

    const quote = createMockQuote(
      current_step,
      top_saved_step,
      top_visited_step
    );
    // our mocked program sets up manage step as step 1
    const program = createMockProgram({first_step_id});
    const session = createMockSession();
    const request = createMockRequest(session);
    const response = createMockResponse();
    const logger = createMockLogger();
    const dao = createMockDao();
    const sut = getSut(response, dao, logger);

    // check that a response is sent to user w/no actions (do not kick back)
    response.from = (_, __, action) => {
      expect(error_is_sent).to.be.false;
      expect(action).to.deep.equal([]);
      expect(given_cur_step_id).to.be.equal(step_to_visit);
      done();

      return '';
    };

    dao.saveQuoteState = (quote, success_cb) => {
      success_cb && success_cb(quote);
      return dao;
    };

    // we want to test the manage quote step
    program.isManageQuoteStep = _ => true;

    // manage step is not visible
    program.isStepVisible = () => false;

    let given_cur_step_id: any = undefined;
    quote.setCurrentStepId = step_id => {
      given_cur_step_id = step_id;

      return quote;
    };

    response.error = (_, __, ___) => {
      error_is_sent = true;

      return '';
    };

    sut.visitStep(<PositiveInteger>step_to_visit, request, quote, program);
  });
});

describe('beforeLoad triggers', () => {
  it('Calls _triggerDapi on dapi events', done => {
    let beforeLoadStep_called = false;
    let triggerDapi_called = false;

    const dapi_name = 'some_dapi_name';

    const quote = createMockQuote();
    const program = createMockProgram();
    const session = createMockSession();
    const request = createMockRequest(session);
    const response = createMockResponse();
    const dao = createMockDao();
    const dapi_processor = createMockDapiProcessor();

    dapi_processor.triggerDapi = (_: any, __: any, ___: any, name: any) => {
      expect(name).to.equal(dapi_name);

      triggerDapi_called = true;
    };

    const sut = getSut(response, dao, undefined, dapi_processor);

    program.meta.fields[dapi_name] = {
      desc: 'foo',
      dapi: {
        name: '',
        value: '',
        when: '',
        mapsrc: {
          input1: 'mapped1',
        },
        mapdest: {
          foo: 'bar',
        },
      },
    };

    program.beforeLoadStep = (_: any, __: any, cb: any) => {
      beforeLoadStep_called = true;

      cb('dapi', '', dapi_name, 0);

      return true;
    };

    request.end = _ => {
      expect(beforeLoadStep_called, 'beforeLoadStep was called').to.be.true;
      expect(triggerDapi_called, 'triggerDapi was called').to.be.true;
      done();

      return request;
    };

    sut.visitStep(<PositiveInteger>1, request, quote, program);
  });
});

describe('Server#publishQuote', () => {
  [
    {
      label: 'Not Internal User can not publish',
      is_internal: false,
      publish_quote_returns_error: false,
      log_level_expected: 0,
      log_msg_expected:
        'Quote %s access update: not allowed for non-internal users',
      post_data: {publish: '1'},
    },
    {
      label: 'calls dao#publishQuote with publish true',
      is_internal: true,
      publish_quote_returns_error: false,
      log_level_expected: 3,
      log_msg_expected: 'Quote %s access update: success',
      post_data: {publish: '1'},
    },
    {
      label: 'calls dao#publishQuote with publish false',
      is_internal: true,
      publish_quote_returns_error: false,
      log_level_expected: 3,
      log_msg_expected: 'Quote %s access update: success',
      post_data: {publish: '0'},
    },
    {
      label: 'missing post data',
      is_internal: true,
      publish_quote_returns_error: false,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: insufficient data received',
      post_data: undefined,
    },
    {
      label: 'missing publish in post data',
      is_internal: true,
      publish_quote_returns_error: false,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: insufficient data received',
      post_data: {},
    },
    {
      label: 'publishQuote returns an error',
      is_internal: true,
      publish_quote_returns_error: true,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: dao error: kaboom!',
      post_data: {publish: '1'},
    },
  ].forEach(
    ({
      label,
      is_internal,
      publish_quote_returns_error,
      log_level_expected,
      log_msg_expected,
      post_data,
    }) => {
      it(label, done => {
        const quote = createMockQuote();
        const session = createMockSession(is_internal);
        const request = createMockRequest(session);
        const response = createMockResponse();
        const logger = createMockLogger();
        const dao = createMockDao();
        const sut = getSut(response, dao, logger);

        let log_level_actual: any;
        let log_msg_actual: any;

        logger.log = (level, message) => {
          log_level_actual = level;
          log_msg_actual = message;

          return logger;
        };

        let get_post_data_called = false;
        request.consumePostData = c => {
          get_post_data_called = true;
          c(post_data);

          return request;
        };

        request.end = _ => {
          expect(log_level_actual).to.equal(log_level_expected);
          expect(log_msg_actual).to.equal(log_msg_expected);
          expect(get_post_data_called).to.equal(is_internal);
          done();

          return request;
        };

        if (publish_quote_returns_error) {
          dao.publishQuote = () => Promise.reject('kaboom!');
        }

        sut.publishQuote(request, quote, is_internal);
      });
    }
  );
});

describe('Server#metaUpdate', () => {
  [
    {
      label: 'user can not perform meta updates',
      can_update_meta: false,
      mock_dao_error: false,
      log_level_expected: 3,
      expect_save_meta_is_called: false,
      log_msg_expected: 'Quote %s access update: access denied',
      post_data: {key: 'retail_agency', value: 'foo'},
    },
    {
      label: 'post w/valid retail_agency calls dao#saveQuoteMeta to set agency',
      can_update_meta: true,
      mock_dao_error: false,
      log_level_expected: 3,
      log_msg_expected: 'Quote %s access update: success',
      expect_save_meta_is_called: true,
      post_data: {key: 'retail_agency', value: 'foo'},
      expected_meta: {retail_agency: ['foo']},
    },
    {
      label: 'post w/invalid meta key is rejected',
      can_update_meta: true,
      mock_dao_error: false,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: insufficient data received',
      expect_save_meta_is_called: false,
      post_data: {key: 'key_off_limits', value: 'foo'},
    },
    {
      label: 'post w/invalid meta value is rejected',
      can_update_meta: true,
      mock_dao_error: false,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: insufficient data received',
      expect_save_meta_is_called: false,
      post_data: {
        key: 'retail_agency',
        value: '<script>this is not alpha numeric</script>',
      },
    },
    {
      label: 'missing post data is rejected',
      can_update_meta: true,
      mock_dao_error: false,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: no data received',
      expect_save_meta_is_called: false,
      post_data: undefined,
    },
    {
      label: 'saveQuoteMeta returns an error on save',
      can_update_meta: true,
      mock_dao_error: true,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: write error',
      expect_save_meta_is_called: true,
      post_data: {key: 'retail_agency', value: 'foo'},
      expected_meta: {retail_agency: ['foo']},
    },
    {
      label:
        'post w/valid liza_access_email calls dao#saveQuoteMeta to set agency',
      can_update_meta: true,
      mock_dao_error: false,
      log_level_expected: 3,
      log_msg_expected: 'Quote %s access update: success',
      expect_save_meta_is_called: true,
      post_data: {key: 'liza_access_email', value: 'foo@example.com'},
      expected_meta: {liza_access_email: ['foo@example.com']},
    },
    {
      label: 'post w/invalid liza_access_email is rejected',
      can_update_meta: true,
      mock_dao_error: false,
      log_level_expected: 0,
      log_msg_expected: 'Quote %s access update: insufficient data received',
      expect_save_meta_is_called: false,
      post_data: {key: 'liza_access_email', value: 'fooexamplecom'},
    },
  ].forEach(
    ({
      label,
      can_update_meta,
      mock_dao_error,
      log_level_expected,
      log_msg_expected,
      expect_save_meta_is_called,
      post_data,
      expected_meta,
    }) => {
      it(label, done => {
        const quote = createMockQuote();
        const session = createMockSession(can_update_meta);
        const request = createMockRequest(session);
        const response = createMockResponse();
        const logger = createMockLogger();
        const dao = createMockDao();

        let log_level_actual: any;
        let log_msg_actual: any;

        logger.log = (level, message) => {
          log_level_actual = level;
          log_msg_actual = message;

          return logger;
        };

        let get_post_data_called = false;
        request.consumePostData = c => {
          get_post_data_called = true;
          c(post_data);

          return request;
        };

        let save_quote_called = false;
        request.end = _ => {
          expect(log_level_actual).to.equal(log_level_expected);
          expect(log_msg_actual).to.equal(log_msg_expected);
          expect(get_post_data_called).to.equal(can_update_meta);
          expect(save_quote_called).to.equal(expect_save_meta_is_called);
          done();

          return request;
        };

        dao.saveQuoteMetaDirect = (quote, meta, cb_success, cb_failure) => {
          save_quote_called = true;
          expect(quote).to.deep.equal(quote);
          expect(meta).to.deep.equal(expected_meta);
          mock_dao_error
            ? cb_failure && cb_failure(quote)
            : cb_success && cb_success(quote);
        };

        const sut = getSut(response, dao, logger);
        sut.metaUpdate(request, quote, can_update_meta);
      });
    }
  );
});

describe('Server#handlePost', () => {
  it('Prevents step save when not next immediate step', done => {
    // attempt to post on a step 2 steps ahead - not ok!
    const step_to_post = 4;
    const current_step = 2;
    const next_visible_step_id = 3;
    const top_saved_step = 2;
    const top_visited_step = 2;
    const expected_kick_back = 3;

    const quote = createMockQuote(
      current_step,
      top_saved_step,
      top_visited_step
    );
    const program = createMockProgram({first_step_id: 1, next_visible_step_id});
    const session = createMockSession();
    const request = createMockRequest(session);
    const response = createMockResponse();
    const autosave = false;
    const sut = getSut(response);

    let get_post_data_called = false;
    request.consumePostData = () => {
      get_post_data_called = true;

      return request;
    };

    const expected_error_action = [
      {action: 'gostep', id: expected_kick_back, title: 'Go Back'},
    ];
    response.error = (message, action, __) => {
      expect(get_post_data_called).to.be.false;
      expect(action).to.deep.equal(expected_error_action);
      expect(message).to.contain(
        'Unable to save step: you have not yet reached'
      );
      done();

      return '';
    };

    sut.handlePost(
      (_: EncryptedKeys, __: string[], to_encrypt: BucketData) => to_encrypt,
      identity,
      <PositiveInteger>step_to_post,
      request,
      quote,
      program,
      session,
      autosave
    );
  });

  it('Clears last_rated_date when classification is true', done => {
    let set_data_called = false;

    const quote = createMockQuote();

    quote.classify = () => ({
      '--liza-clear-rating-results': {is: true},
      '--liza-clear-rating-results-foo': {is: true},
    });

    quote.setMetadata = () => {
      expect(set_data_called).to.be.true;

      return quote;
    };

    quote.getDataByName = () => ['current last_rated_date'];

    quote.setData = (data: any) => {
      const keys = ['foo_last_rated_date', 'last_rated_date'];
      const key = Object.keys(data).shift() ?? '';

      expect(keys.includes(key)).to.be.true;
      expect(data[key]).to.deep.equal(['0']);

      set_data_called = true;
      return quote;
    };

    const dao = createMockDao();

    dao.saveQuote = (
      _quote: ServerSideQuote,
      _class_data: ClassificationResult,
      _success?: Callback,
      _failure?: FailCallback,
      _save_data?: any,
      push_data?: any
    ) => {
      expect(push_data).to.deep.equal({
        'rdelta.data': {
          data: {
            foo_last_rated_date: ['current last_rated_date'],
          },
          concluding_save: false,
          timestamp: 0,
          step_id: 2,
        },
      });

      done();

      return dao;
    };

    getSut(undefined, dao).handlePost(
      (_: EncryptedKeys, __: string[], to_encrypt: BucketData) => to_encrypt,
      identity,
      <PositiveInteger>2,
      createMockRequest(),
      quote,
      createMockProgram(),
      createMockSession(),
      false
    );
  });

  it('Clears last_rated_date properly for multiple steps', done => {
    let set_data_iterations = 0;

    const quote = createMockQuote();

    quote.classify = () => ({
      '--liza-clear-rating-results': {is: true},
      '--liza-clear-rating-results-previous': {is: false},
      '--liza-clear-rating-results-current': {is: true},
    });

    quote.setMetadata = () => {
      expect(set_data_iterations).to.be.gt(0);
      done();
      return quote;
    };

    quote.getBucket = () => <QuoteDataBucket>(<unknown>{
        getData: () => {},

        getDataByName: (name: string) => {
          const lookup: Record<string, unknown> = {
            previous_last_rated_date: ['1234567'],
            current_last_rated_date: ['0'],
          };

          return lookup[name];
        },
      });

    quote.setData = (data: any) => {
      const expectations = [
        {key: 'current_last_rated_date', value: '0'},
        {key: 'last_rated_date', value: '1'},
      ];

      const {key, value} = expectations[set_data_iterations];

      expect(data[key]).to.deep.equal([value]);

      set_data_iterations++;
      return quote;
    };

    getSut().handlePost(
      (_: EncryptedKeys, __: string[], to_encrypt: BucketData) => to_encrypt,
      identity,
      <PositiveInteger>2,
      createMockRequest(),
      quote,
      createMockProgram(),
      createMockSession(),
      false
    );
  });

  it('Clears liza_timestamp_rate_request and retryAttempts on forward_save of rating step', done => {
    const current_step_id = 5;
    const saved_step_id = 5;
    const visited_stop_id = 5;
    const dao = createMockDao();
    const quote = createMockQuote(
      current_step_id,
      saved_step_id,
      visited_stop_id
    );
    const program = createMockProgram({
      first_step_id: 5,
      next_visible_step_id: 6,
    });
    const post_data_non_rating = {
      data: '{"foo": ["bar"]}',
      forward_save: 'false',
    };
    const post_data_rating = {
      data: '{"foo": ["bar"]}',
      forward_save: 'true',
    };
    const {session, request, sut} = getMocks();
    const meta_data_expected: BucketData = {liza_timestamp_rate_request: [0]};
    let metaDataCallCount = 0;
    let retryAttempCallCount = 0;
    let final_setMetadata_called = false;
    let retry_attempts_given: number;

    quote.setMetadata = (meta: BucketData) => {
      metaDataCallCount++;

      if (metaDataCallCount === 2) {
        final_setMetadata_called = true;
      }

      expect(meta).to.deep.equal(meta_data_expected);

      return quote;
    };
    dao.saveQuoteMetaDirect = (quote, meta) => {
      expect(quote).to.deep.equal(quote);
      expect(meta).to.deep.equal(meta_data_expected);
    };

    quote.setRetryAttempts = (attempts: number) => {
      retryAttempCallCount++;
      retry_attempts_given = attempts;

      expect(retry_attempts_given).to.equal(0);

      return quote;
    };

    request.consumePostData = c => {
      if (metaDataCallCount === 0) {
        c(post_data_non_rating);
      } else {
        c(post_data_rating);
      }

      return request;
    };

    // First tested step non rating
    sut
      .handlePost(
        (_: EncryptedKeys, __: string[], to_encrypt: BucketData) => to_encrypt,
        identity,
        <PositiveInteger>5,
        request,
        quote,
        program,
        session,
        false
      )
      // Second tested step rating
      .then(sut => {
        return sut.handlePost(
          (_: EncryptedKeys, __: string[], to_encrypt: BucketData) =>
            to_encrypt,
          identity,
          <PositiveInteger>6,
          request,
          quote,
          program,
          session,
          false
        );
      })
      // Third tested step rating
      .then(sut => {
        return sut.handlePost(
          (_: EncryptedKeys, __: string[], to_encrypt: BucketData) =>
            to_encrypt,
          identity,
          <PositiveInteger>7,
          request,
          quote,
          program,
          session,
          false
        );
      })
      // Final test values
      .then(() => {
        expect(retry_attempts_given).to.equal(0);
        expect(retryAttempCallCount).to.equal(2);
        expect(final_setMetadata_called, 'final setMetadata call made').to.be
          .true;
        done();
      });
  });

  it('Clears liza_timestamp_rate_request and retryAttempts on excess_prerate_rating_complete of rating step', done => {
    const current_step_id = 5;
    const saved_step_id = 5;
    const visited_stop_id = 5;
    const quote = createMockQuote(
      current_step_id,
      saved_step_id,
      visited_stop_id
    );
    const program = createMockProgram({
      first_step_id: 5,
      next_visible_step_id: 6,
    });
    const program2 = createMockProgram({
      first_step_id: 5,
      next_visible_step_id: 7,
    });
    const post_data_non_rating = {
      data: '{"foo": ["bar"]}',
      forward_save: 'false',
    };
    const post_data_rating = {
      data: '{"foo": ["bar"]}',
      forward_save: 'true',
    };
    const post_data_rerating = {
      data: '{"foo": ["bar"], "excess_prerate_rating_complete": ["1"]}',
      forward_save: 'false',
    };
    const {session, request, sut} = getMocks();
    const meta_data_expected: BucketData = {liza_timestamp_rate_request: [0]};
    let metaDataCallCount = 0;
    let retryAttempCallCount = 0;
    let final_setMetadata_called = false;
    let retry_attempts_given: number;

    quote.setMetadata = (meta: BucketData) => {
      metaDataCallCount++;

      if (metaDataCallCount === 3) {
        final_setMetadata_called = true;
      }

      expect(meta).to.deep.equal(meta_data_expected);

      return quote;
    };

    quote.setRetryAttempts = (attempts: number) => {
      retryAttempCallCount++;
      retry_attempts_given = attempts;

      expect(retry_attempts_given).to.equal(0);

      return quote;
    };

    request.consumePostData = c => {
      if (metaDataCallCount === 0) {
        c(post_data_non_rating);
      } else if (metaDataCallCount === 2) {
        c(post_data_rerating);
      } else {
        c(post_data_rating);
      }

      return request;
    };

    // First tested step non rating
    sut
      .handlePost(
        (_: EncryptedKeys, __: string[], to_encrypt: BucketData) => to_encrypt,
        identity,
        <PositiveInteger>5,
        request,
        quote,
        program,
        session,
        false
      )
      // Second tested step rating
      .then(sut => {
        return sut.handlePost(
          (_: EncryptedKeys, __: string[], to_encrypt: BucketData) =>
            to_encrypt,
          identity,
          <PositiveInteger>6,
          request,
          quote,
          program,
          session,
          false
        );
      })
      // Third tested step rating
      .then(sut => {
        return sut.handlePost(
          (_: EncryptedKeys, __: string[], to_encrypt: BucketData) =>
            to_encrypt,
          identity,
          <PositiveInteger>7,
          request,
          quote,
          program2,
          session,
          false
        );
      })
      // Forth tested step previous rating
      .then(sut => {
        return sut.handlePost(
          (_: EncryptedKeys, __: string[], to_encrypt: BucketData) =>
            to_encrypt,
          identity,
          <PositiveInteger>6,
          request,
          quote,
          program,
          session,
          false
        );
      })
      // Final test values
      .then(() => {
        expect(retry_attempts_given).to.equal(0);
        expect(retryAttempCallCount).to.equal(2);
        expect(final_setMetadata_called, 'final setMetadata call made').to.be
          .true;
      });
    done();
  });

  it('Sets imported and top saved step on concluding save', done => {
    const {session, request, sut, quote} = getMocks();

    const program = createMockProgram({first_step_id: 1});
    program.isLastStep = () => {
      return true;
    };

    let imported_set = false;
    let top_saved_step = -1;

    quote.setImported = val => {
      imported_set = val;

      return quote;
    };

    quote.setTopSavedStepId = val => {
      top_saved_step = val;

      return quote;
    };

    const post_data = {
      data: '{"foo": ["bar"]}',
      forward_save: 'true',
    };

    request.consumePostData = c => {
      c(post_data);
      expect(imported_set).to.be.true;
      expect(top_saved_step).to.equal(2);
      done();

      return request;
    };

    program.lockOnImport = true;
    sut.handlePost(
      (_: EncryptedKeys, __: string[], to_encrypt: BucketData) => to_encrypt,
      identity,
      <PositiveInteger>2,
      request,
      quote,
      program,
      session,
      false
    );
  });

  [
    {
      label: 'Allows autosave when not next immediate step',
      autosave: true,
      step_to_post: 4,
      next_visible_step_id: 5,
      current_step: 2,
      top_saved_step: 2,
      top_visited_step: 2,
    },
    {
      label: 'Allows save when next immediate step',
      autosave: false,
      step_to_post: 4,
      next_visible_step_id: 4,
      current_step: 3,
      top_saved_step: 3,
      top_visited_step: 3,
    },
    {
      label: 'Allows save when next immediate step is hidden',
      autosave: false,
      step_to_post: 5,
      next_visible_step_id: 5,
      current_step: 3,
      top_saved_step: 3,
      top_visited_step: 3,
    },
  ].forEach(
    ({
      label,
      autosave,
      step_to_post,
      next_visible_step_id,
      current_step,
      top_saved_step,
      top_visited_step,
    }) => {
      it(label, done => {
        const quote = createMockQuote(
          current_step,
          top_saved_step,
          top_visited_step
        );
        const program = createMockProgram({
          first_step_id: 1,
          next_visible_step_id,
        });
        const session = createMockSession();
        const request = createMockRequest(session);
        const response = createMockResponse();
        const sut = getSut(response);

        let error_is_sent = false;
        response.error = (message, _, __) => {
          error_is_sent = true;
          // no errors are expected
          expect(message).to.be.equal(undefined);
          done();

          return '';
        };

        // TODO: assert on actual saves which requires a lot more mocking
        //  (BucketCipher is directly instantiated in Sut)
        request.consumePostData = () => {
          expect(error_is_sent).to.be.false;
          done();

          return request;
        };

        sut.handlePost(
          (_: EncryptedKeys, __: string[], to_encrypt: BucketData) =>
            to_encrypt,
          identity,
          <PositiveInteger>step_to_post,
          request,
          quote,
          program,
          session,
          autosave
        );
      });
    }
  );

  it('Encrypts delta prior to use of plaintext data', () => {
    const {session, request, sut, quote, program, processor} = getMocks();

    const post_delta = {
      foo: ['bar'],
      bar: ['baz'],
    };

    const expected_enc_delta = {
      foo: ['LookMaImEncrypted!'],
      bar: ['MeTooAndNoHands!'],
    };

    const post_data = {
      data: JSON.stringify(post_delta),
      forward_save: 'true',
    };

    request.consumePostData = c => {
      c(post_data);

      return request;
    };

    const orig_process_diff = processor.processDiff;
    let received_enc_delta = false;

    processor.processDiff = (
      program: Program,
      request: UserRequest,
      quote: ServerSideQuote,
      given_data: BucketData
    ) => {
      expect(given_data).to.deep.equal(expected_enc_delta);
      received_enc_delta = true;

      return orig_process_diff(program, request, quote, given_data);
    };

    program.encryptedKeys = {dummy: ['key']};

    const dummy_skeys = ['session', 'keys'];
    quote.getSessionKeys = () => dummy_skeys;

    const mockEncryptQuoteData = (
      given_keys: EncryptedKeys,
      given_skeys: string[],
      given_data: BucketData
    ) => {
      expect(given_keys).to.deep.equal(program.encryptedKeys);
      expect(given_skeys).to.deep.equal(dummy_skeys);
      expect(given_data).to.deep.equal(post_delta);

      // expected to modify data in-place
      Object.entries(expected_enc_delta).forEach(([field, value]) => {
        given_data[field] = value;
      });
    };

    return sut
      .handlePost(
        mockEncryptQuoteData,
        identity,
        <PositiveInteger>2,
        request,
        quote,
        program,
        session,
        false
      )
      .then(() => {
        // make sure we _actually_ did work for this test
        expect(received_enc_delta).to.be.true;
      });
  });
});

describe('Server#sendInit', () => {
  [
    {
      label: 'Sends init if user is authorized',
      is_authenticated: true,
      expected_action: [{action: 'gostep', id: 0}],
    },
    {
      label: 'Sends quote prompt if user is not authorized',
      is_authenticated: false,
      expected_action: [{action: 'quotePrompt'}],
    },
  ].forEach(({label, is_authenticated, expected_action}) => {
    it(label, done => {
      const response = createMockResponse();
      const dao = createMockDao();

      const session = createMockSession();
      const quote_id = '1234';
      const quote = createMockQuote(0, 0, 0, quote_id);
      const program = createMockProgram();

      quote.isAuthorized = _ => is_authenticated;

      response.from = (_, __, action) => {
        expect(action).to.deep.equal(expected_action);
        done();

        return '';
      };

      const sut = getSut(response, dao);
      const request = createMockRequest(session);
      sut.sendInit(
        request,
        quote,
        program,
        initDoc(quote),
        undefined,
        (_: EncryptedKeys, __: string[], to_decrypt: BucketData) => to_decrypt
      );
    });
  });

  it('Does not decrypt if a skey was used to authenticate', done => {
    const quote = createMockQuote(0, 0, 0, '1234');
    const request = createMockRequest(createMockSession());
    const response = createMockResponse();
    const sut = getSut(response, createMockDao());

    response.from = (_, __, ___) => {
      done();

      return '';
    };

    request.has_skey = () => true;

    const error_callback: any = (
      _: EncryptedKeys,
      __: string[],
      ___: BucketData
    ) => assert.fail();

    sut.sendInit(
      request,
      createMockQuote(0, 0, 0, '1234'),
      createMockProgram(),
      initDoc(quote),
      undefined,
      error_callback
    );
  });

  it('Executes sanitizer functions on the data before sending', done => {
    const data = {
      foo: ['bar'],
      baz: ['asd'],
    };

    const data_expected = {
      baz: ['asd'],
    };

    const request = createMockRequest(createMockSession());
    const response = createMockResponse();
    const quote = createMockQuote(0, 0, 0, '1234');

    let sanitizer_called = false;

    quote.getBucket = () =>
      <any>{
        getData: () => data,
        getDataByName: () => ['0'],
      };

    response.from = (_, given, __) => {
      expect(given?.data).to.deep.equal(data_expected);
      expect(sanitizer_called).to.be.true;
      done();

      return '';
    };

    const sanitizer_fn = (data_given: BucketData) => {
      expect(data_given).to.deep.equal(data);

      sanitizer_called = true;

      delete data_given['foo'];
    };

    getSut(response, createMockDao())
      .registerClientDataSanitizers([sanitizer_fn])
      .sendInit(
        request,
        quote,
        createMockProgram(),
        initDoc(quote),
        undefined,
        (_: EncryptedKeys, __: string[], to_decrypt: BucketData) => to_decrypt
      );
  });

  it('Decrypts fields in response', done => {
    const bucket_data = {
      foo: ['pretend encrypted'],
      bar: ['ciphertext'],
    };

    const expected_data_dec = {
      foo: ['DECRYPTED'],
      bar: ['PLAINTEXT'],
    };

    const request = createMockRequest(createMockSession());
    const response = createMockResponse();
    const quote = createMockQuote(0, 0, 0, '1234');
    const program = createMockProgram();

    quote.getBucket = () =>
      <any>{
        getData: () => bucket_data,
        getDataByName: () => ['0'],
      };

    response.from = (_, given, __) => {
      expect(given?.data).to.deep.equal(expected_data_dec);
      done();

      return '';
    };

    program.encryptedKeys = {dummy: ['key']};

    const dummy_skeys = ['session', 'keys'];
    quote.getSessionKeys = () => dummy_skeys;

    const mockDecryptQuoteData = (
      given_keys: EncryptedKeys,
      given_skeys: string[],
      given_data: BucketData
    ) => {
      expect(given_keys).to.deep.equal(program.encryptedKeys);
      expect(given_skeys).to.deep.equal(dummy_skeys);
      expect(given_data).to.deep.equal(bucket_data);

      return expected_data_dec;
    };

    getSut(response, createMockDao()).sendInit(
      request,
      quote,
      program,
      initDoc(quote),
      undefined,
      mockDecryptQuoteData
    );
  });

  // NB: This makes assumptions about implementation details: it's relying
  // on sanitization being performed just prior to the response.
  it('Does not decrypt fields until just before response is sent', done => {
    // still-encrypted bucket data
    const enc_data = {
      foo: ['encrypted'],
      baz: ['ciphertext'],
    };

    const request = createMockRequest(createMockSession());
    const response = createMockResponse();
    const quote = createMockQuote(0, 0, 0, '1234');

    let sanitizer_called = false;

    quote.getBucket = () =>
      <any>{
        getData: () => enc_data,
        getDataByName: () => ['0'],
      };

    response.from = () => {
      expect(sanitizer_called).to.be.true;
      done();

      return '';
    };

    const mockDecryptQuoteData = (
      _keys: EncryptedKeys,
      _skeys: string[],
      _data: BucketData
    ) => {
      // if this is called before sanitization, then this will replace the
      // data we expected in the sanitizer below
      return {foo: ['bucket was decrypted too early!']};
    };

    const sanitizer_fn = (data_given: BucketData) => {
      // should _not_ have been replaced by `mockDecryptQuoteData` above
      expect(data_given).to.deep.equal(enc_data);
      sanitizer_called = true;
    };

    getSut(response, createMockDao())
      .registerClientDataSanitizers([sanitizer_fn])
      .sendInit(
        request,
        quote,
        createMockProgram(),
        initDoc(quote),
        undefined,
        mockDecryptQuoteData
      );
  });
});

describe('Server#clearImport', () => {
  it('should not work for non-internal users', done => {
    const quote = createMockQuote();
    const session = createMockSession(false);
    const request = createMockRequest(session);
    const response = createMockResponse();
    const logger = createMockLogger();
    const dao = createMockDao();

    let log_level_actual: number;
    let log_msg_actual: string | number | Error;

    logger.log = (level, message: string) => {
      log_level_actual = level;
      log_msg_actual = message;

      return logger;
    };

    request.end = (data: string) => {
      expect(log_level_actual).to.equal(0);
      expect(log_msg_actual).to.contain('not allowed for non-internal users');

      const result = JSON.parse(data);
      expect(result).to.have.lengthOf(1);
      expect(result[0].success).to.be.equal(0);
      done();

      return request;
    };

    const sut = getSut(response, dao, logger);
    sut.clearImport(request, quote);
  });

  it('should not work for unimported quotes', done => {
    const quote = createMockQuote();
    const session = createMockSession(true);
    const request = createMockRequest(session);
    const response = createMockResponse();
    const logger = createMockLogger();
    const dao = createMockDao();

    let log_level_actual: number;
    let log_msg_actual: string | number | Error;

    logger.log = (level, message: string) => {
      log_level_actual = level;
      log_msg_actual = message;

      return logger;
    };

    quote.isImported = () => false;

    request.end = (data: string) => {
      expect(log_level_actual).to.equal(0);
      expect(log_msg_actual).to.contain(
        'cannot clear import on an imported quote'
      );

      const result = JSON.parse(data);
      expect(result).to.have.lengthOf(1);
      expect(result[0].success).to.be.equal(0);
      done();

      return request;
    };

    const sut = getSut(response, dao, logger);
    sut.clearImport(request, quote);
  });

  it('should catch update failures', done => {
    const quote = createMockQuote();
    const session = createMockSession(true);
    const request = createMockRequest(session);
    const response = createMockResponse();
    const logger = createMockLogger();
    const dao = createMockDao();
    const agent_name = 'foo';

    let log_level_actual: number;
    let log_msg_actual: string | number | Error;

    logger.log = (level, message: string) => {
      log_level_actual = level;
      log_msg_actual = message;

      return logger;
    };

    session.userName = () => agent_name;
    dao.unimportQuote = (q, name: string, _: Callback, cb: FailCallback) => {
      expect(q).to.equal(quote);
      expect(name).to.equal(agent_name);
      if (cb) {
        cb(quote, new Error());
      }

      return dao;
    };

    quote.isImported = () => true;

    request.end = (data: string) => {
      expect(log_level_actual).to.equal(0);
      expect(log_msg_actual).to.contain('could not save unlocking metadata');

      const result = JSON.parse(data);
      expect(result).to.have.lengthOf(1);
      expect(result[0].success).to.be.equal(0);
      done();

      return request;
    };

    const sut = getSut(response, dao, logger);
    sut.clearImport(request, quote);
  });

  it('should work for locked quotes', done => {
    const quote = createMockQuote();
    const session = createMockSession(true);
    const request = createMockRequest(session);
    const response = createMockResponse();
    const logger = createMockLogger();
    const dao = createMockDao();
    const agent_name = 'foo';

    let log_level_actual: number;
    let log_msg_actual: string | number | Error;

    logger.log = (level, message: string) => {
      log_level_actual = level;
      log_msg_actual = message;

      return logger;
    };

    session.userName = () => agent_name;
    dao.unimportQuote = (q, name: string, cb: Callback, _: FailCallback) => {
      expect(q).to.equal(quote);
      expect(name).to.equal(agent_name);
      if (cb) {
        cb(quote);
      }

      return dao;
    };

    quote.isImported = () => true;

    request.end = (data: string) => {
      expect(log_level_actual).to.be.undefined;
      expect(log_msg_actual).to.be.undefined;
      expect(data).to.be.undefined;
      done();

      return request;
    };

    const sut = getSut(response, dao, logger);
    sut.clearImport(request, quote);
  });
});

describe('Server#initQuote', async () => {
  const start_date = 1652700889;
  const quote_data = {
    key1: 'value1',
    key2: 'value2',
    currentStepId: 3,
    startDate: start_date,
  };
  const expected_quote_id = '12345';
  const response = createMockResponse();

  const dao = createMockDao();
  dao.pullQuote = () =>
    Promise.resolve(
      (() => {
        return <DocumentData>(<unknown>quote_data);
      })()
    );

  const program = createMockProgram();
  const quote = createMockQuote(0, 0, 0, expected_quote_id);
  const pcache = createMockCache();
  pcache.get = async (_: any) => program;

  const cache = createMockCache(pcache);

  const sut = getSut(response, dao);

  const agent_id = '11111';
  const username = 'foo@bar';
  const agency_num = 'AGT12345';
  const roles = ['bar', 'baz'];
  const agent_name = 'Some Agency';
  const entity_id = '9988776';
  const session = createMockSession(
    true,
    agent_id,
    username,
    agent_name,
    entity_id,
    agency_num,
    roles
  );

  const request = createMockRequest(session);

  it('Should fail without a valid kek', done => {
    sut.init(cache, createMockRater());
    const expected = 'Error: Missing liza kek';

    const error_callback = (e: string) => {
      testCallback([{actual: e, expect: expected}], done);
      done();
    };

    const success_callback = () => {
      done(new Error(`This test should have failed with: ${expected}`));
    };

    sut.initQuote(
      quote,
      program,
      request.getSession(),
      {},
      success_callback,
      error_callback
    );
  });

  it('Should error with no payload on QuoteNotFound errors', done => {
    const fake_kek = {
      [KekService.INTERNAL_KEY]: {
        '1': {
          is_active: true,
          key: Buffer.from(randomBytes(32).toString('base64')),
        },
      },
    };
    dao.pullQuote = () => Promise.reject(QuoteNotFound);
    sut.init(cache, createMockRater());

    const error_callback = (e: string) => {
      // We expect NO data object to return, as no quote was found
      testCallback([{expect: undefined, actual: e}], done);
      done();
    };

    const success_callback = (x: unknown) => {
      done(new Error(`This should not pass: ${x}`));
    };

    sut.initQuote(
      quote,
      program,
      request.getSession(),
      fake_kek,
      success_callback,
      error_callback
    );
  });

  it('Gets existing quote data from db then sets defaults', done => {
    const fake_kek = {
      [KekService.INTERNAL_KEY]: {
        '1': {
          is_active: true,
          key: Buffer.from(randomBytes(32)),
        },
      },
    };
    const quote_data = {
      key1: 'value1',
      key2: 'value2',
      currentStepId: 3,
      startDate: start_date,
    };
    let pull_quote_is_called = false;

    // make sure quote data is retrieved for our quote
    dao.pullQuote = quote_id =>
      Promise.resolve(
        (() => {
          testCallback([{expect: quote_id, actual: expected_quote_id}], done);
          pull_quote_is_called = true;

          return <DocumentData>(<unknown>quote_data);
        })()
      );

    let set_data_calls = 0;
    quote.setData = data_given => {
      set_data_calls++;

      switch (set_data_calls) {
        case 1:
          testCallback(
            [
              {
                expect: {
                  __internal_user: ['1'],
                  __liza_access_roles: ['bar', 'baz'],
                },
                actual: data_given,
              },
            ],
            done
          );
          break;
        case 2:
          testCallback(
            [
              {
                expect: {
                  __liza_step_id_current: ['3'],
                  __liza_start_date: ['1652700889'],
                },
                actual: data_given,
              },
            ],
            done
          );
          expect(data_given).to.deep.equal({
            __liza_step_id_current: ['3'],
            __liza_start_date: ['1652700889'],
          });
          break;
      }

      return quote;
    };

    quote.setMetadata = () => quote;

    // make sure the session username is set on the quote
    let set_username_is_called = false;
    quote.setUserName = given_user_name => {
      testCallback([{expect: given_user_name, actual: username}], done);
      set_username_is_called = true;
      return quote;
    };

    // make sure the session roles is set on the quote
    let set_roles_is_called = false;
    quote.setUserRoles = given_roles => {
      testCallback([{expect: given_roles, actual: roles}], done);
      set_roles_is_called = true;
      return quote;
    };

    quote.setAgencyNumber = given_agency_num => {
      testCallback([{expect: given_agency_num, actual: agency_num}], done);
      return quote;
    };

    quote.setStartDate = given_start_date => {
      testCallback([{expect: given_start_date, actual: start_date}], done);
      return quote;
    };
    quote.getStartDate = () => <UnixTimestamp>start_date;
    pcache.get = async (_: any) => program;
    sut.init(cache, createMockRater());

    const error_callback = (e: unknown) => {
      done(new Error(e instanceof Error ? e.message : `${e}`));
    };

    const success_callback = () => {
      testCallback(
        [
          {expect: true, actual: pull_quote_is_called},
          {expect: 2, actual: set_data_calls},
          {expect: true, actual: set_username_is_called},
          {expect: true, actual: set_roles_is_called},
        ],
        done
      );
      done();
    };

    sut.initQuote(
      quote,
      program,
      request.getSession(),
      fake_kek,
      success_callback,
      error_callback
    );
  });

  it('Defaults data on new quote and saves quote meta', async () => {
    const expected_quote_id = 12345;
    let pull_quote_is_called = false;

    // should not be called for new quote
    dao.pullQuote = quote_id =>
      Promise.resolve(
        (() => {
          expect(quote_id).to.be.equal(expected_quote_id);
          pull_quote_is_called = true;

          return null;
        })()
      );

    const program_id = 'foo';
    const program_ver = 'foover';
    const quote = createMockQuote(0, 0, 0, '' + expected_quote_id, program_id);

    const program = createMockProgram({
      first_step_id: 1,
      program_ver,
      next_visible_step_id: 2,
      group_exclusive_fields: {foogroup: ['foo']},
    });

    // Defaults are derived from this
    program.defaults = {foo: 'foodefault'};
    program.hasKnownType = _ => true;
    const start_date = Math.floor(+new Date() / 1000);

    let save_quote_is_called = false;
    dao.saveQuote = (quote, _class_data, success, _failure, save_data) => {
      save_quote_is_called = true;
      const expected_data = {
        agentId: agent_id,
        agentName: agent_name,
        agentEntityId: entity_id,
        startDate: start_date,
        programId: program_id,
        initialRatedDate: 0,
        importedInd: 0,
        boundInd: 0,
        importDirty: 0,
        syncInd: 0,
        notifyInd: 0,
        syncDate: 0,
        lastPremDate: 0,
        internal: 1,
        pver: program_ver,
        explicitLock: '',
        explicitLockStepId: 0,
      };
      expect(save_data).to.deep.equal(expected_data);
      success && success(quote);

      return dao;
    };

    let save_quote_meta_is_called = false;
    dao.saveQuoteMeta = (given_quote, _, success) => {
      save_quote_meta_is_called = true;
      expect(given_quote).to.be.equal(quote);
      success && success(given_quote);
    };

    const sut = getSut(
      response,
      dao,
      undefined,
      undefined,
      undefined,
      start_date
    );
    sut.init(createMockCache(), createMockRater());

    const request = createMockRequest(session);
    await sut.sendNewQuote(request, program, initDoc(quote));
    expect(pull_quote_is_called).to.be.false;
    expect(save_quote_is_called).to.be.true;
    expect(save_quote_meta_is_called).to.be.true;
  });
});

function getMocks() {
  const quote = createMockQuote();
  const program = createMockProgram();
  const session = createMockSession();
  const request = createMockRequest(session);
  const response = createMockResponse();
  const processor = createMockDapiProcessor();
  const sut = getSut(response, undefined, undefined, processor);

  return {
    quote: quote,
    program: program,
    session: session,
    request: request,
    response: response,
    processor,
    sut: sut,
  };
}

function getSut(
  response?: JsonServerResponse,
  dao?: MongoServerDao,
  logger?: PriorityLog,
  dapi_processor?: DapiProcessor,
  feature_flag?: FeatureFlag,
  start_ts = 0
): Sut {
  return new Sut(
    response || createMockResponse(),
    dao || createMockDao(),
    logger || createMockLogger(),
    {},
    dapi_processor || createMockDapiProcessor(),
    () => <UnixTimestamp>start_ts,
    feature_flag ?? createMockFeatureFlag(),
    <(program: Program) => (quote: ServerSideQuote) => any>{}
  );
}

function createMockQuote(
  current_step_id?: number,
  saved_step_id?: number,
  visited_step_id?: number,
  quote_id?: string,
  program_id?: string
): ServerSideQuote {
  const quote = {
    setProgramId: () => quote,
    getSessionKeys: () => ['foo'],
    getDownstreamSessionKeys: () => [''],
    setSessionKeys: () => quote,
    setDownstreamSessionKeys: () => quote,
    setLatestKeks: () => quote,
    getLatestKeks: () => {},
    setData: () => quote,
    setMetadata: () => quote,
    setUserName: () => quote,
    setUserRoles: () => quote,
    setAgentId: () => quote,
    setAgentName: () => quote,
    setAgentEntityId: () => quote,
    setAgencyNumber: () => quote,
    setInitialRatedDate: () => quote,
    setStartDate: () => quote,
    setImported: () => quote,
    setBound: () => quote,
    needsImport: () => quote,
    setCurrentStepId: () => quote,
    setTopVisitedStepId: () => quote,
    setTopSavedStepId: () => quote,
    setProgramVersion: () => quote,
    setExplicitLock: () => quote,
    setError: () => quote,
    setSaveError: () => quote,
    hasError: () => false,
    hasExpired: () => false,
    setCreditScoreRef: () => quote,
    setLastPremiumDate: () => quote,
    setRatedDate: () => quote,
    setRatingData: () => quote,
    setRetryAttempts: () => quote,
    on: () => quote,
    getAgentId: () => quote,
    getAgentName: () => quote,
    getUserRoles: () => quote,
    getInitialRatedDate: () => quote,
    getLastPremiumDate: () => quote,
    getExplicitLockReason: () => '',
    getExplicitLockStep: () => 0,
    getCurrentStepId: () => current_step_id || 0,
    getTopSavedStepId: () => saved_step_id || 0,
    getTopVisitedStepId: () => visited_step_id || 0,
    refreshData: () => quote,
    isLocked: () => false,
    getId: () => quote_id || '111111',
    getStartDate: () => '',
    getProgramId: () => program_id || 'foo',
    getProgramVersion: () => 'foo',
    isImported: () => false,
    isBound: () => false,
    getBucket: () => {
      return {
        getData: () => {},
        getDataByName: () => ['0'],
      };
    },
    classify: () => ({}),
    setLastPersistedFieldState: () => quote,
    isAuthorized: () => true,
    getDataByName: () => ['foo'],
    setNonpersistent: () => quote,
    getNonpersistent: () => [],
    refreshStepDelta: () => quote,
  };

  return <ServerSideQuote>(<unknown>quote);
}

function createMockLogger(cb = (_level: any, _msg: any) => {}): PriorityLog {
  const log_cb = (level: any, msg: any) => {
    cb(level, msg);
  };

  return <PriorityLog>(<unknown>{
    PRIORITY_ERROR: 0,
    PRIORITY_IMPORTANT: 1,
    PRIORITY_DB: 2,
    PRIORITY_INFO: 3,
    PRIORITY_SOCKET: 4,
    debug: log_cb,
    info: log_cb,
    notice: log_cb,
    warning: log_cb,
    error: log_cb,
    critical: log_cb,
    alert: log_cb,
    emergency: log_cb,
    log: log_cb,
  });
}

function createMockRequest(
  session?: UserSession,
  end_callback = () => {}
): UserRequest {
  return <UserRequest>(<unknown>{
    end: end_callback,
    getSession: () => session || createMockSession(),
    consumePostData: (c: any) => {
      c({});
    },
    has_skey: () => false,
  });
}

function createMockResponse(
  from_cb = (_actions: any) => {},
  error_cb = (_msg: any, _actions: any) => {}
): JsonServerResponse {
  return <JsonServerResponse>(<unknown>{
    from: (_quote: any, _data: any, action: any) => from_cb(action),
    error: (msg: any, action: any, _caption: any) => error_cb(msg, action),
  });
}

function createMockSession(
  internal?: boolean,
  agent_id?: string,
  username?: string,
  agent_name?: string,
  entity_id?: string,
  agency_num?: string,
  roles?: string[]
): UserSession {
  return <UserSession>(<unknown>{
    isInternal: () => internal,
    agentId: () => agent_id || '',
    userName: () => username || '',
    userRoles: () => roles || [],
    agentName: () => agent_name || '',
    agentEntityId: () => entity_id || '',
    agencyNumber: () => agency_num || '',
  });
}

function createMockDapiProcessor(): DapiProcessor {
  const processor: any = createStubInstance(DapiProcessor);
  const filtered = {};
  const dapis: any[] = [];
  const meta_clear = {};
  const rdiff = {};

  processor.processDiff = () => {
    return {
      filtered: filtered,
      dapis: dapis,
      meta_clear: meta_clear,
      rdiff: rdiff,
    };
  };

  processor.monitorMetadataPromise = () => {};
  processor.triggerDapi = () => {};

  return processor;
}

function createMockProgram(
  overrides: {
    first_step_id?: number;
    program_ver?: string;
    next_visible_step_id?: number;
    group_exclusive_fields?: Record<string, any>;
    classes?: Record<string, {is: boolean}>;
  } = {}
): Program {
  return <Program>(<unknown>{
    getId: () => 'foo',
    version: overrides?.program_ver || 'foo',
    initQuote: () => {},
    getFirstStepId: () => overrides?.first_step_id || 1,
    isStepVisible: () => true,
    isLastStep: () => false,
    isManageQuoteStep: () => false,
    isIndicationQuoteStep: () => false,
    // eslint-disable-next-line no-sparse-arrays
    steps: [
      ,
      {id: 'Manage Quote', type: 'manage'},
      {id: 'Step1', type: ''},
      {id: 'Step2', type: ''},
      {id: 'Step3', type: ''},
      {id: 'Step4', type: ''},
      {id: 'Step5', type: ''},
      {id: 'Step6', type: ''},
    ],
    rateSteps: [, false, false, false, false, false, true, true],
    processNaFields: () => {},
    getNextVisibleStep: () => overrides?.next_visible_step_id || 2,
    classify: () => {
      return overrides?.classes || {};
    },
    submit: () => null,
    secureFields: [],
    groupExclusiveFields: overrides?.group_exclusive_fields || {},
    beforeLoadStep: () => {},
    meta: {
      groups: {},
      fields: {},
    },
    downstreamServices: ['someService'],
    whens: {},
  });
}

function createMockCache(pcache?: any): any {
  return {
    get: (_: any) => Promise.resolve(pcache || createMockCache()),
  };
}

function createMockRater() {
  return {
    init: () => {},
  };
}

function createMockFeatureFlag(override = false): FeatureFlag {
  return <FeatureFlag>(<unknown>{
    isEnabled: (val: string) =>
      Promise.resolve(val === 'liza_autosave' || override),
  });
}

function createMockDao(
  min_quote_id?: number,
  max_quote_id?: number
): MongoServerDao {
  return <MongoServerDao>(<unknown>{
    on: () => createMockDao(),
    init: () => {},
    pullQuote: () => {},
    saveQuote: () => {},
    saveQuoteMeta: () => {},
    saveQuoteMetaDirect: () => {},
    saveQuoteState: (_: any, cb: any) => {
      cb && cb();
    },
    getMinQuoteId: (c: any) => c(min_quote_id || 0),
    getMaxQuoteId: (c: any) => c(max_quote_id || 999999),
    mergeData: () => {},
    publishQuote: () => Promise.resolve(),
  });
}

const initDoc =
  (quote: ServerSideQuote) => (_flag: FeatureFlag) => (_data: DocumentData) =>
    TE.of(quote);

type TestCallback = {
  expect: any;
  actual: any;
};
const testCallback = (tests: TestCallback[], done: Function) => {
  try {
    for (const test of tests) {
      expect(test.actual).to.deep.equal(test.expect);
    }
  } catch (e) {
    const err = e instanceof Error ? e.message : e;
    done(new Error(`${err}`));
  }
};
