/**
 * Tests DapiProcessor
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {DataProcessor} from '../../../src/server/request/DataProcessor';
import {UserRequest} from '../../../src/server/request/UserRequest';
import {UserSession} from '../../../src/server/request/UserSession';
import {DapiProcessor as Sut} from '../../../src/server/dapi/DapiProcessor';
import {Program} from '../../../src/program/Program';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {PriorityLog} from '../../../src/server/log/PriorityLog';
import {MongoServerDao} from '../../../src/server/db/MongoServerDao';
import {FieldData} from '../../../src/server/meta/DapiMetaSource';
import {EmptyBucketData} from '../../../src/bucket/Bucket';

/* eslint-disable @typescript-eslint/no-var-requires, node/no-unpublished-require */
const chai = require('chai');
const expect = chai.expect;
const sinon = require('sinon');
/* eslint-enable @typescript-eslint/no-var-requires, node/no-unpublished-require */

describe('DapiProcessor', () => {
  describe('#processDiff', () => {
    it('', () => {
      Sut;
    });
  });

  describe('#monitorMetadataPromise', () => {
    it('', () => {
      Sut;
    });
  });

  describe('#triggerDapi', () => {
    it('triggers a dapi', done => {
      let processDiff_called = false;

      const override_expected = {
        mapped1: ['foo'],
      };

      const meta_clear = {foo: 'meta_to_clear'};

      const quote = createMockQuote();
      const program = createMockProgram();
      const session = createMockSession();
      const request = createMockRequest(session);
      const dao = createMockDao();
      const data_processor = createMockDataProcessor();

      dao.mergeData = (_: any, data: any) => {
        expect(data).to.deep.equal(meta_clear);

        done();

        return dao;
      };

      data_processor.processDiff = (
        _: any,
        __: any,
        ___: any,
        ____: any,
        _____: any,
        override: any
      ) => {
        if (!processDiff_called) {
          expect(override).to.deep.equal(override_expected);
        }

        processDiff_called = true;

        return {
          dapis: [Promise.resolve(<FieldData>{})],
          meta_update: <EmptyBucketData>(<unknown>meta_clear),
          filtered: {},
          rdiff: {},
        };
      };

      const sut = getSut(dao, undefined, data_processor);

      program.meta.fields['some_dapi_name'] = {
        desc: 'foo',
        dapi: {
          name: '',
          value: '',
          when: '',
          mapsrc: {
            input1: 'mapped1',
          },
          mapdest: {
            foo: 'bar',
          },
        },
      };

      sut.triggerDapi(program, request, quote, 'some_dapi_name');
    });

    it('handles unknown dapis', done => {
      const quote = createMockQuote();
      const program = createMockProgram();
      const session = createMockSession();
      const request = createMockRequest(session);
      const dao = createMockDao();
      const logger = createMockLogger();
      const data_processor = createMockDataProcessor();

      logger.log = (level: any, message: any) => {
        expect(level).to.equal(0);
        expect(message).to.equal('Cannot trigger unknown dapi some_dapi_name');
        done();

        return logger;
      };

      const sut = getSut(dao, logger, data_processor);

      sut.triggerDapi(program, request, quote, 'some_dapi_name');
    });

    it('handles dapis with no map field', done => {
      const quote = createMockQuote();
      const program = createMockProgram();
      const session = createMockSession();
      const request = createMockRequest(session);
      const dao = createMockDao();
      const logger = createMockLogger();
      const data_processor = createMockDataProcessor();

      logger.log = (level: any, message: any) => {
        expect(level).to.equal(0);
        expect(message).to.equal(
          'Cannot trigger dapi some_dapi_name because it has no defined sources'
        );
        done();

        return logger;
      };

      const sut = getSut(dao, logger, data_processor);

      program.meta.fields['some_dapi_name'] = {
        desc: 'foo',
        dapi: {
          name: '',
          value: '',
          when: '',
          mapsrc: {},
          mapdest: {},
        },
      };

      sut.triggerDapi(program, request, quote, 'some_dapi_name');
    });
  });
});

function getSut(
  dao?: MongoServerDao,
  logger?: PriorityLog,
  data_processor?: DataProcessor
): Sut {
  return new Sut(
    dao || createMockDao(),
    logger || createMockLogger(),
    () => <UnixTimestamp>123,
    data_processor || createMockDataProcessor()
  );
}

function createMockQuote(
  current_step_id?: number,
  saved_step_id?: number,
  visited_step_id?: number,
  quote_id?: string,
  program_id?: string
): ServerSideQuote {
  const quote = {
    setData: () => {},
    setMetadata: () => {},
    setUserName: () => {},
    setUserRoles: () => {},
    setAgentId: () => {},
    setAgentName: () => {},
    setAgentEntityId: () => {},
    setAgencyNumber: () => {},
    setInitialRatedDate: () => {},
    setStartDate: () => {},
    setImported: () => {},
    setBound: () => {},
    needsImport: () => quote,
    setCurrentStepId: () => {},
    setTopVisitedStepId: () => {},
    setTopSavedStepId: () => {},
    setProgramVersion: () => {},
    setExplicitLock: () => {},
    setError: () => {},
    setSaveError: () => quote,
    hasError: () => false,
    hasExpired: () => false,
    setCreditScoreRef: () => {},
    setLastPremiumDate: () => {},
    setRatedDate: () => {},
    setRatingData: () => {},
    setRetryAttempts: () => {},
    on: () => {},
    getAgentId: () => {},
    getAgentName: () => {},
    getUserRoles: () => {},
    getInitialRatedDate: () => {},
    getLastPremiumDate: () => {},
    getExplicitLockReason: () => '',
    getExplicitLockStep: () => 0,
    getCurrentStepId: () => current_step_id || 0,
    getTopSavedStepId: () => saved_step_id || 0,
    getTopVisitedStepId: () => visited_step_id || 0,
    refreshData: () => {},
    isLocked: () => false,
    getId: () => quote_id || '111111',
    getStartDate: () => '',
    getProgramId: () => program_id || 'foo',
    getProgramVersion: () => 'foo',
    isImported: () => false,
    isBound: () => false,
    getBucket: () => {
      return {
        getData: () => {},
        getDataByName: () => ['0'],
      };
    },
    classify: () => ({}),
    setLastPersistedFieldState: () => quote,
    isAuthorized: () => true,
    getDataByName: () => ['foo'],
    setNonpersistent: () => quote,
    getNonpersistent: () => [],
    refreshStepDelta: () => quote,
  };

  return <ServerSideQuote>(<unknown>quote);
}

function createMockLogger(cb = (_level: any, _msg: any) => {}): PriorityLog {
  const log_cb = (level: any, msg: any) => {
    cb(level, msg);
  };

  return <PriorityLog>(<unknown>{
    PRIORITY_ERROR: 0,
    PRIORITY_IMPORTANT: 1,
    PRIORITY_DB: 2,
    PRIORITY_INFO: 3,
    PRIORITY_SOCKET: 4,
    debug: log_cb,
    info: log_cb,
    notice: log_cb,
    warning: log_cb,
    error: log_cb,
    critical: log_cb,
    alert: log_cb,
    emergency: log_cb,
    log: log_cb,
  });
}

function createMockRequest(
  session?: UserSession,
  end_callback = () => {}
): UserRequest {
  return <UserRequest>(<unknown>{
    end: end_callback,
    getSession: () => session || createMockSession(),
    consumePostData: (c: any) => {
      c({});
    },
  });
}

function createMockSession(
  internal?: boolean,
  agent_id?: string,
  username?: string,
  agent_name?: string,
  entity_id?: string,
  agency_num?: string,
  roles?: string[]
): UserSession {
  return <UserSession>(<unknown>{
    isInternal: () => internal,
    agentId: () => agent_id || '',
    userName: () => username || '',
    userRoles: () => roles || [],
    agentName: () => agent_name || '',
    agentEntityId: () => entity_id || '',
    agencyNumber: () => agency_num || '',
  });
}

function createMockDataProcessor(): DataProcessor {
  const processor = sinon.createStubInstance(DataProcessor);
  const filtered = {};
  const dapis: any[] = [];
  const meta_clear = {};
  const rdiff = {};

  processor.processDiff = () => {
    return {
      filtered: filtered,
      dapis: dapis,
      meta_clear: meta_clear,
      rdiff: rdiff,
    };
  };

  processor.monitorMetadataPromise = () => {};
  processor.triggerDapi = () => {};

  return processor;
}

function createMockProgram(
  overrides: {
    first_step_id?: number;
    program_ver?: string;
    next_visible_step_id?: number;
    group_exclusive_fields?: Record<string, any>;
    classes?: Record<string, {is: boolean}>;
  } = {}
): Program {
  return <Program>(<unknown>{
    getId: () => 'foo',
    version: overrides?.program_ver || 'foo',
    getFirstStepId: () => overrides?.first_step_id || 1,
    isStepVisible: () => true,
    isLastStep: () => false,
    isManageQuoteStep: () => false,
    isIndicationQuoteStep: () => false,
    // eslint-disable-next-line no-sparse-arrays
    steps: [
      ,
      {id: 'Manage Quote', type: 'manage'},
      {id: 'Step1', type: ''},
      {id: 'Step2', type: ''},
      {id: 'Step3', type: ''},
      {id: 'Step4', type: ''},
      {id: 'Step5', type: ''},
      {id: 'Step6', type: ''},
    ],
    processNaFields: () => {},
    getNextVisibleStep: () => overrides?.next_visible_step_id || 2,
    classify: () => {
      return overrides?.classes || {};
    },
    secureFields: [],
    groupExclusiveFields: overrides?.group_exclusive_fields || {},
    beforeLoadStep: () => {},
    meta: {
      groups: {},
      fields: {},
    },
  });
}

function createMockDao(
  min_quote_id?: number,
  max_quote_id?: number
): MongoServerDao {
  return <MongoServerDao>(<unknown>{
    on: () => createMockDao(),
    init: () => {},
    pullQuote: () => {},
    saveQuote: () => {},
    saveQuoteMeta: () => {},
    saveQuoteState: (_: any, cb: any) => {
      cb && cb();
    },
    getMinQuoteId: (c: any) => c(min_quote_id || 0),
    getMaxQuoteId: (c: any) => c(max_quote_id || 999999),
    mergeData: () => {},
    publishQuote: () => Promise.resolve(),
  });
}
