/**
 * Test GcpDataApi
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {GoogleAuth, IdTokenClient} from 'google-auth-library';
import {GcpDataApi as Sut} from '../../../src/server/dapi/GcpDataApi';
import {expect} from 'chai';

describe('GcpDataApi', () => {
  it('Uses id token client to make call', done => {
    const url = 'https://example.com';
    const data = {foo: 'bar'};

    const idTokenClient = getMockClient({});
    const auth = getMockAuth(idTokenClient);

    const callback = (err: NullableError, response: any) => {
      expect(err).to.be.null;
      expect(response).to.be.null;
      done();
    };

    new Sut(url, auth).request(data, callback);
  });

  it('Gets id token client with url', done => {
    const url = 'https://example.com';
    const data = {foo: 'bar'};

    const idTokenClient = getMockClient({});
    const auth = getMockAuth(idTokenClient);

    auth.getIdTokenClient = (url_given: string) => {
      expect(url_given).to.equal(url);

      return Promise.resolve(idTokenClient);
    };

    const callback = (err: NullableError, response: any) => {
      expect(err).to.be.null;
      expect(response).to.be.null;
      done();
    };

    new Sut(url, auth).request(data, callback);
  });

  it('Makes POST request to the correct url with the provided data', done => {
    const url = 'https://example.com';
    const data = {foo: 'bar'};

    const idTokenClient = getMockClient({});
    const auth = getMockAuth(idTokenClient);

    idTokenClient.request = <typeof idTokenClient.request>((
      opts: any,
      cb: any
    ) => {
      expect(opts.url).to.equal(url);
      expect(opts.data).to.equal(data);
      expect(opts.method).to.equal('POST');

      cb(null, {});
    });

    const callback = (err: NullableError, response: any) => {
      expect(err).to.be.null;
      expect(response).to.be.null;
      done();
    };

    new Sut(url, auth).request(data, callback);
  });

  it('Calls callback with error getting id token client', done => {
    const url = 'https://example.com';
    const data = {foo: 'bar'};
    const error = new Error('Could not get id token client');

    const idTokenClient = getMockClient({});
    const auth = getMockAuth(idTokenClient);

    auth.getIdTokenClient = () => Promise.reject(error);

    const callback = (err: NullableError, response: any) => {
      expect(err).deep.equal(error);
      expect(response).to.be.null;
      done();
    };

    new Sut(url, auth).request(data, callback);
  });

  it('Calls callback with error client request error', done => {
    const url = 'https://example.com';
    const data = {foo: 'bar'};
    const error = new Error('Could not make request');

    const idTokenClient = getMockClient({});
    const auth = getMockAuth(idTokenClient);

    idTokenClient.request = <typeof idTokenClient.request>((
      _opts: any,
      cb: any
    ) => {
      cb(error, null);
    });

    const callback = (err: NullableError, response: any) => {
      expect(err).deep.equal(error);
      expect(response).to.be.null;
      done();
    };

    new Sut(url, auth).request(data, callback);
  });

  it('Calls callback with error when client returns no data', done => {
    const url = 'https://example.com';
    const data = {foo: 'bar'};
    const expected_error_msg = 'No response received from GCP';

    const idTokenClient = getMockClient({});
    const auth = getMockAuth(idTokenClient);

    idTokenClient.request = <typeof idTokenClient.request>((
      _opts: any,
      cb: any
    ) => {
      cb(null, null);
    });

    const callback = (err: NullableError, response: any) => {
      expect(err?.message).equal(expected_error_msg);
      expect(response).to.be.null;
      done();
    };

    new Sut(url, auth).request(data, callback);
  });
});

const getMockAuth = (client: IdTokenClient): GoogleAuth => <GoogleAuth>(<
    unknown
  >{
    getIdTokenClient: () => Promise.resolve(client),
  });

const getMockClient = (response: any, err: any = null): IdTokenClient =>
  <IdTokenClient>{
    request: (_opts: any, callback: any) => {
      callback(err, response);
    },
  };
