/**
 * Test SessionKeyDataApi
 *
 *  Copyright (C) 2010-2025 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {SessionKeyDataApi as Sut} from '../../../src/server/dapi/SessionKeyDataApi';
import {DataApi, DataApiInput} from '../../../src/dapi/DataApi';
import {DataApiResult} from '../../../src/dapi/DataApi';
import {DapiAction} from '../../../src/dapi/DataApi';
import {assert, expect} from 'chai';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';

describe('SessionKeyDataApi', () => {
  it('Augments data with active session key for downstream system ID', done => {
    const data_given = {foo: 'bar'};
    const id = 'some_id';
    const action = DapiAction.TRIGGER;
    const mock_quote = createMockQuote();
    const session_keys = ['Encryptedkeys: secret-13'];
    const downstream_id = 'gl_property';
    let expected_to_call_get_downstream_session_keys = false;

    const downStreamKeys: Record<string, string[]> = {
      liza: ['Encryptedkeys: secret-11'],
      gl_property: session_keys,
    };

    const data_expected = {
      foo: 'bar',
      downstream_session_keys: session_keys,
    };

    mock_quote.getDownstreamSessionKeys = () => {
      expected_to_call_get_downstream_session_keys = true;
      return downStreamKeys;
    };

    const api: DataApi = {
      request: (
        data_actual: DataApiInput,
        _callback: NodeCallback<DataApiResult>
      ) => {
        expect(data_actual).to.deep.equal(data_expected);
        expect(expected_to_call_get_downstream_session_keys).to.be.true;
        done();

        return api;
      },
    };

    new Sut(mock_quote, downstream_id, api).request(
      data_given,
      () => {},
      id,
      action
    );
  });

  it('Does not augment data when downstream system ID is not provided', done => {
    const data_given = {foo: 'bar'};
    const id = 'some_id';
    const action = DapiAction.TRIGGER;
    const mock_quote = createMockQuote();
    const empty_downstream_id = '';
    let expected_to_call_get_downstream_session_keys = false;

    const downStreamKeys: Record<string, string[]> = {
      liza: ['Encryptedkeys: secret-11'],
    };

    const data_expected = {
      foo: 'bar',
    };

    mock_quote.getDownstreamSessionKeys = () => {
      expected_to_call_get_downstream_session_keys = true;
      return downStreamKeys;
    };

    const api: DataApi = {
      request: (
        data_actual: DataApiInput,
        _callback: NodeCallback<DataApiResult>
      ) => {
        expect(data_actual).to.deep.equal(data_expected);
        expect(expected_to_call_get_downstream_session_keys).to.be.false;
        done();

        return api;
      },
    };

    new Sut(mock_quote, empty_downstream_id, api).request(
      data_given,
      () => {},
      id,
      action
    );
  });

  it('Calls back with error when data is not a valid object', done => {
    const expected_err_msg =
      'SessionKey http dapi triggered without a valid data object';

    const api: DataApi = {
      request: (
        _data: DataApiInput,
        _callback: NodeCallback<DataApiResult>,
        _id?: string,
        _action?: DapiAction
      ) => {
        assert.fail('Should not have called api');

        return api;
      },
    };

    const callback = (err: NullableError, response: any) => {
      expect(err?.message).to.equal(expected_err_msg);
      expect(response).to.be.null;
      done();
    };

    new Sut(createMockQuote(), 'foo', api).request('', callback);
  });

  it('Calls back with error with invalid downstream session keys', done => {
    const expected_err_msg =
      "No session keys found for DataAPI 'some_id' downstream system 'invalid-downstream-system'";

    const mock_quote = createMockQuote();
    const data_given = {foo: 'bar'};
    const id = 'some_id';
    const action = DapiAction.TRIGGER;

    const downStreamKeys: Record<string, string[]> = {
      liza: ['Encryptedkeys: secret-11'],
    };

    mock_quote.getDownstreamSessionKeys = () => {
      return downStreamKeys;
    };

    const api: DataApi = {
      request: (
        _data: DataApiInput,
        _callback: NodeCallback<DataApiResult>,
        _id?: string,
        _action?: DapiAction
      ) => {
        assert.fail('Should not have called api');

        return api;
      },
    };

    const callback = (err: NullableError, response: any) => {
      expect(err?.message).to.equal(expected_err_msg);
      expect(response).to.be.null;
      done();
    };

    new Sut(mock_quote, 'invalid-downstream-system', api).request(
      data_given,
      callback,
      id,
      action
    );
  });
});

function createMockQuote(): ServerSideQuote {
  return <ServerSideQuote>(<unknown>{
    getDownstreamSessionKeys: () => [],
  });
}
