/**
 * Test WebhookHttpDataApi
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {WebhookHttpDataApi as Sut} from '../../../src/server/dapi/WebhookHttpDataApi';
import {DataApi, DataApiInput} from '../../../src/dapi/DataApi';
import {DataApiResult} from '../../../src/dapi/DataApi';
import {DapiAction} from '../../../src/dapi/DataApi';
import {TokenId} from '../../../src/server/token/Token';
import {assert, expect} from 'chai';

describe('WebhookHttpDataApi', () => {
  it('Augments data with a token', done => {
    const data_given = {foo: 'bar'};
    const token = <TokenId>'some token';
    const id = 'some_id';
    const action = DapiAction.TRIGGER;

    const data_expected = {foo: 'bar', webhook_trigger_some_id: token};

    const api: DataApi = {
      request: (
        data_actual: DataApiInput,
        _callback: NodeCallback<DataApiResult>,
        id_actual?: string,
        action_actual?: DapiAction,
        token_actual?: TokenId
      ) => {
        expect(data_actual).to.deep.equal(data_expected);
        expect(id_actual).to.equal(id);
        expect(action_actual).to.equal(action);
        expect(token_actual).to.equal(token);
        done();

        return api;
      },
    };

    new Sut(api).request(data_given, () => {}, id, action, token);
  });

  it('Calls back with error on missing token', done => {
    const expected_err_msg =
      'Webhook http dapi triggered without a valid token';

    const api: DataApi = {
      request: (
        _data: DataApiInput,
        _callback: NodeCallback<DataApiResult>,
        _id?: string,
        _action?: DapiAction,
        _token?: TokenId
      ) => {
        assert.fail('Should not have called api');

        return api;
      },
    };

    const callback = (err: NullableError, response: any) => {
      expect(err?.message).to.equal(expected_err_msg);
      expect(response).to.be.null;
      done();
    };

    new Sut(api).request({foo: 'bar'}, callback);
  });

  it('Calls back with error on string input', done => {
    const expected_err_msg =
      'Webhook http dapi triggered without a valid data object';

    const api: DataApi = {
      request: (
        _data: DataApiInput,
        _callback: NodeCallback<DataApiResult>,
        _id?: string,
        _action?: DapiAction,
        _token?: TokenId
      ) => {
        assert.fail('Should not have called api');

        return api;
      },
    };

    const callback = (err: NullableError, response: any) => {
      expect(err?.message).to.equal(expected_err_msg);
      expect(response).to.be.null;
      done();
    };

    new Sut(api).request(
      'foo',
      callback,
      'some_id',
      DapiAction.TRIGGER,
      <TokenId>'foo'
    );
  });
});
