/**
 * Tests Document ID Lock Manager
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Semaphore as SemaphoreType} from '../../../src/server/lock/Semaphore';
import {DocumentIdLockManager} from '../../../src/server/lock/DocumentIdLockManager';
import {UserRequest} from '../../../src/server/request/UserRequest';
import {assert, expect} from 'chai';

describe('DocumentIdLockManager', () => {
  describe('#init', () => {
    it('should free stale locks on an interval', () => {
      let wlockFreeStaleCalled = false;
      let rlockFreeStaleCalled = false;
      let setIntervalCalled = false;

      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const lockManager = new DocumentIdLockManager(wlock, rlock);

      // Mock the freeStale method of the write lock semaphore
      wlock.freeStale = (interval: number, callback: (id: number) => void) => {
        wlockFreeStaleCalled = true;

        expect(interval).to.equal(
          DocumentIdLockManager.CLEAR_STALE_INTERVAL_MS
        );

        callback(123);

        return wlock;
      };

      // Mock the freeStale method of the read lock semaphore
      rlock.freeStale = (interval: number, callback: (id: number) => void) => {
        rlockFreeStaleCalled = true;

        expect(interval).to.equal(
          DocumentIdLockManager.CLEAR_STALE_INTERVAL_MS
        );

        callback(234);

        return rlock;
      };
      const old_setInterval = global.setInterval;

      global.setInterval = (
        callback: (...args: any[]) => void,
        interval: number
      ) => {
        expect(interval).to.equal(
          DocumentIdLockManager.CLEAR_STALE_INTERVAL_MS
        );

        setIntervalCalled = true;

        callback();

        return <NodeJS.Timeout>{};
      };

      lockManager.init();

      expect(wlockFreeStaleCalled, '#freeStale called for write lock').to.be
        .true;
      expect(rlockFreeStaleCalled, '#freeStale called for read lock').to.be
        .true;
      expect(setIntervalCalled, 'setInterval called').to.be.true;

      global.setInterval = old_setInterval;
    });

    it('should only setup an interval once', () => {
      const old_setInterval = global.setInterval;

      try {
        let setIntervalCallCount = 0;

        global.setInterval = (_: (...args: any[]) => void, __: number) => {
          setIntervalCallCount++;

          return <NodeJS.Timeout>{};
        };

        new DocumentIdLockManager(getMockSemaphore(), getMockSemaphore())
          .init()
          .init();

        expect(setIntervalCallCount).to.equal(1);
      } catch (e) {
        assert.fail(e);
      } finally {
        global.setInterval = old_setInterval;
      }
    });
  });

  describe('#onLockAquired', () => {
    it('should register a callback to be called when a lock is acquired', () => {
      const quote_id = 123;

      let callbackCalled = false;

      const request = getMockUserRequest();
      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const lockManager = new DocumentIdLockManager(wlock, rlock);

      const cb = (id: number) => {
        expect(id).to.equal(quote_id);

        callbackCalled = true;
      };

      lockManager.onLockAquired(cb);

      lockManager.acquireRwLock(quote_id, request, () => {});

      expect(callbackCalled).to.be.true;
    });
  });

  describe('locking', () => {
    it('#acquireRwLock aquires read and write locks', () => {
      const quote_id = 123;

      let wlockAcquireCalled = false;
      let rlockAcquireCalled = false;

      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const lockManager = new DocumentIdLockManager(rlock, wlock);

      wlock.acquire = (id: number, callback: (free: () => void) => void) => {
        wlockAcquireCalled = true;

        expect(id).to.equal(quote_id);

        callback(() => {});
      };

      rlock.acquire = (id: number, callback: (free: () => void) => void) => {
        rlockAcquireCalled = true;

        expect(id).to.equal(quote_id);

        callback(() => {});
      };

      lockManager.acquireRwLock(quote_id, getMockUserRequest(), () => {});

      expect(wlockAcquireCalled, 'write lock acquired').to.be.true;
      expect(rlockAcquireCalled, 'read lock acquired').to.be.true;
    });

    it('#acquireWriteLock aquires just a write lock', () => {
      const quote_id = 123;

      let wlockAcquireCalled = false;

      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const lockManager = new DocumentIdLockManager(rlock, wlock);

      wlock.acquire = (id: number, callback: (free: () => void) => void) => {
        wlockAcquireCalled = true;

        expect(id).to.equal(quote_id);

        callback(() => {});
      };

      rlock.acquire = (_: number, __: (free: () => void) => void) => {
        assert.fail('A read lock should not have been acquired');
      };

      lockManager.acquireWriteLock(quote_id, getMockUserRequest(), () => {});

      expect(wlockAcquireCalled, 'write lock acquired').to.be.true;
    });

    it('#acquireReadLock aquires just a read lock', () => {
      const quote_id = 123;

      let rlockAcquireCalled = false;

      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const lockManager = new DocumentIdLockManager(rlock, wlock);

      rlock.acquire = (id: number, callback: (free: () => void) => void) => {
        rlockAcquireCalled = true;

        expect(id).to.equal(quote_id);

        callback(() => {});
      };

      wlock.acquire = (_: number, __: (free: () => void) => void) => {
        assert.fail('A write lock should not have been acquired');
      };

      lockManager.acquireReadLock(quote_id, getMockUserRequest(), () => {});

      expect(rlockAcquireCalled, 'read lock acquired').to.be.true;
    });

    it('after acquiring a lock, the lock is freed on request end', () => {
      const quote_id = 123;

      let freeCalled = false;
      let freeCb = () => {};

      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const request = getMockUserRequest();
      const lockManager = new DocumentIdLockManager(rlock, wlock);

      rlock.acquire = (id: number, callback: (free: () => void) => void) => {
        expect(id).to.equal(quote_id);

        callback(() => {
          freeCalled = true;
        });
      };

      request.once = (_event: string, cb: () => void) => {
        freeCb = cb;

        return request;
      };

      lockManager.acquireReadLock(quote_id, request, () => {});

      freeCb();

      expect(freeCalled, 'free was not called automatically').to.be.true;
    });

    it('if manual is specified, the lock is not automatically freed', () => {
      const quote_id = 123;

      let freeCalled = false;

      const wlock = getMockSemaphore();
      const rlock = getMockSemaphore();
      const request = getMockUserRequest();
      const lockManager = new DocumentIdLockManager(rlock, wlock);

      rlock.acquire = (id: number, callback: (free: () => void) => void) => {
        expect(id).to.equal(quote_id);

        callback(() => {
          freeCalled = true;
        });
      };

      request.once = (_: string, __: () => void) => {
        assert.fail('the request should not have been hooked');

        return request;
      };

      lockManager.acquireReadLock(
        quote_id,
        request,
        (free: () => void) => {
          expect(freeCalled, 'free was called automatically').to.be.false;
          free();
          expect(freeCalled, 'free was not called manually').to.be.true;
        },
        true
      );
    });
  });
});

function getMockUserRequest(): UserRequest {
  return <UserRequest>(<unknown>{
    getQuoteId: () => {
      return 123;
    },
    once: (_event: string, cb: () => void) => {
      cb();
    },
  });
}

function getMockSemaphore(): SemaphoreType {
  const semaphore = {
    acquire: (_id: number, callback: (free: () => void) => void) => {
      callback(() => {});
    },
    freeStale: (_interval: number, callback: (id: number) => void) => {
      callback(123);

      return semaphore;
    },
    isLocked: (_id: number) => {
      return false;
    },
  };

  return semaphore;
}
