/**
 * Tests Data-API-based metadata population
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';
import {StagingBucket} from '../../../src/bucket/StagingBucket';
import {DataApiManager} from '../../../src/dapi/DataApiManager';
import {PositiveInteger} from '../../../src/numeric';
import {DapiMetaSource as Sut} from '../../../src/server/meta/DapiMetaSource';
import {ServerSideDapi} from '../../../src/program/Program';
import {DapiAction} from '../../../src/dapi/DataApi';
import {BucketData} from '../../../src/bucket/Bucket';

const expect = require('chai').expect;

describe('DapiMetaSource', () => {
  it('populates field with dapi response', () => {
    const dapim = createStubDapiManager();
    const field_name = 'field_foo';
    const index = 1;

    const dapi = {
      name: 'dapi_name',
      value: 'foo',
      mapdest: {map: 'dest'},
    };

    // input data to dapi
    const given_data = {};

    // dapi output data (response)
    const ret_data: Record<string, any>[] = [{}];

    const bucket_result: BucketData = <BucketData>(<unknown>{
      [dapi.name]: {
        [index]: ret_data[0][dapi.value],
      },
    });

    const metabucket = getStubBucket();

    // g prefix = "given"
    // all these show why we want to encapsulate this garbage
    dapim.getApiData = (gapi, gdata, gcallback, gname, gindex) => {
      expect(gapi).to.equal(dapi.name);
      expect(gdata).to.equal(given_data);
      expect(gname).to.equal(field_name);
      expect(gindex).to.equal(index);

      // make sure we handle async
      process.nextTick(() => gcallback(null, ret_data));

      return dapim;
    };

    dapim.setFieldData = (gname, gindex, gdata, gvalue, glabel, gunchanged) => {
      expect(gname).to.equal(dapi.name);
      expect(gindex).to.equal(index);
      expect(gdata).to.equal(ret_data);
      expect(gvalue).to.equal(dapi.value);
      expect(glabel).to.equal(''); // unused
      expect(gunchanged).to.equal(false);

      return dapim;
    };

    dapim.expandFieldData = (
      gname,
      gindex,
      gbucket,
      gmap,
      gpredictive,
      gdiff
    ) => {
      expect(gname).to.equal(dapi.name);
      expect(gindex).to.equal(index);
      expect(gbucket).to.equal(metabucket);
      expect(gmap).to.equal(dapi.mapdest);
      expect(gpredictive).to.equal(true);
      expect(gdiff).to.deep.equal(bucket_result);

      metabucket.getData = () => bucket_result;

      return dapim;
    };

    return new Sut(() => metabucket)
      .getFieldData(
        field_name,
        <PositiveInteger>index,
        dapim,
        <ServerSideDapi>(<unknown>dapi),
        given_data,
        DapiAction.TRIGGER,
        <StagingBucket>{getDiff: () => {}},
        () => {}
      )
      .then((result: any) => {
        expect(result.field).to.equal(field_name);
        expect(result.index).to.equal(index);
        expect(result.data).to.equal(bucket_result);
      });
  });

  it('rejects promise on error', () => {
    const e = Error('Test error');
    const dapim = createStubDapiManager();

    dapim.getApiData = (_, __, ___, ____, _____, ______, failc) => {
      failc(e);
      return dapim;
    };

    return expect(
      new Sut(() => getStubBucket()).getFieldData(
        'name',
        <PositiveInteger>0,
        dapim,
        <ServerSideDapi>(<unknown>{}),
        {},
        DapiAction.TRIGGER,
        <StagingBucket>{getDiff: () => {}},
        () => {}
      )
    ).to.eventually.be.rejectedWith(e);
  });

  it('rejects if more than one result is returned from dapi', () => {
    const dapim = createStubDapiManager();

    dapim.getApiData = (_, __, callback) => {
      // more than one result
      callback(null, [{}, {}]);
      return dapim;
    };

    return expect(
      new Sut(() => getStubBucket()).getFieldData(
        'name',
        <PositiveInteger>0,
        dapim,
        <ServerSideDapi>(<unknown>{}),
        {},
        DapiAction.TRIGGER,
        <StagingBucket>{getDiff: () => {}},
        () => {}
      )
    ).to.eventually.be.rejectedWith(
      'Data API request for name produced 2 results'
    );
  });

  it('executes error callback when an error is returned', () => {
    const dapim = createStubDapiManager();

    const error_to_throw = new Error('Something happened');

    dapim.getApiData = (_, __, callback) => {
      callback(error_to_throw, [{}]);
      return dapim;
    };

    let error_callback_executed = false;

    new Sut(() => getStubBucket()).getFieldData(
      'name',
      <PositiveInteger>0,
      dapim,
      <ServerSideDapi>(<unknown>{}),
      {},
      DapiAction.TRIGGER,
      <StagingBucket>{getDiff: () => {}},
      (e: Error) => {
        expect(e.message).to.equal(error_to_throw.message);
        error_callback_executed = true;
      }
    );

    expect(error_callback_executed).to.equal(true);
  });
});

function createStubDapiManager(): DataApiManager {
  return <DataApiManager>(<unknown>{
    getApiData() {},
    setFieldData() {},
    expandFieldData() {},
  });
}

function getStubBucket(): QuoteDataBucket {
  return <QuoteDataBucket>(<unknown>{
    setValues() {},
    getData() {},
  });
}
