/**
 * Tests ServerSideQuote
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

'use strict';
const {QuoteDataBucket} = require('../../../src/bucket/QuoteDataBucket');
const Sut = require('../../..').server.quote.ServerSideQuote;
const expect = require('chai').expect;

describe('ServerSideQuote', () => {
  describe('accessors & mutators', () => {
    [
      {
        property: 'startDate',
        default: 0,
        value: 946684800,
      },
      {
        property: 'initialRatedDate',
        default: 0,
        value: 946684800,
      },
      {
        property: 'agentId',
        default: 0,
        value: 12345678,
      },
      {
        property: 'agentEntityId',
        default: 0,
        value: 12345678,
      },
      {
        property: 'agentName',
        default: '',
        value: 'name',
      },
      {
        property: 'imported',
        default: false,
        value: true,
        accessor: 'is',
      },
      {
        property: 'bound',
        default: false,
        value: true,
        accessor: 'is',
      },
      {
        property: 'currentStepId',
        default: 1,
        value: 2,
      },
      {
        property: 'topVisitedStepId',
        default: 1,
        value: 2,
      },
      {
        property: 'topSavedStepId',
        value: 1,
      },
      {
        property: 'error',
        default: '',
        value: 'ERROR',
      },
      {
        property: 'programVersion',
        default: '',
        value: '1.0.0',
      },
      {
        property: 'creditScoreRef',
        default: 0,
        value: 800,
      },
      {
        property: 'lastPremiumDate',
        default: 0,
        value: 946684800,
      },
      {
        property: 'ratedDate',
        default: 0,
        value: 946684800,
      },
      {
        property: 'rateBucket',
        default: null,
        value: new QuoteDataBucket(),
      },
      {
        property: 'nonpersistent',
        default: [],
        value: ['foo', 'bar'],
      },
    ].forEach(testCase => {
      const quote = Sut(123, {}, {}, {getFirstStepId: () => 0});
      const property = testCase.property;
      const title_cased = property.charAt(0).toUpperCase() + property.slice(1);
      const setter = (testCase.mutator || 'set') + title_cased;
      const getter = (testCase.accessor || 'get') + title_cased;

      it(property + ' can be mutated and accessed', () => {
        expect(quote[getter].call(quote)).to.deep.equal(testCase.default);
        quote[setter].call(quote, testCase.value);
        expect(quote[getter].call(quote)).to.equal(testCase.value);
      });
    });
  });

  describe('#isAuthorized', () => {
    [
      {
        label: 'Internal user is authorized',
        is_internal: true,
        session_agent_id: 900000,
        session_agent_num: 'FOO',
        session_username: undefined,
        quote_agent_id: 911111,
        meta_data: {liza_access_groups: ['BAR']},
        expected: true,
      },
      {
        label:
          'Non-internal user is authorized if session and quote agentId match',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: undefined,
        session_username: undefined,
        quote_agent_id: 911111,
        meta_data: undefined,
        expected: true,
      },
      {
        label:
          'Non-internal user is NOT authorized when agentId is not a match',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: undefined,
        session_username: undefined,
        quote_agent_id: 922222,
        meta_data: undefined,
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized when access_groups exist and do not contain agt num, even if agent ID matches',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: 'FOO',
        session_username: undefined,
        quote_agent_id: 911111,
        meta_data: {liza_access_groups: ['']},
        expected: false,
      },
      {
        label: 'Non-internal user is authorized if agency is in access group',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: 'BAR',
        session_username: undefined,
        quote_agent_id: undefined,
        meta_data: {liza_access_groups: ['FOO', 'BAR', 'BAZ']},
        expected: true,
      },
      {
        label:
          'Non-internal user is NOT authorized if agency is not in access group',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: 'QUX',
        session_username: undefined,
        quote_agent_id: undefined,
        meta_data: {liza_access_groups: ['FOO', 'BAR', 'BAZ']},
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized if agentId and agency number are undefined',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: undefined,
        session_username: undefined,
        quote_agent_id: 911111,
        meta_data: {liza_access_groups: ['FOO', 'BAR', 'BAZ']},
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized if quote agentId and agency number are undefined',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: 'FOO',
        session_username: undefined,
        quote_agent_id: undefined,
        meta_data: undefined,
        expected: false,
      },
      {
        label:
          'Non-internal user is authorized when agentId is not a match, but liza_access_email matches',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: undefined,
        session_username: 'foo@example.com',
        quote_agent_id: 922222,
        meta_data: {liza_access_email: ['foo@example.com']},
        expected: true,
      },
      {
        label:
          'Non-internal user is authorized when access_groups exist and do not contain agt num, even if agent ID matches, but liza_access_email matches',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: 'FOO',
        session_username: 'foo@example.com',
        quote_agent_id: 911111,
        meta_data: {
          liza_access_groups: [''],
          liza_access_email: ['foo@example.com'],
        },
        expected: true,
      },
      {
        label:
          'Non-internal user is authorized if agency is not in access group, but liza_access_email matches',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: 'QUX',
        session_username: 'foo@example.com',
        quote_agent_id: undefined,
        meta_data: {
          liza_access_groups: ['FOO', 'BAR', 'BAZ'],
          liza_access_email: ['foo@example.com'],
        },
        expected: true,
      },
      {
        label:
          'Non-internal user is authorized if agentId and agency number are undefined, but liza_access_email matches',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: undefined,
        session_username: 'foo@example.com',
        quote_agent_id: 911111,
        meta_data: {
          liza_access_groups: ['FOO', 'BAR', 'BAZ'],
          liza_access_email: ['foo@example.com'],
        },
        expected: true,
      },
      {
        label:
          'Non-internal user is authorized if quote agentId and agency number are undefined, but liza_access_email matches',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: 'FOO',
        session_username: 'foo@example.com',
        quote_agent_id: undefined,
        meta_data: {liza_access_email: ['foo@example.com']},
        expected: true,
      },
      {
        label:
          'Non-internal user is NOT authorized when agentId is not a match and liza_access_email does not match',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: undefined,
        session_username: 'foo@example.com',
        quote_agent_id: 922222,
        meta_data: {liza_access_email: ['bar@example.com']},
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized when access_groups exist and do not contain agt num, even if agent ID matches and liza_access_email does not match',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: 'FOO',
        session_username: 'foo@example.com',
        quote_agent_id: 911111,
        meta_data: {
          liza_access_groups: [''],
          liza_access_email: ['bar@example.com'],
        },
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized if agency is not in access group and liza_access_email does not match',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: 'QUX',
        session_username: 'foo@example.com',
        quote_agent_id: undefined,
        meta_data: {
          liza_access_groups: ['FOO', 'BAR', 'BAZ'],
          liza_access_email: ['bar@example.com'],
        },
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized if agentId and agency number are undefined and liza_access_email does not match',
        is_internal: false,
        session_agent_id: undefined,
        session_agent_num: undefined,
        session_username: 'foo@example.com',
        quote_agent_id: 911111,
        meta_data: {
          liza_access_groups: ['FOO', 'BAR', 'BAZ'],
          liza_access_email: ['bar@example.com'],
        },
        expected: false,
      },
      {
        label:
          'Non-internal user is NOT authorized if quote agentId and agency number are undefined and liza_access_email does not match',
        is_internal: false,
        session_agent_id: 911111,
        session_agent_num: 'FOO',
        session_username: 'foo@example.com',
        quote_agent_id: undefined,
        meta_data: {liza_access_email: ['bar@example.com']},
        expected: false,
      },
    ].forEach(
      ({
        label,
        is_internal,
        session_agent_id,
        session_agent_num,
        session_username,
        quote_agent_id,
        meta_data,
        expected,
      }) => {
        it(label, () => {
          const sut = Sut();
          const session = createMockSession();
          session.isInternal = () => is_internal;
          session.agentId = () => session_agent_id;
          session.agencyNumber = () => session_agent_num;
          session.userName = () => session_username;
          const mock_meta_bucket = {
            setValues: () => {},
            getData: () => meta_data,
          };

          sut.setMetabucket(mock_meta_bucket);
          sut.setAgentId(quote_agent_id);

          expect(sut.isAuthorized(session)).to.equal(expected);
        });
      }
    );
  });

  describe('rating data', () => {
    it(`#setRatingData throws an error if no bucket is set`, () => {
      const data = {foo: 'bar'};
      const sut = Sut();

      expect(function () {
        sut.setRatingData(data);
      }).to.throw(Error);
    });

    it(`Bucket values setters/getters work correctly`, () => {
      const data = {foo: 'bar'};
      let bucket_data = null;
      const sut = Sut();
      var called = false;

      const bucket = {
        setValues(gdata) {
          expect(gdata).to.deep.equal(data);

          bucket_data = gdata;
          called = true;
        },

        getData() {
          return bucket_data;
        },
      };

      sut.setRateBucket(bucket);
      sut.setRatingData(data);

      expect(called).to.equal(true);
    });
  });

  describe('#refreshStepDelta', () => {
    [
      {
        label: 'Delta is zero when visiting the current step',
        current_step: 1,
        new_step: 1,
        delta_expected: '0',
      },
      {
        label: 'Delta is negative when visiting a previous step',
        current_step: 2,
        new_step: 1,
        delta_expected: '-1',
      },
      {
        label: 'Delta is positive when visiting the next step',
        current_step: 1,
        new_step: 2,
        delta_expected: '1',
      },
      {
        label: 'String values behave properly',
        current_step: '9',
        new_step: '10',
        delta_expected: '1',
      },
    ].forEach(({label, current_step, new_step, delta_expected}) => {
      it(label, () => {
        let setData_called = false;

        const bucket = {
          setValues(given_data) {
            expect(given_data.__liza_step_delta[0]).to.equal(delta_expected);
            expect(given_data.__liza_step_id_current[0]).to.equal(
              '' + new_step
            );

            setData_called = true;
          },

          getDataByName: () => [current_step],
        };

        const sut = Sut(123, bucket);

        sut.setRateBucket(bucket);
        sut.refreshStepDelta(new_step);

        expect(setData_called).to.be.true;
      });
    });
  });
});

function createMockSession(
  internal,
  agent_id,
  username,
  agent_name,
  entity_id,
  agency_num
) {
  return {
    isInternal: () => internal,
    agentId: () => agent_id || '',
    userName: () => username || '',
    userRoles: () => [],
    agentName: () => agent_name || '',
    agentEntityId: () => entity_id || '',
    agencyNumber: () => agency_num || '',
  };
}
