/**
 * Tests document loading
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {expect} from 'chai';
import * as E from 'fp-ts/Either';
import * as IOE from 'fp-ts/IOEither';
import {pipe} from 'fp-ts/function';
import {___Writable} from 'naughty';

import {
  applyDocumentDefaults,
  cleanDocument,
  kickBackToNewlyApplicable,
  mergeDocumentData,
  loadSessionIntoQuote,
  loadDocumentIntoQuote,
  mergeFlagData,
  initNewDocument,
  createSessions,
  loadSessionKeysIntoDocumentData,
  fixNonEncryptedFields,
  updateLatestVersions,
  retrieveOrGenerateSessionKey,
} from '../../../src/server/quote/loader';
import {ClassData} from '../../../src/client/Cmatch';
import {Program} from '../../../src/program/Program';
import {DocumentData} from '../../../src/server/db/MongoServerDao';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {PositiveInteger, ts_ctor} from '../../../src/numeric';
import {invalidate} from '../../../src/validate/invalid';
import {UserSession} from '../../../src/server/request/UserSession';
import {FeatureFlag} from '../../../src/system/flags/FeatureFlag';
import {ServerDao} from '../../../src/server/db/ServerDao';
import {randomBytes} from 'crypto';
import {
  ServiceSecretMap,
  KekService,
} from '../../../src/server/service/KekService';
import {
  createSessionKey,
  encryptSessionKey,
  isPossiblyEncrypted,
  Key,
} from '../../../liza-crypto';
import {keyBytes} from '../../../liza-crypto/src/config';

describe('loader.applyDocumentDefaults', () => {
  [
    {
      label: 'initializes defaults',
      defaults: {a: 'one', b: 'two'},
      meta: {
        groups: {},
        qtypes: {
          a: {type: 'noyes'},
          b: {type: 'noyes'},
        },
      },
      groupExclusiveFields: {
        Something: ['a', 'b'],
      },
      doc_data: {},
      expected: {
        a: ['one'],
        b: ['two'],
      },
    },
    // for ancient versions of liza-proguic (before it was even called
    // "liza")
    {
      label: 'initializes defaults for ancient data representations',
      defaults: {a: 'one', b: 'two'},
      meta: {
        groups: {},
        qtypes: {
          a: 'noyes',
          b: 'noyes',
        },
      },
      groupExclusiveFields: {
        Something: ['a', 'b'],
      },
      doc_data: {},
      expected: {
        a: ['one'],
        b: ['two'],
      },
    },
    {
      label: 'does nothing with no data or defaults',
      defaults: {},
      meta: {
        groups: {},
      },
      groupExclusiveFields: {},
      doc_data: {},
      expected: {},
    },
    {
      label: 'produces empty object given undefined data',
      defaults: {},
      meta: {
        groups: {},
      },
      groupExclusiveFields: {},
      doc_data: undefined,
      expected: {},
    },
    {
      label: 'keeps existing data with defaults',
      defaults: {
        foo: 'init',
        bar: 'test',
      },
      meta: {
        groups: {},
        qtypes: {
          foo: {type: 'noyes'},
          bar: {type: 'noyes'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
        SomethingElse: ['bar'],
      },
      doc_data: {bar: ['baz']},
      expected: {
        foo: ['init'],
        bar: ['baz'],
      },
    },
    {
      label: 'keeps existing data with no defaults',
      defaults: {},
      meta: {
        groups: {},
        qtypes: {
          bar: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        SomethingElse: ['bar'],
      },
      doc_data: {bar: ['baz']},
      expected: {
        bar: ['baz'],
      },
    },
    {
      label: 'does not overwrite existing data with defaults',
      defaults: {foo: 'init'},
      meta: {
        groups: {},
        qtypes: {
          foo: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {foo: ['bar']},
      expected: {
        foo: ['bar'],
      },
    },
    {
      label: 'does not overwrite existing data with defaults and multiple rows',
      defaults: {foo: 'init'},
      meta: {
        groups: {
          Something: {
            min: 3,
          },
        },
        qtypes: {
          foo: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {foo: ['bar', 'baz', 'test']},
      expected: {
        foo: ['bar', 'baz', 'test'],
      },
    },
    {
      label: 'initializes with default bucket data',
      defaults: {foo: 'init'},
      meta: {
        groups: {
          Something: {
            min: 5,
          },
        },
        qtypes: {
          foo: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {},
      expected: {
        foo: ['init', 'init', 'init', 'init', 'init'],
      },
    },
    {
      label: 'does not reset applicable fields on init',
      defaults: {foo: 'init'},
      whens: {},
      meta: {
        groups: {
          Something: {
            min: 2,
          },
        },
        qtypes: {
          foo: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {},
      expected: {
        foo: ['init', 'init'],
      },
      clearNaFields: true,
      naFieldValue: '',
    },
    {
      label: 'does not reset retained fields on init',
      defaults: {foo: 'init'},
      cretain: {foo: true},
      meta: {
        groups: {
          Something: {
            min: 2,
          },
        },
        qtypes: {
          foo: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {},
      expected: {
        foo: ['init', 'init'],
      },
      clearNaFields: true,
      naFieldValue: '',
    },
    {
      label: 'fix missing bucket data values',
      defaults: {foo: 'init'},
      meta: {
        groups: {
          Something: {
            min: 5,
          },
        },
        qtypes: {
          foo: {type: 'text'},
        },
      },
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {
        foo: ['1', '2', '3', '4'],
      },
      expected: {
        foo: ['1', '2', '3', '4', 'init'],
      },
    },
    {
      label: 'ignores undefined types',
      defaults: {foo: 'default'},
      meta: {
        groups: {},
        qtypes: {
          foo: {type: 'undefined'},
        },
      },
      hasKnownType: () => false,
      groupExclusiveFields: {
        Something: ['foo'],
      },
      doc_data: {},
      expected: {},
    },
  ].forEach(
    ({
      label,
      doc_data,
      defaults,
      meta,
      groupExclusiveFields,
      expected,
      clearNaFields,
      naFieldValue,
      whens,
      cretain,
      hasKnownType,
    }) => {
      it(label, () => {
        const fake_kek = Buffer.from(randomBytes(keyBytes));
        const kek_map = {
          [KekService.INTERNAL_KEY]: {
            '1': {
              is_active: true,
              key: fake_kek,
            },
          },
        };
        const sessions = createSessions(createSessionKey(), kek_map, []);
        const program = <Program>(<unknown>{
          id: 'foo',
          defaults: defaults,
          meta: meta,
          groupExclusiveFields: groupExclusiveFields,
          whens: whens ? whens : defaults,
          clearNaFields,
          naFieldValue,
          cretain: cretain ? cretain : {},
          encryptedKeys: {},
          hasKnownType: hasKnownType ? hasKnownType : () => true,
        });

        const either = applyDocumentDefaults(program)(kek_map)(<DocumentData>{
          data: doc_data,
          sessionKeys: sessions,
        })();
        expect(either._tag).to.equal('Right');
        expect(either._tag === 'Right' && either.right.data).to.deep.equal(
          expected
        );
      });
    }
  );
});

describe('loader.cleanQuote', () => {
  const liza_kek = Buffer.from(randomBytes(keyBytes));
  const fake_session_key = encryptSessionKey(liza_kek, createSessionKey());
  describe('group cleaning', () => {
    [
      {
        label: 'expands indexes of linked and non-linked groups',

        group_index: {
          one: 'field11', // linked
          two: 'field11', // linked
          three: 'field31',
        },

        exclusive: {
          one: ['field11', 'field12'],
          two: ['field21', 'field22'],
          three: ['field31', 'field32', 'unknown_ignore_me'],
        },

        defaults: {
          field12: '12default',
        },

        qtypes: {
          field11: {type: 'text'},
          field12: {type: 'text'},
          field21: {type: 'text'},
          field22: {type: 'text'},
          field31: {type: 'text'},

          // ancient pre-"liza" data representation
          field32: 'text',
        },

        existing: {
          field11: ['1', '', '3'], // leader one, two
          field12: ['a', 'b'],
          field21: ['e'],
          field22: ['I', 'II'],
          field31: ['i', 'ii'], // leader three
          field32: ['x'],
        },

        expected: {
          field12: [, , '12default'],
          field21: [, '', ''],
          field22: [, , ''],
          field32: [, ''],
        },

        hasKnownType: (name: string) => name !== 'unknown_ignore_me',
      },
    ].forEach(test =>
      it(test.label, () => {
        const quote = createStubQuote(test.existing, {}, fake_session_key);
        const program = createStubProgram({});

        program.defaults = test.defaults;
        program.groupIndexField = test.group_index;
        program.groupExclusiveFields = test.exclusive;
        program.meta.qtypes = test.qtypes;
        program.whens = {
          field11: ['--vis-field11'],
          field12: ['--vis-field12'],
        };
        program.hasNaField = () => false;
        program.hasKnownType = test.hasKnownType
          ? test.hasKnownType
          : () => true;

        const updates: Record<string, any> = {};

        quote.setData = given => {
          Object.keys(given).forEach(k => (updates[k] = given[k]));
          return quote;
        };

        return cleanDocument({liza_kek})(<Program>program)(quote)().then(
          result => {
            expect(E.isRight(result)).to.be.true;
            expect(updates).to.deep.equal(test.expected);
          }
        );
      })
    );
  });

  describe('fixGroup', () => {
    [
      {
        label: 'clears NA fields when vector',

        group_index: {
          one: 'field11', // linked
        },

        exclusive: {
          one: ['field11', 'field12'],
        },

        defaults: {
          field12: '12default',
        },

        qtypes: {
          field11: {type: 'text'},
          field12: {type: 'text'},
        },

        existing: {
          field11: ['1', '', '3'], // leader one, two
          field12: ['a', 'b'],
        },

        expected: {
          field12: [, , ''],
        },

        bucket: {
          field11: [1, 1, 0],
          field12: [1, 1],
        },

        classify: {
          '--vis-field11': {is: true, indexes: [1, 1, 0]},
          '--vis-field12': {is: true, indexes: [1, 1]},
        },
      },

      {
        label: 'clears NA fields when matrix',

        group_index: {
          one: 'field11', // linked
        },

        exclusive: {
          one: ['field11', 'field12'],
        },

        defaults: {
          field12: '12default',
        },

        qtypes: {
          field11: {type: 'text'},
          field12: {type: 'text'},
        },

        existing: {
          field11: ['1', '', '3'], // leader one, two
          field12: ['a', 'b'],
        },

        expected: {
          field12: [, , ''],
        },

        bucket: {
          field11: [1, 1, 0],
          field12: [1, 1],
        },

        classify: {
          '--vis-field11': {
            is: true,
            indexes: [
              [1, 1, 1],
              [1, 1, 1],
              [0, 0, 0],
            ],
          },
          '--vis-field12': {
            is: true,
            indexes: [
              [1, 1, 1],
              [1, 1, 1],
            ],
          },
        },
      },
      {
        label: 'applies default values when visible as vector',

        group_index: {
          one: 'field11', // linked
        },

        exclusive: {
          one: ['field11', 'field12'],
        },

        defaults: {
          field12: '12default',
        },

        qtypes: {
          field11: {type: 'text'},
          field12: {type: 'text'},
        },

        existing: {
          field11: ['1', '', '3'], // leader one, two
          field12: ['a', 'b'],
        },

        expected: {
          field12: [, , '12default'],
        },

        bucket: {
          field11: [1, 1, 1],
          field12: [1, 1],
        },

        classify: {
          '--vis-field11': {is: true, indexes: [1, 1, 1]},
          '--vis-field12': {is: true, indexes: [1, 1]},
        },

        hasNaField: () => false,
      },
      {
        label: 'applies default values when visible as matrix',

        group_index: {
          one: 'field11', // linked
        },

        exclusive: {
          one: ['field11', 'field12'],
        },

        defaults: {
          field12: '12default',
        },

        qtypes: {
          field11: {type: 'text'},
          field12: {type: 'text'},
        },

        existing: {
          field11: ['1', '', '3'], // leader one, two
          field12: ['a', 'b'],
        },

        expected: {
          field12: [, , '12default'],
        },

        bucket: {
          field11: [1, 1, 1],
          field12: [1, 1],
        },

        classify: {
          '--vis-field11': {
            is: true,
            indexes: [
              [1, 1, 1],
              [1, 1, 1],
              [0, 1, 0],
            ],
          },
          '--vis-field12': {
            is: true,
            indexes: [
              [1, 1, 1],
              [1, 1, 1],
            ],
          },
        },

        hasNaField: () => false,
      },
    ].forEach(test =>
      it(test.label, () => {
        const quote = createStubQuote(test.existing, {}, fake_session_key);
        const program = createStubProgram({});

        program.defaults = test.defaults;
        program.groupIndexField = test.group_index;
        program.groupExclusiveFields = test.exclusive;
        program.meta.qtypes = test.qtypes;
        program.clearNaFields = true;
        program.naFieldValue = '';
        program.whens = {
          field11: ['--vis-field11'],
          field12: ['--vis-field12'],
        };

        program.hasNaField = test.hasNaField ? test.hasNaField : () => true;
        program.hasKnownType = () => true;

        program.classify = () => <ClassData>(<unknown>test.classify);

        quote.getBucket = () =>
          <any>{
            getData() {
              return test.bucket;
            },
          };

        const updates: Record<string, any> = {};

        quote.setData = given => {
          Object.keys(given).forEach(k => (updates[k] = given[k]));
          return quote;
        };

        return cleanDocument({liza_kek})(<Program>program)(quote)().then(
          result => {
            expect(E.isRight(result)).to.be.true;
            expect(updates).to.deep.equal(test.expected);
          }
        );
      })
    );
  });

  describe('metadata cleaning', () => {
    [
      {
        label: 'populates all fields when empty',
        existing: {},
        fields: {foo: {}, bar: {}},
        expected: {foo: [], bar: []},
      },
      {
        label: 'populates only missing fields when non-empty',
        existing: {foo: [1], baz: [2]},
        fields: {foo: {}, bar: {}},
        expected: {foo: [1], bar: [], baz: [2]},
      },
      {
        label: 'does nothing with no fields',
        existing: {foo: [1], baz: [2]},
        fields: {},
        expected: {foo: [1], baz: [2]},
      },
    ].forEach(({label, existing, fields, expected}) =>
      it(label, () => {
        const quote = createStubQuote({}, existing, fake_session_key);
        const program = createStubProgram(fields);

        return cleanDocument({liza_kek})(<Program>program)(quote)().then(
          result => {
            expect(E.isRight(result)).to.be.true;
            expect(quote.getMetabucket().getData()).to.deep.equal(expected);
          }
        );
      })
    );
  });

  type Test = {
    label: string;
    encryptedKeys: Record<string, any[]>;
    expected: Record<string, boolean>;
  };
  describe('fixEncryptedFields', () => {
    [
      {
        label: 'Doesnt encrypt empty items',
        encryptedKeys: {foo: [], bar: []},
        expected: {foo: false, bar: false},
      },
      {
        label: 'Doesnt encrypt non-string types',
        encryptedKeys: {foo: [1], bar: [2]},
        expected: {foo: false, bar: false},
      },
      {
        label: 'Encrypts required new keys',
        encryptedKeys: {foo: ['some-val'], bar: ['some-val-2']},
        expected: {foo: true, bar: true},
      },
      {
        label: 'Does not encrypt the empty string',
        encryptedKeys: {foo: ['nonempty'], bar: ['']},
        expected: {foo: true, bar: false},
      },
    ].forEach(({label, encryptedKeys, expected}: Test) =>
      it(label, () => {
        const quote = createStubQuote(encryptedKeys, {}, fake_session_key);
        const program = createStubProgram({}, {encryptedKeys});

        fixNonEncryptedFields({liza_kek})(<Program>program)(quote)();

        Object.keys(expected).forEach(name => {
          const bucketVal = quote.getDataByName(name);
          bucketVal.forEach(
            data =>
              typeof data === 'string' &&
              expect(isPossiblyEncrypted(data)).to.be.equal(expected[name])
          );
        });
      })
    );
  });
});

describe('loader.kickBackToNewlyApplicable', () => {
  type TestData = {
    label: string;
    field_state?: Record<string, number | number[]>;
    last_field_state?: Record<string, number | number[]>;
    qstep: Record<string, number>;
    data: Record<string, string[]>;
    required: Record<string, boolean>;
    internal?: Record<string, boolean>;
    locked?: boolean;
    lock_step?: PositiveInteger;
    current_step_id: number;
    top_saved_step_id: number;
    expected_current_step_id?: number;
    expected_saved_step_id?: number;
    expected_error?: boolean;
    expected_data: Record<string, Array<undefined | string>>;
  };

  // Simply to make the test cases read more clearly
  const no_kickback = -Infinity;

  (<TestData[]>[
    {
      label: 'fails if field state is not available',

      // Current field state is not available
      field_state: undefined,
      last_field_state: {},

      // We'll include this anyway just to make sure that we do nothing even
      // if a kickback would have otherwise occurred.
      qstep: {foo: 2},
      required: {foo: true},

      current_step_id: 3,
      top_saved_step_id: 3,

      // No current field data represents a bug in the system---the
      // classifier was not invoked when it was expected to have been, or
      // the logic that populates the value is not functioning correctly.
      expected_error: true,

      // And we should invalidate nothing
      expected_data: {},
    },

    // If no fields have changed applicability, then we want to leave the user
    // at whatever step they left off on.
    {
      label: 'keeps document at current step if no fields changed',

      // Identical field states
      field_state: {foo: [1, 0]},
      last_field_state: {foo: [1, 0]},

      // We must provide some mapping to ensure the system fails if it
      // _thinks_ there is some sort of change, and _this must be less than
      // top_saved_step_id below_
      qstep: {foo: 2},
      required: {foo: true},

      current_step_id: 2,
      top_saved_step_id: 3,

      // Note that current and saved started and should remain different
      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'kicks back to step on single index change',

      // Second index changes
      field_state: {foo: [1, 1]},
      last_field_state: {foo: [1, 0]},

      // And the field is required
      required: {foo: true},

      // And so should kick back to this step
      qstep: {foo: 2},

      current_step_id: 3,
      top_saved_step_id: 3,

      expected_current_step_id: 2,
      expected_saved_step_id: 2,

      // And should invalidate only the changed index
      expected_data: {foo: [undefined, invalidate('')]},
    },

    {
      label: 'kicks back to step on multiple index change',

      // Both indexes change
      field_state: {foo: [1, 1]},
      last_field_state: {foo: [0, 0]},

      // And the field is required
      required: {foo: true},

      // And so should kick back to this step
      qstep: {foo: 2},

      current_step_id: 3,
      top_saved_step_id: 3,

      expected_current_step_id: 2,
      expected_saved_step_id: 2,

      // And should invalidate both changed indexes
      expected_data: {foo: [invalidate(''), invalidate('')]},
    },

    // Kicking back to a field that is not required would make it required,
    // since the field would be invalidated and force the user to answer
    // it.  By not kicking back required fields, we are also able to
    // add fields without triggering kickback.
    {
      label: 'does not kick back on index change to non-required field',

      // Second index changes
      field_state: {foo: [1, 1]},
      last_field_state: {foo: [1, 0]},

      // But the field is _not_ required
      required: {foo: false},

      // And so should _not_ kick back to this step
      qstep: {foo: 2},

      current_step_id: 3,
      top_saved_step_id: 3,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'does not kick back on single index change to N/A',

      // The first index changed, but we're now N/A, so there's no use
      // kicking the user back to address something that can't be seen
      field_state: {foo: [0, 0]},
      last_field_state: {foo: [1, 0]},

      // And so there should _not_ be a kickback to this step
      qstep: {foo: 2},

      // Even though the field is required
      required: {foo: true},

      current_step_id: 3,
      top_saved_step_id: 3,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'does not kick back on multiple index change to N/A',

      // Both index changed, but we're now N/A, so there's no use
      // kicking the user back to address something that can't be seen
      field_state: {foo: [0, 0]},
      last_field_state: {foo: [1, 1]},

      // And so there should _not_ be a kickback to this step
      qstep: {foo: 2},

      // Even though the field is required
      required: {foo: true},

      current_step_id: 3,
      top_saved_step_id: 3,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'missing prior field state is implicitly zero (no kickback)',

      // No change, because it should be implicitly [0, 0]
      field_state: {foo: [0, 0]},
      last_field_state: {},

      // And so there should _not_ be a kickback to this step
      qstep: {foo: 2},

      // Even though the field is required
      required: {foo: true},

      current_step_id: 3,
      top_saved_step_id: 3,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'missing prior field state is implicitly zero (kickback)',

      // Change, because it should be implicitly [0, 0]
      field_state: {foo: [1, 0]},
      last_field_state: {},

      // And the field is required
      required: {foo: true},

      // And so there should be a kickback to this step
      qstep: {foo: 1},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: 1,
      expected_saved_step_id: 1,

      // And so should invalidate only the changed index
      expected_data: {foo: [invalidate(''), undefined]},
    },

    {
      label: 'missing current field state is ignored (implicitly NA)',

      // It's okay that we're now implicitly [0, 0]; we shouldn't kick back
      // to something that's N/A
      field_state: {},
      last_field_state: {foo: [1, 1]},

      // And so there should _not_ be a kickback to this step
      qstep: {foo: 1},

      // Even though the field is required
      required: {foo: true},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'current field state is scalar, last was vector, no kickback',

      // Current field state is broadcasted into [1, 1], which matches the
      // last field state
      field_state: {foo: 1},
      last_field_state: {foo: [1, 1]},

      // And so there should _not_ be a kickback to this step
      qstep: {foo: 1},

      // Even though the field is required
      required: {foo: true},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'current field state is scalar, last was vector, kickback',

      // Current field state is broadcasted into [1, 1], which differs from
      // the last field state
      field_state: {foo: 1},
      last_field_state: {foo: [1, 0]},

      // And the field is required
      required: {foo: true},

      // And so there should be a kickback to this step
      qstep: {foo: 1},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: 1,
      expected_saved_step_id: 1,

      // And so should invalidate only the changed index
      expected_data: {foo: [undefined, invalidate('')]},
    },

    {
      label: 'last field state was scalar, current is vector, no kickback',

      // Last field state is broadcasted into [1, 1], which matches the
      // current field state
      field_state: {foo: [1, 1]},
      last_field_state: {foo: 1},

      // And so there should _not_ be a kickback to this step
      qstep: {foo: 1},

      // Even though the field is required
      required: {foo: true},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'last field state was scalar, current is vector, kickback',

      // Last field state is broadcasted into [1, 1], which differs from
      // the current field state
      field_state: {foo: [1, 0]},
      // (Though note that this won't actually happen, since it'd be stripped
      // out, and if this were 1, this test would have to make current 0,
      // which means N/A, and therefore wouldn't kick back)
      last_field_state: {foo: 0},

      // And the field is required
      required: {foo: true},

      // And so there should be a kickback to this step
      qstep: {foo: 1},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: 1,
      expected_saved_step_id: 1,

      // And so should invalidate only the changed index
      expected_data: {foo: [invalidate(''), undefined]},
    },

    {
      label: 'does not kick forward past top saved step',

      // Field state differs
      field_state: {foo: [1]},
      last_field_state: {foo: [0]},

      // And the field is required
      required: {foo: true},

      // And the field is on this step
      qstep: {foo: 7},

      // But that's _past_ the saved step, and so we should not make the
      // change, otherwise we'd be kicking _forward_
      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'kicks back to lowest of applicable step ids',

      // Multiple field states differ
      field_state: {foo: [1], bar: [1], baz: [1], quux: [1]},
      last_field_state: {quux: [1]},

      // Each located on these steps (one intentionally > current_step_id)
      qstep: {foo: 7, bar: 2, baz: 3, quux: 1},

      // And each is required
      required: {foo: true, bar: true, baz: true, quux: true},

      // We should kick back to the lowest of each of the above, which will
      // be all except for quux, having filtered out foo for being greater
      // than the top_saved_step_id
      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: 2,
      expected_saved_step_id: 2,

      // And so should invalidate only the fields prior to the current step
      expected_data: {bar: [invalidate('')], baz: [invalidate('')]},
    },

    {
      label: 'kicks back to lowest of required field step ids',

      // Multiple field states differ
      field_state: {foo: [1], bar: [1], baz: [1], quux: [1]},
      last_field_state: {quux: [1]},

      // Each is below the current step
      qstep: {foo: 4, bar: 3, baz: 2, quux: 1},

      // But note that two of them are not required
      required: {foo: false, bar: true, baz: true, quux: false},

      current_step_id: 5,
      top_saved_step_id: 5,

      // And so we should kick back to the step id associated with the
      // lowest _required_ field, which omits the lowest one (`quux`)
      expected_current_step_id: 2,
      expected_saved_step_id: 2,

      // We should only invalidate the required fields, which omits both
      // `foo` and `quux`
      expected_data: {bar: [invalidate('')], baz: [invalidate('')]},
    },

    // Kicking back based on internal fields is problematic because field
    // visibility differs when an internal users views a document
    {
      label: 'does not kick back internal fields',

      // One field changed
      field_state: {foo: [1]},
      last_field_state: {foo: [0]},

      // Field is below the current step
      qstep: {foo: 1},

      // And is required
      required: {foo: true},

      // But is internal
      internal: {foo: true},

      current_step_id: 5,
      top_saved_step_id: 5,

      // And so there should be no kickback
      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And no fields should be invalidated
      expected_data: {},
    },

    {
      label: 'kicks back to lowest of non-internal step ids',

      // Multiple field states differ
      field_state: {foo: [1], bar: [1], baz: [1], quux: [1]},
      last_field_state: {quux: [1]},

      // Each is below the current step
      qstep: {foo: 4, bar: 3, baz: 2, quux: 1},

      // All are required
      required: {foo: true, bar: true, baz: true, quux: true},

      // But two are internal
      internal: {foo: false, bar: true, baz: false, quux: true},

      current_step_id: 5,
      top_saved_step_id: 5,

      // And so we should kick back to the step id associated with the
      // lowest _non-internal_ field, which omits the lowest one (`quux`)
      expected_current_step_id: 2,
      expected_saved_step_id: 2,

      // We should only invalidate the required fields, which omits both
      // `bar` and `quux`
      expected_data: {foo: [invalidate('')], baz: [invalidate('')]},
    },

    {
      label: 'kicks back before top saved even if current step is less',

      // Multiple field states differ
      field_state: {foo: [1], bar: [1], baz: [1], quux: [1]},
      last_field_state: {quux: [1]},

      // Each is required
      required: {foo: true, bar: true, baz: true, quux: true},

      // Each located on these steps (one intentionally > top_saved_step_id)
      qstep: {foo: 7, bar: 2, baz: 3, quux: 1},

      // current_step_id < kickback < top_saved_step_id, so we should be
      // kicked back but the current step id should not change
      current_step_id: 1,
      top_saved_step_id: 4,

      expected_current_step_id: 1,
      expected_saved_step_id: 2,

      // And so should invalidate only the fields prior to the top saved
      // step
      expected_data: {bar: [invalidate('')], baz: [invalidate('')]},
    },

    {
      label: 'does not kick back if quote is locked without lock step',

      // Field state differs
      field_state: {foo: [1]},
      last_field_state: {foo: [0]},

      // And the field is required
      required: {foo: true},

      // And the field is on this step
      qstep: {foo: 1},

      // But our quote is locked, without an explicit lock step, meaning no
      // modifications to any step are possible
      locked: true,
      lock_step: 0,

      // And so we should not kick back at all, despite the kickback step
      // being less than our top saved step
      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: no_kickback,
      expected_saved_step_id: no_kickback,

      // And so should not invalidate anything
      expected_data: {},
    },

    {
      label: 'kicks back, but not before explicit lock step',

      // Field state differs
      field_state: {foo: [1], bar: [1], baz: [1]},
      last_field_state: {foo: [0], bar: [0], baz: [0]},

      // Each field is required
      required: {foo: true, bar: true, baz: true},

      // Foo is before the lock step, bar is after but before the top saved
      // step, and baz is after the top saved step
      qstep: {foo: 1, bar: 3, baz: 9},
      //                   ^ kick back here

      // Our quote is locked at this step
      locked: true,
      lock_step: 2,

      // And so we should not kick back to before the lock, but we should
      // kick back after the lock but before the top saved step
      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: 3,
      expected_saved_step_id: 3,

      // And so should invalidate the fields after the lock but before the
      // current step
      expected_data: {bar: [invalidate('')]},
    },

    // Be careful with this one.  This _should_ be safe, because if we have
    // a scalar 0, that surely means that all indexes were hidden.
    //
    // There is a situation where this could potentially be wrong: if the
    // indexes within a group somehow got out-of-sync and the cleaner fixed
    // them up.  That would be a bug, but it could potentially cause data to
    // be overwritten.  But the user would at least be forced to re-enter
    // those data, so the error is recoverable in a safe (albeit
    // inconvenient) way.
    {
      label: 'invalidation broadcasts scalar state onto bucket field',

      // Both states are scalar, where foo is now applicable for all
      // indexes, whereas before it was applicable for none
      field_state: {foo: 1},
      last_field_state: {foo: 0},

      // And the field is required
      required: {foo: true},

      // And so there should be a kickback to this step
      qstep: {foo: 1},

      current_step_id: 4,
      top_saved_step_id: 4,

      expected_current_step_id: 1,
      expected_saved_step_id: 1,

      // The bucket contains two indexes for this field
      data: {
        foo: ['bar', 'baz'],
      },

      // And so we should broadcast the above state to two indexes so that
      // it overwrites both bucket values (since neither was applicable
      // previously, according to `last_field_state`)
      expected_data: {foo: [invalidate(''), invalidate('')]},
    },

    //
    // Missing field state tests
    //
    // The idea is to automatically infer, to the best of our ability, what
    // the field state of old documents (prior to field state existing)
    // might have been.  That way, we can hopefully do something reasonable
    // even with older documents.
    //
    // Note that this system assumes non-empty fields mean that the user has
    // provided an answer (or accepted a default), but this is only strictly
    // true after the Liza version that introduced clearing bucket values
    // for non-applicable fields.  Documents even older than that,
    // unfortunately, we can't say for certain what happend; however, it so
    // happens that the circumstances under which this system was written
    // are such that default values were not present for such documents that
    // we care about, and so it ends up working out as a viable migration
    // path.

    {
      label: 'Infers missing field state from required, empty fields',

      // Current field state, but no prior field state
      field_state: {foo: [1, 1], bar: [1, 1], baz: [0, 0]},
      last_field_state: undefined,

      // `foo` is before our top saved step, `bar` is after, `baz` is well
      // before (but note above that it's not applicable)
      qstep: {foo: 2, bar: 5, baz: 1},

      current_step_id: 4,
      top_saved_step_id: 4,

      // We are inferring field state, so we should _not_ cause an
      // error.  This was not true in previous implementations.
      expected_error: false,

      data: {
        // The first index must have been applicable since the user provided
        // an answer.
        foo: ['some_value', ''],

        bar: ['', ''],
        baz: ['', ''],
      },

      // We'll test non-required fields in a separate test to avoid too many
      // confusing variables.
      required: {
        foo: true,
        bar: true,
        baz: true,
      },

      // `foo` and `baz` are the only ones before the top saved step, so we
      // only consider them.  `baz` is not applicable for any index, and
      // `foo` is applicable for both but, based on the inferred data, was
      // already applicable previously for the first index.
      expected_current_step_id: 2,
      expected_saved_step_id: 2,
      expected_data: {
        foo: [undefined, invalidate('')],
      },
    },

    {
      label: 'Omits non-required fields from state inference',

      // Current field state, but no prior field state
      field_state: {foo: [1, 1], bar: [1, 1]},
      last_field_state: undefined,

      // Both are before the current step
      qstep: {foo: 1, bar: 2},

      current_step_id: 4,
      top_saved_step_id: 4,

      // We are inferring field state, so we should _not_ cause an
      // error.  This was not true in previous implementations.
      expected_error: false,

      // Neither field was previously answered
      data: {
        foo: ['', ''],
        bar: ['', ''],
      },

      // But only `bar` is required
      required: {
        foo: false,
        bar: true,
      },

      // And so we should only act on `bar`
      expected_current_step_id: 2,
      expected_saved_step_id: 2,
      expected_data: {
        bar: [invalidate(''), invalidate('')],
      },
    },

    {
      label: 'Omits internal fields from state inference',

      // Current field state, but no prior field state
      field_state: {foo: [1, 1], bar: [1, 1]},
      last_field_state: undefined,

      // Both are before the current step
      qstep: {foo: 1, bar: 2},

      current_step_id: 4,
      top_saved_step_id: 4,

      // We are inferring field state, so we should _not_ cause an
      // error.  This was not true in previous implementations.
      expected_error: false,

      // Neither field was previously answered
      data: {
        foo: ['', ''],
        bar: ['', ''],
      },

      // Both are required
      required: {foo: true, bar: true},

      // But one is internal
      internal: {foo: true},

      // And so we should only act on `bar`
      expected_current_step_id: 2,
      expected_saved_step_id: 2,
      expected_data: {
        bar: [invalidate(''), invalidate('')],
      },
    },
  ]).forEach(({label, ...tdata}) =>
    it(label, () => {
      const program = createStubProgram({});
      const quote = createStubQuote({}, {});

      quote.getFieldState = () => tdata.field_state;
      quote.getLastPersistedFieldState = () => tdata.last_field_state;
      quote.getCurrentStepId = () => <PositiveInteger>tdata.current_step_id;
      quote.getTopVisitedStepId = () => <PositiveInteger>tdata.current_step_id;
      quote.getTopSavedStepId = () => <PositiveInteger>tdata.top_saved_step_id;

      program.qstep = <Record<string, PositiveInteger>>tdata.qstep;
      program.qrequired = tdata.required || {};
      program.internal = tdata.internal || {};

      let given_current = no_kickback;
      let given_top = no_kickback;
      let given_saved = no_kickback;
      let given_data = {};

      // These will be called regardless of whether the step actually differs
      // from the current
      quote.setCurrentStepId = given_id => {
        given_current = given_id;
        return quote;
      };
      quote.setTopVisitedStepId = given_id => {
        given_top = given_id;
        return quote;
      };
      quote.setTopSavedStepId = given_id => {
        given_saved = given_id;
        return quote;
      };

      const {data} = tdata;
      quote.getDataByName = (field: string) => data[field] || [];

      // These are only relevent to a couple of test cases above, meant to
      // play well with the locking system
      quote.isLocked = () => tdata.locked || false;
      quote.getExplicitLockStep = () => tdata.lock_step || <PositiveInteger>0;
      quote.setData = given => {
        given_data = given;
        return quote;
      };

      pipe(
        kickBackToNewlyApplicable(program)(quote),

        IOE.fold(
          left => () => {
            expect(tdata.expected_error).to.equal(
              true,
              'unexpected Left: ' + left
            );
          },

          right => () => {
            expect(given_current).to.equal(tdata.expected_current_step_id);
            expect(given_top).to.equal(tdata.expected_current_step_id);
            expect(given_saved).to.equal(tdata.expected_saved_step_id);
            expect(given_data).to.deep.equal(tdata.expected_data);
            expect(right).to.equal(quote);
          }
        )
      )();
    })
  );

  describe('loader.mergeDocumentData', () => {
    it('merges two partial document data sets', () => {
      const now = ts_ctor();

      const sub_data = {startDate: now, agentName: 'foo'};
      const super_data = {startDate: now, agentId: 123};
      const expected = {startDate: now, agentName: 'foo', agentId: 123};

      expect(mergeDocumentData(sub_data)(super_data)).to.deep.equal(expected);
    });

    it('overwrites members of the sub_data', () => {
      const now = ts_ctor();

      const sub_data = {startDate: now, data: {foo: 'bar'}, bound: 1};
      const super_data = {
        startDate: (now + 10) as UnixTimestamp,
        data: {bar: 'foo'},
      };
      const expected = {
        startDate: (now + 10) as UnixTimestamp,
        data: {bar: 'foo'},
        bound: 1,
      };

      expect(mergeDocumentData(sub_data)(super_data)).to.deep.equal(expected);
    });
  });
});

describe('loader.loadSessionKeysIntoDocumentData', () => {
  (<
    {
      label: string;
      data: Record<string, unknown>;
      expected: Record<string, unknown>;
      kekMap?: ServiceSecretMap;
    }[]
  >[
    {
      label: 'no session keys in document results in them being created',
      data: {},
      expected: {
        sessionKeys: {liza: ['Generated:foo']},
        latestKekVersion: {liza: 1},
      },
    },
    {
      label: 'undefined session keys in document results in them being created',
      data: {sessionKeys: undefined},
      expected: {
        sessionKeys: {liza: ['Generated:foo']},
        latestKekVersion: {liza: 1},
      },
    },
    {
      label: 'session keys in document results in them not being created',
      data: {sessionKeys: {liza: ['foo']}, latestKekVersion: {liza: 1}},
      expected: {sessionKeys: {liza: ['foo']}, latestKekVersion: {liza: 1}},
    },
    {
      label: 'the same session key is encrypted multiple times',
      data: {sessionKeys: {liza: ['foo']}, latestKekVersion: {liza: 1}},
      expected: {sessionKeys: {liza: ['foo']}, latestKekVersion: {liza: 1}},
    },
  ]).forEach(({label, data, expected}) => {
    it(label, () => {
      const fake_kek = Buffer.from(randomBytes(keyBytes));
      const kek_map = {
        [KekService.INTERNAL_KEY]: {
          '1': {
            is_active: true,
            key: fake_kek,
          },
        },
      };
      const result = loadSessionKeysIntoDocumentData([])(kek_map)(
        <DocumentData>(<unknown>data)
      );

      // Overwrite generated session keys with something deterministic
      if (result?.sessionKeys?.liza?.[0].startsWith('Encrypted:')) {
        result.sessionKeys.liza[0] = 'Generated:foo';
      }

      expect(result).to.deep.equal(expected);
    });
  });

  it('the same session key is encrypted multiple times', () => {
    const fake_kek = Buffer.from(randomBytes(keyBytes));
    const kek_map = {
      [KekService.INTERNAL_KEY]: {
        '1': {
          is_active: true,
          key: fake_kek,
        },
      },
    };

    const result = loadSessionKeysIntoDocumentData([''])(kek_map)(
      <DocumentData>(<unknown>{})
    );

    const keys = Object.values(result.sessionKeys ?? {}).flat();

    const unique = new Set(keys);

    expect(unique.size).to.equal(1);
  });
});

describe('loader.loadSessionIntoQuote', () => {
  type TestData = {
    label: string;
    session_data: {
      agent_id: number;
      user_name: string;
      agency_number: string;
      agent_name: string;
      is_internal: boolean;
      roles?: string[] | undefined;
    };
    data: DocumentData;
    expected: {
      agent_id: number;
      user_name: string;
      agency_number: string;
      agent_name: string;
      internal_user: string[];
      user_roles?: string[] | [];
    };
  };

  (<TestData[]>[
    {
      label: 'sets values from session when no data is supplied',
      session_data: {
        agent_id: 123,
        user_name: 'Foo User',
        agency_number: 'AGT123',
        agent_name: 'Foo Agency',
        is_internal: false,
      },
      data: {},
      expected: {
        agent_id: 123,
        user_name: 'Foo User',
        agency_number: 'AGT123',
        agent_name: 'Foo Agency',
        internal_user: ['0'],
        user_roles: [],
      },
    },
    {
      label: 'prefer values from data when data is supplied',
      session_data: {
        agent_id: 123,
        user_name: 'Foo User',
        agency_number: 'AGT123',
        agent_name: 'Foo Agency',
        is_internal: false,
      },
      data: {
        agentId: 234,
        agentName: 'Bar Agency',
      },
      expected: {
        agent_id: 234,
        user_name: 'Foo User',
        agency_number: 'AGT123',
        agent_name: 'Bar Agency',
        internal_user: ['0'],
        user_roles: [],
      },
    },
    {
      label: 'default when values are not in session or data',
      session_data: {
        agent_id: undefined,
        user_name: undefined,
        agency_number: undefined,
        agent_name: undefined,
        is_internal: false,
      },
      data: {},
      expected: {
        agent_id: 0,
        user_name: '',
        agency_number: '',
        agent_name: '',
        internal_user: ['0'],
        user_roles: [],
      },
    },
    {
      label: 'internal session sets __internal_user bucket data to "1"',
      session_data: {
        is_internal: true,
      },
      data: {},
      expected: {
        agent_id: 123,
        user_name: 'foo user',
        agency_number: 'AGT234',
        agent_name: 'foo agent',
        internal_user: ['1'],
        user_roles: [],
      },
    },
    {
      label: 'non-internal session sets __internal_user bucket data to "0"',
      session_data: {
        is_internal: false,
      },
      data: {},
      expected: {
        agent_id: 123,
        user_name: 'foo user',
        agency_number: 'AGT234',
        agent_name: 'foo agent',
        internal_user: ['0'],
        user_roles: [],
      },
    },
    {
      label: 'user roles from session are set in __liza_access_roles',
      session_data: {
        roles: ['foo', 'bar'],
      },
      data: {},
      expected: {
        agent_id: 123,
        user_name: 'foo user',
        agency_number: 'AGT234',
        agent_name: 'foo agent',
        internal_user: ['0'],
        user_roles: ['foo', 'bar'],
      },
    },
    {
      label:
        'user roles are not available in session will not set any __liza_access_roles',
      session_data: {
        roles: undefined,
      },
      data: {},
      expected: {
        agent_id: 123,
        user_name: 'foo user',
        agency_number: 'AGT234',
        agent_name: 'foo agent',
        internal_user: ['0'],
        user_roles: [],
      },
    },
  ]).forEach(({label, ...tdata}) =>
    it(label, () => {
      const session = createStubSession(tdata.session_data);
      const quote = createStubQuote({}, {});
      const data = tdata.data;

      let setter_called = 0;

      quote.setAgentId = val => {
        expect(val).to.equal(tdata.expected.agent_id);
        setter_called++;
        return quote;
      };

      quote.setUserName = val => {
        expect(val).to.equal(tdata.expected.user_name);
        setter_called++;
        return quote;
      };

      quote.setUserRoles = val => {
        expect(val).to.deep.equal(tdata.expected.user_roles);
        setter_called++;
        return quote;
      };

      quote.setAgencyNumber = val => {
        expect(val).to.equal(tdata.expected.agency_number);
        setter_called++;
        return quote;
      };

      quote.setAgentName = val => {
        expect(val).to.equal(tdata.expected.agent_name);
        setter_called++;
        return quote;
      };

      quote.setData = (data: any) => {
        expect(data.__internal_user).to.deep.equal(
          tdata.expected.internal_user
        );
        expect(data.__liza_access_roles).to.deep.equal(
          tdata.expected.user_roles
        );
        setter_called++;
        return quote;
      };

      pipe(loadSessionIntoQuote(session)(quote)(data))();

      expect(setter_called).to.equal(6);
    })
  );
});

describe('loader.initNewDocument', () => {
  it('should initialize a document', async () => {
    const fake_kek = Buffer.from(randomBytes(keyBytes));
    const kek_map = {
      [KekService.INTERNAL_KEY]: {
        '1': {
          is_active: true,
          key: fake_kek,
        },
      },
    };
    const mockDao = <ServerDao>{
      getNextQuoteId: async () => 1,
    };
    const quoteCreator = createStubQuote({}, {}),
      program = {...createStubProgram(), groupExclusiveFields: {}},
      session = createStubSession(),
      feature_flag = createStubFeatureFlag(),
      init_doc_data = <DocumentData>{};
    const mockNewQuote = (_id: number) => quoteCreator;
    const actual = await initNewDocument(mockDao)(mockNewQuote)(program)(
      session
    )(kek_map)(feature_flag)(init_doc_data)();

    expect(E.isRight(actual)).to.be.true;
  });
});

describe('loader.createSessions', () => {
  const valid_kek = Buffer.from(randomBytes(keyBytes));

  it('Should generate a set of session keys given a list of services and keks', async () => {
    const kek_map = {
      liza: {'1': {key: valid_kek, is_active: true}},
      frankenrater: {'1': {key: valid_kek, is_active: true}},
    };

    const actual = await createSessions(createSessionKey(), kek_map, [
      'liza',
      'frankenrater',
    ]);

    expect(actual.liza?.[0]).to.match(/Encrypted:/);
    expect(actual.frankenrater?.[0]).to.match(/Encrypted:/);
  });

  it('Should filter out any inactive keks', async () => {
    const kek_map = {
      liza: {
        '1': {key: valid_kek, is_active: false},
        '2': {key: valid_kek, is_active: false},
        '3': {key: valid_kek, is_active: true},
      },
      frankenrater: {
        '1': {key: valid_kek, is_active: false},
        '2': {key: valid_kek, is_active: true},
        '3': {key: valid_kek, is_active: false},
      },
    };

    const actual = await createSessions(createSessionKey(), kek_map, [
      'liza',
      'frankenrater',
    ]);

    expect(actual.liza.length).to.eq(1);
    expect(actual.frankenrater.length).to.eq(1);
    expect(actual.liza?.[0]).to.match(/Encrypted:/);
    expect(actual.frankenrater?.[0]).to.match(/Encrypted:/);
  });
});

describe('loader.loadDocumentIntoQuote', () => {
  type TestData = {
    label: string;
    data: DocumentData;
    expected: {
      program_id: string;
    };
  };

  (<TestData[]>[
    {
      label: 'sets program id using data',
      data: {
        programId: 'xxxx',
      },
      expected: {
        program_id: 'xxxx',
      },
    },
  ]).forEach(({label, ...tdata}) =>
    it(label, () => {
      const quote = createStubQuote({}, {});
      const data = tdata.data;

      let setter_called = 0;

      quote.setProgramId = val => {
        expect(val).to.equal(tdata.expected.program_id);
        setter_called++;
        return quote;
      };

      pipe(loadDocumentIntoQuote(quote)(data))();

      expect(setter_called).to.equal(1);
    })
  );
});

describe('loader.mergeFlagData', () => {
  type TestData = {
    label: string;
    flags: Record<string, string>;
    flag_response: () => Promise<boolean>;
    input_data: Record<string, Record<string, any>>;
    expected: Record<string, Record<string, any>>;
  };
  (<TestData[]>[
    {
      label: 'A flag is resolved as true and added to the bucket',
      flags: {foo: 'bar'},
      flag_response: () => Promise.resolve(true),
      input_data: {},
      expected: {data: {foo: ['1']}},
    },
    {
      label: 'A flag is resolved as false and added to the bucket',
      flags: {foo: 'bar'},
      flag_response: () => Promise.resolve(false),
      input_data: {},
      expected: {data: {foo: ['0']}},
    },
    {
      label: 'Current bucket data is untouched',
      flags: {foo: 'bar'},
      flag_response: () => Promise.resolve(true),
      input_data: {data: {baz: ['whatever']}},
      expected: {
        data: {baz: ['whatever'], foo: ['1']},
      },
    },
    {
      label: 'Rejected feature flag promise defaults to zero',
      flags: {foo: 'bar'},
      flag_response: () => Promise.reject(),
      input_data: {},
      expected: {data: {foo: ['0']}},
    },
  ]).forEach(({label, flags, flag_response, input_data, expected}) => {
    it(label, done => {
      const program = createStubProgram(),
        session = createStubSession(),
        feature_flag = createStubFeatureFlag(flag_response);

      program.flags = flags;

      mergeFlagData(program)(session)(feature_flag)(
        <DocumentData>(<unknown>input_data)
      )().then((result: any) => {
        expect(result).to.deep.equal(expected);

        done();
      });
    });
  });
});

describe('loader.updateLatestVersions', () => {
  type TestData = {
    label: string;
    doc_data: {
      latestKekVersion?: Record<string, string> | undefined;
    };
    keks: ServiceSecretMap;
    expected: boolean;
  };
  (<TestData[]>[
    {
      label: 'should return true if no keks are present on doc',
      doc_data: {},
      keks: {
        'a-thing': {
          1: {key: Buffer.from('secret'), is_active: true},
        },
      },
      expected: true,
    },
    {
      label:
        'should return true if keks versions differ between doc and provided',
      doc_data: {
        latestKekVersion: {
          'a-thing': 2,
        },
      },
      keks: {
        'a-thing': {
          1: {key: Buffer.from('secret'), is_active: true},
        },
      },
      expected: true,
    },
    {
      label:
        'should return false if keks versions are the same between doc and provided',
      doc_data: {
        latestKekVersion: {
          'a-thing': 1,
        },
      },
      keks: {
        'a-thing': {
          1: {key: Buffer.from('secret'), is_active: true},
          2: {key: Buffer.from('secret'), is_active: false},
        },
      },
      expected: false,
    },
  ]).forEach(({label, doc_data, keks, expected}) => {
    it(label, () => {
      expect(updateLatestVersions(<DocumentData>doc_data, keks)).to.eq(
        expected
      );
    });
  });
});

describe('loader.retrieveOrGenerateSessionKey', () => {
  const valid_kek = Buffer.from(randomBytes(32));

  type TestData = {
    label: string;
    session_keys: string[];
    kek: Key;
    expected: number;
  };
  (<TestData[]>[
    {
      label: 'should generate a new session key if no keys are provided',
      session_keys: [],
      kek: valid_kek,
      expected: 0,
    },
    {
      label:
        'should generate a new session key if unable to decrypt any session keys',
      session_keys: ['1', '2', '3'],
      kek: valid_kek,
      expected: 0,
    },
  ]).forEach(({label, session_keys, kek, expected}) => {
    it(label, () => {
      expect(
        retrieveOrGenerateSessionKey(session_keys)(kek).getDecryptedText()
          .length
      ).to.be.gt(expected);
    });
  });

  it('should use a session key if its able to decrypt', () => {
    const session_key = createSessionKey();
    const encrypted_session_key = encryptSessionKey(valid_kek, session_key);

    const session_keys = ['1', '2', encrypted_session_key, '4'];

    expect(
      retrieveOrGenerateSessionKey(session_keys)(valid_kek).getDecryptedText()
    ).to.equal(session_key.getDecryptedText());
  });
});

function createStubQuote(
  data: Record<string, any[]>,
  metadata: Record<string, any>,
  session_key: string = ''
) {
  const quote = <ServerSideQuote>(<unknown>{
    getProgramId: () => 'foo',
    getId: () => 1,
    getSessionKeys: () => [session_key],
    setSessionKeys: () => quote,
    getLatestKeks: () => {},
    setLatestKeks: () => quote,
    getDownstreamSessionKeys: () => [],
    setDownstreamSessionKeys: () => quote,
    setData: (newData: Record<string, any> = {}) => {
      Object.keys(newData).forEach(
        field_name => (data[field_name] = newData[field_name])
      );
      return quote;
    },
    setUserRoles: () => quote,
    setAgencyNumber: () => quote,
    setAgentName: () => quote,
    setNonpersistent: () => quote,
    getDataByName: (name: string) => data[name],
    getMetabucket: () => ({
      getDataByName: (name: string) => metadata[name],
      getData: () => metadata,
      setValues: (data: Record<string, any>) => {
        Object.keys(data).forEach(
          field_name => (metadata[field_name] = data[field_name])
        );
      },
    }),
    getBucket: () => ({
      getData() {
        return data;
      },
      getDataByName: (name: string) => data[name],
    }),
    getCurrentStepId: () => 0,
    setCurrentStepId: () => quote,
    setTopVisitedStepId: () => quote,
    getTopVisitedStepId: () => 0,
    getTopSavedStepId: () => 0,
    setTopSavedStepId: () => quote,
    classify: () => ({}),
    getFieldState: () => ({}),
    getLastPersistedFieldState: () => ({}),
    isLocked: () => false,
    getExplicitLockStep: () => 0,
    setProgramId: () => quote,
    setLastPersistedFieldState: () => quote,
    setMetadata: () => quote,
    setAgentEntityId: () => quote,
    setInitialRatedDate: () => quote,
    setStartDate: () => quote,
    setImported: () => quote,
    setBound: () => quote,
    setAgentId: () => quote,
    setUserName: () => quote,
    needsImport: () => quote,
    setProgramVersion: () => quote,
    setExplicitLock: () => quote,
    setError: () => quote,
    setSaveError: () => quote,
    setCreditScoreRef: () => quote,
    setLastPremiumDate: () => quote,
    setRatedDate: () => quote,
    setRatingData: () => quote,
    setRetryAttempts: () => quote,
    getStartDate: () => 0,
  });

  return quote;
}

function createStubFeatureFlag(
  response: any = () => Promise.resolve(true)
): FeatureFlag {
  return <FeatureFlag>(<unknown>{
    isEnabled: response,
  });
}

function createStubProgram(
  meta_fields: Record<string, any> = {},
  fields: Record<string, any> = {}
) {
  return <___Writable<Program>>(<unknown>{
    getId: () => 'foo',
    meta: {fields: meta_fields},
    defaults: {},
    classify: () => ({}),
    processNaFields: () => {},
    downstreamServices: [''],
    encryptedKeys: {},
    ...fields,
  });
}

function createStubSession(data: Record<string, any> = {}) {
  return <UserSession>(<unknown>{
    // Use hasOwnProperty so we can explicitly set property to undefined
    agentId: () =>
      Object.prototype.hasOwnProperty.call(data, 'agent_id')
        ? data.agent_id
        : 123,
    userName: () =>
      Object.prototype.hasOwnProperty.call(data, 'user_name')
        ? data.user_name
        : 'foo user',
    userRoles: () =>
      Object.prototype.hasOwnProperty.call(data, 'roles')
        ? data.roles
        : undefined,
    agencyNumber: () =>
      Object.prototype.hasOwnProperty.call(data, 'agency_number')
        ? data.agency_number
        : 'AGT234',
    agentName: () =>
      Object.prototype.hasOwnProperty.call(data, 'agent_name')
        ? data.agent_name
        : 'foo agent',
    isInternal: () =>
      Object.prototype.hasOwnProperty.call(data, 'is_internal')
        ? data.is_internal
        : false,
  });
}
