/**
 * Tests DslRater
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {DslRaterContext} from '../../../src/server/rater/DslRaterContext';
import {DslRater as Sut} from '../../../src/server/rater/DslRater';
import {ResultSet} from '../../../src/server/rater/ResultSet';
import {Rater} from '../../../src/server/rater/Rater';
import {expect} from 'chai';

describe('DslRater', () => {
  describe('rate', () => {
    it('Calls context#rate', () => {
      let called = false;

      const raters = [
        getRaterStub((_: unknown, __: unknown) => {}, []),
        getRaterStub((_: unknown, __: unknown) => {}, []),
      ];
      const resultSet = getResultSetStub();
      const override = () => {
        called = true;
      };

      const context = getDslContext(override);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
    });

    it('Undefined rater calls context#complete', () => {
      let called = false;
      let rateCalled = false;

      const raters = <Rater[]>(<unknown>[undefined]);
      const resultSet = getResultSetStub();
      const override_rate = () => {
        rateCalled = true;
      };

      const processCompleted = () => {
        called = true;
      };

      const context = getDslContext(override_rate, processCompleted);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
      expect(rateCalled).to.equal(false);
    });

    it('Only rates applicable supplier if a name is passed in', () => {
      const given_supplier = 'bar';
      const raters_expected = [given_supplier];
      const raters_given: string[] = [];

      const rater_foo = getRaterStub(undefined, {'--elig-suppliers-foo': true});
      const rater_bar = getRaterStub(undefined, {'--elig-suppliers-bar': true});
      const rater_baz = getRaterStub(undefined, {'--elig-suppliers-baz': true});

      rater_foo.supplier = 'foo';
      rater_bar.supplier = 'bar';
      rater_baz.supplier = 'baz';

      const raters: Rater[] = [rater_foo, rater_bar, rater_baz];

      const override_rate = (name: string, _: any, __: any, cb: any) => {
        raters_given.push(name);
        cb();
      };

      const context = getDslContext(override_rate, () =>
        expect(raters_given).to.deep.equal(raters_expected)
      );
      const sut = new Sut(raters, getResultSetStub());

      overrideNextTick(() => sut.rate(context, given_supplier));
    });

    it('Only rates applicable suppliers passed in as requested', () => {
      const suppliers = ['baz', 'foo'];
      const raters_invoked: string[] = [];

      const rater_foo = getRaterStub(undefined, {'--elig-suppliers-foo': true});
      const rater_bar = getRaterStub(undefined, {'--elig-suppliers-bar': true});
      const rater_baz = getRaterStub(undefined, {'--elig-suppliers-baz': true});

      rater_foo.supplier = 'foo';
      rater_bar.supplier = 'bar';
      rater_baz.supplier = 'baz';

      const raters: Rater[] = [rater_foo, rater_bar, rater_baz];

      const override_rate = (name: string, _: any, __: any, cb: any) => {
        raters_invoked.push(name);
        cb();
      };

      const context = getDslContext(override_rate, () =>
        expect(raters_invoked).to.deep.equal(suppliers)
      );
      const sut = new Sut(raters, getResultSetStub());

      overrideNextTick(() => sut.rate(context, suppliers));
    });

    it('Calls rater with canTerm from context', () => {
      let actual;
      let called = false;

      const expected = false;
      const callback = (_data: any, canTerm: boolean) => {
        actual = canTerm;
        called = true;
      };

      const classes = {
        '--elig-suppliers-bar': true,
      };

      const raters = [getRaterStub(callback, classes)];
      const resultSet = getResultSetStub();
      const context = new DslRaterContext(null, expected);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
      expect(actual).to.equal(expected);
    });

    it('Handle undefined classes gracefully', () => {
      let called = false;

      const callback = (_: any, __: any) => {
        called = true;
      };

      const raters = [getRaterStub(callback, undefined)];
      const resultSet = getResultSetStub();
      const context = new DslRaterContext(null, false);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
    });

    it('Submit sets _unavailable flag', () => {
      let called = false;
      let actual = {};

      const expected = {
        _unavailable: '1',
        ineligible: '',
        submit: 'true',
        __classes: {
          submit: true,
          'submit-foo': true,
          '--elig-suppliers-bar': true,
        },
      };
      const raterCb = (_: any, __: any) => {};
      const resultCb = (name: any, _set: any) => {
        actual = name;
        called = true;
      };
      const classes = {
        'submit-foo': true,
        submit: true,
        '--elig-suppliers-bar': true,
      };
      const raters = [getRaterStub(raterCb, classes)];
      const resultSet = getResultSetStub(resultCb);
      const context = new DslRaterContext(null, true);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
      expect(actual).to.deep.equal(expected);
    });

    it('Prohibit sets _unavailable flag', () => {
      let called = false;
      let actual = {};

      const expected = {
        _unavailable: '1',
        ineligible: 'foo prohibit; baz prohibit',
        submit: '',
        __classes: {
          'inelig-foo': 'foo prohibit',
          'inelig-baz': 'baz prohibit',
          '--elig-suppliers-bar': false,
        },
      };
      const raterCb = (_: any, __: any) => {};
      const resultCb = (name: any, _set: any) => {
        actual = name;
        called = true;
      };

      const classes = {
        'inelig-foo': 'foo prohibit',
        'inelig-baz': 'baz prohibit',
        '--elig-suppliers-bar': false,
      };

      const raters = [getRaterStub(raterCb, classes)];
      const resultSet = getResultSetStub(resultCb);
      const context = getDslContext(undefined, undefined, expected, false);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
      expect(actual).to.deep.equal(expected);
    });

    it('Assertions sets _unavailable flag', () => {
      let called = false;
      let actual = {};

      const expected = {
        _unavailable: '1',
        ineligible: 'foo assertion; baz assertion',
        submit: '',
        __classes: {
          '-assert-foo': 'foo assertion',
          '-assert-bar': false,
          '-assert-baz': 'baz assertion',
          '--elig-suppliers-bar': false,
        },
      };
      const raterCb = (_: any, __: any) => {};
      const resultCb = (name: any, _set: any) => {
        actual = name;
        called = true;
      };

      const classes = {
        '-assert-foo': 'foo assertion',
        '-assert-bar': false,
        '-assert-baz': 'baz assertion',
        '--elig-suppliers-bar': false,
      };

      const raters = [getRaterStub(raterCb, classes)];
      const resultSet = getResultSetStub(resultCb);
      const context = getDslContext(undefined, undefined, expected, false);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
      expect(actual).to.deep.equal(expected);
    });

    it('Domain invalid sets _unavailable flag', () => {
      let called = false;
      let actual = {};

      const expected = {
        _unavailable: '1',
        ineligible: 'Domain violation: invalid value for foo',
        submit: '',
        __classes: {
          'domain-invalid-foo': 'Domain violation: invalid value for foo',
          '--elig-suppliers-bar': false,
        },
      };
      const raterCb = (_: any, __: any) => {};
      const resultCb = (name: any, _set: any) => {
        actual = name;
        called = true;
      };

      const classes = {
        'domain-invalid-foo': 'Domain violation: invalid value for foo',
        '--elig-suppliers-bar': false,
      };

      const raters = [getRaterStub(raterCb, classes)];
      const resultSet = getResultSetStub(resultCb);
      const context = getDslContext(undefined, undefined, expected, false);
      const sut = new Sut(raters, resultSet);

      sut.rate(context);

      expect(called).to.equal(true);
      expect(actual).to.deep.equal(expected);
    });
  });

  it('_unavailable flag defaults to false', () => {
    let called = false;
    let actual = {};

    const expected = {
      _unavailable: '0',
      ineligible: '',
      submit: '',
      __classes: {
        foo: true,
        submit: false,
        '--elig-suppliers-bar': true,
      },
    };
    const raterCb = (_: any, __: any) => {};
    const resultCb = (name: any, _set: any) => {
      actual = name;
      called = true;
    };
    const classes = {
      foo: true,
      submit: false,
      '--elig-suppliers-bar': true,
    };
    const raters = [getRaterStub(raterCb, classes)];
    const resultSet = getResultSetStub(resultCb);
    const context = new DslRaterContext(null, true);
    const sut = new Sut(raters, resultSet);

    sut.rate(context);

    expect(called).to.equal(true);
    expect(actual).to.deep.equal(expected);
  });

  it('--elig-supplier-{supplier} causes ineligibility', () => {
    let called = false;
    let actual = {};

    const expected = {
      _unavailable: '1',
      ineligible: 'This supplier is ineligible.',
      submit: '',
      __classes: {
        '--elig-suppliers-bar': false,
      },
    };
    const raterCb = (_: any, __: any) => {};
    const resultCb = (name: any, _set: any) => {
      actual = name;
      called = true;
    };

    const classes = {
      '--elig-suppliers-bar': false,
    };

    const raters = [getRaterStub(raterCb, classes)];
    const resultSet = getResultSetStub(resultCb);
    const context = getDslContext(undefined, undefined, expected, false);
    const sut = new Sut(raters, resultSet);

    sut.rate(context);

    expect(called).to.equal(true);
    expect(actual).to.deep.equal(expected);
  });

  it('missing --elig-supplier-{supplier} throws error', () => {
    const expected = {
      _unavailable: '1',
      ineligible: 'This supplier is ineligible.',
      submit: '',
      __classes: {},
    };
    const raterCb = () => {};
    const resultCb = () => {};

    const classes = {};

    const raters = [getRaterStub(raterCb, classes)];
    const resultSet = getResultSetStub(resultCb);
    const context = getDslContext(undefined, undefined, expected, false);
    const sut = new Sut(raters, resultSet);

    expect(() => sut.rate(context)).to.throw(
      'Missing supplier eligibility field: --elig-suppliers-bar'
    );
  });

  function getRaterStub(callback: any = () => {}, classes: any) {
    const raterStub: any = (() => {
      const raterStub = (data: any, canTerm: boolean) => {
        callback(data, canTerm);

        return {
          __classes: classes,
        };
      };

      raterStub.supplier = 'bar';
      raterStub.rater = {
        meta: 'foo',
        classify: {
          desc: classes,
        },
      };

      return raterStub;
    })();

    return raterStub;
  }

  function getResultSetStub(
    callback = (_: any, __: any) => {}
  ): (id: string) => ResultSet {
    return (_str: string) => {
      return <ResultSet>(<unknown>{
        addResult(name: any, set: any) {
          callback(name, set);
        },
      });
    };
  }

  function getDslContext(
    override_rate?: any,
    override_complete?: any,
    data?: any,
    canTerm?: boolean
  ): DslRaterContext {
    const context: any = new DslRaterContext(data, canTerm);

    if (override_rate) {
      context.rate = override_rate;
    }

    if (override_complete) {
      context.complete = override_complete;
    }

    context.addResultSet = () => {};

    return <DslRaterContext>context;
  }

  function overrideNextTick(cb: any) {
    const old_nextTick = global.process.nextTick;

    global.process.nextTick = tick_cb => tick_cb();

    cb();

    global.process.nextTick = old_nextTick;
  }
});
