/**
 * Tests Concurrent Session Manager
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {ConcurrentSessionManager} from '../../../src/server/request/ConcurrentSessionManager';
import {UserSession} from '../../../src/server/request/UserSession';
import {assert, expect} from 'chai';

describe('ConcurrentSessionManager', () => {
  describe('#init', () => {
    it('should delete sessions on an interval', () => {
      const old_setInterval = global.setInterval;

      try {
        const quoteId = 123;
        let setIntervalCalled = false;
        let intervalCb = () => {};
        let tsReturned = <UnixTimestamp>0;

        const session1 = getMockUserSession();
        const session2 = getMockUserSession();

        session2.userName = () => 'Frank';

        const sessionManager = new ConcurrentSessionManager(() => tsReturned);

        global.setInterval = (
          callback: (...args: any[]) => void,
          interval: number
        ) => {
          expect(interval).to.equal(
            ConcurrentSessionManager.CLEAR_INTERVAL_SEC
          );

          setIntervalCalled = true;

          intervalCb = callback;

          return <NodeJS.Timeout>{};
        };

        // Create a session with the timestamp of 0
        sessionManager.touchSession(quoteId, session1);

        expect(
          sessionManager.getConcurrentSessionUser(quoteId, session2),
          'Ensure we have the session we just created by ' +
            'checking with a session that has a different user'
        ).to.equal(session1.userName());

        // Start the interval
        sessionManager.init();
        expect(setIntervalCalled, 'setInterval called').to.be.true;

        // Advance our time by 1s and trigger the interval
        tsReturned = <UnixTimestamp>1;
        intervalCb();

        expect(
          sessionManager.getConcurrentSessionUser(quoteId, session2),
          'Ensure the session is still there because not enough time has lapsed'
        ).to.equal(session1.userName());

        // Advance our time by 100s and trigger the interval
        tsReturned = <UnixTimestamp>101;
        intervalCb();

        expect(
          sessionManager.getConcurrentSessionUser(quoteId, session2),
          'Expect that the session has now been cleared'
        ).to.equal('');
      } catch (e) {
        assert.fail(e);
      } finally {
        global.setInterval = old_setInterval;
      }
    });

    it('should only setup an interval once', () => {
      const old_setInterval = global.setInterval;

      try {
        let setIntervalCallCount = 0;

        global.setInterval = (_: (...args: any[]) => void, __: number) => {
          setIntervalCallCount++;

          return <NodeJS.Timeout>{};
        };

        new ConcurrentSessionManager(() => <UnixTimestamp>0).init().init();

        expect(setIntervalCallCount).to.equal(1);
      } catch (e) {
        assert.fail(e);
      } finally {
        global.setInterval = old_setInterval;
      }
    });
  });

  describe('touching and getting sessions', () => {
    it('should return empty string if no session flag exists', () => {
      const sessionManager = new ConcurrentSessionManager(
        () => <UnixTimestamp>0
      );
      const userSession = getMockUserSession();

      expect(sessionManager.getConcurrentSessionUser(1, userSession)).to.equal(
        ''
      );
    });

    it('should return empty string if session flag belongs to the same user', () => {
      const sessionManager = new ConcurrentSessionManager(
        () => <UnixTimestamp>0
      );
      const userSession = getMockUserSession();

      sessionManager.touchSession(1, userSession);
      expect(sessionManager.getConcurrentSessionUser(1, userSession)).to.equal(
        ''
      );
    });

    it('should return userName if session flag belongs to a different user', () => {
      const sessionManager = new ConcurrentSessionManager(
        () => <UnixTimestamp>0
      );
      const userSession = getMockUserSession();
      const differentUserSession = getMockUserSession();

      differentUserSession.userName = () => 'Frank';

      sessionManager.touchSession(1, userSession);
      expect(
        sessionManager.getConcurrentSessionUser(1, differentUserSession)
      ).to.equal(userSession.userName());
    });
  });
});

function getMockUserSession(): UserSession {
  return <UserSession>{
    userName: () => 'Shadynasty',
  };
}
