/**
 * Tests UserSession
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {expect} from 'chai';
import {UserSession as Sut} from '../../../src/server/request/UserSession';

describe('UserSession', () => {
  describe('UserSession#getters', () => {
    [
      {
        label: 'session data is unserialized',
        session_data:
          'agentID|s:6:"900000";' +
          'retail_agency_num|s:8:"AGT12345";' +
          'user_name|s:11:"foo@bar.com";' +
          'agentNAME|s:3:"Foo";' +
          'user_roles|a:2:{i:0;s:13:"ROLE_EMPLOYEE";i:1;s:10:"ROLE_ADMIN";};',
        expected_username: 'foo@bar.com',
        expected_user_roles: ['ROLE_EMPLOYEE', 'ROLE_ADMIN'],
        expected_agent_name: 'Foo',
        expected_agent_id: 900000,
        expected_agency_num: 'AGT12345',
        expected_data: {
          agentID: '900000',
          agentNAME: 'Foo',
          retail_agency_num: 'AGT12345',
          user_name: 'foo@bar.com',
          user_roles: ['ROLE_EMPLOYEE', 'ROLE_ADMIN'],
        },
      },
      {
        label: 'session data is undefined when unavailable',
        session_data: '',
        expected_username: undefined,
        expected_user_roles: undefined,
        expected_agent_name: undefined,
        expected_agent_id: undefined,
        expected_agency_num: undefined,
        expected_data: {},
      },
    ].forEach(
      ({
        label,
        session_data,
        expected_username,
        expected_user_roles,
        expected_agent_name,
        expected_agent_id,
        expected_agency_num,
        expected_data,
      }) => {
        it(label, () => {
          const session_id = '12345';
          const memcache = {
            get: (id: any, callback: (_err: any, data: any) => void) => {
              expect(id).to.be.equal(session_id);
              callback(undefined, session_data);
            },
          };
          const sut = new Sut(session_id, memcache);
          expect(sut.getData()).to.deep.equal(expected_data);
          expect(sut.userName()).to.equal(expected_username);
          expect(sut.userRoles()).to.deep.equal(expected_user_roles);
          expect(sut.agentName()).to.equal(expected_agent_name);
          expect(sut.agentId()).to.equal(expected_agent_id);
          expect(sut.agencyNumber()).to.equal(expected_agency_num);
        });
      }
    );
  });

  describe('UserSession#isInternal', () => {
    [
      {
        label: 'Session is internal with agentId',
        agentId: 900000,
        session_data: '',
        expected: true,
      },
      {
        label: 'Session is internal with role',
        agentId: 900000,
        session_data:
          'user_roles|a:2:{i:0;s:13:"ROLE_EMPLOYEE";i:1;s:10:"ROLE_ADMIN";};',
        expected: true,
      },
      {
        label: 'Session is not internal with incorrect role',
        agentId: 900001,
        session_data:
          'user_roles|a:2:{i:0;s:15:"ROLE_AGGREGATOR";i:1;s:10:"ROLE_ADMIN";};',
        expected: false,
      },
      {
        label: 'Session is not internal with no roles',
        agentId: 900001,
        session_data: '',
        expected: false,
      },
      {
        label: 'Session is not internal with employee role',
        agentId: 900001,
        session_data:
          'user_roles|a:2:{i:0;s:13:"ROLE_EMPLOYEE";i:1;s:10:"ROLE_ADMIN";};',
        expected: true,
      },
    ].forEach(({label, agentId, session_data, expected}) => {
      it(label, () => {
        const session_id = '2';
        const memcache = {
          get: (id: any, callback: (_err: any, data: any) => void) => {
            expect(id).to.be.equal(session_id);
            callback(undefined, session_data);
          },
        };

        const sut = new Sut(session_id, memcache);
        sut.setAgentId(agentId);
        expect(sut.agentId()).to.equal(agentId);
        expect(sut.isInternal()).to.equal(expected);
      });
    });
  });

  describe('UserSession#canUpdateMeta', () => {
    [
      {
        label: 'Session is internal with agentId',
        agentId: 900000,
        session_data: '',
        expected: true,
      },
      {
        label: 'Session is internal with role',
        agentId: 900000,
        session_data:
          'user_roles|a:2:{i:0;s:13:"ROLE_EMPLOYEE";i:1;s:10:"ROLE_ADMIN";};',
        expected: true,
      },
      {
        label: 'Session is not internal with incorrect role',
        agentId: 900001,
        session_data:
          'user_roles|a:2:{i:0;s:8:"ROLE_FOO";i:1;s:10:"ROLE_ADMIN";};',
        expected: false,
      },
      {
        label: 'Session is not internal with no roles',
        agentId: 900001,
        session_data: '',
        expected: false,
      },
      {
        label: 'Session is not internal with employee role',
        agentId: 900001,
        session_data:
          'user_roles|a:2:{i:0;s:13:"ROLE_EMPLOYEE";i:1;s:10:"ROLE_ADMIN";};',
        expected: true,
      },
      {
        label:
          'Session is not internal or employee role but has aggregator role',
        agentId: 900001,
        session_data:
          'user_roles|a:2:{i:0;s:15:"ROLE_AGGREGATOR";i:1;s:10:"ROLE_ADMIN";};',
        expected: true,
      },
    ].forEach(({label, agentId, session_data, expected}) => {
      it(label, () => {
        const session_id = '2';
        const memcache = {
          get: (id: any, callback: (_err: any, data: any) => void) => {
            expect(id).to.be.equal(session_id);
            callback(undefined, session_data);
          },
        };

        const sut = new Sut(session_id, memcache);
        sut.setAgentId(agentId);
        expect(sut.agentId()).to.equal(agentId);
        expect(sut.canUpdateMeta()).to.equal(expected);
      });
    });
  });
});
