/**
 * Tests Encryption Service
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {
  decryptDeltaFieldData,
  decryptQuoteData as sutDecrypt,
  encryptQuoteData as sutEncrypt,
} from '../../../src/server/service/EncryptionService';
import {EncryptedKeys} from '../../../src/program/Program';
import {BucketData} from '../../../src/bucket/Bucket';
import {Decryptor, Encryptor} from '../../../liza-crypto';
import {ProtectedString} from '../../../liza-crypto/src/types';
import {invalidate} from '../../../src/validate/invalid';

const expect = require('chai').expect;

describe('EncryptionService', () => {
  // Note: each of these tests is intentionally in their own block so that
  // you can observe, via an LSP, the types of the results, which differ in
  // each example.
  describe('decryptDeltaFieldData', () => {
    const dec = getMockDecryptor();

    it('decrypts string values', () => {
      const given = 'foo';

      const result = decryptDeltaFieldData(dec)(given);
      expect(result).to.equal('Decrypted:' + given);
    });

    it('decrypts array of strings', () => {
      const given = ['foo', 'bar', 'baz'];

      const result = decryptDeltaFieldData(dec)(given);
      expect(result).to.deep.equal(given.map(x => 'Decrypted:' + x));
    });

    it('does not modify other values', () => {
      const given = {foo: {bar: 'baz'}};

      const result = decryptDeltaFieldData(dec)(given);
      expect(result).to.deep.equal(given);
    });
  });

  describe('encryptQuoteData', () => {
    (<
      {
        label: string;
        encrypted_keys: EncryptedKeys;
        to_encrypt: BucketData;
        expected: BucketData;
      }[]
    >[
      {
        label: 'should function with no data supplied',
        encrypted_keys: {},
        to_encrypt: {},
        expected: {},
      },
      {
        label: 'should change nothing if no encrypted keys are supplied',
        encrypted_keys: {},
        to_encrypt: {foo: ['bar']},
        expected: {foo: ['bar']},
      },
      {
        label: 'only encrypts fields specified by encrypted_keys',
        encrypted_keys: {foo: ['service']},
        to_encrypt: {foo: ['bar'], moo: ['cow']},
        expected: {foo: ['Encrypted:bar'], moo: ['cow']},
      },
      {
        label: 'ignores missing fields specified by encrypted_keys',
        encrypted_keys: {missing: ['service']},
        to_encrypt: {foo: ['bar']},
        expected: {foo: ['bar']},
      },
      {
        label: 'encrypts every index for a specified field',
        encrypted_keys: {foo: ['service']},
        to_encrypt: {foo: ['bar', 'baz', 'yolo']},
        expected: {foo: ['Encrypted:bar', 'Encrypted:baz', 'Encrypted:yolo']},
      },
      // TODO: We shouldn't have to worry about delta implementation details
      // here; deltas should provide a mapping operation that skips these
      // values itself.
      {
        label:
          'does not encrypt nulls outside of tail position (non-modifications)',
        encrypted_keys: {foo: ['service']},
        to_encrypt: {foo: ['bar', null, 'baz']}, // index 1 is unmodified
        expected: {foo: ['Encrypted:bar', null, 'Encrypted:baz']},
      },
      // TODO: Here too.
      {
        label: 'does not encrypt trailing nulls (truncation)',
        encrypted_keys: {foo: ['service']},
        to_encrypt: {foo: ['bar', null]}, // truncated at index 1
        expected: {foo: ['Encrypted:bar', null]},
      },
      // TODO: Similarly, we should not have to make assumptions about field
      // applicability.  Here, we (a) assume that empty means "not
      // applicable", and (b) that the system enforces a configuration
      // (through proguic, which Liza is not supposed to know exists)
      // prohibiting the user from supplying empty values.
      {
        label: 'does not encrypt the empty string',
        encrypted_keys: {foo: ['service']},
        to_encrypt: {foo: ['', 'foo', '']},
        expected: {foo: ['', 'Encrypted:foo', '']},
      },
      // TODO: We also shouldn't have to worry about every possible field
      // value variant
      {
        label: 'does not encrypt invalid field states',
        encrypted_keys: {foo: ['service']},
        to_encrypt: {foo: ['ok', invalidate('ignore')]},
        expected: {foo: ['Encrypted:ok', invalidate('ignore')]},
      },
    ]).forEach(({label, encrypted_keys, to_encrypt, expected}) => {
      it(label, () => {
        sutEncrypt(() => getMockEncryptor())(encrypted_keys, [], to_encrypt);

        expect(to_encrypt).to.deep.equal(expected);
      });
    });
  });

  describe('decryptQuoteData', () => {
    (<
      {
        label: string;
        encrypted_keys: EncryptedKeys;
        to_decrypt: BucketData;
        expected: BucketData;
      }[]
    >[
      {
        label: 'should function with no data supplied',
        encrypted_keys: {},
        to_decrypt: {},
        expected: {},
      },
      {
        label: 'should change nothing if no encrypted keys are supplied',
        encrypted_keys: {},
        to_decrypt: {foo: ['bar']},
        expected: {foo: ['bar']},
      },
      {
        label: 'only decrypts fields specified by encrypted_keys',
        encrypted_keys: {foo: ['service']},
        to_decrypt: {foo: ['bar'], moo: ['cow']},
        expected: {foo: ['Decrypted:bar'], moo: ['cow']},
      },
      {
        label: 'ignores missing fields specified by encrypted_keys',
        encrypted_keys: {missing: ['service']},
        to_decrypt: {foo: ['bar']},
        expected: {foo: ['bar']},
      },
      {
        label: 'decrypts every index for a specified field',
        encrypted_keys: {foo: ['service']},
        to_decrypt: {foo: ['bar', 'baz', 'yolo']},
        expected: {foo: ['Decrypted:bar', 'Decrypted:baz', 'Decrypted:yolo']},
      },
      {
        label: 'does not attempt to encrypt empty strings',
        encrypted_keys: {foo: ['service']},
        to_decrypt: {foo: ['']},
        expected: {foo: ['']},
      },
      {
        label: 'does not error if encrypted fields are not in the bucket',
        encrypted_keys: {foo: ['service']},
        to_decrypt: {},
        expected: {},
      },
      // TODO: We shouldn't have to worry about every possible field value
      // variant
      {
        label: 'does not encrypt invalid field states',
        encrypted_keys: {foo: ['service']},
        to_decrypt: {foo: ['ok', invalidate('ignore')]},
        expected: {foo: ['Decrypted:ok', invalidate('ignore')]},
      },
    ]).forEach(({label, encrypted_keys, to_decrypt, expected}) => {
      it(label, () => {
        expect(
          sutDecrypt(() => getMockDecryptor())(encrypted_keys, [], to_decrypt)
        ).to.deep.equal(expected);
      });
    });
  });
});

function getMockEncryptor(): Encryptor {
  return <Encryptor>{
    encrypt(val: string) {
      return 'Encrypted:' + val;
    },
  };
}

function getMockDecryptor(): Decryptor {
  return <Decryptor>{
    decrypt(val: string) {
      return new ProtectedString('Decrypted:' + val);
    },
  };
}
