import {SecretManagerServiceClient} from '@google-cloud/secret-manager';
import {GcpSecretStore} from '../../../src/server/service/GcpSecretStore';
import {expect} from 'chai';

describe('GcpSecretStore', () => {
  describe('getSecret', () => {
    it('Should error if no kek is found for a given service-name', () => {
      const versioned_secret = 'foo';
      const client: any = getMockClient(() => {});
      client.accessSecretVersion = () => {};
      const sut = new GcpSecretStore('', client);

      return expect(
        sut.getSecret(versioned_secret)
      ).to.eventually.be.rejectedWith(
        Error,
        `No kek found for ${versioned_secret}`
      );
    });
    it('should request a secret version with a given uri', async () => {
      const project_id = 'foo';
      const service = 'bar';
      const versioned_secret = `projects/${project_id}/secrets/${service}-kek`;

      let accessCalled = false;
      const client = getMockClient((res: {name: string}) => {
        expect(res).to.deep.equal({name: versioned_secret});
        accessCalled = true;
      }, 'fake-kek');

      const sut = new GcpSecretStore(project_id, client);
      await sut.getSecret(versioned_secret);

      expect(accessCalled).to.be.true;
    });
  });

  describe('listSecretVersions', () => {
    it('Should return a list of secret names/versions', async () => {
      const service = 'foo';
      const proj_id = 'some-id';
      const expected = [
        {
          name: `projects/${proj_id}/secrets/${service}-kek/versions/1`,
          version: 1,
        },
      ];

      const client = getMockClient(() => {});
      const actual = await new GcpSecretStore(
        proj_id,
        client
      ).listSecretVersions(service);

      expect(actual).to.deep.equal(expected);
    });

    it('Should error if an un-named version is returned', () => {
      const service = 'foo';

      const client: any = getMockClient(() => {});
      client.listSecretVersionsAsync = () => [{name: undefined}];

      const sut = new GcpSecretStore('', client);

      return expect(
        sut.listSecretVersions(service)
      ).to.eventually.be.rejectedWith(
        Error,
        'Secret version is un-named and cannot be retrieved'
      );
    });

    it('Should error if a version number is unavailable', () => {
      const service = 'foo';

      const client: any = getMockClient(() => {});
      client.listSecretVersionsAsync = () => [{name: 'totes-invalid'}];

      const sut = new GcpSecretStore('', client);

      return expect(
        sut.listSecretVersions(service)
      ).to.eventually.be.rejectedWith(
        Error,
        'Version number is less than one, tried to convert: null'
      );
    });
  });
});

const getMockClient = (cb: (res: {name: string}) => void, ret?: string) =>
  <SecretManagerServiceClient>(<unknown>{
    listSecretVersionsAsync: (res: {filter: string; parent: string}) => {
      return [
        {
          name: `${res.parent}/versions/1`,
        },
      ];
    },
    accessSecretVersion: async (res: {name: string}) => {
      cb(res);

      return {
        0: {
          payload: {
            data: {
              toString: () => ret,
            },
          },
        },
      };
    },
  });
