/**
 * Tests Kek Service
 *
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import {KekService} from '../../../src/server/service/KekService';
import {expect, use as chai_use} from 'chai';
import {SecretStore} from '../../../src/server/service/SecretStore';

chai_use(require('chai-as-promised'));

describe('KekSevice', () => {
  let blacklisted_keks: string[] = [];
  describe('#fetch', () => {
    it('should return the key from the secret manager', async () => {
      const expected = 'super secret key';
      const client = getMockClient(() => {}, Buffer.from(expected));

      const kekService = new KekService(client, () => {}, blacklisted_keks);

      const key = await kekService.fetch('bar');
      expect(key[0]?.[1]?.key.toString()).to.equal(expected);
    });

    it('should throw an error on an invalid key', () => {
      const errMsg = 'Invalid Key';
      const client = getMockClient(() => {}, Buffer.from('something illegal'));
      const kekService = new KekService(
        client,
        () => {
          throw new Error(errMsg);
        },
        blacklisted_keks
      );

      return expect(kekService.fetch('bar')).to.eventually.be.rejectedWith(
        Error,
        errMsg
      );
    });

    it('should throw an error if no keys are retrieved', () => {
      const client: any = getMockClient(() => {});
      client.listSecretVersions = () => [];
      const kekService = new KekService(client, () => {}, blacklisted_keks);

      return expect(kekService.fetch('bar')).to.eventually.be.rejectedWith(
        Error,
        'No valid keks found for service: bar'
      );
    });

    it('should throw an error if no active keys are retrieved', () => {
      const service = 'foo';
      blacklisted_keks = [`${service}_kek:1`];
      const client = getMockClient(() => {});
      const kekService = new KekService(client, () => {}, blacklisted_keks);

      return expect(kekService.fetch(service)).to.eventually.be.rejectedWith(
        Error,
        `No active keks found for service: ${service}`
      );
    });

    it('client errors are allowed to bubble up to the caller', () => {
      const client = getMockClient(() => {
        throw new Error('Boom!');
      });
      const kekService = new KekService(client, () => {}, blacklisted_keks);

      return expect(kekService.fetch('bar')).to.eventually.be.rejectedWith(
        Error,
        'Boom!'
      );
    });
  });

  describe('#fetchAll', () => {
    it('should create a map of system list to kek', async () => {
      const expected = 'itsakek!';
      const client = getMockClient(() => 'some-name');

      const kekService = new KekService(client, () => {}, blacklisted_keks);
      const actual = await kekService.fetchAll(['a']);

      expect(Object.keys(actual).length).to.be.greaterThan(0);
      Object.keys(actual).forEach(service =>
        expect(actual[service]?.[1]?.key.toString()).to.equal(expected)
      );
    });
  });
});

const getMockClient = (
  cb: (name: string) => void,
  kek?: Buffer
): SecretStore => ({
  listSecretVersions: async service => {
    return [
      {
        name: service,
        version: 1,
      },
    ];
  },
  getSecret: async (secret_name: string) => {
    cb(secret_name);

    return kek ?? Buffer.from('itsakek!');
  },
});
