/**
 * Tests RatingService
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {
  RatingService as Sut,
  RateRequestResult,
} from '../../../src/server/service/RatingService';

import {ClientActions} from '../../../src/client/action/ClientAction';
import {PriorityLog} from '../../../src/server/log/PriorityLog';
import {ProcessManager} from '../../../src/server/rater/ProcessManager';
import {Program, ClassificationResult} from '../../../src/program/Program';
import {QuoteId} from '../../../src/quote/Quote';
import {
  Rater,
  RateResult,
  WorksheetData,
} from '../../../src/server/rater/Rater';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {UserSession} from '../../../src/server/request/UserSession';
import {QuoteDataBucket} from '../../../src/bucket/QuoteDataBucket';
import {PositiveInteger} from '../../../src/numeric';
import {Kv} from '../../../src/bucket/delta';

import {
  ServerDao,
  Callback as ServerDaoCallback,
} from '../../../src/server/db/ServerDao';

import {expect, use as chai_use} from 'chai';
chai_use(require('chai-as-promised'));

describe('RatingService', () => {
  it('returns rating results', () => {
    const {logger, raters, dao, session, quote, stub_rate_data, ts_ctor} =
      getStubs();

    const sut = new Sut(logger, dao, raters, ts_ctor);

    return expect(sut.request(session, quote)).to.eventually.deep.equal(
      stub_rate_data
    );
  });

  it('fetches rating results', () => {
    const {logger, raters, dao, quote, stub_rate_data, ts_ctor} = getStubs();

    let merge_called = false;

    const sut = new Sut(logger, dao, raters, ts_ctor);

    dao.mergeBucket = (_: any, __: any, cb: any) => {
      merge_called = true;

      cb();

      return dao;
    };

    const expected = {
      content: {
        data: stub_rate_data,
        initialRatedDate: quote.getRatedDate(),
        lastRatedDate: quote.getLastPremiumDate(),
      },
      actions: [],
    };

    return expect(sut.fetchRate(quote))
      .to.eventually.deep.equal(expected)
      .then(() => expect(merge_called).to.be.true);
  });

  it('updates rating dates before serving to client', () => {
    const {logger, raters, dao, session, quote, stub_rate_data, ts_ctor} =
      getStubs();

    const sut = new Sut(logger, dao, raters, ts_ctor);

    let last_prem_called = false;
    let rated_date_called = false;

    const stub_last_prem_ts = <UnixTimestamp>0;
    const stub_rated_date_ts = <UnixTimestamp>0;

    quote.setLastPremiumDate = () => {
      last_prem_called = true;
      return quote;
    };

    quote.setRatedDate = () => {
      rated_date_called = true;
      return quote;
    };

    quote.getLastPremiumDate = () => stub_last_prem_ts;
    quote.getRatedDate = () => stub_rated_date_ts;

    return expect(sut.request(session, quote))
      .to.eventually.deep.equal(stub_rate_data)
      .then(() => {
        expect(last_prem_called).to.be.true;
        expect(rated_date_called).to.be.true;
      });
  });

  it('sets last premium date to zero if there was an override', () => {
    const {logger, raters, dao, session, quote, stub_rate_data, ts_ctor} =
      getStubs();

    const rater: Rater = <any>new (class {
      rate(
        _quote: ServerSideQuote,
        _session: UserSession,
        _indv: string,
        success: (
          data: RateResult,
          actions: ClientActions,
          override: boolean
        ) => void,
        _failure: (message: string) => void
      ) {
        // force to be async so that the tests resemble how the code
        // actually runs
        process.nextTick(() => success(stub_rate_data, [], true));

        return this;
      }
    })();

    raters.byId = _ => rater;

    const sut = new Sut(logger, dao, raters, ts_ctor);

    let last_prem_called = false;
    let rated_date_called = false;

    let stub_last_prem_ts = <UnixTimestamp>123;
    let stub_rated_date_ts = <UnixTimestamp>2592001;

    quote.setLastPremiumDate = ts => {
      stub_last_prem_ts = ts;
      last_prem_called = true;
      return quote;
    };

    quote.setRatedDate = ts => {
      stub_rated_date_ts = ts;
      rated_date_called = true;
      return quote;
    };

    quote.getLastPremiumDate = () => stub_last_prem_ts;
    quote.getRatedDate = () => stub_rated_date_ts;

    return expect(sut.request(session, quote))
      .to.eventually.deep.equal(stub_rate_data)
      .then((_: RateRequestResult) => {
        expect(last_prem_called, 'baz').to.be.true;
        expect(rated_date_called, 'baax').to.be.true;
      });
  });

  it('saves rate data to its own field', () => {
    const {logger, raters, dao, session, quote, stub_rate_data, ts_ctor} =
      getStubs();

    let saved_rates = false;

    let saveQuote_call_count = 0;

    dao.saveQuote = (
      quote: ServerSideQuote,
      _class_data: ClassificationResult,
      success: ServerDaoCallback,
      _failure: ServerDaoCallback,
      save_data: Record<string, any>,
      _push_data: Record<string, any>
    ) => {
      expect(save_data.ratedata).to.deep.equal(stub_rate_data);

      saveQuote_call_count++;
      saved_rates = true;
      success(quote);

      return dao;
    };

    const sut = new Sut(logger, dao, raters, ts_ctor);

    return sut.request(session, quote).then(() => {
      expect(saved_rates).to.be.true;
      expect(saveQuote_call_count).to.equal(1);
    });
  });

  it('saves delta to its own field', () => {
    const {logger, raters, dao, session, quote, stub_rate_delta, ts_ctor} =
      getStubs();

    let saved_quote = false;

    let timestamp = 0;

    quote.setLastPremiumDate = (ts: UnixTimestamp) => {
      timestamp = ts;
      return quote;
    };

    let saveQuote_call_count = 0;

    dao.saveQuote = (
      quote: ServerSideQuote,
      _class_data: ClassificationResult,
      success: ServerDaoCallback,
      _failure: ServerDaoCallback,
      _save_data: Record<string, any>,
      push_data: Record<string, any>
    ) => {
      stub_rate_delta['rdelta.ratedata'].timestamp = timestamp;
      stub_rate_delta['rdelta.ratedata'].data = {};
      saved_quote = true;

      expect(push_data).to.deep.equal(stub_rate_delta);

      saveQuote_call_count++;
      success(quote);

      return dao;
    };

    const sut = new Sut(logger, dao, raters, ts_ctor);

    return sut.request(session, quote).then(() => {
      expect(saved_quote).to.be.true;
      expect(saveQuote_call_count).to.equal(1);
    });
  });

  it('invokes classify if it has not been run yet', () => {
    const {logger, raters, dao, session, quote, ts_ctor} = getStubs();

    let classify_called = false;
    let field_state_checked = false;

    quote.getFieldState = () => {
      field_state_checked = true;

      return undefined;
    };

    quote.classify = () => {
      classify_called = true;

      return {};
    };

    const sut = new Sut(logger, dao, raters, ts_ctor);

    return sut.request(session, quote).then(() => {
      expect(classify_called).to.be.true;
      expect(field_state_checked).to.be.true;
    });
  });

  it('does not invoke classify if it has already been invoked', () => {
    const {logger, raters, dao, session, quote, ts_ctor} = getStubs();

    let classify_called = false;
    let field_state_checked = false;

    quote.getFieldState = () => {
      field_state_checked = true;

      return {};
    };

    quote.classify = () => {
      classify_called = true;

      return {};
    };

    const sut = new Sut(logger, dao, raters, ts_ctor);

    return sut.request(session, quote).then(() => {
      expect(classify_called).to.be.false;
      expect(field_state_checked).to.be.true;
    });
  });

  it('rejects and responds with error', () => {
    const {dao, logger, program, quote, rater, raters, session, ts_ctor} =
      getStubs();

    const expected_error = new Error('expected error');

    rater.rate = () => {
      throw expected_error;
    };

    const sut = new Sut(logger, dao, raters, ts_ctor);

    let logged = false;

    logger.log = function (
      priority: number,
      _format: string,
      qid: QuoteId,
      program_id: string,
      message: string
    ) {
      if (typeof message === 'string') {
        expect(priority).to.equal(PriorityLog.PRIORITY_ERROR);
        expect(qid).to.equal(quote.getId());
        expect(program_id).to.equal(program.getId());
        expect(message).to.contain(expected_error.message);

        logged = true;
      }

      return logger;
    };

    return expect(sut.request(session, quote))
      .to.eventually.rejectedWith(expected_error)
      .then(() => expect(logged).to.be.true);
  });

  it('returns error message from rater', () => {
    const {dao, logger, quote, rater, raters, session, ts_ctor} = getStubs();

    const expected_message = 'expected foo';

    const sut = new Sut(logger, dao, raters, ts_ctor);

    rater.rate = (
      _quote: ServerSideQuote,
      _session: UserSession,
      _indv: string,
      _success: (data: RateResult, actions: ClientActions) => void,
      failure: (message: string) => void
    ) => {
      failure(expected_message);
      return rater;
    };

    return expect(sut.request(session, quote)).to.eventually.rejectedWith(
      Error,
      expected_message
    );
  });

  it('Consider missing retry flag to still be pending', () => {
    // The stub rate data returned on rate has a total of 2 suppliers and
    // we are specifying that there in only 1 retry field, so one must be
    // missing. Missing retry attempts should be treated as pending so we
    // expect the retry attempt counter to be incremented
    const stub_rate_data = {
      foo__unavailable: '0',
      bar__unavailable: '0',
      bar___retry: '0',
      __unavailable_all: '0',
    };

    const {dao, logger, quote, raters, session, ts_ctor} =
      getStubs(stub_rate_data);

    const retry_attempts_current = 24;
    const retry_attempts_expected = 25;

    let retry_attempts_given: number;

    quote.setRetryAttempts = (attempts: number) => {
      retry_attempts_given = attempts;

      return quote;
    };

    quote.getRetryAttempts = () => retry_attempts_current;

    const sut = new Sut(logger, dao, raters, ts_ctor);
    return sut.request(session, quote).then(() => {
      expect(retry_attempts_given).to.equal(retry_attempts_expected);
    });
  });

  it('determines max attempts for an indication step', () => {
    const stub_rate_data = {
      foo__unavailable: '0',
      bar__unavailable: '0',
      foo___retry: '1',
      bar___retry: '1',
      __unavailable_all: '0',
    };

    const {dao, logger, quote, raters, session, program, ts_ctor} =
      getStubs(stub_rate_data);

    const retry_attempts_current = 34;

    let logged_called = false;

    // Set max attempts for indication step to be the max
    program.ratingPoll.maxAttempts = 100;
    program.ratingPoll.indicationMaxAttempts = 34;
    program.isIndicationQuoteStep = () => {
      return true;
    };

    logger.log = function (
      priority: number,
      format: string,
      _qid: QuoteId,
      _program_id: string,
      _step_id: string,
      message: string
    ) {
      if (typeof message === 'string') {
        expect(format).to.equal(
          "Rating timeout reached for program '%s' quote #%s step %s and supplier(s) %s"
        );
        expect(priority).to.equal(PriorityLog.PRIORITY_ERROR);
        logged_called = true;
      }

      return logger;
    };

    quote.getRetryAttempts = () => retry_attempts_current;

    const sut = new Sut(logger, dao, raters, ts_ctor);
    return sut.request(session, quote).then(() => {
      expect(logged_called).to.equal(true);
    });
  });

  it('determines max attempts for a non-indication step', () => {
    const stub_rate_data = {
      foo__unavailable: '0',
      bar__unavailable: '0',
      foo___retry: '1',
      bar___retry: '1',
      __unavailable_all: '0',
    };

    const {dao, logger, quote, raters, session, program, ts_ctor} =
      getStubs(stub_rate_data);

    const retry_attempts_current = 34;

    let logged_called = false;

    // Set max attempts for quote step to be the max
    program.ratingPoll.maxAttempts = 34;
    program.ratingPoll.indicationMaxAttempts = 100;
    program.isIndicationQuoteStep = () => {
      return false;
    };

    logger.log = function (
      priority: number,
      format: string,
      _qid: QuoteId,
      _program_id: string,
      _step_id: string,
      message: string
    ) {
      if (typeof message === 'string') {
        expect(format).to.equal(
          "Rating timeout reached for program '%s' quote #%s step %s and supplier(s) %s"
        );
        expect(priority).to.equal(PriorityLog.PRIORITY_ERROR);
        logged_called = true;
      }

      return logger;
    };

    quote.getRetryAttempts = () => retry_attempts_current;

    const sut = new Sut(logger, dao, raters, ts_ctor);
    return sut.request(session, quote).then(() => {
      expect(logged_called).to.equal(true);
    });
  });

  it('Ignore missing retry flag logic if no retry flags exist', () => {
    const {dao, logger, quote, raters, session, ts_ctor} = getStubs();

    // The stub rate data returned on rate has a total of 1 supplier and
    // we are specifying that there are no retry fields, so one must be
    // missing. Missing retry attempts should be treated as pending so we
    // expect the retry attempt counter to be incremented
    const retry_attempts_current = 24;
    const retry_attempts_expected = 0;

    let retry_attempts_given: number;

    quote.setRetryAttempts = (attempts: number) => {
      retry_attempts_given = attempts;

      return quote;
    };

    quote.getRetryAttempts = () => retry_attempts_current;

    const sut = new Sut(logger, dao, raters, ts_ctor);
    return sut.request(session, quote).then((_: RateResult) => {
      expect(retry_attempts_given).to.equal(retry_attempts_expected);
    });
  });

  it('Default to retry field count if __result_ids is undefined', () => {
    const rate_data = {};

    const {dao, logger, quote, raters, session, ts_ctor} = getStubs(rate_data);
    let retry_attempts_given: number;

    quote.setRetryAttempts = (attempts: number) => {
      retry_attempts_given = attempts;

      return quote;
    };

    const sut = new Sut(logger, dao, raters, ts_ctor);
    return sut.request(session, quote).then((_: RateResult) => {
      // Because there were no missing fields and zero true retries returned
      // we will expect the retry attempts to be cleared because there are no
      // more pending
      expect(retry_attempts_given).to.equal(0);
    });
  });

  (<
    [
      string,
      Record<string, any>,
      string[],
      string[],
      boolean[],
      number,
      number,
      boolean
    ][]
  >[
    [
      'delay action is returned when raters are pending',
      {
        'supplier-a___retry': [0],
        'supplier-b___retry': [1],
        'supplier-c___retry': [1],
        'supplier-d___retry': [0],
        'supplier-a__unavailable': [0],
        'supplier-b__unavailable': [0],
        'supplier-c__unavailable': [0],
        'supplier-d__unavailable': [0],
      },
      ['supplier-b', 'supplier-c'],
      ['supplier-a', 'supplier-b', 'supplier-c', 'supplier-d'],
      [true],
      0,
      0,
      true,
    ],
    [
      'delay action is not returned when all raters are completed',
      {
        'supplier-a___retry': [0],
        'supplier-b___retry': [0],
        'supplier-c___retry': [0],
        'supplier-d___retry': [[0]],
        'supplier-a__unavailable': [0],
        'supplier-b__unavailable': [0],
        'supplier-c__unavailable': [0],
        'supplier-d__unavailable': [0],
      },
      [],
      ['supplier-a', 'supplier-b', 'supplier-c', 'supplier-d'],
      [true],
      0,
      0,
      true,
    ],
    [
      'Undefined ratesteps defaults gracefully',
      {
        'supplier-a___retry': [0],
        'supplier-b___retry': [0],
        'supplier-c___retry': [1],
        'supplier-d___retry': [1],
        'supplier-a__unavailable': [0],
        'supplier-b__unavailable': [0],
        'supplier-c__unavailable': [0],
        'supplier-d__unavailable': [0],
      },
      ['supplier-c', 'supplier-d'],
      ['supplier-a', 'supplier-b', 'supplier-c', 'supplier-d'],
      undefined,
      0,
      2,
      false,
    ],
    [
      'Set __rate_pending to zero after max attempts are reached',
      {
        'supplier-a___retry': [0],
        'supplier-b___retry': [0],
        'supplier-c___retry': [1],
        'supplier-d___retry': [1],
        'supplier-a__unavailable': [0],
        'supplier-b__unavailable': [0],
        'supplier-c__unavailable': [0],
        'supplier-d__unavailable': [0],
      },
      [],
      ['supplier-a', 'supplier-b', 'supplier-c', 'supplier-d'],
      undefined,
      0,
      30,
      false,
    ],
  ]).forEach(
    ([
      label,
      supplier_data,
      expected_pending_suppliers,
      expected_total_suppliers,
      rate_steps,
      step_id,
      attempts,
      expected_save_meta,
    ]) => {
      it(label, () => {
        const {
          dao,
          logger,
          quote,
          raters,
          session,
          stub_rate_data,
          program,
          ts_ctor,
        } = getStubs(supplier_data);

        let meta_save_called = false;

        Object.assign(stub_rate_data, supplier_data);

        if (expected_save_meta) {
          dao.saveQuoteMeta = (_, data, __, ___) => {
            expect(data?.['liza_timestamp_rate_request']).to.deep.equal([
              ts_ctor(),
            ]);

            meta_save_called = true;

            return dao;
          };
        } else {
          meta_save_called = true;
        }

        program.rateSteps = rate_steps;
        quote.getProgram = () => {
          return program;
        };
        quote.getCurrentStepId = () => {
          return <PositiveInteger>step_id;
        };
        quote.getRetryAttempts = () => {
          return attempts;
        };

        const sut = new Sut(logger, dao, raters, ts_ctor);
        return sut
          .request(session, quote)
          .then((result: RateResult) => {
            expect(result['__rate_pending']).to.deep.equal(
              expected_pending_suppliers
            );
            expect(result['suppliers_pending_rate']).to.deep.equal(
              expected_pending_suppliers
            );
            expect(result['suppliers_rated']).to.deep.equal(
              expected_total_suppliers
            );
          })
          .then(() => expect(meta_save_called).to.be.true);
      });
    }
  );

  describe('protected API', () => {
    it('calls #postProcessRaterData after rating before save', done => {
      let processed = false;

      const {logger, raters, dao, session, quote, ts_ctor} = getStubs();

      dao.saveQuote = () => {
        expect(processed).to.equal(true);
        done();

        return dao;
      };

      const sut = new (class extends Sut {
        postProcessRaterData() {
          processed = true;
        }
      })(logger, dao, raters, ts_ctor);

      sut.request(session, quote, ['something']);
    });

    it('merges saved and returned ratedata', () => {
      const merged_rating_data = <RateResult>(<unknown>{
        whatever: ['doesnt matter'],
      });

      const {logger, raters, dao, session, quote, ts_ctor} = getStubs();

      quote.getRatingData = () => merged_rating_data;
      quote.getDataByName = () => ['1'];

      let saved_rates = false;

      dao.saveQuote = (
        quote: ServerSideQuote,
        _class_data: ClassificationResult,
        success: ServerDaoCallback,
        _failure: ServerDaoCallback,
        save_data: Record<string, any>,
        _push_data: Record<string, any>
      ) => {
        expect(save_data.ratedata).to.deep.equal(merged_rating_data);

        saved_rates = true;
        success(quote);

        return dao;
      };

      const sut = new Sut(logger, dao, raters, ts_ctor);

      return sut.request(session, quote).then((result: any) => {
        expect(result).to.deep.equal(merged_rating_data);
        expect(saved_rates).to.be.true;
      });
    });

    it('merges saved and returned ratedata in a single mongo call', () => {
      const merged_rating_data = <RateResult>(<unknown>{
        whatever: ['doesnt matter'],
      });

      const {logger, raters, dao, session, quote, ts_ctor} = getStubs();

      quote.getRatingData = () => merged_rating_data;
      quote.getDataByName = () => ['1'];

      let saved_rates = false;

      dao.saveQuote = (
        quote: ServerSideQuote,
        _class_data: ClassificationResult,
        success: ServerDaoCallback,
        _failure: ServerDaoCallback,
        save_data: Record<string, any>,
        _push_data: Record<string, any>
      ) => {
        expect(save_data).to.deep.equal({
          'data.__rate_pending': [],
          'data.__unavailable_all': '0',
          'data.bar__unavailable': '0',
          'data.foo__unavailable': '0',
          'data.suppliers_pending_rate': [],
          'data.suppliers_rated': ['foo', 'bar'],
          ratedata: merged_rating_data,
        });

        saved_rates = true;
        success(quote);

        return dao;
      };

      const sut = new Sut(logger, dao, raters, ts_ctor);

      return sut.request(session, quote, undefined).then((result: any) => {
        expect(result).to.deep.equal(merged_rating_data);
        expect(saved_rates).to.be.true;
      });
    });

    it('classes are stored as classData and are returned to the client', () => {
      const merged_rating_data = <RateResult>(<unknown>{
        whatever: ['doesnt matter'],
        foo___classes: ['these are stored separately'],
      });

      const expected_client_return = {
        whatever: ['doesnt matter'],
        foo___classes: ['these are stored separately'],
      };

      const expected_dao_save = {
        ratedata: {
          whatever: ['doesnt matter'],
          foo___classes: ['these are stored separately'],
        },
        'classData.foo': ['these are stored separately'],
        'data.__rate_pending': [],
        'data.__unavailable_all': '0',
        'data.bar__unavailable': '0',
        'data.foo__unavailable': '0',
        'data.suppliers_pending_rate': [],
        'data.suppliers_rated': ['foo', 'bar'],
      };

      const {logger, raters, dao, session, quote, ts_ctor} = getStubs();

      quote.getRatingData = () => merged_rating_data;
      quote.getDataByName = () => ['1'];

      let saved_rates = false;

      dao.saveQuote = (
        quote: ServerSideQuote,
        _class_data: ClassificationResult,
        success: ServerDaoCallback,
        _failure: ServerDaoCallback,
        save_data: Record<string, any>,
        _push_data: Record<string, any>
      ) => {
        expect(save_data).to.deep.equal(expected_dao_save);
        saved_rates = true;
        success(quote);

        return dao;
      };

      const sut = new Sut(logger, dao, raters, ts_ctor);

      return sut.request(session, quote).then((result: any) => {
        expect(result).to.deep.equal(expected_client_return);
        expect(saved_rates).to.be.true;
      });
    });
  });
});

function getStubs(rate_data?: any) {
  const program_id = 'foo';

  const program = <Program>{
    getId: () => program_id,
    isIndicationQuoteStep: (_number: PositiveInteger) => false,
    ineligibleLockCount: 0,
    rateSteps: [true],
    ratingPoll: {},
  };

  // rate reply
  const stub_rate_data: RateResult = rate_data || {
    foo__unavailable: '0',
    bar__unavailable: '0',
    __unavailable_all: '0',
  };

  const stub_rate_delta: any = {
    'rdelta.ratedata': {
      data: {
        _unavailable_all: [undefined],
      },
      concluding_save: false,
      timestamp: 123,
    },
  };

  const createDelta = (_src: Kv, _dest: Kv) => {
    return stub_rate_delta['rdelta.ratedata']['data'];
  };

  const rater = <any>new (class {
    rate(
      _quote: ServerSideQuote,
      _session: UserSession,
      _indv: string,
      success: (data: RateResult, actions: ClientActions) => void,
      _failure: (message: string) => void
    ) {
      // force to be async so that the tests resemble how the code
      // actually runs
      process.nextTick(() => success(stub_rate_data, []));

      return this;
    }
  })();

  const raters = <ProcessManager>(<unknown>{
    byId: () => rater,
    init: () => {},
  });

  const logger = <PriorityLog>(<unknown>{log: () => logger});

  const dao = <ServerDao>(<unknown>new (class {
    saveQuote(
      quote: ServerSideQuote,
      _class_data: ClassificationResult,
      success: ServerDaoCallback,
      _failure: ServerDaoCallback,
      _save_data: Record<string, any>,
      _push_data: Record<string, any>
    ): this {
      success(quote);
      return this;
    }

    mergeBucket(_: any, __: any, cb: any): this {
      cb();
      return this;
    }

    saveQuoteClasses(): this {
      return this;
    }

    setWorksheets(): this {
      return this;
    }

    saveQuoteState(): this {
      return this;
    }

    unimportQuote(): this {
      return this;
    }

    saveQuoteRateRetries(): this {
      return this;
    }

    syncRatingState(quote: ServerSideQuote): Promise<ServerSideQuote> {
      return Promise.resolve(quote);
    }

    ensurePendingRate(_quote: ServerSideQuote): Promise<string[]> {
      return Promise.resolve([]);
    }

    saveQuoteMeta(_: any, __: any, ___: any, ____: any): this {
      return this;
    }

    mergeData(): this {
      return this;
    }

    saveQuoteLockState(): this {
      throw new Error('Unused method');
    }

    getWorksheet(): Promise<WorksheetData> {
      throw new Error('Unused method');
    }
  })());

  const session = <UserSession>{
    isInternal: () => false,
  };

  const quote = <ServerSideQuote>(<unknown>{
    getDataByName: () => ['0'],
    getProgramId: () => program_id,
    getProgram: () => program,
    getId: () => <QuoteId>0,
    setLastPremiumDate: () => quote,
    setRatedDate: () => quote,
    getRatedDate: () => <UnixTimestamp>0,
    getLastPremiumDate: () => <UnixTimestamp>0,
    getCurrentStepId: () => 0,
    setCurrentStepId: () => quote,
    setExplicitLock: () => quote,
    setRateBucket: () => quote,
    setRatingData: () => quote,
    getRatingData: () => stub_rate_data,
    getBucket: () => new QuoteDataBucket(),
    getMetabucket: () => new QuoteDataBucket(),
    getMetaUpdatedDate: () => <UnixTimestamp>0,
    getProgramVersion: () => 'Foo',
    getExplicitLockReason: () => 'Reason',
    getExplicitLockStep: () => <PositiveInteger>1,
    isImported: () => true,
    isBound: () => true,
    getTopVisitedStepId: () => <PositiveInteger>1,
    getTopSavedStepId: () => <PositiveInteger>1,
    setTopSavedStepId: () => quote,
    setRetryAttempts: () => quote,
    needsImport: () => quote,
    getRetryAttempts: () => 1,
    retryAttempted: () => quote,
    setMetadata: () => quote,
    getRetryCount: () => {
      return {field_count: 2, true_count: 0};
    },
    setInitialRatedDate: () => quote,
    getExpirationDate: () => 123,
    getFieldState: () => ({}),
  });

  const ts_ctor = () => {
    return <UnixTimestamp>2592001;
  };

  return {
    program: program,
    stub_rate_data: stub_rate_data,
    stub_rate_delta: stub_rate_delta,
    createDelta: createDelta,
    rater: rater,
    raters: raters,
    logger: logger,
    dao: dao,
    session: session,
    quote: quote,
    ts_ctor: ts_ctor,
  };
}
