/**
 * Tests WebhookService
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {StagingBucket} from '../../../src/bucket/StagingBucket';
import {DataApiResultItem} from '../../../src/dapi/DataApi';
import {DataApiManager} from '../../../src/dapi/DataApiManager';
import {PositiveInteger} from '../../../src/numeric';
import {ServerSideDapi} from '../../../src/program/Program';
import {Program} from '../../../src/program/Program';
import {DapiProcessor} from '../../../src/server/dapi/DapiProcessor';
import {PriorityLog} from '../../../src/server/log/PriorityLog';
import {FieldData} from '../../../src/server/meta/DapiMetaSource';
import {DapiMetaSource} from '../../../src/server/meta/DapiMetaSource';
import {TokenStore} from '../../../src/server/token/store/TokenStore';
import {WebhookService as Sut} from '../../../src/server/service/WebhookService';
import {ServerSideQuote} from '../../../src/server/quote/ServerSideQuote';
import {UserRequest} from '../../../src/server/request/UserRequest';
import {
  Token,
  TokenId,
  TokenNamespace,
  TokenState,
} from '../../../src/server/token/Token';
import {WebhookBucketDataApi} from '../../../src/dapi/WebhookBucketDataApi';

// eslint-disable-next-line node/no-unpublished-import
import {expect} from 'chai';

describe('WebhookService', () => {
  it('Fulfills a successful webhook request', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = {foo: 'bar'};
    const field = <TokenNamespace>'field_name_0';
    const token_id = <TokenId>'123token123';
    const token = {data: 'some token', last_created: true};
    const dapi_def = <ServerSideDapi>{name: 'field name'};

    let token_looked_up = false;
    let token_completed = false;
    let token_accepted = false;

    token_store.lookupToken = (token_given: TokenId) => {
      expect(token_given).to.deep.equal(token_id);

      token_looked_up = true;

      return Promise.resolve(<Token<TokenState>>token);
    };

    token_store.completeToken = (
      token_given: Token<TokenState>,
      data_given: string
    ) => {
      expect(token_given).to.deep.equal(token);
      expect(data_given).to.deep.equal(JSON.stringify([data]));

      token_completed = true;

      return Promise.resolve(<Token<TokenState.DONE>>token);
    };

    token_store.acceptToken = (
      token_given: Token<TokenState>,
      data_given: string
    ) => {
      expect(token_given).to.deep.equal(token);
      expect(data_given).to.deep.equal(JSON.stringify([data]));

      token_accepted = true;

      return Promise.resolve(<Token<TokenState.ACCEPTED>>token);
    };

    dapi_meta_source.mapDapiResponse = (
      name: string,
      index: PositiveInteger,
      dapi_def_given: ServerSideDapi,
      _bucket: StagingBucket,
      _dapi_manger: DataApiManager,
      data_given: DataApiResultItem[],
      success_cb: (mapped_data: FieldData) => void,
      _failure_cb: (_reason?: string | Error) => void
    ) => {
      expect(name).to.equal('field_name');
      expect(index).to.equal(0);
      expect(data_given).to.deep.equal([data]);
      expect(dapi_def_given).to.deep.equal(dapi_def);

      success_cb(<FieldData>(<unknown>data));
    };

    program.meta.fields = {
      field_name: {
        dapi: dapi_def,
        desc: 'some meta field',
      },
    };

    return sut
      .fulfillWebhook(program, field, token_id, data, quote, request)
      .then(() => {
        expect(token_looked_up).to.be.true;
        expect(token_completed).to.be.true;
        expect(token_accepted).to.be.true;
      });
  });

  it('Clears tokens in bucket', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = {foo: 'bar'};
    const field = <TokenNamespace>'field_name_0';
    const token_id = <TokenId>'123token123';

    dapi_meta_source.mapDapiResponse = (
      _name: string,
      _index: PositiveInteger,
      _dapi_def_given: ServerSideDapi,
      _bucket: StagingBucket,
      _dapi_manger: DataApiManager,
      _data_given: DataApiResultItem[],
      success_cb: (mapped_data: FieldData) => void,
      _failure_cb: (_reason?: string | Error) => void
    ) => {
      success_cb(<FieldData>(<unknown>data));
    };

    let monitorMetadataPromise_called = false;
    let setValues_called = false;

    dapi_processor.monitorMetadataPromise = (
      _: any,
      __: any,
      data_given: any
    ) => {
      expect(data_given).to.deep.equal({
        'data.webhook_trigger_field_name_0': [''],
        [`data.${WebhookBucketDataApi.WEBHOOK_AGG_KEY}`]: ['preserved_key'],
      });

      monitorMetadataPromise_called = true;
    };

    program.meta.fields = {
      field_name: {
        dapi: <ServerSideDapi>{name: 'field name'},
        desc: 'some meta field',
      },
    };

    quote.getDataByName = () => [
      'preserved_key',
      'webhook_trigger_field_name_0',
    ];

    bucket.setValues = (values: any) => {
      expect(values).to.deep.equal({
        webhook_trigger_field_name_0: [''],
        [WebhookBucketDataApi.WEBHOOK_AGG_KEY]: ['preserved_key'],
      });

      setValues_called = true;

      return bucket;
    };

    return sut
      .fulfillWebhook(program, field, token_id, data, quote, request)
      .then(() => {
        expect(monitorMetadataPromise_called).to.be.true;
        expect(setValues_called).to.be.true;
      });
  });

  it('Throws an error on an invalid token', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = {foo: 'bar'};
    const field = <TokenNamespace>'field_name_0';
    const token_id = <TokenId>'123token123';
    const token = {data: 'some token', last_created: false};
    const dapi_def = <ServerSideDapi>{name: 'field name'};

    token_store.lookupToken = (token_given: TokenId) => {
      expect(token_given).to.deep.equal(token_id);

      return Promise.resolve(<Token<TokenState>>token);
    };

    program.meta.fields = {
      field_name: {
        dapi: dapi_def,
        desc: 'some meta field',
      },
    };

    let error_thrown = false;

    try {
      await sut.fulfillWebhook(program, field, token_id, data, quote, request);
    } catch (e) {
      if (e instanceof Error) {
        expect(e.message).to.equal('The token is not valid');

        error_thrown = true;
      }
    }

    expect(error_thrown).to.be.true;
  });

  it('Throws an error on an invalid field requested', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = {foo: 'bar'};
    const field = <TokenNamespace>'field_name';
    const token_id = <TokenId>'123token123';
    const token = {data: 'some token', last_created: false};

    token_store.lookupToken = (token_given: TokenId) => {
      expect(token_given).to.deep.equal(token_id);

      return Promise.resolve(<Token<TokenState>>token);
    };

    let error_thrown = false;

    try {
      await sut.fulfillWebhook(program, field, token_id, data, quote, request);
    } catch (e) {
      if (e instanceof Error) {
        expect(e.message).to.equal('Could not parse field: bad input');

        error_thrown = true;
      }
    }

    expect(error_thrown).to.be.true;
  });

  it('Throws an error on an invalid dapi requested', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = {foo: 'bar'};
    const field = <TokenNamespace>'bad_field_name_0';
    const token_id = <TokenId>'123token123';
    const token = {data: 'some token', last_created: false};
    const dapi_def = <ServerSideDapi>{name: 'field name'};

    token_store.lookupToken = (token_given: TokenId) => {
      expect(token_given).to.deep.equal(token_id);

      return Promise.resolve(<Token<TokenState>>token);
    };

    program.meta.fields = {
      field_name: {
        dapi: dapi_def,
        desc: 'some meta field',
      },
    };

    let error_thrown = false;

    try {
      await sut.fulfillWebhook(program, field, token_id, data, quote, request);
    } catch (e) {
      if (e instanceof Error) {
        expect(e.message).to.equal('Unknown Data API');

        error_thrown = true;
      }
    }

    expect(error_thrown).to.be.true;
  });

  it('Throws an error on an invalid payload', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = 'string instead of object';
    const field = <TokenNamespace>'field_name_0';
    const token_id = <TokenId>'123token123';
    const token = {data: 'some token', last_created: false};
    const dapi_def = <ServerSideDapi>{name: 'field name'};

    token_store.lookupToken = (token_given: TokenId) => {
      expect(token_given).to.deep.equal(token_id);

      return Promise.resolve(<Token<TokenState>>token);
    };

    program.meta.fields = {
      field_name: {
        dapi: dapi_def,
        desc: 'some meta field',
      },
    };

    let error_thrown = false;

    try {
      await sut.fulfillWebhook(program, field, token_id, data, quote, request);
    } catch (e) {
      if (e instanceof Error) {
        expect(e.message).to.equal('Invalid payload sent');

        error_thrown = true;
      }
    }

    expect(error_thrown).to.be.true;
  });

  it('Throws an error on an invalid payload', async () => {
    const {
      logger,
      token_store,
      dapi_manager,
      dapi_meta_source,
      bucket,
      dapi_processor,
      program,
      quote,
      request,
    } = getMocks();

    const sut = new Sut(
      logger,
      () => token_store,
      () => bucket,
      () => dapi_manager,
      dapi_meta_source,
      dapi_processor
    );

    const data = [1, 2, 3];
    const field = <TokenNamespace>'field_name_0';
    const token_id = <TokenId>'123token123';
    const token = {data: 'some token', last_created: false};
    const dapi_def = <ServerSideDapi>{name: 'field name'};

    token_store.lookupToken = (token_given: TokenId) => {
      expect(token_given).to.deep.equal(token_id);

      return Promise.resolve(<Token<TokenState>>token);
    };

    program.meta.fields = {
      field_name: {
        dapi: dapi_def,
        desc: 'some meta field',
      },
    };

    let error_thrown = false;

    try {
      await sut.fulfillWebhook(program, field, token_id, data, quote, request);
    } catch (e) {
      if (e instanceof Error) {
        expect(e.message).to.equal('Invalid payload sent: Array not expected');

        error_thrown = true;
      }
    }

    expect(error_thrown).to.be.true;
  });
});

function getMocks() {
  const program = createMockProgram();
  const logger = createMockLogger();
  const token_store = createMockTokenStore();
  const dapi_manager = createMockDataApiManager();
  const dapi_meta_source = createMockDapiMetaSource();
  const bucket = createMockBucket();
  const dapi_processor = createMockDapiProcessor();
  const quote = createMockQuote();
  const request = createMockRequest();

  return {
    logger,
    token_store,
    dapi_manager,
    dapi_meta_source,
    bucket,
    dapi_processor,
    program,
    quote,
    request,
  };
}

function createMockTokenStore(): TokenStore {
  return <TokenStore>(<unknown>{
    lookupToken: () => Promise.resolve({last_created: true}),
    completeToken: () => Promise.resolve({}),
    acceptToken: () => Promise.resolve({}),
    isDoneable: () => true,
  });
}

function createMockQuote(): ServerSideQuote {
  return <ServerSideQuote>(<unknown>{
    getId: () => 123,
    getDataByName: () => ['123'],
  });
}

function createMockRequest(): UserRequest {
  return <UserRequest>{};
}

function createMockBucket(): StagingBucket {
  const bucket = <StagingBucket>(<unknown>{
    setValues: () => bucket,
  });

  return bucket;
}

function createMockDapiProcessor(): DapiProcessor {
  return <DapiProcessor>(<unknown>{
    monitorMetadataPromise: () => {},
  });
}

function createMockProgram(
  overrides: {
    first_step_id?: number;
    program_ver?: string;
    next_visible_step_id?: number;
    group_exclusive_fields?: Record<string, any>;
    classes?: Record<string, {is: boolean}>;
  } = {}
): Program {
  return <Program>(<unknown>{
    getId: () => 'foo',
    version: overrides?.program_ver || 'foo',
    getFirstStepId: () => overrides?.first_step_id || 1,
    isStepVisible: () => true,
    isLastStep: () => false,
    isManageQuoteStep: () => false,
    isIndicationQuoteStep: () => false,
    // eslint-disable-next-line no-sparse-arrays
    steps: [
      ,
      {id: 'Manage Quote', type: 'manage'},
      {id: 'Step1', type: ''},
      {id: 'Step2', type: ''},
      {id: 'Step3', type: ''},
      {id: 'Step4', type: ''},
      {id: 'Step5', type: ''},
      {id: 'Step6', type: ''},
    ],
    processNaFields: () => {},
    getNextVisibleStep: () => overrides?.next_visible_step_id || 2,
    classify: () => {
      return overrides?.classes || {};
    },
    secureFields: [],
    groupExclusiveFields: overrides?.group_exclusive_fields || {},
    beforeLoadStep: () => {},
    meta: {
      groups: {},
      fields: {},
    },
  });
}

function createMockLogger(cb = (_level: any, _msg: any) => {}): PriorityLog {
  const log_cb = (level: any, msg: any) => {
    cb(level, msg);
  };

  return <PriorityLog>(<unknown>{
    PRIORITY_ERROR: 0,
    PRIORITY_IMPORTANT: 1,
    PRIORITY_DB: 2,
    PRIORITY_INFO: 3,
    PRIORITY_SOCKET: 4,
    debug: log_cb,
    info: log_cb,
    notice: log_cb,
    warning: log_cb,
    error: log_cb,
    critical: log_cb,
    alert: log_cb,
    emergency: log_cb,
    log: log_cb,
  });
}

function createMockDataApiManager() {
  const dapi_manager = <DataApiManager>(<unknown>{
    on: () => dapi_manager,
  });

  return dapi_manager;
}

function createMockDapiMetaSource(): DapiMetaSource {
  return <DapiMetaSource>(<unknown>{
    getFieldData(
      _field: string,
      _index: PositiveInteger,
      _dapi_manager: any,
      _dapi: any,
      _data: Record<string, any>
    ) {
      return new Promise<any>(() => {});
    },

    mapDapiResponse(
      _name: string,
      _index: string,
      _dapi_definition: ServerSideDapi,
      _bucket: StagingBucket,
      _dapi_manger: DataApiManager,
      _data: DataApiResultItem[],
      _success_cb: (mapped_data: FieldData) => void,
      _failure_cb: () => void
    ) {},
  });
}
