/**
 * Tests AutoObjectStore
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {AutoObjectStore as Sut} from '../../src/store/AutoObjectStore';
import {MemoryStore as Store} from '../../src/store/MemoryStore';

const chai = require('chai');
const expect = chai.expect;

chai.use(require('chai-as-promised'));

describe('AutoObjectStore', () => {
  describe('given an object value', () => {
    it('applies given ctor to objects', () => {
      const obj = new Store();

      const sut = new (Sut(Store, () => obj))();

      const foo = sut.add('foo', {}).then(_ => sut.get('foo'));

      return expect(foo).to.eventually.deep.equal(obj);
    });

    it('adds object values to new store', () => {
      const obj = {bar: 'baz'};

      const sut = new (Sut(Store, () => new Store()))();

      const bar = sut
        .add('foo', obj)
        .then(_ => sut.get('foo'))
        .then(substore => (<Store>substore).get('bar'));

      return expect(bar).to.eventually.equal(obj.bar);
    });

    it('caches sub-store until key changes', () => {
      const sut = new (Sut(Store, () => new Store()))();

      return sut
        .add('foo', {})
        .then(_ => sut.get('foo'))
        .then(store1 =>
          expect(sut.get('foo'))
            .to.eventually.equal(store1)
            .then((_: Store) => sut.add('foo', 'new'))
            .then((_: Store) => sut.get('foo'))
            .then((store2: Store) => expect(store2).to.not.equal(store1))
        );
    });
  });

  it('leaves non-objects untouched', () => {
    const expected = 'bar';
    const sut = new (Sut(Store, () => new Store()))();

    const foo = sut.add('foo', expected).then((_: Store) => sut.get('foo'));

    return expect(foo).to.eventually.equal(expected);
  });

  // includes class instances, since easejs generates prototypes
  it('leaves prototype instances untouched', () => {
    const expected = new (class {})();
    const sut = new (Sut(Store, () => new Store()))();

    const foo = sut.add('foo', expected).then((_: Store) => sut.get('foo'));

    return expect(foo).to.eventually.equal(expected);
  });
});
