/**
 * Test case for Cascading store
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Cascading as Sut} from '../../src/store/Cascading';
import {MemoryStore as Store} from '../../src/store/MemoryStore';

const chai = require('chai'),
  expect = chai.expect;

chai.use(require('chai-as-promised'));

describe('store.Cascading', () => {
  describe('#add', () => {
    it('does not allow attaching non-store objects', () => {
      expect(new (Sut(Store))().add('invalid', {})).to.be.rejectedWith(
        TypeError
      );
    });

    it('allows attaching Store objects', () => {
      return new (Sut(Store))().add('valid', new Store());
    });
  });

  describe('#clear', () => {
    it('invokes #clear on all contained stores', () => {
      const cleared: Store[] = [];

      const MockStore = class extends Store {
        clear() {
          cleared.push(this);

          return Promise.resolve(this);
        }
      };

      const stores = [1, 2, 3].map(() => new MockStore());
      const sut = new (Sut(Store))();

      stores.forEach((store, i) => sut.add(i + '', store));

      // should trigger clear on all stores
      return sut.clear().then(() => {
        expect(stores.every(store => cleared.some(item => item === store))).to
          .be.true;
      });
    });

    it('does not clear self', () => {
      const sut = new (Sut(Store))();
      const substore = new Store();

      return sut
        .add('foo', substore)
        .then(() => sut.clear())
        .then(() => {
          return expect(sut.get('foo')).to.eventually.equal(substore);
        });
    });
  });
});
