/**
 * Test case for MemoryStore
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {StoreMissError} from '../../src/store/StoreMissError';
import {MemoryStore as Sut} from '../../src/store/MemoryStore';

const chai_required = require('chai');
const expect = chai_required.expect;
chai_required.use(require('chai-as-promised'));

describe('store.MemoryStore', () => {
  describe('#add', () => {
    it('adds item to store when missing', () => {
      const sut = new Sut();
      const item = {};

      return expect(
        sut.add('foo', item).then(() => sut.get('foo'))
      ).to.eventually.equal(item);
    });

    it('replaces item in store if existing', () => {
      const sut = new Sut();
      const item = {};

      return expect(
        sut
          .add('foo', [])
          .then(() => sut.add('foo', item))
          .then(() => sut.get('foo'))
      ).to.eventually.equal(item);
    });

    it('returns self with promise', () => {
      const sut = new Sut();

      return expect(sut.add('foo', 'bar')).to.eventually.equal(sut);
    });
  });

  describe('#populate', () => {
    it("#add's each element of object to store", () => {
      const obj = {foo: {}, bar: {}};
      const sut = new Sut();

      return sut.populate(obj).then(_ => {
        // by reference
        expect(sut.get('foo')).to.eventually.equal(obj.foo);
        expect(sut.get('bar')).to.eventually.equal(obj.bar);
      });
    });

    it('fails if any add fails', () => {
      const e = Error('ok');

      const ErroringSut = class extends Sut {
        add(_k: string, _v: unknown) {
          return Promise.reject(e);
        }
      };

      const sut = new ErroringSut();

      return expect(sut.populate({a: 1})).to.eventually.be.rejectedWith(e);
    });
  });

  // most things implicitly tested above
  describe('#get', () => {
    it('rejects promise if store item does not exist', () => {
      return expect(
        new Sut().get('unknown')
      ).to.eventually.be.rejected.and.be.instanceof(StoreMissError);
    });
  });

  describe('#clear', () => {
    it('removes all items from store', () => {
      const sut = new Sut();
      const keys = ['foo', 'bar', 'baz'];

      keys.forEach(key => sut.add(key, 'foo'));

      // should remove all items
      return sut.clear().then(() => {
        return Promise.all(
          keys.map(key => {
            expect(sut.get(key)).to.eventually.be.rejected;
          })
        );
      });
    });

    it('returns self with promise', () => {
      const sut = new Sut();

      return expect(sut.clear()).to.eventually.equal(sut);
    });
  });

  describe('#reduce', () => {
    it('folds each stored item', () => {
      const vals: Record<string, number> = {
        one: 1,
        two: 2,
        three: 3,
      };

      const SummingSut = class extends Sut {
        sum() {
          return this.reduce((accum, item, key) => {
            // correct key for item?
            expect(item).to.equal(vals[key]);

            return <number>accum + <number>item;
          }, 5);
        }
      };

      const sut = new SummingSut();

      Object.keys(vals).forEach((key, _i) => sut.add(key, vals[key]));

      // implicitly tests initial
      return expect(sut.sum()).to.eventually.equal(11);
    });
  });
});
