/**
 * Delta Processor test
 *
 *  Copyright (C) 2010-2019 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {DeltaProcessor as Sut} from '../../src/system/DeltaProcessor';
import {AmqpPublisher} from '../../src/system/AmqpPublisher';
import {DeltaDao} from '../../src/system/db/DeltaDao';
import {DeltaDocument, applyDelta, mergeDeltas} from '../../src/bucket/delta';
import {DocumentId} from '../../src/document/Document';
import {PositiveInteger} from '../../src/numeric';
import {EventEmitter} from 'events';

import {expect, use as chai_use} from 'chai';
chai_use(require('chai-as-promised'));

type TestInput = {
  label: string;
  given: any;
  expected: any;
};

describe('system.DeltaProcessor', () => {
  describe('#process', () => {
    (<TestInput[]>[
      {
        label: 'No deltas are processed',
        given: {
          id: 123,
          lastUpdate: 123123123,
          data: {},
          ratedata: {},
          rdelta: {},
          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
            liza_timestamp_owner_vis_updated: [987654321],
            bound_date_ts: 123456789,
            bound_by_username: 'bound user',
            unlocked_by_username: undefined,
            unlocked_date_ts: undefined,
            sessionKeys: {
              foo: ['bar', 'baz'],
            },
          },
        },
        expected: [],
      },

      // when quote is initialized: { foo: [ '' ], state: [ 'a' ] }
      {
        label: 'Publishes deltas in order',

        given: {
          id: 123,
          lastUpdate: 123123123,
          agentEntityId: 12345,

          data: {
            bar: ['initial and unchanged'],
            foo: ['third'],
            state: ['a', 'b', 'c', 'd'],
          },

          ratedata: {
            prem: ['rate_second'],
            state: ['i', 'ii', 'iii'],
          },

          sessionKeys: {
            foo: ['bar', 'baz'],
          },

          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
            liza_timestamp_owner_vis_updated: [987654321],
            bound_date_ts: 123456789,
            bound_by_username: 'bound user',
            sessionKeys: {
              foo: ['bar', 'baz'],
            },
          },

          rdelta: {
            data: [
              {
                timestamp: 1,
                data: {
                  foo: [''],
                  state: [undefined, null],
                },
              },
              {
                timestamp: 3,
                data: {
                  foo: ['first'],
                  state: [undefined, undefined, null],
                },
              },
              {
                timestamp: 5,
                data: {
                  foo: ['second'],
                  state: [undefined, undefined, undefined, null],
                },
              },
            ],

            ratedata: [
              {
                timestamp: 2,
                data: {
                  prem: [''],
                  state: [undefined, null],
                },
              },
              {
                timestamp: 4,
                data: {
                  prem: ['rate_first'],
                  state: [undefined, undefined, null],
                },
              },
            ],
          },
        },

        expected: [
          // bucket
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 123,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: undefined,
              liza_timestamp_owner_vis_updated: [987654321],
              bound_date_ts: 123456789,
              bound_by_username: 'bound user',
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {
              foo: [''],
              state: [undefined, null],
            },
            bucket: {
              bar: ['initial and unchanged'],
              foo: ['first'],
              state: ['a', 'b'],
            },
            ratedata: {
              prem: [''],
              state: ['i'],
            },
          },

          // rate
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 123,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: undefined,
              liza_timestamp_owner_vis_updated: [987654321],
              bound_date_ts: 123456789,
              bound_by_username: 'bound user',
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {
              prem: [''],
              state: [undefined, null],
            },
            bucket: {
              bar: ['initial and unchanged'],
              foo: ['first'],
              state: ['a', 'b'],
            },
            ratedata: {
              prem: ['rate_first'],
              state: ['i', 'ii'],
            },
          },

          // bucket
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 123,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: undefined,
              liza_timestamp_owner_vis_updated: [987654321],
              bound_date_ts: 123456789,
              bound_by_username: 'bound user',
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {
              foo: ['first'],
              state: [undefined, undefined, null],
            },
            bucket: {
              bar: ['initial and unchanged'],
              foo: ['second'],
              state: ['a', 'b', 'c'],
            },
            ratedata: {
              prem: ['rate_first'],
              state: ['i', 'ii'],
            },
          },

          // rate
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 123,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: undefined,
              liza_timestamp_owner_vis_updated: [987654321],
              bound_date_ts: 123456789,
              bound_by_username: 'bound user',
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {
              prem: ['rate_first'],
              state: [undefined, undefined, null],
            },
            bucket: {
              bar: ['initial and unchanged'],
              foo: ['second'],
              state: ['a', 'b', 'c'],
            },
            ratedata: {
              prem: ['rate_second'],
              state: ['i', 'ii', 'iii'],
            },
          },

          // bucket
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 123,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: undefined,
              liza_timestamp_owner_vis_updated: [987654321],
              bound_date_ts: 123456789,
              bound_by_username: 'bound user',
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {
              foo: ['second'],
              state: [undefined, undefined, undefined, null],
            },
            bucket: {
              bar: ['initial and unchanged'],
              foo: ['third'],
              state: ['a', 'b', 'c', 'd'],
            },
            ratedata: {
              prem: ['rate_second'],
              state: ['i', 'ii', 'iii'],
            },
          },
        ],
      },
      {
        label: 'trims delta array based on index',
        given: {
          id: 111,
          lastUpdate: 123123123,
          topSavedStepId: 2,
          agentEntityId: 12345,
          deltaPublishedTs: {data: <UnixTimestamp>0},
          deltaPublishedIndex: {data: 0},
          data: {foo: ['second']},
          ratedata: {},
          sessionKeys: {
            foo: ['bar', 'baz'],
          },
          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
          },
          rdelta: {
            data: [
              {
                data: {foo: ['']},
                timestamp: 123,
              },
              {
                data: {foo: ['first']},
                timestamp: 234,
              },
            ],
          },
          totalPublishDelta: {
            data: 1,
          },
        },
        expected: [
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 111,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: 2,
              liza_timestamp_owner_vis_updated: undefined,
              bound_date_ts: undefined,
              bound_by_username: undefined,
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {foo: ['first']},
            bucket: {foo: ['second']},
            ratedata: {},
          },
        ],
      },
      {
        label:
          'Data deltas with step_id higher than topSavedStepId ' +
          'are not published',
        given: {
          id: 111,
          lastUpdate: 123123123,
          agentEntityId: 12345,
          deltaPublishedTs: {data: <UnixTimestamp>0},
          topSavedStepId: 2,
          data: {foo: ['second']},
          ratedata: {},
          sessionKeys: {
            foo: ['bar', 'baz'],
          },
          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
          },
          rdelta: {
            data: [
              {
                data: {foo: ['']},
                step_id: 2,
                timestamp: 123,
              },
              {
                data: {foo: ['first']},
                step_id: 3,
                timestamp: 234,
              },
            ],
          },
          totalPublishDelta: {
            data: 1,
          },
        },

        expected: [
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 111,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: 2,
              liza_timestamp_owner_vis_updated: undefined,
              bound_date_ts: undefined,
              bound_by_username: undefined,
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {foo: ['']},
            bucket: {foo: ['first']},
            ratedata: {},
          },
        ],
      },

      {
        label:
          'Ratedata deltas with step_id higher than topSavedStepId ' +
          'are published',
        given: {
          id: 111,
          agentEntityId: 12345,
          lastUpdate: 123123123,
          deltaPublishedTs: {data: <UnixTimestamp>0},
          topSavedStepId: 2,
          data: {foo: ['current']},
          ratedata: {bar: ['current']},
          sessionKeys: {
            foo: ['bar', 'baz'],
          },
          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
          },
          rdelta: {
            data: [
              {
                data: {foo: ['']},
                step_id: 2,
                timestamp: 123,
              },
              {
                data: {foo: ['data_update_1']},
                step_id: 3,
                timestamp: 234,
              },
              {
                data: {foo: ['data_update_2']},
                step_id: 3,
                timestamp: 456,
              },
            ],
            ratedata: [
              {
                data: {bar: ['ratedata_update']},
                step_id: 3,
                timestamp: 345,
              },
            ],
          },
        },
        expected: [
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 111,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: 2,
              liza_timestamp_owner_vis_updated: undefined,
              bound_date_ts: undefined,
              bound_by_username: undefined,
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {foo: ['']},
            bucket: {foo: ['data_update_1']},
            ratedata: {bar: ['ratedata_update']},
          },
          {
            meta: {
              created_by_username: 'foo@bar.com',
              created_by_roles: ['EMPLOYEE'],
              entity_id: 12345,
              entity_name: undefined,
              expDate: undefined,
              id: 111,
              lastUpdate: 123123123,
              last_updated_by_username: 'bar@bar.com',
              last_updated_by_roles: ['EMPLOYEE'],
              liza_access_groups: undefined,
              liza_access_email: undefined,
              program: undefined,
              quoteSetId: undefined,
              retail_agency: 'AGT123',
              startDate: undefined,
              topSavedStepId: 2,
              liza_timestamp_owner_vis_updated: undefined,
              bound_date_ts: undefined,
              bound_by_username: undefined,
              unlocked_by_username: undefined,
              unlocked_date_ts: undefined,
              sessionKeys: {
                foo: ['bar', 'baz'],
              },
            },
            rdelta: {bar: ['ratedata_update']},
            bucket: {foo: ['data_update_2']},
            ratedata: {bar: ['current']},
          },
        ],
      },
    ]).forEach(({label, given, expected}) =>
      it(label, () => {
        const published: any = [];
        const dao = createMockDeltaDao();
        const publisher = createMockDeltaPublisher();
        const emitter = new EventEmitter();
        const mergeDeltasByType = (deltas: any) => deltas;

        dao.getUnprocessedDocument = (): Promise<DeltaDocument> => {
          return Promise.resolve(given);
        };

        publisher.publish = (meta, delta, bucket, ratedata): Promise<void> => {
          published.push({
            meta: meta,
            rdelta: delta.data,
            bucket: bucket,
            ratedata: ratedata,
          });

          return Promise.resolve();
        };

        const sut = new Sut(
          dao,
          publisher,
          emitter,
          applyDelta,
          mergeDeltasByType
        );

        return expect(sut.process())
          .to.eventually.deep.equal(undefined)
          .then(_ => expect(published).to.deep.equal(expected));
      })
    );

    (<TestInput[]>[
      {
        label: 'handles empty delta data',
        given: {
          id: 111,
          lastUpdate: 123123123,
          agentEntityId: 12345,
          data: {foo: ['bar']},
          ratedata: {},
          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
          },
          rdelta: {
            data: [
              {
                data: {},
                timestamp: 234,
                concluding_save: true,
              },
            ],
          },
        },
        expected: [
          {
            doc_id: 111,
            bucket: {foo: ['bar']},
            ratedata: {},
            rdelta: {
              concluding_save: true,
              data: {},
              index: 0,
              timestamp: 234,
              type: 'data',
            },
          },
        ],
      },
    ]).forEach(({label, given, expected}) =>
      it(label, () => {
        const published: any = [];
        const dao = createMockDeltaDao();
        const publisher = createMockDeltaPublisher();
        const emitter = new EventEmitter();
        const applyDelta = (bucket: any, _: any) => bucket;
        const mergeDeltasByType = (deltas: any) => deltas;

        dao.getUnprocessedDocument = (): Promise<DeltaDocument> => {
          return Promise.resolve(given);
        };

        publisher.publish = (meta, delta, bucket, ratedata): Promise<void> => {
          published.push({
            doc_id: meta.id,
            rdelta: delta,
            bucket: bucket,
            ratedata: ratedata,
          });

          return Promise.resolve();
        };

        const sut = new Sut(
          dao,
          publisher,
          emitter,
          applyDelta,
          mergeDeltasByType
        );

        return expect(sut.process())
          .to.eventually.deep.equal(undefined)
          .then(_ => expect(published).to.deep.equal(expected));
      })
    );

    (<TestInput[]>[
      {
        label: 'testing',
        given: {
          id: 111,
          agentEntityId: 12345,
          lastUpdate: 123123123,
          deltaPublishedTs: {data: <UnixTimestamp>0},
          topSavedStepId: 28,
          data: {foo: ['current']},
          ratedata: {bar: ['current']},
          sessionKeys: {
            foo: ['bar', 'baz'],
          },
          meta: {
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
          },
          rdelta: {
            data: [
              {
                data: {
                  agency_contact_id: [''],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420345,
                step_id: 27,
              },
              {
                data: {
                  agency_contact_name: [''],
                  agency_contact_email: [''],
                  agency_contact_phone: [''],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420356,
                step_id: 27,
              },
              {
                data: {
                  moratorium_returned: [''],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420357,
                step_id: 27,
              },
              {
                data: {
                  moratorium_returned: ['0'],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420358,
                step_id: 28,
              },
              {
                data: {},
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420359,
                step_id: 28,
              },
              {
                data: {
                  min_premium_owed: ['0'],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420361,
                step_id: 28,
              },
              {
                data: {
                  digital_signatures_required: ['0'],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420361,
                step_id: 28,
              },
              {
                data: {
                  terms_and_conditions: [''],
                },
                concluding_save: false,
                is_indication: false,
                timestamp: 1635420363,
                step_id: 28,
              },
              {
                data: {},
                concluding_save: true,
                is_indication: false,
                timestamp: 1635420364,
                step_id: 28,
              },
            ],
            ratedata: [],
          },
        },
        expected: [
          {
            doc_id: 111,
            rdelta: {
              type: 'data',
              timestamp: 1635420357,
              index: 2,
              concluding_save: false,
              is_indication: false,
              step_id: 27,
              data: {
                agency_contact_email: [''],
                agency_contact_id: [''],
                agency_contact_name: [''],
                agency_contact_phone: [''],
                moratorium_returned: [''],
              },
            },
            bucket: {
              foo: ['current'],
              moratorium_returned: ['0'],
              min_premium_owed: ['0'],
              digital_signatures_required: ['0'],
              terms_and_conditions: [''],
            },
            ratedata: {bar: ['current']},
          },
          {
            doc_id: 111,
            rdelta: {
              type: 'data',
              timestamp: 1635420364,
              index: 8,
              concluding_save: true,
              is_indication: false,
              step_id: 28,
              data: {
                digital_signatures_required: ['0'],
                min_premium_owed: ['0'],
                moratorium_returned: ['0'],
                terms_and_conditions: [''],
              },
            },
            bucket: {foo: ['current']},
            ratedata: {bar: ['current']},
          },
        ],
      },
    ]).forEach(({label, given, expected}) =>
      it(label, () => {
        const published: any = [];
        const dao = createMockDeltaDao();
        const publisher = createMockDeltaPublisher();
        const emitter = new EventEmitter();

        let published_called = false;

        dao.getUnprocessedDocument = (): Promise<DeltaDocument> => {
          return Promise.resolve(given);
        };

        dao.markDocumentAsDonePublishing = (_: any) => {
          published_called = true;

          return Promise.resolve();
        };

        publisher.publish = (meta, delta, bucket, ratedata): Promise<void> => {
          published.push({
            doc_id: meta.id,
            rdelta: delta,
            bucket: bucket,
            ratedata: ratedata,
          });

          return Promise.resolve();
        };

        const sut = new Sut(dao, publisher, emitter, applyDelta, mergeDeltas);

        return expect(sut.process())
          .to.eventually.deep.equal(undefined)
          .then(_ => expect(published).to.deep.equal(expected))
          .then(_ => expect(published_called).to.be.true);
      })
    );
  });

  describe('Error handling', () => {
    it('Marks document in error state and continues', () => {
      const published: any = [];
      let error_flag_set = false;
      const dao = createMockDeltaDao();
      const publisher = createMockDeltaPublisher();
      const emitter = new EventEmitter();
      const applyDelta = (bucket: any, _: any) => bucket;
      const mergeDeltasByType = (deltas: any) => deltas;
      const entity_num = 'Some Agency';
      const entity_id = 4321;
      const lastUpdate = <UnixTimestamp>123123123;
      const createdData = <UnixTimestamp>234234234;
      const exp_date = <UnixTimestamp>123;
      const doc = <DeltaDocument>(<unknown>{
        id: <DocumentId>123,
        programId: 'mega',
        agentName: entity_num,
        agentEntityId: entity_id,
        startDate: createdData,
        lastUpdate: lastUpdate,
        quoteExpDate: exp_date,
        quoteSetId: <DocumentId>123,
        topSavedStepId: <PositiveInteger>1,
        deltaPublishedTs: {data: <UnixTimestamp>1},
        data: {foo: ['start_bar']},
        ratedata: {},
        sessionKeys: {
          foo: ['bar', 'baz'],
        },
        meta: {
          created_by_username: 'foo@bar.com',
          created_by_roles: ['EMPLOYEE'],
          last_updated_by_username: 'bar@bar.com',
          last_updated_by_roles: ['EMPLOYEE'],
          retail_agency: 'AGT123',
          liza_access_groups: [],
          liza_access_email: [],
        },
        rdelta: {
          data: [
            {
              data: {foo: ['first_bar']},
              timestamp: <UnixTimestamp>123123,
              type: 'data',
              step_id: <PositiveInteger>1,
              concluding_save: false,
              is_indication: false,
            },
          ],
          ratedata: [],
        },
      });

      const expected_published = [
        {
          meta: {
            entity_id: 4321,
            entity_name: 'Some Agency',
            expDate: 123,
            id: 123,
            quoteSetId: 123,
            program: 'mega',
            lastUpdate: 123123123,
            startDate: 234234234,
            topSavedStepId: 1,
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
            liza_access_groups: [],
            liza_access_email: [],
            liza_timestamp_owner_vis_updated: undefined,
            bound_date_ts: undefined,
            bound_by_username: undefined,
            unlocked_by_username: undefined,
            unlocked_date_ts: undefined,
            sessionKeys: {
              foo: ['bar', 'baz'],
            },
          },
          delta: {foo: ['first_bar']},
          bucket: {foo: ['start_bar']},
          ratedata: {},
        },
      ];

      const expected_error = 'Uh oh';

      dao.getUnprocessedDocument = (): Promise<DeltaDocument> =>
        Promise.resolve(doc);

      dao.markDocumentAsProcessed = (_doc_id, _ts): Promise<void> =>
        Promise.reject(new Error(expected_error));

      dao.setErrorFlag = (): Promise<void> => {
        error_flag_set = true;
        return Promise.resolve();
      };

      publisher.publish = (meta, delta, bucket, ratedata): Promise<void> => {
        published.push({
          meta: meta,
          delta: delta.data,
          bucket: bucket,
          ratedata: ratedata,
        });

        return Promise.resolve();
      };

      // Prevent node from converting an error event into an error
      emitter.on('error', () => {});

      const sut = new Sut(
        dao,
        publisher,
        emitter,
        applyDelta,
        mergeDeltasByType
      );

      return expect(sut.process())
        .to.eventually.deep.equal(undefined)
        .then(_ => {
          expect(error_flag_set).to.be.true;
          expect(published).to.deep.equal(expected_published);
        });
    });
  });

  describe('Error handling', () => {
    it('Failure to set document error state further processing', () => {
      const published: any = [];
      let caught_error = '';
      const dao = createMockDeltaDao();
      const publisher = createMockDeltaPublisher();
      const emitter = new EventEmitter();
      const applyDelta = (bucket: any, _: any) => bucket;
      const mergeDeltasByType = (deltas: any) => deltas;
      const doc = <DeltaDocument>(<unknown>{
        id: <DocumentId>123,
        programId: 'mega',
        agentName: 'Some Agency',
        agentEntityId: 4321,
        startDate: <UnixTimestamp>234234234,
        lastUpdate: <UnixTimestamp>123123123,
        quoteExpDate: <UnixTimestamp>123123123,
        quoteSetId: <DocumentId>123,
        topSavedStepId: <PositiveInteger>1,
        deltaPublishedTs: {data: <UnixTimestamp>1},
        data: {foo: ['start_bar']},
        ratedata: {},
        sessionKeys: {
          foo: ['bar', 'baz'],
        },
        meta: {
          created_by_username: 'foo@bar.com',
          created_by_roles: ['EMPLOYEE'],
          last_updated_by_username: 'bar@bar.com',
          last_updated_by_roles: ['EMPLOYEE'],
          retail_agency: 'AGT123',
          liza_access_groups: [],
          liza_access_email: [],
        },
        rdelta: {
          data: [
            {
              data: {foo: ['first_bar']},
              timestamp: <UnixTimestamp>123123,
              type: 'data',
              step_id: <PositiveInteger>1,
              concluding_save: false,
            },
          ],
          ratedata: [],
        },
      });

      // Only one is published
      const expected_published = [
        {
          meta: {
            entity_id: 4321,
            entity_name: 'Some Agency',
            expDate: 123123123,
            id: 123,
            quoteSetId: 123,
            program: 'mega',
            lastUpdate: 123123123,
            startDate: 234234234,
            topSavedStepId: 1,
            created_by_username: 'foo@bar.com',
            created_by_roles: ['EMPLOYEE'],
            last_updated_by_username: 'bar@bar.com',
            last_updated_by_roles: ['EMPLOYEE'],
            retail_agency: 'AGT123',
            liza_access_groups: [],
            liza_access_email: [],
            liza_timestamp_owner_vis_updated: undefined,
            bound_date_ts: undefined,
            bound_by_username: undefined,
            unlocked_by_username: undefined,
            unlocked_date_ts: undefined,
            sessionKeys: {
              foo: ['bar', 'baz'],
            },
          },
          delta: {foo: ['first_bar']},
          bucket: {foo: ['start_bar']},
          ratedata: {},
        },
      ];

      const expected_error = 'Uh oh';

      dao.getUnprocessedDocument = (): Promise<DeltaDocument> =>
        Promise.resolve(doc);

      dao.markDocumentAsProcessed = (_doc_id, _ts): Promise<void> =>
        Promise.reject(new Error("Couldn't mark document"));

      dao.setErrorFlag = (): Promise<void> =>
        Promise.reject(new Error(expected_error));

      publisher.publish = (meta, delta, bucket, ratedata): Promise<void> => {
        published.push({
          meta,
          delta: delta.data,
          bucket: bucket,
          ratedata: ratedata,
        });

        return Promise.resolve();
      };

      // Prevent node from converting an error event into an error
      emitter.on('error', () => {});

      const sut = new Sut(
        dao,
        publisher,
        emitter,
        applyDelta,
        mergeDeltasByType
      );

      return expect(
        sut.process().catch(e => {
          caught_error = e.message;
        })
      )
        .to.eventually.deep.equal(undefined)
        .then(_ => {
          expect(caught_error).to.equal(expected_error);
          expect(published).to.deep.equal(expected_published);
        });
    });
  });
});

function createMockDeltaDao(): DeltaDao {
  return <DeltaDao>{
    getUnprocessedDocument() {
      return Promise.resolve(<DeltaDocument>{});
    },
    updatePublishedTracking() {
      return Promise.resolve();
    },
    markDocumentAsPublishing() {
      return Promise.resolve();
    },
    markDocumentAsDonePublishing() {
      return Promise.resolve();
    },
    markDocumentAsProcessed() {
      return Promise.resolve();
    },
    setErrorFlag() {
      return Promise.resolve();
    },
    getErrorCount() {
      return Promise.resolve(0);
    },
    getOldestUnpublishedAge() {
      return Promise.resolve(0);
    },
  };
}

function createMockDeltaPublisher(): AmqpPublisher {
  return <AmqpPublisher>{
    publish() {
      return Promise.resolve();
    },
  };
}
