/**
 * Delta data access test
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {MongoDeltaDao as Sut} from '../../../src/system/db/MongoDeltaDao';
import {Collection} from 'mongodb';
import {DaoError} from '../../../src/error/DaoError';
import {expect, use as chai_use} from 'chai';
chai_use(require('chai-as-promised'));

describe('system.db.MongoDeltaDao', () => {
  describe('#getUnprocessedDocuments', () => {
    it('finds unprocessed documents', () => {
      const stub_env = 'foo';
      let find_called = false;

      const coll = getMockCollection();

      coll.findOneAndUpdate = (selector: any) => {
        expect(selector.deltaError).to.deep.equal({$ne: true});
        expect(selector.published).to.be.false;
        expect(selector.env).to.equal(stub_env);

        find_called = true;

        return Promise.resolve({});
      };

      const sut = new Sut(coll, stub_env);
      return sut
        .getUnprocessedDocument()
        .then(() => expect(find_called).to.be.true);
    });

    it('rejects with error on query failure', () => {
      let find_called = false;

      const stub_env = 'foo';
      const expected_error_msg = 'expected error!';
      const coll = getMockCollection();

      coll.findOneAndUpdate = (_: any) => {
        find_called = true;

        return Promise.reject(new Error(expected_error_msg));
      };

      const sut = new Sut(coll, stub_env);
      return expect(sut.getUnprocessedDocument())
        .to.eventually.rejectedWith(DaoError, expected_error_msg)
        .then(() => expect(find_called).to.be.true);
    });
  });

  describe('#getOldestUnpublishedAge', () => {
    it('returns the diff of the found ts and the provided ts', () => {
      const coll = getMockCollection();

      const provided_ts = <UnixTimestamp>150;
      const found_ts = 100;
      const expected_age = 50;
      let findone_called = false;

      coll.findOne = (_: any) => {
        findone_called = true;

        return Promise.resolve({lastUpdate: found_ts});
      };

      const sut = new Sut(coll, 'foo');
      return sut.getOldestUnpublishedAge(provided_ts).then(actual_age => {
        expect(findone_called).to.be.true;
        expect(actual_age).to.equal(expected_age);
      });
    });

    it('returns zero when no unpublished documents are found', () => {
      const coll = getMockCollection();

      const provided_ts = <UnixTimestamp>150;
      const expected_age = 0;
      let findone_called = false;

      coll.findOne = (_: any) => {
        findone_called = true;

        return Promise.resolve();
      };

      const sut = new Sut(coll, 'foo');
      return sut.getOldestUnpublishedAge(provided_ts).then(actual_age => {
        expect(findone_called).to.be.true;
        expect(actual_age).to.equal(expected_age);
      });
    });
  });
});

function getMockCollection(): Collection {
  return <Collection>(<unknown>{
    findAndModify() {},
    findOne() {},
    update() {},
    createIndex() {},
    insert() {},
    find() {},
    findOneAndUpdate() {},
  });
}
