/**
 * Test case for FieldContext
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {
  FieldContext as Sut,
  ContextContent,
  NullableContextContent,
} from '../../../src/ui/context/FieldContext';
import {FieldStyler} from '../../../src/ui/context/styler/FieldStyler';
import {FieldsStore} from '../../../src/ui/context/FieldsStore';

import {expect} from 'chai';

before(function () {
  this.jsdom = require('jsdom-global')();
});

after(function () {
  this.jsdom();
});

describe('FieldContext', () => {
  it('sibling is null when label does not exist', () => {
    const element_id = 'qcontainer_checkbox_no_label';
    const group_content = getGroupContent();
    const content = group_content.querySelector('#' + element_id);
    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      element_id,
      <ContextContent>content
    );

    const expected_content =
      '<dd id="qcontainer_checkbox_no_label">' +
      '<input type="checkbox" id="q_checkbox_no_label_n_0">' +
      '</dd>';

    const given = <ContextContent>sut.getFirstOfContentSet();

    expect(given?.outerHTML as string).to.equal(expected_content);
  });

  it('gets element ID', () => {
    const element_id = 'qcontainer_checkbox_no_label';
    const group_content = getGroupContent();
    const content = group_content.querySelector('#' + element_id);
    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      element_id,
      <ContextContent>content
    );

    expect(sut.getElementId()).to.equal(element_id);
  });

  [
    {
      label: 'attaches text field to DOM with next element',
      next_element: 'div',
      element_id: 'qcontainer_baz',
      sibling_id: 'qlabel_baz',
      expected:
        '<dl>' +
        '<dt id="qlabel_baz">Baz</dt>' +
        '<dd id="qcontainer_baz">' +
        '<input type="text" id="foo_baz_0">' +
        '</dd>' +
        '<div></div>' +
        '</dl>',
    },
    {
      label: 'attaches text field to DOM w/out next element',
      next_element: null,
      element_id: 'qcontainer_baz',
      sibling_id: 'qlabel_baz',
      expected:
        '<dl>' +
        '<dt id="qlabel_baz">Baz</dt>' +
        '<dd id="qcontainer_baz">' +
        '<input type="text" id="foo_baz_0">' +
        '</dd>' +
        '</dl>',
    },
    {
      label: 'attaches checkbox field to DOM with next element',
      next_element: 'div',
      element_id: 'qcontainer_checkbox_foo',
      sibling_id: 'qlabel_checkbox_foo',
      expected:
        '<dl>' +
        '<dt id="qlabel_checkbox_foo">Foo</dt>' +
        '<dd id="qcontainer_checkbox_foo">' +
        '<input type="checkbox" id="q_checkbox_foo_n_0">' +
        '<input type="checkbox" id="q_checkbox_foo_y_0">' +
        '</dd>' +
        '<div></div>' +
        '</dl>',
    },
  ].forEach(({label, next_element, element_id, sibling_id, expected}) => {
    it(label, () => {
      const group_content = getGroupContent();
      const content = group_content.querySelector('#' + element_id);
      const sibling = group_content.querySelector('#' + sibling_id);
      const sut = new Sut(
        document,
        getStylerStub(),
        getFieldsStoreStub(),
        'foo',
        element_id,
        <ContextContent>content,
        <NullableContextContent>sibling
      );

      const to_content = document.createElement('dl');
      let dummy_next_element = null;
      if (next_element !== null) {
        dummy_next_element = document.createElement(next_element);
        to_content.appendChild(dummy_next_element);
      }

      sut.show(to_content, dummy_next_element);
      expect(to_content.outerHTML).to.equal(expected);
    });
  });

  it('detaches field and sibling from DOM', () => {
    const from_content = getGroupContent();
    let content = from_content.querySelector('#qcontainer_checkbox_foo');
    let sibling = from_content.querySelector('#qlabel_checkbox_foo');
    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'q_checkbox_foo_n_0',
      <ContextContent>content,
      <NullableContextContent>sibling
    );

    expect(from_content.contains(content)).to.be.true;
    expect(from_content.contains(sibling)).to.be.true;
    sut.hide();
    expect(from_content.contains(content)).to.be.false;
    expect(from_content.contains(sibling)).to.be.false;
  });

  it('detaches field from DOM when attached', () => {
    const from_content = document.createElement('dl');
    const child = document.createElement('div');
    from_content.appendChild(child);
    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'foo',
      <ContextContent>child
    );

    sut.hide();
    expect(from_content.contains(child)).to.be.false;
    expect(from_content.outerHTML).to.equal('<dl></dl>');
  });

  it('does not detach field from DOM after already detached', () => {
    const from_content = document.createElement('dl');
    const child = document.createElement('div');
    from_content.appendChild(child);

    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'foo',
      <ContextContent>child
    );

    // now simulate detaching
    from_content.removeChild(child);

    sut.hide();
    expect(from_content.contains(child)).to.be.false;
    expect(from_content.outerHTML).to.equal('<dl></dl>');
  });

  it('isAttached and isVisible is false when not attached to DOM', () => {
    const from_content = document.createElement('dl');
    const child = document.createElement('div');
    from_content.appendChild(child);

    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'foo',
      <ContextContent>child
    );

    // now detach
    sut.hide();

    expect(sut.isAttached()).to.be.false;

    expect(sut.isVisible()).to.be.false;
  });

  it('isAttached and isVisible is true when attached to DOM', () => {
    const from_content = document.createElement('dl');
    const child = document.createElement('div');
    from_content.appendChild(child);

    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'foo',
      <ContextContent>child
    );

    const to_content = document.createElement('dl');

    sut.show(to_content, null);

    expect(sut.isAttached()).to.be.true;

    expect(sut.isVisible()).to.be.true;
  });

  it('throws error trying to attach field to an undefined element', () => {
    const from_content = document.createElement('dl');
    const child = document.createElement('div');
    from_content.appendChild(child);

    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'foo_1',
      <ContextContent>child
    );

    const expected_err_msg = 'Error inserting field on DOM: foo, foo_1';

    // Pretend the "parent" DOM element is undefined
    const undefinedParentElement: unknown = undefined;
    expect(() =>
      sut.show(undefinedParentElement as ContextContent, null)
    ).to.throw(Error, expected_err_msg);
  });

  it('getFirstOfContentSet returns the field content', () => {
    const from_content = document.createElement('dl');
    const child = document.createElement('div');
    from_content.appendChild(child);

    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'foo',
      <ContextContent>child
    );

    expect(sut.getFirstOfContentSet()).to.equal(child);
  });

  it('getFirstOfContentSet returns the sibling if exists', () => {
    const group_content = getGroupContent();
    let content = group_content.querySelector('#qcontainer_checkbox_foo');
    let sibling = group_content.querySelector('#qlabel_checkbox_foo');

    const sut = new Sut(
      document,
      getStylerStub(),
      getFieldsStoreStub(),
      'foo',
      'q_checkbox_foo_n_0',
      <ContextContent>content,
      <NullableContextContent>sibling
    );

    expect(sut.getFirstOfContentSet()).to.equal(sibling);
  });

  [
    {
      label: 'setOptions sets new value',
      options: [
        {value: 0, label: 'Foo goes here', label_id: 'foo_label'},
        {value: 'bar', label: 'Bar goes here', label_id: 'bar_label'},
        {value: 'baz', label: 'Baz goes here', label_id: 'baz_label'},
      ],
      value: 'baz',
      expected:
        '<dd id="qcontainer_select_element">' +
        '<select id="q_select_element_0">' +
        '<option value="0">Foo goes here</option>' +
        '<option value="bar">Bar goes here</option>' +
        '<option value="baz">Baz goes here</option>' +
        '</select>' +
        '</dd>',
      expected_set_val_calls: 1,
      expected_set_label_calls: 3,
      expected_val: 'baz',
    },
    {
      label: 'setOptions sets new value when empty string',
      options: [
        {value: '', label: 'Foo goes here', label_id: 'foo_label'},
        {value: 'bar', label: 'Bar goes here', label_id: 'bar_label'},
        {value: 'baz', label: 'Baz goes here', label_id: 'baz_label'},
      ],
      value: '',
      expected:
        '<dd id="qcontainer_select_element">' +
        '<select id="q_select_element_0">' +
        '<option value="">Foo goes here</option>' +
        '<option value="bar">Bar goes here</option>' +
        '<option value="baz">Baz goes here</option>' +
        '</select>' +
        '</dd>',
      expected_set_val_calls: 1,
      expected_set_label_calls: 3,
      expected_val: '',
    },
    {
      label:
        'setOptions sets previous value when new value is not provided and old value still exists',
      options: [
        {value: 'foo', label: 'Foo goes here', label_id: 'foo_label'},
        {value: 'bar', label: 'Bar goes here', label_id: 'bar_label'},
      ],
      value: undefined,
      expected:
        '<dd id="qcontainer_select_element">' +
        '<select id="q_select_element_0">' +
        '<option value="foo">Foo goes here</option>' +
        '<option value="bar">Bar goes here</option>' +
        '</select>' +
        '</dd>',
      expected_set_val_calls: 1,
      expected_set_label_calls: 2,
      expected_val: 'bar',
    },
    {
      label:
        'setOptions sets default value when requested value is not an option',
      options: [
        {value: '', label: 'I am the default', label_id: 'foo_label'},
        {value: 'bar', label: 'Bar goes here', label_id: 'bar_label'},
      ],
      value: 'baz',
      expected:
        '<dd id="qcontainer_select_element">' +
        '<select id="q_select_element_0">' +
        '<option value="">I am the default</option>' +
        '<option value="bar">Bar goes here</option>' +
        '</select>' +
        '</dd>',
      expected_set_val_calls: 0,
      expected_set_label_calls: 2,
      expected_val: '',
    },
    {
      label:
        'setOptions sets default value when new value is not provided and old value no longer exists',
      options: [
        {value: '', label: 'I am the default', label_id: 'foo_label'},
        {value: 'baz', label: 'Baz', label_id: 'baz_label'},
      ],
      value: undefined,
      expected:
        '<dd id="qcontainer_select_element">' +
        '<select id="q_select_element_0">' +
        '<option value="">I am the default</option>' +
        '<option value="baz">Baz</option>' +
        '</select>' +
        '</dd>',
      expected_set_val_calls: 0,
      expected_set_label_calls: 2,
      expected_val: '',
    },
  ].forEach(
    ({
      label,
      options,
      value,
      expected,
      expected_set_val_calls,
      expected_set_label_calls,
      expected_val,
    }) => {
      it(label, () => {
        const group_content = getGroupContent();
        const content = group_content.querySelector(
          '#qcontainer_select_element'
        );
        const styler = getStylerStub();
        const fields_store = getFieldsStoreStub();
        const name = 'foo';

        let set_value_calls = 0;
        let given_set_value = '';

        styler.setValue = (_content: any, value: string) => {
          set_value_calls++;
          given_set_value = value;
        };

        let set_label_calls = 0;
        let given_set_label_name = '';
        fields_store.setSelectLabel = (
          name: string,
          _value: any,
          _label: any
        ) => {
          set_label_calls++;
          given_set_label_name = name;
        };

        const sut = new Sut(
          document,
          styler,
          fields_store,
          name,
          'q_select_element_0',
          <ContextContent>content
        );

        sut.setOptions(options, value);

        const final_content = sut.getFirstOfContentSet();
        const selected_option = <HTMLSelectElement>(
          final_content.querySelector('#q_select_element_0')
        );
        expect(selected_option?.value).to.equal(expected_val);
        expect(final_content.outerHTML).to.equal(expected);
        expect(set_value_calls).to.equal(expected_set_val_calls);
        expect(given_set_value).to.equal(expected_val);
        expect(set_label_calls).to.equal(expected_set_label_calls);
        expect(given_set_label_name).to.equal(name);
      });
    }
  );

  it('setOptions calls fields store with each label and value', () => {
    const group_content = getGroupContent();
    const content = group_content.querySelector('#qcontainer_select_element');
    const styler = getStylerStub();
    const fields_store = getFieldsStoreStub();
    const name = 'zoom';
    const value = 'baz';

    let expected_set_values: string[] = ['0', 'bar', 'baz'];
    let expected_set_labels: string[] = [
      'Foo goes here',
      'Bar goes here',
      'Baz goes here',
    ];

    let options = [
      {value: 0, label: 'Foo goes here', label_id: 'foo_label'},
      {value: 'bar', label: 'Bar goes here', label_id: 'bar_label'},
      {value: 'baz', label: 'Baz goes here', label_id: 'baz_label'},
    ];

    let set_label_calls = 0;
    let given_set_label_name = '';
    let given_set_values: string[] = [];
    let given_set_labels: string[] = [];

    fields_store.setSelectLabel = (
      name: string,
      value: string,
      label: string
    ) => {
      set_label_calls++;
      given_set_label_name = name;
      given_set_values.push(value);
      given_set_labels.push(label);
    };

    const sut = new Sut(
      document,
      styler,
      fields_store,
      name,
      'q_select_element_0',
      <ContextContent>content
    );

    sut.setOptions(options, value);

    expect(set_label_calls).to.equal(3);
    expect(given_set_label_name).to.equal(name);
    expect(given_set_values).to.deep.equal(expected_set_values);
    expect(given_set_labels).to.deep.equal(expected_set_labels);
  });

  it('disable calls field styler', () => {
    const group_content = getGroupContent();
    const content = group_content.querySelector('#qcontainer_select_element');
    const styler = getStylerStub();
    const fields_store = getFieldsStoreStub();
    const name = 'zoom';
    const disable = true;
    let disable_calls = 0;
    let given_disable_value = false;

    styler.disable = (_content: any, disabled: boolean) => {
      disable_calls++;
      given_disable_value = disabled;
    };

    const sut = new Sut(
      document,
      styler,
      fields_store,
      name,
      'q_select_element_0',
      <ContextContent>content
    );

    sut.disable(disable);

    expect(disable_calls).to.equal(1);
    expect(given_disable_value).to.equal(disable);
  });
});

function getStylerStub() {
  return <FieldStyler>{
    setValue: (_: any, __: any) => {},
    disable: (_: any, __: any) => {},
  };
}

function getFieldsStoreStub() {
  return <FieldsStore>{
    setSelectLabel: (_: any, __: any, ___: any) => {},
    getSelectLabel: (_: any, __: any) => {},
  };
}

function getGroupContent() {
  // Mock group content
  var group = document.createElement('dl');

  group.innerHTML =
    '<dt id="qlabel_checkbox_foo">Foo</dt>' +
    '<dd id="qcontainer_checkbox_foo">' +
    '<input type="checkbox" id="q_checkbox_foo_n_0">' +
    '<input type="checkbox" id="q_checkbox_foo_y_0">' +
    '</dd>' +
    '<dd id="qcontainer_checkbox_no_label">' +
    '<input type="checkbox" id="q_checkbox_no_label_n_0">' +
    '</dd>' +
    '<dd id="qcontainer_subfield">' +
    '<select id="q_bi_risk_type_0" class="foo widget">' +
    '<option id="q_bar_subfield_0" value="Bar">Bar</option>' +
    '<option id="q_baz_subfield_0" value="Baz">Baz</option>' +
    '<option id="q_qux_subfield_0" value="Qux">Qux</option>' +
    '</select>' +
    '</dd>' +
    '<dt id="qlabel_foo_bar_long_name">Bar</dt>' +
    '<dd id="qcontainer_foo_bar_long_name">' +
    '<input type="text" id="foo_bar_0">' +
    '</dd>' +
    '<dd id="qcontainer_subfield_single">' +
    '<select id="q_subfield_single_type_0" class="foo widget">' +
    '<option id="q_subfield_single_0" value="Foo">Foo</option>' +
    '</select>' +
    '</dd>' +
    '<dt id="qlabel_baz">Baz</dt>' +
    '<dd id="qcontainer_baz">' +
    '<input type="text" id="foo_baz_0">' +
    '</dd>' +
    '<dd id="qcontainer_select_element">' +
    '<select id="q_select_element_0">' +
    '<option value="bar" default="true">Existing option bar</option>';
  '</select>' + '</dd>';

  return group;
}
