/**
 * Test case for FieldStylerFactory
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {FieldStylerFactory as Sut} from '../../../../src/ui/context/styler/FieldStylerFactory';
import {PositiveInteger} from '../../../../src/numeric';
import {FieldsStore} from '../../../../src/ui/context/FieldsStore';
import {DefaultFieldStyler} from '../../../../src/ui/context/styler/DefaultFieldStyler';
import {CheckboxFieldStyler} from '../../../../src/ui/context/styler/CheckboxFieldStyler';
import {DatePickerFieldStyler} from '../../../../src/ui/context/styler/DatePickerFieldStyler';
import {DisplayFieldStyler} from '../../../../src/ui/context/styler/DisplayFieldStyler';
import {ElementStyler} from '../../../../src/ui/ElementStyler';
import {Renderer} from 'react-dom';

import {expect} from 'chai';

describe('FieldStylerFactory', () => {
  [
    {
      label: 'creates new DefaultFieldStyler',
      field_name: 'baz',
      type: 'text',
      expected: DefaultFieldStyler,
    },
    {
      label: 'creates new DefaultFieldStyler for static without qtype',
      field_name: 'foo-label',
      type: 'text',
      expected: DefaultFieldStyler,
    },
    {
      label: 'creates new CheckboxFieldStyler for legacyradio type',
      field_name: 'baz',
      type: 'legacyradio',
      expected: CheckboxFieldStyler,
    },
    {
      label: 'creates new CheckboxFieldStyler for radio question type',
      field_name: 'baz',
      type: 'radio',
      expected: CheckboxFieldStyler,
    },
    {
      label: 'creates new CheckboxFieldStyler for noyes question type',
      field_name: 'baz',
      type: 'noyes',
      expected: CheckboxFieldStyler,
    },
    {
      label: 'creates new CheckboxFieldStyler for checkbox question type',
      field_name: 'baz',
      type: 'checkbox',
      expected: CheckboxFieldStyler,
    },
    {
      label:
        'creates new CheckboxFieldStyler for radio with deprecated qtypes structure',
      field_name: 'baz',
      type: 'radio',
      expected: CheckboxFieldStyler,
    },
    {
      label: 'creates new DatePickerFieldStyler for datePicker question type',
      field_name: 'baz',
      type: 'datePicker',
      expected: DatePickerFieldStyler,
    },
    {
      label:
        'creates new DefaultFieldStyler for currency with deprecated qtypes structure',
      field_name: 'baz',
      type: 'currency',
      expected: DefaultFieldStyler,
    },
  ].forEach(({label, field_name, type, expected}) => {
    it(label, () => {
      const styler = <ElementStyler>{};
      const arefs = {foo: 'bar', qux: 'boo'};
      const createElement = () => {};
      const render = () => {};
      const fields_store = getFieldsStoreStub();
      let fields_store_call_count = 0;

      fields_store.getQuestionType = (name: string) => {
        fields_store_call_count++;
        expect(name).to.equal(field_name);
        return type;
      };

      const sut = new Sut(
        arefs,
        styler,
        fields_store,
        <Document>{},
        <CreateElement>(<unknown>createElement),
        <Renderer>(<unknown>render)
      );

      const given = sut.create(field_name, <PositiveInteger>0);

      expect(given).to.be.instanceOf(expected);
      expect(fields_store_call_count).to.equal(1);
    });
  });

  it('creates new DisplayFieldStyler for display/answer type', () => {
    const styler = <ElementStyler>{};
    const aref = {baz: 'bar'};
    const createElement = () => {};
    const render = () => {};
    const fields_store = getFieldsStoreStub();

    const sut = new Sut(
      aref,
      styler,
      fields_store,
      <Document>{},
      <CreateElement>(<unknown>createElement),
      <Renderer>(<unknown>render)
    );

    const given = sut.create('baz', <PositiveInteger>0);

    expect(given).to.be.instanceOf(DisplayFieldStyler);
  });
});

function getFieldsStoreStub() {
  return <FieldsStore>{
    setSelectLabel: (_: any, __: any, ___: any) => {},
    getSelectLabel: (_: any, __: any) => {},
    getQuestionType: (_: any) => {},
  };
}
