/**
 * Test case for AccordionGroupUi
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

const expect = require('chai').expect,
  sinon = require('sinon');

import {AccordionGroupUi as Sut} from '../../../src/ui/group/AccordionGroupUi';
import {
  createSut,
  createQuote,
  createGroup,
  createContent,
  createContainer,
  createContext,
  getDomElement,
} from './CommonResources';

interface AccordionContainer {
  querySelector: sinon.SinonStub;
  querySelectorAll: sinon.SinonStub;
  classList: {
    contains: sinon.SinonStub;
  };
}
interface AccordionContent {
  querySelector: sinon.SinonStub;
  querySelectorAll: sinon.SinonStub;
  getAttribute: sinon.SinonStub;
  classList: {
    contains: sinon.SinonStub;
    add: sinon.SinonStub;
    remove: sinon.SinonStub;
  };
  addEventListener: sinon.SinonStub;
  getElementsByClassName: sinon.SinonStub;
}

let content: AccordionContent;
let container: AccordionContainer;

before(function () {
  this.jsdom = require('jsdom-global')();
});

after(function () {
  this.jsdom();
});

describe('ui.group.AccordionGroupUi', () => {
  beforeEach('set up container, row, and add button', () => {
    container = createContainer();
    content = createContent();

    content.querySelector
      .withArgs('div.liza-accordion-group')
      .returns(container);

    content.classList.contains.withArgs('locked').returns(false);

    const accordionItem = {
      parentNode: {removeChild: () => {}},
    };

    const button = {
      onclick: () => {},
      style: {display: ''},
    };

    container.querySelector
      .withArgs('.liza-accordion-row-container')
      .returns(accordionItem);

    container.querySelectorAll
      .withArgs('.liza-accordion-row-container')
      .returns([accordionItem]);

    container.querySelectorAll
      .withArgs('.liza-remove-accordion')
      .returns([button]);

    container.querySelector.withArgs('.liza-add-accordion').returns(button);
  });

  describe('#showField', function () {
    [
      {
        field: 'foo_bar',
        index: 0,
      },
      {
        field: 'bar_foo',
        index: 1,
      },
    ].forEach(({field, index}) => {
      it(`it shows the header for ${field} when group fields are visible`, () => {
        const sut = createSut(Sut, {
          field: field,
          content: content,
        });

        const removeClass = sinon.stub();

        const header = {
          classList: {
            remove: removeClass,
          },
        };

        sut.getCurrentIndex = sinon.stub().returns(10);
        sut.hasVisibleField = sinon.stub().withArgs(index).returns(true);

        container.querySelectorAll
          .withArgs('.liza-accordion-row-container')
          .returns([header, header]);

        const quote = createQuote();

        content.querySelectorAll
          .withArgs('.liza-accordion-row-container')
          .returns([header, header]);

        sut.init(quote);
        sut.showField(field, index);

        expect(removeClass.calledOnce).to.be.true;
      });
    });
  });

  describe('#hideField', function () {
    [
      {
        field: 'foo_baz',
        index: 0,
      },
      {
        field: 'baz_foo',
        index: 1,
      },
    ].forEach(({field, index}) => {
      it(`it hides the header for ${field} when no group fields are visible`, () => {
        const sut = createSut(Sut, {
          field: field,
          content: content,
        });

        const addClass = sinon.stub();

        const header = {
          classList: {
            add: addClass,
          },
        };

        sut.getCurrentIndex = sinon.stub().returns(10);
        sut.hasVisibleField = sinon.stub().withArgs(index).returns(false);

        container.querySelectorAll
          .withArgs('.liza-accordion-row-container')
          .returns([header, header]);

        const quote = createQuote();

        content.querySelectorAll
          .withArgs('.liza-accordion-row-container')
          .returns([header, header]);

        sut.init(quote);
        sut.hideField(field, index);

        expect(addClass.calledOnce).to.be.true;
      });
    });
  });

  describe('removeIndex', () => {
    it('calls GroupContext and ElementStyler remove methods', () => {
      const index = 1;
      const fakeRow = <HTMLInputElement>(<unknown>getDomElement());

      container.querySelectorAll
        .withArgs('.liza-accordion-expand')
        .returns([fakeRow]);

      let styler_remove_called = false;
      let context_remove_index_is_called = false;
      let group_get_fields_is_called = false;

      const styler = {
        remove: () => {
          styler_remove_called = true;
          return;
        },
      };

      const context = createContext();

      context.removeIndex = () => {
        context_remove_index_is_called = true;
      };

      const group = createGroup('foo');

      group.getExclusiveFieldNames = () => {
        group_get_fields_is_called = true;
      };

      const sut = createSut(Sut, {
        context: context,
        content: content,
        group: group,
        styler: styler,
      });

      sut.getCurrentIndex = sinon.stub().returns(1);
      sut.hasVisibleField = sinon.stub().withArgs(index).returns(false);

      const quote = createQuote();

      sut.init(quote);
      sut.removeIndex(index);

      expect(styler_remove_called).to.be.true;
      expect(context_remove_index_is_called).to.be.true;
      expect(group_get_fields_is_called).to.be.true;
    });
  });

  describe('displayField', () => {
    it('sets input to checked of row and calls element styler', () => {
      const foundElement = getDomElement();
      const rowInput = <HTMLInputElement>(<unknown>getDomElement());
      const index = 0;

      // Set row to HTMLInputElement to satisfy if statement
      Object.setPrototypeOf(rowInput, HTMLInputElement.prototype);
      rowInput.click = () => {
        was_clicked = true;
      };

      container.querySelectorAll
        .withArgs('.liza-accordion-head')
        .returns([rowInput]);

      let get_widget_called = false;
      let was_clicked = false;

      const styler = {
        getWidgetByName: () => {
          get_widget_called = true;
          return foundElement;
        },
      };

      const context = createContext();

      const group = createGroup('foo');
      const field = 'foo_bar';

      const sut = createSut(Sut, {
        field: field,
        context: context,
        content: content,
        group: group,
        styler: styler,
      });

      sut.getCurrentIndex = sinon.stub().returns(10);

      const quote = createQuote();

      sut.init(quote);
      sut.displayField(field, index);

      expect(get_widget_called).to.be.true;
      expect(was_clicked).to.be.true;
    });
  });
});
