/**
 * Test case for LiveSearch
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {LiveSearch as Sut} from '../../../src/ui/question/LiveSearch';
import * as React from 'react';
import {expect} from 'chai';
import {InputActionTypes, ActionMeta, OptionTypeBase} from 'react-select';

const jsdom = require('jsdom');
const {JSDOM} = jsdom;

let document: Document | null;

beforeEach(() => {
  document = new JSDOM().window.document;
});

describe('ui.question.LiveSearch', () => {
  describe('#updateOptions', () => {
    [
      {
        label: 'generates a state option from an HTMLSelectOption element',
        select_options: [
          {
            value: 123,
            text: 'foo',
          },
        ],
        expected_state: {
          options: [
            {
              value: 123,
              label: 'foo',
            },
          ],
          selected_option: null,
        },
      },
      {
        label: 'sets selected options',
        select_options: [
          {
            value: 123,
            text: 'foo',
            selected: true,
          },
        ],
        expected_state: {
          options: [
            {
              value: 123,
              label: 'foo',
            },
          ],
          selected_option: {
            value: 123,
            label: 'foo',
          },
        },
      },
      {
        label: 'ignores options with a missing text or undefined value',
        select_options: [
          {
            value: 123,
            text: 'foo',
          },
          {
            value: 234,
            text: undefined,
          },
          {
            value: undefined,
            text: 'bar',
          },
          {
            value: '',
            text: 'baz',
          },
        ],
        expected_state: {
          options: [
            {
              value: 123,
              label: 'foo',
            },
            {
              value: '',
              label: 'baz',
            },
          ],
          selected_option: null,
        },
      },
    ].forEach(({label, select_options, expected_state}) => {
      it(label, () => {
        const mock_input = <HTMLInputElement>(<unknown>{});
        const mock_select = <HTMLSelectElement>(<unknown>select_options);

        mock_select.addEventListener = () => {};
        mock_select.getAttribute = () => '';

        const props = {
          input: mock_input,
          select: mock_select,
          document: document!,
          disabled: false,
        };

        const sut = new Sut(props);

        let given_state = null;

        // Override this so that we can test the state functionality without
        // needing to render this on a simulated dom
        sut.setState = (state: any) => {
          given_state = state;
        };

        sut.updateOptions();

        expect(given_state).to.deep.equal(expected_state);
      });
    });
  });

  describe('#onInputChange', () => {
    [
      {
        label: 'update input element on input-change action',
        value: 'foo',
        action: 'input-change',
        expected: 'foo',
      },
      {
        label: 'ignore actions that are not input-change',
        value: 'foo',
        action: 'something-else',
        expected: '',
      },
    ].forEach(({label, value, action, expected}) => {
      it(label, () => {
        const {input, sut} = createMocks();

        sut.onInputChange(value, {action: <InputActionTypes>action});

        expect(input.value).to.deep.equal(expected);
      });
    });
  });

  describe('#onChange', () => {
    [
      {
        label: 'update select element on select-option action',
        action: 'select-option',
        value: '123',
        expected_input: '',
        expected_select: '123',
        expected_selected: true,
      },
      {
        label: 'clear values on clear action',
        action: 'clear',
        value: '123',
        expected_input: '',
        expected_select: '',
        expected_selected: false,
      },
    ].forEach(
      ({
        label,
        action,
        value,
        expected_input,
        expected_select,
        expected_selected,
      }) => {
        it(label, () => {
          let given_state = null;

          const {select, input, sut} = createMocks();
          const option = {
            value: value,
            label: 'foo',
          };

          // Override this so that we can test the state functionality without
          // needing to render this on a simulated dom
          sut.setState = (state: any) => {
            given_state = state;
          };

          sut.onChange(option, <ActionMeta<OptionTypeBase>>{action: action});

          const expected_state = {
            selected_option: expected_selected ? option : null,
          };

          expect(given_state, 'state').to.deep.equal(expected_state);
          expect(input.value, 'input').to.equal(expected_input);
          expect(select.value, 'select').to.equal(expected_select);
        });
      }
    );
  });

  it('formats option labels as html', () => {
    const {sut} = createMocks();

    const expected = '<div>foo</div';
    const given = sut.formatOptionLabel({value: 123, label: expected});

    expect(given).to.deep.equal(
      React.createElement('div', {
        dangerouslySetInnerHTML: {
          __html: expected,
        },
      })
    );
  });

  it('filterOption never removes options', () => {
    const {sut} = createMocks();
    expect(sut.filterOption({value: 123, label: 'foo'}, 'foo')).to.be.true;
  });

  function createMocks() {
    const mock_input = <HTMLInputElement>(<unknown>{
      value: '',
      dispatchEvent: () => {},
    });
    const mock_select = <HTMLSelectElement>(<unknown>{
      value: '',
      dispatchEvent: () => {},
      addEventListener: () => {},
      getAttribute: () => {},
    });

    const props = {
      input: mock_input,
      select: mock_select,
      document: document!,
      disabled: false,
    };

    return {
      input: mock_input,
      select: mock_select,
      sut: new Sut(props),
    };
  }
});
