/**
 * Test case for GeneralStepUi
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Collection} from '../../../src/ui/step/Collection';
import {GeneralStepUi as Sut} from '../../../src/ui/step/GeneralStepUi';
import {GroupUi} from '../../../src/ui/group/GroupUi';

const expect = require('chai').expect,
  sinon = require('sinon');

let cleanup_jsdom: () => void;

before(() => {
  cleanup_jsdom = require('jsdom-global')();
});

after(() => {
  cleanup_jsdom();
});

describe('ui.GeneralStepUi', () => {
  describe('on answer data change', () => {
    [
      {
        label: 'group will style answer',
        field_refs: ['foo_d1022e47903'],
        data: {foo: ['bar']},
        expected_names: ['foo_d1022e47903'],
        expected_indexes: [0],
        expected_values: ['bar'],
        expected_calls: 1,
      },
      {
        label: 'group will style multiple answers ',
        field_refs: ['foo_d1022e47903', 'foo_d222222222'],
        data: {foo: ['bar']},
        expected_names: ['foo_d1022e47903', 'foo_d222222222'],
        expected_indexes: [0, 0],
        expected_values: ['bar', 'bar'],
        expected_calls: 2,
      },
    ].forEach(
      ({
        label,
        field_refs,
        data,
        expected_names,
        expected_indexes,
        expected_values,
        expected_calls,
      }) => {
        it(label, done => {
          const formatter = createFormatter(data);
          const group = createGroupUi();

          let group_set_value_calls = 0;
          const given_values: string[] = [];
          const given_names: string[] = [];
          const given_indexes: number[] = [];
          let answer_field_ref_calls = 0;
          let async_flag = false;

          const styler = createElementStyler(() => {});
          styler.styleAnswer = (_: any) => {
            done(new Error('ElementStyler should not be called'));
          };

          group.setValueByName = (name, index, value) => {
            group_set_value_calls++;
            given_values.push(value);
            given_names.push(name);
            given_indexes.push(index);

            if (async_flag && group_set_value_calls === expected_calls) {
              // value is set w/answer ref
              expect(given_names).to.deep.equal(expected_names);
              expect(given_indexes).to.deep.equal(expected_indexes);
              expect(given_values).to.deep.equal(expected_values);
              expect(answer_field_ref_calls).to.equal(1);
              done();
            }

            return group;
          };

          const sut = createSut(formatter, styler, {}, group);
          sut.getAnswerFieldRefs = _ => {
            answer_field_ref_calls++;
            return field_refs;
          };

          let get_group_call_count = 0;
          sut.getElementGroup = name => {
            expect(name).to.equal(field_refs[get_group_call_count]);
            get_group_call_count++;
            return group;
          };
          sut.setActive(true);
          sut.answerDataUpdate(data);
          async_flag = true;
        });
      }
    );
  });

  describe('#scrollTo', () => {
    [
      {
        field: '',
        index: 0,
        cause: 'foocause',
      },
      {
        field: undefined,
        index: 0,
        cause: 'foocause2',
      },
      {
        field: '',
        index: 0,
        cause: '',
      },
      {
        field: '',
        index: 0,
        cause: undefined,
      },
      {
        field: 'foo',
        index: -1,
        cause: undefined,
      },
      {
        field: 'foo',
        index: undefined,
        cause: undefined,
      },
      {
        field: 'foo',
        index: undefined,
        cause: 'index cause',
      },
    ].forEach((args, i) => {
      it('emits error given invalid field (' + i + ')', done => {
        const step = {
          getValidCause: () => {
            return args.cause;
          },
        };

        const sut = createSut({}, {}, step);

        // should only throw a single error
        sut.once('error', (error: any) => {
          expect(error).to.be.instanceof(Error);
          expect(error.message).to.have.string('Could not scroll');

          if (args.cause) {
            expect(error.message).to.have.string('cause: ');
            expect(error.message).to.have.string(args.cause);
          } else {
            expect(error.message).to.not.have.string('cause:');
          }

          done();
        });

        sut.scrollTo(
          <string>args.field,
          <number>args.index,
          false,
          <string>args.cause
        );
      });
    });

    describe('#monitorDom', () => {
      describe('change.program', () => {
        it('stages data', () => {
          let bind_cb = (_: any) => {};
          let staged_data_called = false;

          // return a mock input
          const styler = {
            getNameElement: () => {
              staged_data_called = true;

              return {
                attr: () => undefined,
                val: () => undefined,
              };
            },
          };

          // add listener
          const sut = createSut({}, styler);
          sut.$content = {
            bind: (_: any, cb: any) => {
              bind_cb = cb;
            },
          };

          sut.monitorDom();

          // Call the bind callback with a mock event
          bind_cb({});

          expect(staged_data_called).to.be.true;
        });

        it('does not stage data when locked', () => {
          let bind_cb = (_: any) => {};
          let staged_data_called = false;

          // return a mock input
          const styler = {
            getNameElement: () => {
              staged_data_called = true;

              return {
                attr: () => undefined,
                val: () => undefined,
              };
            },
          };

          const step = {
            getExclusiveFieldNames: () => [],
            getBucket: () => createBucket({}),
          };

          const sut = createSut({}, styler, step);
          sut.collections = [];
          sut.$content = {
            bind: (_: any, cb: any) => {
              bind_cb = cb;
            },
          };

          sut.monitorDom();
          sut.lock(true);

          // Call the bind callback with a mock event
          bind_cb({});

          expect(staged_data_called).to.be.false;
        });
      });

      describe('keyup', () => {
        it('stages data', () => {
          let data_actual: Record<string, any> = {};

          const staged_name = 'foo123[]';
          const expected_name = 'foo123';
          const expected_val = 'foo123';
          const elem = document.createElement('input');
          const event = document.createEvent('KeyboardEvent');

          elem.setAttribute('data-stage-event', 'keyup');
          document.body.appendChild(elem);
          event.initEvent('keyup', true, true);

          const styler = {
            getNameElement: () => {
              return {
                0: elem,
                attr: () => staged_name,
                val: () => expected_val,
              };
            },
          };

          const sut = createSut({}, styler);
          sut.$content = {
            bind: () => {},
          };

          sut.on('dataChange', (data: any) => {
            data_actual = data;
          });

          sut.monitorDom();

          elem.dispatchEvent(event);

          expect(data_actual[expected_name]).to.deep.equal([expected_val]);
        });

        it('checks checkboxes in a group without staging', () => {
          let data_change_emitted = false;

          const elem = document.createElement('input');
          const event = document.createEvent('KeyboardEvent');

          elem.setAttribute('type', 'checkbox');
          elem.setAttribute('data-stage-event', 'keyup');
          elem.setAttribute('data-question-length', '3');
          elem.checked = false;
          document.body.appendChild(elem);
          event.initEvent('keyup', true, true);

          const styler = {
            getNameElement: () => {
              return {
                0: elem,
                attr: () => 'foo123[]',
                val: () => 'foo123',
              };
            },
          };

          const sut = createSut({}, styler);
          sut.$content = {
            bind: () => {},
          };

          sut.on('dataChange', () => {
            data_change_emitted = true;
          });

          sut.monitorDom();

          elem.dispatchEvent(event);

          expect(elem.checked, 'checked').to.be.true;
          expect(data_change_emitted, 'emitted').to.be.false;
        });

        it('sets value to 0 on singular unchecked checkbox', () => {
          let data_actual: Record<string, any> = {};

          const staged_name = 'foo123[]';
          const expected_name = 'foo123';
          const expected_val = 0;
          const elem = document.createElement('input');
          const event = document.createEvent('KeyboardEvent');

          elem.setAttribute('type', 'checkbox');
          elem.setAttribute('data-stage-event', 'keyup');
          elem.setAttribute('data-question-length', '1');
          elem.checked = false;
          document.body.appendChild(elem);
          event.initEvent('keyup', true, true);

          const styler = {
            getNameElement: () => {
              return {
                0: elem,
                attr: () => staged_name,
                val: () => 12345,
              };
            },
          };

          const sut = createSut({}, styler);
          sut.$content = {
            bind: () => {},
          };

          sut.on('dataChange', (data: any) => {
            data_actual = data;
          });

          sut.monitorDom();

          elem.dispatchEvent(event);

          expect(data_actual[expected_name]).to.deep.equal([expected_val]);
        });
      });
    });

    it('emits error when element is not found', done => {
      const field = 'foo',
        index = 5;

      const styler = {
        getProperIndex: () => {},

        getWidgetByName: () => {
          // no element
          return [];
        },
      };

      const sut = createSut({}, styler, {});

      sut.once('error', (error: any) => {
        expect(error).to.be.instanceof(Error);
        expect(error.message).to.have.string('Could not scroll');
        expect(error.message).to.have.string('could not locate');

        expect(error.message).to.have.string(field + '[' + index + ']');

        done();
      });

      sut.scrollTo(field, index, false, '');
    });

    it('emits error when element is not visible', done => {
      const field = 'foo',
        index = 5;

      const styler = {
        getProperIndex: () => {},

        getWidgetByName: () => {
          const element = {
            is: (selector: string) => {
              expect(selector).to.equal(':visible');
              return false;
            },
          };

          return element;
        },
      };

      const sut = createSut({}, styler, {});

      sut.once('error', (error: any) => {
        expect(error).to.be.instanceof(Error);
        expect(error.message).to.have.string('Could not scroll');
        expect(error.message).to.have.string('not visible');

        expect(error.message).to.have.string(field + '[' + index + ']');

        done();
      });

      sut.scrollTo(field, index, false, '');
    });

    it('scrolls to the failed element', done => {
      const field = 'foo',
        index = 5,
        showm = true,
        message = 'whatacluster',
        element = {
          0: {
            attributes: [],
          },
          is: (selector: any) => {
            expect(selector).to.equal(':visible');
            return true;
          },
        };

      const styler = {
        getProperIndex: () => {},
        getWidgetByName: () => element,
        focus: (_: any, __: any, ___: any) => {},
      };

      const content = {
        parent: () => {
          return {
            scrollTo: scroll_mock,
          };
        },
      };

      function scroll_mock(given_element: any, duration: any, options: any) {
        expect(given_element).to.equal(element);
        expect(duration).to.equal(100);
        expect(options.offset.top).to.equal(-150);

        styler.focus = function (given_element, given_showm, given_msg) {
          expect(given_element).to.equal(element);
          expect(given_showm).to.equal(showm);
          expect(given_msg).to.equal(message);

          done();
        };

        options.onAfter();
      }

      const sut = createSut({}, styler, {});

      // XXX: SUT needs refactoring!
      sut.$content = content;

      const result = sut.scrollTo(field, index, true, message);
      expect(result).to.equal(sut);
    });
  });

  describe('#lock', () => {
    it('locks all collections', () => {
      let set_lock_calls = 0;
      const collection_0 = createCollection();
      const collection_1 = createCollection();

      const step = {
        getExclusiveFieldNames: () => {
          return [];
        },
        getBucket: () => createBucket({}),
      };

      const sut = createSut({}, {}, step);

      sut.collections = [collection_0, collection_1];

      collection_0.lock = (lock: any) => {
        set_lock_calls++;
        expect(lock).to.equal(true);
      };

      collection_1.lock = (lock: any) => {
        set_lock_calls++;
        expect(lock).to.equal(true);
      };

      sut.lock(true);

      expect(set_lock_calls).to.equal(2);
    });

    it('calls GroupUi disableField for each exclusive field when locked', () => {
      const field_names = {
        foo: true,
        qux: true,
      };
      const data = {foo: ['bar'], qux: ['1', '2']};
      const expected_names = ['foo', 'qux'];
      const expected_disabled = [true, true];
      const expected_disable_fields_calls = 2;
      const expected_group_calls = 2;

      let get_group_call_count = 0;
      let group_disable_field_calls = 0;
      const group = createGroupUi();
      const step = {
        getExclusiveFieldNames: () => {
          return field_names;
        },
        getBucket: () => createBucket(data),
      };

      const given_names: string[] = [];
      const given_disabled: boolean[] = [];

      group.disableField = (name, disable) => {
        group_disable_field_calls++;
        given_names.push(name);
        given_disabled.push(disable);

        return group;
      };

      const sut = createSut({}, {}, step, group);

      sut.getElementGroup = name => {
        expect(name).to.equal(expected_names[get_group_call_count]);
        get_group_call_count++;
        return group;
      };

      sut.collections = [];
      sut.lock(true);

      expect(get_group_call_count).to.equal(expected_group_calls);
      expect(group_disable_field_calls).to.equal(expected_disable_fields_calls);
      expect(given_names).to.deep.equal(expected_names);
      expect(given_disabled).to.deep.equal(expected_disabled);
    });
  });
});

/**
 * Create new SUT with formatter FORMATTER and generated element
 * styler
 *
 * @param {Object} formatter    validator/formatter mock
 * @param {Object} styler       mock ElementStyler
 * @param {Object} step         mock step
 * @param {Object} group        mocked Group Ui
 * @param {Object} jquery       mocked jquery object
 *
 * @return {Sut}
 */
function createSut(
  formatter: any,
  styler: any,
  step?: any,
  _group?: any,
  jquery?: any
): Sut {
  const sut = new Sut(
    step || {},
    styler,
    formatter,
    document,
    jquery || sinon.stub()
  );

  sut.getAnswerContext = _ => {
    return {};
  };
  sut.getAnswerFieldRefs = _ => {
    return [];
  };
  sut.getElementGroup = _ => <GroupUi>{};

  return sut;
}

/**
 * Create mock Collection
 *
 * @return {Object} Collection
 */
function createCollection() {
  return <Collection>{
    visit: () => {},
    lock: (_: any) => {},
  };
}

/**
 * Create mock GroupUi
 *
 * @return {Object} GroupUi
 */
function createGroupUi(): GroupUi {
  return <GroupUi>(<unknown>{
    setValueByName: (_: any, __: any, ___: any) => {},
    disableField: (_: any, __: any) => {},
  });
}

/**
 * Create mock ElementStyler
 *
 * styleAnswer method is defined by STYLE_CALLBACK
 *
 * @param {function(string,*)} style_callback styler styleAnswer method dfn
 *
 * @return {Object} ElementStyler mock
 */
function createElementStyler(style_callback: any) {
  return {
    getAnswerElementByName: () => {
      // jQuery element
      const $element = {
        0: {
          attributes: [],
        },
        length: 1,
        text: () => {},
      };

      return $element;
    },

    getProperIndex: () => {},

    getWidgetByName: () => {
      return [];
    },

    styleAnswer: style_callback,
  };
}

/**
 * Create mock validator/formatter
 *
 * The only method provided is `format', which contains no definition;
 * it is expected to be mocked by the caller.
 *
 * @param {Object}
 */
function createFormatter(return_data?: any): any {
  return {
    format: function (_: any) {
      return return_data;
    },
  };
}

/**
 * Create mock bucket
 */
const createBucket = (data: Record<string, string[]>) => {
  return {
    getDataByName: (field: string) => data[field] || [],
  };
};
