/**
 * Test case for NaFieldStyler
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework
 *
 *  Liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {DomField} from '../../../src/ui/field/DomField';
import {NaFieldStyler as Sut} from '../../../src/ui/styler/NaFieldStyler';

const expect = require('chai').expect;

describe('ui.styler.NaFieldStyler', () => {
  function testApplyHidden() {
    const element = stubEle({className: ''}),
      r1 = stubEle({className: ''}),
      r2 = stubEle({className: ''}),
      row = [r1, r2];

    new Sut().applyStyle(getStubField(element), element, row);

    [element, r1, r2].forEach(ele => {
      expect(ele.className).to.match(/\bhidden\b/);
    });
  }

  function testApplyClear() {
    const element = stubEle({style: 'foo'}),
      r1 = stubEle({style: 'foo'}),
      r2 = stubEle({style: 'foo'}),
      row = [r1, r2];

    new Sut().applyStyle(getStubField(element), element, row);

    [element, r1, r2].forEach(ele => {
      expect(ele.style).to.equal('');
    });
  }

  function testRevokeHidden() {
    const element = stubEle({className: 'foo hidden'}),
      r1 = stubEle({className: 'foo hidden'}),
      r2 = stubEle({className: 'foo hidden'}),
      row = [r1, r2];

    new Sut().revokeStyle(getStubField(element), element, row);

    [element, r1, r2].forEach(ele => {
      expect(ele.className).to.not.match(/\bhidden\b/);
      expect(ele.className).to.match(/foo/);
    });
  }

  function testRevokeStyle() {
    const element = stubEle({style: 'foo'}),
      r1 = stubEle({style: 'foo'}),
      r2 = stubEle({style: 'foo'}),
      row = [r1, r2];

    new Sut().revokeStyle(getStubField(element), element, row);

    [element, r1, r2].forEach(ele => {
      expect(ele.style).to.equal('foo');
    });
  }

  describe('#getId', () => {
    it('returns unique identifier', () => {
      expect(new Sut().getId()).to.equal('na');
    });
  });

  describe('#applyStyle', () => {
    it('sets hidden class on all elements', testApplyHidden);
    it('clears style on all elements', testApplyClear);

    it('does not set class on subfield parents', () => {
      const element = stubEle({
        className: '',
        parentElement: {
          className: 'widget',
          removeChild: () => {},
        },
      });

      const r1 = stubEle({className: ''}),
        r2 = stubEle({className: ''}),
        row = [r1, r2];

      new Sut().applyStyle(getStubField(element), element, row);

      expect(element.className).to.match(/\bhidden\b/);

      [r1, r2].forEach(ele => {
        expect(ele.className).to.equal('');
      });
    });

    it('does not clears style subfield parents', () => {
      const element = stubEle({
        style: 'foo',
        parentElement: {
          className: 'widget',
          removeChild: () => {},
        },
      });

      const r1 = stubEle({style: 'foo'}),
        r2 = stubEle({style: 'foo'}),
        row = [r1, r2];

      new Sut().applyStyle(getStubField(element), element, row);

      expect(element.style).to.equal('');

      [r1, r2].forEach(ele => {
        expect(ele.style).to.equal('foo');
      });
    });

    // f@#(& IE
    it('removes subfield from DOM', done => {
      const element = stubEle({
        style: '',
        parentElement: {
          className: 'widget',
          removeChild: (ele: HTMLElement) => {
            expect(ele).to.equal(element);
            done();
          },
        },
      });

      new Sut().applyStyle(getStubField(element), element, []);
    });
  });

  describe('#revokeStyle', () => {
    it('removes hidden class on all elements', testRevokeHidden);
    it('does not clear style on all elements', testRevokeStyle);

    it('does not remove hidden class on subfield parents', () => {
      const element = <HTMLElement>(<unknown>{
        className: 'foo hidden',
        parentElement: {
          className: 'widget',
          appendChild: () => {},
        },
      });

      const r1 = <HTMLElement>(<unknown>{className: 'foo hidden'}),
        r2 = <HTMLElement>(<unknown>{className: 'foo hidden'}),
        row = [r1, r2];

      new Sut().revokeStyle(getStubField(element), element, row);

      expect(element.className).to.not.match(/\bhidden\b/);
      expect(element.className).to.match(/foo/);

      [r1, r2].forEach(ele => {
        expect(ele.className).to.equal('foo hidden');
      });
    });

    // we eventually need to care about where it's re-attached
    it('re-attaches subfield to DOM', done => {
      const element = <HTMLElement>(<unknown>{
        className: '',
        parentElement: {
          className: 'widget',
          appendChild: (ele: HTMLElement) => {
            expect(ele).to.equal(element);
            done();
          },
        },
      });

      new Sut().revokeStyle(getStubField(element), element, []);
    });
  });

  describe('#isApplied', () => {
    it('recognizes when applied', () => {
      const element = stubEle({
        className: '',
      });

      const sut = new Sut(),
        field = getStubField(element);

      sut.applyStyle(field, element, []);

      expect(sut.isApplied(field, element)).to.be.true;

      sut.revokeStyle(field, element, []);

      expect(sut.isApplied(field, element)).to.be.false;
    });
  });

  describe('protected API', () => {
    describe('#isSubField', () => {
      it('recognizes parent widget class as subfield', () => {
        const element = <HTMLElement>(<unknown>{
          className: '',
          parentElement: {
            className: 'widget',
            removeChild: () => {},
          },
        });

        expect(protSut().protIsSubField(getStubField(element))).to.be.true;
      });

      it('missing parent widget class is non-subfield', () => {
        const element = <HTMLElement>(<unknown>{
          className: '',
        });

        expect(protSut().protIsSubField(getStubField(element))).to.be.false;
      });
    });

    describe('#hideField', () => {
      it('sets hidden class on all elements', testApplyHidden);
      it('clears style on all elements', testApplyClear);
    });

    describe('#showField', () => {
      it('removes hidden class on all elements', testRevokeHidden);
      it('does not clear style on all elements', testRevokeStyle);
    });
  });
});

function getStubField(element: HTMLElement): DomField {
  return <DomField>{
    getParent: () => {
      return element.parentElement;
    },
  };
}

function protSut() {
  return new (class extends Sut {
    protIsSubField(element: DomField) {
      return this.isSubField(element);
    }
  })();
}

function stubEle(obj: Record<string, unknown>): HTMLElement {
  obj.setAttribute = function (name: string, value: string) {
    obj[name] = value;
  };

  return <HTMLElement>(<unknown>obj);
}
