/**
 *  Copyright (C) 2010-2024 R-T Specialty, LLC.
 *
 *  This file is part of the Liza Data Collection Framework.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as
 *  published by the Free Software Foundation, either version 3 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import * as Sut from '../../src/util/common_validators';
import {expect} from 'chai';
import fc from 'fast-check';

const non_string_arb = fc
  .anything()
  .filter(r => typeof r === 'object' || r !== String(r));

// I tried to make the regex different than the SUT, but it is still
// very similar. I will add concrete tests as well to try to minimize
// the risk of reusing similar code.
const bucket_key_arb = fc
  .string({minLength: 1})
  .filter(x => !/[^A-Za-z0-9_-]/.test(x));

const bucket_value_arb = fc.array(fc.string(), {minLength: 1});

const bucket_arb = fc.dictionary(bucket_key_arb, bucket_value_arb);

describe('CommonValidators', () => {
  describe('isString', () => {
    const sut = Sut.isString;

    it('should be true for strings', () => {
      fc.assert(
        fc.property(fc.string(), str => {
          expect(sut(str)).to.be.true;
        })
      );
    });

    it('should be false for non-strings', () => {
      fc.assert(
        fc.property(non_string_arb, given => {
          expect(sut(given)).to.be.false;
        })
      );
    });
  });

  describe('isAlphaNumeric', () => {
    const sut = Sut.isAlphaNumeric;

    it('should be true for an empty string', () => {
      expect(sut('')).to.be.true;
    });

    it('should be true for a non-empty, alphanumeric string', () => {
      expect(sut('fa342oo')).to.be.true;
    });

    it('should be false for an empty string', () => {
      expect(sut(' ')).to.be.false;
    });

    it('should be false for a string with non-alphanumeric chars', () => {
      expect(sut('a$b2')).to.be.false;
    });

    it('should be true for alphanumeric strings', () => {
      fc.assert(
        fc.property(
          // I tried to make the regex different than the SUT, but it is still
          // very similar. I will add concrete tests as well to try to minimize
          // the risk of reusing similar code.
          fc.string().filter(x => !/[^A-Za-z0-9]/.test(x)),
          str => {
            expect(sut(str)).to.be.true;
          }
        )
      );
    });

    it('should be false for non-strings', () => {
      fc.assert(
        fc.property(non_string_arb, given => {
          expect(sut(given)).to.be.false;
        })
      );
    });
  });

  describe('isNonEmptyString', () => {
    const sut = Sut.isNonEmptyString;

    it('should be false for an empty string', () => {
      expect(sut('')).to.be.false;
    });

    it('should be true for nonempty strings', () => {
      fc.assert(
        fc.property(fc.string({minLength: 1}), str => {
          expect(sut(str)).to.be.true;
        })
      );
    });
  });

  describe('isBucketKey', () => {
    const sut = Sut.isBucketKey;

    it('should be false for an empty string', () => {
      expect(sut('')).to.be.false;
    });

    it('should be false for an string with spaces', () => {
      expect(sut(' df')).to.be.false;
    });

    it('should be false for an string invalid chars', () => {
      expect(sut('foo$34@#%Fd')).to.be.false;
    });

    it('should be true for an string with all different types of valid chars', () => {
      expect(sut('foo-bar_baz-12-4')).to.be.true;
    });

    it('should be true for valid strings', () => {
      fc.assert(
        fc.property(bucket_key_arb, str => {
          expect(sut(str)).to.be.true;
        })
      );
    });
  });

  describe('isNonEmptyArray', () => {
    const sut = Sut.isNonEmptyArray;

    it('should be false for an empty array', () => {
      expect(sut([])).to.be.false;
    });

    it('should be true for nonempty arrays', () => {
      fc.assert(
        fc.property(fc.array(fc.anything(), {minLength: 1}), arr => {
          expect(sut(arr)).to.be.true;
        })
      );
    });
  });

  describe('isBucketValue', () => {
    const sut = Sut.isBucketValue;

    it('should be false for an empty array', () => {
      expect(sut([])).to.be.false;
    });

    it('should be false for a nonempty array of non-strings', () => {
      fc.assert(
        fc.property(fc.array(non_string_arb, {minLength: 1}), arr => {
          expect(sut(arr)).to.be.false;
        })
      );
    });

    it('should be true for a nonempty array of strings', () => {
      fc.assert(
        fc.property(bucket_value_arb, arr => {
          expect(sut(arr)).to.be.true;
        })
      );
    });
  });

  describe('isBucket', () => {
    const sut = Sut.isBucket;

    it('should be false for null', () => {
      expect(sut(null)).to.be.false;
    });

    it('should be false for non-object', () => {
      fc.assert(
        fc.property(
          fc.anything().filter(x => typeof x !== 'object'),
          arr => {
            expect(sut(arr)).to.be.false;
          }
        )
      );
    });

    it('should be true for valid buckets', () => {
      fc.assert(
        fc.property(bucket_arb, bucket => {
          expect(sut(bucket)).to.be.true;
        })
      );
    });
  });
});
