/**
 * Test validation failure
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {Failure, Failure as Sut} from '../../src/validate/Failure';
import {Field} from '../../src/field/Field';

const expect = require('chai').expect;

const DummyField = () =>
  new (class implements Field {
    getName(): string {
      return '';
    }

    getIndex(): number {
      return 0;
    }
  })();

describe('Failure', () => {
  it('throws error when not given Field for failure', () => {
    expect(() => {
      new Sut(<Field>{});
    }).to.throw(TypeError);
  });

  it('throws error when not given a Field for causes', () => {
    expect(() => {
      // not an array
      new Sut(DummyField(), '', <Field[]>{});
    }).to.throw(TypeError);

    expect(() => {
      // one not a Field
      new Sut(DummyField(), '', [DummyField(), <Field>{}]);
    }).to.throw(TypeError);
  });

  it('does not throw error for empty clause list', () => {
    expect(() => {
      new Sut(DummyField(), '', []);
    }).to.not.throw(TypeError);
  });

  describe('#getField', () => {
    it('returns original field', () => {
      const field = DummyField();

      expect(new Sut(field).getField()).to.equal(field);
    });
  });

  describe('#getReason', () => {
    it('returns original failure reason', () => {
      const reason = 'solar flares';

      expect(new Sut(DummyField(), reason).getReason()).to.equal(reason);
    });

    it('returns empty string by default', () => {
      expect(new Sut(DummyField()).getReason()).to.equal('');
    });
  });

  describe('#getCauses', () => {
    it('returns original cause fields', () => {
      const causes = [DummyField(), DummyField()];

      expect(new Sut(DummyField(), '', causes).getCauses()).to.equal(causes);
    });

    // in other words: field caused itself to fail
    it('returns field by default', () => {
      const field = DummyField();

      expect(new Sut(field).getCauses()).to.deep.equal([field]);
    });
  });

  describe('when converted to a string', () => {
    it('produces failure reason', () => {
      const reason = 'bogons';

      expect('' + new Sut(DummyField(), reason)).to.equal(reason);
    });
  });

  describe('#merge', () => {
    it('rejects non-Failure merges', () => {
      expect(() => {
        new Sut(DummyField()).merge(<Failure>{});
      }).to.throw(TypeError);
    });

    it('merges causes', () => {
      const cause1 = DummyField(),
        cause2 = DummyField(),
        cause3 = DummyField();

      const result = new Sut(DummyField(), '', [cause1]).merge(
        new Sut(DummyField(), '', [cause2, cause3])
      );

      expect(result.getCauses()).to.deep.equal([cause1, cause2, cause3]);
    });

    it('merges reasons', () => {
      const msg1 = 'message 1',
        msg2 = 'message 2';

      const result = new Sut(DummyField(), msg1)
        .merge(new Sut(DummyField(), msg2))
        .getReason();

      expect(result).to.contain(msg1);
      expect(result).to.contain(msg2);
    });
  });
});
