/**
 * @license
 * VIN (Vehicle Identification Number) formatter test
 *
 *  Copyright (C) 2010-2022 R-T Specialty, LLC.
 *
 *  This file is part of liza.
 *
 *  liza is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import {assert, expect} from 'chai';
import {VinFormatter as Sut} from '../../../src/validate/formatter/VinFormatter';

describe('validate.formatter.VinFormatter', () => {
  describe('parse', () => {
    Object.entries({
      'no change': ['5NPDH4AE3FH588001', '5NPDH4AE3FH588001'],
      'space is trimmed': [' 5NPDH4AE3FH588001 ', '5NPDH4AE3FH588001'],
      'whitespace is removed': ['5NPDH4AE3 FH588001', '5NPDH4AE3FH588001'],
      'lowercase is capitalized': ['5npdH4AE3FH588001', '5NPDH4AE3FH588001'],
    }).forEach(([label, [given, expected]]) =>
      it(label, () => assert.equal(new Sut().parse(given), expected))
    );

    Object.entries({
      'too short': [
        '123',
        'VIN must be exactly 17 characters long. Verify that the VIN entered is correct.',
      ],
      'too long': [
        '123456789012345678',
        'VIN must be exactly 17 characters long. Verify that the VIN entered is correct.',
      ],
      'does not end in digits': [
        '5NPDH4AE3FH58800F',
        'The last 3 characters must be numbers. Verify that the VIN entered is correct.',
      ],
      'e is not valid in last 3 digit check': [
        '5NPDH4AE3FH5881E1',
        'The last 3 characters must be numbers. Verify that the VIN entered is correct.',
      ],
      'I is not a valid VIN letter': [
        'INPDH4AE3FH588001',
        'VIN should not contain I, O, or Q. Verify that the VIN entered is correct.',
      ],
      'O is not a valid VIN letter': [
        'ONPDH4AE3FH588001',
        'VIN should not contain I, O, or Q. Verify that the VIN entered is correct.',
      ],
      'Q is not a valid VIN letter': [
        'QNPDH4AE3FH588001',
        'VIN should not contain I, O, or Q. Verify that the VIN entered is correct.',
      ],
      'Multiple invalid VIN letters': [
        'IOQNH4AE3FH588001',
        'VIN should not contain I, O, or Q. Verify that the VIN entered is correct.',
      ],
      'Check digit is wrong': [
        '1J4RR4GG6BC714274',
        'Invalid VIN. Verify that the VIN entered is correct.',
      ],
      'Only alphanumeric allowed': [
        '7GZ3$##@7TCG14510',
        'VIN should only contain letters and numbers. Verify that the VIN entered is correct.',
      ],
      'Only alphanumeric allowed, case insensitive': [
        '7gZ3$##@7TcG14510',
        'VIN should only contain letters and numbers. Verify that the VIN entered is correct.',
      ],
    }).forEach(([label, [given, expected]]) =>
      it(label, () => {
        expect(() => new Sut().parse(given)).to.throw(expected);
      })
    );
  });

  describe('retrieve', () => {
    it('returns data without changing it', () => {
      assert.equal(new Sut().retrieve('foo'), 'foo');
    });
  });
});
